#include <sstream>
#include <unordered_map>
#include "barcode.h"
#include "stringutils.h"

namespace
{
	const std::unordered_map<std::string, Barcode::Symbology> g_nameToSymbologyMap = {
		{ "UPC-A", Barcode::Symbology::UPCA },
		{ "UPC-E", Barcode::Symbology::UPCE },
		{ "EAN-8", Barcode::Symbology::EAN8 },
		{ "EAN-13", Barcode::Symbology::EAN13 },
		{ "Int2of5", Barcode::Symbology::Int2of5 },
		{ "Int2of5Chk", Barcode::Symbology::Int2of5Chk },
		{ "Ind2of5", Barcode::Symbology::Ind2of5 },
		{ "Ind2of5Chk", Barcode::Symbology::Ind2of5Chk },
		{ "Mat2of5", Barcode::Symbology::Mat2of5 },
		{ "Mat2of5Chk", Barcode::Symbology::Mat2of5Chk },
		{ "Code39", Barcode::Symbology::Code39 },
		{ "Code39Chk", Barcode::Symbology::Code39Chk },
		{ "Code39Ext", Barcode::Symbology::Code39Ext },
		{ "Code39ExtChk", Barcode::Symbology::Code39ExtChk },
		{ "Code93", Barcode::Symbology::Code93 },
		{ "Code128", Barcode::Symbology::Code128 },
		{ "Code128B", Barcode::Symbology::Code128B },
		{ "UCC-128", Barcode::Symbology::UCC128 },
		{ "EAN-128", Barcode::Symbology::EAN128 },
		{ "Codabar", Barcode::Symbology::Codabar },
		{ "CodabarChk", Barcode::Symbology::CodabarChk },
		{ "MSI", Barcode::Symbology::MSI },
		{ "MSIChk10", Barcode::Symbology::MSIChk10 },
		{ "MSIChk10+10", Barcode::Symbology::MSIChk10_10 },
		{ "MSIChk11+10", Barcode::Symbology::MSIChk11_10 },
		{ "QR", Barcode::Symbology::QR }
	};

	const std::unordered_map<Barcode::Symbology, std::string> g_symbologyToNameMap = {
		{ Barcode::Symbology::UPCA, "UPC-A" },
		{ Barcode::Symbology::UPCE, "UPC-E" },
		{ Barcode::Symbology::EAN8, "EAN-8" },
		{ Barcode::Symbology::EAN13, "EAN-13" },
		{ Barcode::Symbology::Int2of5, "Int2of5" },
		{ Barcode::Symbology::Int2of5Chk, "Int2of5Chk" },
		{ Barcode::Symbology::Ind2of5, "Ind2of5" },
		{ Barcode::Symbology::Ind2of5Chk, "Ind2of5Chk" },
		{ Barcode::Symbology::Mat2of5, "Mat2of5" },
		{ Barcode::Symbology::Mat2of5Chk, "Mat2of5Chk" },
		{ Barcode::Symbology::Code39, "Code39" },
		{ Barcode::Symbology::Code39Chk, "Code39Chk" },
		{ Barcode::Symbology::Code39Ext, "Code39Ext" },
		{ Barcode::Symbology::Code39ExtChk, "Code39ExtChk" },
		{ Barcode::Symbology::Code93, "Code93" },
		{ Barcode::Symbology::Code128, "Code128" },
		{ Barcode::Symbology::Code128B, "Code128B" },
		{ Barcode::Symbology::UCC128, "UCC-128" },
		{ Barcode::Symbology::EAN128, "EAN-128" },
		{ Barcode::Symbology::Codabar, "Codabar" },
		{ Barcode::Symbology::CodabarChk, "CodabarChk" },
		{ Barcode::Symbology::MSI, "MSI" },
		{ Barcode::Symbology::MSIChk10, "MSIChk10" },
		{ Barcode::Symbology::MSIChk10_10, "MSIChk10+10" },
		{ Barcode::Symbology::MSIChk11_10, "MSIChk11+10" },
		{ Barcode::Symbology::QR, "QR" }
	};
}

Barcode::Barcode()
: m_rectangleUnit(Measurement::Unit::kUserUnits)
{
}

Barcode::Barcode(Measurement::Unit mu, const LemonPDF::Rectangle &r, Symbology s, const std::string &t)
: m_rectangleUnit(mu)
, m_rectangle(r)
, m_symbology(s)
, m_text(t)
{
}

bool Barcode::TryGetSymbologyFromName(const std::string &name, Symbology &s)
{
	auto iter = g_nameToSymbologyMap.find(name);
	if (iter != g_nameToSymbologyMap.end())
	{
		s = iter->second;
		return true;
	}

	return false;
}

const std::string &Barcode::GetSymbologyName(Symbology s)
{
	auto iter = g_symbologyToNameMap.find(s);
	if (iter != g_symbologyToNameMap.end())
		return iter->second;

	throw std::domain_error("invalid symbology value");
}

const LemonPDF::Rectangle &Barcode::getRectangle() const
{
	return m_rectangle;
}

Measurement::Unit Barcode::getRectangleUnit() const
{
	return m_rectangleUnit;
}

Barcode::Symbology Barcode::getSymbology() const
{
	return m_symbology;
}

const std::string &Barcode::getSymbologyName() const
{
	return GetSymbologyName(m_symbology);
}

const std::string &Barcode::getText() const
{
	return m_text;
}

void Barcode::setRectangle(const LemonPDF::Rectangle &r)
{
	m_rectangle = r;
}

void Barcode::setRectangleUnit(Measurement::Unit ru)
{
	m_rectangleUnit = ru;
}

void Barcode::setSymbology(Symbology s)
{
	m_symbology = s;
}

void Barcode::setText(const std::string &t)
{
	m_text = t;
}

std::istream &operator>>(std::istream &in, Barcode &b)
{
	std::string symbologyName;
	if (!std::getline(in, symbologyName, ','))
		return in;

	Barcode::Symbology symbology;
	if (!Barcode::TryGetSymbologyFromName(symbologyName, symbology))
	{
		in.setstate(std::ios::failbit);
		return in;
	}

	b.setSymbology(symbology);
	
	// get rectangle measurements
	LemonPDF::Rectangle rect;
	Measurement::Unit unit;

	if (!StringUtilities::TryParseRectangle(in, rect, unit))
	{
		in.setstate(std::ios::failbit);
		return in;
	}

	b.setRectangleUnit(unit);
	b.setRectangle(rect);

	// use up the remainder of the line as the text content
	std::string text;
	std::getline(in, text);
	b.setText(text);

	return in;
}

std::ostream &operator<<(std::ostream &out, const Barcode &b)
{
	std::string unit;
	if (b.getRectangleUnit() == Measurement::Unit::kPercent)
		unit = "%";
	const LemonPDF::Rectangle &rect = b.getRectangle();
	out << b.getSymbology() << ",";
	out << rect.lowerLeft.x << unit << ",";
	out << rect.lowerLeft.y << unit << ",";
	out << rect.getWidth() << unit << ",";
	out << rect.getHeight() << unit << ",";
	out << b.getText();
	return out;
}

std::ostream &operator<<(std::ostream &out, Barcode::Symbology s)
{
	try
	{
		out << Barcode::GetSymbologyName(s);
	}
	catch (std::domain_error &)
	{
		out << "invalid";
	}

	return out;
}
