#ifndef STRIPEDHORSE_GREENSQUAREDETECTOR_BARCODE_H
#define STRIPEDHORSE_GREENSQUAREDETECTOR_BARCODE_H

#include <lemonpdf/pdftypes.h>
#include "measurement.h"

/**
 * This class encapsulates various attributes about a barcode that is
 * intended to be inserted onto a page of a document.
 */
class Barcode
{
public:
	enum class Symbology
	{
		UPCA,
		UPCE,
		EAN8,
		EAN13,
		Int2of5,
		Int2of5Chk,
		Ind2of5,
		Ind2of5Chk,
		Mat2of5,
		Mat2of5Chk,
		Code39,
		Code39Chk,
		Code39Ext,
		Code39ExtChk,
		Code93,
		Code128,
		Code128B,
		UCC128,
		EAN128,
		Codabar,
		CodabarChk,
		MSI,
		MSIChk10,
		MSIChk10_10,
		MSIChk11_10,
		QR
	};

	/**
	 * @function Barcode
	 * Construct an unusable empty barcode instance
	 */
	Barcode();

	/**
	 * @function Barcode
	 * @param mu the units of the bounding box of this barcode
	 * @param r the bounding box of the barcode
	 * @param symbology the name of the barcode symbology as expected by the RIP, e.g. "QR", or "Codabar" etc.
	 * @oaram text the value to encode
	 */
	Barcode(Measurement::Unit mu, const LemonPDF::Rectangle &r, Symbology symbology, const std::string &text);

	/**
	 * @function TryGetSymbologyFromName
	 * @param name the name of the barcode symbology
	 * @param[out] sym the output argument that is set to the value of the Symbology enum if the name is valid
	 * @description returns "true" if the supplied name represents a known symbology, and sets the @a sym
	 * output argument. The name needs to match one of the names specified in the PDF Barcode Support document.
	 */
	static bool TryGetSymbologyFromName(const std::string &name, Symbology &sym);

	/**
	 * @function GetSymbologyName
	 * @param s the enum value representing a barcode symbology
	 * @description returns the name of the symbology as expected by the RIP in the TTECPart of the PDF page
	 * dictionary
	 */
	static const std::string &GetSymbologyName(Symbology s);

	const LemonPDF::Rectangle &getRectangle() const;
	Measurement::Unit getRectangleUnit() const;
	Symbology getSymbology() const;
	const std::string &getSymbologyName() const;
	const std::string &getText() const;
	void setRectangleUnit(Measurement::Unit mu);
	void setRectangle(const LemonPDF::Rectangle &rect);
	void setSymbology(Symbology symbology);
	void setText(const std::string &text);
	
private:
	Measurement::Unit m_rectangleUnit;
	LemonPDF::Rectangle m_rectangle;
	Symbology m_symbology;
	std::string m_text;
};

std::istream &operator>>(std::istream &in, Barcode &);
std::ostream &operator<<(std::ostream &out, const Barcode &);
std::ostream &operator<<(std::ostream &out, Barcode::Symbology);

#endif

