#include <cmath>
#include "greensquare.h"

namespace
{
	const std::unordered_map<std::string, GreenSquare::Color> s_nameToColorMap = {
		{ "red", GreenSquare::Color::kRed },
		{ "orange", GreenSquare::Color::kOrange },
		{ "yellow", GreenSquare::Color::kYellow },
		{ "chartreuse", GreenSquare::Color::kChartreuse },
		{ "green", GreenSquare::Color::kGreen },
		{ "springgreen", GreenSquare::Color::kSpringGreen },
		{ "cyan", GreenSquare::Color::kCyan },
		{ "azure", GreenSquare::Color::kAzure },
		{ "blue", GreenSquare::Color::kBlue },
		{ "violet", GreenSquare::Color::kViolet },
		{ "magenta", GreenSquare::Color::kMagenta },
		{ "rose", GreenSquare::Color::kRose }
	};
	
	const std::unordered_map<GreenSquare::Color, std::string> s_colorToNameMap = {
		{ GreenSquare::Color::kRed, "red" },
		{ GreenSquare::Color::kOrange, "orange" },
		{ GreenSquare::Color::kYellow, "yellow" },
		{ GreenSquare::Color::kChartreuse, "chartreuse" },
		{ GreenSquare::Color::kGreen, "green" },
		{ GreenSquare::Color::kSpringGreen, "springgreen" },
		{ GreenSquare::Color::kCyan, "cyan" },
		{ GreenSquare::Color::kAzure, "azure" },
		{ GreenSquare::Color::kBlue, "blue" },
		{ GreenSquare::Color::kViolet, "violet" },
		{ GreenSquare::Color::kMagenta, "magenta" },
		{ GreenSquare::Color::kRose, "rose" }
	};

	std::ostream &operator<<(std::ostream &out, const GreenSquare::Rotation &rotation)
	{
		int factor = static_cast<int>(rotation);
		if (factor < 0)
			factor = 0;
		else if (factor > 3)
			factor = 3;
		
		return out << (90 * factor) << "d";
	}
}

std::ostream &operator<<(std::ostream &out, const GreenSquare &gs)
{
	const LemonPDF::Rectangle &rect = gs.getRectangle();
	out << gs.getColor() << ",";
	out << rect.lowerLeft.x << ",";
	out << rect.lowerLeft.y << ",";
	out << rect.getWidth() << ",";
	out << rect.getHeight() << ",";
	out << gs.getRotation() << ",";
	out << gs.getText();
	return out;
}

std::ostream &operator<<(std::ostream &out, GreenSquare::Color c)
{
	auto iter = s_colorToNameMap.find(c);
	if (iter == s_colorToNameMap.end())
		out << "unknown";
	else
		out << iter->second;
	return out;
}

GreenSquare::GreenSquare(Color c, const LemonPDF::Rectangle &rect, const std::string &text, double rad)
: m_color(c)
, m_rectangle(rect)
, m_text(text)
, m_rawAngle(rad)
{
}

bool GreenSquare::TryGetColorFromName(const std::string &name, GreenSquare::Color &c)
{
	auto iter = s_nameToColorMap.find(name);
	if (iter == s_nameToColorMap.end())
		return false;

	c = iter->second;
	return true;
}

GreenSquare::Color GreenSquare::getColor() const
{
	return m_color;
}

double GreenSquare::getRawAngle() const
{
	return m_rawAngle;
}

GreenSquare::Rotation GreenSquare::getRotation() const
{
	// quantise the angle and force it to the nearest multiple of pi/4 radians
	return StripedHorse::SnapAngleToFixedRotation(m_rawAngle, M_PI_4);
}

const LemonPDF::Rectangle &GreenSquare::getRectangle() const
{
	return m_rectangle;
}

const std::string &GreenSquare::getText() const
{
	return m_text;
}

std::string &GreenSquare::getText()
{
	return m_text;
}

void GreenSquare::setColor(Color c)
{
	m_color = c;
}

void GreenSquare::setRawAngle(double rad)
{
	m_rawAngle = rad;
}

void GreenSquare::setRectangle(const LemonPDF::Rectangle &r)
{
	m_rectangle = r;
}

void GreenSquare::setRotation(Rotation r)
{
	int factor = static_cast<int>(r);

	if (factor < 0 || factor > 3)
		throw std::domain_error("Invalid rotation");
	
	m_rawAngle = factor * M_PI_2;
}

void GreenSquare::setText(const std::string &t)
{
	m_text = t;
}
