#ifndef STRIPEDHORSE_GREENSQUAREDETECTOR_GREENSQUARE_H
#define STRIPEDHORSE_GREENSQUAREDETECTOR_GREENSQUARE_H

#include <string>
#include <libcore/rotation.h>
#include <lemonpdf/pdftypes.h>

/**
 * Green Square is the unofficial name of a coloured rectangle on a page that
 * contains text, they are not necessarily green in colour, nor are they
 * necessarily squares, however, colloquially this has been the name used
 * when discussing this concept.
 * 
 * This class basically encapsulates all the relevant information about the
 * detected Green Square, such as the bounding box of the Green Square, the
 * text containined within, and the colour itself, which is quantised into
 * one of 12 known colour names.
 */
class GreenSquare
{
public:
	using Rotation = StripedHorse::Rotation;

	enum class Color
	{
		kRed,
		kOrange,
		kYellow,
		kChartreuse,
		kGreen,
		kSpringGreen,
		kCyan,
		kAzure,
		kBlue,
		kViolet,
		kMagenta,
		kRose
	};

	/**
	 * @function TryGetColorFromName
	 * @param name the name of the colour, expected to be represented in lowercase
	 * @param[out] c an output argument that is set to the enum value if the name was successfully parsed
	 * @description Use this function to convert the name of a colour into the appropriate
	 * enum value. If the name does not represent a known colour, this function will return false,
	 * otherwise, if the name was successfully parsed, it should return true.
	 */
	static bool TryGetColorFromName(const std::string &name, Color &c);
	
	/**
	 * @function GreenSquare
	 * @param color the color of the Green Square
	 * @param rect the bounding box of the Green Square, represented in user units
	 * @param text the text contained within the Green Square.
	 * @param rad the angle (in radians) of the text
	 */
	GreenSquare(Color color, const LemonPDF::Rectangle &rect, const std::string &text, double rad);

	Color getColor() const;
	double getRawAngle() const;
	Rotation getRotation() const;
	const LemonPDF::Rectangle &getRectangle() const;
	const std::string &getText() const;
	std::string &getText();
	void setColor(Color c);
	void setRawAngle(double rad);
	void setRectangle(const LemonPDF::Rectangle &rect);
	void setRotation(Rotation r);
	void setText(const std::string &text);

private:
	Color m_color;
	LemonPDF::Rectangle m_rectangle;
	std::string m_text;
	double m_rawAngle;
};

std::ostream &operator<<(std::ostream &out, const GreenSquare &gs);
std::ostream &operator<<(std::ostream &out, GreenSquare::Color c);

#endif
