#include <cmath>
#include <iostream>
#include <stdexcept>
#include <libcore/rotation.h>
#include "greensquaredetector.h"

namespace
{
	/**
	 * @function isCloseEnough
	 * @param x1 the left operand
	 * @param x2 the right operand
	 * @description determines whether two floating point values are "close enough", which
	 * can be helptful when checking whether coordinates are effectively the same but are
	 * different due to error introduced by successive multiplication of IEEE754 floats
	 */
	bool isCloseEnough(double x1, double x2)
	{
		const double kEpsilon = 0.01;
		return ::abs(x1 - x2) < kEpsilon;
	}

	/**
	 * @function isColorSuitable
	 * @param state the current PDF graphics state
	 * @param[out] c a GreenSquare::Color enum value (if the state's fill colour is suitable)
	 * @description This function determines whether the current fill colour in the graphics
	 * state is suitable for Green Squares. This means that the hue must be close to a
	 * multiple of 30 degrees, and the saturation and value must be rather high (at least 0.9).
	 */
	bool isColorSuitable(const LemonPDF::Graphics::State &state, GreenSquare::Color &c)
	{
		double hue, saturation, value;
		std::tie(hue, saturation, value) = state.fillColor->getHueSaturationValue();

		if (saturation < 0.9)
			return false;

		if (value < 0.9)
			return false;

		// add 30 so that the domain becomes 30..390 to make it easier to detect color and allow
		// wrapping at each side of the spectrum
		hue += 30.0;

		const double kTolerance = 5; // allow 5 degrees of leeway

		for (int i = 0; i < 12; i++)
		{
			const double lowerBound = (i + 1) * 30 - kTolerance;
			const double upperBound = (i + 1) * 30 + kTolerance;
			
			if (hue > lowerBound && hue < upperBound)
			{
				c = static_cast<GreenSquare::Color>(i);
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * @function isTopPathRectangularOnPage
	 * @param state the current PDF graphics state
	 * @param[out] r the detected rectangle (if the top path in the graphics state is rectangular)
	 * @description This function determines whether the top path on the graphics state is rectangular
	 * *as it appears on the page*. For example, if the path is rectangular in the current unit space,
	 * but the unit space has been rotated 45 degrees, then this function should return false. Otherwise,
	 * this will detect a rectangle if the current path has 5 nodes (newpath, three lines, and a closepath),
	 * and the coordinates of the newpath and three lines form two sets of parallel lines in page space.
	 */
	bool isTopPathRectangularOnPage(const LemonPDF::Graphics::State &state, LemonPDF::Rectangle &r)
	{
		using namespace LemonPDF;
		using namespace LemonPDF::Graphics;
		using PathNodeType = Path::PathNodeType;

		Path topPath = state.path;

		topPath.simplify(); // remove redundant nodes
		
		if (topPath.size() != 5) // move, line, line, line, close
			return false;

		PathNodeType requiredTypes[5] = {
			PathNodeType::kNewPath,
			PathNodeType::kLine,
			PathNodeType::kLine,
			PathNodeType::kLine,
			PathNodeType::kClosePath
		};

		// check that our path comprises of the right path node types
		Point p[4];
		const std::vector<Path::PathNode> &nodes = topPath.getPathNodes();
		for (size_t i = 0; i < nodes.size(); i++)
		{
			if (nodes[i].getPathNodeType() != requiredTypes[i])
				return false;

			// cache the real page points of the NewPath and Line nodes
			if (nodes[i].getPathNodeType() != PathNodeType::kClosePath)
				p[i] = state.ctm * nodes[i].getPoints().front();
		}

		// check that the coordinates are approximately square *on the page*
		if ((isCloseEnough(p[0].x, p[1].x) && isCloseEnough(p[2].x, p[3].x) && isCloseEnough(p[0].y, p[3].y) && isCloseEnough(p[1].y, p[2].y)) ||
		    (isCloseEnough(p[0].x, p[3].x) && isCloseEnough(p[1].x, p[2].x) && isCloseEnough(p[0].y, p[1].y) && isCloseEnough(p[2].y, p[3].y)))
		{
			r = Rectangle { p[0], p[2] }; // any two opposing points should work to create a rectangle
			if (r.lowerLeft.x > r.upperRight.x)
				std::swap(r.lowerLeft.x, r.upperRight.x);
			if (r.lowerLeft.y > r.upperRight.y)
				std::swap(r.lowerLeft.y, r.upperRight.y);

			return true;
		}

		return false;
	}
}

void GreenSquareDetector::checkTopPath()
{
	LemonPDF::Rectangle rect;
	GreenSquare::Color c;
	
	if (isTopPathRectangularOnPage(topState(), rect) && isColorSuitable(topState(), c))
		m_greenSquares.emplace_back(c, rect, "", 0.0);
}

void GreenSquareDetector::closePath()
{
	StandardManagedContext::closePath();
}

void GreenSquareDetector::fill(enum LemonPDF::Graphics::FillType)
{
	checkTopPath();
	topState().path.clear();
}

void GreenSquareDetector::fillAndStroke(enum LemonPDF::Graphics::FillType)
{
	checkTopPath();
	topState().path.clear();
}

const std::vector<GreenSquare> &GreenSquareDetector::getGreenSquares() const
{
	return m_greenSquares;
}

void GreenSquareDetector::intersectClippingPath(enum LemonPDF::Graphics::FillType fillType)
{
	(void) fillType;
	topState().path.clear();
}

void GreenSquareDetector::setTextFont(LemonPDF::Graphics::FontPtr font, double fontSize)
{
	(void) fontSize;
	m_font = font;
}

void GreenSquareDetector::showText(const LemonPDF::Graphics::GlyphData &glyphData)
{
	showText({{ 0.0, glyphData }});
}

void GreenSquareDetector::showText(const std::vector<LemonPDF::Graphics::SpacedText> &spacedTexts)
{
	if (m_font == nullptr)
		throw std::runtime_error("No font has been set!");

	LemonPDF::Point textPoint = topState().ctm * topState().textState.ctm * LemonPDF::Point { 0, 0 };
	for (size_t i = 0; i < m_greenSquares.size(); i++)
	{
		GreenSquare &currentSquare = m_greenSquares[i];
		if (currentSquare.getRectangle().contains(textPoint))
		{
			std::string &currentText = m_greenSquares[i].getText();
			for (const LemonPDF::Graphics::SpacedText &st : spacedTexts)
			{
				LemonPDF::Graphics::GlyphIndices indices = m_font->getGlyphIndicesForData(st.glyphData);
			    currentText.append(m_font->decodeText(indices));
			}
			
			currentSquare.setRawAngle(StripedHorse::DetectAngle(topState().textState.ctm));
		}
	}
}

void GreenSquareDetector::stroke()
{
	topState().path.clear();
}
