#include <iostream>
#include <set>
#include <sstream>
#include <stdexcept>
#include <vector>
#include <lemonpdf/pdffile.h>
#include <lemonpdf/gfxstandardmanagedcontext.h>
#include <lemonpdf/gfxparser.h>
#include "barcode.h"
#include "greensquare.h"
#include "greensquaredetector.h"
#include "templater.h"
#include "textobject.h"

using namespace LemonPDF;

namespace
{
	enum class RuntimeMode
	{
		kPerformGreenSquare,
		kDetectOnly,
		kInsertCustomBarcodesOnly
	};

	using SymbologyMap = std::unordered_map<GreenSquare::Color, Barcode::Symbology>;
	
	class GreenSquareException : public std::runtime_error
	{
	public:
		GreenSquareException(const std::string &msg) : std::runtime_error(msg) {}
	};
	
	/**
	 * @function addBaseFontToPDF
	 * @param file the file to add a font to
	 * @param baseFont the name of one of the base 14 fonts available for PDF (no validation is done)
	 * @description use this function to add a font dictionary to the PDF that utilises
	 * the base 14 fonts. This allows text to be added to pages without the need to embed
	 * an entire font
	 */
	ObjectReference addBaseFontToPDF(PDFFilePtr file, const std::string &baseFont)
	{
		DictionaryPtr fontDict = std::make_shared<Dictionary>();
		fontDict->add("BaseFont", Literal::FromName(baseFont));
		fontDict->add("Type", Literal::FromName("Font"));
		fontDict->add("Encoding", Literal::FromName("WinAnsiEncoding"));
		fontDict->add("Subtype", Literal::FromName("Type1"));
		return file->addObject(fontDict);
	}

	/**
	 * @function detectGreenSquaresOnPage
	 * @param page the page to inspect
	 * @description A simple wrapper around the GreenSquareDetector that takes care of
	 * obtaining the page content stream and feeding it into a tokerniser and parser.
	 */
	std::vector<GreenSquare> detectGreenSquaresOnPage(PagePtr page)
	{
		GreenSquareDetector detector;
		page->render(detector);
		return detector.getGreenSquares();
	}

	/**
	 * @function arrayFromPoint
	 * @param pt the point to convert
	 * @description This function converts a Point instance into a PDF array object
	 * containing two elements, the x and y values of the supplied Point instance.
	 * This is typically how an (x, y) coordinate is represented in PDF files.
	 */
	ArrayPtr arrayFromPoint(const Point &pt)
	{
		ArrayPtr result = std::make_shared<Array>();
		result->add(Literal::FromReal(pt.x));
		result->add(Literal::FromReal(pt.y));
		return result;
	}

	/**
	 * @function getBarcodeArrayForPage
	 * @param page the PDF page instance
	 * @description The RIP looks at a special entry in the page dictionary called the
	 * "TTECPart". Inside that dictionary there can be a key called "Barcode" which can
	 * either describe an array of dictionaries, or, it can just be a dictionary itself.
	 * This function will create the necessary dictionary objects and then return an
	 * empty array to be filled with barcodes (if the page had no Barcode entries to
	 * begin with) or it will return the existing Barcode array for the supplied page.
	 */
	ArrayPtr getBarcodeArrayForPage(PagePtr page)
	{
		PDFFilePtr file = page->getFile();
		
		DictionaryPtr pageDict = page->getDictionary();
		DictionaryPtr ttecPart = file->resolve<Dictionary>(pageDict->get("TTECPart"));
		if (ttecPart == nullptr)
		{
			ttecPart = std::make_shared<Dictionary>();
			pageDict->add("TTECPart", ttecPart);
		}

		ArrayPtr barcodes = file->resolve<Array>(ttecPart->get("Barcode"));
		if (barcodes == nullptr)
		{
			barcodes = std::make_shared<Array>();
			ttecPart->add("Barcode", barcodes);
		}

		return barcodes;
	}

	/**
	 * @function tryConvertGreenSquareToBarcode
	 * @param gs an instance of a GreenSquare
	 * @param symbologyMap a map that describes which colours map to which symbologies
	 * @param substitutions a dictionary of values to substitute in the GreenSquare's text
	 * @param[out] result set to a Barocde instance if the GreenSquare has a valid mapping
	 * @description This function converts a Green Square into a dedicaed Barcode instance
	 * which can subsequently be converted into a PDF dictionary using @ref
	 * generateBarcodeDictionary. If the Green Square has a colour that is not mapped in the
	 * symbology map, this function will return false.
	 */
	bool tryConvertGreenSquareToBarcode(const GreenSquare &gs, const SymbologyMap &symbologyMap, const Templater::Dictionary &substitutions, Barcode &result)
	{
		auto iter = symbologyMap.find(gs.getColor());
		if (iter == symbologyMap.end())
			return false;

		result = Barcode(Measurement::Unit::kUserUnits, gs.getRectangle(), iter->second, Templater::SubstituteText(gs.getText(), substitutions));
		return true;
	}

	/**
	 * @function tryConvertGreenSquareToTextObject
	 * @param gs an instance of a GreenSquare
	 * @param textColors a set of colours which have been selected for text conversion
	 * @param substitutions a dictionary of values to substitute in the GreenSquare's text
	 * @param[out] result the resulting TextObject (if conversion was successful)
	 * @description This function will only convert a GreenSquare to a TextObject if the
	 * supplied set of colours contains the colour of the GreenSquare.
	 */
	bool tryConvertGreenSquareToTextObject(const GreenSquare &gs, const std::set<GreenSquare::Color> &textColors, const Templater::Dictionary &substitutions, TextObject &result)
	{
		if (textColors.find(gs.getColor()) == textColors.end())
			return false;

		result = TextObject(Measurement::Unit::kUserUnits, gs.getRectangle(), Templater::SubstituteText(gs.getText(), substitutions));
		return true;
	}

	/**
	 * @function adjustRectangleForUnit
	 * @param srcRect the rectangle to adjust
	 * @param mu the measurement unit
	 * @param pageRect the reference rectangle for when the rectangle is measured as a percentage
	 * @description if the supplied measurement unit is user units, then @a srcRect is returned
	 * as-is, however, if the supplied unit is kPercent, then each dimension of the @a srcRect
	 * rectangle is divided by 100 and multiplied against corresponding dimension of the @a
	 * pageRect.
	 */
	Rectangle adjustRectangleForUnit(const Rectangle &srcRect, Measurement::Unit mu, const Rectangle &pageRect)
	{
		if (mu == Measurement::Unit::kPercent)
		{
			Rectangle rect;
			rect.lowerLeft.x = pageRect.getWidth() * srcRect.lowerLeft.x / 100.0;
			rect.lowerLeft.y = pageRect.getHeight() * srcRect.lowerLeft.y / 100.0;
			rect.upperRight.x = rect.lowerLeft.x + pageRect.upperRight.x * srcRect.upperRight.x / 100.0;
			rect.upperRight.y = rect.lowerLeft.y + pageRect.upperRight.y * srcRect.upperRight.y / 100.0;
			return rect;
		}
		else
		{
			return srcRect;
		}
	}

	/**
	 * @function generateBarcodeDictionary
	 * @param b an instance of a Barcode
	 * @param pageRect a rectangle describing the size of the page on which the barcode should be placed
	 * @description This function returns a PDF dictionary object that describes the barcode
	 * in a way that is understood by the RIP. It uses the location and value from the Barcode instance.
	 * If the Barcode rectangle is expressed as a percentage, then, the resulting barcode will be
	 * placed and sized relative to the page. 
	 */
	DictionaryPtr generateBarcodeDictionary(const Barcode &b, const LemonPDF::Rectangle &pageRect)
	{
		Rectangle rect = adjustRectangleForUnit(b.getRectangle(), b.getRectangleUnit(), pageRect);
		DictionaryPtr result = std::make_shared<Dictionary>();

		result->add("AutoScale", Literal::FromName("NonIso"));
		result->add("Location", arrayFromPoint(rect.lowerLeft));
		result->add("Area", arrayFromPoint({ rect.getWidth(), rect.getHeight() }));
		result->add("Data", Literal::FromString(b.getText()));
		result->add("Symbology", Literal::FromName(b.getSymbologyName()));

		return result;
	}

	/**
	 * @function generateContentStreamForTextObject
	 * @param t the text object to convert to a content stream
	 * @param pageRect a rectangle describing the size of the page on which the text should be placed 
	 * This function will create a page content stream that places the provided text
	 * object onto the page using the SHGS1 font (Times-Roman by default). The
	 * stream can be appended to the page's existing list of page content streams.
	 */
	StreamPtr generateContentStreamForTextObject(const TextObject &t, const LemonPDF::Rectangle &pageRect)
	{
		Rectangle r = adjustRectangleForUnit(t.getRectangle(), t.getRectangleUnit(), pageRect);

		std::stringstream streamData;
		streamData << "q\n";
		
		// set fill and stroke to white colour
		streamData << "1 g 1 G\n";

		// fill a rectangle to blank out whatever is already on the page
		// (offset by 0.5 in each direction to ensure we cover everything)
		streamData << (r.lowerLeft.x - 0.5) << " " << (r.lowerLeft.y - 0.5) << " " << (r.getWidth() + 1) << " " << (r.getHeight() + 1) << " re f n\n";

		// set the clipping region to the text object's rectangle to ensure
		// text does not overflow the designated area.
		// (disabled for now because it doesn't factor in descenders etc which may need to appear below the line)
		// streamData << r.lowerLeft.x << " " << r.lowerLeft.y << " " << r.getWidth() << " " << r.getHeight() << " re W n\n";

		// translate to the lower left corner of the text object's rectangle
		streamData << "1 0 0 1 " << r.lowerLeft.x << " " << r.lowerLeft.y << " cm\n";

		// set fill and stroke colour to black
		streamData << "0 g 0 G\n";

		// select the font and output the text
		streamData << "BT\n";
		streamData << "/SHGS1 " << r.getHeight() << " Tf\n";
		Literal::FromString(t.getText())->write(streamData, 0);
		streamData << "Tj\n";
		streamData << "ET\n";
		streamData << "Q\n";

		return Stream::FromIOStream(std::move(streamData));
	}

	/**
	 * @function addTextObjectToPage
	 * @param t the TextObject instance to add to the page, describes both the text content and location
	 * @param page the page to add the text to
	 * @param baseFontRef the font to use
	 * @description This function takes adds the text to the page, but also ensures that the page's resources
	 * include the font described by the @a baseFontRef argument.
	 */
	void addTextObjectToPage(const TextObject &t, PagePtr page, LiteralPtr baseFontRef)
	{
		PDFFilePtr file = page->getFile();
		StreamPtr contentStream = generateContentStreamForTextObject(t, page->getMediaBox());
		ObjectReference contentStreamRef = file->addObject(contentStream);
		
		// "Contents" may be an array of indirect references, or it may be just a single indirect reference
		ObjectPtr currentContentStreams = page->getDictionary()->get("Contents");
		ArrayPtr currentContentStreamsArray = std::dynamic_pointer_cast<Array>(currentContentStreams);
		if (!currentContentStreamsArray)
		{
			// It's not an array, so, we have to make it an array (which will contain the existing indirect
			// reference as well as the one to the new content stream as well).
			currentContentStreamsArray = std::make_shared<Array>();
			currentContentStreamsArray->add(currentContentStreams);
			page->getDictionary()->add("Contents", currentContentStreamsArray);
		}

		currentContentStreamsArray->add(Literal::FromObjectReference(contentStreamRef));

		// Time to update the page's resources
		DictionaryPtr resourcesDict = file->resolve<Dictionary>(page->lookup("Resources"));
		if (resourcesDict == nullptr)
		{
			// page has no resources? That's odd, but we should cater for it
			resourcesDict = std::make_shared<Dictionary>();
			page->getDictionary()->add("Resources", resourcesDict);
		}

		DictionaryPtr fontsDict = file->resolve<Dictionary>(resourcesDict->get("Font"));
		if (fontsDict == nullptr)
		{
			// page has no fonts
			fontsDict = std::make_shared<Dictionary>();
			resourcesDict->add("Font", fontsDict);
		}

		fontsDict->add("SHGS1", baseFontRef);
	}
}

int main(int argc, char *argv[])
{
	enum class OutputVersion
	{
		PDF1_3,
		PDF1_5,
		Original
	} outputVersion = OutputVersion::Original;

	long sequenceStart = 0;
	long sequenceEnd = 0;
	size_t paddingSize = 1;
	bool autoSequencing = false;
	bool detectEmpty = false;
	RuntimeMode runtimeMode = RuntimeMode::kPerformGreenSquare;
	std::vector<Barcode> barcodesForAllPages;
	std::vector<TextObject> textObjectsForAllPages;
	SymbologyMap symbologyMapping;
	std::set<GreenSquare::Color> textColours;
	Templater::Dictionary substitutions;
	std::string prefix, suffix;

	if (argc < 3)
	{
		std::cerr << "Usage: \n";
		std::cerr << argv[0] << " [ options ] pdf-in [ pdf-out ]\n";
		std::cerr << "\n";
		std::cerr <<
		"Options:\n"
		"  --detect-only\n"
		"      Only perform detection (no output file).\n"
		"\n"
		"  --detect-empty\n"
		"      Allow detection of regions with no text.\n"
		"\n"
		"  --insert-only\n"
		"      Only allow custom insertions (do not detect regions).\n"
		"\n"
		"  --insert BARCODE_TYPE,x,y,w,h,VALUE\n"
		"      Insert a barcode at a particular location on all pages (see below for\n"
		"      barcode types).\n"
		"\n"
		"  --pdf1.3        --pdf1.5\n"
		"      Force a particular PDF version (PDF 1.5 will typically be smaller).\n"
		"      The default is to match whatever the input file is.\n"
		"\n"
		"  --substitute key=value\n"
		"      In detected regions, replace occurrences of \"$key\" with \"value\".\n"
		"\n"
		"  --start ###     --end ###\n"
		"      Enable auto-sequencing of barcode content, auto-zero-padded.\n"
		"\n"
		"  --prefix XXX    --suffix XXX\n"
		"      For auto-sequencing, provides a prefix and/or suffix for the barcode\n"
		"      content.\n"
		"\n"
		"  --COLOUR BARCODE_TYPE\n"
		"      Links a barcode type to a region colour, where COLOUR must be one of the\n"
		"      following:\n"
		"        red  orange  yellow  chartreuse  green  springgreen\n"
		"        cyan  azure  blue  violet  magenta  rose\n"
		"      BARCODE_TYPE must be one of the following:\n"
		"         QR  UPC-A  UPC-E  EAN-8  EAN-13  Int2of5  Int2of5Chk\n"
		"         Ind2of5  Ind2of5Chk  Mat2of5  Mat2of5Chk  Code39\n"
		"         Code39Chk  Code39Ext  Code39ExtChk  Code93  Code128\n"
		"         Code128B  UCC-128  EAN-128  Codabar  CodabarChk  MSI\n"
		"         MSIChk10  MSIChk10+10  MSIChk11+10\n";

		return 1;
	}

	// parse the supplied arguments
	int i = 0;
	while (i++ < argc)
	{
		std::string arg = argv[i];

		if (arg == "--detect-only")
		{
			runtimeMode = RuntimeMode::kDetectOnly;
			continue;
		}
		else if (arg == "--detect-empty")
		{
			detectEmpty = true;
			continue;
		}
		else if (arg == "--insert-only")
		{
			runtimeMode = RuntimeMode::kInsertCustomBarcodesOnly;
			continue;
		}
		else if (arg == "--pdf1.3")
		{
			outputVersion = OutputVersion::PDF1_3;
		}
		else if (arg == "--pdf1.5")
		{
			outputVersion = OutputVersion::PDF1_5;
		}

		if (arg.substr(0, 2) != "--")
			break;

		if (i + 1 >= argc)
			throw GreenSquareException("Missing argument to " + arg);

		if (arg == "--substitute")
		{
			std::string keyValue = argv[i + 1];
			std::string::size_type equalsPos = keyValue.find('=');
			if (equalsPos == std::string::npos)
				throw GreenSquareException("Invalid syntax for substitution");
			
			substitutions.emplace(keyValue.substr(0, equalsPos), keyValue.substr(equalsPos + 1));
		}
		else if (arg == "--insert")
		{
			Barcode b;
			std::istringstream serialised(argv[i + 1]);
			if (serialised >> b)
			{
				barcodesForAllPages.emplace_back(b);
			}
			else
			{
				TextObject t;
				serialised.str(argv[i + 1]);
				serialised.clear();
				if (serialised >> t)
					textObjectsForAllPages.emplace_back(t);
			}

			if (serialised.fail())
				throw GreenSquareException("Cannot parse argument to --insert");
		}
		else if (arg == "--prefix")
		{
			prefix = argv[i + 1];
		}
		else if (arg == "--suffix")
		{
			prefix = argv[i + 1];
		}
		else if (arg == "--start")
		{
			std::string nextArg = argv[i + 1];
			sequenceStart = std::stol(nextArg, nullptr, 10);
			if (sequenceEnd < sequenceStart)
				sequenceEnd = sequenceStart;
			autoSequencing = true;
			
			if (nextArg.size() > paddingSize)
				paddingSize = nextArg.size();
		}
		else if (arg == "--end")
		{
			std::string nextArg = argv[i + 1];
			sequenceEnd = std::stol(nextArg, nullptr, 10);
			if (sequenceStart > sequenceEnd)
				sequenceStart = sequenceEnd;
			autoSequencing = true;
			
			if (nextArg.size() > paddingSize)
				paddingSize = nextArg.size();
		}
		else
		{
			GreenSquare::Color c;
			if (!GreenSquare::TryGetColorFromName(arg.substr(2), c))
				throw GreenSquareException("Unrecognised color: " + arg);
			
			std::string nextArg = argv[i + 1];
			if (nextArg == "Text")
			{
				textColours.insert(c);
			}
			else
			{
				Barcode::Symbology symbology;
				if (!Barcode::TryGetSymbologyFromName(nextArg, symbology))
					throw GreenSquareException("Unrecognised symbology: " + nextArg);
				symbologyMapping[c] = symbology;
			}
		}
		
		i++;
	}

	// make sure we have the required input filename (and output filename if required)
	if (runtimeMode == RuntimeMode::kDetectOnly)
	{
		// detecting barcodes doesn't require an output filename
		if (i >= argc)
			throw GreenSquareException("Missing input filename");

		if (barcodesForAllPages.size() > 0 || textObjectsForAllPages.size() > 0)
			throw GreenSquareException("Cannot specify barcodes or text to insert in 'detect' mode");
	}
	else
	{
		if (i >= argc)
			throw GreenSquareException("Missing input filename");
		else if (i >= argc - 1)
			throw GreenSquareException("Missing output filename");
	}
	
	// open up the PDF and begin processing the pages
	PDFFilePtr file = PDFFile::FromFile(argv[i]);

	// currently this is required because by default, LemonPDF will only resolve
	// objects that it needs for its own processing, but this will force it to
	// resolve all objects (the "false" means do not decompress everything).
	file->resolveEntireFile(false);

	// set the output file version to the requested version (PDF 1.5 allows
	// cross reference streams which compress better, so these PDFs are usually
	// smaller than their PDF1.3 counterparts).
	if (outputVersion == OutputVersion::PDF1_3)
		file->setVersion(1, 3);
	else if (outputVersion == OutputVersion::PDF1_5)
		file->setVersion(1, 5);

	std::vector<PagePtr> pages = file->getPages();

	// These maintain the barcodes and text objects for each page in the PDF
	// e.g. pageBarcodes[0] contains the vector of barcodes for the first page.
	std::vector<std::vector<Barcode>> pageBarcodes(pages.size());
	std::vector<std::vector<TextObject>> pageTextObjects(pages.size());

	// Detection phase
	// Here we just want to detect what barcodes or text to insert on each page.
	// Later, at the processing phase, the actual page objects are modified (and
	// possibly duplicated if autosequencing is enabled).
	for (size_t i = 0; i < pages.size(); i++)
	{
		PagePtr page = pages[i];

		std::vector<GreenSquare> greenSquaresOnPage = detectGreenSquaresOnPage(page);
		for (GreenSquare &gs : greenSquaresOnPage)
		{
			if (gs.getText().size() == 0 && !(detectEmpty || autoSequencing))
				continue;
			
			std::cout << gs << std::endl;

			if (runtimeMode == RuntimeMode::kDetectOnly)
				continue;

			Barcode b;
			if (tryConvertGreenSquareToBarcode(gs, symbologyMapping, substitutions, b))
				pageBarcodes[i].emplace_back(std::move(b));
			
			TextObject t;
			if (tryConvertGreenSquareToTextObject(gs, textColours, substitutions, t))
				pageTextObjects[i].emplace_back(std::move(t));
		}
	}

	if (runtimeMode == RuntimeMode::kDetectOnly)
	{
		// nothing to process
		return 0;
	}

	// a reference to the font used for text objects - we only want to add the
	// font to the PDF once and then re-use the same font for all pages
	LiteralPtr baseFontRef;

	// Processing phase
	// Even if autosequencing is not explicitly used, we effectively just do an
	// autosequence of 0 to 0, so that we can share the same page processing logic
	// whether autosequencing is used or not.
	for (long currentSequenceValue = sequenceStart; currentSequenceValue <= sequenceEnd; currentSequenceValue++)
	{
		for (size_t pageNumber = 0; pageNumber < pages.size(); pageNumber++)
		{
			PagePtr page = pages[pageNumber];

			if (autoSequencing && currentSequenceValue > sequenceStart)
			{
				// when autosequencing, operate on a copy of the page instead (since we have
				// already modified the first page)
				ObjectReference ref = file->appendPage(page);
				DictionaryPtr pageDict = file->resolve<Dictionary>(ref);
				page = std::make_shared<LemonPDF::Page>(file, pageDict, nullptr);

				// remove any detected barcodes from the cloned page
				DictionaryPtr ttecPart = file->resolve<Dictionary>(pageDict->get("TTECPart"));
				if (ttecPart)
					ttecPart->remove("Barcode");
			}

			std::vector<TextObject> textObjectsForPage = textObjectsForAllPages;
			textObjectsForPage.insert(textObjectsForPage.end(), pageTextObjects[pageNumber].begin(), pageTextObjects[pageNumber].end());

			for (const TextObject &t : textObjectsForPage)
			{
				if (baseFontRef == nullptr)
				{
					// only add the font dictionary if at least one text object has
					// been added to the PDF (no point adding it if there are no
					// text objects). We only need one instance to be added which
					// can be shared amongst all pages.
					ObjectReference ref = addBaseFontToPDF(file, "Times-Roman");
					baseFontRef = Literal::FromObjectReference(ref);
				}

				addTextObjectToPage(t, page, baseFontRef);
			}
		
			// prevent potentially adding an unnecessary /TTECPart to the PDF if it
			// will just be empty
			std::vector<Barcode> barcodesForPage = barcodesForAllPages;
			barcodesForPage.insert(barcodesForPage.end(), pageBarcodes[pageNumber].begin(), pageBarcodes[pageNumber].end());

			if (barcodesForPage.size() == 0)
				continue;

			ArrayPtr barcodes = getBarcodeArrayForPage(page);

			for (Barcode &b : barcodesForPage)
			{
				if (autoSequencing)
				{
					// replace the value when autosequencing
					std::ostringstream newValue;
					newValue << prefix;
					newValue.fill('0');
					newValue.width(paddingSize);
					newValue << currentSequenceValue;
					newValue << suffix;
					b.setText(newValue.str());

					size_t realPageNumber = (currentSequenceValue - sequenceStart) * pages.size() + pageNumber + 1;
					std::cout << "Autosequence for page " << realPageNumber << ": " << newValue.str() << std::endl;
				}

				barcodes->add(generateBarcodeDictionary(b, page->getMediaBox()));
			}
		}
	}

	file->write(argv[i + 1]);
}
