#include <lemonpdf/gfxtokeniser.h>
#include <lemonpdf/gfxparser.h>
#include <lemonpdf/pdfobject.h>
#include "gfxstreamreformatter.h"

StripedHorse::GFXStreamReformatter::GFXStreamReformatter(size_t indentSize, size_t initialIndent)
  : m_indentSize(indentSize), m_indentLevel(initialIndent)
{
	m_indent = std::string(m_indentLevel * m_indentSize, ' ');
	m_indent.reserve(8 * indentSize);
}

void StripedHorse::GFXStreamReformatter::decreaseIndent()
{
	if (m_indentSize == 0)
		return;
	
	if (m_indentLevel > 0)
		m_indentLevel--;
	
	m_indent.erase(m_indentLevel * m_indentSize);
}

size_t StripedHorse::GFXStreamReformatter::getIndentLevel() const
{
	return m_indentLevel;
}

void StripedHorse::GFXStreamReformatter::increaseIndent()
{
	if (m_indentSize == 0)
		return;
	
	m_indent.insert(m_indentLevel * m_indentSize, m_indentSize, ' ');
	m_indentLevel++;
}

void StripedHorse::GFXStreamReformatter::parse(std::istream &input, std::ostream &out)
{
	using namespace LemonPDF::Graphics;

	bool endOfFileEncountered = false;
	bool firstOperand = true; // used to determine when to indent and when not to
	bool firstLine = true; // used to make sure there are no trailing newlines
	GFXTokeniser tokeniser(input);
	
	while (!endOfFileEncountered)
	{
		GFXToken t = tokeniser.getNextToken();

		if (firstOperand && t.getType() != GFXToken::kOperator && t.getType() != GFXToken::kEndOfFile)
		{
			if (!firstLine)
				out << std::endl;
			out << m_indent;
			firstOperand = false;
		}

		LemonPDF::LiteralPtr literal = nullptr;

		switch (t.getType())
		{
			case GFXToken::kArrayStart:
			{
				out << "[ ";
				break;
			}

			case GFXToken::kArrayEnd:
			{
				out << "] ";
				break;
			}

			case GFXToken::kDictStart:
			{
				out << "<< ";
				break;
			}

			case GFXToken::kDictEnd:
			{
				out << ">> ";
				break;
			}
			
			case GFXToken::kName:
			{
				LemonPDF::Literal::FromName(t.getNameValue())->write(out, 0);
				out << " ";
				break;
			}

			case GFXToken::kNumber:
			{
				LemonPDF::Literal::FromReal(std::stod(t.getText()))->write(out, 0);
				out << " ";
				break;
			}

			case GFXToken::kText:
			{
				LemonPDF::Literal::FromString(t.getText())->write(out, 0);
				out << " ";
				break;
			}
			
			case GFXToken::kOperator:
			{
				firstLine = false;
				unsigned long op = HashOp(t.getText().c_str());

				switch (op)
				{
					case GFXOperator::kOpSaveState:
					case GFXOperator::kOpTextBegin:
					{
						out << std::endl << m_indent << t.getText();
						increaseIndent();
						break;
					}

					case GFXOperator::kOpRestoreState:
					case GFXOperator::kOpTextEnd:
					{
						decreaseIndent();
						out << std::endl << m_indent << t.getText();
						break;
					}

					default:
					{
						if (firstOperand)
							out << std::endl << m_indent;
						out << t.getText();
						break;
					}
				}
				
				firstOperand = true;
				continue;
			}
			
			case GFXToken::kEndOfFile:
			{
				endOfFileEncountered = true;
				break;
			}
		}
	}
}
