#include <iostream>
#include <sstream>
#include <lemonpdf/pdffile.h>
#include "gfxstreamreformatter.h"

using namespace LemonPDF;

namespace
{
	size_t reformatContentStream(StreamPtr stream, size_t initialIndent)
	{
		StripedHorse::GFXStreamReformatter reformatter(2, initialIndent);

		std::stringstream graphicsStreamIn;
		std::ostringstream graphicsStreamOut;

		stream->filter(graphicsStreamIn);
		graphicsStreamIn.seekg(0);

		reformatter.parse(graphicsStreamIn, graphicsStreamOut);

		const std::string &result = graphicsStreamOut.str();
		stream->setData(std::vector<unsigned char> { result.begin(), result.end() });

		return reformatter.getIndentLevel();
	}
}

int main(int argc, char *argv[])
{
	// sensible default values
	bool outputOldschool = true;
	bool performStreamReformatting = true;
	bool performStreamDecompression = true;
	PDFFile::WriteStyle writeStyle = PDFFile::WriteStyle::kPadded;

	int i = 1;
	for (; i < argc; i++)
	{
		std::string arg = argv[i];
		if (arg == "--")
		{
			i++;
			break;
		}
		else if (arg[0] != '-')
		{
			break;
		}

		if (arg == "--compact")
			writeStyle = PDFFile::WriteStyle::kCompact;
		else if (arg == "--no-reformat")
			performStreamReformatting = false;
		else if (arg == "--no-decompress")
			performStreamDecompression = false;
		else if (arg == "--pdf1.5")
			outputOldschool = false;
		else
		{
			std::cerr << "Unknown argument: " << arg << std::endl;
			return 1;
		}
	}

	int argsLeft = argc - i;

	if (argsLeft != 2 && argsLeft != 4)
	{
		std::cerr << "Expected 2 or 4 positinal arguments, got " << (argc - 1) << std::endl;
		std::cerr << "Usage:" << std::endl;
		std::cerr << argv[0] << " [ options ] pdf-in pdf-out" << std::endl;
		std::cerr << argv[0] << " [ options ] pdf-in object-id object-generation object-out" << std::endl;
		std::cerr << std::endl;
		std::cerr << "Options: " << std::endl;
		std::cerr << "  --compact         Do not apply whitespace padding to output" << std::endl;
		std::cerr << "  --pdf1.5          Output PDF 1.5 instead of 1.3 (uses xref stream objects)" << std::endl;
		std::cerr << "  --no-reformat     Do not reformat page content streams" << std::endl;
		std::cerr << "  --no-decompress   Do not decompress any objects" << std::endl;
		return 1;
	}
	
	PDFFilePtr file = PDFFile::FromFile(argv[i]);
	if (outputOldschool)
		file->setVersion(1, 3);
	else
		file->setVersion(1, 5);

	if (argsLeft == 2)
	{
		size_t initialIndent = 0;
		file->resolveEntireFile(performStreamDecompression);

		if (performStreamReformatting)
		{
			for (PagePtr page : file->getPages())
				for (StreamPtr contentStream : page->getContentStreams())
					initialIndent = reformatContentStream(contentStream, initialIndent);
		}

		file->write(argv[i + 1], writeStyle);
	}
	else if (argsLeft == 4)
	{
		size_t objectID = std::stoul(argv[i + 1]);
		size_t generation = std::stoul(argv[i + 2]);
		ObjectReference objReference(objectID, generation);
		ObjectPtr object = file->resolve<Object>(objReference);

		std::ofstream output(argv[i + 3]);

		// The object may not be a stream, so we attempt to cast it first
		StreamPtr stream = std::dynamic_pointer_cast<Stream>(object);
		if (stream)
		{
			if (performStreamDecompression)
			{
				stream->filter(output);
			}
			else
			{
				const std::vector<unsigned char> &data = stream->getData();
				output.write(reinterpret_cast<const char *>(&data[0]), data.size());
			}
		}
		else
		{
			object->write(output, 1);
		}
	}
}
