#include <iostream>
#include <sstream>
#include <libcore/exceptions.h>
#include <lemonpdf/pdffile.h>

namespace
{
	/// sentinel value to denote a blank page
	const size_t kBlankPage = static_cast<size_t>(-1);

	/// converts page size names to actual page size values
	std::unordered_map<std::string, LemonPDF::Point> g_pageSizes = {
		{ "13inLegal",   {  936,  612 } },
		{ "16K",         {  767,  554 } },
		{ "8.5inSquare", {  612,  612 } },
		{ "8K",          { 1107,  767 } },
		{ "A3",          { 1192,  842 } },
		{ "A4",          {  842,  595 } },
		{ "A5",          {  595,  420 } },
		{ "B4",          { 1032,  728 } },
		{ "B5",          {  728,  516 } },
		{ "Computer",    { 1008,  729 } },
		{ "Folio",       {  935,  595 } },
		{ "Ledger",      { 1224,  792 } },
		{ "Legal",       { 1008,  612 } },
		{ "Letter",      {  792,  612 } },
		{ "Statement",   {  612,  396 } }
	};

	const std::vector<std::string> g_resourceTypes = {
		"ColorSpace",
		"ExtGState",
		"Font",
		"XObject"
	};

	/// An exception that is raised whenever making a booklet encounters an error
	class BookletException : public StripedHorse::Exception
	{
	public:
		BookletException(const std::string &msg) : StripedHorse::Exception(msg) {}
	};

	/// Obtain a vector of page indices in the order in which the logical pages
	/// should be placed 2-up onto a real page (the size of the result will
	/// always be a multiple of 4, and each group of four represents the real
	/// page's front left, front right, back left and back right pages).
	std::vector<size_t> bookletPageOrderForNumberOfPages(size_t numberOfPages)
	{
		if (numberOfPages == 0)
			throw BookletException("Cannot make a booklet with 0 pages");

		// round up to the nearest multiple of 4 pages (add 3 and then truncate)
		std::vector<size_t> pageOrder((numberOfPages + 3) & ~0x3);

		// re-order the pages in groups of 2, since we are effectively doing 2-up
		for (std::vector<size_t>::size_type pageIndex = 0; pageIndex < pageOrder.size() / 2; pageIndex++)
		{
			bool isEven = (pageIndex % 2) == 0;
			pageOrder[pageIndex * 2 + (isEven ? 0 : 1)] = pageOrder.size() - pageIndex - 1;
			pageOrder[pageIndex * 2 + (isEven ? 1 : 0)] = pageIndex;
		}

		// go through and mark any non-existent pages (from rounding up) to make it clearer
		for (std::vector<size_t>::size_type pageIndex = 0; pageIndex < pageOrder.size(); pageIndex++)
			if (pageOrder[pageIndex] >= numberOfPages)
				pageOrder[pageIndex] = kBlankPage;

		return pageOrder;
	}

	/// Returns the factor required to scale the source page to half the size of the
	/// destination page (e.g. what amount to we need to scale the page by in order
	/// to fill the left or right side of the destination page).
	double getScaleFactor(const LemonPDF::Rectangle &srcBox, const LemonPDF::Rectangle &dstBox, double outerMargin)
	{
		if (srcBox.getWidth() > srcBox.getHeight())
			return (static_cast<double>(dstBox.getWidth() - 2 * outerMargin) / 2.0) / srcBox.getWidth();
		else
			return static_cast<double>(dstBox.getHeight() - 2 * outerMargin) / srcBox.getHeight();
	}

	/// A simple dictionary type that maps old resource names to new resource names
	using RenameDictionary = std::unordered_map<std::string, std::string>;
	
	/// Takes a content stream, rewriting it to the output but renaming any occurences
	/// of names provided in the rename dictionary
	void renameResourcesInContentStream(std::istream &in, std::ostream &out, const RenameDictionary &renames)
	{
		using namespace LemonPDF;
		using namespace LemonPDF::Graphics;
		using TokenType = GFXToken::TokenType;

		GFXTokeniser tokeniser(in);
		while (true)
		{
			GFXToken token = tokeniser.getNextToken();
			if (token.getType() == TokenType::kEndOfFile)
			{
				out << " ";
				break;
			}

			if (token.getType() == TokenType::kName)
			{
				const std::string &oldName = token.getText();
				auto renameIter = renames.find(oldName);
				if (renameIter != renames.end())
					out << renameIter->second;
				else
					out << oldName;
			}
			else if (token.getType() == TokenType::kText)
			{
				Literal::FromString(token.getText())->write(out, 0);
			}
			else
			{
				out << token.getText();
			}

			out << ' ';
		}
	}

	/// Looks up the most recently inherited resource dictionary in the page
	LemonPDF::DictionaryPtr getResourceDictionary(LemonPDF::PagePtr page)
	{
		LemonPDF::PDFFilePtr file = page->getFile();
		if (!file)
			return nullptr;
		
		LemonPDF::DictionaryPtr resources = file->resolve<LemonPDF::Dictionary>(page->lookup("Resources"));
		if (!resources)
			return nullptr;

		return resources;
	}

	/// Gets a particular type of resource from the resource dictionary (e.g. XObject, Font, etc)
	LemonPDF::DictionaryPtr getPageResourceTypeDictionary(LemonPDF::PagePtr page, const std::string &resourceType)
	{
		LemonPDF::DictionaryPtr resourceDict = getResourceDictionary(page);
		return page->getFile()->resolve<LemonPDF::Dictionary>(resourceDict->get(resourceType));
	}

	/// generate a dictionary containing names of resources that should be renamed
	void getResourceRenamingDictionary(size_t pageIndex, LemonPDF::PagePtr page, RenameDictionary &result)
	{
		if (!page)
			return;
		
		std::map<std::string, size_t> counterMap;

		for (const auto &resourceType : g_resourceTypes)
		{
			counterMap[resourceType] = 0;
			LemonPDF::DictionaryPtr resources = getPageResourceTypeDictionary(page, resourceType);
			if (!resources)
				continue;
			
			resources->enumerate([pageIndex,&resourceType,&result,&counterMap](const std::string &name, LemonPDF::ObjectPtr object){
				(void) object;
				std::ostringstream newName;
				newName << "/SH" << resourceType[0] << pageIndex << "_" << counterMap[resourceType]++;
				result.emplace("/" + name, newName.str());
			});
		}
	}

	/// Iterates through the various PDF page resource types in srcPage (e.g. XObject, Font etc) and
	/// appends the renamed resources to the targetResourceDict dictionary
	void renameResources(LemonPDF::DictionaryPtr targetResourceDict, LemonPDF::PagePtr srcPage, const RenameDictionary &renames)
	{
		if (!srcPage)
			return;
		
		for (const auto &resourceType : g_resourceTypes)
		{
			LemonPDF::DictionaryPtr targetPageResources = targetResourceDict->get<LemonPDF::Dictionary>(resourceType);
			LemonPDF::DictionaryPtr srcPageResources = getPageResourceTypeDictionary(srcPage, resourceType);
			
			if (!srcPageResources)
			{
				// the source page has no resources of this type, so there's nothing to
				// copy over
				continue;
			}

			if (!targetPageResources)
			{
				// the source page has resources to copy, but we don't yet have this
				// section in our target page dictionary
				targetPageResources = std::make_shared<LemonPDF::Dictionary>();
				targetResourceDict->add(resourceType, targetPageResources);
			}

			for (auto rename : renames)
			{
				std::string realName = rename.first.substr(1);
				LemonPDF::ObjectPtr o = srcPageResources->get(realName);
				if (o)
					targetPageResources->add(rename.second.substr(1), o);
			}
		}
	}

	/// traverse the page tree, calling "cb" for each leaf node, providing the object reference
	void traversePageTree(LemonPDF::PDFFilePtr file, LemonPDF::ObjectPtr root, std::function<void(LemonPDF::LiteralPtr)> cb)
	{
		LemonPDF::DictionaryPtr rootAsDict = file->resolve<LemonPDF::Dictionary>(root);
		LemonPDF::LiteralPtr nodeType = file->resolve<LemonPDF::Literal>(rootAsDict->get("Type"));
		if (nodeType->getNameValue() == "Pages")
		{
			LemonPDF::ArrayPtr kids = file->resolve<LemonPDF::Array>(rootAsDict->get("Kids"));
			kids->enumerate([&](LemonPDF::ObjectPtr object){
				LemonPDF::DictionaryPtr kid = file->resolve<LemonPDF::Dictionary>(object);
				LemonPDF::LiteralPtr kidType = file->resolve<LemonPDF::Literal>(kid->get("Type"));
				if (kidType->getNameValue() == "Pages")
				{
					traversePageTree(file, object, cb);
				}
				else
				{
					LemonPDF::LiteralPtr pageReference = std::dynamic_pointer_cast<LemonPDF::Literal>(object);
					if (pageReference)
						cb(pageReference);
				}
			});
		}
		else
		{
			cb(std::dynamic_pointer_cast<LemonPDF::Literal>(root));
		}
	}
	
	/// searches the page tree and returns the references to each page falling within
	/// the specified index range: [firstPageIndex, lastPageIndex)
	std::vector<LemonPDF::ObjectReference> getPageObjectReferences(LemonPDF::PDFFilePtr file, size_t firstPageIndex, size_t lastPageIndex)
	{
		LemonPDF::DictionaryPtr catalog = file->getDocumentCatalog();
		LemonPDF::ObjectPtr pages = catalog->get("Pages");

		size_t currentPageIndex = 0;
		std::vector<LemonPDF::ObjectReference> result;
		traversePageTree(file, pages, [&](LemonPDF::LiteralPtr leaf){
			if (currentPageIndex >= firstPageIndex && currentPageIndex < lastPageIndex)
			{
				result.push_back(leaf->getReference());
				currentPageIndex++;
			}
		});

		return result;
	}

	/// constructs a new page tree based on the provided references
	LemonPDF::DictionaryPtr constructPageTree(const std::vector<LemonPDF::ObjectReference> &pageReferences)
	{
		// current implementation just naively places all kids into a single array,
		// but it could probably be a balanced tree or so (will really only make a
		// difference for documents containing thousands of pages, but, the MFP
		// finisher can't fold this many pages anyway.
		LemonPDF::ArrayPtr kidsArray = std::make_shared<LemonPDF::Array>();
		for (const LemonPDF::ObjectReference &ref : pageReferences)
			kidsArray->add(LemonPDF::Literal::FromObjectReference(ref));

		LemonPDF::DictionaryPtr result = std::make_shared<LemonPDF::Dictionary>();
		result->add("Type", LemonPDF::Literal::FromName("Pages"));
		result->add("Count", LemonPDF::Literal::FromInteger(pageReferences.size()));
		result->add("Kids", kidsArray);
		return result;
	}

	/// writes out the transformation matrix to a page content stream
	void transformContentStream(const LemonPDF::TransformMatrix &ctm, std::ostream &out)
	{
		// we have to use Literal::write() because it will ensure that
		// the floating point value is written in a PDF-compliant way
		// (because PDF does not support scientific notation)
		for (int i = 0; i <= 2; i++)
		{
			for (int j = 0; j <= 1; j++)
			{
				LemonPDF::Literal::FromReal(ctm.m[j][i])->write(out, 0);
				out << " ";
			}
		}
		
		out << "cm\n";
	}

	/// writes a logical page to a physical page content stream based
	/// the supplied transformation, returning a dictionary that specifies how
	/// page resources were renamed (to prevent conflict)
	RenameDictionary rewriteLogicalPage(size_t pageIndex, LemonPDF::PagePtr page, LemonPDF::TransformMatrix ctm, std::ostream &out, LemonPDF::Rectangle physicalPageRect, double outerMargin)
	{
		RenameDictionary resourceRenames;
		std::stringstream pageContents;
		
		for (LemonPDF::StreamPtr stream : page->getContentStreams())
			stream->filter(pageContents);

		double scaleFactor = getScaleFactor(page->getMediaBox(), physicalPageRect, outerMargin);

		// get the resources to rename
		getResourceRenamingDictionary(pageIndex, page, resourceRenames);
		
		/* determine where to position the left page, taking into account margins etc.
		 * 1. Translate the very centre of the left side (e.g. to 1/4 of the width
		 *    of the page and 1/2 the height). If centre margin is supplied, the
		 *    translation is pushed further leftwards by half the centre margin amount
		 * 2. Scale down/up as appropriate so that the page will fit snugly
		 * 3. Translate left by half the width of the source page, and down by half the
		 *    height of the source page
		 * The origin (0, 0) will now be such that when the left side is drawn, it will
		 * appear "centred" on the left side (unless there is a non-zero centre margin,
		 * which will push it further out)
		 */

		ctm.scale(scaleFactor, scaleFactor);
		ctm.translate(-page->getMediaBox().getWidth() / 2, -page->getMediaBox().getHeight() / 2);
		
		// construct a new page content stream with a simple translation/scale applied
		out << "q\n";
		transformContentStream(ctm, out);
		renameResourcesInContentStream(pageContents, out, resourceRenames);
		out << "Q\n";

		return resourceRenames;
	}
}

#ifdef STRIPEDHORSE_TOOLKIT_STANDALONE
int main(int argc, char *argv[])
#else
#include "booklet.h"
int Booklet::main(int argc, char *argv[])
#endif
{
	using namespace LemonPDF;
	bool debugMode = false;
	std::string pageSize = "A4";
	double centreMargin = 0.0;
	double outerMargin = 0.0;
	
	if (argc <= 2)
	{
		std::cerr << "Booklet usage: " << std::endl;
		std::cerr << argv[0] << " [ options ] input-pdf output-pdf" << std::endl;
		std::cerr << std::endl;
		std::cerr << "Options:" << std::endl;
		std::cerr << "  --pagesize SIZE" << std::endl;
		std::cerr << "    Sets the page size. Supported sizes are:" << std::endl;
		for (const auto &iter : g_pageSizes)
			std::cerr << "      " << iter.first << std::endl;
		std::cerr << std::endl;
		std::cerr << "  --centremargin MARGIN" << std::endl;
		std::cerr << "    Sets the centre margin in millimetres" << std::endl;
		std::cerr << std::endl;
		std::cerr << "  --outermargin MARGIN" << std::endl;
		std::cerr << "    Sets the outer margin in millimetres" << std::endl;
		return 1;
	}

	int i = 1;

	std::function<std::string(const std::string &)> getNextArgument = [&i,argc,argv](const std::string &arg)->std::string{
		if (i + 1 >= argc)
			throw std::runtime_error("Missing required argument to " + arg);
		else
			return std::string(argv[i + 1]);
	};

	for (; i < argc; i++)
	{
		std::string arg = argv[i];
		if (arg == "--" || (arg.size() > 0 && arg[0] != '-'))
			break;
		
		
		if (arg == "--centremargin")
		{
			centreMargin = std::stod(getNextArgument("--centremargin"));
			i++;
		}
		else if (arg == "--debug")
		{
			debugMode = true;
		}
		else if (arg == "--outermargin")
		{
			outerMargin = std::stod(getNextArgument("--outermargin"));
			i++;
		}
		else if (arg == "--pagesize")
		{
			pageSize = getNextArgument("--pagesize");
			i++;
		}
		else
		{
			std::cerr << "Unknown argument: " << arg << std::endl;
			return 1;
		}
	}

	auto pageSizeIter = g_pageSizes.find(pageSize);
	if (pageSizeIter == g_pageSizes.end())
	{
		std::cerr << "Unknown page size: " << pageSize << std::endl;
		return 1;
	}

	PDFFilePtr input = PDFFile::FromFile(argv[i]);
	std::vector<PagePtr> pages = input->getPages();
	std::vector<size_t> pageOrder = bookletPageOrderForNumberOfPages(pages.size());

	for (std::vector<size_t>::size_type i = 0; debugMode && i < pageOrder.size(); i += 2)
	{
		std::cerr << "PDF page #" << (i / 2 + 1);
		std::cerr << ", printed page #" << ((i / 4) + 1);
		std::cerr << (i % 4 > 1 ? "  (back)" : " (front)");
		std::cerr << ": ";
		std::cerr << (pageOrder[i] == kBlankPage ? "*" : std::to_string(pageOrder[i]));
		std::cerr << ", ";
		std::cerr << (pageOrder[i + 1] == kBlankPage ? "*" : std::to_string(pageOrder[i + 1]));
		std::cerr << std::endl;
	}

	for (std::vector<size_t>::size_type i = 0; i < pageOrder.size(); i += 2)
	{
		// new media box size for the page
		Rectangle newBox = Rectangle({0, 0}, pageSizeIter->second);

		std::stringstream newPageStream;
		PagePtr leftSide;
		PagePtr rightSide;
		RenameDictionary leftRenames;
		RenameDictionary rightRenames;

		if (pageOrder.at(i) != kBlankPage)
		{
			TransformMatrix leftTransform;
			leftTransform.translate((newBox.getWidth() / 2 - centreMargin) / 2, newBox.getHeight() / 2);

			leftSide = pages.at(pageOrder[i]);
			leftRenames = rewriteLogicalPage(i, leftSide, leftTransform, newPageStream, newBox, outerMargin);
		}

		if (pageOrder.at(i + 1) != kBlankPage)
		{
			TransformMatrix rightTransform;
			rightTransform.translate((3 * newBox.getWidth() / 2 + centreMargin) / 2, newBox.getHeight() / 2);

			rightSide = pages.at(pageOrder[i + 1]);
			rightRenames = rewriteLogicalPage(i + 1, rightSide, rightTransform, newPageStream, newBox, outerMargin);
		}

		PagePtr targetPage = pages.at(i / 2);
		
		targetPage->setArtBox(newBox);
		targetPage->setBleedBox(newBox);
		targetPage->setCropBox(newBox);
		targetPage->setMediaBox(newBox);
		targetPage->setTrimBox(newBox);

		targetPage->getContentStreams().front()->filterIn(newPageStream);

		ArrayPtr allContentStreams = std::dynamic_pointer_cast<Array>(targetPage->getDictionary()->get("Contents"));
		if (allContentStreams)
		{
			// if we have an array, we need to remove the rest of the page content streams
			// because they have all been filtered into a single content stream (newPageStream)
			allContentStreams->truncate(1);
		}
		

		// create a blank resource dictionary so that the left and right pages can be merged
		// into it
		DictionaryPtr targetPageResources = std::make_shared<Dictionary>();
		renameResources(targetPageResources, leftSide, leftRenames);
		renameResources(targetPageResources, rightSide, rightRenames);
		
		// this will override any inherited resource dictionaries, which is kinda what we
		// want because inheritence becomes way too complicated when we begin merging
		// pages together
		targetPage->getDictionary()->add("Resources", targetPageResources);
	}

	input->resolveEntireFile(debugMode);

	size_t lastPage = pageOrder.size() / 2;
	if (lastPage < pages.size())
	{
		if (debugMode)
			std::cerr << "Removing pages: " << (lastPage + 1) << " .. " << pages.size() << std::endl;
		
		std::vector<ObjectReference> pageReferences = getPageObjectReferences(input, 0, lastPage);
		DictionaryPtr newPageTree = constructPageTree(pageReferences);

		// We have to use a reference here (PDF forbids using a direct object)
		// so the easiest thing to do is to simply replace the existing page
		// tree root with our own one
		DictionaryPtr catalog = input->getDocumentCatalog();
		ObjectReference ref = catalog->get<Literal>("Pages")->getReference();
		input->replaceObject(ref, newPageTree);

		// Remove the logical structure of the document (e.g. table-of-contents),
		// this is really only needed by a PDF viewer but since the page objects
		// won't line up any more, it may not work.
		catalog->remove("StructTreeRoot");
	}

	input->write(argv[i + 1], debugMode ? PDFFile::WriteStyle::kPadded : PDFFile::WriteStyle::kCompact);
	
	return 0;
}
