#include <string>
#include <stdexcept>
#include <unordered_set>
#include <lemonpdf/pdffile.h>

using namespace LemonPDF;

namespace
{
	/// a class to represent media type/location
	class Media
	{
	public:
		/// Valid media positions according to WinPD
		static const std::unordered_set<std::string> ValidPositions;

		/// the media type to use, e.g. "Plain", "Thick1", etc.
		std::string type = "Plain";

		/// the drawer to pull the media from, e.g. "Drawer1", "LCF", etc.		
		std::string position = "AutoSelect";
	};

	/// the definition for the set of accepted media positions
	const std::unordered_set<std::string> Media::ValidPositions {
		"Drawer1",
		"Drawer2",
		"Drawer3",
		"Drawer4",
		"LCF",
		"ExternalLCF",
		"Bypass"
	};

	class CoverPage : public Media
	{
	public:
		enum class Style
		{
			/// Do not insert a cover page
			None,

			/// Print on both sides of the page (duplex only)
			PrintOnPage,

			/// Print on only one side of the page (duplex or simplex)
			PrintOnOneSide,

			/// Do not print on the page at all (regardless of duplex setting)
			DoNotPrintOnPage
		} style = Style::None;

		/// Returns a CoverPage::Style based on the input string
		static Style ParseStyle(const std::string &styleName)
		{
			if (styleName == "none")
				return Style::None;
			else if (styleName == "print")
				return Style::PrintOnPage;
			else if (styleName == "print1")
				return Style::PrintOnOneSide;
			else if (styleName == "noprint")
				return Style::DoNotPrintOnPage;
			else
				throw std::runtime_error("Unknown cover page style: " + styleName);
		}
	};

	class CoverPageDirective
	{
	public:
		enum class Action
		{
			// Update existing page
			Update,

			// Inject a new page
			Inject
		};
		
		CoverPageDirective(size_t pageIndex, Action action, DictionaryPtr dict)
		: pageIndex(pageIndex)
		, action(action)
		, privateDict(dict)
		{
		}
	
		size_t pageIndex;
		Action action;
		DictionaryPtr privateDict;
	};

	void printUsage(const std::string &progName)
	{
		std::cerr << "Usage: " << std::endl;
		std::cerr << progName << " [ coverpage-options ] input-file output-file" << std::endl;
		std::cerr << std::endl;
		std::cerr << "Options:" << std::endl;
		std::cerr << "  --back-style STYLE" << std::endl;
		std::cerr << "  --front-style STYLE" << std::endl;
		std::cerr << "    Sets the cover page print style for the back or front cover," << std::endl;
		std::cerr << "    where STYLE can be 'none' (default), 'noprint' (do not print" << std::endl;
		std::cerr << "    on page), 'print' (print on both sides), and 'print1' (print" << std::endl;
		std::cerr << "    on one side)." << std::endl;
		std::cerr << std::endl;
		std::cerr << "  --back-media MEDIA" << std::endl;
		std::cerr << "  --front-media MEDIA" << std::endl;
		std::cerr << "    Sets the cover page media type for the back or front cover," << std::endl;
		std::cerr << "    where MEDIA can be any supported media type, such as 'Plain'," << std::endl;
		std::cerr << "    'Thick1', 'Transparency', etc. (no validation is done on this" << std::endl;
		std::cerr << "    value)." << std::endl;
		std::cerr << std::endl;
		std::cerr << "  --back-source SRC" << std::endl;
		std::cerr << "  --front-source SRC" << std::endl;
		std::cerr << "    Specifies which input slot the back or front cover should" << std::endl;
		std::cerr << "    print from. Valid choices are 'Drawer1', 'Drawer2', 'Drawer3'" << std::endl;
		std::cerr << "    'Drawer4', 'LCF', 'ExternalLCF', and 'Bypass'. Alternatively," << std::endl;
		std::cerr << "    if unspecified, 'AutoSelect' will be used." << std::endl;
		std::cerr << std::endl;
		std::cerr << "  --source SRC" << std::endl;
		std::cerr << "  --media MEDIA" << std::endl;
		std::cerr << "    Specifies the input slot and media type for the document" << std::endl;
		std::cerr << "    itself. The default SRC is 'AutoSelect' and the default MEDIA" << std::endl;
		std::cerr << "    is 'Plain'." << std::endl;
		std::cerr << std::endl;
		std::cerr << "  --duplex MODE" << std::endl;
		std::cerr << "    Specifies whether duplex printing was chosen by the user," << std::endl;
		std::cerr << "    this affects the behaviour of the print style. MODE should" << std::endl;
		std::cerr << "    be either 'true' or 'false'." << std::endl;
	}

	class ArgumentException : public std::runtime_error
	{
	public:
		ArgumentException(const std::string &msg) : std::runtime_error(msg) {}
	};

	/// creates a TTECPart dictionary for the provided cover page
	DictionaryPtr createPrivateDict(const Media &media, bool enableDuplex)
	{
		DictionaryPtr d = std::make_shared<Dictionary>();
		d->add("Type", Literal::FromName("TTECPart"));
		d->add("TTEC_PageDuplex", Literal::FromBoolean(enableDuplex));
		d->add("TTEC_MediaPosition", Literal::FromString(media.position));
		d->add("TTEC_MediaType", Literal::FromString(media.type));
		d->add("TTEC_PagePDFBinding", Literal::FromInteger(1));
		return d;
	}

	std::vector<CoverPageDirective> getDirectives(PDFFilePtr pdf, const Media &document, const CoverPage &front, const CoverPage &back, bool isDuplex)
	{
		using Action = CoverPageDirective::Action;
		using Style = CoverPage::Style;
		
		size_t pageCount = pdf->getPages().size();
		size_t pagesConsumed = 0;
		std::vector<CoverPageDirective> directives;

		switch (front.style)
		{
			case Style::None:
				// do nothing
				break;
			
			case Style::PrintOnPage:
				if (isDuplex)
				{
					// update the media for the first two pages, then reset it back
					directives.emplace_back(0, Action::Update, createPrivateDict(front, true));
					if (pageCount > 2)
					{
						directives.emplace_back(2, Action::Update, createPrivateDict(document, true));
						pagesConsumed = 2;
					}
					else
					{
						pagesConsumed = 1;
					}
				}
				else
				{
					// update the media for only the first page, then reset it back
					directives.emplace_back(0, Action::Update, createPrivateDict(front, false));
					if (pageCount > 1)
						directives.emplace_back(1, Action::Update, createPrivateDict(document, false));
					pagesConsumed = 1;
				}
				break;

			case Style::PrintOnOneSide:
				// duplex is implied for this style, so we disable duplex on the first
				// page so that only one side of it will be printed on, then re-enable it
				// on the second page.
				directives.emplace_back(0, Action::Update, createPrivateDict(front, false));
				if (pageCount > 1)
					directives.emplace_back(1, Action::Update, createPrivateDict(document, true));
				pagesConsumed = 1;
				break;

			case Style::DoNotPrintOnPage:
				// disable duplex and inject a blank page, then re-enable duplex on the
				// second page (if necessary). Note that "0" is used as the index of the
				// page to restore duplex because injected pages are not included in
				// these page indices.
				directives.emplace_back(0, Action::Inject, createPrivateDict(front, false));
				directives.emplace_back(0, Action::Update, createPrivateDict(document, isDuplex));
				break;
		}

		// if nothing left to do, then we can return early
		if (back.style == Style::None)
			return directives;

		// check to see if we need to adjust the back cover style (if there aren't
		// enough pages, etc).
		Style adjustedBackStyle = back.style;
		if (pagesConsumed >= pageCount)
		{
			// there's not enough pages for "PrintOnPage" or "PrintOnOneSide" because
			// the front cover's style used some of them up.
			adjustedBackStyle = Style::DoNotPrintOnPage;
		}
		else if (pagesConsumed >= pageCount - 1 && back.style == Style::PrintOnPage && isDuplex)
		{
			// there's not enough pages to print on both sides, so we adjust it to
			// print only on one side
			adjustedBackStyle = Style::PrintOnOneSide;
		}

		switch (adjustedBackStyle)
		{
			case Style::None:
				// do nothing, technically already handled above
				break;

			case Style::PrintOnPage:
				if (isDuplex)
				{
					// enable duplex on the second to last page so that both of the last
					// pages are printed on the back cover media
					directives.emplace_back(pageCount - 2, Action::Update, createPrivateDict(back, true));
				}
				else
				{
					// set the back cover media to apply on the last page
					directives.emplace_back(pageCount - 1, Action::Update, createPrivateDict(back, false));
				}
				break;

			case Style::PrintOnOneSide:
				// set the back cover media to apply on the last page and also disable duplex
				// to ensure only one side of it is printed on
				directives.emplace_back(pageCount - 1, Action::Update, createPrivateDict(back, false));
				break;

			case Style::DoNotPrintOnPage:
				// inject a new blank page
				directives.emplace_back(pageCount, Action::Inject, createPrivateDict(back, false));
				break;
		}

		return directives;
	}

	/// inserts front and back cover pages into the PDF
	void insertCoverPages(PDFFilePtr pdf, const Media &document, const CoverPage &front, const CoverPage &back, bool isDuplex)
	{
		using Action = CoverPageDirective::Action;
		
		std::vector<PagePtr> pages = pdf->getPages();
		std::vector<CoverPageDirective> directives = getDirectives(pdf, document, front, back, isDuplex);

		LiteralPtr frontCoverRef = nullptr;
		LiteralPtr backCoverRef = nullptr;

		DictionaryPtr catalog = pdf->getDocumentCatalog();
		DictionaryPtr pageTree = pdf->resolve<Dictionary>(catalog->get("Pages"));
		ArrayPtr kids = pdf->resolve<Array>(pageTree->get("Kids"));
		size_t pageCount = pdf->resolve<Literal>(pageTree->get("Count"))->getIntegerValue();
		
		for (const CoverPageDirective &directive : directives)
		{
			switch (directive.action)
			{
				case Action::Inject:
					if (directive.pageIndex == 0)
					{
						// clone the first page and delete the contents and TTEC part (if they exist)
						DictionaryPtr frontCover = std::dynamic_pointer_cast<Dictionary>(pages[0]->getDictionary()->copy());
						frontCover->remove("Contents");
						frontCover->add("TTECPart", directive.privateDict);
						frontCoverRef = Literal::FromObjectReference(pdf->addObject(frontCover));
					}
					else
					{
						// must be the last page
						DictionaryPtr backCover = std::dynamic_pointer_cast<Dictionary>(pages.back()->getDictionary()->copy());
						backCover->remove("Contents");
						backCover->add("TTECPart", directive.privateDict);
						backCoverRef = Literal::FromObjectReference(pdf->addObject(backCover));
					}
					break;

				case Action::Update:
					pages.at(directive.pageIndex)->getDictionary()->add("TTECPart", directive.privateDict);
					break;
			}
		}
		
		if (frontCoverRef)
		{
			pageCount++;
			kids->insert(0, frontCoverRef);
		}

		if (backCoverRef)
		{
			pageCount++;
			kids->add(backCoverRef);
		}

		pageTree->add("Count", Literal::FromInteger(pageCount));
	}
}

#ifdef STRIPEDHORSE_TOOLKIT_STANDALONE
int main(int argc, char *argv[])
#else
#include "coverpages.h"
int CoverPages::main(int argc, char *argv[])
#endif
{
	std::vector<std::string> args(&argv[0], &argv[argc]);

	if (args.size() < 3)
	{
		printUsage(argv[0]);
		return 1;
	}

	bool isDuplex = false;
	Media document;
	CoverPage front, back;

	auto iter = args.begin();
	auto lastOpt = args.end() - 2;
	
	while (++iter != lastOpt)
	{
		if (*iter == "--")
		{
			++iter;
			break;
		}

		// all options require an argument
		if (iter + 1 == lastOpt)
		{
			throw ArgumentException("Missing argument to " + *iter);
		}
		else if (*iter == "--back-style")
		{
			back.style = CoverPage::ParseStyle(*++iter);
		}
		else if (*iter == "--back-media")
		{
			back.type = *++iter;
		}
		else if (*iter == "--back-source")
		{
			back.position = *++iter;
		}
		else if (*iter == "--front-style")
		{
			front.style = CoverPage::ParseStyle(*++iter);
		}
		else if (*iter == "--front-media")
		{
			front.type = *++iter;
		}
		else if (*iter == "--front-source")
		{
			front.position = *++iter;
		}
		else if (*iter == "--media")
		{
			document.type = *++iter;
		}
		else if (*iter == "--source")
		{
			document.position = *++iter;
		}
		else if (*iter == "--duplex")
		{
			std::string duplexSetting = *++iter;
			if (duplexSetting == "true")
				isDuplex = true;
			else if (duplexSetting == "false")
				isDuplex = false;
			else
				throw ArgumentException("Invalid option for --duplex");
		}
	}

	std::string inputFile = *iter++;
	std::string outputFile = *iter++;

	if (iter != args.end())
		throw ArgumentException("Unexpected extra arguments after input and output filenames");
	
	if (front.style == CoverPage::Style::None && back.style == CoverPage::Style::None)
	{
		std::cerr << "Nothing to do, neither front nor back cover pages were requested." << std::endl;
		return 1;
	}

	if (!isDuplex)
	{
		// to simplify some of the logic, for simplex mode there
		// is no difference between PrintOnPage and PrintOnOneSide,
		// so we just set it to "PrintOnPage".
		if (front.style == CoverPage::Style::PrintOnOneSide)
			front.style = CoverPage::Style::PrintOnPage;

		if (back.style == CoverPage::Style::PrintOnOneSide)
			back.style = CoverPage::Style::PrintOnPage;
	}

	PDFFilePtr input = PDFFile::FromFile(inputFile);
	insertCoverPages(input, document, front, back, isDuplex);
	input->resolveEntireFile(false);
	input->write(outputFile);
	
	return 0;
}
