#include <algorithm>
#include <fstream>
#include <iostream>
#include <lemongfx/image-tiff.h>

using namespace LemonGFX;
using namespace LemonGFX::ImageIO;

namespace
{
	uint8_t integerGrayscale(uint8_t r, uint8_t g, uint8_t b)
	{
		// mixes 0.299 of red with 0.587 of green with 0.144 of blue
		// without using floating points
		return (299U * r + 587U * g + 114U * b) / 1000;
	}
}

#ifdef STRIPEDHORSE_TOOLKIT_STANDALONE
int main(int argc, char *argv[])
#else
#include "grayscaler.h"
int Grayscaler::main(int argc, char *argv[])
#endif
{
	if (argc < 3)
	{
		std::cerr << "Grayscaler Usage: " << std::endl;
		std::cerr << argv[0] << " colour-tiff-in grayscale-tiff-out" << std::endl;
		return 0;
	}

	std::ifstream inputFile(argv[1]);
	std::ofstream outputFile(argv[2]);

	LemonGFX::TIFFFileInfo tiffInfo(inputFile);
	
	if (tiffInfo.samplesPerPixel == 1)
	{
		// image is already in the desired format, copy input to output
		inputFile.seekg(0, std::ios_base::beg);
		outputFile << inputFile.rdbuf();
		return 0;
	}
	else if (tiffInfo.samplesPerPixel == 3)
	{
		// assume 4th channel is alpha of RGBA and not that the four channels are CMYK
		size_t newWidth, newHeight;
		size_t xResolution = tiffInfo.xResolution;
		size_t yResolution = tiffInfo.yResolution;
		bool interpolate = false;
		if (xResolution > 300 || yResolution > 300)
		{
			double multiplier = 300.0 / std::max(tiffInfo.xResolution, tiffInfo.yResolution);
			newWidth = tiffInfo.width * multiplier;
			newHeight = tiffInfo.height * multiplier;
			xResolution = 300;
			yResolution = 300;
			interpolate = true;
		}
		else
		{
			newWidth = tiffInfo.width;
			newHeight = tiffInfo.height;
		}

		Image img(8, 1, newWidth, newHeight);
		img.setResolution(xResolution, yResolution);
		Buffer &imgBuffer = img.getData();
		unsigned char *imgBufferData = imgBuffer.data();
		size_t bytesPerRow = img.getBytesPerRow();

		if (interpolate)
		{
			// here we have to determine which source pixels to use in the destination, and
			// at the moment, we are just doing a simple linear reduction
			size_t srcRowCount = 0;
			ReadStripDataFromTIFFFormat(inputFile, tiffInfo, [&](const unsigned char *scanlineData, size_t length){
				if (tiffInfo.width * tiffInfo.samplesPerPixel != length)
					throw std::runtime_error("Decoded row data is an unexpected size");
				
				size_t destRow = (srcRowCount * newHeight / tiffInfo.height);
				if ((srcRowCount + 1) * newHeight / tiffInfo.height == destRow)
				{
					// omit any scanlines where the destination scanline has already been written
					srcRowCount++;
					return;
				}

				for (size_t i = 0; i < bytesPerRow; i++)
				{
					size_t srcRowOffset = (i * tiffInfo.width / bytesPerRow) * tiffInfo.samplesPerPixel;
					uint8_t r = scanlineData[srcRowOffset + 0];
					uint8_t g = scanlineData[srcRowOffset + 1];
					uint8_t b = scanlineData[srcRowOffset + 2];
					imgBufferData[destRow * newWidth + i] = integerGrayscale(r, g, b);
				}

				srcRowCount++;
			});
		}
		else
		{
			// can take a shortcut, direct 1-to-1 mapping of samples
			size_t rowCount = 0;
			ReadStripDataFromTIFFFormat(inputFile, tiffInfo, [&](const unsigned char *scanlineData, size_t length){
				if (tiffInfo.width * tiffInfo.samplesPerPixel != length)
					throw std::runtime_error("Decoded row data is an unexpected size");
				
				for (size_t i = 0; i < tiffInfo.width; i++)
				{
					uint8_t r = scanlineData[i * tiffInfo.samplesPerPixel + 0];
					uint8_t g = scanlineData[i * tiffInfo.samplesPerPixel + 1];
					uint8_t b = scanlineData[i * tiffInfo.samplesPerPixel + 2];
					imgBufferData[rowCount * bytesPerRow + i] = integerGrayscale(r, g, b);
				}

				rowCount++;
			});
		}

		WriteToTIFFFormat(img, outputFile);
	}
	else
	{
		std::cerr << "Don't know how to handle images with " << tiffInfo.samplesPerPixel << " channels" << std::endl;
		return 1;
	}

	return 0;
}
