#include <cstdlib>
#include <cstring>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <lemongfx/bitmap.h>
#include <lemongfx/image.h>
#include <lemongfx/image-tiff.h>
#include <lemonpdf/pdffile.h>
#include <lemonpdf/pdffilter-flate-lzw.h>
#include <lemonpdf/gfxparser.h>
#include <lemonsqueezy/icompressor.h>
#include <libcore/streams.h>
#include <zlib.h>

using namespace LemonGFX;
using namespace LemonPDF;
using namespace LemonSqueezy;
using namespace StripedHorse;

namespace
{
	void showUsage(const char *progName)
	{
		std::cerr << "Halftoner Usage: " << std::endl;
		std::cerr << progName << " input.tiff output.tiff [ halftoner-dpi ]" << std::endl;
		std::cerr << progName << " -p input.pdf output.pdf [ halftoner-dpi ]" << std::endl;
		std::cerr << std::endl;
		std::cerr << " (-p means PDF mode)" << std::endl;
		std::cerr << " (if halftoner-dpi is -1, then Striped Horse halftoner is used)" << std::endl;
	}
}

#ifdef STRIPEDHORSE_TOOLKIT_STANDALONE
int main(int argc, char *argv[])
#else
#include "halftoner.h"
int Halftoner::main(int argc, char *argv[])
#endif
{
	if (argc <= 2)
	{
		showUsage(argv[0]);
		return 1;
	}

	bool pdfMode = false;
	std::vector<std::string> args(&argv[0], &argv[argc]);

	std::vector<std::string>::const_iterator argsIter = args.begin();
	argsIter++; // skip over argv[0] (executable name)

	while (argsIter->size() > 0 && (*argsIter)[0] == '-')
	{
		if (*argsIter == "-p")
			pdfMode = true;

		argsIter++;
	}

	// check that there are at least two arguments left
	if (args.end() - argsIter < 2)
	{
		showUsage(argv[0]);
		return 1;
	}
	
	const float dpmm = 10.0; // required for compatibility with Striped Horse
	size_t resolution = 300; // dpi determines the halftoner to use (defaults to 300 dpi halftoner)
	std::ifstream inputFile(*argsIter++, std::ios_base::binary);
	std::ofstream outputFile(*argsIter++, std::ios_base::binary);
	
	// if we still have one left over, interpret it as a dpmm or dpi value
	if (argsIter != args.end())
	{
		size_t pos = 0;
		int result = std::stoi(*argsIter, &pos);
		if (pos < argsIter->size())
		{
			std::cerr << "Unknown number: " << argv[3] << std::endl;
			return 1;
		}
		
		resolution = static_cast<size_t>(result);
	}
	
	HalftoningMethod hm = HalftoningMethod::kDefault;
	if (resolution <= 200)
		hm = HalftoningMethod::kReuss200dpi;
	else if (resolution <= 400)
		hm = HalftoningMethod::kReuss300dpi;
	else
		hm = HalftoningMethod::kReuss600dpi;

	if (pdfMode)
	{
		// the stream is moved into the PDFFile because the PDFFile handles it (requires r-value reference)
		PDFFilePtr pdf = PDFFile::FromStream(std::move(inputFile));
		pdf->resolveEntireFile();

		pdf->enumerateStreamsOfType("XObject", "Image", [&](const ObjectReference &objRef, StreamPtr stream){
			// don't care about its ID
			(void) objRef;

			Image inputImage = LemonPDF::Graphics::GetImageFromXObject(pdf, stream);
			Bitmap outputImage = Bitmap::Downsample(inputImage, RectangleU { 0, 0, inputImage.getWidth(), inputImage.getHeight() }, dpmm, BitmapRotation::kRotate0, hm, HalftoningCurve::kNormal, nullptr);

			stream->setDictionaryValue("BitsPerComponent", Literal::FromInteger(1));
			stream->setDictionaryValue("ColorSpace", Literal::FromName("DeviceGray"));

			ArrayPtr decodeArray = std::make_shared<LemonPDF::Array>();
			decodeArray->add(Literal::FromInteger(1));
			decodeArray->add(Literal::FromInteger(0));
			stream->setDictionaryValue("Decode", decodeArray);
			stream->getDictionary()->remove("DecodeParms");
			stream->setFilterNames({ "FlateDecode" });

			InputVectorStreamBuffer inputBuffer(outputImage.getData());
			std::istream inputStream(&inputBuffer);
			stream->filterIn(inputStream);
		});

		pdf->write(outputFile);
	}
	else
	{
		// read the tiff, halftone it, write it out again. easy!
		Image inputImage = ImageIO::ReadFromTIFFFormat(inputFile);
		Bitmap outputImage = Bitmap::Downsample(inputImage, RectangleU { 0, 0, inputImage.getWidth(), inputImage.getHeight() }, dpmm, BitmapRotation::kRotate0, hm, HalftoningCurve::kNormal, nullptr);
		ImageIO::WriteToTIFFFormat(outputImage, outputFile);
	}

	return 0;
}
