#include <fstream>
#include <lemonpdf/pdffile.h>

using namespace LemonPDF;

namespace
{
	std::ostream &operator<<(std::ostream &out, const LemonPDF::ObjectReference &objRef)
	{
		return (out << objRef.getObjectID() << " " << objRef.getGeneration() << " obj");
	}
}

#ifdef STRIPEDHORSE_TOOLKIT_STANDALONE
int main(int argc, char *argv[])
#else
#include "icc-injector.h"
int ICCInjector::main(int argc, char *argv[])
#endif
{
	if (argc < 3)
	{
		std::cerr << "ICC Injector Usage:" << std::endl;
		std::cerr << argv[0] << " input-pdf output-pdf icc-profile" << std::endl;
		return 0;
	}
	
	char *inFilename = argv[1];
	char *outFilename = argv[2];
	char *iccFilename = argv[3];

	PDFFilePtr file = PDFFile::FromFile(inFilename);
	
	// create the ICC stream object
	StreamPtr iccStream = Stream::FromIOStream(std::ifstream { iccFilename }, { "FlateDecode" });
	iccStream->setDictionaryValue("N", Literal::FromInteger(3));
	
	// Add the the stream object to the file and get a reference to it
	ObjectReference iccStreamRef = file->addObject(iccStream);

	// Create a name for the profile to be used in the resource dictionary
	std::string iccProfileName = "ScanPlusProfile";
	LiteralPtr iccProfileNameLiteral = Literal::FromName(iccProfileName);
	
	// color profiles are specified as an array, for ICC profiles, it must be
	// [ /ICCProfile x y R ]
	ArrayPtr colorSpaceArray = std::make_shared<Array>();
	colorSpaceArray->add(Literal::FromName("ICCBased"));
	colorSpaceArray->add(Literal::FromObjectReference(iccStreamRef));

	ObjectReference colorSpaceArrayRef = file->addObject(colorSpaceArray);
	LiteralPtr colorSpaceArrayRefLiteral = Literal::FromObjectReference(colorSpaceArrayRef);

	// Go through all image xobjects in the file and set its color space 
	// to the new name
	file->enumerateStreamsOfType("XObject", "Image", [&](const ObjectReference &objRef, StreamPtr stream){
		// check to see if existing color space is /DeviceRGB
		
		ObjectPtr existingColorSpace = stream->getDictionary()->get("ColorSpace");
		if (existingColorSpace == nullptr)
		{
			std::cerr << objRef << " does not have an existing colorspace, ignoring" << std::endl;
			return;
		}
		
		LiteralPtr existingColorSpaceName = file->resolve<Literal>(existingColorSpace);
		if (!existingColorSpaceName)
		{
		 	// might be missing for JPXDecode images, or might also be
		 	// an array for non-device colorspaces (in which case we
		 	// still want to ignore it)
		 	std::cerr << objRef << " does not have device colorspace or colorspace is missing, ignoring" << std::endl;
			return;
		}
		
		if (existingColorSpaceName->getNameValue() != "DeviceRGB")
		{
			std::cerr << objRef << " is in " << existingColorSpaceName->getNameValue() << " colorspace, ignoring" << std::endl;
			return;
		}
		
		// if we get to this point, then the existing colorspace is DeviceRGB
		// and is therefore suitable for replacing
		stream->getDictionary()->add("ColorSpace", colorSpaceArrayRefLiteral);
	});

	// Go through all page resources and add the the colour space as a
	// resource
	file->enumeratePages([&](PagePtr page){
		ResourceManagerPtr resources = page->getResources();
		resources->addColorSpace(iccProfileName, colorSpaceArray);
	});

	// resolving the entire file is necessary for now :(
	file->resolveEntireFile(false);
	file->write(outFilename);
	
	return 0;
}

