#include <cstring>
#include <ctime>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <lemongfx/image-tiff.h>
#include <lemonpdf/pdffile.h>
#include <lemonsqueezy/ccitt.h>

#include "image-injector.h"

using namespace LemonGFX;
using namespace LemonPDF;

/*
A lot of the work here was based on the work by mmarks in the toucan app
(that is, working out what actually goes into a PDF and PDF/A). Plagiarsed
here by dlemonby.
*/

namespace
{
	const std::string &kProducerCreator = "Striped Horse";

	StreamPtr makePageContentStreamForImage(const std::string &imageName, double imageWidth, double imageHeight)
	{
		std::stringstream pageContents;
		pageContents << "q\n";
		pageContents << "  " << imageWidth << " 0 0 " << imageHeight << " 0 0 cm\n";
		pageContents << "  /" << imageName << " Do\n";
		pageContents << "Q\n";
		return Stream::FromIOStream(std::move(pageContents));//, { "FlateDecode" });
	}

	std::string toHex(const std::string &input)
	{
		std::stringstream output;
		output << std::setfill('0');
		output << std::hex;
		for (const char &ch : input)
		{
			unsigned uch = static_cast<unsigned>(ch);
			output.width(2); // width is lost every output
			output << uch;
		}

		return output.str();
	}

	StreamPtr makeMetadataStream(time_t timestamp, const std::string &documentID, const std::string &instanceID)
	{
		struct tm t;
#ifdef _WIN32
		gmtime_s(&t, &timestamp);
#else
		gmtime_r(&timestamp, &t);
#endif

		std::stringstream metadata;

		metadata << "<?xpacket begin=\"\xef\xbb\xbf\" id=\"W5M0MpCehiHzreSzNTczkc9d\"?>\n";
		metadata << "  <x:xmpmeta xmlns:x=\"adobe:ns:meta/\" x:xmptk=\"" << kProducerCreator << "\">\n";
		metadata << "    <rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">\n";
		metadata << "      <rdf:Description rdf:about=\"\" xmlns:xmp=\"http://ns.adobe.com/xap/1.0/\">\n";
		metadata << "        <xmp:CreateDate>" << std::put_time(&t, "%FT%T") << "+00:00</xmp:CreateDate>\n";
		metadata << "        <xmp:CreatorTool>" << kProducerCreator << "</xmp:CreatorTool>\n";
		metadata << "      </rdf:Description>\n";
		metadata << "      <rdf:Description rdf:about=\"\" xmlns:pdf=\"http://ns.adobe.com/pdf/1.3/\">\n";
		metadata << "        <pdf:Producer>" << kProducerCreator << "</pdf:Producer>\n";
		metadata << "      </rdf:Description>\n";
		metadata << "      <rdf:Description rdf:about=\"\" xmlns:pdfaid=\"http://www.aiim.org/pdfa/ns/id/\">\n";
		metadata << "        <pdfaid:part>1</pdfaid:part>\n";
		metadata << "        <pdfaid:conformance>B</pdfaid:conformance>\n";
		metadata << "      </rdf:Description>\n";
		metadata << "      <rdf:Description rdf:about=\"\" xmlns:dc=\"http://purl.org/dc/elements/1.1/\">\n";
		metadata << "        <dc:format>application/pdf</dc:format>\n";
		metadata << "      </rdf:Description>\n";
		metadata << "      <rdf:Description rdf:about=\"\" xmlns:xmpMM=\"http://ns.adobe.com/xap/1.0/mm/\">\n";
		metadata << "        <xmpMM:DocumentID>uuid:" << toHex(documentID) << "</xmpMM:DocumentID>\n";
		metadata << "        <xmpMM:InstanceID>uuid:" << toHex(instanceID) << "</xmpMM:InstanceID>\n";
		metadata << "      </rdf:Description>\n";
		metadata << "    </rdf:RDF>\n";
		metadata << "  </x:xmpmeta>\n";

		// not sure why but almost every PDF has all these extra blank lines
		for (size_t i = 0; i < 21; i++)
			metadata << "                                                                                                   \n";
		
		metadata << "<?xpacket end=\"w\"?>";

		StreamPtr metadataStream = Stream::FromIOStream(std::move(metadata));
		metadataStream->setType("Metadata");
		metadataStream->setDictionaryValue("Subtype", Literal::FromName("XML"));
		return metadataStream;	
	}

	ArrayPtr createBilevelDecodeArray(uint32_t photometricInterpretation, DictionaryPtr decodeParms)
	{
		ArrayPtr decodeArray = std::make_shared<Array>();
		switch (photometricInterpretation)
		{
			case TIFFPhotometricInterpretation::kBlackIsZero:
				decodeArray->add(Literal::FromInteger(0));
				decodeArray->add(Literal::FromInteger(1));
				if (decodeParms)
					decodeParms->add("BlackIs1", Literal::FromBoolean(false));
				break;
		
			case TIFFPhotometricInterpretation::kWhiteIsZero:
				decodeArray->add(Literal::FromInteger(1));
				decodeArray->add(Literal::FromInteger(0));
				if (decodeParms)
					decodeParms->add("BlackIs1", Literal::FromBoolean(true));
				break;
			
			default:
				throw std::runtime_error("Unsupported photometric interpretation used in TIFF file");
		}

		return decodeArray;
	}

	int embedTIFFInsideNewPDF(const std::string &inFilename, const std::string &outFilename, bool outputPDFA, const char *iccFilename)
	{
		static const std::string imageName = "Im0";

		std::ifstream inStream(inFilename);
		TIFFFileInfo tiffInfo;
		Buffer stripData = ImageIO::ReadStripDataFromTIFFFormat(inStream, false, &tiffInfo);

		PDFFilePtr file = PDFFile::EmptyFile(1, 4);

		double pdfWidth = 72.0 * tiffInfo.width / tiffInfo.xResolution;
		double pdfHeight = 72.0 * tiffInfo.height / tiffInfo.yResolution;

		// create the XObject
		// since the data is already encoded, we don't want the Stream creator to try and encode it
		// again (which it will do if we tell it what filters to apply), so we create the stream object
		// without any filters and manually set the filters afterwards.
		StreamPtr image = Stream::FromBuffer(stripData);
		image->setType("XObject");
		image->setDictionaryValue("Subtype", Literal::FromName("Image"));
		image->setDictionaryValue("Name", Literal::FromName(imageName));
		image->setDictionaryValue("Width", Literal::FromInteger(tiffInfo.width));
		image->setDictionaryValue("Height", Literal::FromInteger(tiffInfo.height));
		
		if (tiffInfo.compression == TIFFCompression::kJPEG)
		{
			image->setDictionaryValue("BitsPerComponent", Literal::FromInteger(tiffInfo.bitsPerSample.front()));
			if (tiffInfo.samplesPerPixel == 1)
			{
				image->setDictionaryValue("ColorSpace", Literal::FromName("DeviceGray"));
			}
			else if ((tiffInfo.photometricInterpretation == 2 || tiffInfo.photometricInterpretation == 6) && tiffInfo.samplesPerPixel == 3)
			{
				image->setDictionaryValue("ColorSpace", Literal::FromName("DeviceRGB"));
			}
			else
			{
				std::cerr << "Unable to determine which colorspace to use for this TIFF";
				std::cerr << " (photometric interpretation " << tiffInfo.photometricInterpretation;
				std::cerr << ", samples per pixel " << tiffInfo.samplesPerPixel << ")" << std::endl;
				return 1;
			}
			
			image->setFilterNames({ "DCTDecode" });
		}
		else if (tiffInfo.compression == TIFFCompression::kCCITT4)
		{
			image->setDictionaryValue("BitsPerComponent", Literal::FromInteger(1));
			image->setDictionaryValue("ColorSpace", Literal::FromName("DeviceGray"));
			
			DictionaryPtr decodeParms = std::make_shared<Dictionary>();
			decodeParms->add("K", Literal::FromInteger(-1));
			decodeParms->add("Columns", Literal::FromInteger(tiffInfo.width));
			
			image->setDictionaryValue("Decode", createBilevelDecodeArray(tiffInfo.photometricInterpretation, decodeParms));
			image->setDictionaryValue("DecodeParms", decodeParms);
			image->setFilterNames({ "CCITTFaxDecode" });
		}
		else if (tiffInfo.compression == TIFFCompression::kNoCompression)
		{
			image->setDictionaryValue("BitsPerComponent", Literal::FromInteger(tiffInfo.bitsPerSample.front()));
			if (tiffInfo.samplesPerPixel == 1)
			{
				image->setDictionaryValue("ColorSpace", Literal::FromName("DeviceGray"));
				if (tiffInfo.bitsPerSample.front() == 1)
					image->setDictionaryValue("Decode", createBilevelDecodeArray(tiffInfo.photometricInterpretation, nullptr));
			}
			else if ((tiffInfo.photometricInterpretation == 2 || tiffInfo.photometricInterpretation == 6) && tiffInfo.samplesPerPixel == 3)
			{
				image->setDictionaryValue("ColorSpace", Literal::FromName("DeviceRGB"));
			}
		}
		else
		{
			std::cerr << "Unsupported compression type used in TIFF file: " << tiffInfo.compression << std::endl;
			return 1;
		}
		
		ObjectReference imageRef = file->addObject(image);

		// Add the content stream for the image
		StreamPtr pageContents = makePageContentStreamForImage(imageName, pdfWidth, pdfHeight);
		ObjectReference pageContentsRef = file->addObject(pageContents);

		// Get the document root but don't populate it yet
		DictionaryPtr pageRoot = std::make_shared<Dictionary>();
		ObjectReference pageRootRef = file->addObject(pageRoot);

		// Create the resources dictionary
		DictionaryPtr xObjects = std::make_shared<Dictionary>();
		xObjects->add(imageName, Literal::FromObjectReference(imageRef));

		ArrayPtr procSet = std::make_shared<Array>();
		procSet->add(Literal::FromName("PDF"));
		procSet->add(Literal::FromName("ImageC"));

		DictionaryPtr resources = std::make_shared<Dictionary>();
		resources->add("XObject", xObjects);
		resources->add("ProcSet", procSet);

		// Add the page itself
		ArrayPtr mediaBox = std::make_shared<Array>();
		mediaBox->add(Literal::FromInteger(0));
		mediaBox->add(Literal::FromInteger(0));
		mediaBox->add(Literal::FromReal(pdfWidth));
		mediaBox->add(Literal::FromReal(pdfHeight));
		
		DictionaryPtr page = std::make_shared<Dictionary>();
		page->add("Type", Literal::FromName("Page"));
		page->add("MediaBox", mediaBox);
		page->add("Contents", Literal::FromObjectReference(pageContentsRef));
		page->add("Parent", Literal::FromObjectReference(pageRootRef));
		page->add("Resources", resources);
		ObjectReference pageRef = file->addObject(page);

		// now populate the page tree root
		ArrayPtr kids = std::make_shared<Array>();
		kids->add(Literal::FromObjectReference(pageRef));
		pageRoot->add("Type", Literal::FromName("Pages"));
		pageRoot->add("Count", Literal::FromInteger(1));
		pageRoot->add("Kids", kids);
		
		// create info dictionary
		time_t timestamp = ::time(nullptr);
		DictionaryPtr infoDictionary = std::make_shared<Dictionary>();
		infoDictionary->add("Creator", Literal::FromString(kProducerCreator));
		infoDictionary->add("CreationDate", Literal::FromString(Types::CreateTimestamp(timestamp)));
		infoDictionary->add("Producer", Literal::FromString(kProducerCreator));
		file->setInfoDictionary(infoDictionary);

		// update document catalog
		DictionaryPtr catalog = file->getDocumentCatalog();
		catalog->add("Pages", Literal::FromObjectReference(pageRootRef));

		if (outputPDFA)
		{
			// 3 extra objects needed:
			//   Metadata dictionary
			//   OutputIntents dictionary
			//   Destination ICC profile stream
			StreamPtr iccProfile = Stream::FromIOStream(std::ifstream { iccFilename }, { "FlateDecode" });
			iccProfile->setDictionaryValue("N", Literal::FromInteger(3));
			ObjectReference iccProfileRef = file->addObject(iccProfile);

			DictionaryPtr outputIntent = std::make_shared<Dictionary>();
			outputIntent->add("Type", Literal::FromName("OutputIntent"));
			outputIntent->add("S", Literal::FromName("GTS_PDFA1"));
			outputIntent->add("OutputConditionIdentifier", Literal::FromString("Custom"));
			outputIntent->add("Info", Literal::FromString("3 Channel RGB IccProfile"));
			outputIntent->add("DestOutputProfile", Literal::FromObjectReference(iccProfileRef));

			ArrayPtr outputIntents = std::make_shared<Array>();
			outputIntents->add(outputIntent);

			std::string documentID;
			std::string instanceID;

			documentID.resize(32); // all zeroes
			instanceID.resize(32); // all zeroes
			
			StreamPtr metadata = makeMetadataStream(timestamp, documentID, instanceID);
			ObjectReference metadataRef = file->addObject(metadata);

			catalog->add("Metadata", Literal::FromObjectReference(metadataRef));
			catalog->add("OutputIntents", outputIntents);

			DictionaryPtr trailer = file->getTrailer();
			ArrayPtr id = std::make_shared<Array>();
			id->add(Literal::FromString(documentID));
			id->add(Literal::FromString(instanceID));
			trailer->add("ID", id);
		}

		file->write(outFilename);
		return 0;
	}

	int embedTIFFInsideExistingPDF(const std::string &inputTIFFName, const std::string &inputPDFName, const std::string &outputPDFName)
	{
		std::ifstream tiffFile(inputTIFFName);
		TIFFFileInfo tiffInfo(tiffFile);

		size_t pageCount = 0;
		StreamPtr imageStream;

		PDFFilePtr pdf = PDFFile::FromFile(inputPDFName);
		pdf->resolveEntireFile();

		pdf->enumeratePages([&](PagePtr page)->void {
			pageCount++;
			if (pageCount > 1)
				throw std::runtime_error("Cannot handle PDFs with more than one page");
			
			DictionaryPtr dictionary = page->getDictionary();
			if (!dictionary)
				throw std::runtime_error("Page does not have a dictionary");
			
			DictionaryPtr resources = pdf->resolve<Dictionary>(dictionary->get("Resources"));
			if (!resources)
				throw std::runtime_error("Page does not have any resources");
			
			DictionaryPtr xobjects = pdf->resolve<Dictionary>(resources->get("XObject"));
			if (!xobjects)
				throw std::runtime_error("Page does not have any XObject resources");
			
			if (xobjects->getSize() != 1)
				throw std::runtime_error("Can't handle pages with more than one XObject");
			
			xobjects->enumerate([&](const std::string &name, LemonPDF::ObjectPtr object){
				imageStream = pdf->resolve<Stream>(object);
				if (!imageStream)
					throw std::runtime_error("The XObject is not a stream");
				
				std::cerr << "Found XObject " << name << ", assuming it is a page image." << std::endl;
			});
		});

		// Can't do anything if we don't have the image stream
		if (!imageStream)
			throw std::runtime_error("No usable image stream was found in the PDF");

		DictionaryPtr streamDict = imageStream->getDictionary();

		streamDict->add("Width", Literal::FromInteger(tiffInfo.width));
		streamDict->add("Height", Literal::FromInteger(tiffInfo.height));
		streamDict->add("BitsPerComponent", Literal::FromInteger(tiffInfo.bitsPerSample.front()));
		streamDict->remove("Decode");

		// make sure the colorspace matches the TIFF
		if (tiffInfo.samplesPerPixel == 1)
			streamDict->add("ColorSpace", Literal::FromName("DeviceGray"));
		else if ((tiffInfo.photometricInterpretation == 2 || tiffInfo.photometricInterpretation == 6) && tiffInfo.samplesPerPixel == 3)
			streamDict->add("ColorSpace", Literal::FromName("DeviceRGB"));

		if (tiffInfo.compression == TIFFCompression::kJPEG)
		{
			LemonGFX::Buffer jpegData = ImageIO::ReadStripDataFromTIFFFormat(tiffFile, tiffInfo, false);
			streamDict->remove("DecodeParms");
			imageStream->setData(jpegData);
			imageStream->setFilterNames({ "DCTDecode" });
		}
		else
		{
			LemonGFX::Buffer imageData = ImageIO::ReadStripDataFromTIFFFormat(tiffFile, tiffInfo, true);
			imageStream->setData(imageData);
		}
		
		pdf->write(outputPDFName);
		return 0;
	}
}

#ifdef STRIPEDHORSE_TOOLKIT_STANDALONE
int main(int argc, char *argv[])
#else
int ImageInjector::main(int argc, char *argv[])
#endif
{
	if (argc < 3 || argc > 5)
	{
		std::cerr << "Image Injector Usage:" << std::endl;
		std::cerr << argv[0] << " [ -a destination-icc-profile ] input-tiff output-pdf" << std::endl;
		std::cerr << argv[0] << " -e input-tiff input-pdf output-pdf" << std::endl;
		return 0;
	}
	
	bool createNew = true;
	bool outputPDFA = false;
	size_t nextArg = 1;
	const char *iccFilename = nullptr;

	if (strcmp(argv[1], "-e") == 0)
	{
		nextArg++;
		createNew = false;
	}

	if (createNew)
	{
		if (argc > 4 && strcmp(argv[nextArg], "-a") == 0)
		{
			outputPDFA = true;
			iccFilename = argv[2];
			nextArg += 2;
		}

		std::string inFilename = argv[nextArg];
		std::string outFilename = argv[nextArg + 1];

		return embedTIFFInsideNewPDF(inFilename, outFilename, outputPDFA, iccFilename);
	}
	else
	{
		std::string inputTIFF = argv[nextArg];
		std::string inputPDF = argv[nextArg + 1];
		std::string outputPDF = argv[nextArg + 2];

		return embedTIFFInsideExistingPDF(inputTIFF, inputPDF, outputPDF);
	}

	return 0;
}

