#include <cstring>
#include <iostream>
#include <map>
#include <vector>
#include "booklet.h"
#include "coverpages.h"
#include "grayscaler.h"
#include "halftoner.h"
#include "icc-injector.h"
#include "image-injector.h"

// simple wrapper around 
int main(int argc, char *argv[])
{
	using MainFunction = int (*)(int, char **);
	static const std::map<std::string, MainFunction> toolMap {
		{ "booklet", Booklet::main },
		{ "coverpages", CoverPages::main },
		{ "grayscaler", Grayscaler::main },
		{ "halftoner", Halftoner::main },
		{ "icc-injector", ICCInjector::main },
		{ "image-injector", ImageInjector::main }
	};
	
	if (argc < 2)
	{
		for (const auto &tool : toolMap)
		{
			std::string execName = argv[0];
			execName.push_back(' ');
			execName.append(tool.first);
			execName.push_back('\0');

			char *args[2] = { &execName[0], nullptr };
			tool.second(1, args);

			std::cerr << std::endl;
		}

		return 1;
	}

	std::vector<char *> args;

	// create an argument encompassing both the executable name and tool name
	// (this is only so that when the tools are not called in standalone mode, their
	// help message shows the correct usage).
	std::string execName = argv[0];
	execName.push_back(' ');
	execName.append(argv[1]);
	
	// c_str() is valid until end of main() and casting away the const will work OK
	// provided none of the subtools try and change the value (which they have no reason to).
	args.push_back(const_cast<char *>(execName.c_str()));

	// push the rest of the arguments into the list
	for (int i = 2; i <= argc; i++)
		args.push_back(argv[i]);

	auto toolMapIter = toolMap.find(argv[1]);
	if (toolMapIter == toolMap.end())
	{
		std::cerr << "Unknown tool: " << argv[1] << std::endl;
		return 1;
	}
	
	// we give args.size() - 1 because the last element is the null pointer (which
	// is supposed to be there but is not included in the argument count).
	try
	{
		return toolMapIter->second(args.size() - 1, &args[0]);
	}
	catch (std::exception &ex)
	{
		std::cerr << "Terminating due to unhandled exception. " << ex.what() << std::endl;
		return 6; // SIGABRT = 6
	}
	catch (...)
	{
		std::cerr << "Terminating due to unknown exception." << std::endl;
		return 6;
	}
}
