#!/usr/bin/env python

import re
import sys
import optparse

# for some reason this one goes this way around
special_sizes = { (2, 1.5) : "2x1.5" }

parser = optparse.OptionParser()
parser.add_option("--xrange", help="specifies the range of x values to generate in inches (e.g. '1-4')", action="store", default="1-4")
parser.add_option("--xstep", help="specifies how to increment the x value in inches (e.g. '1')", action="store", type="float", default=1.0)
parser.add_option("--yrange", help="specifies the range of y values to generate in inches (e.g. '0.5-4')", action="store", default="0.5-4")
parser.add_option("--ystep", help="specifies how to increment the y value in inches (e.g. '0.5')", action="store", type="float", default=0.5)
parser.add_option("--extra", help="add an extra page size (e.g. 'AddressLabel:4x6')", action="append", default=[])
parser.add_option("--default", help="specifies which page size is the default one, (e.g. '3x6Rotated')", action="store", default=None)

(opts, args) = parser.parse_args()

if len(args) > 0:
    output = open(args[0], "w")
else:
    output = sys.stdout

output.write("// GENERATED FILE, DO NOT EDIT\n")

(xstart, xstop) = (float(x) for x in opts.xrange.split("-"))
(ystart, ystop) = (float(y) for y in opts.yrange.split("-"))

# CUPS seems to care if "1" is printed as "1.0", so avoid
# printing decimal places if possible
def trunc(x):
    if int(x) == float(x):
        return "%d" % (x)
    else:
        return "%.1f" % (x)

def write_size(out, x, y, known_as=None, default_size=None):
    xstr = trunc(x)
    ystr = trunc(y)

    if (x, y) in special_sizes:
        name = special_sizes[(x, y)]
    elif x > y:
        # x and y should swap here
        name = "%sx%sRotated" % (ystr, xstr)
    else:
        name = "%sx%s" % (xstr, ystr)

    if known_as is None:
        known_as = '%s\\" x %s\\"' % (xstr, ystr)
    
    out.write('#media "%s.FullBleed/%s" %sin %sin\n' % (name, known_as, xstr, ystr))
    if name == default_size:
        default_marker = "*"
    else:
        default_marker = ""
    out.write("%sMediaSize %s.FullBleed\n" % (default_marker, name))

# to prevent compounding floating point error, we multiply the step
# by the current iteration and add that to the start value rather
# than simply adding the step
yiter = 0
y = ystart + yiter * opts.ystep
while y <= ystop:
    xiter = 0
    x = xstart + xiter * opts.xstep
    while x <= xstop:
        write_size(output, x, y, default_size=opts.default)
        xiter += 1
        x = xstart + xiter * opts.xstep
    yiter += 1
    y = ystart + yiter * opts.ystep

extra_re = re.compile("((.+):)?([0-9.])x([0-9.])")

for extra in opts.extra:
    m = extra_re.match(extra)
    if m is None:
        raise RuntimeError("Can't parse this size: %s" % (extra,))
    write_size(output, x=float(m.group(3)), y=float(m.group(4)), known_as=m.group(2), default_size=opts.default)

output.write("\n")
output.close()
