#!/usr/bin/env python

import os
import re
import sys

class UnsupportedPPDException(Exception):
    pass

def processPPD(ppdfile):
    """Returns a 3-tuple (vendor ID, product ID, [list of model names])"""
    usbdetails = {}
    modelnames = []

    vidpidre = re.compile("^\\*TECBarcodeUSB(Product|Vendor)ID:\\s+\"(.+)\"")
    productre = re.compile("^\\*Product:\\s+\"\\((.+)\\)\"")
    tectargetre = re.compile("^\\*TECTarget:\\s+\"(.+)\"")

    for line in ppdfile:
        m = productre.match(line)
        if m is not None:
            modelnames.append(m.group(1))
            continue
        
        m = vidpidre.match(line)
        if m is not None:
            usbdetails[m.group(1)] = m.group(2)
            continue
        
        m = tectargetre.match(line)
        if m is not None:
            if m.group(1) == "unsupported":
                raise UnsupportedPPDException()
    
    return (usbdetails.get("Vendor"), usbdetails.get("Product"), modelnames)

def main():
    idlist = {}

    if len(sys.argv) < 3:
        sys.stderr.write("Usage: %s out:quirks-filename out:udev-rules in:ppd-filename [ in:ppdfilename ... ]\n")
        exit(1)

    for filename in sys.argv[3:]:
        ppdfile = open(filename, "r")
        try:
            vendorID, productID, modelnames = processPPD(ppdfile)
            if vendorID is None or productID is None:
                raise RuntimeError("%s does not contain a USB vendor and/or product ID" % (filename))
            
            if (vendorID, productID) not in idlist:
                idlist[(vendorID, productID)] = []
            idlist[(vendorID, productID)].extend(modelnames)
        except UnsupportedPPDException:
            sys.stderr.write("Ignoring %s (product not supported)\n" % (filename))
        ppdfile.close()
    
    keys = idlist.keys()
    keys.sort()

    quirks = open(sys.argv[1], "w")
    quirks.write("# GENERATED FILE\n\n")
    for key in keys:
        modelnames = idlist[key]
        modelnames.sort()
        for modelname in modelnames:
            quirks.write("# %s\n" % (modelname,))
        quirks.write("0x%s 0x%s blacklist\n\n" % key)
    quirks.close()
    
    udevrules = open(sys.argv[2], "w")
    udevrules.write("# GENERATED FILE\n")
    udevrules.write("#\n")
    udevrules.write("# Rules in this file should work with older and newer versions of udev\n\n")
    for key in keys:
        modelnames = idlist[key]
        udevrules.write("\n")
        for modelname in modelnames:
            udevrules.write("# %s\n" % (modelname,))
        udevrules.write("SUBSYSTEM==\"usb_device\", ACTION==\"add\", SYSFS{idVendor}==\"%s\", SYSFS{idProduct}==\"%s\", GROUP=\"lp\", MODE=\"0660\", RUN:=\"/usr/lib/toshiba-tec/barcode-printer-drivers/queuemanager\"\n" % key)
        udevrules.write("SUBSYSTEM==\"usb\", ENV{ID_VENDOR_ID}==\"%s\", ENV{ID_MODEL_ID}==\"%s\", RUN:=\"/usr/lib/toshiba-tec/barcode-printer-drivers/queuemanager\"\n" % key)
        udevrules.write("SUBSYSTEM==\"usb\", ACTION==\"add\", SYSFS{idVendor}==\"%s\", SYSFS{idProduct}==\"%s\", RUN:=\"\"\n" % key)
    
    udevrules.write("\n\n# All devices\n")
    udevrules.write("SUBSYSTEM==\"usb_device\", ACTION==\"remove\", RUN+=\"/usr/lib/toshiba-tec/barcode-printer-drivers/queuemanager\"\n")
    udevrules.close()

if __name__ == '__main__':
    main()
