#include <algorithm>
#include <cmath>
#include "bezier.h"

namespace
{
	inline double distance(const LemonGFX::PointF &p1, const LemonGFX::PointF &p2)
	{
		LemonGFX::RectangleF r { p1, p2 };
		return ::sqrt(::pow(r.getWidth(), 2) + ::pow(r.getHeight(), 2)); 
	}
}

LemonGFX::PointF LemonGFX::Bezier::MidPoint(const LemonGFX::PointF &p1, const LemonGFX::PointF &p2, double t)
{
	return LemonGFX::PointF { p1.x + t * (p2.x - p1.x), p1.y + t * (p2.y - p1.y) };
}

std::vector<LemonGFX::PointF> LemonGFX::Bezier::ApproximateCurve(const PointF &p1, const PointF &cp, const PointF &p2)
{
	std::vector<PointF> result;
	const size_t steps = std::max(static_cast<size_t>(5), static_cast<size_t>(distance(p1, cp) + distance(cp, p2) / 25.0));

	result.push_back(p1);
	for (size_t i = 0; i < steps; i++)
	{
		double t = static_cast<double>(i) / steps;

		PointF q1 = MidPoint(p1, cp, t);
		PointF q2 = MidPoint(cp, p2, t);

		result.emplace_back(MidPoint(q1, q2, t));
	}
	result.push_back(p2);

	return result;
}

std::vector<LemonGFX::PointF> LemonGFX::Bezier::ApproximateCurve(const PointF &p1, const PointF &cp1, const PointF &cp2, const PointF &p2)
{
	std::vector<PointF> result;
	double totalDistance = distance(p1, cp1) + distance(cp1, cp2) + distance(cp2, p2);
	const size_t steps = std::max(static_cast<size_t>(5), static_cast<size_t>(totalDistance / 25.0));

	result.push_back(p1);
	for (size_t i = 0; i < steps; i++)
	{
		double t = static_cast<double>(i) / steps;
		
		PointF q1 = MidPoint(p1, cp1, t);
		PointF q2 = MidPoint(cp1, cp2, t);
		PointF q3 = MidPoint(cp2, p2, t);

		PointF r1 = MidPoint(q1, q2, t);
		PointF r2 = MidPoint(q2, q3, t);

		result.emplace_back(MidPoint(r1, r2, t));
	}

	result.push_back(p2);
	return result;
}
