#include <algorithm>
#include <bitset>
#include <cstring>
#include <iterator>
#include <sstream>
#include <unordered_map>
#include <unordered_set>
#include "bitmap.h"
#include "draw.h"
#include "halftoner.h"
#include "reuss-halftoner.h"
#include "thresholder.h"

namespace
{
	inline size_t getByteIndex(size_t bytesPerRow, size_t x, size_t y)
	{
		return bytesPerRow * y + (x / 8);
	}

	using SourceCoordinateFunction = LemonGFX::PointU(*)(size_t, size_t, size_t, size_t, size_t, size_t);

	inline LemonGFX::PointU getSourceCoordinates_Rotate0(size_t srcWidth, size_t srcHeight, size_t destWidth, size_t destHeight, size_t destX, size_t destY)
	{
		return LemonGFX::PointU(destX * srcWidth / destWidth, destY * srcHeight / destHeight);
	}

	inline LemonGFX::PointU getSourceCoordinates_Rotate90(size_t srcWidth, size_t srcHeight, size_t destWidth, size_t destHeight, size_t destX, size_t destY)
	{
		return LemonGFX::PointU(destY * srcWidth / destHeight, (destWidth - 1 - destX) * srcHeight / destWidth);
	}

	inline LemonGFX::PointU getSourceCoordinates_Rotate180(size_t srcWidth, size_t srcHeight, size_t destWidth, size_t destHeight, size_t destX, size_t destY)
	{
		return LemonGFX::PointU((srcWidth - 1) - (destX * srcWidth / destWidth), (srcHeight - 1) - (destY * srcHeight / destHeight));
	}

	inline LemonGFX::PointU getSourceCoordinates_Rotate270(size_t srcWidth, size_t srcHeight, size_t destWidth, size_t destHeight, size_t destX, size_t destY)
	{
		return LemonGFX::PointU((destHeight - 1 - destY) * srcWidth / destHeight, destX * srcHeight / destWidth);
	}

	// TOPIX compression
	StripedHorse::Buffer compressTOPIXRowData(const StripedHorse::Buffer &thisLine, const StripedHorse::Buffer &lastLine)
	{
		StripedHorse::Buffer result;
		std::bitset<8> block512;
		std::vector<std::bitset<8>> block64(8);
		std::vector<std::vector<std::bitset<8>>> block8(8);
		for (size_t i = 0; i < 8; i++)
			block8.at(i).resize(8);
		
		for (size_t i = 0; i < thisLine.size(); i++)
		{	
			unsigned char ch = thisLine.at(i) ^ lastLine.at(i);
			if (!ch)
				continue;
			
			size_t b512 = i / 64;
			size_t b64 = (i % 64) / 8;
			size_t b8 = i % 8;
			
			block8.at(b512).at(b64).set(7 - b8);
			block64.at(b512).set(7 - b64);
			block512.set(7 - b512);
		}
		
		result.push_back(block512.to_ulong());
		if (block512.none())
			return result;
		
		for (size_t i = 0; i < 8; i++)
		{
			if (!block512.test(7 - i))
				continue;
		
			const std::bitset<8> &thisBlock64 = block64.at(i);
			result.push_back(thisBlock64.to_ulong());
			
			for (size_t j = 0; j < 8; j++)
			{
				if (!thisBlock64.test(7 - j))
					continue;
				
				const std::bitset<8> &thisBlock8 = block8.at(i).at(j);
				result.push_back(thisBlock8.to_ulong());
				
				for (size_t k = 0; k < 8; k++)
				{
					if (!thisBlock8.test(7 - k))
						continue;
					
					size_t idx = (i * 64) + (j * 8) + k;
					unsigned char ch = thisLine[idx] ^ lastLine[idx];
					if (ch == '\0')
						throw StripedHorse::InternalInconsistencyException("Outputting unnecessary byte");
					result.push_back(thisLine[idx] ^ lastLine[idx]);
				}
			}
		}
		
		return result;
	}
		
	size_t expandTOPIXRowData(const StripedHorse::Buffer &buffer, size_t offset, StripedHorse::Buffer &output)
	{
		size_t n = 0;
		uint8_t h512 = buffer.at(offset + n++);
		
		std::fill(output.begin(), output.end(), 0);
		
		for (size_t i = 0; i < 8; i++)
		{
			if (!(h512 & (1 << (7 - i))))
				continue;
			
			uint8_t h64 = buffer.at(offset + n++);
			
			for (size_t j = 0; j < 8; j++)
			{
				if (!(h64 & (1 << (7 - j))))
					continue;
				
				uint8_t h8 = buffer.at(offset + n++);
				
				for (size_t k = 0; k < 8; k++)
				{
					if (!(h8 & (1 << (7 - k))))
						continue;
					
					output.at(i * 64 + j * 8 + k) = buffer.at(offset + n++);
				}
			}
		}
		
		return n;
	}
	
	void writeTOPIXCompressedFormat(const LemonGFX::Bitmap &bitmap, std::ostream &out)
	{
		StripedHorse::Buffer lastLine(bitmap.getBytesPerRow());
		
		out.put(0x00);
		out.put(0x00);
		size_t dataSize = 0;

		if (bitmap.getWidth() > 8 * 512)
			throw LemonGFX::BitmapException("Bitmap width exceeds supported width for TOPIX compression");
		
		if (bitmap.getHeight() == 0)
			throw LemonGFX::BitmapException("Bitmap has no height");
		
		for (size_t r = 0; r < bitmap.getHeight(); r++)
		{
			StripedHorse::Buffer thisLine = bitmap.getRow(r);
			StripedHorse::Buffer compressed = compressTOPIXRowData(thisLine, lastLine);
			out.write(reinterpret_cast<const char *>(&compressed[0]), compressed.size());
			lastLine = thisLine;
			dataSize += compressed.size();
		}
		
		if (dataSize > 65536)
		{
			throw LemonGFX::BitmapException("Unable to compress bitmap to under maximum size of 65536 bytes");
		}
		else
		{
			out.seekp(0);
			out.put((dataSize & 0xFF00) >> 8);
			out.put(dataSize & 0x00FF);
			out.seekp(0, std::ios::end);
		}
	}

	LemonGFX::Bitmap readTOPIXCompressedFormat(const StripedHorse::Buffer &buffer, size_t width, size_t height)
	{
		(void) height;
		
		LemonGFX::Bitmap result(width, 0);
		StripedHorse::Buffer rowData(result.getBytesPerRow());
		
		for (size_t n = 0; n < buffer.size();)
		{
			size_t rowDataSize = expandTOPIXRowData(buffer, n, rowData);
			if (rowDataSize == 0)
				throw LemonGFX::BitmapException("TOPIX data invalid");
			
			result.appendRowXor(rowData);
			n += rowDataSize;
		}
		
		return result;
	}

	uint8_t getLuminance(uint8_t r, uint8_t g, uint8_t b)
	{
		return (299U * r + 587U * g + 114U * b) / 1000;
	}

	struct DownsamplingParameters
	{
		LemonGFX::HalftoningCurve halftoningCurve;
		const unsigned char *srcData;
		size_t srcBytesPerRow;
		size_t srcWidth;
		size_t srcHeight;
		size_t spp;
		bool blackIs1;
		unsigned char *destData;
		size_t destBytesPerRow;
		size_t destWidth;
		size_t destHeight;
		size_t nullBits;
		SourceCoordinateFunction getSourceCoordinates;
	};

	template<typename Halftoner>
	inline unsigned char getAndAnalyseSample(size_t srcX, size_t srcY, const struct DownsamplingParameters &dp, Halftoner &halftoner, LemonGFX::DownsamplingAnalysis *analysis)
	{
		const unsigned char sample = dp.srcData[srcY * dp.srcBytesPerRow + srcX];
		if (halftoner.isPure(sample))
			analysis->pureSamples++;
		return sample;
	}

	template<typename Halftoner>
	LemonGFX::DownsamplingAnalysis simpleDownsample_8bps_1spp(const struct DownsamplingParameters &dp)
	{
		LemonGFX::DownsamplingAnalysis analysis;
		Halftoner halftoner(dp.halftoningCurve);

		for (size_t destY = 0; destY < dp.destHeight; destY++)
		{
			for (size_t x = 0; x < dp.destBytesPerRow - (dp.nullBits > 0 ? 1 : 0); x++)
			{
				// work on a whole byte at once rather than single bits
				unsigned char byte = 0;
				for (size_t c = 0; c < 8; c++)
				{
					const size_t destX = x * 8 + c;
					unsigned char sample = getAndAnalyseSample(destX, destY, dp, halftoner, &analysis);
					byte = (byte << 1) | halftoner.isEnabledForPoint(destX, destY, sample);
				}

				dp.destData[destY * dp.destBytesPerRow + x] = byte;
			}
		}

		// for images that don't end on a byte boundary we need to fill
		// in the remainder bits. Having this external to the other loop
		// is faster since we only have to check if there are null bits
		// once
		if (dp.nullBits)
		{
			for (size_t destY = 0; destY < dp.destHeight; destY++)
			{
				unsigned char byte = 0;
				for (size_t c = 0; c < 8 - dp.nullBits; c++)
				{
					const size_t destX = (dp.destBytesPerRow - 1) * 8 + c;
					unsigned char sample = getAndAnalyseSample(destX, destY, dp, halftoner, &analysis);
					byte = (byte << 1) | halftoner.isEnabledForPoint(destX, destY, sample);
				}

				byte <<= dp.nullBits;
				dp.destData[destY * dp.destBytesPerRow + dp.destBytesPerRow - 1] = byte;
			}
		}

		return analysis;
	}

	template<typename Halftoner>
	LemonGFX::DownsamplingAnalysis interpolatedDownsample_8bps_1spp(const struct DownsamplingParameters &dp)
	{
		LemonGFX::DownsamplingAnalysis analysis;
		Halftoner halftoner(dp.halftoningCurve);

		for (size_t destY = 0; destY < dp.destHeight; destY++)
		{
			for (size_t x = 0; x < dp.destBytesPerRow; x++)
			{
				// work on a whole byte at once rather than single bits
				unsigned char byte = 0;
				size_t limit = (x == dp.destBytesPerRow - 1 ? 8 - dp.nullBits : 8);
				for (size_t c = 0; c < limit; c++)
				{
					size_t destX = x * 8 + c;
					LemonGFX::PointU srcPoint(dp.getSourceCoordinates(dp.srcWidth, dp.srcHeight, dp.destWidth, dp.destHeight, destX, destY));
					unsigned char sample = getAndAnalyseSample(srcPoint.x, srcPoint.y, dp, halftoner, &analysis);
					byte = (byte << 1) | halftoner.isEnabledForPoint(destX, destY, sample);
				}

				if (limit != 8)
					byte <<= dp.nullBits;
				
				dp.destData[destY * dp.destBytesPerRow + x] = byte;
			}
		}

		return analysis;
	}

	template<typename Halftoner>
	LemonGFX::DownsamplingAnalysis interpolatedDownsample_8bps_3or4spp(const struct DownsamplingParameters &dp)
	{
		LemonGFX::DownsamplingAnalysis analysis;
		Halftoner halftoner(dp.halftoningCurve);
		
		for (size_t destY = 0; destY < dp.destHeight; destY++)
		{
			for (size_t x = 0; x < dp.destBytesPerRow; x++)
			{
				unsigned char byte = 0;
				size_t limit = (x == dp.destBytesPerRow - 1 ? 8 - dp.nullBits : 8);
				for (size_t c = 0; c < limit; c++)
				{
					size_t destX = x * 8 + c;
					LemonGFX::PointU srcPoint(dp.getSourceCoordinates(dp.srcWidth, dp.srcHeight, dp.destWidth, dp.destHeight, destX, destY));

					const size_t byteOffset = srcPoint.y * dp.srcBytesPerRow + srcPoint.x * dp.spp;
					uint8_t r, g, b;
					r = dp.srcData[byteOffset + 0];
					g = dp.srcData[byteOffset + 1];
					b = dp.srcData[byteOffset + 2];
					unsigned char sample = getLuminance(r, g, b);
					analysis.pureSamples += halftoner.isPure(sample);
					byte = (byte << 1) | halftoner.isEnabledForPoint(destX, destY, sample);
				}

				if (limit != 8)
					byte <<= dp.nullBits;
				
				dp.destData[destY * dp.destBytesPerRow + x] = byte;
			}
		}

		return analysis;
	}

	template<typename Halftoner>
	LemonGFX::DownsamplingAnalysis interpolatedDownsample_1bps_1spp(const struct DownsamplingParameters &dp)
	{
		// downsampling a bilevel image
		for (size_t destY = 0; destY < dp.destHeight; destY++)
		{
			for (size_t x = 0; x < dp.destBytesPerRow; x++)
			{
				unsigned char byte = 0;
				size_t limit = (x == dp.destBytesPerRow - 1 ? 8 - dp.nullBits : 8);
				for (size_t c = 0; c < limit; c++)
				{
					size_t destX = x * 8 + c;
					LemonGFX::PointU srcPoint(dp.getSourceCoordinates(dp.srcWidth, dp.srcHeight, dp.destWidth, dp.destHeight, destX, destY));

					// in the context of Images, 1 bits represent white (this is consistent with
					// how PDFs describe bitmaps), however, in the context of Bitmaps, 1 bits typically
					// represent black, so they are negated here
					size_t byteOffset = getByteIndex(dp.srcBytesPerRow, srcPoint.x, srcPoint.y);
					size_t bitOffset = srcPoint.x % 8;
					bool bit = (dp.srcData[byteOffset] & (1U << (7 - bitOffset)));
					if (dp.blackIs1)
						byte = (byte << 1) | bit;
					else
						byte = (byte << 1) | !bit;
				}

				if (limit != 8)
					byte <<= dp.nullBits;
				
				dp.destData[destY * dp.destBytesPerRow + x] = byte;
			}
		}

		// all samples are pure
		return LemonGFX::DownsamplingAnalysis(dp.destWidth * dp.destHeight);
	}
}

LemonGFX::Bitmap LemonGFX::Bitmap::Downsample(const Image &other, const RectangleU &destination, double dpmm, BitmapRotation rotation, HalftoningMethod hm, HalftoningCurve hc, struct DownsamplingAnalysis *analysisResult)
{
	const size_t spp = other.getSamplesPerPixel();
	if (spp != 1 && spp != 3 && spp != 4)
		throw BitmapException("Cannot downsample images that do not have 1, 3 or 4 channels");

	const size_t bps = other.getBitsPerSample();
	if (!(bps == 8 || (spp == 1 && bps == 1)))
		throw BitmapException("Cannot downsample images that do not have 8 bits per sample or 1-bit monochrome");

	// analyse as we go, information could be useful for determining compression
	DownsamplingAnalysis analysis;

	// one-off cached calculations
	DownsamplingParameters dp;
	dp.halftoningCurve = hc;
	dp.srcData = other.getData().data();
	dp.srcBytesPerRow = other.getBytesPerRow();
	dp.srcWidth = other.getWidth();
	dp.srcHeight = other.getHeight();
	dp.spp = other.getSamplesPerPixel();
	dp.blackIs1 = other.blackIs1();
	dp.destWidth = dpmm * destination.getWidth() / 10;
	dp.destHeight = dpmm * destination.getHeight() / 10;
	dp.destBytesPerRow = (dp.destWidth + 7) / 8;
	dp.nullBits = (8 - (dp.destWidth % 8)) % 8;
	dp.getSourceCoordinates = nullptr;

	// make sure the data size matches what we want
	if (other.getDataSize() != dp.srcHeight * dp.srcBytesPerRow)
		throw BitmapException("Source data does not match size");

	// choose the function for getting the source coordinates based on image rotation
	switch (rotation)
	{
		case BitmapRotation::kRotate0:
			// only for bilevel identical sized images can we omit sample interpolation
			if (!(bps == 8 && spp == 1 && dp.destHeight == dp.srcHeight && dp.destWidth == dp.srcWidth))
				dp.getSourceCoordinates = getSourceCoordinates_Rotate0;
			break;
		
		case BitmapRotation::kRotate90:
			dp.getSourceCoordinates = getSourceCoordinates_Rotate90;
			break;
		
		case BitmapRotation::kRotate180:
			dp.getSourceCoordinates = getSourceCoordinates_Rotate180;
			break;
		
		case BitmapRotation::kRotate270:
			dp.getSourceCoordinates = getSourceCoordinates_Rotate270;
			break;
		
		default:
			throw BitmapException("Unknown rotation");
	}

	// allocate enough room so that we can halftone "on-the-fly"
	std::vector<unsigned char> destData;
	destData.resize(dp.destBytesPerRow * dp.destHeight);
	dp.destData = destData.data();
	
	if (bps == 8 && spp == 1)
	{
		// single 8-bit channel
		if (dp.getSourceCoordinates)
		{
			switch (hm)
			{
				case HalftoningMethod::kThresholding:
					analysis = interpolatedDownsample_8bps_1spp<Thresholder>(dp);
					break;
				
				case HalftoningMethod::kReuss200dpi:
					analysis = interpolatedDownsample_8bps_1spp<ReussHalftoner200dpi>(dp);
					break;
				
				case HalftoningMethod::kReuss300dpi:
					analysis = interpolatedDownsample_8bps_1spp<ReussHalftoner300dpi>(dp);
					break;
				
				case HalftoningMethod::kReuss600dpi:
					analysis = interpolatedDownsample_8bps_1spp<ReussHalftoner600dpi>(dp);
					break;
					
				case HalftoningMethod::kStripedHorse9Toner:
				default:
					analysis = interpolatedDownsample_8bps_1spp<SimpleHalftoner>(dp);
					break;
			}
		}
		else
		{
			switch (hm)
			{
				case HalftoningMethod::kThresholding:
					analysis = simpleDownsample_8bps_1spp<Thresholder>(dp);
					break;
				
				case HalftoningMethod::kReuss200dpi:
					analysis = simpleDownsample_8bps_1spp<ReussHalftoner200dpi>(dp);
					break;
				
				case HalftoningMethod::kReuss300dpi:
					analysis = simpleDownsample_8bps_1spp<ReussHalftoner300dpi>(dp);
					break;
				
				case HalftoningMethod::kReuss600dpi:
					analysis = simpleDownsample_8bps_1spp<ReussHalftoner600dpi>(dp);
					break;
				
				case HalftoningMethod::kStripedHorse9Toner:
				default:
					analysis = simpleDownsample_8bps_1spp<SimpleHalftoner>(dp);
					break;
			}
		}
	}
	else if (bps == 8 && (spp == 3 || spp == 4))
	{
		// multiple 8-bit channels (3 or 4, and if 4, assume RGBA)
		switch (hm)
		{
			case HalftoningMethod::kThresholding:
				analysis = interpolatedDownsample_8bps_3or4spp<Thresholder>(dp);
				break;
			
			case HalftoningMethod::kReuss200dpi:
				analysis = interpolatedDownsample_8bps_3or4spp<ReussHalftoner200dpi>(dp);
				break;
			
			case HalftoningMethod::kReuss300dpi:
				analysis = interpolatedDownsample_8bps_3or4spp<ReussHalftoner300dpi>(dp);
				break;
			
			case HalftoningMethod::kReuss600dpi:
				analysis = interpolatedDownsample_8bps_3or4spp<ReussHalftoner600dpi>(dp);
				break;
			
			case HalftoningMethod::kStripedHorse9Toner:
			default:
				analysis = interpolatedDownsample_8bps_3or4spp<SimpleHalftoner>(dp);
				break;
		}

	}
	else if (bps == 1 && spp == 1)
	{
		analysis = interpolatedDownsample_1bps_1spp<SimpleHalftoner>(dp);
	}
	else
	{
		throw BitmapException("Unsupported source image format for downsampling");
	}
	
	if (analysisResult)
		*analysisResult = analysis;

	Bitmap bitmap(std::move(destData), dp.destWidth, dp.destHeight);
	bitmap.setResolution(other.getXResolution(), other.getYResolution());
	bitmap.m_position = destination.lowerLeft;
	return bitmap;
}

LemonGFX::Bitmap LemonGFX::Bitmap::InvalidBitmap()
{
	return Bitmap(); // empty bitmap is invalid
}

LemonGFX::Bitmap LemonGFX::Bitmap::Load(const StripedHorse::Buffer &data, size_t width, size_t height, enum CompressionFormat cf)
{
	switch (cf)
	{
		case CompressionFormat::kUncompressed:
			return Bitmap(data, width, height);

		case CompressionFormat::kTOPIX:
			return readTOPIXCompressedFormat(data, width, height);
		
		default:
			throw BitmapException("Unknown compression format specified");
	}
}

LemonGFX::Bitmap::Bitmap()
  : Image(1, 1, 0, 0), m_halftoningCurve(HalftoningCurve::kNormal)
{
}

LemonGFX::Bitmap::Bitmap(size_t width, size_t height)
  : Image(1, 1, width, height), m_halftoningCurve(HalftoningCurve::kNormal)
{
}

LemonGFX::Bitmap::Bitmap(const StripedHorse::PointU &size)
  : Bitmap(size.x, size.y)
{
}

LemonGFX::Bitmap::Bitmap(const StripedHorse::Buffer &buffer, size_t width, size_t height)
  : Image(1, 1, width, height, buffer), m_halftoningCurve(HalftoningCurve::kNormal)
{
}

LemonGFX::Bitmap::Bitmap(StripedHorse::Buffer &&buffer, size_t width, size_t height)
  : Image(1, 1, width, height, std::move(buffer)), m_halftoningCurve(HalftoningCurve::kNormal)
{
}

LemonGFX::Bitmap::Bitmap(const Bitmap &bitmap)
  : Image(bitmap), m_position(bitmap.m_position), m_halftoningCurve(bitmap.m_halftoningCurve)
{
}

LemonGFX::Bitmap::Bitmap(Bitmap &&bitmap)
  : Image(std::move(bitmap)), m_position(bitmap.m_position), m_halftoningCurve(bitmap.m_halftoningCurve)
{
}

LemonGFX::Bitmap &LemonGFX::Bitmap::operator=(const Bitmap &bitmap)
{
	this->Image::operator=(bitmap);
	m_position = bitmap.m_position;
	return *this;
}

void LemonGFX::Bitmap::appendRow(const StripedHorse::Buffer &rowData)
{
	if (rowData.size() != m_bytesPerRow)
		throw BitmapException("Row data is the wrong length");
	
	m_data.insert(m_data.end(), rowData.begin(), rowData.end());
	m_height++;
}

void LemonGFX::Bitmap::appendRowXor(const StripedHorse::Buffer &rowData)
{
	if (rowData.size() != m_bytesPerRow)
		throw BitmapException("Row data is the wrong length");
	
	if (m_height == 0)
	{
		m_data.insert(m_data.end(), rowData.begin(), rowData.end());
		m_height++;
	}
	else
	{
		m_data.resize(m_data.size() + m_bytesPerRow);
		size_t oldRow = (m_height - 1) * m_bytesPerRow;
		size_t newRow = m_height * m_bytesPerRow;
		for (size_t i = 0; i < m_bytesPerRow; i++)
			m_data.at(newRow + i) = m_data.at(oldRow + i) ^ rowData.at(i);
		
		m_height++;
	}
}

void LemonGFX::Bitmap::applyMask(const Bitmap &mask)
{
	if (m_data.size() != mask.m_data.size())
		throw BitmapException("Cannot mask image if the mask is a different size");
	
	for (size_t i = 0; i < m_data.size(); i++)
		m_data[i] &= ~mask.m_data[i];
}

bool LemonGFX::Bitmap::blackIs1() const
{
	return true;
}

void LemonGFX::Bitmap::clear()
{
	memset(m_data.data(), 0, m_data.size());
}

void LemonGFX::Bitmap::cloneRow(size_t srcRow, size_t destRow)
{
	cloneRowXor(srcRow, destRow, StripedHorse::Buffer());
}

void LemonGFX::Bitmap::cloneRowXor(size_t srcRow, size_t destRow, const StripedHorse::Buffer &buffer)
{
	if (srcRow >= m_height || destRow >= m_height)
		throw BitmapException("Invalid row");
	
	size_t srcRowOffset = m_bytesPerRow * srcRow;
	size_t destRowOffset = m_bytesPerRow * destRow;
	
	for (size_t i = 0; i < m_bytesPerRow; i++)
	{
		StripedHorse::Buffer::value_type xorValue {};
		if (i < buffer.size())
			xorValue = buffer[i];
		m_data[destRowOffset + i] = m_data[srcRowOffset + i] ^ xorValue;
	}
}

void LemonGFX::Bitmap::combine(const Bitmap &other)
{
	StripedHorse::PointU pos = other.m_position;
	
	size_t realX = 0;	
	if (pos.x & 0x04)
		realX = (pos.x + (pos.x % 8)) / 8;
	else
		realX = (pos.x - (pos.x % 8)) / 8;
	
	for (size_t y = 0; y < other.m_height && (y + pos.y < m_height); y++)
	{
		for (size_t x = 0; x < other.m_bytesPerRow && (x + realX < m_bytesPerRow); x++)
		{
			m_data.at((pos.y + y) * m_bytesPerRow + x + realX) = other.m_data.at(y * other.m_bytesPerRow + x);
		}
	}
}

void LemonGFX::Bitmap::compress(enum CompressionFormat bf, std::ostream &out) const
{
	switch (bf)
	{
		case CompressionFormat::kUncompressed:
			out.write(reinterpret_cast<const char *>(&m_data[0]), m_data.size());
			break;

		case CompressionFormat::kTOPIX:
			writeTOPIXCompressedFormat(*this, out);
			break;
		
		default:
			throw BitmapException("Unknown bitmap format specified");
	}
}

LemonGFX::Bitmap LemonGFX::Bitmap::cut(const StripedHorse::RectangleU &size) const
{
	Bitmap result;
	
	if (size.lowerLeft.x > size.upperRight.x || size.lowerLeft.y > size.upperRight.y || size.upperRight.x == 0 || size.upperRight.y == 0)
		throw BitmapException("Invalid size");
	
	// data is top-down (y ascends positively downwards) but
	// coordinates are bottom-up (y ascends positively upwards)
	StripedHorse::RectangleU dataRect { size.lowerLeft.x, size.upperRight.y, size.upperRight.x, size.lowerLeft.y };

	result.m_width = size.getWidth();
	result.m_height = size.getHeight();
	result.m_bytesPerRow = ((result.m_width - 1) / 8) + 1;
	result.m_data.resize(result.m_bytesPerRow * result.m_height);

	for (size_t y = 0; y < result.m_height; y++)
		for (size_t x = 0; x < result.m_width; x++)
			if (getBit(dataRect.lowerLeft.x + x, dataRect.upperRight.y + y))
				result.setBit(x, y);
	
	result.m_position = size.lowerLeft;
	return result;
}

LemonGFX::RectangleU LemonGFX::Bitmap::getAutoCroppedDimensions() const
{
	LemonGFX::RectangleU result;
	bool found = false;
	
	for (size_t y = 0; y < m_height; y++)
	{
		if (!isRowBlank(y))
		{
			result.upperRight.y = y;
			found = true;
			break;
		}
	}
	
	if (!found)
	{
		// image is empty
		return result;
	}
	
	for (size_t x = 0; x < m_width; x++)
	{
		if (!isColumnBlank(x))
		{
			result.lowerLeft.x = x;
			break;
		}
	}
	
	for (size_t x = m_width; x > 0; x--)
	{
		if (!isColumnBlank(x - 1))
		{
			result.upperRight.x = x;
			break;
		}
	}
	
	for (size_t y = m_height; y > 0; y--)
	{
		if (!isRowBlank(y - 1))
		{
			result.lowerLeft.y = y;
			break;
		}
	}
	
	return LemonGFX::RectangleU {
		result.lowerLeft.x, result.upperRight.y,
		result.upperRight.x, result.lowerLeft.y
	};
}

bool LemonGFX::Bitmap::getBit(size_t x, size_t y) const
{
	if (x >= m_width || y >= m_height)
		throw BitmapException("Index out of range");
	
	const unsigned char &byte = m_data[getByteIndex(m_bytesPerRow, x, y)];
	return byte & (0x80U >> (x % 8));
}

LemonGFX::HalftoningCurve LemonGFX::Bitmap::getHalftoningCurve() const
{
	return m_halftoningCurve;
}

LemonGFX::Buffer LemonGFX::Bitmap::getRow(size_t r) const
{
	if (r >= m_height)
		throw BitmapException("Index out of range");
	LemonGFX::Buffer::const_iterator it = m_data.begin();
	std::advance(it, r * m_bytesPerRow);
	return LemonGFX::Buffer(it, it + m_bytesPerRow);
}

const LemonGFX::PointU &LemonGFX::Bitmap::getPosition() const
{
	return m_position;
}

void LemonGFX::Bitmap::invert()
{
	for (unsigned char &byte : m_data)
		byte = ~byte;
}

bool LemonGFX::Bitmap::isColumnBlank(size_t c) const
{
	// getBit will throw for out of range
	
	for (size_t y = 0; y < m_height; y++)
		if (getBit(c, y))
			return false;
	
	return true;
}

bool LemonGFX::Bitmap::isRowBlank(size_t r) const
{
	if (r >= m_height)
		throw BitmapException("Index out of range");
	
	for (size_t i = (r * m_bytesPerRow); i < ((r + 1) * m_bytesPerRow); i++)
		if (m_data[i])
			return false;
	
	return true;
}

void LemonGFX::Bitmap::rotate180()
{
	// Swap the top-left for the bottom-right, and slowly repeat
	// until reaching the centre (swapping only needs to occur if
	// the bits are different)
	for (size_t y = 0; y < m_height / 2; y++)
	{
		for (size_t x = 0; x < m_width; x++)
		{
			size_t xt = m_width - x - 1;
			size_t yt = m_height - y - 1;

			bool p1 = getBit(x, y);
			bool p2 = getBit(xt, yt);

			if (p1 != p2)
			{
				setBit(x, y, p2);
				setBit(xt, yt, p1);
			}
		}
	}

	// If the number of rows is odd, then swap the bits of the
	// centre row
	if (m_height % 2 == 1)
	{
		size_t y = (m_height + 1) / 2;
		for (size_t x = 0; x < m_width / 2; x++)
		{
			size_t xt = m_width - x - 1;
			bool p1 = getBit(x, y);
			bool p2 = getBit(xt, y);

			if (p1 != p2)
			{
				setBit(x, y, p2);
				setBit(xt, y, p1);
			}
		}
	}
}

void LemonGFX::Bitmap::setBit(size_t x, size_t y, bool v)
{
	unsigned char &byte = m_data[getByteIndex(m_bytesPerRow, x, y)];
	byte = (byte & ~(static_cast<unsigned char>(0x80) >> (x % 8))) | (static_cast<unsigned char>(v) << (7 - (x % 8)));
}

void LemonGFX::Bitmap::setHalftoningCurve(HalftoningCurve hc)
{
	m_halftoningCurve = hc;
}

void LemonGFX::Bitmap::setPosition(const LemonGFX::PointU &position)
{
	m_position = position;
}

void LemonGFX::Bitmap::setRow(size_t r, const LemonGFX::Buffer &data)
{
	if (r >= m_height)
		throw BitmapException("Index out of range");
	
	if (data.size() < m_bytesPerRow)
		throw BitmapException("Row data not large enough");
	
	LemonGFX::Buffer::iterator dstStart = m_data.begin() + (r * m_bytesPerRow);
	
	std::copy(data.begin(), data.begin() + m_bytesPerRow, dstStart);
}

