#ifndef LEMONGFX_BITMAP_H
#define LEMONGFX_BITMAP_H 1

#include <functional>
#include <iostream>
#include <limits>
#include "exception.h"
#include "gfxtypes.h"
#include "halftoner.h"
#include "image.h"

namespace LemonGFX
{
	enum class BitmapRotation
	{
		kRotate0,
		kRotate90,
		kRotate180,
		kRotate270
	};
	
	enum class HalftoningMethod
	{
		kDefault,
		kStripedHorse9Toner,
		kReuss200dpi,
		kReuss300dpi,
		kReuss600dpi,
		kThresholding
	};

	struct DownsamplingAnalysis
	{
		DownsamplingAnalysis(size_t p)
		: pureSamples(p) {}

		DownsamplingAnalysis() {}

		size_t pureSamples = 0;
	};

	class Bitmap : public Image
	{
	public:
		enum class CompressionFormat
		{
			kUncompressed,
			kTOPIX
		};

		static Bitmap Downsample(const Image &other, const RectangleU &destination, double dpmm, BitmapRotation rotation = BitmapRotation::kRotate0, HalftoningMethod ht = HalftoningMethod::kDefault, HalftoningCurve hc = HalftoningCurve::kNormal, struct DownsamplingAnalysis *analysis = nullptr);

		static Bitmap InvalidBitmap();

		static Bitmap Load(const Buffer &data, size_t width, size_t height, enum CompressionFormat bf);

		/**
		 Constructs an empty, invalid bitmap. Any operations on this bitmap
		 will fail.
		 */
		Bitmap();
		
		/**
		 Constructs a blank bitmap object
		 @param[in] width the width of the bitmap (in dots)
		 @param[in] height the height of the bitmap (in dots)
		 */
		Bitmap(size_t width, size_t height);
		
		/**
		 Constructs a blank bitmap object
		 @param[in] size the size of the bitmap to create (in dots)
		 */
		Bitmap(const PointU &size);
		
		/**
		 Constructs a bitmap object
		 @param[in] bits a buffer containing the bits of the bitmap (1 is black, 0 is white)
		 @param[in] width the width of the bitmap (in dots)
		 @param[in] height the height of the bitmap (in dots)
		 */
		Bitmap(const Buffer &bits, size_t width, size_t height);

		/**
		 Constructs a bitmap by moving the supplied buffer
		 @param[in] bits the buffer that should be moved into this bitmap (1 is black, 0 is white)
		 @param[in] width the width of the bitmap (in dots)
		 @param[in] height the height of the bitmap (in dots)
		*/
		Bitmap(Buffer &&bits, size_t width, size_t height);
		
		Bitmap(const Bitmap &);
		Bitmap(Bitmap &&);
		
		Bitmap &operator=(const Bitmap &);

		/**
		 Appends a row of data to the bitmap, increasing its height.
		 @param[in] rowData the row data to append
		 */
		void appendRow(const Buffer &buf);
		
		/**
		 Appends a row of data to the bitmap, xoring it against the previous
		 row. If there is no previous row, then this call is equivalent to
		 appendRow.
		 @param[in] rowData the row data to append
		 */
		void appendRowXor(const Buffer &buf);
		
		/**
		 Performs a bitwise-and operation on the bitmap using the bits from the
		 parameter (effectively applying a mask)
		*/
		void applyMask(const Bitmap &other);

		/**
		 Returns true (since in the context of bitmaps, 1 is considered black)
		 */
		bool blackIs1() const override;

		/**
		Erases the entire bitmap
		*/
		void clear();
		
		/**
		 Overwrites the contents of destRow with the contents of srcRow
		 @param[in] srcRow the row to copy data from
		 @param[in] destRow the row to overwrite with the data from srcRow
		 */
		void cloneRow(size_t srcRow, size_t destRow);
		
		/**
		 Overwrites the contents of destRow with the contents of srcRow after
		 xor'ing with the provided row buffer
		 @param[in] srcRow the row to copy data from
		 @param[in] destRwo the row to copy data to
		 @param[in] buffer the buffer to xor against
		 */
		void cloneRowXor(size_t srcRow, size_t destRow, const Buffer &buffer);

		/**
		 Overwrites the bits of this bitmap with the bits of the provided
		 bitmap.
		 @param[in] b the bitmap to copy bits from. The position of the
		            provided bitmap determines where it is placed on this
		            bitmap.
		 */
		void combine(const Bitmap &b);
		
		/**
		 Writes 1-bit bitmap data uncompressed, or compressed with TOPIX or RLE 
		 @param[in] cf the format to write the the compressed data in
		 @param[in] o the output stream to write the bitmap data to
		 */
		void compress(enum CompressionFormat cf, std::ostream &o) const;
		
		/**
		 Extracts a rectangular region of the bitmap
		 @param[in] size a rectangle representing the region to cut
		 @return A bitmap of the requested size containing the extracted data
		 */
		Bitmap cut(const RectangleU &size) const;
		
		/**
		 Determines the size of the bitmap after removing leading and
		 trailing blank rows and columns.
		 @return A rectangle containing the region of the bitmap after leading
		         and trailing blank rows and columns. If no rows or columns
		         were excluded, the result is the size of the entire bitmap.
		 */
		RectangleU getAutoCroppedDimensions(void) const;
		
		/**
		 Gets a single bit of the bitmap
		 @param[in] x the x coordinate
		 @param[in] y the y coordinate
		 @return true if the specified bit is set, otherwise false
		 */
		bool getBit(size_t x, size_t y) const;

		/**
		 Gets the halftoning curve that is currently set (default is kNormal)
		 */
		HalftoningCurve getHalftoningCurve() const;
		
		
		/**
		 The position of the bitmap.
		 @return the absolute position of the bitmap
		 */
		const PointU &getPosition() const;
		
		/**
		 Extracts a single row from the bitmap
		 @param[in] row the 0-based index of the row to return (starting from
		            the top).
		 @return a buffer object containing the bitmap data. The size of the
		         buffer will be the same as this object's bytes-per-row.
		 */
		Buffer getRow(size_t row) const;
		
		/**
		 Flips all the bits in the image.
		*/
		void invert() override;
		
		/**
		 Determines if the specified column is blank (has no bits set)
		 @param[in] c the 0-based index of the column to check (starting from
		            the left).
		 @return true if no bits in the specified column are set
		 */
		bool isColumnBlank(size_t c) const;
		
		/**
		 Determines if the specified row is blank (has no bits set)
		 @param[in] r the 0-based index of the row to check (starting from the
		            top).
		 @return true if no bits in the specificed row are set
		 */
		bool isRowBlank(size_t r) const;

		/**
		 Rotates the bitmap 180 degrees
		 */
		void rotate180();
		
		/**
		 Sets the specified bit in the bitmap
		 @param[in] x the x coordinate of the bit to set
		 @param[in] y the y coordinate of the bit to set
		 @param[in] v if true, the bit is set, if false, the bit is cleared
		              (the default is true) 
		 */
		void setBit(size_t x, size_t y, bool v = true);

		/**
		 Sets the halftoning curve for any rendering operations (such as DrawPathsOnBitmap)
		 @param[in] hc the Halftoning Curve
		 */
		void setHalftoningCurve(HalftoningCurve hc);
		
		/**
		 Sets the position of the bitmap. This is not used internally
		 in any way, it is just a way to be able to pair a bitmap with a set
		 of coordinates.
		 @param[in] p the position to set
		 */
		void setPosition(const PointU &p);
		
		/**
		 Sets the specified row data
		 @param[in] r the row index to set
		 @param[in] data the data to set for the row
		 */
		void setRow(size_t r, const Buffer &data);

	private:
		PointU m_position;
		HalftoningCurve m_halftoningCurve;
	};
	
	class BitmapException : public LemonGFX::Exception
	{
	public:
		BitmapException(const std::string &msg) : LemonGFX::Exception(msg) {}
	};
}

#endif

