#include <algorithm>
#include <bitset>
#include <functional>
#include <set>
#include "draw.h"
#include "halftoner.h"
#include "stroke.h"

namespace
{
	struct Edge
	{
		using type = LemonGFX::PointF::type;

		Edge(type yMin, type yMax, type x1, type x2, int direction, size_t shapeIndex)
		  : yMin(yMin), yMax(yMax), x1(x1), x2(x2), direction(direction), shapeIndex(shapeIndex)
		{
		}
		
		type yMin;
		type yMax;
		type x1;
		type x2;
		int direction;
		size_t shapeIndex;
	};

	struct ActiveEdge
	{
		using type = Edge::type;

		ActiveEdge(type x, int direction, size_t shapeIndex)
		  : x(x), direction(direction), shapeIndex(shapeIndex)
		{
		}

		bool operator<(const ActiveEdge &other) const
		{
			return x < other.x;
		}

		type x;
		int direction;
		size_t shapeIndex;
	};

	struct EdgeSorter
	{
		bool operator()(const Edge &e1, const Edge &e2) const
		{
			return e1.yMin < e2.yMin;
		}
	};

	std::vector<Edge> getEdgeTable(const LemonGFX::Path &path)
	{
		std::vector<Edge> et;

		for (size_t i = 0; i < path.getSize(); i++)
		{
			const LemonGFX::Path::Subpath &subpath = path[i];
			LemonGFX::PointF lastPoint = subpath.back();
			for (const LemonGFX::PointF &thisPoint : subpath)
			{
				if (std::abs(thisPoint.y - lastPoint.y) > 0.001)
				{
					if (thisPoint.y < lastPoint.y)
						et.emplace_back(thisPoint.y, lastPoint.y, lastPoint.x, thisPoint.x, 1, i);
					else
						et.emplace_back(lastPoint.y, thisPoint.y, thisPoint.x, lastPoint.x, -1, i);
				}
				
				lastPoint = thisPoint;
			}
		}

		std::sort(et.begin(), et.end(), EdgeSorter());
		return et;
	}

	std::vector<ActiveEdge> getActiveEdges(const std::vector<Edge> &edges, Edge::type y)
	{
		std::vector<ActiveEdge> aet;
		for (const Edge &e : edges)
		{
			if (e.yMin <= y && y <= e.yMax)
			{
				Edge::type ratio = ((y - e.yMin) / (e.yMax - e.yMin));
				Edge::type xEdge = e.x2 + (e.x1 - e.x2) * ratio;
				aet.emplace_back(xEdge, e.direction, e.shapeIndex);
			}
		}
		std::sort(aet.begin(), aet.end());
		return aet;
	}

	void fillRange(LemonGFX::Bitmap &b, size_t y, ActiveEdge::type onEdge, ActiveEdge::type offEdge, uint16_t intensity)
	{
		if (onEdge < 0.0)
			onEdge = 0.0;
		if (onEdge >= b.getWidth())
			return;

		for (size_t x = static_cast<size_t>(onEdge); x < static_cast<size_t>(offEdge); x++)
		{
			if (x < b.getWidth())
				b.setBit(x, y, intensity & (0x8000U >> ((y % 4) * 4 + (x % 4))));
			else
				return;
		}
	}

	LemonGFX::Path preparePathForDevice(const LemonGFX::Path &path, const LemonGFX::RectangleF &globalBounds, std::function<double(double)> toDeviceSpace)
	{
		LemonGFX::TransformMatrix ctm;

		ctm.scale(toDeviceSpace(1.0), toDeviceSpace(1.0));
		ctm.translate(-globalBounds.lowerLeft.x, -globalBounds.lowerLeft.y);
	
		LemonGFX::Path result(path); // copy attributes
		result.clearSubpaths(); // but erase subpaths so we can add the translated ones
		for (const LemonGFX::Path::Subpath &subpath : path)
		{
			LemonGFX::Path::Subpath newSubpath;
			newSubpath.reserve(subpath.size());
			for (const LemonGFX::PointF &subpoint : subpath)
				newSubpath.emplace_back(ctm * subpoint);
			result.addSubpath(newSubpath);
		}

		result.setStrokeWidth(toDeviceSpace(path.getStrokeWidth()));
		return result;
	}

	void fillOnBitmap(LemonGFX::Bitmap &b, const LemonGFX::Path &path, double dpmm)
	{
		(void) dpmm;

		LemonGFX::SimpleHalftoner halftoner(b.getHalftoningCurve());
		enum LemonGFX::Fill::FillType fillType = path.getFillType();
		const uint16_t intensity = halftoner.getBitPatternForLuminance(static_cast<unsigned char>(255.0 - path.getFillShade() * 255.0));
		std::vector<Edge> edgeTable = getEdgeTable(path);
	
		for (size_t y = 0; y < b.getHeight(); y++)
		{
			std::vector<ActiveEdge> activeEdges = getActiveEdges(edgeTable, y);
	
			if (activeEdges.size() == 0)
				continue;
	
			if (fillType == LemonGFX::Fill::FillType::kEvenOddFill)
			{
				for (size_t i = 0; i < activeEdges.size(); i += 2)
				{
					ActiveEdge::type onEdge = activeEdges[i].x;
					ActiveEdge::type offEdge = activeEdges[i + 1].x;
					
					fillRange(b, y, onEdge, offEdge, intensity);
				}
			}
			else if (fillType == LemonGFX::Fill::FillType::kNonzeroWindingFill)
			{
				int directionSum = 0;
				int onEdgeIndex = -1;
	
				for (size_t i = 0; i < activeEdges.size(); i++)
				{
					if (directionSum == 0)
						onEdgeIndex = i;
					
					directionSum += activeEdges[i].direction;
					if (directionSum == 0)
					{
						if (onEdgeIndex >= 0)
						{
							ActiveEdge::type onEdge = activeEdges[onEdgeIndex].x;
							ActiveEdge::type offEdge = activeEdges[i].x;
							
							fillRange(b, y, onEdge, offEdge, intensity);
							onEdgeIndex = -1;
						}
					}
				}
			}
			else if (fillType == LemonGFX::Fill::FillType::kUnionFill)
			{
				for (size_t i = 0; i < activeEdges.size(); i++)
				{
					ActiveEdge::type onEdge = activeEdges[i].x;
					std::set<size_t> shapeStack { activeEdges[i].shapeIndex };
					
					while (++i < activeEdges.size())
					{
						auto it = shapeStack.find(activeEdges[i].shapeIndex);
						if (it != shapeStack.end())
							shapeStack.erase(it);
						else
							shapeStack.emplace(activeEdges[i].shapeIndex);
						
						if (shapeStack.size() == 0)
							break;
					}
	
					if (i >= activeEdges.size())
						i = activeEdges.size() - 1;
					
					ActiveEdge::type offEdge = activeEdges[i].x;
					fillRange(b, y, onEdge, offEdge, intensity);
				}
			}
		}
	}
}

LemonGFX::Bitmap LemonGFX::Draw::DrawPath(const Path &path, double dpmm)
{
	return DrawPaths(std::vector<Path> { path }, dpmm);
}

LemonGFX::Bitmap LemonGFX::Draw::DrawPaths(const std::vector<Path> &paths, double dpmm)
{
	if (paths.size() == 0)
		return Bitmap::InvalidBitmap();
	
	RectangleF globalBounds = paths[0].getBounds();

	std::function<double(double)> toDeviceSpace = [dpmm](double value)
	{
		return value * dpmm / 10.0;
	};

	for (std::vector<Path>::size_type i = 1; i < paths.size(); i++)
	{
		RectangleF thisBounds = paths[i].getBounds();
		double strokeWidthOffset = toDeviceSpace(paths[i].getStrokeWidth()) / 2.0;
		globalBounds.lowerLeft.x = std::min(thisBounds.lowerLeft.x - strokeWidthOffset, globalBounds.lowerLeft.x);
		globalBounds.lowerLeft.y = std::min(thisBounds.lowerLeft.y - strokeWidthOffset, globalBounds.lowerLeft.y);
		globalBounds.upperRight.x = std::max(thisBounds.upperRight.x + strokeWidthOffset, globalBounds.upperRight.x);
		globalBounds.upperRight.y = std::max(thisBounds.upperRight.y + strokeWidthOffset, globalBounds.upperRight.y);
	}

	PointU lowerLeft = static_cast<PointU>(globalBounds.lowerLeft);
	unsigned offsetX = static_cast<unsigned>(globalBounds.lowerLeft.x) % 8;

	double bitmapWidth = toDeviceSpace(std::ceil(globalBounds.getWidth() + offsetX));
	double bitmapHeight = toDeviceSpace(std::ceil(globalBounds.getHeight()));

	Bitmap b = Bitmap(static_cast<size_t>(bitmapWidth), static_cast<size_t>(bitmapHeight));
	if (!b.isValidSize())
		return Bitmap::InvalidBitmap();
	
	b.setPosition(PointU { lowerLeft.x - offsetX, lowerLeft.y });
	
	DrawPathsOnBitmap(b, paths, dpmm);
	
	return b;
}

void LemonGFX::Draw::DrawPathsOnBitmap(Bitmap &bmp, const std::vector<Path> &paths, double dpmm)
{
	PointF position = static_cast<PointF>(bmp.getPosition());
	RectangleF globalBounds(position.x, position.y, bmp.getWidth(), bmp.getHeight());
	std::function<double(double)> toDeviceSpace = [dpmm](double value)
	{
		return value * dpmm / 10.0;
	};

	for (const Path &path : paths)
	{
		Path preparedPath = preparePathForDevice(path, globalBounds, toDeviceSpace);
		if (path.isFilled())
			fillOnBitmap(bmp, preparedPath, dpmm);
		if (path.getStrokeWidth() > 0.0)
			fillOnBitmap(bmp, LemonGFX::Stroke::GetStrokePath(preparedPath), dpmm);
	}
}
