#include <libcore/endian.h>
#include <libcore/streams.h>
#include <lemonsqueezy/deflate.h>
#include <lemonsqueezy/jpeg-dct.h>
#include <lemonsqueezy/lzw.h>
#include <lemonsqueezy/tiffprediction.h>
#include <algorithm>
#include <functional>
#include "image-tiff.h"

// TIFF file format
namespace
{
	// Intel/Motorola byte order terminology used by TIFF standards
	enum TIFFByteOrderMark : uint16_t
	{
		kIntelByteOrder = 0x4949,   // "II"
		kMotorolaByteOrder = 0x4D4D // "MM"
	};

	enum TIFFResolutionUnit : uint16_t
	{
		kNoUnit = 1,
		kInches = 2,
		kCentimetres = 3
	};

	enum TIFFTagID : uint16_t
	{
		kNewSubfileType = 254,
		kImageWidth = 256,
		kImageHeight = 257,
		kBitsPerSample = 258,
		kCompression = 259,
		kPhotometricInterpretation = 262,
		kStripOffsets = 273,
		kSamplesPerPixel = 277,
		kRowsPerStrip = 278,
		kStripByteCounts = 279,
		kXResolution = 282,
		kYResolution = 283,
		kResolutionUnit = 296,
		kPredictor = 317
	};

	enum TIFFDataType : uint16_t
	{
		kByte = 1,
		kASCII = 2,
		kShort = 3,
		kLong = 4,
		kRational = 5
	};

	struct TIFFTagData
	{
		TIFFTagData(uint16_t type, uint32_t count, uint32_t offset)
			: type(type), count(count), offset(offset) {}
		uint16_t type;
		uint32_t count;
		uint32_t offset;
	};

	inline size_t getBytesPerRow(const LemonGFX::TIFFFileInfo &tiff)
	{
		switch (tiff.bitsPerSample.front())
		{
			case 1:
				return (tiff.samplesPerPixel * tiff.width + 7) / 8;
			case 2:
				return (tiff.samplesPerPixel * tiff.width + 3) / 4;
			case 4:
				return (tiff.samplesPerPixel * tiff.width + 1) / 2;
			case 8:
				return (tiff.samplesPerPixel * tiff.width);
			case 16:
				return (tiff.samplesPerPixel * tiff.width * 2);
			default:
				throw LemonGFX::ImageException("Unsupported bit-depth for this TIFF file");
		}
	}

	template<typename T>
	void writeToStream(std::ostream &o, const T &val)
	{
		o.write(reinterpret_cast<const char *>(&val), sizeof val);
	}

	template<typename T>
	T readFromStream(std::istream &i)
	{
		T val;
		if (i.read(reinterpret_cast<char *>(&val), sizeof val))
			return val;
		
		throw LemonGFX::ImageException("Read error reading from TIFF");
	}

	template<typename T>
	bool tryGetTagValues(std::istream &in, enum TIFFByteOrderMark bo, const std::map<enum TIFFTagID, TIFFTagData> &tags, enum TIFFTagID tagID, std::vector<T> &values)
	{
		values.clear();

		auto tag = tags.find(tagID);
		if (tag == tags.end())
			return false;
		
		if (tag->second.count == 1 && tag->second.type != kRational)
		{
			values.emplace_back(tag->second.offset);
		}
		else
		{
			in.seekg(tag->second.offset);
			
			// rationals are just two longs
			uint32_t limit = tag->second.count;
			if (tag->second.type == TIFFDataType::kRational)
				limit *= 2;
			
			for (uint32_t i = 0; i < limit; i++)
			{
				// we can't use templates at runtime, but the decisions made here
				// cannot be made at compile time.

				T value;
				switch (tag->second.type)
				{
					case TIFFDataType::kByte:
						value = static_cast<T>(readFromStream<uint8_t>(in));
						break;
					
					case TIFFDataType::kShort:
						if (bo == kIntelByteOrder)
							value = static_cast<T>(StripedHorse::FromLittleEndian(readFromStream<uint16_t>(in)));
						else
							value = static_cast<T>(StripedHorse::FromBigEndian(readFromStream<uint16_t>(in)));
						break;
					
					case TIFFDataType::kRational:
					case TIFFDataType::kLong:
						if (bo == kIntelByteOrder)
							value = static_cast<T>(StripedHorse::FromLittleEndian(readFromStream<uint32_t>(in)));
						else
							value = static_cast<T>(StripedHorse::FromBigEndian(readFromStream<uint32_t>(in)));
						break;
					
					default:
						throw LemonGFX::ImageException("Cannot read fields of this type");
				}

				values.emplace_back(value);
			}
		}

		return true;
	}

	template<typename T>
	bool tryGetTagValue(const std::map<enum TIFFTagID, TIFFTagData> &tags, enum TIFFTagID tagID, T &val)
	{
		auto tag = tags.find(tagID);
		if (tag == tags.end() || tag->second.count != 1)
			return false;
		
		val = tag->second.offset;
		return true;
	}

	template<typename T>
	void confirmTagValue(const std::map<enum TIFFTagID, TIFFTagData> &tags, enum TIFFTagID tagID, const T &expectedValue)
	{
		T valueFromTable;
		if (!tryGetTagValue(tags, tagID, valueFromTable))
			throw LemonGFX::ImageException("TIFF file tag missing or unusable " + std::to_string(tagID));
		
		if (valueFromTable != expectedValue)
			throw LemonGFX::ImageException("TIFF file does not match expected format");
	}

	size_t compress(LemonSqueezy::ICompressor &compressor, const std::vector<unsigned char> &inputData, std::vector<unsigned char> &outputData)
	{
		StripedHorse::BasicInputVectorStreamBuffer<> inputBuffer(inputData);
		StripedHorse::BasicOutputVectorStreamBuffer<> outputBuffer(outputData);
		std::istream inStream(&inputBuffer);
		std::ostream outStream(&outputBuffer);
		compressor.write(inStream, outStream);
		return outputData.size();
	}

	size_t decompress(LemonSqueezy::IDecompressor &decompressor, const std::vector<unsigned char> &inputData, std::vector<unsigned char> &outputData)
	{
		StripedHorse::BasicInputVectorStreamBuffer<> inputBuffer(inputData);
		StripedHorse::BasicOutputVectorStreamBuffer<> outputBuffer(outputData);
		std::istream inStream(&inputBuffer);
		std::ostream outStream(&outputBuffer);
		decompressor.read(inStream, outStream);
		return outputData.size();
	}
	
	LemonGFX::Buffer readRawStripData(std::istream &in, const LemonGFX::TIFFFileInfo &tiff, size_t stripIndex)
	{
		uint32_t bytesToRead = tiff.stripByteCounts[stripIndex];
		std::vector<unsigned char> thisData(bytesToRead);

		// seek and read the data
		in.seekg(tiff.stripOffsets[stripIndex]);
		in.read(reinterpret_cast<char *>(&thisData[0]), bytesToRead);

		// gotta bomb out if we can't read the data
		if (static_cast<uint32_t>(in.gcount()) != bytesToRead)
			throw LemonGFX::ImageException("Unable to read required number of bytes");
		return thisData;
	}

	void writeToCallback(const std::vector<unsigned char> &data, size_t bytesPerRow, LemonGFX::ImageIO::ScanlineCallback callback)
	{
		if (data.size() % bytesPerRow != 0)
			throw LemonGFX::ImageException("Data in this TIFF file is not stored correctly");
		
		for (size_t i = 0; i < data.size(); i += bytesPerRow)
			callback(data.data() + i, bytesPerRow);
	}

	void readStripData(std::istream &in, const LemonGFX::TIFFFileInfo &tiff, size_t stripIndex, LemonGFX::ImageIO::ScanlineCallback callback)
	{
		const size_t bytesPerRow = getBytesPerRow(tiff);
		
		LemonGFX::Buffer thisData = readRawStripData(in, tiff, stripIndex);
		LemonGFX::Buffer decompressed;

		switch (tiff.compression)
		{
			case LemonGFX::TIFFCompression::kNoCompression:
			{
				writeToCallback(thisData, bytesPerRow, callback);
				return; // skip prediction
			}

			case LemonGFX::TIFFCompression::kLZW:
			{
				LemonSqueezy::LZWCompression lzwDecompressor(LemonSqueezy::LZWCompression::TablingSystem::kOffByOne);
				decompress(lzwDecompressor, thisData, decompressed);
				break;
			}

			case LemonGFX::TIFFCompression::kDeflate:
			{
				LemonSqueezy::DeflateCompression deflateDecompressor;
				decompress(deflateDecompressor, thisData, decompressed);
				break;
			}

			case LemonGFX::TIFFCompression::kJPEG:
			{
				LemonSqueezy::JPEGCompression jpegDecompressor;
				StripedHorse::BasicInputVectorStreamBuffer<> inBuffer(thisData);
				std::istream inStream(&inBuffer);
				jpegDecompressor.read(inStream, callback);
				return; // skip prediction
			}

			default:
			{
				throw LemonGFX::ImageException("Unknown TIFF compression method used");
			}
		}

		switch (tiff.prediction)
		{
			case 1:
			{
				// no prediction
				writeToCallback(decompressed, bytesPerRow, callback);
				break;
			}
			
			case 2:
			{
				// horizontal differencing
				std::vector<unsigned char> predicted;
				LemonSqueezy::TIFFPrediction predictor(tiff.samplesPerPixel, tiff.bitsPerSample.front(), tiff.width);
				decompress(predictor, decompressed, predicted);
				writeToCallback(predicted, bytesPerRow, callback);
				break;
			}
			
			default:
				throw LemonGFX::ImageException("Unknown prediction method used in TIFF");
		}
	}
}

LemonGFX::TIFFFileInfo::TIFFFileInfo()
{
}

LemonGFX::TIFFFileInfo::TIFFFileInfo(std::istream &in)
{
	using ImageException = LemonGFX::ImageException;

	enum TIFFByteOrderMark orderMark;
	in.read(reinterpret_cast<char *>(&orderMark), sizeof orderMark);

	std::function<uint16_t()> readUInt16;
	std::function<uint32_t()> readUInt32;

	if (orderMark == kIntelByteOrder)
	{
		readUInt16 = [&](){ return StripedHorse::FromLittleEndian(readFromStream<uint16_t>(in)); };
		readUInt32 = [&](){ return StripedHorse::FromLittleEndian(readFromStream<uint32_t>(in)); };
	}
	else if (orderMark == kMotorolaByteOrder)
	{
		readUInt16 = [&](){ return StripedHorse::FromBigEndian(readFromStream<uint16_t>(in)); };
		readUInt32 = [&](){ return StripedHorse::FromBigEndian(readFromStream<uint32_t>(in)); };
	}
	else
	{
		throw LemonGFX::ImageException("Not a valid TIFF file");
	}

	uint16_t version = readUInt16();
	if (version != 0x2A)
		throw LemonGFX::ImageException("Unknown TIFF version");
	
	std::map<enum TIFFTagID, TIFFTagData> allTags;
	uint32_t headerOffset;

	do
	{
		headerOffset = readUInt32();
		in.seekg(headerOffset);

		uint16_t numberOfTags = readUInt16();
		for (uint16_t i = 0; i < numberOfTags; i++)
		{
			TIFFTagID tagID = static_cast<enum TIFFTagID>(readUInt16());
			TIFFTagData data(0, 0, 0);
			data.type = readUInt16();
			data.count = readUInt32();

			// if "count" is 1 and the value can fit within the offset field,
			// then the offset is considered the value of the field (aligned
			// to the left)

			if (data.count > 1)
			{
				data.offset = readUInt32();
			}
			else
			{
				switch (data.type)
				{
					case TIFFDataType::kByte:
						data.offset = static_cast<uint8_t>(in.get());
						in.seekg(3, std::ios::cur);
						break;
					
					case TIFFDataType::kShort:
						data.offset = readUInt16();
						in.seekg(2, std::ios::cur);
						break;
					
					default:
						data.offset = readUInt32();
				}
			}

			allTags.emplace(tagID, data);
		}
		
		headerOffset = readUInt32();
	}
	while (headerOffset != 0);
	
	if (!tryGetTagValue(allTags, kPredictor, prediction))
		prediction = 1;

	if (!tryGetTagValue(allTags, kCompression, compression))
		throw ImageException("Cannot determine compression of this TIFF");

	if (!(tryGetTagValue(allTags, kImageWidth, width) && tryGetTagValue(allTags, kImageHeight, height)))
		throw ImageException("TIFF dimensions are invalid");
	
	if (!tryGetTagValue(allTags, kSamplesPerPixel, samplesPerPixel))
		throw ImageException("File does not specify number of samples per pixel");
	
	if (!tryGetTagValues(in, orderMark, allTags, kBitsPerSample, bitsPerSample))
		throw ImageException("Cannot find information about bits per sample");
	
	if (bitsPerSample.size() != samplesPerPixel || bitsPerSample.size() == 0)
		throw ImageException("Number of bits per sample fields does not match number of samples per pixel");

	if (!std::all_of(bitsPerSample.begin(), bitsPerSample.end(), [&](uint32_t val) { return val == bitsPerSample.front(); }))
		throw ImageException("Variably-sized bit-depths are not supported");

	if(!tryGetTagValue(allTags, kPhotometricInterpretation, photometricInterpretation))
		throw ImageException("TIFF file does not contain a photometric interpretation");
	
	if (!tryGetTagValues(in, orderMark, allTags, kStripOffsets, stripOffsets))
		throw ImageException("Unable to get strip offsets");
	
	if (!tryGetTagValues(in, orderMark, allTags, kStripByteCounts, stripByteCounts))
		throw ImageException("Unable to get strip byte counts");
	
	if (!tryGetTagValue(allTags, kRowsPerStrip, rowsPerStrip))
		throw ImageException("Could not find strip information");
	
	if (stripOffsets.size() != stripByteCounts.size())
		throw ImageException("Strip offsets and strip byte counts are different sizes");
	
	std::vector<uint32_t> rational;
	uint32_t resolutionUnit = 0;
	if (!tryGetTagValues(in, orderMark, allTags, kXResolution, rational) || rational.size() != 2)
		throw ImageException("This TIFF file does not contain the required x-resolution information");	
	xResolution = static_cast<double>(rational[0]) / rational[1];

	if (!tryGetTagValues(in, orderMark, allTags, kYResolution, rational) || rational.size() != 2)
		throw ImageException("This TIFF file does not contain the required y-resolution information");
	yResolution = static_cast<double>(rational[0]) / rational[1];

	if (!tryGetTagValue(allTags, kResolutionUnit, resolutionUnit))
		throw ImageException("This TIFF file does not contain the resolution unit of measurement");
	
	switch (resolutionUnit)
	{
		case TIFFResolutionUnit::kNoUnit:
		case TIFFResolutionUnit::kInches:
			// nothing to do
			break;
		
		case TIFFResolutionUnit::kCentimetres:
			// convert to DPI
			xResolution *= 2.54;
			yResolution *= 2.54;
			break;
		
		default:
			throw ImageException("Unknown unit of measurement used in TIFF file");
	}
}

LemonGFX::Buffer LemonGFX::ImageIO::ReadStripDataFromTIFFFormat(std::istream &in, bool decompress, LemonGFX::TIFFFileInfo *info)
{
	TIFFFileInfo tiff(in);
	if (info)
		*info = tiff;
	
	return ReadStripDataFromTIFFFormat(in, tiff, decompress);
}


LemonGFX::Buffer LemonGFX::ImageIO::ReadStripDataFromTIFFFormat(std::istream &in, const TIFFFileInfo &tiff, bool decompress)
{
	LemonGFX::Buffer result;

	if (decompress)
	{
		for (size_t i = 0; i < tiff.stripOffsets.size(); i++)
		{
			readStripData(in, tiff, i, [&result](const unsigned char *newData, size_t length){
				result.insert(result.end(), newData, newData + length);
			});
		}
	}
	else
	{
		for (size_t i = 0; i < tiff.stripOffsets.size(); i++)
		{
			LemonGFX::Buffer thisData = readRawStripData(in, tiff, i);
			result.insert(result.end(), thisData.begin(), thisData.end());
		}
	}

	return result;
}

void LemonGFX::ImageIO::ReadStripDataFromTIFFFormat(std::istream &in, ScanlineCallback callback, TIFFFileInfo *info)
{
	TIFFFileInfo tiff(in);
	if (info)
		*info = tiff;
	
	return ReadStripDataFromTIFFFormat(in, tiff, callback);
}

void LemonGFX::ImageIO::ReadStripDataFromTIFFFormat(std::istream &in, const TIFFFileInfo &tiff, ScanlineCallback callback)
{
	for (size_t i = 0; i < tiff.stripOffsets.size(); i++)
		readStripData(in, tiff, i, callback);
}

LemonGFX::Image LemonGFX::ImageIO::ReadFromTIFFFormat(std::istream &in, LemonGFX::TIFFFileInfo *info)
{
	TIFFFileInfo tiff(in);
	if (info)
		*info = tiff;
	
	LemonGFX::Buffer data;
	for (size_t i = 0; i < tiff.stripOffsets.size(); i++)
	{
		readStripData(in, tiff, i, [&data](const unsigned char *newData, size_t length){
			data.insert(data.end(), newData, newData + length);
		});
	}

	LemonGFX::Image image = LemonGFX::Image(tiff.bitsPerSample.front(), tiff.samplesPerPixel, tiff.width, tiff.height, data);
	image.setResolution(tiff.xResolution, tiff.yResolution);
	return image;
}

void LemonGFX::ImageIO::WriteToTIFFFormat(const LemonGFX::Image &img, std::ostream &out, enum TIFFCompression compression)
{
	if (StripedHorse::GetHostEndianness() == StripedHorse::Endian::kBigEndian)
		writeToStream(out, kMotorolaByteOrder);
	else
		writeToStream(out, kIntelByteOrder);
	
	writeToStream(out, static_cast<uint16_t>(0x2A)); // TIFF version number
	writeToStream(out, static_cast<uint32_t>(0x00000008)); // Offset to first (and only) IFD

	// default is "bilevel" (1 is white)
	// photometricInterpretation values according to TIFF
	//  0  : 1 bits considered black
	//  1  : 1 bits considered white
	//  2  : RGB
	//  3  : palette
	//  4  : mask
	//  5  : separated colour (typically CMYK)

	uint32_t photometricInterpretation = 1;
	if (img.getSamplesPerPixel() == 1)
		photometricInterpretation = img.blackIs1() ? 0 : 1;
	else if (img.getSamplesPerPixel() == 3)
		photometricInterpretation = 2; // assume RGB
	else if (img.getSamplesPerPixel() == 4)
		photometricInterpretation = 5; // assume CMYK

	std::map<enum TIFFTagID, TIFFTagData> allTags {
		{ kNewSubfileType,            { kLong,     1, 0 } }, // no specific filetype flags
		{ kImageWidth,                { kLong,     1, static_cast<uint32_t>(img.getWidth()) } },
		{ kImageHeight,               { kLong,     1, static_cast<uint32_t>(img.getHeight()) } },
		{ kBitsPerSample,             { kShort,    1, 0xDEADBEEF } }, // filted later
		{ kCompression,               { kShort,    1, static_cast<uint32_t>(compression) } },
		{ kPhotometricInterpretation, { kShort,    1, photometricInterpretation } },
		{ kStripOffsets,              { kLong,     1, 0xDEADBEEF } }, // filled later
		{ kSamplesPerPixel,           { kShort,    1, static_cast<uint32_t>(img.getSamplesPerPixel()) } },
		{ kRowsPerStrip,              { kLong,     1, static_cast<uint32_t>(img.getHeight()) } }, // put all data in one strip
		{ kStripByteCounts,           { kLong,     1, 0xDEADBEEF } }, // strip contains all bytes, but needs to wait until after compression
		{ kXResolution,               { kRational, 1, 0xDEADBEEF } }, // filled later
		{ kYResolution,               { kRational, 1, 0xDEADBEEF } }, // filled later
		{ kResolutionUnit,            { kShort,    1, TIFFResolutionUnit::kInches } }
	};

	writeToStream(out, static_cast<uint16_t>(allTags.size()));

	// 2 bytes for BOM, 2 bytes for version, plus 4 bytes for offset, plus 12 bytes for each tag
	const uint32_t sizeOfHeaderInformation = 2 + 2 + 4 + 2 + allTags.size() * 12 + 4;

	allTags.find(kXResolution)->second.offset = sizeOfHeaderInformation;
	allTags.find(kYResolution)->second.offset = sizeOfHeaderInformation + 8;

	// for single-channel images, the bits-per-sample is stored in the offset field, otherwise
	// there needs to be one "bits-per-sample" field for each channel, even if they are all the
	// same
	if (img.getSamplesPerPixel() == 1)
	{
		allTags.find(kBitsPerSample)->second.offset = img.getBitsPerSample();
		allTags.find(kStripOffsets)->second.offset = sizeOfHeaderInformation + 16;
	}
	else
	{
		allTags.find(kBitsPerSample)->second.count = img.getSamplesPerPixel();
		allTags.find(kBitsPerSample)->second.offset = sizeOfHeaderInformation + 16;
		allTags.find(kStripOffsets)->second.offset = sizeOfHeaderInformation + 16 + 2 * img.getSamplesPerPixel();
	}

	const std::vector<unsigned char> &inputData = img.getData();
	std::vector<unsigned char> outputData;

	size_t dataSize = 0;
	switch (compression)
	{
		case kNoCompression:
		{
			dataSize = img.getDataSize();
			break;
		}
		
		case kLZW:
		{
			LemonSqueezy::LZWCompression lzwCompressor(LemonSqueezy::LZWCompression::TablingSystem::kOffByOne);
			dataSize = compress(lzwCompressor, inputData, outputData);
			break;
		}

		case kDeflate:
		{
			LemonSqueezy::DeflateCompression deflateCompressor;
			dataSize = compress(deflateCompressor, inputData, outputData);
			break;
		}

		default:
		{
			throw ImageException("Unsupported TIFF compression method for writing");
		}
	}

	allTags.find(kStripByteCounts)->second.offset = dataSize;

	for (const std::pair<enum TIFFTagID, TIFFTagData> &tag : allTags)
	{
		writeToStream(out, tag.first);
		writeToStream(out, tag.second.type);
		writeToStream(out, tag.second.count);
		writeToStream(out, tag.second.offset);
	}

	writeToStream(out, static_cast<uint32_t>(0)); // pointer to next IFD (0 = no more IFDs)

	// x and y resolution
	writeToStream(out, static_cast<uint32_t>(img.getXResolution()));
	writeToStream(out, static_cast<uint32_t>(1));
	writeToStream(out, static_cast<uint32_t>(img.getYResolution()));
	writeToStream(out, static_cast<uint32_t>(1));

	// bits per sample (must be one for each channel), and only output it if there is more
	// than one channel (otherwise it is stored in the offset field above)
	if (img.getSamplesPerPixel() > 1)
		for (size_t i = 0; i < img.getSamplesPerPixel(); i++)
			writeToStream(out, static_cast<uint16_t>(img.getBitsPerSample()));

	if (compression == TIFFCompression::kNoCompression)
	{
		// output the image data directly (no copy)
		out.write(reinterpret_cast<const char *>(inputData.data()), dataSize);
	}
	else
	{
		// output the compressed output
		out.write(reinterpret_cast<const char *>(outputData.data()), dataSize);
	}
}

