#ifndef STRIPEDHORSE_IMAGEIO_TIFF_H
#define STRIPEDHORSE_IMAGEIO_TIFF_H 1

#include <functional>
#include <iostream>
#include "image.h"

namespace LemonGFX
{
	/** compression values (not all are supported by this library) */
	enum TIFFCompression : uint16_t
	{
		kNoCompression = 1,
		kCCITTModifiedHuffmanRLE = 2,
		kCCITT3 = 3,
		kCCITT4 = 4,
		kLZW = 5,
		kJPEG = 7,
		kDeflate = 8,
	};
	
	enum TIFFPhotometricInterpretation : uint16_t
	{
		kWhiteIsZero = 0,
		kBlackIsZero = 1,
		kRGB = 2,
		kPalette = 3,
		kTransparencyMask = 4,
		kSeparated = 5,
		kYCbCr = 6,
		kCIELab = 8,
		kCIELabAlternate = 9,
		kITULab = 10
	};
	
	/** Helper class */
	struct TIFFFileInfo
	{
		TIFFFileInfo();
		TIFFFileInfo(std::istream &in);
		
		uint32_t photometricInterpretation = 0;
		uint32_t samplesPerPixel = 0;
		uint32_t width = 0;
		uint32_t height = 0;
		uint32_t compression = 0;
		uint32_t prediction = 0;
		uint32_t rowsPerStrip = 0;
		double xResolution = 0.0;
		double yResolution = 0.0;
		std::vector<uint32_t> bitsPerSample;
		std::vector<uint32_t> stripOffsets;
		std::vector<uint32_t> stripByteCounts;
	};

	namespace ImageIO
	{
		using ScanlineCallback = std::function<void(const unsigned char *data, size_t length)>;
		
		/**
		 * Reads the strip data from a TIFF file
		 * @param[in] in the stream containing the TIFF formatted file
		 * @param[in] decompress a boolean determining whether to read the strip data as-is, or to decompress it
		 * @param[out] info an optional argument that can receive the TIFF header information
		 */
		LemonGFX::Buffer ReadStripDataFromTIFFFormat(std::istream &in, bool decompress, TIFFFileInfo *info = nullptr);

		/**
		 * Reads the strip data from a TIFF file where the header has already been read
		 * @param[in] in the stream containing the TIFF formatted file, with the header already read
		 * @param[in] info the previously-read TIFF header
		 * @param[in] decompress a boolean determining whether to read the strip data as-is, or to decompress it
		 */
		LemonGFX::Buffer ReadStripDataFromTIFFFormat(std::istream &in, const TIFFFileInfo &info, bool decompress);

		/**
		 * Reads the strip data from a TIFF file, calling back for each scanline in the image
		 * @param[in] in the stream containing the TIFF formatted file
		 * @param[in] callback the function to call back for each row of samples in the image
		 * @param[out] info an optional argument that can receive the TIFF header information
		 */
		void ReadStripDataFromTIFFFormat(std::istream &in, ScanlineCallback callback, TIFFFileInfo *info = nullptr);

		/**
		 * Reads the strip data from a TIFF file where the header has already been read, calling back for each scanline in the image
		 * @param[in] in the stream containing the TIFF formatted file, with the header already read
		 * @param[in] info the previously-read TIFF header
		 * @param[in] callback the function to call back for each row of samples in the image
		 */
		void ReadStripDataFromTIFFFormat(std::istream &in, const TIFFFileInfo &info, ScanlineCallback callback);

		/**
		 * Reads a TIFF file into a LemonGFX::Image
		 * @param[in] in the stream containing the TIFF formatted file
		 * @param[out] info an optional argument that can receive the TIFF header information
		 */
		LemonGFX::Image ReadFromTIFFFormat(std::istream &in, TIFFFileInfo *info = nullptr);

		/**
		 * Writes a LemonGFX::Image to a TIFF container
		 * @param[in] img the image to write
		 * @param[in] out the stream to write the TIFF file to
		 * @param[in] compression the compression scheme to use
		 */
		void WriteToTIFFFormat(const LemonGFX::Image &img, std::ostream &out, enum TIFFCompression compression = TIFFCompression::kNoCompression);
	}
}

#endif
