#include <climits>
#include "image.h"
#include "image-tiff.h"

namespace
{
	size_t calculateBytesPerRow(size_t bitsPerSample, size_t samplesPerPixel, size_t pixelsPerRow)
	{
		size_t bitsPerRow = bitsPerSample * samplesPerPixel * pixelsPerRow;
		if (bitsPerRow == 0)
			return 0;
		else
			return ((((bitsPerRow - 1) >> 3) + 1) << 3) / CHAR_BIT;
	}
}

LemonGFX::Image::Image()
  : m_bitsPerSample(0), m_samplesPerPixel(0), m_width(0), m_height(0), m_xResolution(96), m_yResolution(96)
{
}

LemonGFX::Image::Image(size_t bps, size_t spp, size_t width, size_t height)
  : m_bitsPerSample(bps), m_samplesPerPixel(spp), m_width(width), m_height(height),
    m_xResolution(96), m_yResolution(96), m_bytesPerRow(calculateBytesPerRow(bps, spp, width))
{
	m_data.resize(m_bytesPerRow * m_height);
}

LemonGFX::Image::Image(size_t bps, size_t spp, size_t width, size_t height, const Buffer &data)
  : m_bitsPerSample(bps), m_samplesPerPixel(spp), m_width(width), m_height(height),
    m_xResolution(96), m_yResolution(96), m_data(data), m_bytesPerRow(calculateBytesPerRow(bps, spp, width))
{
}

LemonGFX::Image::Image(size_t bps, size_t spp, size_t width, size_t height, Buffer &&data)
  : m_bitsPerSample(bps), m_samplesPerPixel(spp), m_width(width), m_height(height), 
    m_xResolution(96), m_yResolution(96), m_data(std::move(data)),
	m_bytesPerRow(calculateBytesPerRow(bps, spp, width))
{
}

LemonGFX::Image::Image(const Image &other)
  : m_bitsPerSample(other.m_bitsPerSample), m_samplesPerPixel(other.m_samplesPerPixel),
    m_width(other.m_width), m_height(other.m_height), m_xResolution(other.m_xResolution),
	m_yResolution(other.m_yResolution),m_data(other.m_data), m_bytesPerRow(other.m_bytesPerRow)
{
}

LemonGFX::Image::Image(Image &&other)
  : m_bitsPerSample(other.m_bitsPerSample), m_samplesPerPixel(other.m_samplesPerPixel),
    m_width(other.m_width), m_height(other.m_height), m_xResolution(other.m_xResolution),
	m_yResolution(other.m_yResolution), m_data(std::move(other.m_data)),
	m_bytesPerRow(other.m_bytesPerRow)
{
}

LemonGFX::Image &LemonGFX::Image::operator=(const Image &other)
{
	m_bitsPerSample = other.m_bitsPerSample;
	m_samplesPerPixel = other.m_samplesPerPixel;
	m_width = other.m_width;
	m_height = other.m_height;
	m_data = other.m_data;
	m_bytesPerRow = other.m_bytesPerRow;
	m_xResolution = other.m_xResolution;
	m_yResolution = other.m_yResolution;
	return *this;
}

bool LemonGFX::Image::blackIs1() const
{
	// For Images, 1 is considered white
	return false;
}

size_t LemonGFX::Image::getBitsPerSample() const
{
	return m_bitsPerSample;
}

size_t LemonGFX::Image::getBytesPerRow() const
{
	return m_bytesPerRow;
}

const LemonGFX::Buffer &LemonGFX::Image::getData() const
{
	return m_data;
}

LemonGFX::Buffer &LemonGFX::Image::getData()
{
	return m_data;
}

size_t LemonGFX::Image::getDataSize() const
{
	return m_data.size();
}

size_t LemonGFX::Image::getHeight() const
{
	return m_height;
}

size_t LemonGFX::Image::getSamplesPerPixel() const
{
	return m_samplesPerPixel;
}

size_t LemonGFX::Image::getXResolution() const
{
	return m_xResolution;
}

size_t LemonGFX::Image::getYResolution() const
{
	return m_yResolution;
}

size_t LemonGFX::Image::getWidth() const
{
	return m_width;
}

void LemonGFX::Image::invert()
{
	if (m_bitsPerSample != 8)
		throw ImageException("Unable to invert images if the number of bits per sample is not 8");
	
	for (unsigned char &sample : m_data)
		sample ^= 0xFFU;
}

bool LemonGFX::Image::isValidSize() const
{
	return m_width > 0 && m_height > 0;
}

void LemonGFX::Image::read(std::istream &in)
{
	*this = std::move(ImageIO::ReadFromTIFFFormat(in));
}

void LemonGFX::Image::reserve(size_t size)
{
	m_data.reserve(size);
}

void LemonGFX::Image::setResolution(size_t x, size_t y)
{
	m_xResolution = x;
	m_yResolution = y;
}

void LemonGFX::Image::write(Format f, std::ostream &out) const
{
	switch (f)
	{
		case Format::kTIFF:
			ImageIO::WriteToTIFFFormat(*this, out);
			break;
		
		default:
			throw ImageException("Unsupported format");
	}
}
