#include <algorithm>
#include <cmath>
#include <set>
#include "path.h"

LemonGFX::Path::Path()
  : m_isFilled(true), m_fillShade(1.0), m_fillType(Fill::FillType::kEvenOddFill), m_strokeShade(1.0), m_strokeWidth(0.0)
{
}

LemonGFX::Path::Path(const Path &other)
  : m_subpaths(other.m_subpaths), m_isFilled(other.m_isFilled), m_fillShade(other.m_fillShade),
    m_fillType(other.m_fillType), m_strokeShade(other.m_strokeShade), m_strokeWidth(other.m_strokeWidth)
{
}

LemonGFX::Path::Path(Path &&other)
  : m_subpaths(std::move(other.m_subpaths)), m_isFilled(other.m_isFilled), m_fillShade(other.m_fillShade),
    m_fillType(other.m_fillType), m_strokeShade(other.m_strokeShade), m_strokeWidth(other.m_strokeWidth)
{
}

LemonGFX::Path::SubpathCollection::const_iterator LemonGFX::Path::begin() const
{
	return m_subpaths.begin();
}

LemonGFX::Path &LemonGFX::Path::operator=(const Path &other)
{
	m_subpaths = other.m_subpaths;
	m_isFilled = other.m_isFilled;
	m_fillShade = other.m_fillShade;
	m_fillType = other.m_fillType;
	m_strokeShade = other.m_strokeShade;
	m_strokeWidth = other.m_strokeWidth;
	return *this;
}

LemonGFX::Path::Subpath &LemonGFX::Path::operator[](SubpathCollection::size_type i)
{
	return m_subpaths.at(i);
}

const LemonGFX::Path::Subpath &LemonGFX::Path::operator[](SubpathCollection::size_type i) const
{
	return m_subpaths.at(i);
}

void LemonGFX::Path::addSubpath(const Subpath &subpath)
{
	m_subpaths.push_back(subpath);
}

LemonGFX::Path::Subpath &LemonGFX::Path::back()
{
	return m_subpaths.back();
}

const LemonGFX::Path::Subpath &LemonGFX::Path::back() const
{
	return m_subpaths.back();
}

void LemonGFX::Path::clearSubpaths()
{
	m_subpaths.clear();
}

void LemonGFX::Path::emplaceSubpath(Subpath &&subpath)
{
	m_subpaths.emplace_back(subpath);
}

LemonGFX::Path::SubpathCollection::const_iterator LemonGFX::Path::end() const
{
	return m_subpaths.end();
}

LemonGFX::Path::Subpath &LemonGFX::Path::front()
{
	return m_subpaths.front();
}

const LemonGFX::Path::Subpath &LemonGFX::Path::front() const
{
	return m_subpaths.front();
}

StripedHorse::RectangleF LemonGFX::Path::getBounds() const
{
	StripedHorse::PointF lowerLeft;
	StripedHorse::PointF upperRight;

	if (m_subpaths.size() == 0)
		throw StripedHorse::InvalidArgumentException("Invalid number of points in path");
	
	lowerLeft = m_subpaths.at(0).at(0);
	upperRight = m_subpaths.at(0).at(0);

	for (typename SubpathCollection::size_type i = 0; i < m_subpaths.size(); i++)
	{
		const Subpath &subpath = m_subpaths.at(i);
		if (subpath.size() == 0)
			throw StripedHorse::InvalidArgumentException("Invalid number of points in path");
		
		for (const StripedHorse::PointF &p : subpath)
		{
			lowerLeft.x = std::min(lowerLeft.x, p.x);
			lowerLeft.y = std::min(lowerLeft.y, p.y);
			upperRight.x = std::max(upperRight.x, p.x);
			upperRight.y = std::max(upperRight.y, p.y);
		}
	}

	return StripedHorse::RectangleF { lowerLeft, upperRight };
}

double LemonGFX::Path::getFillShade() const
{
	return m_fillShade;
}

enum LemonGFX::Fill::FillType LemonGFX::Path::getFillType() const
{
	return m_fillType;
}

LemonGFX::Path::SubpathCollection::size_type LemonGFX::Path::getSize() const
{
	return m_subpaths.size();
}

double LemonGFX::Path::getStrokeShade() const
{
	return m_strokeShade;
}

double LemonGFX::Path::getStrokeWidth() const
{
	return m_strokeWidth;
}
	
bool LemonGFX::Path::isDesiredShapeForTemplating() const
{
	if (m_subpaths.size() != 1)
		return false;
	
	const Subpath &subpath = m_subpaths[0];

	if (!(subpath.size() == 4 || (subpath.size() == 5 && subpath.front() == subpath.back())))
		return false;
	
	const double epsilon = 0.01;
	double hypotenuse1 = std::sqrt(std::pow(subpath[2].x - subpath[0].x, 2.0) + std::pow(subpath[2].y - subpath[0].y, 2.0));
	double hypotenuse2 = std::sqrt(std::pow(subpath[1].x - subpath[3].x, 2.0) + std::pow(subpath[1].y - subpath[3].y, 2.0));

	return std::abs(std::abs(hypotenuse1) - std::abs(hypotenuse2)) < epsilon;
}

bool LemonGFX::Path::isFilled() const
{
	return m_isFilled;
}

void LemonGFX::Path::setFillShade(double fillShade)
{
	if (fillShade < 0.0)
		m_fillShade = 0.0;
	else if (fillShade > 1.0)
		m_fillShade = 1.0;
	else
		m_fillShade = fillShade;
}

void LemonGFX::Path::setFillType(enum Fill::FillType ft)
{
	m_fillType = ft;
}

void LemonGFX::Path::setIsFilled(bool isFilled)
{
	m_isFilled = isFilled;
}

void LemonGFX::Path::setStrokeShade(double strokeShade)
{
	if (strokeShade < 0.0)
		m_strokeShade = 0.0;
	else if (strokeShade > 1.0)
		m_strokeShade = 1.0;
	else
		m_strokeShade = strokeShade;
}

void LemonGFX::Path::setStrokeWidth(double strokeWidth)
{
	if (strokeWidth < 0.0)
		m_strokeWidth = 0.0;
	else
		m_strokeWidth = strokeWidth;
}
