#include <cmath>
#include "stroke.h"

namespace
{
	double getAngleOfLine(const StripedHorse::PointF &p1, const StripedHorse::PointF &p2)
	{
		return atan2(p2.y - p1.y, p2.x - p1.x);
	}

	LemonGFX::Path::Subpath thickenLine(const StripedHorse::PointF &p1, const StripedHorse::PointF &p2, double thickness)
	{
		LemonGFX::Path::Subpath result(4);
		const double overAngle = getAngleOfLine(p1, p2) + M_PI_2;
		const double underAngle = overAngle + M_PI;
		const double offset = thickness / 2.0;
		
		result[0].x = p1.x + (offset * cos(overAngle) - offset * sin(overAngle));
		result[0].y = p1.y + (offset * sin(overAngle) + offset * cos(overAngle));
		
		result[1].x = p2.x + (offset * cos(overAngle) - offset * sin(overAngle));
		result[1].y = p2.y + (offset * sin(overAngle) + offset * cos(overAngle));
		
		result[2].x = p2.x + (offset * cos(underAngle) - offset * sin(underAngle));
		result[2].y = p2.y + (offset * sin(underAngle) + offset * cos(underAngle));
		
		result[3].x = p1.x + (offset * cos(underAngle) - offset * sin(underAngle));
		result[3].y = p1.y + (offset * sin(underAngle) + offset * cos(underAngle));

		return result;
	}
}

LemonGFX::Path LemonGFX::Stroke::GetStrokePath(const LemonGFX::Path &path)
{
	std::vector<std::vector<StripedHorse::PointF>> strokeObject;
	
#ifndef USE_NEW_STROKING
	// thicken the line
	for (const auto &subpath : path)
	{
		for (size_t i = 1; i < subpath.size(); i++)
		{
			const PointF &lastPoint = subpath[i - 1];
			const PointF &thisPoint = subpath[i];

			if (lastPoint != thisPoint)
				strokeObject.emplace_back(thickenLine(lastPoint, thisPoint, path.getStrokeWidth()));
		}
	}

	// join the gaps
	Path finalStrokeObject;
	if (strokeObject.size() > 0)
	{
		finalStrokeObject.addSubpath(strokeObject.front());
		for (size_t i = 1; i < strokeObject.size(); i++)
		{
			const std::vector<StripedHorse::PointF> &lastStrokeElement = strokeObject[i - 1];
			const std::vector<StripedHorse::PointF> &thisStrokeElement = strokeObject[i];
			
			finalStrokeObject.addSubpath(std::vector<StripedHorse::PointF> { lastStrokeElement[1], thisStrokeElement[0], thisStrokeElement[3], lastStrokeElement[2] });
			finalStrokeObject.addSubpath(thisStrokeElement);
		}
	}

	finalStrokeObject.setFillShade(path.getStrokeShade());
	finalStrokeObject.setFillType(LemonGFX::Fill::FillType::kUnionFill);
#else
	Path finalStrokeObject;
	for (const auto &subpath : path)
	{
		Path::Subpath outerPath;
		Path::Subpath innerPath;

		for (size_t i = 1; i < subpath.size(); i++)
		{
			const PointF &lastPoint = subpath[i - 1];
			const PointF &thisPoint = subpath[i];

			if (thisPoint == lastPoint)
				continue;
			
			Path::Subpath thickenedPath = thickenLine(lastPoint, thisPoint, path.getStrokeWidth());
			outerPath.push_back(thickenedPath[0]);
			outerPath.push_back(thickenedPath[1]);
			innerPath.push_back(thickenedPath[2]);
			innerPath.push_back(thickenedPath[3]);
			lastPoint = thisPoint;
		}

		finalStrokeObject.addSubpath(outerPath);
		finalStrokeObject.addSubpath(innerPath);
	}

	finalStrokeObject.setFillShade(path.getStrokeShade());
	finalStrokeObject.setFillType(LemonGFX::Fill::FillType::kEvenOddFill);
#endif

	return finalStrokeObject;
}
