#include <cctype>
#include <climits>
#include <regex>
#include <sstream>
#include <unordered_map>
#include "cmap.h"
#include "libtokenisation.h"
#include "pdfexception.h"

namespace
{
	std::string getNextToken(std::istream &in)
	{
		char ch;
		while (in.get(ch) && ::isspace(ch))
			continue;
		
		switch (ch)
		{
			case '<':
				return LemonPDF::Tokenisation::TokeniseHexString(in);
			
			case '(':
				return LemonPDF::Tokenisation::TokeniseString(in);
			
			default:
				throw LemonPDF::PDFSyntaxException("Unexpected character while parsing CMap", in);
		}
	}


	// characters in str must be converted to unsigned char first,
	// otherwise if they are converted straight to char16_t or
	// char32_t, and the value of the char is negative, then -1
	// gets converted to 0xFFFF (or 0xFFFFFFFF) when it should
	// actually be converted to 0x00FF
	char32_t getGlyphIndexFromString(const std::string &str)
	{
		std::string glyphIndex = str;
		if (glyphIndex.size() > 4)
			throw LemonPDF::PDFParseException("Glyph index too large");
		
		glyphIndex.insert(0, 4 - glyphIndex.size(), '\0');
		return static_cast<char32_t>(static_cast<unsigned char>(glyphIndex.at(0))) << 24 |
		       static_cast<char32_t>(static_cast<unsigned char>(glyphIndex.at(1))) << 16 |
		       static_cast<char32_t>(static_cast<unsigned char>(glyphIndex.at(2))) << 8  |
		       static_cast<char32_t>(static_cast<unsigned char>(glyphIndex.at(3)));
	}

	char32_t getCharValueFromString(const std::string &str)
	{
		if (str.size() == 1)
		{
			return static_cast<char32_t>(static_cast<unsigned char>(str[0]));
		}
		else if (str.size() == 2)
		{
			unsigned char bytes[] {
				static_cast<unsigned char>(str.at(0)),
				static_cast<unsigned char>(str.at(1))
			};
			return { (static_cast<char32_t>(bytes[0]) << CHAR_BIT) | bytes[1] };			
		}
		else
		{
			throw LemonPDF::PDFParseException("Unsupported character size");
		}
	}

	std::vector<char32_t> getCharValuesFromString(const std::string &str)
	{
		// return the simple case
		if (str.size() == 1)
			return { getCharValueFromString(str) };

		if (str.size() % 2 == 1)
			throw LemonPDF::PDFParseException("Cannot handle odd number of bytes");

		// this string is composed of multiple code points, potentially containing
		// surrogate pairs.
		std::vector<char32_t> result;
		result.reserve(str.size() / 2);

		char32_t highSurrogate = 0;
		for (size_t i = 0; i < str.size(); i += 2)
		{
			char32_t thisChar = getCharValueFromString(str.substr(i, 2));

			if (highSurrogate != 0)
			{
				if ((thisChar & 0xFC00) != 0xDC00)
					throw LemonPDF::PDFParseException("Invalid surrogate pair encoding");
				
				result.push_back(0x10000 + ((highSurrogate - 0xD800) << 10) + (thisChar - 0xDC00));
				highSurrogate = 0;
			}
			else if ((thisChar & 0xFC00) == 0xD800)
			{
				highSurrogate = thisChar;
			}
			else
			{
				result.push_back(thisChar);
			}
		}

		return result;
	}

	bool parseCMap(std::istream &in, std::map<char32_t, std::string> &forward, std::map<std::string, char32_t> &reverse)
	{
		std::string line;

		std::regex startLineOneToOne("([0-9]+) beginbfchar");
		std::regex startLineRange("([0-9]+) beginbfrange");
		
		while (LemonPDF::Tokenisation::ReadLine(in, line))
		{
			std::smatch m;
			if (std::regex_match(line, m, startLineOneToOne))
			{
				while (LemonPDF::Tokenisation::ReadLine(in, line))
				{
					if (line == "endbfchar")
						break;
					 
					std::istringstream lineStream { line };
					std::string glyphIndexToken = getNextToken(lineStream);
					std::string charValue = getNextToken(lineStream);
					char32_t glyphIndex = getGlyphIndexFromString(glyphIndexToken);

					std::string newCharValue;
					for (char32_t wcharValue : getCharValuesFromString(charValue))
						newCharValue.append(LemonPDF::CMap::ToUTF8(wcharValue));

					forward.emplace(glyphIndex, newCharValue);
					reverse.emplace(newCharValue, glyphIndex);
				}
			}
			else if (std::regex_match(line, m, startLineRange))
			{
				while (LemonPDF::Tokenisation::ReadLine(in, line))
				{
					if (line == "endbfrange")
						break;
					
					std::istringstream lineStream { line };
					std::string beginGlyphToken = getNextToken(lineStream);
					std::string endGlyphToken = getNextToken(lineStream);
					std::string charValue = getNextToken(lineStream);
					char32_t beginGlyphIndex = getGlyphIndexFromString(beginGlyphToken);
					char32_t endGlyphIndex = getGlyphIndexFromString(endGlyphToken);
					char32_t wcharValue = getCharValueFromString(charValue);

					for (char32_t gi = beginGlyphIndex; gi <= endGlyphIndex; gi++)
					{
						char32_t ch = static_cast<char32_t>(wcharValue + gi - beginGlyphIndex);
						std::string newCharValue = LemonPDF::CMap::ToUTF8(ch);
						forward.emplace(gi, newCharValue);
						reverse.emplace(newCharValue, gi);
					}
				}
			}
		}

		return true;
	}

	template<typename CharT1, typename CharT2>
	std::string convertFromCMap(const std::basic_string<CharT1> &in, const std::map<CharT2, std::string> &map, const std::string &missing = "?")
	{
		std::ostringstream result;

		for (const CharT1 &ch : in)
		{
			auto i = map.find(ch);
			if (i != map.end())
				result << i->second;
			else
				result << missing;
		}

		return result.str();
	}
}

LemonPDF::CMap::CMap()
  : m_isKnownMapping(false)
{
}

// Private constructor, these mappings are known to be valid
LemonPDF::CMap::CMap(char16_t maxCode, char16_t offset, const std::vector<char16_t> &enc)
  : m_isKnownMapping(true)
{
	for (char16_t i = 0; i < offset; i++)
	{
		std::string utf8 = ToUTF8(i);
		m_forwardMapping.emplace(i, utf8);
		m_reverseMapping.emplace(utf8, i);
	}

	for (char16_t i = 0; i < enc.size(); i++)
	{
		std::string utf8 = ToUTF8(enc[i]);
		m_forwardMapping.emplace(offset + i, utf8);
		m_reverseMapping.emplace(utf8, offset + i);
	}
	
	for (char16_t i = offset + enc.size(); i < maxCode; i++)
	{
		std::string utf8 = ToUTF8(i);
		m_forwardMapping.emplace(i, utf8);
		m_reverseMapping.emplace(utf8, i);
	}
}

LemonPDF::CMap LemonPDF::CMap::FromPostscriptStyle(std::istream &in)
{
	CMap result;
	parseCMap(in, result.m_forwardMapping, result.m_reverseMapping);
	return result;
}

LemonPDF::CMap LemonPDF::CMap::MacRomanEncoding()
{
	static CMap s_macRomanEncoding(0x100, 0x80, { 
		0x00C4, 0x00C5, 0x00C7, 0x00C9, 0x00D1, 0x00D6, 0x00DC, 0x00E1, 0x00E0, 0x00E2, 0x00E4, 0x00E3, 0x00E5, 0x00E7, 0x00E9, 0x00E8,
		0x00EA, 0x00EB, 0x00ED, 0x00EC, 0x00EE, 0x00EF, 0x00F1, 0x00F3, 0x00F2, 0x00F4, 0x00F6, 0x00F5, 0x00FA, 0x00F9, 0x00DB, 0x00FC,
		0x2020, 0x00B0, 0x00A2, 0x00A3, 0x00A7, 0x2022, 0x00B6, 0x00DF, 0x00AE, 0x00A9, 0x2122, 0x00B4, 0x00A8, 0x2260, 0x00C6, 0x00D8,
		0x221E, 0x00B1, 0x2264, 0x2265, 0x00A5, 0x00B5, 0x2202, 0x2211, 0x220F, 0x03C0, 0x222B, 0x00AA, 0x00BA, 0x03A9, 0x00E6, 0x00F8,
		0x00BF, 0x00A1, 0x00AC, 0x221A, 0x0192, 0x2248, 0x2206, 0x00AB, 0x00BB, 0x2026, 0x00A0, 0x00C0, 0x00C3, 0x00D5, 0x0152, 0x0153,
		0x2013, 0x2014, 0x201C, 0x201D, 0x2018, 0x2019, 0x00F7, 0x25CA, 0x00FF, 0x0178, 0x2044, 0x20AC, 0x2039, 0x203A, 0xFB01, 0xFB02,
		0x2021, 0x00B7, 0x201A, 0x201E, 0x2030, 0x00C2, 0x00CA, 0x00C1, 0x00CB, 0x00C8, 0x00CD, 0x00CE, 0x00CF, 0x00CC, 0x00D3, 0x00D4,
		0xF8FF, 0x00D2, 0x00DA, 0x00DB, 0x00D9, 0x0131, 0x0236, 0x02DC, 0x00AD, 0x02D8, 0x02D9, 0x02DA, 0x00B8, 0x02DD, 0x02DB, 0x02C7
	});

	return s_macRomanEncoding;
}

LemonPDF::CMap LemonPDF::CMap::WinAnsiEncoding()
{
	static CMap s_winAnsiEncoding(0x100, 0x80, {
		0x20AC, 0x0000, 0x201A, 0x0192, 0x201E, 0x2026, 0x2020, 0x2021, 0x02C6, 0x2030, 0x0160, 0x2039, 0x0152, 0x0000, 0x017D, 0x0000,
		0x0000, 0x2018, 0x2019, 0x201C, 0x201D, 0x2022, 0x2013, 0x2014, 0x02DC, 0x2122, 0x0161, 0x203A, 0x0153, 0x0000, 0x017E, 0x0178,
	});

	return s_winAnsiEncoding;
}

std::string LemonPDF::CMap::ToUTF8(char32_t ch)
{
	if (ch <= 0xFFFF)
		return ToUTF8(static_cast<char16_t>(ch));

	std::string result;
	result.resize(4);

	result[0] = 0xF0 | ((ch >> 18) & 0x03);
	result[1] = 0xC0 | ((ch >> 12) & 0x1F);
	result[2] = 0xC0 | ((ch >> 6) & 0x1F);
	result[3] = 0x80 | ((ch >> 0) & 0x3F);

	return result;
}

std::string LemonPDF::CMap::ToUTF8(char16_t ch)
{
	std::string result;
	result.reserve(3);

	if (ch < 0x80)
	{
		result.push_back(static_cast<char>(ch));
	}
	else if (ch >= 0x80 && ch < 0x800)
	{
		result.push_back(0xC0 | ((ch >> 6) & 0x1F));
		result.push_back(0x80 | ((ch >> 0) & 0x3F));
	}
	else
	{
		result.push_back(0xE0 | ((ch >> 12) & 0x0F));
		result.push_back(0x80 | ((ch >>  6) & 0x3F));
		result.push_back(0x80 | ((ch >>  0) & 0x3F));
	}

	return result;
}

LemonPDF::CMap::CMap(const CMap &other)
  : m_forwardMapping(other.m_forwardMapping), m_reverseMapping(other.m_reverseMapping),
    m_isKnownMapping(other.m_isKnownMapping)
{
}

LemonPDF::CMap::CMap(CMap &&other)
  : m_forwardMapping(std::move(other.m_forwardMapping)), m_reverseMapping(std::move(other.m_reverseMapping)),
    m_isKnownMapping(other.m_isKnownMapping)
{
}

LemonPDF::CMap &LemonPDF::CMap::operator=(const CMap &other)
{
	m_forwardMapping = other.m_forwardMapping;
	m_reverseMapping = other.m_reverseMapping;
	m_isKnownMapping = other.m_isKnownMapping;
	return *this;
}

void LemonPDF::CMap::addMapping(char32_t ch, const std::string &s)
{
	m_forwardMapping.emplace(ch, s);
	m_reverseMapping.emplace(s, ch);
}

std::string LemonPDF::CMap::getCharactersForGlyphIndex(char32_t glyphIndex) const
{
	auto it = m_forwardMapping.find(glyphIndex);
	if (it == m_forwardMapping.end())
		throw PDFInvalidOperationException("Invalid glyph index");
	
	return it->second;
}

std::string LemonPDF::CMap::getCharactersFromGlyphIndexString(const std::u16string &input, const std::string &missing) const
{
	return convertFromCMap(input, m_forwardMapping, missing);
}

std::string LemonPDF::CMap::getCharactersFromGlyphIndexString(const std::string &input, const std::string &missing) const
{
	return convertFromCMap(input, m_forwardMapping, missing);
}

char32_t LemonPDF::CMap::getGlyphIndexForString(const std::string &str) const
{
	auto it = m_reverseMapping.find(str);
	if (it == m_reverseMapping.end())
		throw PDFInvalidOperationException("No glyph index for specified characters");
	
	return it->second;
}

bool LemonPDF::CMap::hasMappingForGlyphIndex(char32_t glyphIndex) const
{
	return m_forwardMapping.find(glyphIndex) != m_forwardMapping.end();
}

bool LemonPDF::CMap::hasMappings() const
{
	return m_forwardMapping.size() > 0;
}

bool LemonPDF::CMap::isKnownMapping() const
{
	return m_isKnownMapping;
}

bool LemonPDF::CMap::tryGetGlyphIndexForString(const std::string &str, char32_t &result) const
{
	auto it = m_reverseMapping.find(str);
	if (it == m_reverseMapping.end())
	{
		result = 0;
		return false;
	}

	result = it->second;
	return true;
}
