#include <algorithm>
#include <cmath>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <unordered_map>
#include "gfxcolor.h"

LemonPDF::Graphics::Color::Color(ColorSpacePtr cs, const std::vector<double> &comps)
  : m_colorSpace(cs), m_components(comps)
{
}

LemonPDF::Graphics::Color::Color(ColorSpacePtr cs, const std::initializer_list<double> &comps)
  : m_colorSpace(cs), m_components(comps)
{
}

LemonPDF::Graphics::Color::~Color()
{
}

LemonPDF::Graphics::ColorSpacePtr LemonPDF::Graphics::Color::getColorSpace() const
{
	return m_colorSpace;
}

const std::vector<double> &LemonPDF::Graphics::Color::getComponents() const
{
	return m_components;
}

std::tuple<double, double, double> LemonPDF::Graphics::Color::getHueSaturationValue() const
{
	const double epsilon = 0.0001;

	if (!m_colorSpace)
		throw PDFInvalidOperationException("No colorspace is set");
	
	double r, g, b;
	std::tie(r, g, b) = getRedGreenBlue();

	double hue = 0.0, saturation = 0.0, value = 0.0;

	// HSV requires finding the largest and smallest of the component values
	const double max = std::max({r, g, b});
	const double min = std::min({r, g, b});
	const double difference = max - min;

	value = max;
	if (max < epsilon)
	{
		// if the largest component is virtually zero, the colour is black (no hue)
		hue = NAN;
		value = 0.0; // clamp to 0
	}
	else if (difference < epsilon)
	{
		// if the difference between highest and lowest is small, it's a shade of gray,
		// and therefore has no hue
		hue = NAN;
	}
	else
	{
		saturation = difference / max;
		if (r == max) // if red is the max
			hue = 000.0 + 60.0 * (g - b) / difference;
		else if (g == max) // green is max
			hue = 120.0 + 60.0 * (b - r) / difference;
		else if (b == max) // blue is max
			hue = 240.0 + 60.0 * (r - g) / difference;
		
		hue = (hue < 0.0 ? hue + 360.0 : hue);
	}
	
	return std::make_tuple(hue, saturation, value);
}

std::tuple<double, double, double> LemonPDF::Graphics::Color::getRedGreenBlue() const
{
	if (!m_colorSpace)
		throw PDFInvalidOperationException("No color space set");
	
	switch (m_colorSpace->getColorModel())
	{
		case ColorSpace::ColorModel::kRGB:
			return std::make_tuple(m_components[0], m_components[1], m_components[2]);
		
		case ColorSpace::ColorModel::kGray:
			return std::make_tuple(m_components[0], m_components[0], m_components[0]);
		
		case ColorSpace::ColorModel::kCMYK:
		{
			double r, g, b;
			r = 1.0 - std::min(1.0, m_components[0] * (1.0 - m_components[3]) + m_components[3]);
			g = 1.0 - std::min(1.0, m_components[1] * (1.0 - m_components[3]) + m_components[3]);
			b = 1.0 - std::min(1.0, m_components[2] * (1.0 - m_components[3]) + m_components[3]);
			return std::make_tuple(r, g, b);
		}

		default:
			throw PDFUnimplementedException("Don't know how to convert this color to RGB");
	}
}

std::string LemonPDF::Graphics::Color::toRGBString() const
{
	double r;
	double g;
	double b;

	std::tie(r, g, b) = getRedGreenBlue();

	std::ostringstream result;
	result << std::hex;
	result << "#";
	result.fill('0');
	result << std::setw(2) << static_cast<unsigned>(r * 255.0);
	result << std::setw(2) << static_cast<unsigned>(g * 255.0);
	result << std::setw(2) << static_cast<unsigned>(b * 255.0);
	return result.str();
}

LemonPDF::Graphics::ColorException::ColorException(const std::string &msg)
  : PDFException(msg)
{
}
