#include <unordered_map>

#include "gfxcolorspace.h"
#include "gfxcolor.h"

namespace
{
	class CalibratedGrayColorSpace : public LemonPDF::Graphics::ColorSpace
	{
	public:
		CalibratedGrayColorSpace(const LemonPDF::Graphics::XYZPoint &white, const LemonPDF::Graphics::XYZPoint &black, double gg)
		  : ColorSpace("Calibrated Gray"), m_whitePoint(white), m_blackPoint(black), m_grayGamma(gg)
		{
			(void) m_grayGamma;
		}
		
		ColorModel getColorModel() const override
		{
			return ColorModel::kGray;
		}
		
		size_t getNumberOfComponents() const override
		{
			return 1;
		}

		bool isSupported() const override
		{
			return true;
		}
		
	private:
		LemonPDF::Graphics::XYZPoint m_whitePoint;
		LemonPDF::Graphics::XYZPoint m_blackPoint;
		double m_grayGamma;
	};
	
	class DeviceGrayColorSpace : public CalibratedGrayColorSpace
	{
	public:
		DeviceGrayColorSpace()
		  : CalibratedGrayColorSpace(LemonPDF::Graphics::XYZPoint{1.0, 1.0, 1.0}, LemonPDF::Graphics::XYZPoint{0.0, 0.0, 0.0}, 1.0)
		{
		}
		
		const std::string &getName() const override
		{
			return s_name;
		}
		
	private:
		static std::string s_name;
	};
	
	std::string DeviceGrayColorSpace::s_name = "Device Gray";
	
	class CalibratedRGBColorSpace : public LemonPDF::Graphics::ColorSpace
	{
	public:
		CalibratedRGBColorSpace(const LemonPDF::Graphics::XYZPoint &white, const LemonPDF::Graphics::XYZPoint &black, double rg, double gg, double bg)
		  : ColorSpace("Calibrated RGB"), m_whitePoint(white), m_blackPoint(black), m_redGamma(rg), m_greenGamma(gg), m_blueGamma(bg)
		{
			(void) m_redGamma;
			(void) m_greenGamma;
			(void) m_blueGamma;
		}
		
		ColorModel getColorModel() const override
		{
			return ColorModel::kRGB;
		}

		size_t getNumberOfComponents() const override
		{
			return 3;
		}

		bool isSupported() const override
		{
			return true;
		}
		
	private:
		LemonPDF::Graphics::XYZPoint m_whitePoint;
		LemonPDF::Graphics::XYZPoint m_blackPoint;
		double m_redGamma;
		double m_greenGamma;
		double m_blueGamma;
	};
	
	class DeviceRGBColorSpace : public CalibratedRGBColorSpace
	{
	public:
		DeviceRGBColorSpace()
		  : CalibratedRGBColorSpace(LemonPDF::Graphics::XYZPoint{1.0, 1.0, 1.0}, LemonPDF::Graphics::XYZPoint{0.0, 0.0, 0.0}, 1.0, 1.0, 1.0)
		{
		}
		
		const std::string &getName() const override
		{
			return s_name;
		}
		
	private:
		static std::string s_name;
	};
	
	std::string DeviceRGBColorSpace::s_name = "Device RGB";

	class ICCBasedColorSpace : public LemonPDF::Graphics::ColorSpace
	{
	public:
		ICCBasedColorSpace(const std::vector<unsigned char> &iccProfileData, size_t numberOfComponents)
		  : ColorSpace("ICC Based"), m_iccProfileData(iccProfileData), m_numberOfComponents(numberOfComponents), m_colorModel(ColorModel::kOther)
		{
			if (m_numberOfComponents == 1)
				m_colorModel = ColorModel::kGray;
			else if (m_numberOfComponents == 3)
				m_colorModel = ColorModel::kRGB;
			else if (m_numberOfComponents == 4)
				m_colorModel = ColorModel::kCMYK;
			else
				throw LemonPDF::Graphics::ColorException("Invalid number of components for ICC Profile");
		}
		
		ColorModel getColorModel() const override
		{
			return m_colorModel;
		}

		size_t getNumberOfComponents() const override
		{
			return m_numberOfComponents;
		}
		
		bool isSupported() const override
		{
			return true;
		}
		
	private:
		std::vector<unsigned char> m_iccProfileData;
		size_t m_numberOfComponents;
		ColorModel m_colorModel;
	};
	
	class DeviceCMYKColorSpace : public LemonPDF::Graphics::ColorSpace
	{
	public:
		DeviceCMYKColorSpace()
		  : LemonPDF::Graphics::ColorSpace("Device CMYK")
		{
		}
		
		ColorModel getColorModel() const override
		{
			return ColorModel::kCMYK;
		}

		size_t getNumberOfComponents() const override
		{
			return 4;
		}

		bool isSupported() const override
		{
			return true;
		}
	};

	class PatternColorSpace : public LemonPDF::Graphics::ColorSpace
	{
	public:
		PatternColorSpace()
		  : LemonPDF::Graphics::ColorSpace("Pattern")
		{
		}

		ColorModel getColorModel() const override
		{
			return ColorModel::kOther;
		}

		size_t getNumberOfComponents() const override
		{
			return 0;
		}

		bool isSupported() const override
		{
			return true;
		}
	};

	class UnsupportedColorSpace : public LemonPDF::Graphics::ColorSpace
	{
	public:
		UnsupportedColorSpace(const std::string &name)
		  : LemonPDF::Graphics::ColorSpace("Unsupported Colorspace " + name)
		{
		}

		ColorModel getColorModel() const override
		{
			return ColorModel::kOther;
		}

		size_t getNumberOfComponents() const override
		{
			return 1;
		}

		bool isSupported() const override
		{
			return false;
		}
	};
	
	static std::unordered_map<std::string, LemonPDF::Graphics::ColorSpacePtr> g_namedColorSpaces
	{
		{ "DeviceGray", std::make_shared<DeviceGrayColorSpace>() },
		{ "DeviceCMYK", std::make_shared<DeviceCMYKColorSpace>() },
		{ "DeviceRGB",  std::make_shared<DeviceRGBColorSpace>() },
		{ "Pattern", std::make_shared<PatternColorSpace>() },
		{ "DeviceN", std::make_shared<UnsupportedColorSpace>("DeviceN") },
		{ "Indexed", std::make_shared<UnsupportedColorSpace>("Indexed") },
		{ "Separation", std::make_shared<UnsupportedColorSpace>("Separation") }
	};
}

LemonPDF::Graphics::XYZPoint::XYZPoint(double x, double y, double z)
  : m_x(x), m_y(y), m_z(z)
{
}

double LemonPDF::Graphics::XYZPoint::getX() const
{
	return m_x;
}

double LemonPDF::Graphics::XYZPoint::getY() const
{
	return m_y;
}

double LemonPDF::Graphics::XYZPoint::getZ() const
{
	return m_z;
}

LemonPDF::Graphics::ColorSpace::ColorSpace(const std::string &name)
  : m_name(name)
{
}

LemonPDF::Graphics::ColorSpace::ColorSpace()
{
}

LemonPDF::Graphics::ColorSpace::~ColorSpace()
{
}

LemonPDF::Graphics::ColorSpacePtr LemonPDF::Graphics::ColorSpace::CreateCalibratedGrayColorSpace(const XYZPoint &white, const XYZPoint &black, double gg)
{
	return std::make_shared<CalibratedGrayColorSpace>(white, black, gg);
}

LemonPDF::Graphics::ColorSpacePtr LemonPDF::Graphics::ColorSpace::CreateCalibratedRGBColorSpace(const XYZPoint &white, const XYZPoint &black, double rg, double gg, double bg)
{
	return std::make_shared<CalibratedRGBColorSpace>(white, black, rg, gg, bg);
}

LemonPDF::Graphics::ColorSpacePtr LemonPDF::Graphics::ColorSpace::CreateICCBasedColorSpace(const std::vector<unsigned char> &iccProfileData, size_t numberOfComponents)
{
	return std::make_shared<ICCBasedColorSpace>(iccProfileData, numberOfComponents);
}

LemonPDF::Graphics::ColorSpacePtr LemonPDF::Graphics::ColorSpace::GetNamedColorSpace(const std::string &name)
{
	std::unordered_map<std::string, ColorSpacePtr>::iterator i = g_namedColorSpaces.find(name);
	if (i != g_namedColorSpaces.end())
		return i->second;
	else
		return nullptr;
}

const std::string &LemonPDF::Graphics::ColorSpace::getName() const
{
	return m_name;
}
