#ifndef LEMONPDF_GFXCONTEXT_H
#define LEMONPDF_GFXCONTEXT_H

#include <lemongfx/bitmap.h>
#include "gfxcolor.h"
#include "gfxfont.h"
#include "gfxpattern.h"
#include "gfxpath.h"
#include "pdftypes.h"

namespace LemonPDF
{
	namespace Graphics
	{
		enum FillType
		{
			kNonzeroWindingFill,
			kEvenOddFill
		};
		
		enum LineCapStyle
		{
			kButtCap,
			kRoundCap,
			kProjectingSquareCap
		};
		
		enum LineJoinStyle
		{
			kMiterJoin,
			kRoundJoin,
			kBevelJoin
		};
		
		enum RenderingIntent
		{
			kRelativeColorimetricIntent,
			kAbsoluteColorimetricIntent,
			kPerceptualIntent,
			kSaturatedIntent
		};
		
		enum TextRenderingMode
		{
			kFillTextRenderingMode,
			kStrokeTextRenderingMode,
			kFillAndStrokeTextRenderingMode,
			kNoFillOrStrokeTextRenderingMode,
			kFillAndClipTextRenderingMode,
			kStrokeAndClipTextRenderingMode
		};
		
		struct TextState
		{
			TransformMatrix ctm;
			TransformMatrix lineCtm;
			TransformMatrix renderingCtm;
			double characterSpacing;
			double wordSpacing;
			double horizontalScaling;
			double leading;
			FontPtr font;
			double fontSize;
			enum TextRenderingMode renderingMode;
			double rise;
			bool knockout;
			
			TextState();
		};
		
		struct State
		{
			TransformMatrix ctm;
			Point point;
			Path path;
			double lineWidth;
			enum LineCapStyle lineCapStyle;
			enum LineJoinStyle lineJoinStyle;
			double flatness;
			double miterLimit;
			enum RenderingIntent renderingIntent;
			TextState textState;
			ColorSpacePtr fillColorSpace;
			ColorSpacePtr strokeColorSpace;
			ColorPtr fillColor;
			ColorPtr strokeColor;
			double fillAlpha;
			double strokeAlpha;
			
			State();
		};
		
		struct SpacedText
		{
			double spacing;
			GlyphData glyphData;

			SpacedText();
			SpacedText(double, const GlyphData &);
		};

		class IContext
		{
		public:
			// General graphics state commands
			virtual void concat(const TransformMatrix &) {}
			virtual void restoreState() {}
			virtual void saveState() {}
			virtual void setFlatness(double) {}
			virtual void setLineCapStyle(enum LineCapStyle) {}
			virtual void setLineJoinStyle(enum LineJoinStyle) {}
			virtual void setLineWidth(double) {}
			virtual void setMiterLimit(double) {}
			virtual void setRenderingIntent(enum RenderingIntent) {}
			
			// Color commands
			virtual void setDefaultColorSpace(ColorSpacePtr) {}
			virtual void setFillAlpha(double) {}
			virtual void setFillColorComponents(const std::vector<double> &) {}
			virtual void setFillColorSpace(ColorSpacePtr) {}
			virtual void setFillPattern(PatternPtr) {}
			virtual void setStrokeAlpha(double) {}
			virtual void setStrokeColorComponents(const std::vector<double> &) {}
			virtual void setStrokeColorSpace(ColorSpacePtr) {}
			virtual void setStrokePattern(PatternPtr) {}
			
			// Path commands
			virtual void closePath() {}
			virtual void curveFrom(const Point &, const Point &) {}
			virtual void curveTo(const Point &, const Point &) {}
			virtual void curveTo(const Point &, const Point &, const Point &) {}
			virtual void endPath() {}
			virtual void intersectClippingPath(enum FillType = kNonzeroWindingFill) {}
			virtual void fill(enum FillType = kNonzeroWindingFill) {}
			virtual void fillAndStroke(enum FillType = kNonzeroWindingFill) {}
			virtual void lineTo(const Point &) {}
			virtual void moveTo(const Point &) {}
			virtual void rectangle(const Rectangle &) {}
			virtual void stroke() {}
			
			// Text state commands
			virtual void setTextCharacterSpacing(double) {}
			virtual void setTextWordSpacing(double) {}
			virtual void setTextScale(double) {}
			virtual void setTextLeading(double) {}
			virtual void setTextFont(FontPtr, double) {}
			virtual void setTextRenderingMode(enum TextRenderingMode) {}
			virtual void setTextRise(double) {}
			
			// Text object commands
			virtual void beginText() {}
			virtual void endText() {}
			
			// Text positioning
			virtual void moveToNextLine() {}
			virtual void moveToNextLine(const Point &) {}
			virtual void setTextMatrix(const TransformMatrix &) {}
			
			// Text showing
			virtual void showText(const GlyphData &) {}
			virtual void showText(const std::vector<SpacedText> &) {}
			
			// Marked content
			virtual void beginMarkedContent(const std::string &) {}
			virtual void beginMarkedContent(const std::string &, const std::string &) {}
			virtual void markedContentPoint(const std::string &) {}
			virtual void markedContentPoint(const std::string &, const std::string &) {}
			virtual void endMarkedContent() {};

			// Other
			virtual void putImage(const LemonGFX::Image &) {};
			virtual void putMaskedImage(const LemonGFX::Image &, const LemonGFX::Image &) {};

			virtual ~IContext() {};
		};
	}
}

#endif
