#include <algorithm>
#include <sstream>
#include <libcore/logging.h>
#include "gfxfont.h"
#include "pdfexception.h"

LemonPDF::Graphics::Font::Font(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw)
  : m_fontCompleteness(fc), m_fontType(ft), m_fontWeight(fw), m_name(name), m_defaultGlyphWidth(0),
    m_canConvertGlyphIndicesToPaths(false), m_canConvertGlyphIndicesToText(true), m_doubleByteEncoding(false)
{
}

bool LemonPDF::Graphics::Font::canConvertGlyphIndicesToPaths() const
{
	return m_canConvertGlyphIndicesToPaths;
}

bool LemonPDF::Graphics::Font::canConvertGlyphIndicesToText() const
{
	return m_canConvertGlyphIndicesToText;
}

std::string LemonPDF::Graphics::Font::decodeText(const GlyphIndices &glyphIndices) const
{
	std::ostringstream result;
	
	for (GlyphIndices::value_type gi : glyphIndices)
	{
		std::string chars = m_cmap.getCharactersForGlyphIndex(gi);
		result << chars;
	}

	return result.str();
}

const LemonPDF::CMap &LemonPDF::Graphics::Font::getCMap() const
{
	return m_cmap;
}

double LemonPDF::Graphics::Font::getDefaultGlyphWidth() const
{
	return m_defaultGlyphWidth;
}

const std::string &LemonPDF::Graphics::Font::getEncodingName() const
{
	return m_encodingName;
}

enum LemonPDF::Graphics::Font::FontCompleteness LemonPDF::Graphics::Font::getFontCompleteness() const
{
	return m_fontCompleteness;
}

enum LemonPDF::Graphics::Font::FontType LemonPDF::Graphics::Font::getFontType() const
{
	return m_fontType;
}

int LemonPDF::Graphics::Font::getFontWeight() const
{
	return m_fontWeight;
}

LemonPDF::Graphics::GlyphIndices LemonPDF::Graphics::Font::getGlyphIndicesForData(const GlyphData &glyphData) const
{
	GlyphIndices result;
	result.reserve(m_doubleByteEncoding ? glyphData.size() / 2 : glyphData.size());

	for (std::string::size_type i = 0; i < glyphData.size(); i++)
	{
		char16_t ch;
		if (m_doubleByteEncoding && i + 1 < glyphData.size())
		{
			ch = (static_cast<char16_t>(glyphData.at(i)) << 8) | (glyphData.at(i + 1));
			i++;
		}
		else
		{
			ch = glyphData.at(i);
		}

		result.emplace_back(ch);
	}

	return result;
}

const LemonPDF::Graphics::GlyphOutlineMap &LemonPDF::Graphics::Font::getGlyphOutlineMap() const
{
	throw PDFInvalidOperationException("Cannot get glyph outlines of this font type");
}

const LemonPDF::Graphics::GlyphWidthMap &LemonPDF::Graphics::Font::getGlyphWidths() const
{
	return m_glyphWidths;
}

const std::string &LemonPDF::Graphics::Font::getName() const
{
	return m_name;
}

LemonPDF::Graphics::Path LemonPDF::Graphics::Font::getPathForGlyphIndices(const GlyphIndices &indices, double *width) const
{
	Path p;
	TransformMatrix ctm;
	GlyphWidthMap widths = getGlyphWidths();
	GlyphOutlineMap outlines = getGlyphOutlineMap();

	if (width)
		*width = 0;

	for (char16_t g : indices)
	{
		double glyphWidth;
		auto glyphWidthIter = widths.find(g);
		if (glyphWidthIter == widths.end())
			glyphWidth = getDefaultGlyphWidth();
		else
			glyphWidth = glyphWidthIter->second;
		
		auto glyphIter = outlines.find(g);
		if (glyphIter != outlines.end())
		{
			if (glyphIter->second != nullptr)
			{
				const Path &thisGlyph = glyphIter->second->getPath();
				Path translated = Path::Transform(thisGlyph, ctm);
				p.combine(translated);
			}
		}
		else
		{
			StripedHorse::Logging::Warning() << "No glyph outline for glyph with ID (" << g << ")" << std::endl;
		}

		ctm.translate(glyphWidth, 0);
		if (width)
			*width += glyphWidth;
	}

	return p;
}

double LemonPDF::Graphics::Font::getWidthsForGlyphIndices(const GlyphIndices &glyphIndices, std::vector<double> *individualWidths) const
{
	double grandTotal = 0.0;

	if (individualWidths != nullptr)
		individualWidths->reserve(glyphIndices.size());

	for (GlyphIndices::value_type gi : glyphIndices)
	{
		auto gw = m_glyphWidths.find(gi);
		double thisGlyphWidth = (gw == m_glyphWidths.end() ? m_defaultGlyphWidth : gw->second);

		grandTotal += thisGlyphWidth;
		if (individualWidths)
			individualWidths->push_back(thisGlyphWidth);
	}

	return grandTotal;
}

bool LemonPDF::Graphics::Font::hasFixedWidthGlyphs() const
{
	const GlyphWidthMap &glyphWidths = getGlyphWidths();

	if (glyphWidths.size() == 0)
		return false;

	auto start = glyphWidths.begin();
	auto end = glyphWidths.end();
	double width = start->second;
	return std::all_of(std::next(start), end, [&](const GlyphWidthMap::value_type &kv) { return kv.second == width || kv.second == 0; } );
}

void LemonPDF::Graphics::Font::setCanConvertGlyphIndicesToPaths(bool b)
{
	m_canConvertGlyphIndicesToPaths = b;
}

void LemonPDF::Graphics::Font::setCanConvertGlyphIndicesToText(bool b)
{
	m_canConvertGlyphIndicesToText = b;
}

void LemonPDF::Graphics::Font::setCMap(const CMap &cmap)
{
	m_cmap = cmap;

	if (m_cmap.isKnownMapping())
		return;
	
	if (m_cmap.hasMappings())
	{
		if (m_cmap.hasMappingForGlyphIndex(1))
		{
			// It is VERY unlikely for a legit /ToUnicode mapping to map
			// charcode 1 to charvalue 1
			const std::string &result = m_cmap.getCharactersForGlyphIndex(1);
			if (result.size() == 1 && result[0] == '\x01')
			{
				m_canConvertGlyphIndicesToText = false;
				return;
			}
		}

		m_canConvertGlyphIndicesToText = true;
	}
}

void LemonPDF::Graphics::Font::setDefaultGlyphWidth(double defaultGlyphWidth)
{
	m_defaultGlyphWidth = defaultGlyphWidth;
}

void LemonPDF::Graphics::Font::setEncodingName(const std::string &encodingName)
{
	m_encodingName = encodingName;
	if (m_encodingName == "Identity-H")
		m_doubleByteEncoding = true;
}

void LemonPDF::Graphics::Font::setGlyphWidths(const GlyphWidthMap &glyphWidths)
{
	m_glyphWidths = glyphWidths;
}

LemonPDF::Graphics::Type0Font::Type0Font(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw)
  : Font(name, fc, ft, fw)
{
}

LemonPDF::Graphics::Type1Font::Type1Font(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw)
  : Font(name, fc, ft, fw)
{
}

LemonPDF::Graphics::MMType1Font::MMType1Font(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw)
  : Type1Font(name, fc, ft, fw)
{
}

LemonPDF::Graphics::Type3Font::Type3Font(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw)
  : Font(name, fc, ft, fw)
{
}
