#ifndef LEMONPDF_GFXFONT_H
#define LEMONPDF_GFXFONT_H

#include <memory>
#include <string>
#include <vector>

#include "cmap.h"
#include "gfxglyph.h"

namespace LemonPDF
{
	namespace Graphics
	{
		class Font;
		using FontPtr = std::shared_ptr<Font>;
		
		using GlyphData = std::vector<unsigned char>;
		using GlyphIndices = std::vector<char16_t>;
		using GlyphWidthMap = std::map<char32_t, double>;
		using GlyphOutlineMap = std::map<char16_t, IGlyphPtr>;

		class Font
		{
		public:
			enum FontCompleteness
			{
				kComplete,
				kSubsetted
			};

			enum FontType
			{
				kFixedPitch  = 1 << 0,
				kSerif       = 1 << 1,
				kSymbolic    = 1 << 2,
				kScript      = 1 << 3,
				kNonSymbolic = 1 << 5,
				kItalic      = 1 << 6,
				kAllCap      = 1 << 16,
				kSmallCap    = 1 << 17,
				kForceBold   = 1 << 18
			};
			
			Font(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw);
			virtual ~Font() = default;

			virtual bool canConvertGlyphIndicesToPaths() const;
			virtual bool canConvertGlyphIndicesToText() const;
			virtual std::string decodeText(const GlyphIndices &) const;
			virtual const CMap &getCMap() const;
			virtual double getDefaultGlyphWidth() const;
			virtual const std::string &getEncodingName() const;
			virtual enum FontCompleteness getFontCompleteness() const;
			virtual enum FontType getFontType() const;
			virtual int getFontWeight() const;
			virtual GlyphIndices getGlyphIndicesForData(const GlyphData &data) const;
			virtual const GlyphOutlineMap &getGlyphOutlineMap() const;
			virtual const GlyphWidthMap &getGlyphWidths() const;
			virtual const std::string &getName() const;
			virtual Path getPathForGlyphIndices(const GlyphIndices &glyphIndices, double *width) const;
			virtual double getWidthsForGlyphIndices(const GlyphIndices &glyphIndices, std::vector<double> *individualWidths) const;
			virtual bool hasFixedWidthGlyphs() const;
			virtual void setCanConvertGlyphIndicesToPaths(bool);
			virtual void setCanConvertGlyphIndicesToText(bool);
			virtual void setCMap(const CMap &cmap);
			virtual void setDefaultGlyphWidth(double);
			virtual void setEncodingName(const std::string &encodingName);
			virtual void setGlyphWidths(const GlyphWidthMap &glyphWidths);
			
		private:
			CMap m_cmap;
			enum FontCompleteness m_fontCompleteness;
			enum FontType m_fontType;
			int m_fontWeight;
			std::string m_name;
			std::string m_encodingName;
			GlyphWidthMap m_glyphWidths;
			double m_defaultGlyphWidth;
			bool m_canConvertGlyphIndicesToPaths;
			bool m_canConvertGlyphIndicesToText;
			bool m_doubleByteEncoding;
		};
		
		class Type0Font : public Font
		{
		public:
			Type0Font(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw);
		
		private:
		};
		
		class Type1Font : public Font
		{
		public:
			Type1Font(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw);
		
		private:
		};
		
		class MMType1Font : public Type1Font
		{
		public:
			MMType1Font(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw);
		
		private:
		};
		
		class Type3Font : public Font
		{
		public:
			Type3Font(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw);
		
		private:
		};
	}
}

#endif
