#include <iostream>
#include <sstream>
#include <unordered_set>

#include <libcore/logging.h>

#include "gfxcontext.h"
#include "gfxparser.h"
#include "gfxtokeniser.h"
#include "pdffile.h"

namespace
{
	std::unordered_map<std::string, enum LemonPDF::Graphics::RenderingIntent> g_renderingIntentMap {
		{ "AbsoluteColorimetric", LemonPDF::Graphics::RenderingIntent::kAbsoluteColorimetricIntent },
		{ "RelativeColorimetric", LemonPDF::Graphics::RenderingIntent::kRelativeColorimetricIntent },
		{ "Perceptual", LemonPDF::Graphics::RenderingIntent::kPerceptualIntent },
		{ "Saturation", LemonPDF::Graphics::RenderingIntent::kSaturatedIntent }
	};

	enum LemonPDF::Graphics::GFXOperator HashOp(const std::string &op)
	{
		return static_cast<LemonPDF::Graphics::GFXOperator>(LemonPDF::Graphics::HashOp(op.c_str()));
	}
	
	double toDouble(const LemonPDF::Graphics::GFXToken &t)
	{
		return std::stod(t.getText());
	}
	
	long toLong(const LemonPDF::Graphics::GFXToken &t)
	{
		return std::stol(t.getText());
	}
	
	int toInt(const LemonPDF::Graphics::GFXToken &t)
	{
		return std::stoi(t.getText());
	}
	
	void notImplemented(const std::string &op)
	{
		static std::unordered_set<std::string> s_alreadyWarned;
		if (s_alreadyWarned.insert(op).second)
			StripedHorse::Logging::Warning() << "PDF graphics operator \"" << op << "\" used" << std::endl;
	}

	void parseExtGState(LemonPDF::DictionaryPtr extGState, LemonPDF::Graphics::IContext &ctx, std::shared_ptr<const LemonPDF::ResourceManager> resourceManager)
	{
		using namespace LemonPDF;
		using namespace LemonPDF::Graphics;
		
		PDFFilePtr pdf = resourceManager->getFile();
		LiteralPtr value;
		
		value = pdf->resolve<Literal>(extGState->get("LW"));
		if (value)
			ctx.setLineWidth(value->getRealValue());
		
		value = pdf->resolve<Literal>(extGState->get("LC"));
		if (value)
			ctx.setLineCapStyle(static_cast<enum LineCapStyle>(value->getIntegerValue()));

		value = pdf->resolve<Literal>(extGState->get("LJ"));
		if (value)
			ctx.setLineJoinStyle(static_cast<enum LineJoinStyle>(value->getIntegerValue()));

		value = pdf->resolve<Literal>(extGState->get("ML"));
		if (value)
			ctx.setMiterLimit(value->getRealValue());

		value = pdf->resolve<Literal>(extGState->get("RI"));
		if (value)
		{
			auto renderingIntentIter = g_renderingIntentMap.find(value->getStringValue());
			if (renderingIntentIter != g_renderingIntentMap.end())
				ctx.setRenderingIntent(renderingIntentIter->second);
			else
				notImplemented("Rendering intent: " + value->getStringValue());
		}
		
		ArrayPtr fontValue;
		fontValue = pdf->resolve<Array>(extGState->get("Font"));
		if (fontValue != nullptr && fontValue->getSize() == 2)
		{
			LiteralPtr fontName = pdf->resolve<Literal>(fontValue->get(0));
			FontPtr font = resourceManager->getFont(fontName->getNameValue());

			value = pdf->resolve<Literal>(fontValue->get(1));
			if (font && value)
				ctx.setTextFont(font, value->getRealValue());
		}

		value = pdf->resolve<Literal>(extGState->get("CA"));
		if (value)
			ctx.setFillAlpha(value->getRealValue());

		value = pdf->resolve<Literal>(extGState->get("ca"));
		if (value)
			ctx.setStrokeAlpha(value->getRealValue());
		
		std::vector<std::string> unsupportedKeys {
			"D", "OP", "op", "OPM",
			"BG", "BG2", "UCR", "UCR2", "TR", "TR2", "HT", "FL"
			"SM", "SA", "BM", "SMask", "AIS", "TK"
		};

		for (const std::string &unsupportedKey : unsupportedKeys)
		{
			if (extGState->get(unsupportedKey) != nullptr)
				notImplemented("ExtGState:" + unsupportedKey);
		}
	}

	LemonPDF::Graphics::ColorSpacePtr getColorSpaceFromXObject(LemonPDF::PDFFilePtr pdf, LemonPDF::StreamPtr stream)
	{
		if (!stream)
			return nullptr;
		
		LemonPDF::DictionaryPtr streamDict = stream->getDictionary();
		LemonPDF::ObjectPtr csObject = streamDict->get("ColorSpace");
		if (!csObject)
			return nullptr;
		
		LemonPDF::LiteralPtr csLiteral = pdf->resolve<LemonPDF::Literal>(csObject);
		if (csLiteral)
			return LemonPDF::Graphics::ColorSpace::GetNamedColorSpace(csLiteral->getNameValue());
		
		LemonPDF::ArrayPtr csArray = pdf->resolve<LemonPDF::Array>(csObject);
		if (!csArray || csArray->getSize() != 2)
			return nullptr;
		
		csLiteral = pdf->resolve<LemonPDF::Literal>(csArray->get(0));
		LemonPDF::Graphics::ColorSpacePtr colorSpace = LemonPDF::Graphics::ColorSpace::GetNamedColorSpace(csLiteral->getNameValue());
		if (colorSpace)
			return colorSpace;
		
		LemonPDF::StreamPtr profileStream = pdf->resolve<LemonPDF::Stream>(csArray->get(1));
		if (!profileStream)
			return nullptr;
		
		LemonPDF::DictionaryPtr profileStreamDict = profileStream->getDictionary();
		LemonPDF::LiteralPtr alternateProfile = pdf->resolve<LemonPDF::Literal>(profileStreamDict->get("Alternate"));
		if (!alternateProfile)
		{
			// as a guess, if there is one channel, we'll assume DeviceGray,
			// if three channels, RGB, and four, CMYK.
			LemonPDF::LiteralPtr channelCount = pdf->resolve<LemonPDF::Literal>(profileStreamDict->get("N"));
			if (!channelCount)
				return nullptr;

			if (channelCount->getIntegerValue() == 1)
				return LemonPDF::Graphics::ColorSpace::GetNamedColorSpace("DeviceGray");
			else if (channelCount->getIntegerValue() == 3)
				return LemonPDF::Graphics::ColorSpace::GetNamedColorSpace("DeviceRGB");
			else if (channelCount->getIntegerValue() == 4)
				return LemonPDF::Graphics::ColorSpace::GetNamedColorSpace("DeviceCMYK");
			else
				return nullptr;
		}
		
		return LemonPDF::Graphics::ColorSpace::GetNamedColorSpace(alternateProfile->getNameValue());
	}
	
	void performOperation(const std::string &op, const std::vector<LemonPDF::Graphics::GFXToken> &tokens, LemonPDF::Graphics::IContext &ctx, std::shared_ptr<const LemonPDF::ResourceManager> resourceManager)
	{
		using namespace LemonPDF;
		using namespace LemonPDF::Graphics;
		
		unsigned long hop = HashOp(op);
		switch (hop)
		{
			case kOpSaveState:
			{
				ctx.saveState();
				break;
			}
			
			case kOpRestoreState:
			{
				ctx.restoreState();
				break;
			}
			
			case kOpConcatMatrix:
			{
				if (tokens.size() < 6)
					throw GFXOperandException(op);
				double args[6];
				for (size_t i = 0; i < 6; i++)
					args[i] = toDouble(tokens[i]);
				ctx.concat(TransformMatrix(args[0], args[1], args[2], args[3], args[4], args[5]));
				break;
			}
				
			case kOpSetLineWidth:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				ctx.setLineWidth(toDouble(tokens[0]));
				break;
			}
				
			case kOpSetLineCap:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				ctx.setLineCapStyle(static_cast<enum LineCapStyle>(toLong(tokens[0])));
				break;
			}
			
			case kOpSetLineJoin:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				ctx.setLineJoinStyle(static_cast<enum LineJoinStyle>(toLong(tokens[0])));
				break;
			}
			
			case kOpSetMiterLimit:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				ctx.setMiterLimit(toDouble(tokens[0]));
				break;
			}
			
			case kOpSetDashPattern:
			{
				notImplemented(op);
				break;
			}
			
			case kOpSetRenderingIntent:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				
				auto renderingIntentIter = g_renderingIntentMap.find(tokens[0].getNameValue());
				if (renderingIntentIter == g_renderingIntentMap.end())
					throw GFXOperandException(op);
				
				ctx.setRenderingIntent(renderingIntentIter->second);
				break;
			}
			
			case kOpSetFlatness:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				ctx.setFlatness(toDouble(tokens[0]));
				break;
			}
			
			case kOpSetGraphicsState:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				
				PDFFilePtr pdf = resourceManager->getFile();
				DictionaryPtr resources = resourceManager->getDictionary();
				if (resources == nullptr)
					throw GFXOperandException(op);
				
				DictionaryPtr extGStates = pdf->resolve<Dictionary>(resources->get("ExtGState"));
				if (!extGStates)
					throw GFXOperandException(op);

				DictionaryPtr extGState = pdf->resolve<Dictionary>(extGStates->get(tokens[0].getNameValue()));
				if (!extGState)
					throw GFXOperandException(op);
				
				parseExtGState(extGState, ctx, resourceManager);
				
				break;
			}
			
			case kOpMoveTo:
			{
				if (tokens.size() < 2)
					throw GFXOperandException(op);
				ctx.moveTo(LemonPDF::Point {toDouble(tokens[0]), toDouble(tokens[1])});
				break;
			}
			
			case kOpLineTo:
			{
				if (tokens.size() < 2)
					throw GFXOperandException(op);
				ctx.lineTo(LemonPDF::Point {toDouble(tokens[0]), toDouble(tokens[1])});
				break;
			}
			
			case kOpCurveTo:
			{
				if (tokens.size() < 6)
					throw GFXOperandException(op);
				ctx.curveTo(LemonPDF::Point {toDouble(tokens[0]), toDouble(tokens[1])}, LemonPDF::Point {toDouble(tokens[2]), toDouble(tokens[3])}, LemonPDF::Point {toDouble(tokens[4]), toDouble(tokens[5])});
				break;
			}
			
			case kOpCurveToV:
			{
				if (tokens.size() < 4)
					throw GFXOperandException(op);
				ctx.curveTo(LemonPDF::Point {toDouble(tokens[0]), toDouble(tokens[1])}, LemonPDF::Point {toDouble(tokens[0]), toDouble(tokens[1])}, LemonPDF::Point {toDouble(tokens[2]), toDouble(tokens[3])});
				break;
			}
			
			case kOpCurveToY:
			{
				if (tokens.size() < 4)
					throw GFXOperandException(op);
				ctx.curveTo(LemonPDF::Point {toDouble(tokens[0]), toDouble(tokens[1])}, LemonPDF::Point {toDouble(tokens[0]), toDouble(tokens[1])}, LemonPDF::Point {toDouble(tokens[2]), toDouble(tokens[3])});
				break;
			}
			
			case kOpClosePath:
			{
				ctx.closePath();
				break;
			}
			
			case kOpRectangle:
			{
				if (tokens.size() < 4)
					throw GFXOperandException(op);

				double x = toDouble(tokens[0]);
				double y = toDouble(tokens[1]);
				double w = toDouble(tokens[2]);
				double h = toDouble(tokens[3]);
				
				ctx.rectangle(LemonPDF::Rectangle {x, y, x + w, y + h});
				break;
			}
			
			case kOpStroke:
			{
				ctx.stroke();
				break;
			}
			
			case kOpCloseAndStroke:
			{
				ctx.closePath();
				ctx.stroke();
				break;
			}
			
			case kOpFillDeprecated:
			case kOpFill:
			{
				ctx.fill(FillType::kNonzeroWindingFill);
				break;
			}
			
			case kOpEvenOddFill:
			{
				ctx.fill(FillType::kEvenOddFill);
				break;
			}
			
			case kOpFillAndStroke:
			{
				ctx.fill(FillType::kNonzeroWindingFill);
				ctx.stroke();
				break;
			}
			
			case kOpEvenOddFillAndStroke:
			{
				ctx.fill(FillType::kEvenOddFill);
				ctx.stroke();
				break;
			}
			
			case kOpCloseAndFillAndStroke:
			{
				ctx.closePath();
				ctx.fill(FillType::kNonzeroWindingFill);
				ctx.stroke();
				break;
			}
			
			case kOpCloseAndEvenOddFillAndStroke:
			{
				ctx.closePath();
				ctx.fill(FillType::kEvenOddFill);
				ctx.stroke();
				break;
			}
			
			case kOpEndPath:
			{
				break;
			}

			case kOpShade:
			{
				notImplemented(op);
				break;
			}
			
			case kOpIntersectClippingPath:
			{
				ctx.intersectClippingPath(FillType::kNonzeroWindingFill);
				break;
			}
			
			case kOpEvenOddIntersectClippingPath:
			{
				ctx.intersectClippingPath(FillType::kEvenOddFill);
				break;
			}

			case kOpTextBegin:
			{
				ctx.beginText();
				break;
			}
			
			case kOpTextEnd:
			{
				ctx.endText();
				break;
			}
			
			case kOpSetTextCharacterSpacing:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				ctx.setTextCharacterSpacing(toDouble(tokens[0]));
				break;
			}
			
			case kOpSetTextWordSpacing:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				ctx.setTextWordSpacing(toDouble(tokens[0]));
				break;
			}
			
			case kOpSetTextScale:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				ctx.setTextScale(toDouble(tokens[0]));
				break;
			}
			
			case kOpSetTextLeading:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				ctx.setTextLeading(toDouble(tokens[0]));
				break;
			}
			
			case kOpSetTextFont:
			{
				if (tokens.size() < 2)
					throw GFXOperandException(op);
				
				std::string fontName = tokens.at(0).getNameValue();
				double fontSize = toDouble(tokens.at(1));
				
				FontPtr font = resourceManager->getFont(fontName);
				ctx.setTextFont(font, fontSize);
				break;
			}
			
			case kOpSetTextRenderingMode:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				enum TextRenderingMode tm = static_cast<enum TextRenderingMode>(toInt(tokens[0]));
				ctx.setTextRenderingMode(tm);
				break;
			}
			
			case kOpTextNextLineOffset:
			case kOpTextNextLineOffsetSetLeading:
			{
				if (tokens.size() < 2)
					throw GFXOperandException(op);
				double x = toDouble(tokens[0]);
				double y = toDouble(tokens[1]);
				if (hop == kOpTextNextLineOffsetSetLeading)
					ctx.setTextLeading(-y);
				ctx.moveToNextLine(LemonPDF::Point(x, y));
				break;
			}
			
			case kOpSetTextMatrix:
			{
				if (tokens.size() < 6)
					throw GFXOperandException(op);
				double args[6];
				for (size_t i = 0; i < 6; i++)
					args[i] = toDouble(tokens[i]);
				ctx.setTextMatrix(TransformMatrix(args[0], args[1], args[2], args[3], args[4], args[5]));
				break;
			}
			
			case kOpTextNextLine:
			{
				ctx.moveToNextLine();
				break;
			}
			
			case kOpTextShow:
			case kOpTextNextLineShow:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				if (hop == kOpTextNextLineShow)
					ctx.moveToNextLine();
				const std::string &token = tokens[0].getText();
				ctx.showText(GlyphData(token.begin(), token.end()));
				break;
			}
			
			case kOpTextNextLineShowSpacing:
			{
				if (tokens.size() < 3)
					throw GFXOperandException(op);
				ctx.setTextWordSpacing(toDouble(tokens[0]));
				ctx.setTextCharacterSpacing(toDouble(tokens[1]));
				ctx.moveToNextLine();

				const std::string &token = tokens[2].getText();
				ctx.showText(GlyphData(token.begin(), token.end()));
				break;
			}
			
			case kOpTextSpacedShow:
			{
				std::vector<SpacedText> spacedTexts;
				SpacedText st;
				
				st.spacing = 0;
				for (GFXToken t : tokens)
				{
					if (t.getType() == GFXToken::kNumber)
					{
						st.spacing = toDouble(t);
					}
					else if (t.getType() == GFXToken::kText)
					{
						st.glyphData = GlyphData(t.getText().begin(), t.getText().end());
						spacedTexts.push_back(st);
						
						st.spacing = 0;
						st.glyphData.clear();
					}
				}
				
				ctx.showText(spacedTexts);
				break;
			}
			
			case kOpXObject:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				
				PDFFilePtr pdf = resourceManager->getFile();
				if (!pdf)
					throw PDFInvalidOperationException("PDF Page object is not rooted in a file");
				
				StreamPtr stream = resourceManager->getXObject(tokens.at(0).getNameValue());
				if (!stream)
					throw GFXOperandException(op);
				
				DictionaryPtr streamDict = stream->getDictionary();
				if (!streamDict)
					throw GFXOperandException(op);
				
				LiteralPtr typeLiteral = pdf->resolve<Literal>(streamDict->get("Type"));
				if (!typeLiteral || typeLiteral->getNameValue() != "XObject")
					StripedHorse::Logging::Debug() << "Strange, this object may not be an XObject but it is treated as one" << std::endl;
				
				LiteralPtr subtypeLiteral = pdf->resolve<Literal>(streamDict->get("Subtype"));
				if (!subtypeLiteral)
					throw PDFInvalidOperationException("Expected /Subtype key in XObject");
				
				std::string subtype = subtypeLiteral->getNameValue();
				if (subtype == "Image")
				{
					if (!stream->filteringSupported())
					{
						StripedHorse::Logging::Warning() << "Cannot handle this image data" << std::endl;
						break;
					}

					LemonGFX::Image image = GetImageFromXObject(pdf, stream);
					LemonGFX::Image mask;

					StreamPtr maskStream = pdf->resolve<Stream>(streamDict->get("SMask"));
					if (maskStream != nullptr)
					{
						mask = GetImageFromXObject(pdf, maskStream);
						if (mask.getWidth() != image.getWidth() || mask.getHeight() != image.getHeight())
							StripedHorse::Logging::Warning() << "The image's mask is a different size to the image!" << std::endl;

						ctx.putMaskedImage(image, mask);
					}
					else
					{
						ctx.putImage(image);
					}
				}
				else if (subtype == "Form")
				{
					DictionaryPtr resourceDictionary = pdf->resolve<Dictionary>(stream->getDictionary()->get("Resources"));
					
					std::stringstream graphicsStream;
					stream->filter(graphicsStream);
					GFXTokeniser tokeniser(graphicsStream);
					GFXParser parser(tokeniser, ctx);

					ctx.saveState();
					if (resourceDictionary)
						parser.parseAll(std::make_shared<ResourceManager>(pdf, resourceDictionary));
					else
						parser.parseAll(resourceManager); // inherit
					ctx.restoreState();
				}
				else
				{
					StripedHorse::Logging::Error() << "Unknown XObject subtype " << subtype << std::endl;
				}

				break;
			}
			
			case kOpSetFillColor:
			{
				std::vector<double> comps;
				for (const GFXToken &t : tokens)
					comps.push_back(toDouble(t));
				ctx.setFillColorComponents(comps);
				break;
			}
			
			case kOpSetStrokeColor:
			{
				std::vector<double> comps;
				for (const GFXToken &t : tokens)
					comps.push_back(toDouble(t));
				ctx.setStrokeColorComponents(comps);
				break;
			}
			
			case kOpSetCMYKFill:
			{
				if (tokens.size() < 4)
					throw GFXOperandException(op);
				
				ColorSpacePtr cmyk = ColorSpace::GetNamedColorSpace("DeviceCMYK");
				ctx.setFillColorSpace(cmyk);
				ctx.setFillColorComponents(std::vector<double>{ toDouble(tokens[0]), toDouble(tokens[1]), toDouble(tokens[2]), toDouble(tokens[3]) });
				break;
			}
			
			case kOpSetCMYKStroke:
			{
				if (tokens.size() < 4)
					throw GFXOperandException(op);
				
				ColorSpacePtr cmyk = ColorSpace::GetNamedColorSpace("DeviceCMYK");
				ctx.setStrokeColorSpace(cmyk);
				ctx.setStrokeColorComponents(std::vector<double>{ toDouble(tokens[0]), toDouble(tokens[1]), toDouble(tokens[2]), toDouble(tokens[3]) });
				break;
			}
			
			case kOpSetGrayFill:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				
				ColorSpacePtr gray = ColorSpace::GetNamedColorSpace("DeviceGray");
				ctx.setFillColorSpace(gray);
				ctx.setFillColorComponents(std::vector<double>{ toDouble(tokens[0]) });
				break;
			}
			
			case kOpSetGrayStroke:
			{
				if (tokens.size() < 1)
					throw GFXOperandException(op);
				
				ColorSpacePtr gray = ColorSpace::GetNamedColorSpace("DeviceGray");
				ctx.setStrokeColorSpace(gray);
				ctx.setStrokeColorComponents(std::vector<double>{ toDouble(tokens[0]) });
				break;
			}
			
			case kOpSetRGBFill:
			{
				if (tokens.size() < 3)
					throw GFXOperandException(op);

				ColorSpacePtr rgb = ColorSpace::GetNamedColorSpace("DeviceRGB");
				ctx.setFillColorSpace(rgb);
				ctx.setFillColorComponents(std::vector<double>{ toDouble(tokens[0]), toDouble(tokens[1]), toDouble(tokens[2]) });
				break;
			}
			
			case kOpSetRGBStroke:
			{
				if (tokens.size() < 3)
					throw GFXOperandException(op);
				
				ColorSpacePtr rgb = ColorSpace::GetNamedColorSpace("DeviceRGB");
				ctx.setStrokeColorSpace(rgb);
				ctx.setStrokeColorComponents(std::vector<double>{ toDouble(tokens[0]), toDouble(tokens[1]), toDouble(tokens[2]) });
				break;
			}
			
			case kOpSetFillColorN:
			case kOpSetStrokeColorN:
			{
				if (tokens.size() == 1 && tokens.at(0).getType() == GFXToken::TokenType::kName)
				{
					// must be a pattern?
					PatternPtr pattern = resourceManager->getPattern(tokens.at(0).getNameValue());
					if (hop == kOpSetFillColorN)
						ctx.setFillPattern(pattern);
					else
						ctx.setStrokePattern(pattern);
				}
				else
				{
					// could be an index or deviceN or somethingerather like that
					std::vector<double> comps;
					for (const GFXToken &t : tokens)
						comps.push_back(toDouble(t));
					
					if (hop == kOpSetFillColorN)
						ctx.setFillColorComponents(comps);
					else if (hop == kOpSetStrokeColorN)
						ctx.setStrokeColorComponents(comps);
				}

				break;
			}
			
			case kOpSetStrokeColorSpace:
			{
				if (tokens.size() != 1)
					throw GFXOperandException(op);
				
				const std::string &val = tokens.at(0).getText();
				
				ColorSpacePtr cs = resourceManager->getColorSpace(tokens.at(0).getNameValue());
				if (cs == nullptr)
					throw GFXOperandException("Unknown color space: " + val);

				ctx.setStrokeColorSpace(cs);
				break;
			}
			
			case kOpSetFillColorSpace:
			{
				if (tokens.size() != 1)
					throw GFXOperandException(op);
				
				const std::string &val = tokens.at(0).getText();
				
				ColorSpacePtr cs = resourceManager->getColorSpace(tokens.at(0).getNameValue());
				if (cs == nullptr)
					throw GFXOperandException("Unknown color space: " + val);

				ctx.setFillColorSpace(cs);
				break;
			}

			case kOpMarkedContentBegin:
			{
				if (tokens.size() != 1)
					throw GFXOperandException(op);
				
				ctx.beginMarkedContent(tokens.at(0).getText());
				break;
			}

			case kOpMarkedContentBeginWithProperties:
			{
				if (tokens.size() < 2)
					throw GFXOperandException(op);
				
				ctx.beginMarkedContent(tokens.at(0).getText(), tokens.at(1).getText());
				break;
			}

			case kOpMarkedContentPoint:
			{
				if (tokens.size() != 1)
					throw GFXOperandException(op);
				
				ctx.markedContentPoint(tokens.at(0).getText());
				break;
			}

			case kOpMarkedContentPointWithProperties:
			{
				if (tokens.size() < 2)
					throw GFXOperandException(op);
				
				ctx.markedContentPoint(tokens.at(0).getText(), tokens.at(1).getText());
				break;
			}

			case kOpMarkedContentEnd:
			{
				ctx.endMarkedContent();
				break;
			}

			case kOpBeginExtension:
			case kOpEndExtension:
			{
				// these operators should be ignored for now, until
				// the PDF specification fleshes out how to use them
				// properly
				break;
			}
			
			default:
			{
				throw GFXParserException("Unknown operator: " + op);
			}
		}
	}
}

LemonPDF::Graphics::GFXParser::GFXParser(IGFXTokeniser &tokeniser, IContext &context)
  : m_tokeniser(tokeniser), m_context(context)
{
}

bool LemonPDF::Graphics::GFXParser::parseNextOperation(std::shared_ptr<const ResourceManager> resourceManager)
{
	bool withinExtensionArea = false;
	bool opCompleted = false;
		
	std::vector<GFXToken> args;
	
	while (!opCompleted)
	{
		GFXToken t = m_tokeniser.getNextToken();

		// If we encountered a "BX", we need to ignore everything
		// up until "EX" (this is how PDF handles future extensions
		// within a page content stream).
		if (withinExtensionArea)
		{
			if (::HashOp(t.getText()) == kOpEndExtension)
			{
				withinExtensionArea = false;
				continue;
			}
		}
		
		switch (t.getType())
		{
			case GFXToken::kArrayStart:
			{
				while (t.getType() != GFXToken::kArrayEnd)
				{
					args.push_back(t);
					t = m_tokeniser.getNextToken();
					
					if (t.getType() == GFXToken::kEndOfFile)
						throw PDFIOException("Unexpected end of file parsing array");
				}
				
				break;
			}

			case GFXToken::kArrayEnd:
			{
				throw GFXParserException("Unexpected array end token");
			}

			case GFXToken::kDictStart:
			{
				while (t.getType() != GFXToken::kDictEnd)
				{
					args.push_back(t);
					t = m_tokeniser.getNextToken();

					if (t.getType() == GFXToken::kEndOfFile)
						throw PDFIOException("Unexpected end of file parsing dictionary");
				}

				break;
			}
			
			case GFXToken::kDictEnd:
			{
				throw GFXParserException("Unexpected dictionary end token");
			}
			
			case GFXToken::kEndOfFile:
			{
				return false;
			}
			
			case GFXToken::kName:
			case GFXToken::kNumber:
			case GFXToken::kText:
			{
				args.push_back(t);
				break;
			}
			
			case GFXToken::kOperator:
			{
				if (::HashOp(t.getText()) == kOpBeginExtension)
				{
					withinExtensionArea = true;
					break;
				}
				else
				{
					performOperation(t.getText(), args, m_context, resourceManager);
					opCompleted = true;
					return true;
				}
			}
		}
	}
	
	return true;
}

void LemonPDF::Graphics::GFXParser::parseAll(std::shared_ptr<const ResourceManager> resourceManager)
{
	while (parseNextOperation(resourceManager));
}

LemonPDF::Graphics::GFXParserException::GFXParserException(const std::string &msg, const std::vector<GFXToken> &tokens)
  : PDFException("")
{
	std::ostringstream msgstr;
	msgstr << msg << '\n';
	for (size_t i = 0; i < tokens.size(); i++)
		msgstr << " -- operand[" << i << "] = " << tokens.at(i).getText() << '\n';
	m_msg = msgstr.str();
}

LemonGFX::Image LemonPDF::Graphics::GetImageFromXObject(LemonPDF::PDFFilePtr pdf, LemonPDF::StreamPtr stream)
{
	LemonPDF::Graphics::ColorSpacePtr colorSpace = getColorSpaceFromXObject(pdf, stream);
	LemonPDF::DictionaryPtr streamDict = stream->getDictionary();
	LemonPDF::LiteralPtr bitsPerSampleLiteral = pdf->resolve<LemonPDF::Literal>(streamDict->get("BitsPerComponent"));
	LemonPDF::LiteralPtr widthLiteral = pdf->resolve<LemonPDF::Literal>(streamDict->get("Width"));
	LemonPDF::LiteralPtr heightLiteral = pdf->resolve<LemonPDF::Literal>(streamDict->get("Height"));
	LemonPDF::ArrayPtr decodeArray = pdf->resolve<LemonPDF::Array>(streamDict->get("Decode"));
	if (!bitsPerSampleLiteral || !widthLiteral || !heightLiteral)
		throw LemonPDF::Graphics::GFXOperandException("Do");
	
	size_t bps = bitsPerSampleLiteral->getIntegerValue();
	size_t spp = colorSpace ? colorSpace->getNumberOfComponents() : 1;
	size_t width = widthLiteral->getIntegerValue();
	size_t height = heightLiteral->getIntegerValue();

	std::vector<unsigned char> imageData = stream->getFilteredData();
	if (decodeArray != nullptr)
	{
		if (decodeArray->getSize() != spp * 2)
			throw LemonPDF::Graphics::GFXOperandException("Do");
		
		if (spp == 1 && bps == 1)
		{
			// for monochrome bitmaps, if the /Decode array is [1 0] then that
			// means 1 should be interpreted as black and 0 should be
			// interpreted as white. This means that the image data should be
			// negated in the context of PDFs, which expect 1 to be white
			// and 0 to be black (see Table 11, §7.5.6, PDF 32000-1:2008).
			LemonPDF::LiteralPtr mapFrom = pdf->resolve<LemonPDF::Literal>(decodeArray->get(0));
			LemonPDF::LiteralPtr mapTo = pdf->resolve<LemonPDF::Literal>(decodeArray->get(1));

			if (mapFrom && mapTo && mapFrom->getIntegerValue() == 1 && mapTo->getIntegerValue() == 0)
				for (unsigned char &b : imageData)
					b = ~b;
		}
		else
		{
			// currently /Decode only supported for monochrome bitmaps
			StripedHorse::Logging::Warning() << "Cannot use /Decode for this image" << std::endl;
		}
	}

	return LemonGFX::Image(bps, spp, width, height, imageData);
}
