#ifndef LEMONPDF_GFXPARSER_H
#define LEMONPDF_GFXPARSER_H

#include "gfxcontext.h"
#include "gfxtokeniser.h"
#include "pdfpage.h"
#include "pdfexception.h"

namespace LemonPDF
{
	namespace Graphics
	{
		inline constexpr unsigned long HashOp(const char name[])
		{
			return name[0] == '\0' ? 0 :
			       name[1] == '\0' ? name[0] :
			       name[2] == '\0' ? ((name[0] << 8) | name[1]) :
			       name[3] == '\0' ? ((name[0] << 16) | (name[1] << 8) | name[2]) :
			                         ((name[0] << 24) | (name[1] << 16) | (name[2] << 8) | name[3]);
		}
		
		enum GFXOperator : unsigned long
		{
			// graphics state operators
			kOpSaveState = HashOp("q"),
			kOpRestoreState = HashOp("Q"),
			kOpConcatMatrix = HashOp("cm"),
			kOpSetLineWidth = HashOp("w"),
			kOpSetLineCap = HashOp("J"),
			kOpSetLineJoin = HashOp("j"),
			kOpSetMiterLimit = HashOp("M"),
			kOpSetDashPattern = HashOp("d"),
			kOpSetRenderingIntent = HashOp("ri"),
			kOpSetFlatness = HashOp("i"),
			kOpSetGraphicsState = HashOp("gs"),
			
			// path operators
			kOpMoveTo = HashOp("m"),
			kOpLineTo = HashOp("l"),
			kOpCurveTo = HashOp("c"),
			kOpCurveToV = HashOp("v"),
			kOpCurveToY = HashOp("y"),
			kOpClosePath = HashOp("h"),
			kOpRectangle = HashOp("re"),
			
			// paint operators
			kOpStroke = HashOp("S"),
			kOpCloseAndStroke = HashOp("s"),
			kOpFillDeprecated = HashOp("F"),
			kOpFill = HashOp("f"),
			kOpEvenOddFill = HashOp("f*"),
			kOpFillAndStroke = HashOp("B"),
			kOpEvenOddFillAndStroke = HashOp("B*"),
			kOpCloseAndFillAndStroke = HashOp("b"),
			kOpCloseAndEvenOddFillAndStroke = HashOp("b*"),
			kOpEndPath = HashOp("n"),
			kOpShade = HashOp("sh"),
			
			// clipping path operators
			kOpIntersectClippingPath = HashOp("W"),
			kOpEvenOddIntersectClippingPath = HashOp("W*"),
			
			// text object operators
			kOpTextBegin = HashOp("BT"),
			kOpTextEnd = HashOp("ET"),
			
			// text state operators
			kOpSetTextCharacterSpacing = HashOp("Tc"),
			kOpSetTextWordSpacing = HashOp("Tw"),
			kOpSetTextScale = HashOp("Tz"),
			kOpSetTextLeading = HashOp("TL"),
			kOpSetTextFont = HashOp("Tf"),
			kOpSetTextRenderingMode = HashOp("Tr"),
			kOpSetTextRise = HashOp("Ts"),
			
			// text positioning operators
			kOpTextNextLineOffset = HashOp("Td"),
			kOpTextNextLineOffsetSetLeading = HashOp("TD"),
			kOpSetTextMatrix = HashOp("Tm"),
			kOpTextNextLine = HashOp("T*"),
			kOpTextShow = HashOp("Tj"),
			kOpTextNextLineShow = HashOp("\'"),
			kOpTextNextLineShowSpacing = HashOp("\""),
			kOpTextSpacedShow = HashOp("TJ"),
			
			// XObject
			kOpXObject = HashOp("Do"),
			
			// colorspace operators
			kOpSetStrokeColorSpace = HashOp("CS"),
			kOpSetFillColorSpace = HashOp("cs"),
			kOpSetStrokeColor = HashOp("SC"),
			kOpSetFillColor = HashOp("sc"),
			kOpSetStrokeColorN = HashOp("SCN"),
			kOpSetFillColorN = HashOp("scn"),
			kOpSetGrayStroke = HashOp("G"),
			kOpSetGrayFill = HashOp("g"),
			kOpSetRGBStroke = HashOp("RG"),
			kOpSetRGBFill = HashOp("rg"),
			kOpSetCMYKStroke = HashOp("K"),
			kOpSetCMYKFill = HashOp("k"),

			// Marked content operators
			kOpMarkedContentPoint = HashOp("MP"),
			kOpMarkedContentPointWithProperties = HashOp("DP"),
			kOpMarkedContentBegin = HashOp("BMC"),
			kOpMarkedContentBeginWithProperties = HashOp("BDC"),
			kOpMarkedContentEnd = HashOp("EMC"),

			// Extension operators
			kOpBeginExtension = HashOp("BX"),
			kOpEndExtension = HashOp("EX")
		};
		
		class GFXTokeniser;
		
		class GFXParser
		{
		public:
			GFXParser(IGFXTokeniser &, IContext &);
			
			void parseAll(std::shared_ptr<const ResourceManager>);
			bool parseNextOperation(std::shared_ptr<const ResourceManager>);
		
		private:
			IGFXTokeniser &m_tokeniser;
			IContext &m_context;
		};
		
		class GFXParserException : public LemonPDF::PDFException
		{
		public:
			explicit GFXParserException(const std::string &msg) : PDFException(msg) {}
			explicit GFXParserException(const std::string &msg, const std::vector<GFXToken> &tokens);
		};
		
		class GFXOperandException : public LemonPDF::PDFException
		{
		public:
			explicit GFXOperandException(const std::string &op) : PDFException("Invalid operands for operator: " + op) {}
		};

		LemonGFX::Image GetImageFromXObject(LemonPDF::PDFFilePtr pdf, LemonPDF::StreamPtr stream);
	}
}

#endif
