#include <algorithm>
#include <cmath>
#include "gfxpath.h"
#include "pdfexception.h"

namespace
{
	// three points are collinear if they form a triangle with 0 area
	bool isCollinear(const LemonPDF::Point &p1, const LemonPDF::Point &p2, const LemonPDF::Point &p3, const double epsilon = 0.01)
	{
		return std::abs((p1.x * (p2.y - p3.y) + p2.x * (p3.y - p1.y) + p3.x * (p1.y - p2.y)) / 2) < epsilon;
	}
}

LemonPDF::Graphics::Path LemonPDF::Graphics::Path::Transform(const Path &path, const TransformMatrix &ctm)
{
	Path result;
	result.m_pathNodes.reserve(path.m_pathNodes.size());

	for (auto pathIter = path.cbegin(); pathIter != path.cend(); pathIter++)
	{
		std::vector<Point> points = pathIter->getPoints();
		for (auto &p : points)
			p = ctm * p;
		result.m_pathNodes.emplace_back(pathIter->getPathNodeType(), points);
	}
	
	return result;
}

bool LemonPDF::Graphics::Path::PathNode::operator==(const LemonPDF::Graphics::Path::PathNode &other) const
{
	return m_pathNodeType == other.m_pathNodeType && m_points == other.m_points;
}

enum LemonPDF::Graphics::Path::PathNodeType LemonPDF::Graphics::Path::PathNode::getPathNodeType() const
{
	return m_pathNodeType;
}

std::vector<LemonPDF::Point> &LemonPDF::Graphics::Path::PathNode::getPoints()
{
	return m_points;
}

const std::vector<LemonPDF::Point> &LemonPDF::Graphics::Path::PathNode::getPoints() const
{
	return m_points;
}

void LemonPDF::Graphics::Path::PathNode::setPathNodeType(enum PathNodeType pnt)
{
	m_pathNodeType = pnt;
}

LemonPDF::Graphics::Path::Path()
  : m_lastMove(0)
{
}

bool LemonPDF::Graphics::Path::operator==(const Path &other) const
{
	return m_pathNodes == other.m_pathNodes;
}

void LemonPDF::Graphics::Path::appendLine(const LemonPDF::Point &p)
{
	if (m_pathNodes.size() >= 1)
	{
		auto &lastNode = m_pathNodes.back();
		enum PathNodeType lastNodeType = lastNode.getPathNodeType();
		if ((lastNodeType == PathNodeType::kLine || lastNodeType == PathNodeType::kNewPath) && lastNode.getPoints().at(0) == p)
			return;
				
		if (m_pathNodes.size() >= 2)
		{
			const auto &secondToLastNode = *(m_pathNodes.end() - 2);

			if (lastNode.getPathNodeType() == PathNodeType::kLine && isCollinear(secondToLastNode.getPoints().back(), lastNode.getPoints().front(), p))
			{
				lastNode.getPoints().at(0) = p;
				return;
			}
		}
	}

	m_pathNodes.emplace_back(PathNodeType::kLine, p);
}

void LemonPDF::Graphics::Path::appendBezierCurve(const LemonPDF::Point &c1, const LemonPDF::Point &c2, const LemonPDF::Point &e)
{
	m_pathNodes.emplace_back(PathNodeType::kBezierCurve, c1, c2, e);
}

void LemonPDF::Graphics::Path::appendBezierCurve(const LemonPDF::Point &c, const LemonPDF::Point &e)
{
	// convert quadratic to cubic
	LemonPDF::Point p1;
	if (m_pathNodes.size() >= 1)
	{
		auto &lastNode = m_pathNodes.back();
		p1 = lastNode.getPoints().back();
	}
	else
	{
		p1 = c;
	}

	LemonPDF::Point c1, c2;

	c1.x = p1.x + 2 * (c.x - p1.x) / 3;
	c1.y = p1.y + 2 * (c.y - p1.y) / 3;

	c2.x = e.x + 2 * (c.x - e.x) / 3;
	c2.y = e.y + 2 * (c.y - e.y) / 3;

	m_pathNodes.emplace_back(PathNodeType::kBezierCurve, c1, c2, e);
}

std::vector<LemonPDF::Graphics::Path::PathNode>::iterator LemonPDF::Graphics::Path::begin()
{
	return m_pathNodes.begin();
}

std::vector<LemonPDF::Graphics::Path::PathNode>::const_iterator LemonPDF::Graphics::Path::cbegin() const
{
	return m_pathNodes.cbegin();
}

std::vector<LemonPDF::Graphics::Path::PathNode>::const_iterator LemonPDF::Graphics::Path::cend() const
{
	return m_pathNodes.cend();
}

void LemonPDF::Graphics::Path::clear()
{
	m_pathNodes.clear();
	m_lastMove = 0;
}

void LemonPDF::Graphics::Path::closePath()
{
	// optimisation - remove empty paths
	if (m_pathNodes.back().getPathNodeType() == PathNodeType::kNewPath)
	{
		m_pathNodes.pop_back();
		auto it = std::find_if(m_pathNodes.rbegin(), m_pathNodes.rend(), [](const PathNode &pn) { return pn.getPathNodeType() == PathNodeType::kNewPath; });
		if (it != m_pathNodes.rend())
			m_lastMove = it.base() - m_pathNodes.begin();
		return;
	}

	// if the last node matches the first node, the close path should handle that
	auto lastPathNode = m_pathNodes.end() - 1;
	if (lastPathNode->getPathNodeType() == PathNodeType::kLine && lastPathNode->getPoints().front() == m_pathNodes.at(m_lastMove).getPoints().front())
		m_pathNodes.pop_back();
	
	// collinear nodes are removed by appendLine() but the case of a closed path is not handled there
	if (m_pathNodes.size() >= 3)
	{
		std::vector<PathNode> lastNodes { m_pathNodes.back(), m_pathNodes.at(m_lastMove), m_pathNodes.at(m_lastMove + 1) };
		
		if (lastNodes[2].getPathNodeType() == PathNodeType::kLine)
		{
			if (isCollinear(lastNodes[0].getPoints().back(), lastNodes[1].getPoints().front(), lastNodes[2].getPoints().front()))
			{
				m_pathNodes.at(m_lastMove + 1).setPathNodeType(PathNodeType::kNewPath);
				m_pathNodes.erase(m_pathNodes.begin() + m_lastMove);
			}
		}
	}

	m_pathNodes.emplace_back(PathNodeType::kClosePath);
}

void LemonPDF::Graphics::Path::combine(const Path &other)
{
	m_pathNodes.insert(m_pathNodes.end(), other.m_pathNodes.begin(), other.m_pathNodes.end());
}

std::vector<LemonPDF::Graphics::Path::PathNode>::iterator LemonPDF::Graphics::Path::end()
{
	return m_pathNodes.end();
}

void LemonPDF::Graphics::Path::endPath()
{
	m_pathNodes.emplace_back(PathNodeType::kTerminal);
}

const std::vector<LemonPDF::Graphics::Path::PathNode> &LemonPDF::Graphics::Path::getPathNodes() const
{
	return m_pathNodes;
}

bool LemonPDF::Graphics::Path::isDrawable() const
{
	if (m_pathNodes.size() < 2)
		return false;
	
	bool allPointsAreEqual = true;
	size_t i = 0;

	while (allPointsAreEqual && ++i < m_pathNodes.size())
		if (m_pathNodes[i].getPoints().size() > 0)
			allPointsAreEqual = (m_pathNodes[i].getPoints() == m_pathNodes[0].getPoints());

	return !allPointsAreEqual;
}

void LemonPDF::Graphics::Path::move(const LemonPDF::Point &p)
{
	// two consecutive moves should collapse to one
	if (m_pathNodes.size() > 0 && m_pathNodes.back().getPathNodeType() == PathNodeType::kNewPath)
	{
		m_pathNodes.back().getPoints().at(0) = p;
		return;
	}

	m_lastMove = m_pathNodes.size();
	m_pathNodes.emplace_back(PathNodeType::kNewPath, p);
}

bool LemonPDF::Graphics::Path::simplify()
{
	bool changed = false;

	if (m_pathNodes.size() > 0 && m_pathNodes.back().getPathNodeType() == PathNodeType::kNewPath)
	{
		m_pathNodes.pop_back();
		changed = true;
	}

	return changed;
}

std::vector<LemonPDF::Graphics::Path::PathNode>::size_type LemonPDF::Graphics::Path::size() const
{
	return m_pathNodes.size();
}
