#include <sstream>
#include "gfxstandardmanagedcontext.h"

LemonPDF::Graphics::StandardManagedContext::StandardManagedContext()
{
	m_states.emplace_back();
}

void LemonPDF::Graphics::StandardManagedContext::beginText()
{
	topState().textState.ctm = TransformMatrix {};
	topState().textState.lineCtm = TransformMatrix {};
}

void LemonPDF::Graphics::StandardManagedContext::closePath()
{
	topState().path.closePath();
}

void LemonPDF::Graphics::StandardManagedContext::concat(const LemonPDF::TransformMatrix &m)
{
	topState().ctm.concat(m);
}

void LemonPDF::Graphics::StandardManagedContext::curveFrom(const LemonPDF::Point &c, const LemonPDF::Point &ep)
{
	curveTo(topState().point, c, ep);
}

void LemonPDF::Graphics::StandardManagedContext::curveTo(const LemonPDF::Point &c, const LemonPDF::Point &ep)
{
	curveTo(c, ep, ep);
}

void LemonPDF::Graphics::StandardManagedContext::curveTo(const LemonPDF::Point &c1, const LemonPDF::Point &c2, const LemonPDF::Point &ep)
{
	topState().point = ep;
	topState().path.appendBezierCurve(c1, c2, ep);
}

void LemonPDF::Graphics::StandardManagedContext::endPath()
{
	topState().path.endPath();
}

void LemonPDF::Graphics::StandardManagedContext::endText()
{
}

void LemonPDF::Graphics::StandardManagedContext::lineTo(const LemonPDF::Point &p)
{
	topState().point = p;
	topState().path.appendLine(topState().point);
}

void LemonPDF::Graphics::StandardManagedContext::moveTo(const LemonPDF::Point &p)
{
	topState().point = p;
	topState().path.move(topState().point);
}

void LemonPDF::Graphics::StandardManagedContext::moveToNextLine()
{
	topState().textState.lineCtm.translate(0, -topState().textState.leading);
	topState().textState.ctm = topState().textState.lineCtm;
}

void LemonPDF::Graphics::StandardManagedContext::moveToNextLine(const LemonPDF::Point &p)
{
	topState().textState.lineCtm.translate(p.x, p.y);
	topState().textState.ctm = topState().textState.lineCtm;
}

void LemonPDF::Graphics::StandardManagedContext::rectangle(const LemonPDF::Rectangle &r)
{
	Path &path = topState().path;

	path.move(LemonPDF::Point{ r.lowerLeft.x, r.lowerLeft.y });
	path.appendLine(LemonPDF::Point{ r.upperRight.x, r.lowerLeft.y });
	path.appendLine(LemonPDF::Point{ r.upperRight.x, r.upperRight.y });
	path.appendLine(LemonPDF::Point{ r.lowerLeft.x, r.upperRight.y });
	path.closePath();
}

void LemonPDF::Graphics::StandardManagedContext::restoreState()
{
	if (m_states.size() == 1)
		throw PDFInvalidOperationException("restoreState/saveState is unbalanced");
	m_states.pop_back();
}

void LemonPDF::Graphics::StandardManagedContext::saveState()
{
	m_states.push_back(topState());
}

void LemonPDF::Graphics::StandardManagedContext::setDefaultColorSpace(ColorSpacePtr cs)
{
	(void) cs;
	throw PDFUnimplementedException("setDefaultColorSpace");
}

void LemonPDF::Graphics::StandardManagedContext::setFillAlpha(double alpha)
{
	topState().fillAlpha = alpha;
}

void LemonPDF::Graphics::StandardManagedContext::setFillColorComponents(const std::vector<double> &comps)
{
    ColorSpacePtr fillColorSpace = topState().fillColorSpace;
	
	if (fillColorSpace == nullptr)
		throw PDFInvalidOperationException("No fill color space set");
	
	if (comps.size() != fillColorSpace->getNumberOfComponents())
	{
		std::ostringstream msg;
		msg << "Wrong number of components for fill color space (" << fillColorSpace->getName() << "), expected " << fillColorSpace->getNumberOfComponents() << ", got " << comps.size();
		throw PDFInvalidOperationException(msg.str());
	}
	
	ColorPtr c = std::make_shared<Color>(fillColorSpace, comps);
	topState().fillColor = c;
}

void LemonPDF::Graphics::StandardManagedContext::setFillColorSpace(ColorSpacePtr cs)
{
	topState().fillColorSpace = cs;
}

void LemonPDF::Graphics::StandardManagedContext::setFillPattern(PatternPtr pattern)
{
	ColorSpacePtr fillColorSpace = topState().fillColorSpace;

	if (fillColorSpace == nullptr)
		throw PDFInvalidOperationException("No fill color space set");
	
	if (fillColorSpace->getName() != "Pattern")
		throw PDFInvalidOperationException("Pattern color space is not set, but a pattern has been used for filling");
	
	topState().fillColor = pattern;
}

void LemonPDF::Graphics::StandardManagedContext::setFlatness(double flatness)
{
	topState().flatness = flatness;
}

void LemonPDF::Graphics::StandardManagedContext::setLineCapStyle(enum LineCapStyle lcs)
{
	topState().lineCapStyle = lcs;
}

void LemonPDF::Graphics::StandardManagedContext::setLineJoinStyle(enum LineJoinStyle ljs)
{
	topState().lineJoinStyle = ljs;
}

void LemonPDF::Graphics::StandardManagedContext::setLineWidth(double thickness)
{
	topState().lineWidth = thickness;
}

void LemonPDF::Graphics::StandardManagedContext::setMiterLimit(double miterLimit)
{
	topState().miterLimit = miterLimit;
}

void LemonPDF::Graphics::StandardManagedContext::setRenderingIntent(enum RenderingIntent ri)
{
	topState().renderingIntent = ri;
}

void LemonPDF::Graphics::StandardManagedContext::setStrokeAlpha(double alpha)
{
	topState().strokeAlpha = alpha;
}

void LemonPDF::Graphics::StandardManagedContext::setStrokeColorComponents(const std::vector<double> &comps)
{
	ColorSpacePtr strokeColorSpace = topState().strokeColorSpace;
	
	if (strokeColorSpace == nullptr)
		throw PDFInvalidOperationException("No stroke color space set");
	
	if (comps.size() != strokeColorSpace->getNumberOfComponents())
	{
		std::ostringstream msg;
		msg << "Wrong number of components for stroke color space (" << strokeColorSpace->getName() << "), expected " << strokeColorSpace->getNumberOfComponents() << ", got " << comps.size();
		throw PDFInvalidOperationException(msg.str());
	}

	ColorPtr c = std::make_shared<Color>(strokeColorSpace, comps);
	topState().strokeColor = c;
}

void LemonPDF::Graphics::StandardManagedContext::setStrokeColorSpace(ColorSpacePtr cs)
{
	topState().strokeColorSpace = cs;
}

void LemonPDF::Graphics::StandardManagedContext::setStrokePattern(PatternPtr pattern)
{
	ColorSpacePtr strokeColorSpace = topState().strokeColorSpace;

	if (strokeColorSpace == nullptr)
		throw PDFInvalidOperationException("No stroke color space set");
	
	if (strokeColorSpace->getName() != "Pattern")
		throw PDFInvalidOperationException("Pattern color space is not set, but a pattern has been used for stroking");
	
	topState().strokeColor = pattern;
}

void LemonPDF::Graphics::StandardManagedContext::setTextCharacterSpacing(double cs)
{
	topState().textState.characterSpacing = cs;
}

void LemonPDF::Graphics::StandardManagedContext::setTextMatrix(const LemonPDF::TransformMatrix &m)
{
	topState().textState.ctm = m;
	topState().textState.lineCtm = m;
}

void LemonPDF::Graphics::StandardManagedContext::setTextWordSpacing(double ws)
{
	topState().textState.wordSpacing = ws;
}

void LemonPDF::Graphics::StandardManagedContext::setTextScale(double ts)
{
	topState().textState.horizontalScaling = ts;
}

void LemonPDF::Graphics::StandardManagedContext::setTextLeading(double tl)
{
	topState().textState.leading = tl;
}

void LemonPDF::Graphics::StandardManagedContext::setTextFont(FontPtr f, double size)
{
	topState().textState.font = f;
	topState().textState.fontSize = size;
}

void LemonPDF::Graphics::StandardManagedContext::setTextRenderingMode(enum TextRenderingMode trm)
{
	topState().textState.renderingMode = trm;
}

void LemonPDF::Graphics::StandardManagedContext::setTextRise(double tr)
{
	topState().textState.rise = tr;
}

LemonPDF::Graphics::State &LemonPDF::Graphics::StandardManagedContext::topState()
{
	return *(m_states.end() - 1);
}
