#include <cctype>
#include <sstream>
#include "gfxtokeniser.h"
#include "libtokenisation.h"
#include "pdfexception.h"

namespace
{
	std::string parseOperator(std::istream &stream)
	{
		std::string result;
		char c;
		
		while (stream.get(c))
		{
			if (::isspace(c))
				continue;
			else
				break;
		}
		
		do
		{
			if (::isalpha(c))
			{
				result.push_back(c);
			}
			else if (c == '*' || c == '\'' || c == '\"')
			{
				result.push_back(c);
				return result;
			}
			else
			{
				stream.putback(c);
				break;
			}
		}
		while (stream.get(c));
		
		if (!stream && !stream.eof())
			throw LemonPDF::PDFIOException("Read error");
		
		return result;
	}
	
	void tokenise(std::istream &stream, std::deque<LemonPDF::Graphics::GFXToken> &tokens)
	{
		using namespace LemonPDF::Tokenisation;
		using LemonPDF::Graphics::GFXToken;
		
		while (1)
		{
			size_t offset = stream.tellg();
			char ch;
			if (!stream.get(ch))
			{
				tokens.emplace_back(offset, GFXToken::kEndOfFile);
				break;
			}
			
			switch (ch)
			{
				case '<':
				{
					char next = '\0';
					if (!stream.get(next))
						throw LemonPDF::Graphics::GFXTokeniserException("Unexpected syntax", stream);
					
					if (next == '<')
					{
						tokens.emplace_back(offset, GFXToken::kDictStart, "<<");
					}
					else
					{
						// go back
						stream.seekg(offset);
						tokens.emplace_back(offset, GFXToken::kText, TokeniseHexString(stream));
					}
					break;
				}

				case '>':
				{
					char next;
					if (!stream.get(next))
						throw LemonPDF::Graphics::GFXTokeniserException("Unexpected read error", stream);
					
					if (next != '>')
						throw LemonPDF::Graphics::GFXTokeniserException("Unexpected token", stream);
					
					tokens.emplace_back(offset, GFXToken::kDictEnd, ">>");
					break;
				}

				case '%':
				{
					// read a comment (ignore everything until the end of the line)
					while (stream.get(ch) && ch != '\r' && ch != '\n');
					continue;
				}
				
				case '+':
				case '-':
				case '.':
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
				{
					stream.putback(ch);
					std::pair<bool, std::string> t = TokeniseNumber(stream);
					tokens.emplace_back(offset, GFXToken::kNumber, t.second);
					break;
				}
				
				case '(':
				{
					tokens.emplace_back(offset, GFXToken::kText, TokeniseString(stream));
					break;
				}
				
				case '[':
				{
					tokens.emplace_back(offset, GFXToken::kArrayStart, "[");
					break;
				}
				
				case ']':
				{
					tokens.emplace_back(offset, GFXToken::kArrayEnd, "]");
					break;
				}
				
				case '/':
				{
					stream.putback(ch);
					tokens.emplace_back(offset, GFXToken::kName, TokeniseName(stream));
					break;
				}

				case '\0':
				case ' ':
				case '\r':
				case '\n':
				case '\t':
				{
					continue;
				}
				
				/**
				 * try parsing an operator
				 */
				default:
				{
					stream.putback(ch);
					std::string result = parseOperator(stream);
					if (result.size())
						tokens.emplace_back(offset, GFXToken::kOperator, result);
					else
						throw LemonPDF::Graphics::GFXTokeniserException("Unable to tokenise input stream...", stream);
				}
			}
			
			break;
		}
	}
}

LemonPDF::Graphics::GFXToken::GFXToken(size_t offset, enum TokenType type, const std::string &text)
  : m_offset(offset), m_type(type), m_text(text)
{
}

size_t LemonPDF::Graphics::GFXToken::getOffset() const
{
	return m_offset;
}

enum LemonPDF::Graphics::GFXToken::TokenType LemonPDF::Graphics::GFXToken::getType() const
{
	return m_type;
}

std::string LemonPDF::Graphics::GFXToken::getNameValue() const
{
	if (m_type != TokenType::kName)
		throw LemonPDF::Graphics::GFXTokenException("Cannot get name value of non-name token");
	return m_text.substr(1);
}

const std::string &LemonPDF::Graphics::GFXToken::getText() const
{
	return m_text;
}

LemonPDF::Graphics::GFXTokeniser::GFXTokeniser(std::istream &input)
  : m_stream(input)
{
}

void LemonPDF::Graphics::GFXTokeniser::consume()
{
	m_tokens.pop_front();
}

LemonPDF::Graphics::GFXToken LemonPDF::Graphics::GFXTokeniser::getNextToken()
{
	if (m_tokens.size() == 0)
		tokenise(m_stream, m_tokens);
	
	GFXToken result = m_tokens.front();
	m_tokens.pop_front();
	return result;
}

LemonPDF::Graphics::GFXToken LemonPDF::Graphics::GFXTokeniser::peekNextToken()
{
	if (m_tokens.size() == 0)
		tokenise(m_stream, m_tokens);
	
	return m_tokens.front();
}

LemonPDF::Graphics::GFXTokeniserException::GFXTokeniserException(const std::string &msg, std::istream &input)
  : PDFException("")
{
	std::ostringstream ss;
	ss << msg << ": ... ";
	
	size_t seekTo = input.tellg();
	if (seekTo < 10)
		seekTo = 0;
	else
		seekTo -= 10;
	
	for (size_t i = 0; i < 50; i++)
	{
		char ch;
		if (!input.get(ch))
			break;
		ss.put(ch);
	}
	
	m_msg = ss.str();
}

LemonPDF::Graphics::GFXTokenException::GFXTokenException(const std::string &msg)
  : PDFException(msg)
{
}
