#include <bitset>
#include <cmath>
#include <libcore/logging.h>
#include "gfxtruetypefont.h"
#include "gfxtruetypefontglyphcollection.h"
#include "gfxtruetypefontglyphparser.h"
#include "gfxtruetypefonttablereader.h"
#include "pdfexception.h"

namespace
{
	uint16_t clearAllButMSB(uint16_t value)
	{
		std::bitset<16> bits { value };
		bool foundMSB = false;
		
		for (size_t i = 0; i < 16; i++)
		{
			if (!foundMSB)
			{
				if (bits.test(15 - i))
					foundMSB = true;
			}
			else
			{
				bits.reset(15 - i);
			}
		}

		return bits.to_ulong();
	}

	inline void writeUInt16(std::ostream &out, uint16_t val)
	{
		out.put((val >> 8) & 0xFF);
		out.put((val >> 0) & 0xFF);
	}

	inline void writeUInt32(std::ostream &out, uint32_t val)
	{
		writeUInt16(out, (val >> 16) & 0xFFFF);
		writeUInt16(out, (val >>  0) & 0xFFFF);
	}

	inline void putUInt16(std::vector<unsigned char> &data, std::vector<unsigned char>::size_type i, uint16_t value)
	{
		data[i + 0] = (value >> 8) & 0xFF;
		data[i + 1] = (value >> 0) & 0xFF;
	}

	inline void putUInt32(std::vector<unsigned char> &data, std::vector<unsigned char>::size_type i, uint32_t value)
	{
		putUInt16(data, i + 0, (value >> 16) & 0xFFFF);
		putUInt16(data, i + 2, (value >>  0) & 0xFFFF);
	}

	uint32_t calculateChecksum(const std::vector<unsigned char> &data)
	{
		uint32_t result = 0;
		
		LemonPDF::Graphics::TrueTypeFontTableReader reader(data);
		LemonPDF::Graphics::TrueTypeFontTableReader::BufferType::size_type max = reader.getData().size();
		LemonPDF::Graphics::TrueTypeFontTableReader::BufferType::size_type i;

		for (i = 0; i + 4 <= max; i += 4)
			result += reader.readUInt32();
		
		if (max > i)
		{
			uint32_t lastWord = 0;
			for (LemonPDF::Graphics::TrueTypeFontTableReader::BufferType::size_type j = 0; j < 4; j++)
			{
				lastWord <<= 8;
				if (i + j < max)
					lastWord |= reader.readUInt8();
			}

			result += lastWord;
		}

		return result;
	}

	std::map<std::string, LemonPDF::Graphics::TrueTypeFont::Table> readTables(const std::vector<unsigned char> &fontData)
	{
		std::map<std::string, LemonPDF::Graphics::TrueTypeFont::Table> result;
		
		LemonPDF::Graphics::TrueTypeFontTableReader fontReader(fontData);
		uint32_t header = fontReader.readUInt32();

		if (header != 0x74727565 && header != 0x00010000 && header != 0x4F54544F)
			throw LemonPDF::PDFInvalidOperationException("Unsupported font type");

		uint16_t numberOfTables = fontReader.readUInt16();

		fontReader.advance(sizeof(uint16_t) * 3); // skip over "searchRange", "entrySelector" and "rangeShift"
		
		for (uint16_t i = 0; i < numberOfTables; i++)
		{
			char tag[4] = {};
			for (size_t i = 0; i < sizeof tag; i++)
				tag[i] = fontReader.readUInt8();
			std::string tableName(tag, tag + 4);
			
			fontReader.advance(sizeof(uint32_t)); // skip over "checksum"
			uint32_t offset = fontReader.readUInt32();
			uint32_t length = fontReader.readUInt32();

			const std::vector<unsigned char>::const_iterator start = fontData.begin() + offset;
			const std::vector<unsigned char>::const_iterator end = start + length;
			result.emplace(tableName, std::vector<unsigned char>(start, end));
		}
		
		return result;
	}

	struct MaximumProfile
	{
		MaximumProfile(LemonPDF::Graphics::TrueTypeFontTableReader &&maxp)
		{
			version = maxp.readUInt32();
			numberOfGlyphs = maxp.readUInt16();
		}

		uint32_t version;
		uint16_t numberOfGlyphs;
	};

	LemonPDF::Graphics::TrueTypeFont::FontHeader parseFontHeader(LemonPDF::Graphics::TrueTypeFontTableReader &&head)
	{
		LemonPDF::Graphics::TrueTypeFont::FontHeader fh;
		fh.majorVersion = head.readUInt16();
		fh.minorVersion = head.readUInt16();
		fh.fontRevision = head.readUInt32();
		fh.checkSumAdjustment = head.readUInt32();
		fh.magicNumber = head.readUInt32();
		fh.flags = head.readUInt16();
		fh.unitsPerEm = head.readUInt16();
		fh.created.high = head.readUInt32();
		fh.created.low = head.readUInt32();
		fh.modified.high = head.readUInt32();
		fh.modified.low = head.readUInt32();
		fh.xMin = head.readInt16();
		fh.yMin = head.readInt16();
		fh.xMax = head.readInt16();
		fh.yMax = head.readInt16();
		fh.macStyle = head.readUInt16();
		fh.lowestRecPPEM = head.readUInt16();
		fh.fontDirectionHint = head.readInt16();
		fh.indexToLocFormat = head.readInt16();
		fh.glyphDataFormat = head.readInt16();
		return fh;
	}
}

LemonPDF::Graphics::TrueTypeFont::Table::Table(const std::vector<unsigned char> &data)
  : m_checksum(0)
{
	setData(data);
}

void LemonPDF::Graphics::TrueTypeFont::Table::setData(const std::vector<unsigned char> &data)
{
	m_data = data;
	m_checksum = calculateChecksum(m_data);
}

uint32_t LemonPDF::Graphics::TrueTypeFont::Table::getChecksum() const
{
	return m_checksum;
}

const std::vector<unsigned char> &LemonPDF::Graphics::TrueTypeFont::Table::getData() const
{
	return m_data;
}

LemonPDF::Graphics::TrueTypeFont::TrueTypeFont(const std::string &name, enum FontCompleteness fc, enum FontType ft, int fw, const std::vector<unsigned char> &data)
  : Font(name, fc, ft, fw), m_tables(readTables(data)), m_glyphOutlinesLoaded(false)
{
	auto headIter = m_tables.find("head");
	if (headIter != m_tables.end())
		m_fontHeader = parseFontHeader(TrueTypeFontTableReader(headIter->second.getData()));

	// The presence of these tables means we can parse the glyphs
	// (particularly, "glyf" is needed, because the other type of
	// font outline data in "CFF" etc cannot be parsed presently
	auto maxpIter = m_tables.find("maxp");
	auto locaIter = m_tables.find("loca");
	auto glyfIter = m_tables.find("glyf");

	bool usable = maxpIter != m_tables.end() && locaIter != m_tables.end() && glyfIter != m_tables.end();
	setCanConvertGlyphIndicesToPaths(usable);

	try
	{
		parseCMap();
	}
	catch (const PDFInvalidOperationException &)
	{
		// CMap may be available from another part of the PDF, so
		// not an error (or even a warning really).
	}
}

LemonPDF::Graphics::TrueTypeFont::~TrueTypeFont()
{
}

void LemonPDF::Graphics::TrueTypeFont::enableEmbeddability()
{
	std::vector<unsigned char> tableData;
	std::map<std::string, Table>::iterator i = m_tables.find("OS/2");
	
	if (i == m_tables.end())
	{
		tableData.resize(78); // length of version 0 OS/2 table

		putUInt16(tableData,  4, 400); // weight class
		putUInt16(tableData,  6,   5); // width class
		putUInt16(tableData,  8,   0); // licensing rights
		putUInt32(tableData, 58, 0x20202020); // vendor tag
		putUInt16(tableData, 68,   1); // typo-ascender (can't be zero)
		m_tables.emplace("OS/2", Table(tableData));
	}
	else
	{
		tableData = i->second.getData();
		putUInt16(tableData,  8,   0); // licensing rights
		i->second.setData(tableData);
	}
}

const LemonPDF::Graphics::GlyphOutlineMap &LemonPDF::Graphics::TrueTypeFont::getGlyphOutlineMap() const
{
	if (!m_glyphOutlinesLoaded)
	{
		auto maxpIter = m_tables.find("maxp");
		auto locaIter = m_tables.find("loca");
		auto glyfIter = m_tables.find("glyf");

		if (maxpIter == m_tables.end() || locaIter == m_tables.end() || glyfIter == m_tables.end())
			throw LemonPDF::PDFInvalidOperationException("Required tables missing in font (one or more of: maxp, head, loca, glyf)");

		MaximumProfile maximumProfile(TrueTypeFontTableReader(maxpIter->second.getData()));

		TrueTypeFontTableReader locaReader(locaIter->second.getData());
		std::vector<uint32_t> offsets(maximumProfile.numberOfGlyphs + 1); // extra offset at end for calculating glyf entry size
		if (m_fontHeader.indexToLocFormat == 0)
		{
			for (uint16_t i = 0; i < offsets.size(); i++)
				offsets[i] = locaReader.readUInt16() * 2U;
		}
		else
		{
			for (uint16_t i = 0; i < offsets.size(); i++)
				offsets[i] = locaReader.readUInt32();
		}
		
		TrueTypeFontTableReader glyfReader(glyfIter->second.getData());
		TrueTypeFontGlyphParser glyphParser(m_fontHeader.unitsPerEm);
		TrueTypeFontGlyphCollection allGlyphs = glyphParser.parseGlyfTable(offsets, glyfReader);

		allGlyphs.composeAllGlyphs();
		for (uint32_t i = 0; i < maximumProfile.numberOfGlyphs; i++)
		{
			auto chIter = m_glyphMapping.find(i);
			if (chIter != m_glyphMapping.end())
				m_glyphOutlines[chIter->second] = allGlyphs.glyphForID(i);
		}
		
		m_glyphOutlinesLoaded = true;
	}

	return m_glyphOutlines;
}

const LemonPDF::Graphics::TrueTypeFont::Table &LemonPDF::Graphics::TrueTypeFont::getTable(const std::string &name) const
{
	if (name.size() != 4)
		throw PDFInvalidArgumentException("name");
	
	std::map<std::string, Table>::const_iterator i = m_tables.find(name);
	if (i == m_tables.end())
		throw PDFInvalidArgumentException("name");

	return i->second;
}

bool LemonPDF::Graphics::TrueTypeFont::hasTable(const std::string &name) const
{
	if (name.size() != 4)
		throw LemonPDF::PDFInvalidArgumentException("name");
	return m_tables.find(name) != m_tables.end();
}

void LemonPDF::Graphics::TrueTypeFont::parseCMap()
{
	auto i = m_tables.find("cmap");
	if (i == m_tables.end())
		return;
	
	TrueTypeFontTableReader tableReader(i->second.getData());

	uint16_t version = tableReader.readUInt16();
	if (version != 0x0000)
		StripedHorse::Logging::Debug() << "Unsupported cmap table version:" << version << std::endl;
	
	uint16_t numberOfTables = tableReader.readUInt16();
	uint32_t offsetOfDesiredTable = 0;
	
	for (uint16_t i = 0; i < numberOfTables && offsetOfDesiredTable == 0; i++)
	{
		uint16_t platformID = tableReader.readUInt16();
		tableReader.advance(sizeof(uint16_t)); // skip over encodingID
		uint32_t offset = tableReader.readUInt32(); // need to read the offset anyway

		if (platformID == 1 || platformID == 3) // Mac or Windows
			offsetOfDesiredTable = offset;
	}

	if (offsetOfDesiredTable == 0)
	{
		StripedHorse::Logging::Debug() << "No suitable cmap found" << std::endl;
		return;
	}

	tableReader.seek(offsetOfDesiredTable);
	uint16_t format = tableReader.readUInt16();

	switch (format)
	{
		// byte encoding (character code maps to glyph ID)
		case 0:
		{
			tableReader.advance(sizeof(uint16_t) * 2); // skip over "length" and "language" fields
			for (char32_t i = 0; i < 256; i++)
			{
				uint8_t characterCode = tableReader.readUInt8();
				m_glyphMapping[characterCode] = i;
			}

			break;
		}

		// Segment mapping to delta values
		case 4:
		{
			tableReader.advance(sizeof(uint16_t) * 2); // skip over "length" and "language" fields

			uint16_t segmentCount = tableReader.readUInt16() / 2;

			if (segmentCount < 2)
				throw LemonPDF::PDFInvalidOperationException("Segment count too small");
			
			tableReader.advance(sizeof(uint16_t) * 3); // skip over "searchRange", "entrySelector" and "rangeShift" fields

			std::vector<uint16_t> endCodes(segmentCount);
			std::vector<uint16_t> startCodes(segmentCount);
			std::vector<int16_t> idDeltas(segmentCount);
			
			for (uint16_t i = 0; i < segmentCount; i++)
				endCodes[i] = tableReader.readUInt16();

			tableReader.advance(sizeof(uint16_t)); // padding
			
			for (uint16_t i = 0; i < segmentCount; i++)
				startCodes[i] = tableReader.readUInt16();
			
			for (uint16_t i = 0; i < segmentCount; i++)
				idDeltas[i] = tableReader.readInt16();

			for (uint16_t i = 0; i < segmentCount; i++)
			{
				for (uint16_t code = startCodes.at(i); code <= endCodes.at(i); code++)
				{
					if (code == 0xFFFF) // specific terminal value
						break;
					
					char ch = static_cast<char>(code + idDeltas.at(i));
					m_glyphMapping[ch] = code;
				}
			}

			break;
		}

		// Trimmed table mapping
		case 6:
		{
			tableReader.advance(sizeof(uint16_t) * 2); // skip over "length" and "language" fields

			char16_t firstCode = static_cast<char16_t>(tableReader.readUInt16());
			uint16_t codeCount = tableReader.readUInt16();
			
			for (uint16_t i = 0; i < codeCount; i++)
			{
				char16_t thisCode = static_cast<char16_t>(firstCode + i);
				uint16_t glyphCode = tableReader.readUInt16();
				if (glyphCode != 0)
					m_glyphMapping[glyphCode] = thisCode;
			}

			break;
		}

		default:
		{
			// only support format 4 and 6 for now
			StripedHorse::Logging::Debug() << "Unsupported cmap table format: " << format << std::endl;
		}
	}
}

void LemonPDF::Graphics::TrueTypeFont::setTable(const std::string &name, const Table &table)
{
	if (name.size() != 4)
		throw LemonPDF::PDFInvalidArgumentException("name");
	m_tables.emplace(name, table);
}

void LemonPDF::Graphics::TrueTypeFont::write(std::ostream &out) const
{
	out.write("\x00\x01\x00\x00", 4);
	
	uint16_t numTables = m_tables.size();
	uint16_t maxPowerOfTwoOfNumTables = clearAllButMSB(numTables);
	uint16_t searchRange = maxPowerOfTwoOfNumTables * 16UL;
	uint16_t entrySelector = static_cast<uint16_t>(std::log2(maxPowerOfTwoOfNumTables));
	uint16_t rangeShift = numTables * 16UL - searchRange;
	
	writeUInt16(out, numTables);
	writeUInt16(out, searchRange);
	writeUInt16(out, entrySelector);
	writeUInt16(out, rangeShift);
	
	uint32_t runningOffset = 0; // size of header data

	const uint32_t headerSize = 12;
	const uint32_t entrySize = 16;
	
	for (const std::pair<std::string, Table> &table : m_tables)
	{
		uint32_t dataOffset = headerSize + m_tables.size() * entrySize + runningOffset;
		uint32_t dataSize = table.second.getData().size();
		
		out << table.first;
		writeUInt32(out, table.second.getChecksum());
		writeUInt32(out, dataOffset);
		writeUInt32(out, dataSize);
		
		runningOffset += (dataSize + 3) & ~static_cast<uint32_t>(3);
	}

	for (const std::pair<std::string, Table> &table : m_tables)
	{
		const std::vector<unsigned char> &data = table.second.getData();
		out.write(reinterpret_cast<const char *>(&data[0]), data.size());

		switch (table.second.getData().size() % 4)
		{
			case 1:
				out.put(0x00);
				// fallthrough
			case 2:
				out.put(0x00);
				// fallthrough
			case 3:
				out.put(0x00);
		}
	}
}

