#include <libcore/logging.h>
#include <libcore/types.h>
#include <lemongfx/bezier.h>
#include "gfxtruetypefontglyphparser.h"
#include "pdfexception.h"

namespace
{
	template<typename T>
	inline const T& clamp(const T &v, const T &low, const T &high)
	{
		if (v < low)
			return low;
		if (high < v)
			return high;
		return v;
	}

	enum SimpleGlyphFlag : uint8_t
	{
		kOnCurvePoint      = (1 << 0),
		kXIsShort          = (1 << 1),
		kYIsShort          = (1 << 2),
		kRepeatFlag        = (1 << 3),
		kXIsSame           = (1 << 4),
		kYIsSame           = (1 << 5),
		kSimpleReserved    = (3 << 6), // top bits are reserved

		// Composed flags
		kXIsPositiveShort  = kXIsShort | kXIsSame,
		kXIsNegativeShort  = kXIsShort,
		kYIsPositiveShort  = kYIsShort | kYIsSame,
		kYIsNegativeShort  = kYIsShort,

		kXFlags            = kXIsPositiveShort,
		kYFlags            = kYIsPositiveShort,

		kNoFlags           = 0
	};

	enum CompositeGlyphFlag : uint16_t
	{
		kArg1And2AreWords        = (1 <<  0),
		kArgsAreXYValues         = (1 <<  1),
		kRoundXYToGrid           = (1 <<  2),
		kWeHaveAScale            = (1 <<  3),
		kMoreComponents          = (1 <<  5),
		kWeHaveAnXAndYScale      = (1 <<  6),
		kWeHaveATwoByTwo         = (1 <<  7),
		kWeHaveInstructions      = (1 <<  8),
		kUseMyMetrics            = (1 <<  9),
		kOverlapCompound         = (1 << 10),
		kScaledComponentOffset   = (1 << 11),
		kUnscaledComponentOffset = (1 << 12),

		kCompositeReserved       = (0xE010)
	};

	struct AnnotatedPoint
	{
		AnnotatedPoint()
		  : onCurve(false)
		{
		}

		AnnotatedPoint(const LemonPDF::Point &point, bool onCurve)
		  : point(point), onCurve(onCurve)
		{
		}

		LemonPDF::Point point;
		bool onCurve;
	};

	using AnnotatedPoints = std::vector<AnnotatedPoint>;

	std::vector<AnnotatedPoints> interpolateContours(const AnnotatedPoints &coordinates, const std::vector<uint16_t> &endPointsOfContours)
	{
		std::vector<AnnotatedPoints> interpolatedContours;
		uint16_t lastEndPoint = 0;
		for (uint16_t c = 0; c < endPointsOfContours.size(); c++)
		{
			AnnotatedPoints thisContour;
			AnnotatedPoint lastPoint = coordinates[endPointsOfContours[c]];
			for (uint16_t p = lastEndPoint; p <= endPointsOfContours[c]; p++)
			{
				const AnnotatedPoint &thisPoint = coordinates[p];
				if (!thisPoint.onCurve && !lastPoint.onCurve)
					thisContour.emplace_back(LemonGFX::Bezier::MidPoint(lastPoint.point, thisPoint.point), true);
				thisContour.emplace_back(thisPoint);
				lastPoint = thisPoint;
			}

			interpolatedContours.emplace_back(std::move(thisContour));
			lastEndPoint = endPointsOfContours[c] + 1;
		}

		return interpolatedContours;
	}

	LemonPDF::Graphics::TrueTypeFontGlyphPtr parseSimpleGlyph(int16_t numberOfContours, uint16_t unitsPerEm, LemonPDF::Graphics::TrueTypeFontTableReader &glyfTable)
	{
		std::vector<uint16_t> endPointsOfContours(numberOfContours);
		for (uint16_t n = 0; n < numberOfContours; n++)
			endPointsOfContours[n] = glyfTable.readUInt16();
		
		uint16_t numberOfInstructions = glyfTable.readUInt16();
		glyfTable.advance(numberOfInstructions); // gloss over instructions (not needed for approximation)
		
		// some glyphs are legitimately empty (e.g. space character)
		if (numberOfContours == 0)
			return std::make_shared<LemonPDF::Graphics::TrueTypeFontGlyph>();

		uint16_t maxPoints = endPointsOfContours.back() + 1;
		std::vector<SimpleGlyphFlag> flags(maxPoints);
		std::vector<AnnotatedPoint> coordinates(maxPoints);
		std::vector<int16_t> xCoordinates(maxPoints);
		std::vector<int16_t> yCoordinates(maxPoints);

		for (uint16_t i = 0; i < maxPoints; i++)
		{
			SimpleGlyphFlag flag = static_cast<SimpleGlyphFlag>(glyfTable.readUInt8());
			flags[i] = flag;
			coordinates[i].onCurve = flag & SimpleGlyphFlag::kOnCurvePoint;
			if (flag & SimpleGlyphFlag::kRepeatFlag)
			{
				uint8_t repeatCount = glyfTable.readUInt8();
				for (uint8_t r = 0; r < repeatCount; r++)
				{
					i++;
					flags[i] = flag;
					coordinates[i].onCurve = flag & SimpleGlyphFlag::kOnCurvePoint;
				}
			}
		}

		int16_t currentX = 0;
		for (uint16_t i = 0; i < maxPoints; i++)
		{
			switch (flags[i] & SimpleGlyphFlag::kXFlags)
			{
				case SimpleGlyphFlag::kNoFlags:
					currentX += glyfTable.readInt16();
					break;

				case SimpleGlyphFlag::kXIsNegativeShort:
					currentX -= static_cast<int16_t>(glyfTable.readUInt8());
					break;
				
				case SimpleGlyphFlag::kXIsPositiveShort:
					currentX += static_cast<int16_t>(glyfTable.readUInt8());
					break;

				case SimpleGlyphFlag::kXIsSame:
					// do nothing
					break;
				
				default:
					throw LemonPDF::PDFParseException("This should not have happened!");
			}

			coordinates[i].point.x = 1000.0 * currentX / unitsPerEm;
		}

		int16_t currentY = 0;
		for (uint16_t i = 0; i < maxPoints; i++)
		{
			switch (flags[i] & SimpleGlyphFlag::kYFlags)
			{
				case SimpleGlyphFlag::kNoFlags:
					currentY += glyfTable.readInt16();
					break;

				case SimpleGlyphFlag::kYIsNegativeShort:
					currentY -= static_cast<int16_t>(glyfTable.readUInt8());
					break;
				
				case SimpleGlyphFlag::kYIsPositiveShort:
					currentY += static_cast<int16_t>(glyfTable.readUInt8());
					break;

				case SimpleGlyphFlag::kYIsSame:
					// Do nothing
					break;
				
				default:
					throw LemonPDF::PDFParseException("This should not have happened!");
			}

			coordinates[i].point.y = 1000.0 * currentY / unitsPerEm;
		}
		
		LemonPDF::Graphics::Path result;
		for (const AnnotatedPoints &thisContour : interpolateContours(coordinates, endPointsOfContours))
		{
			size_t firstOnPoint = 0;
			for (AnnotatedPoints::size_type i = 0; i < thisContour.size(); i++)
			{
				if (thisContour[i].onCurve)
				{
					result.move(thisContour[i].point);
					firstOnPoint = i;
					break;
				}
			}
			
			for (AnnotatedPoints::size_type i = 0; i < thisContour.size(); i++)
			{
				const AnnotatedPoint &thisPoint = thisContour[(firstOnPoint + i) % thisContour.size()];
				if (thisPoint.onCurve)
				{
					result.appendLine(thisPoint.point);
				}
				else
				{
					AnnotatedPoint endPoint = thisContour[(firstOnPoint + i + 1) % thisContour.size()];
					result.appendBezierCurve(thisPoint.point, endPoint.point);
					i++;
				}
			}
		
			result.closePath();
		}

		return std::make_shared<LemonPDF::Graphics::TrueTypeFontGlyph>(result);
	}

	LemonPDF::Graphics::TrueTypeFontGlyphPtr parseCompositeGlyphData(LemonPDF::Graphics::TrueTypeFontTableReader &glyfTable)
	{
		std::vector<LemonPDF::Graphics::CompositeGlyphInfo> infos;
		uint16_t flags;

		do
		{
			flags = glyfTable.readUInt16();
			uint16_t glyphIndex = glyfTable.readUInt16();
			LemonPDF::Graphics::CompositeGlyphInfo info;
			info.glyphIndex = glyphIndex;
			info.xOffset = 0;
			info.yOffset = 0;

			if (flags & CompositeGlyphFlag::kArg1And2AreWords)
			{
				glyfTable.advance(sizeof(uint16_t) * 2); //ignore
			}
			else
			{
				glyfTable.advance(sizeof(uint8_t) * 2); //ignore
			}

			if (flags & CompositeGlyphFlag::kWeHaveAScale)
			{
				glyfTable.advance(2);
			}
			else if (flags & CompositeGlyphFlag::kWeHaveAnXAndYScale)
			{
				glyfTable.advance(4);
			}
			else if (flags & CompositeGlyphFlag::kWeHaveATwoByTwo)
			{
				glyfTable.advance(8);
			}

			infos.emplace_back(info);
		}
		while (flags & CompositeGlyphFlag::kMoreComponents);

		if (flags & CompositeGlyphFlag::kWeHaveInstructions)
		{
			uint16_t numberOfInstructions = glyfTable.readUInt16();
			StripedHorse::Logging::Warning() << "Ignoring " << numberOfInstructions << " TTF instructions for composite glyph" << std::endl;
			glyfTable.advance(numberOfInstructions);
		}

		return std::make_shared<LemonPDF::Graphics::TrueTypeFontGlyph>(infos);
	}
}

LemonPDF::Graphics::TrueTypeFontGlyphParser::TrueTypeFontGlyphParser(uint16_t unitsPerEm)
  : m_unitsPerEm(unitsPerEm)
{
}

LemonPDF::Graphics::TrueTypeFontGlyphCollection LemonPDF::Graphics::TrueTypeFontGlyphParser::parseGlyfTable(const std::vector<uint32_t> &offsets, TrueTypeFontTableReader &glyfTable)
{
	if (offsets.size() < 2)
		throw PDFParseException("At least two offsets should be present");
	
	const uint32_t numberOfGlyphs = offsets.size() - 1;

	TrueTypeFontGlyphCollection result;
	result.reserve(numberOfGlyphs);
	
	for (uint16_t i = 0; i < numberOfGlyphs; i++)
	{
		TrueTypeFontGlyphPtr glyph;
		uint32_t offset = offsets[i];
		uint32_t nextOffset = offsets[i + 1];

		if (nextOffset > offset && nextOffset - offset > 0)
		{
			glyfTable.seek(offset);

			int16_t numberOfContours = glyfTable.readInt16();
			Rectangle bounds;
			bounds.lowerLeft.x = 1000.0 * glyfTable.readInt16() / m_unitsPerEm;
			bounds.lowerLeft.y = 1000.0 * glyfTable.readInt16() / m_unitsPerEm;
			bounds.upperRight.x = 1000.0 * glyfTable.readInt16() / m_unitsPerEm;
			bounds.upperRight.y = 1000.0 * glyfTable.readInt16() / m_unitsPerEm;

			if (numberOfContours > 0)
				glyph = parseSimpleGlyph(numberOfContours, m_unitsPerEm, glyfTable);
			else if (numberOfContours < 0)
				glyph = parseCompositeGlyphData(glyfTable);

			glyph->setBounds(bounds);
		}

		if (glyph != nullptr)
			result.addGlyph(i, glyph);
	}
	
	return result;
}
