#include <cstdlib>
#include "pdfexception.h"
#include "libtokenisation.h"

namespace
{
	bool ishex(char ch)
	{
		return (ch >= '0' && ch <= '9') || (ch >= 'A' && ch <= 'F') || (ch >= 'a' && ch <= 'f');
	}
	
	bool isoct(char ch)
	{
		return (ch >= '0' && ch <= '7');
	}
	
	int fromhex(char ch)
	{
		if (ch >= '0' && ch <= '9')
			return ch - '0';
		else if (ch >= 'A' && ch <= 'F')
			return ch - 'A' + 10;
		else if (ch >= 'a' && ch <= 'f')
			return ch - 'a' + 10;
		return 0;
	}
	
	int fromoct(char ch)
	{
		if (ch >= '0' && ch <= '7')
			return ch - '0';
		return 0;
	}
}

std::istream &LemonPDF::Tokenisation::ReadLine(std::istream &in, std::string &out)
{
	out.clear();
	char buffer[256];

	do
	{
		in.read(buffer, sizeof buffer);
		std::streamsize bytesRead = in.gcount();
		if (bytesRead == 0)
			break;
		
		for (std::streamsize i = 0; i < bytesRead; i++)
		{
			bool lineEnded = false;
			bool skipTwo = false;
			if (buffer[i] == '\r')
			{
				if (i == (bytesRead - 1))
				{
					if (in.peek() == '\n')
						skipTwo = true;
				}
				else
				{
					if (buffer[i + 1] == '\n')
						skipTwo = true;
				}

				lineEnded = true;
			}
			else if (buffer[i] == '\n')
			{
				lineEnded = true;
			}

			if (lineEnded)
			{
				out.insert(out.end(), buffer, buffer + i);
				
				// in case we hit EOF, we need to clear so we can seek before EOF again
				in.clear();
				in.seekg(-(bytesRead - i - (skipTwo ? 2 : 1)), std::ios_base::cur);
				return in;
			}
		}

		out.insert(out.end(), buffer, buffer + bytesRead);
	}
	while (in);
	
	return in;
}

bool LemonPDF::Tokenisation::IsDelimeter(char ch)
{
	switch (ch)
	{
		case '(':
		case ')':
		case '<':
		case '>':
		case '[':
		case ']':
		case '{':
		case '}':
		case '/':
		case '%':
			return true;
		
		default:
			return false;
	}
}

std::string LemonPDF::Tokenisation::TokeniseHexString(std::istream &input)
{
	std::string result;
	
	int nibbles[2];
	int nc = 0;
	char ch;
	while (input.get(ch))
	{
		if (ch == '>')
			break;
		
		if (ishex(ch))
			nibbles[nc++] = fromhex(ch);
		
		if (nc == 2)
		{
			result.push_back(((nibbles[0] & 0xF) << 4) | (nibbles[1] & 0xF));
			nc = 0;
		}
	}
	
	if (nc)
		throw LemonPDF::PDFSyntaxException("Odd number of characters in hex string", input);
	
	if (input.eof())
		throw LemonPDF::PDFSyntaxException("Unexpected end of file while parsing hex string", input);
	
	return result;
}

std::string LemonPDF::Tokenisation::TokeniseName(std::istream &input)
{
	std::string result;
	char ch;
	
	if (!input.get(ch))
		throw LemonPDF::PDFIOException("Read error");
	
	if (ch != '/')
		throw LemonPDF::PDFSyntaxException("Invalid name", input);
	
	result.push_back(ch);
	
	while (input.get(ch))
	{
		if (isspace(ch) || IsDelimeter(ch))
		{
			input.putback(ch);
			break;
		}
		
		if (ch == '#')
		{
			int nibbles[2];
			for (int i = 0; i < 2; i++)
			{
				char nch;
				if (!input.get(nch))
					throw LemonPDF::PDFIOException("Read error");
				if (!ishex(nch))
					throw LemonPDF::PDFSyntaxException("Invalid hex sequence in name", input);
				nibbles[i] = fromhex(nch);
			}
			
			result.push_back(((nibbles[0] & 0xF) << 4) | (nibbles[1] & 0xF));
		}
		else
		{
			result.push_back(ch);
		}
	}
	
	if (!input)
		throw LemonPDF::PDFIOException("Read error");
	
	return result;
}

std::pair<bool, std::string> LemonPDF::Tokenisation::TokeniseNumber(std::istream &input)
{
	std::string result;
	
	char ch;
	bool stop = false;
	while (!stop && input.get(ch))
	{
		switch (ch)
		{
			case '+':
			case '-':
			case '.':
			case 'e':
			case 'E':
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				result.push_back(ch);
				break;
			
			default:
				input.putback(ch);
				stop = true;
				break;
		}
	}
	
	char *end = nullptr;
	const char *begin = result.c_str();
	strtoll(begin, &end, 10);
	if (end == begin + result.size())
		return std::make_pair(true, result);

	end = nullptr;
	strtod(begin, &end);
	if (end == begin + result.size())
		return std::make_pair(false, result);
	
	throw LemonPDF::PDFSyntaxException("Invalid number literal", input);
}

std::string LemonPDF::Tokenisation::TokeniseString(std::istream &input)
{
	std::string result;
	
	int brackets = 1;
	bool escape = false;
	int oct = 0;
	char ch;

	// From the PDF 1.7 Reference (p. 15):
	// 
	// A conforming writer may split a literal string across multiple lines.
	// The REVERSE SOLIDUS (5Ch) (backslash character) at the end of a line
	// shall be used to indicate that the string continues on the following
	// line. A conforming reader shall disregard the REVERSE SOLIDUS and the
	// end-of-line marker following it when reading the string; the resulting
	//  string value shall be identical to that which would be read if the
	// string were not split.
	
	while (input.get(ch))
	{
		if (escape)
		{
			switch (ch)
			{
				case '\r':
					// an escaped end-of-line should be ignored by a tokeniser
					if (input.peek() == '\n')
						input.get();
					break;

				case '\n':
					break;

				case 'a':
					result.push_back('\a');
					break;
				
				case 'b':
					result.push_back('\b');
					break;
				
				case 'f':
					result.push_back('\f');
					break;
				
				case 'n':
					result.push_back('\n');
					break;
				
				case 'r':
					result.push_back('\r');
					break;
				
				case 't':
					result.push_back('\t');
					break;
				
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
					oct = ch - '0';
					for (int i = 0; i < 2; i++)
					{
						if (!input.get(ch))
							throw LemonPDF::PDFIOException("Read error");
						if (!isoct(ch))
							throw LemonPDF::PDFSyntaxException("Invalid octal constant", input);
						oct *= 8;
						oct += fromoct(ch);
					}
					result.push_back((char) oct);
					break;
				
				default:
					result.push_back(ch);
			}
			
			escape = false;
		}
		else
		{
			switch (ch)
			{
				case '(':
					brackets++;
					break;
				
				case ')':
					brackets--;
					break;
				
				case '\\':
					escape = true;
					continue;
			}
			
			if (brackets > 0)
				result.push_back(ch);
		}
		
		if (brackets == 0)
			break;
	}
	
	return result;
}

