#include <iterator>
#include <sstream>
#include "pdfexception.h"
#include "pdfobject.h"
#include "pdftokeniser.h"

LemonPDF::ArrayPtr LemonPDF::Array::Parse(Tokeniser &tokeniser)
{
	Token t = tokeniser.peekNextToken();
	if (t.getType() != Token::kArrayStart)
		throw PDFSyntaxException("Array parse error", tokeniser.getInputStream());
	
	tokeniser.consume();
	
	ArrayPtr result = std::make_shared<Array>();
	
	while (1)
	{
		t = tokeniser.peekNextToken();
		if (t.getType() == Token::kArrayEnd)
		{
			tokeniser.consume();
			break;
		}
		else if (t.getType() == Token::kEndOfFile)
		{
			throw PDFSyntaxException("Unexpected end of file", tokeniser.getInputStream());
		}
		
		ObjectPtr o = Object::Parse(tokeniser);
		result->m_objects.push_back(o);
	}
	
	return result;
}

LemonPDF::Array::~Array()
{
}

LemonPDF::Array::operator std::string() const
{
	std::ostringstream ss;
	ss << "<Array, ";
	ss << m_objects.size();
	ss << " entries>";
	return ss.str();
}

void LemonPDF::Array::add(ObjectPtr o)
{
	m_objects.push_back(o);
}

LemonPDF::ObjectPtr LemonPDF::Array::copy() const
{
	ArrayPtr result = std::make_shared<Array>();
	enumerate([&result](size_t index, ObjectPtr object){
		(void) index;
		result->add(object->copy());
	});
	return result;
}

void LemonPDF::Array::enumerate(std::function<void(size_t, ObjectPtr)> func) const
{
	for (size_t i = 0; i < m_objects.size(); i++)
		func(i, m_objects[i]);
}

void LemonPDF::Array::enumerate(std::function<void(ObjectPtr)> func) const
{
	for (ObjectPtr o : m_objects)
		func(o);
}

LemonPDF::ObjectPtr LemonPDF::Array::get(size_t index) const
{
	if (index >= m_objects.size())
		throw PDFInvalidOperationException("Index is out of range");
	return m_objects[index];
}

size_t LemonPDF::Array::getSize() const
{
	return m_objects.size();
}

void LemonPDF::Array::insert(size_t i, ObjectPtr o)
{
	auto it = m_objects.begin();
	std::advance(it, i);
	m_objects.insert(it, o);
}

void LemonPDF::Array::remove(size_t i)
{
	m_objects.erase(m_objects.begin() + i);
}

LemonPDF::ArrayPtr LemonPDF::Array::shallowCopy() const
{
	ArrayPtr result = std::make_shared<Array>();
	result->m_objects = m_objects;
	return result;
}

void LemonPDF::Array::truncate(size_t i)
{
	if (i > m_objects.size())
		throw std::out_of_range("Cannot truncate array, not enough elements");

	m_objects.erase(m_objects.begin() + i, m_objects.end());
}

void LemonPDF::Array::write(std::ostream &out, size_t depth) const
{
	if (depth == 0)
	{
		out << "[";
		for (size_t i = 0; i < m_objects.size(); i++)
		{
			ObjectPtr object = m_objects[i];
			if (i > 0)
			{
				LiteralPtr literal = std::dynamic_pointer_cast<Literal>(object);
				if (literal)
				{
					Literal::Type t = literal->getType();
					if (t != Literal::Type::kName && t != Literal::Type::kString)
						out << " ";
				}
			}

			object->write(out, 0);
		}
		out << "]";
	}
	else
	{
		std::string prefix((depth - 1) * 2, ' ');
		std::string indent(depth * 2, ' ');

		out << "[\n";

		for (auto o : m_objects)
		{
			out << indent;
			o->write(out, depth + 1);
			out << "\n";
		}

		out << prefix << "]";
	}
}
