#include <set>
#include <sstream>
#include "pdfexception.h"
#include "pdfobject.h"
#include "pdftokeniser.h"

LemonPDF::DictionaryPtr LemonPDF::Dictionary::Parse(Tokeniser &tokeniser)
{
	DictionaryPtr result = std::make_shared<Dictionary>();
	
	Token t = tokeniser.peekNextToken();
	if (t.getType() != Token::kDictStart)
		throw PDFSyntaxException("Dictionary parse error", tokeniser.getInputStream());
	
	tokeniser.consume();
	
	while (1)
	{
		Token t = tokeniser.peekNextToken();
		
		if (t.getType() == Token::kDictEnd)
		{
			tokeniser.consume();
			return result;
		}
		
		if (t.getType() == Token::kEndOfFile)
			throw PDFSyntaxException("Unexpected end of file", tokeniser.getInputStream());
		
		if (t.getType() != Token::kName)
			throw PDFSyntaxException("Expected name object", tokeniser.getInputStream());
		
		LiteralPtr name = std::dynamic_pointer_cast<Literal>(Object::Parse(tokeniser));
		if (!name)
			throw PDFSyntaxException("Expected name object, or name object invalid", tokeniser.getInputStream());
		
		ObjectPtr value = Object::Parse(tokeniser);
		result->m_objects[name->getNameValue()] = value;
	}
}

LemonPDF::Dictionary::~Dictionary()
{
}

LemonPDF::Dictionary::operator std::string() const
{
	std::ostringstream ss;
	ss << "<Dictionary, ";
	ss << m_objects.size();
	ss << " entries>";
	return ss.str();
}

void LemonPDF::Dictionary::add(const std::string &name, ObjectPtr o)
{
	m_objects[name] = o;
}

void LemonPDF::Dictionary::add(const LiteralPtr name, ObjectPtr o)
{
	add(name->getNameValue(), o);
}

bool LemonPDF::Dictionary::contains(const std::string &name) const
{
	return m_objects.find(name) != m_objects.end();
}

bool LemonPDF::Dictionary::contains(const LiteralPtr name) const
{
	return contains(name->getNameValue());
}

LemonPDF::ObjectPtr LemonPDF::Dictionary::copy() const
{
	DictionaryPtr result = std::make_shared<Dictionary>();
	enumerate([&result](const std::string &key, ObjectPtr value){
		result->add(key, value->copy());
	});
	return result;
}

void LemonPDF::Dictionary::enumerate(std::function<void(const std::string &, ObjectPtr)> func) const
{
	for (auto kv : m_objects)
		func(kv.first, kv.second);
}

LemonPDF::ObjectPtr LemonPDF::Dictionary::get(const std::string &name) const
{
	auto i = m_objects.find(name);
	if (i == m_objects.end())
		return nullptr;
	else
		return i->second;
}

LemonPDF::ObjectPtr LemonPDF::Dictionary::get(LiteralPtr name) const
{
	return get(name->getNameValue());
}

size_t LemonPDF::Dictionary::getSize() const
{
	return m_objects.size();
}

const std::string &LemonPDF::Dictionary::getType() const
{
	auto i = m_objects.find("Type");
	if (i == m_objects.end())
		throw PDFInvalidOperationException("Dictionary does not have a type");
	
	LiteralPtr l = std::dynamic_pointer_cast<Literal>(i->second);
	if (!l)
		throw PDFInvalidOperationException("Dictionary type is invalid");
	
	return l->getNameValue();
}

void LemonPDF::Dictionary::remove(const std::string &name)
{
	auto i = m_objects.find(name);
	if (i != m_objects.end())
		m_objects.erase(i);
}

void LemonPDF::Dictionary::remove(const LiteralPtr name)
{
	remove(name->getNameValue());
}

LemonPDF::DictionaryPtr LemonPDF::Dictionary::shallowCopy() const
{
	DictionaryPtr result = std::make_shared<Dictionary>();
	result->m_objects = m_objects;
	return result;
}

void LemonPDF::Dictionary::write(std::ostream &out, size_t depth) const
{
	std::set<std::string> sortedKeys;
	for (auto kvp : m_objects)
		sortedKeys.insert(kvp.first);
		
	if (depth == 0)
	{
		out << "<<";
		for (auto key : sortedKeys)
		{
			const auto kvp = m_objects.find(key);
			out << "/" << kvp->first;
			LiteralPtr literal = std::dynamic_pointer_cast<Literal>(kvp->second);

			if (literal)
			{
				Literal::Type t = literal->getType();
				if (t != Literal::Type::kName && t != Literal::Type::kString)
					out << " ";
			}

			kvp->second->write(out, 0);
		}
		out << ">>";
	}
	else
	{
		std::string prefix((depth - 1) * 2, ' ');
		std::string indent(depth * 2, ' ');
		out << "<<\n";

		for (auto key : sortedKeys)
		{
			const auto kvp = m_objects.find(key);
			out << indent << "/" << kvp->first << " ";
			kvp->second->write(out, depth + 1);
			out << "\n";
		}

		out << prefix << ">>";
	}
}
