#include <algorithm>
#include <set>
#include <sstream>
#include <tuple>
#include <libcore/logging.h>
#include "libtokenisation.h"
#include "pdfexception.h"
#include "pdffile.h"
#include "pdfpage.h"
#include "pdftokeniser.h"
#include "pjlreader.h"

namespace LemonPDF
{
	class PDFCircularReferenceException : public PDFException
	{
	public:
		PDFCircularReferenceException(const std::string &msg) : PDFException(msg) {}
	};
}

namespace
{
	class CircularReferenceChecker
	{
	public:
		CircularReferenceChecker(LemonPDF::ObjectReferenceSet &objlist, const LemonPDF::ObjectReference &obj)
		  : m_list(objlist), m_obj(obj)
		{
			if (!m_list.insert(obj).second)
				throw LemonPDF::PDFCircularReferenceException("Circular reference detected");
		}
		
		CircularReferenceChecker(const CircularReferenceChecker &) = delete;
		CircularReferenceChecker &operator=(const CircularReferenceChecker &) = delete;
		
		~CircularReferenceChecker()
		{
			m_list.erase(m_obj);
		}
		
	private:
		LemonPDF::ObjectReferenceSet &m_list;
		const LemonPDF::ObjectReference &m_obj;
	};
	
	class AutoStreamRestorer
	{
	public:
		explicit AutoStreamRestorer(std::istream &stream)
		  : m_stream(stream), m_pos(m_stream.tellg())
		{
		}
		
		AutoStreamRestorer(const AutoStreamRestorer &) = delete;
		AutoStreamRestorer &operator=(const AutoStreamRestorer &) = delete;
		
		~AutoStreamRestorer()
		{
			m_stream.clear();
			m_stream.seekg(m_pos);
		}
		
	private:
		std::istream &m_stream;
		std::istream::pos_type m_pos;
	};
	
	class FileLevelObject
	{
	public:
		FileLevelObject(LemonPDF::Tokeniser &tokeniser, LemonPDF::ResolverFunction resolver, const LemonPDF::ObjectReference &expected)
		{
			AutoStreamRestorer restorer(tokeniser.getInputStream());
			LemonPDF::Token t = tokeniser.peekNextToken();
			if (t.getType() != LemonPDF::Token::kObjectStart)
			{
				std::ostringstream msg;
				msg << "Expected \"" << expected.getObjectID() << " " << expected.getGeneration() << " obj\", got \"" << t.getText() << "\"";
				throw LemonPDF::PDFParseException(msg.str(), tokeniser.getInputStream());
			}
			
			tokeniser.consume();
			
			std::pair<size_t, size_t> details = t.getObjectIDAndGeneration();
			m_id = details.first;
			m_generation = details.second;
			m_object = LemonPDF::Object::Parse(tokeniser, resolver);
			
			// if ObjectID is 0, it's because we're constructing the xref table from this file
			// level object, which means we don't know what the expected object ID is.
			if (expected.getObjectID() != 0)
			{
				if (m_id != expected.getObjectID() || m_generation != expected.getGeneration())
				{
					std::ostringstream msg;
					msg << "Expected \"" << expected.getObjectID() << " " << expected.getGeneration() << " obj\", ";
					msg << "got \"" << m_id << " " << m_generation << " obj\"";
					throw LemonPDF::PDFParseException(msg.str(), tokeniser.getInputStream());
				}
			}

			t = tokeniser.getNextToken();
			if (t.getType() != LemonPDF::Token::kObjectEnd)
				throw LemonPDF::PDFSyntaxException("Expected \"endobj\"", tokeniser.getInputStream());
		}
		
		LemonPDF::ObjectPtr getObject() const
		{
			return m_object;
		}
		
		size_t getGeneration() const
		{
			return m_generation;
		}
		
		size_t getID() const
		{
			return m_id;
		}
		
	private:
		LemonPDF::ObjectPtr m_object;
		size_t m_id;
		size_t m_generation;
	};
	
	/// Essentially calculate floor(log(val)/log(256)) + 1
	size_t numberOfBytesNeededForValue(size_t value)
	{
		size_t bytesNeeded = 1;
		while (value >>= 8)
			bytesNeeded++;
		return bytesNeeded;
	}

	void writeFixedWidthFieldValue(std::ostream &out, size_t fieldSize, size_t value)
	{
		while (fieldSize > 0)
			out.put(static_cast<char>((value >> (--fieldSize * 8)) & 0xFF));
	}

	size_t parseAsUnsignedInteger(const unsigned char *buf, size_t length)
	{
		size_t result = 0;
		for (size_t i = 0; i < length; i++)
			result |= buf[i] << (8 * (length - i - 1));
		return result;
	}
	
	size_t parseAsUnsignedInteger(const std::string &str)
	{
		size_t pos = 0;
		long long value = stoll(str, &pos, 10);
		if (value < 0 || pos != str.size())
			throw LemonPDF::PDFParseException("Expected non-negative integer");
		return value;
	}

	bool decompressStreamInPlace(LemonPDF::ObjectPtr o)
	{
		LemonPDF::StreamPtr stream = std::dynamic_pointer_cast<LemonPDF::Stream>(o);
		if (!stream)
			return true;

		if (!stream->filteringSupported())
			return false;
		
		LemonPDF::DictionaryPtr streamDict = stream->getDictionary();
		
		std::vector<std::string> filterNames = stream->getFilterNames();
		if (filterNames.size() == 1)
		{
			streamDict->add("LemonPDF:OriginalFilter", LemonPDF::Literal::FromName(filterNames[0]));
		}
		else if (filterNames.size() > 1)
		{
			LemonPDF::ArrayPtr filterArray = std::make_shared<LemonPDF::Array>();
			for (auto filterName : filterNames)
				filterArray->add(LemonPDF::Literal::FromName(filterName));
			streamDict->add("LemonPDF:OriginalFilter", filterArray);
		}

		std::stringstream ss;
		stream->filter(ss);
		std::vector<unsigned char> buffer(ss.tellp());
		ss.read(reinterpret_cast<char *>(&buffer[0]), buffer.size());
		stream->setData(buffer);
		return true;
	}

	template<size_t N>
	void convertPDFArrayToDoubleArray(LemonPDF::PDFFile &pdf, LemonPDF::ArrayPtr array, double (&vals)[N])
	{
		if (array == nullptr)
			throw LemonPDF::PDFInvalidOperationException("Array is null");

		if (array->getSize() != N)
			throw LemonPDF::PDFInvalidOperationException("Cannot convert array, wrong number of objects");
		
		array->enumerate([&vals, &pdf](size_t i, LemonPDF::ObjectPtr o)->bool{
			LemonPDF::LiteralPtr literal = pdf.resolve<LemonPDF::Literal>(o);
			if (!literal)
				throw LemonPDF::PDFInvalidOperationException("Cannot convert PDF array to double array, object is not a numeric value.");
			vals[i] = literal->getRealValue();
			return true;
		});
	}

	/**
	 * Creates an object stream (a PDF stream object that contains other PDF objects)
	 * @param objects a mapping that maps object references (i.e. IDs) to actual objects.
	 * @param writeStyle determines whether to format the contents in a more human readable way.
	 * @return a tuple, the first element is the stream object containing the sub-objects,
	 *   and the second element is a map that maps each object ID to an index into this
	 *   stream.
	 */
	std::tuple<LemonPDF::StreamPtr, LemonPDF::ObjectReferenceMap<size_t>> createObjectStream(const LemonPDF::ObjectReferenceMap<LemonPDF::ObjectPtr> &objects, LemonPDF::PDFFile::WriteStyle writeStyle)
	{
		LemonPDF::ObjectReferenceMap<size_t> offsets;
		LemonPDF::ObjectReferenceMap<size_t> indices;
		std::set<LemonPDF::ObjectReference> sortedReferences;

		std::stringstream header;
		std::stringstream objectData;

		size_t index = 0;
		for (const auto iter : objects)
			sortedReferences.insert(iter.first);

		for (const auto iter : sortedReferences)
		{
			offsets.emplace(iter, objectData.tellp());
			indices.emplace(iter, index++);
			if (writeStyle == LemonPDF::PDFFile::WriteStyle::kPadded)
			{
				objectData << "% " << iter.getObjectID() << " " << iter.getGeneration() << " obj\n";
				objects.find(iter)->second->write(objectData, 1);
				objectData << "\n";
			}
			else
			{
				objects.find(iter)->second->write(objectData, 0);
			}
		}

		for (const auto iter : sortedReferences)
		{
			size_t offset = offsets.find(iter)->second;
			header << iter.getObjectID() << " " << offset << " ";
			if (writeStyle == LemonPDF::PDFFile::WriteStyle::kPadded)
				header << "\n";
		}

		size_t headerSize = header.tellp();

		header << objectData.rdbuf();

		LemonPDF::StreamPtr result = std::make_shared<LemonPDF::Stream>();

		if (writeStyle == LemonPDF::PDFFile::WriteStyle::kCompact)
			result->setFilterNames({"FlateDecode"});

		LemonPDF::DictionaryPtr streamDict = result->getDictionary();
		streamDict->add("Type", LemonPDF::Literal::FromName("ObjStm"));
		streamDict->add("First", LemonPDF::Literal::FromInteger(headerSize));
		streamDict->add("N", LemonPDF::Literal::FromInteger(objects.size()));

		header.seekg(0);
		header.clear();
		result->filterIn(header);

		return std::make_tuple(result, indices);
	}

	/**
	 * Create a cross-reference stream
	 * @param largestOffset the largest object offset within the file, used to
	 *   calculate field widths in the cross-reference stream
	 * @param largestObjectID the largest object ID within the file, used to
	 *   calculate field widths in the cross-reference stream
	 * @param objectStreamID the ID of the stream containing compressed objects
	 * @param offsets a mapping between object references and file offsets, used
	 *   to construct the actual cross-reference table
	 * @param compressedObjectIndices a mapping between object references and the
	 *   cardinal position within the object stream
	 * @return a Stream object that can be appended to the file
	 */
	LemonPDF::StreamPtr createXrefStream(size_t largestOffset, size_t largestObjectID, size_t objectStreamID, const LemonPDF::ObjectReferenceMap<size_t> &offsets, const LemonPDF::ObjectReferenceMap<size_t> &compressedObjectIndices)
	{
		size_t fieldSizes[3] = { 1, 0, 0 };
		size_t maxBytesForOffsets = numberOfBytesNeededForValue(largestOffset);
		size_t maxBytesForObjectIDs = numberOfBytesNeededForValue(largestObjectID);

		fieldSizes[1] = std::max(maxBytesForOffsets, maxBytesForObjectIDs);
		fieldSizes[2] = numberOfBytesNeededForValue(compressedObjectIndices.size());

		std::stringstream xrefStreamData;
		for (size_t i = 0; i <= largestObjectID; i++)
		{
			LemonPDF::ObjectReference thisObject = LemonPDF::ObjectReference(i);
			auto offsetIter = offsets.find(thisObject);
			auto compressedObjectIndexIter = compressedObjectIndices.find(thisObject);

			if (offsetIter != offsets.end())
			{
				// this is a file level object
				writeFixedWidthFieldValue(xrefStreamData, fieldSizes[0], 1);
				writeFixedWidthFieldValue(xrefStreamData, fieldSizes[1], offsetIter->second);
				writeFixedWidthFieldValue(xrefStreamData, fieldSizes[2], 0);
			}
			else if (compressedObjectIndexIter != compressedObjectIndices.end())
			{
				// this object is embedded inside another object
				writeFixedWidthFieldValue(xrefStreamData, fieldSizes[0], 2);
				writeFixedWidthFieldValue(xrefStreamData, fieldSizes[1], objectStreamID);
				writeFixedWidthFieldValue(xrefStreamData, fieldSizes[2], compressedObjectIndexIter->second);
			}
			else
			{
				// this object ID does not exist
				writeFixedWidthFieldValue(xrefStreamData, fieldSizes[0], 0);
				writeFixedWidthFieldValue(xrefStreamData, fieldSizes[1], i);
				writeFixedWidthFieldValue(xrefStreamData, fieldSizes[2], 0);
			}
		}

		LemonPDF::StreamPtr xrefStream = LemonPDF::Stream::FromIOStream(std::move(xrefStreamData));
		LemonPDF::DictionaryPtr streamDict = xrefStream->getDictionary();

		LemonPDF::ArrayPtr fieldWidths = std::make_shared<LemonPDF::Array>();
		fieldWidths->add(LemonPDF::Literal::FromInteger(fieldSizes[0]));
		fieldWidths->add(LemonPDF::Literal::FromInteger(fieldSizes[1]));
		fieldWidths->add(LemonPDF::Literal::FromInteger(fieldSizes[2]));
		streamDict->add("W", fieldWidths);

		return xrefStream;
	}
}

LemonPDF::PDFFilePtr LemonPDF::PDFFile::EmptyFile(size_t major, size_t minor)
{
	return std::make_shared<PDFFile>(ConstructionToken(), major, minor);
}

LemonPDF::PDFFilePtr LemonPDF::PDFFile::FromFile(const std::string &filename)
{
	return std::make_shared<PDFFile>(ConstructionToken(), filename);
}

LemonPDF::PDFFilePtr LemonPDF::PDFFile::FromStream(std::ifstream &&stream)
{
	return std::make_shared<PDFFile>(ConstructionToken(), std::move(stream));
}

LemonPDF::PDFFile::PDFFile(const ConstructionToken &token, size_t major, size_t minor)
  : m_majorVersion(major), m_minorVersion(minor), m_dataOffset(0), m_largestObjectID(0)
{
	(void) token;

	m_root = std::make_shared<Dictionary>();
	m_root->add("Type", Literal::FromName("Catalog"));

	m_trailer = std::make_shared<Dictionary>();

	addObject(m_root);
}

LemonPDF::PDFFile::PDFFile(const ConstructionToken &token, const std::string &filename)
  : m_majorVersion(0), m_minorVersion(0), m_input(filename, std::ios_base::binary), m_dataOffset(0), m_largestObjectID(0)
{
	(void) token;

	m_input.rdbuf()->pubsetbuf(m_fileBuffer.data(), m_fileBuffer.size());
	m_resolver = [&](ObjectPtr object){
		return resolveObject(object, true);
	};

	parse();
}

LemonPDF::PDFFile::PDFFile(const ConstructionToken &token, std::ifstream &&input)
  : m_majorVersion(0), m_minorVersion(0), m_input(std::move(input)), m_dataOffset(0), m_largestObjectID(0)
{
	(void) token;

	m_resolver = [&](ObjectPtr object){
		return resolveObject(object, true);
	};

	parse();
}

LemonPDF::ObjectReference LemonPDF::PDFFile::addObject(ObjectPtr object)
{
	// although we should pick the first free object ID, that is O(n), whereas
	// choosing the highest and adding one is amortized constant.
	ObjectReference addedObjectID(m_largestObjectID + 1, 0);

	if (!m_objects.emplace(addedObjectID, object).second)
	{
		// object was not added, probably the largestObjectID was already used,
		// in which case we are out of sync, not a lot we can do without making
		// the situation worse (potentially).
		throw PDFInvalidOperationException("Unable to add the object, because the added object ID is already in used");
	}

	// don't increment this unless it was successfully added
	m_largestObjectID++;

	m_xref[addedObjectID] = 0;
	return addedObjectID;
}

void LemonPDF::PDFFile::addCompressedXref(const ObjectReference &ref, const CompressedObjectReference &cref)
{
	m_compressedXref.emplace(ref, cref);
	m_largestObjectID = std::max(m_largestObjectID, ref.getObjectID());
}

void LemonPDF::PDFFile::addXref(const ObjectReference &ref, size_t offset)
{
	m_xref.emplace(ref, offset);
	m_largestObjectID = std::max(m_largestObjectID, ref.getObjectID());
}

LemonPDF::ObjectReference LemonPDF::PDFFile::appendPage(PagePtr page)
{
	ObjectPtr clonedPageDictionary = page->getDictionary()->copy();
	ObjectReference newPageRef = addObject(clonedPageDictionary);

	DictionaryPtr pages = resolve<Dictionary>(m_root->get("Pages"));
	if (!pages)
	{
		pages = std::make_shared<Dictionary>();
		pages->add("Type", Literal::FromName("Pages"));
		pages->add("Kids", std::make_shared<Array>());
		pages->add("Count", Literal::FromInteger(0));
		m_root->add("Pages", pages);
	}
	
	ArrayPtr kidsArray = resolve<Array>(pages->get("Kids"));
	kidsArray->add(Literal::FromObjectReference(newPageRef));

	long long pageCount = pages->get<Literal>("Count")->getIntegerValue();
	pageCount++;
	pages->add("Count", Literal::FromInteger(pageCount));

	return newPageRef;
}

LemonPDF::Rectangle LemonPDF::PDFFile::convertArrayToRectangle(ArrayPtr array)
{
	double vals[4] = {};
	convertPDFArrayToDoubleArray(*this, array, vals);
	return LemonPDF::Rectangle { vals[0], vals[1], vals[2], vals[3] };
}

LemonPDF::TransformMatrix LemonPDF::PDFFile::convertArrayToTransformMatrix(ArrayPtr array)
{
	double vals[6] = {};
	convertPDFArrayToDoubleArray(*this, array, vals);
	return LemonPDF::TransformMatrix { vals[0], vals[1], vals[2], vals[3], vals[4], vals[5] };
}

void LemonPDF::PDFFile::enumeratePages(PageEnumerationFunction function)
{
	std::vector<PagePtr> pages = getPages();
	for (PagePtr page : pages)
		function(page);
}

void LemonPDF::PDFFile::enumerateStreamsOfType(const std::string &type, StreamEnumerationFunction function)
{
	enumerateStreamsOfType(type, "", function);
}

void LemonPDF::PDFFile::enumerateStreamsOfType(const std::string &type, const std::string &subtype, StreamEnumerationFunction function)
{
	// streams cannot be embedded in compressed object streams
	for (auto xref : m_xref)
	{
		LemonPDF::StreamPtr stream = resolve<Stream>(xref.first, true);
		if (stream)
		{
			const DictionaryPtr streamDict = stream->getDictionary();
			LiteralPtr typeLiteral = resolve<Literal>(streamDict->get("Type"));
			LiteralPtr subtypeLiteral = resolve<Literal>(streamDict->get("Subtype"));

			if ((type.size() || (typeLiteral && typeLiteral->getNameValue() == type)) && (subtype.size() == 0 || (subtypeLiteral && subtypeLiteral->getNameValue() == subtype)))
				function(xref.first, stream);
		}
	}
}

LemonPDF::DictionaryPtr LemonPDF::PDFFile::getDocumentCatalog() const
{
	return m_root;
}

LemonPDF::DictionaryPtr LemonPDF::PDFFile::getInfoDictionary() const
{
	return m_info;
}

LemonPDF::DictionaryPtr LemonPDF::PDFFile::getTrailer() const
{
	return m_trailer;
}

size_t LemonPDF::PDFFile::getMajorVersion() const
{
	return m_majorVersion;
}

size_t LemonPDF::PDFFile::getMinorVersion() const
{
	return m_minorVersion;
}

std::vector<LemonPDF::PagePtr> LemonPDF::PDFFile::getPages()
{
	DictionaryPtr pages = resolve<Dictionary>(m_root->get("Pages"));
	if (!pages)
		throw PDFInvalidOperationException("The document does not contain any pages.");
	
	PageTreeNodePtr pageRoot = PageTreeNode::ConstructTree(shared_from_this(), pages, nullptr);
	return PageTreeNode::GetPageObjects(pageRoot);
}

const LemonPDF::ObjectReferenceMap<size_t> &LemonPDF::PDFFile::getXrefTable() const
{
	return m_xref;
}

const LemonPDF::ObjectReferenceMap<LemonPDF::CompressedObjectReference> &LemonPDF::PDFFile::getCompressedXrefTable() const
{
	return m_compressedXref;
}

void LemonPDF::PDFFile::parse()
{
	PJLReader pjlReader(m_input);
	m_dataOffset = pjlReader.getPrintDataOffset();

	std::string header;
	
	if (!Tokenisation::ReadLine(m_input, header))
		throw PDFIOException("Unable to read from file");
	
	if (header.substr(0, 4) != "%PDF")
		throw PDFParseException("Invalid header for PDF file");
	
	if (header.size() >= 8)
	{
		if (header[4] != '-')
			throw PDFParseException("Invalid header for PDF file");
		
		std::istringstream ss(header.substr(5));
		char period;
		if (!(ss >> m_majorVersion >> period >> m_minorVersion) || period != '.')
			throw PDFParseException("Invalid header for PDF file");
	}
	
	std::string line;
	bool parsed = false;
	while (Tokenisation::ReadLine(m_input, line))
	{
		if (line == "startxref")
		{
			if (!Tokenisation::ReadLine(m_input, line))
				throw PDFIOException("Unexpected end of file");
			
			AutoStreamRestorer restorer(m_input);
			size_t offset = parseAsUnsignedInteger(line);
			if (!m_input.seekg(offset + m_dataOffset))
				throw PDFIOException("Unable to seek to cross-reference table");
			
			parseXrefTables();
			parsed = true;
		}
	}
	
	if (!parsed)
		throw PDFParseException("Unexpected end of file while looking for Xref table");
	
	if (!m_root)
		throw PDFInvalidOperationException("There is no document catalog");
}

void LemonPDF::PDFFile::parseXrefTables()
{
	using namespace LemonPDF;
	
	Tokeniser tokeniser(m_input);
	Token t = tokeniser.peekNextToken();
	
	if (t.getType() == Token::kObjectStart)
	{
		// There's no xref table, so the resolver function is a no-op
		FileLevelObject flo(tokeniser, [](ObjectPtr o){ return o; }, ObjectReference(0, 0));
		
		StreamPtr stream = std::dynamic_pointer_cast<Stream>(flo.getObject());
		if (!stream)
			throw PDFInvalidOperationException("Expected stream");
		
		parseXrefObjectStream(stream);
	}
	else if (t.getType() == Token::kXref)
	{
		tokeniser.consume();
		parseXrefTableOldStyle(tokeniser.getInputStream());
	}
}

void LemonPDF::PDFFile::parseXrefObjectStream(StreamPtr stream)
{
	size_t fieldWidths[3] = {};
	size_t id = 0;
	size_t previousXrefTableOffset = 0;

	LiteralPtr previousXrefTableOffsetLiteral = stream->getDictionary()->get<Literal>("Prev");
	if (previousXrefTableOffsetLiteral)
		previousXrefTableOffset = previousXrefTableOffsetLiteral->getIntegerValue();

	if (previousXrefTableOffset > 0)
	{
		AutoStreamRestorer restorer(m_input);
		m_input.seekg(previousXrefTableOffset);
		parseXrefTables();
	}

	ArrayPtr widthArray = stream->getDictionary()->get<Array>("W");
	widthArray->enumerate([&](size_t i, ObjectPtr o){
		if (i > 3)
			return;
		
		LiteralPtr literal = std::dynamic_pointer_cast<Literal>(o);
		fieldWidths[i] = literal->getIntegerValue();
	});
	
	ArrayPtr indexArray = stream->getDictionary()->get<Array>("Index");
	if (indexArray)
	{
		LiteralPtr firstIndexLiteral = indexArray->get<Literal>(0);
		if (!firstIndexLiteral)
			throw PDFInvalidOperationException("First index not found in index array of Xref table");
		
		id = (firstIndexLiteral->getIntegerValue() - 1);
	}

	const size_t totalFieldWidth = fieldWidths[0] + fieldWidths[1] + fieldWidths[2];
	
	const std::vector<unsigned char> &buffer = stream->getFilteredData();
	if ((buffer.size() % totalFieldWidth) != 0)
		throw PDFInvalidOperationException("Xref table data does not match field lengths");
	
	for (size_t i = 0; i < buffer.size(); i += totalFieldWidth)
	{
		size_t fields[3] = { 
			parseAsUnsignedInteger(&buffer[i], fieldWidths[0]),
			parseAsUnsignedInteger(&buffer[i + fieldWidths[0]], fieldWidths[1]),
			parseAsUnsignedInteger(&buffer[i + fieldWidths[0] + fieldWidths[1]], fieldWidths[2])
		};
		
		id++;
		
		switch (fields[0])
		{
			case 0:
				id = fields[1];
				break;
			
			case 1:
				addXref(ObjectReference(id, fields[2]), fields[1]);
				break;
				
			case 2:
				addCompressedXref(ObjectReference(id, 0), CompressedObjectReference(fields[1], fields[2]));
				break;
				
			default:
				break;
		}
	}
	
	DictionaryPtr info = resolve<Dictionary>(stream->getDictionary()->get("Info"), false);
	if (info)
		m_info = info;
	
	DictionaryPtr root = resolve<Dictionary>(stream->getDictionary()->get("Root"), false);
	if (root)
		m_root = root;
}

void LemonPDF::PDFFile::parseXrefTableOldStyle(std::istream &in)
{
	size_t objectID = 0;
	size_t count = 0;
	std::string line;
	
	while (true)
	{
		std::streampos p = in.tellg();
		
		if (!Tokenisation::ReadLine(in, line))
			break;
		
		// discard blank line (may be leftover from tokenisation)
		if (line.size() == 0)
			continue;
		
		std::istringstream ss(line);
		if (!(ss >> objectID >> count))
		{
			in.seekg(p);
			break;
		}
		
		for (size_t i = 0; i < count; i++)
		{
			if (!Tokenisation::ReadLine(in, line))
				throw PDFIOException("Unable to read from input");
			
			if (line.length() != 18 && (line.length() != 19 && line[18] == 0x20))
			{
				std::ostringstream msg;
				msg << "Line invalid \"" << line << "\" (" << line.length() << ")";
				throw LemonPDF::PDFSyntaxException(msg.str(), in);
			}
			
			if (line[17] == 'f')
				continue;
			
			size_t offset = parseAsUnsignedInteger(line.substr(0, 10));
			size_t generation = parseAsUnsignedInteger(line.substr(11, 5));
			
			if (offset > 0)
				addXref(ObjectReference(objectID + i, generation), offset);
		}
	}
	
	Tokeniser tokeniser(in);
	if (tokeniser.peekNextToken().getType() == Token::kTrailer)
	{
		tokeniser.consume();
		DictionaryPtr trailer = Dictionary::Parse(tokeniser);

		DictionaryPtr info = resolve<Dictionary>(trailer->get("Info"), false);
		if (info)
			m_info = info;
		
		DictionaryPtr root = resolve<Dictionary>(trailer->get("Root"), false);
		if (root)
		{
			m_root = root;
			m_trailer = trailer;
			m_encrypt = resolve<Dictionary>(trailer->get("Encrypt"));

			// TODO: add encrypted PDF support
			if (m_encrypt)
				throw PDFUnimplementedException("Encrypted PDFs not supported");
		}

		LiteralPtr prev = resolve<Literal>(trailer->get("Prev"), false);
		if (prev)
		{
			AutoStreamRestorer restorer(m_input);
			m_input.seekg(prev->getIntegerValue());
			parseXrefTables();
		}
	}
}

LemonPDF::ObjectPtr LemonPDF::PDFFile::removeObject(const ObjectReference &ref)
{
	LemonPDF::ObjectPtr result = resolve(ref);
	m_xref.erase(ref);
	m_compressedXref.erase(ref);
	return result;
}

void LemonPDF::PDFFile::replaceObject(const ObjectReference &ref, ObjectPtr object)
{
	// set a dummy value in the xref table (and make sure it doesn't
	// also appear in the compressed xref table).
	m_xref[ref] = 0;
	m_compressedXref.erase(ref);

	// replace whatever is currently there with the new object
	m_objects[ref] = object;
}

LemonPDF::ObjectPtr LemonPDF::PDFFile::resolve(const ObjectReference &ref, const CompressedObjectReference &cref, bool throwOnMissing)
{
	auto cached = m_objects.find(ref);
	if (cached != m_objects.end())
		return cached->second;
	
	CircularReferenceChecker checker(m_resolverSet, ref);
	AutoStreamRestorer restorer(m_input);
	
	ObjectPtr hostObject = resolve(ObjectReference(cref.getObjectID(), 0), throwOnMissing);
	if (!hostObject)
		return nullptr;
	
	StreamPtr stream = std::dynamic_pointer_cast<Stream>(hostObject);
	if (!stream)
		throw PDFInvalidOperationException("Expected Stream");
	
	size_t numberOfObjects;
	size_t firstByte;
	
	LiteralPtr numberLiteral = stream->getDictionary()->get<Literal>("N");
	if (!numberLiteral || numberLiteral->getIntegerValue() <= 0)
		throw PDFInvalidOperationException("The object stream does not have an 'N' key or the 'N' key is invalid.");
	numberOfObjects = numberLiteral->getIntegerValue();
	
	LiteralPtr firstByteLiteral = stream->getDictionary()->get<Literal>("First");
	if (!firstByteLiteral || firstByteLiteral->getIntegerValue() <= 0)
		throw PDFInvalidOperationException("The object stream does not have a 'First' key or the 'First' key is invalid.");
	firstByte = firstByteLiteral->getIntegerValue();
	
	std::stringstream subinput;
	stream->filter(subinput);
	std::map<size_t, size_t> offsets;
	
	for (size_t i = 0; i < numberOfObjects; i++)
	{
		size_t objectID;
		size_t offset;
		
		if (!(subinput >> objectID >> offset))
			throw PDFParseException("Unable to read object ID and offsets");
		
		offsets.emplace(objectID, offset);
	}
	
	for (auto objectOffset : offsets)
	{
		subinput.clear();
		if (!subinput.seekg(firstByte + objectOffset.second))
			throw PDFIOException("Unable to seek within PDF");
		Tokeniser tokeniser(subinput);
		m_objects.emplace(ObjectReference(objectOffset.first, 0), Object::Parse(tokeniser, m_resolver));
	}
	
	auto result = m_objects.find(ref);
	if (result == m_objects.end())
		throw PDFInvalidOperationException("PDF compressed cross-reference table invalid");
	else
		return result->second;
}

LemonPDF::ObjectPtr LemonPDF::PDFFile::resolve(const ObjectReference &objref, bool throwOnMissing)
{
	auto cached = m_objects.find(objref);
	if (cached != m_objects.end())
		return cached->second;
	
	CircularReferenceChecker checker(m_resolverSet, objref);
	AutoStreamRestorer restorer(m_input);
	
	auto offset = m_xref.find(objref);
	if (offset == m_xref.end())
	{
		if (!throwOnMissing)
			return nullptr;
		
		std::ostringstream msg;
		msg << "Cannot find " << objref.getObjectID() << " " << objref.getGeneration() << " obj";
		throw PDFInvalidOperationException(msg.str());
	}
	
	m_input.clear();
	if (!m_input.seekg(offset->second + m_dataOffset, std::ios_base::beg))
		throw PDFIOException("Unable to seek to object offset");
	Tokeniser tokeniser(m_input);
	FileLevelObject flo(tokeniser, m_resolver, objref);
	return m_objects.emplace(objref, flo.getObject()).first->second;
}

void LemonPDF::PDFFile::resolveEntireFile(bool filterObjects)
{
	for (const auto it : m_xref)
	{
		ObjectPtr o = resolve(it.first);
		if (filterObjects)
			if (!decompressStreamInPlace(o))
				StripedHorse::Logging::Warning() << "Unable to decompress " << it.first.getObjectID() << " " << it.first.getGeneration() << " obj" << std::endl;
	}
	
	for (const auto it : m_compressedXref)
	{
		ObjectPtr o = resolve(it.first, it.second);
		if (filterObjects)
			if (!decompressStreamInPlace(o))
				StripedHorse::Logging::Warning() << "Unable to decompress object " << it.first.getObjectID() << " (in " << it.second.getObjectID() << " " << it.second.getGeneration() << " obj)" << std::endl;
	}
}

LemonPDF::ObjectPtr LemonPDF::PDFFile::resolveObject(ObjectPtr object, bool throwOnMissing)
{
	// If this object isn't a literal object, return immediately
	LiteralPtr literal = std::dynamic_pointer_cast<Literal>(object);
	if (!literal)
		return object;
	
	// If this object isn't an indirect object reference, return immediately
	if (literal->getType() != Literal::kReference)
		return object;
	
	// Check to see whether the object ID + generation has already been loaded
	const ObjectReference &ref = literal->getReference();
	auto id = m_objects.find(ref);
	if (id != m_objects.end())
		return id->second;
	
	// Check to see whether it is directly loadable from main Xref table
	ObjectPtr result = resolve(ref, false);
	if (result)
		return result;
	
	// Check to see whether the object is compressed in an object stream
	auto fromCXref = m_compressedXref.find(ref);
	if (fromCXref != m_compressedXref.end())
		return resolve(ref, fromCXref->second, throwOnMissing);
	
	// Throw or return null, depending on situation
	if (!throwOnMissing)
		return nullptr;
	
	std::ostringstream msg;
	msg << "Invalid object reference: " << object;
	throw PDFInvalidOperationException(msg.str());
}

void LemonPDF::PDFFile::setDocumentCatalog(DictionaryPtr catalog)
{
	LiteralPtr type = resolve<Literal>(catalog->get("Type"));
	if (type == nullptr || type->getNameValue() != "Catalog")
		throw PDFInvalidArgumentException("Document catalog has missing or invalid type");
	
	m_root = catalog;
}

void LemonPDF::PDFFile::setInfoDictionary(DictionaryPtr info)
{
	// if no existing info dictionary exists in the file, add
	// the object as a file-level object. We can discard the
	// reference because it gets determined in the write() function
	// anyway
	if (m_info == nullptr)
		addObject(info);
	m_info = info;
}

void LemonPDF::PDFFile::setTrailer(DictionaryPtr trailer)
{
	m_trailer = trailer;
}

void LemonPDF::PDFFile::setVersion(size_t major, size_t minor)
{
	m_majorVersion = major;
	m_minorVersion = minor;
}

void LemonPDF::PDFFile::write(const std::string &filename, WriteStyle writeStyle) const
{
	std::ofstream output(filename, std::ios_base::out | std::ios_base::binary);
	write(output, writeStyle);
}

void LemonPDF::PDFFile::write(std::ofstream &output, WriteStyle writeStyle) const
{
	if (!output.is_open())
		return;
	
	if (m_objects.size() != m_xref.size() + m_compressedXref.size())
		throw PDFUnimplementedException("Cannot write a PDF that has not had all its objects resolved");

	// PDF header
	output << "%PDF-" << m_majorVersion << "." << m_minorVersion << "\n";
	output.write("%\xC3\xA4\xC3\xBC\xC3\xB6\xC3\x9F\n", 10);

	if (m_majorVersion > 1 || m_minorVersion >= 5)
	{
		// PDF versions >= 1.5 can use ObjStm and Xref objects
		writeCompressedXrefPDF(output, writeStyle);
	}
	else
	{
		// PDF versions < 1.5 had to have these oldschool xref tables
		writeOldschoolXrefPDF(output, writeStyle);
	}

	output << "%%EOF\n";
}

void LemonPDF::PDFFile::writeCompressedXrefPDF(std::ofstream &output, WriteStyle writeStyle) const
{
	ObjectReference rootObject;
	ObjectReference infoObject;

	std::set<ObjectReference> discardedObjects;
	ObjectReferenceMap<size_t> offsets;
	ObjectReferenceMap<ObjectPtr> fileLevelObjects;
	ObjectReferenceMap<ObjectPtr> compressableObjects;

	for (const auto it : m_objects)
	{
		StreamPtr stream = std::dynamic_pointer_cast<Stream>(it.second);
		if (stream)
		{
			LiteralPtr type = std::dynamic_pointer_cast<Literal>(stream->getDictionary()->get("Type"));
			if (type && (type->getNameValue() == "XRef" || type->getNameValue() == "ObjStm"))
			{
				// we will replace these with our own objects
				discardedObjects.emplace(it.first);
				continue;
			}

			fileLevelObjects.emplace(it.first, it.second);
		}
		else
		{
			// anything that is not a Stream can be compressed
			compressableObjects.emplace(it.first, it.second);
		}

		if (it.second == m_root)
			rootObject = it.first;
		else if (it.second == m_info)
			infoObject = it.first;
	}

	// we may need to add objects, but this method is const-qualified
	// (because writing a PDF should not necessary modify its internal
	// structure).
	size_t outputLargestObjectID = m_largestObjectID;

	// serialise all compressable objects into a stream, and then either
	// reuse an existing object ID (one that we can safely discard) or
	// generate a new one for it.
	StreamPtr objectStream;
	ObjectReferenceMap<size_t> compressedObjectIndices;
	std::tie(objectStream, compressedObjectIndices) = createObjectStream(compressableObjects, writeStyle);

	size_t objectStreamID;
	if (discardedObjects.size())
	{
		// reuse some old ID that is not being used
		auto iter = discardedObjects.begin();
		objectStreamID = iter->getObjectID();
		fileLevelObjects[ObjectReference(objectStreamID)] = objectStream;
		discardedObjects.erase(iter);
	}
	else
	{
		// create a new ID for this object
		objectStreamID = ++outputLargestObjectID;
		fileLevelObjects[ObjectReference(outputLargestObjectID)] = objectStream;
	}

	std::set<ObjectReference> sortedObjectReferences;
	for (const auto &iter : fileLevelObjects)
		sortedObjectReferences.emplace(iter.first);

	for (const auto iter : sortedObjectReferences)
	{
		offsets.emplace(iter, output.tellp());
		output << iter.getObjectID() << " " << iter.getGeneration() << " obj\n";
		fileLevelObjects[iter]->write(output, writeStyle == WriteStyle::kCompact ? 0 : 1);
		output << "\nendobj\n";
	}

	size_t xrefStreamOffset = output.tellp();
	size_t xrefStreamID;

	if (discardedObjects.size())
	{
		auto iter = discardedObjects.begin();
		xrefStreamID = iter->getObjectID();
		offsets[ObjectReference(xrefStreamID)] = xrefStreamOffset;
		discardedObjects.erase(iter);
	}
	else
	{
		xrefStreamID = ++outputLargestObjectID;
		offsets[ObjectReference(outputLargestObjectID)] = xrefStreamOffset;
	}

	StreamPtr xrefStream = createXrefStream(xrefStreamOffset, outputLargestObjectID, objectStreamID, offsets, compressedObjectIndices);
	DictionaryPtr trailer = xrefStream->getDictionary();

	trailer->add("Type", Literal::FromName("XRef"));
	trailer->add("Size", Literal::FromInteger(outputLargestObjectID + 1));
	trailer->add("Root", Literal::FromObjectReference(rootObject));
	if (infoObject.getObjectID() > 0)
		trailer->add("Info", Literal::FromObjectReference(infoObject));

	output << xrefStreamID << " 0 obj\n";
	xrefStream->write(output, writeStyle == WriteStyle::kCompact ? 0 : 1);
	output << "\nendobj\n";
	output << "startxref\n";
	output << xrefStreamOffset << "\n";

}

void LemonPDF::PDFFile::writeOldschoolXrefPDF(std::ofstream &output, WriteStyle writeStyle) const
{
	std::map<size_t, std::streampos> offsets;
	ObjectReference rootObject;
	ObjectReference infoObject;

	std::map<ObjectReference, ObjectPtr> sortedObjects;
	for (const auto it : m_objects)
		sortedObjects.insert(std::make_pair(it.first, it.second));

	for (const auto it : sortedObjects)
	{
		offsets.emplace(it.first.getObjectID(), output.tellp());
		if (it.second == m_root)
			rootObject = it.first;
		else if (it.second == m_info)
			infoObject = it.first;
		output << it.first.getObjectID() << " " << it.first.getGeneration() << " obj\n";
		if (writeStyle == WriteStyle::kPadded)
			it.second->write(output, 1);
		else if (writeStyle == WriteStyle::kCompact)
			it.second->write(output, 0);
		output << "\nendobj\n";
	}

	DictionaryPtr trailer;
	if (m_trailer)
		trailer = m_trailer;
	else
		trailer = std::make_shared<Dictionary>();

	trailer->add("Size", Literal::FromInteger(m_largestObjectID + 1));
	trailer->add("Root", Literal::FromObjectReference(rootObject));
	if (infoObject.getObjectID() > 0)
		trailer->add("Info", Literal::FromObjectReference(infoObject));

	// remove any reference to an older xref table, since we'll be making our own
	trailer->remove("Prev");

	std::streampos xrefPosition = output.tellp();
	output << "xref\n";
	output << 0 << " " << (m_largestObjectID + 1) << "\n";
	output << "0000000000 65535 f \n";
	output.fill('0');
	for (size_t i = 1; i <= m_largestObjectID; i++)
	{
		auto offsetIterator = offsets.find(i);
		if (offsetIterator == offsets.end())
		{
			output.width(10);
			output << 0 << " ";
			output.width(5);
			output << 0 << " f \n";
		}
		else
		{
			output.width(10);
			output << offsetIterator->second << " ";
			output.width(5);
			output << 0 << " n \n";
		}
	}

	// have to do oldschool trailer output
	output << "trailer\n";
	trailer->write(output, writeStyle == WriteStyle::kCompact ? 0 : 1);
	output << "\nstartxref\n";
	output << xrefPosition << "\n";
}
