#ifndef LEMONPDF_PDFFILE_H
#define LEMONPDF_PDFFILE_H 1

#include <array>
#include <cstddef>
#include <fstream>
#include <functional>
#include <iostream>
#include <string>
#include <map>
#include <memory>
#include <unordered_set>

#include "pdfobject.h"
#include "pdfpage.h"

namespace LemonPDF
{
	using PageEnumerationFunction = std::function<void(PagePtr)>;
	using StreamEnumerationFunction = std::function<void(const ObjectReference &, StreamPtr)>;

	using ObjectReferenceSet = std::unordered_set<ObjectReference, ObjectReferenceHash>;
	
	class PDFFile;
	using PDFFilePtr = std::shared_ptr<PDFFile>;

	class Page;
	
	class PDFFile : public std::enable_shared_from_this<PDFFile>
	{
	private:
		// Constructors need to be public for std::make_shared, but at the same time,
		// the class should not be constructable outside of this class's dedicated
		// creation methods. This private class is a required argument to the public
		// constructors.
		class ConstructionToken
		{
			ConstructionToken() {}
			friend class PDFFile;
		};
	
	public:
		enum WriteStyle
		{
			kPadded,
			kCompact
		};

		PDFFile(const ConstructionToken &, size_t major, size_t minor);
		PDFFile(const ConstructionToken &, const std::string &filename);
		PDFFile(const ConstructionToken &, std::ifstream &&file);

		static PDFFilePtr EmptyFile(size_t major = 1, size_t minor = 4);
		static PDFFilePtr FromFile(const std::string &name);
		static PDFFilePtr FromStream(std::ifstream &&file);

		ObjectReference addObject(ObjectPtr object);
		ObjectReference appendPage(PagePtr page);
		Rectangle convertArrayToRectangle(ArrayPtr array);
		TransformMatrix convertArrayToTransformMatrix(ArrayPtr array);
		void enumeratePages(PageEnumerationFunction);
		void enumerateStreamsOfType(const std::string &type, StreamEnumerationFunction function);
		void enumerateStreamsOfType(const std::string &type, const std::string &subtype, StreamEnumerationFunction function);
		DictionaryPtr getDocumentCatalog() const;
		DictionaryPtr getInfoDictionary() const;
		size_t getMajorVersion() const;
		size_t getMinorVersion() const;
		std::vector<std::shared_ptr<Page>> getPages();
		DictionaryPtr getTrailer() const;
		const ObjectReferenceMap<size_t> &getXrefTable() const;
		const ObjectReferenceMap<CompressedObjectReference> &getCompressedXrefTable() const;
		ObjectPtr removeObject(const ObjectReference &ref);
		void replaceObject(const ObjectReference &ref, ObjectPtr object);
		
		template<typename T>
		std::shared_ptr<T> resolve(ObjectPtr o, bool throwOnMissing = true)
		{
			return std::dynamic_pointer_cast<T>(resolveObject(o, throwOnMissing));
		}

		template<typename T>
		std::shared_ptr<T> resolve(const ObjectReference &ref, bool throwOnMissing = true)
		{
			return std::dynamic_pointer_cast<T>(resolve(ref, throwOnMissing));
		}

		void resolveEntireFile(bool filterObjects = false);
		void setDocumentCatalog(DictionaryPtr);
		void setInfoDictionary(DictionaryPtr);
		void setTrailer(DictionaryPtr);
		void setVersion(size_t major, size_t minor);
		void write(const std::string &filename, WriteStyle ws = WriteStyle::kCompact) const;
		void write(std::ofstream &out, WriteStyle ws = WriteStyle::kCompact) const;
		
	private:
		void addCompressedXref(const ObjectReference &, const CompressedObjectReference &);
		void addXref(const ObjectReference &, size_t offset);
		void parse();
		void parseXrefTables();
		void parseXrefObjectStream(StreamPtr);
		void parseXrefTableOldStyle(std::istream &);
		ObjectPtr resolve(const ObjectReference &, const CompressedObjectReference &, bool throwOnMissing = true);
		ObjectPtr resolve(const ObjectReference &, bool throwOnMissing = true);
		ObjectPtr resolveObject(ObjectPtr, bool throwOnMissing = true);

		void writeOldschoolXrefPDF(std::ofstream &out, WriteStyle ws) const;
		void writeCompressedXrefPDF(std::ofstream &out, WriteStyle ws) const;
		
		size_t m_majorVersion;
		size_t m_minorVersion;
		std::ifstream m_input;
		std::streampos m_dataOffset;
		DictionaryPtr m_info;
		DictionaryPtr m_root;
		DictionaryPtr m_trailer;
		DictionaryPtr m_encrypt;
		ObjectReferenceMap<size_t> m_xref;
		ObjectReferenceMap<CompressedObjectReference> m_compressedXref;
		ObjectReferenceMap<ObjectPtr> m_objects;
		ObjectReferenceSet m_resolverSet;
		size_t m_largestObjectID;
		ResolverFunction m_resolver;
		std::vector<char> m_fileBuffer;
	};
}

#endif
