#include <sstream>
#include <vector>
#include <lemonsqueezy/lzw.h>
#include <lemonsqueezy/pngprediction.h>
#include <lemonsqueezy/tiffprediction.h>
#include <lemonsqueezy/zlib.h>
#include "pdffilter-flate-lzw.h"

LemonPDF::FlateLZWDecodeFilter::FlateLZWDecodeFilter(enum Mode mode, unsigned predictor, unsigned colors, unsigned bitsPerComponent, unsigned columns)
  : m_mode(mode), m_predictor(predictor), m_colors(colors), m_bitsPerComponent(bitsPerComponent), m_columns(columns)
{
}

void LemonPDF::FlateLZWDecodeFilter::decode(std::istream &in, std::ostream &out)
{
	std::unique_ptr<LemonSqueezy::IDecompressor> decompressor;

	if (m_mode == Mode::kFlateDecode)
		decompressor = std::make_unique<LemonSqueezy::ZLibCompression>();
	else if (m_mode == Mode::kLZWDecode)
		decompressor = std::make_unique<LemonSqueezy::LZWCompression>();
	else
		throw PDFInvalidOperationException("Invalid mode for FlateDecode+LZWDecode combo filter");
	
	if (m_predictor == 1)
	{
		decompressor->read(in, out);
	}
	else 
	{
		std::unique_ptr<LemonSqueezy::IDecompressor> prediction;

		if (m_predictor == 2)
			prediction = std::make_unique<LemonSqueezy::TIFFPrediction>(m_colors, m_bitsPerComponent, m_columns);
		else
			prediction = std::make_unique<LemonSqueezy::PNGPrediction>(m_predictor, m_colors, m_bitsPerComponent, m_columns);
		
		std::stringstream intermediate;
		decompressor->read(in, intermediate);
		prediction->read(intermediate, out);
	}
}

void LemonPDF::FlateLZWDecodeFilter::encode(std::istream &in, std::ostream &out)
{
	std::unique_ptr<LemonSqueezy::ICompressor> compressor;

	if (m_mode == Mode::kFlateDecode)
		compressor = std::make_unique<LemonSqueezy::ZLibCompression>();
	else if (m_mode == Mode::kLZWDecode)
		compressor = std::make_unique<LemonSqueezy::LZWCompression>();
	else
		throw PDFInvalidOperationException("Invalid mode for FlateDecode+LZWDecode combo filter");
	
	if (m_predictor == 1)
	{
		compressor->write(in, out);
	}
	else 
	{
		std::unique_ptr<LemonSqueezy::ICompressor> prediction;

		if (m_predictor == 2)
			prediction = std::make_unique<LemonSqueezy::TIFFPrediction>(m_colors, m_bitsPerComponent, m_columns);
		else
			prediction = std::make_unique<LemonSqueezy::PNGPrediction>(m_predictor, m_colors, m_bitsPerComponent, m_columns);
		
		std::stringstream intermediate;
		prediction->write(intermediate, out);
		compressor->write(in, intermediate);
	}
}

std::string LemonPDF::FlateLZWDecodeFilter::getName() const
{
	switch (m_mode)
	{
		case Mode::kFlateDecode:
			return "FlateDecode";
		
		case Mode::kLZWDecode:
			return "LZWDecode";
		
		default:
			throw PDFInvalidOperationException("Invalid decode mode for FlateDecode+LZWDecode combo filter");
	}
}
