#include <algorithm>
#include <cctype>
#include <climits>
#include <iomanip>
#include <limits>
#include <sstream>
#include "libtokenisation.h"
#include "pdfexception.h"
#include "pdfobject.h"
#include "pdftokeniser.h"

namespace
{
	bool isPrintable(const char ch)
	{
		return ::isprint(ch) && !(static_cast<unsigned char>(ch) & 0x80);
	}

	std::string generateReadableString(const std::string &str)
	{
		std::ostringstream result;
		int bracketBalance = 0;
		bool foundOpeningBracket = false;
		
		for (const auto &ch : str)
		{
			if (ch == '(')
			{
				foundOpeningBracket = true;
				bracketBalance++;
			}
			else if (ch == ')')
			{
				// if a closing bracket is found before an opening
				// bracket, then the balance is assumed to be off,
				// meaning all bracket characters will be escaped
				if (!foundOpeningBracket)
				{
					bracketBalance = -1;
					break;
				}

				bracketBalance--;
			}
		}

		result.put('(');
		for (const auto &ch : str)
		{
			switch (ch)
			{
				case '\n':
					result << "\\n";
					break;
				
				case '\r':
					result << "\\r";
					break;
				
				case '\t':
					result << "\\t";
					break;
				
				case '\b':
					result << "\\b";
					break;
				
				case '\f':
					result << "\\f";
					break;
				
				case '(':
				case ')':
					if (bracketBalance != 0)
						result.put('\\');
					result.put(ch);
					break;
				
				case '\\':
					result << "\\\\";
					break;
				
				default:
					if (!isPrintable(ch))
					{
						unsigned char uch = ch;
						result.put('\\');
						result << std::oct << std::setfill('0') << std::setw(3) << static_cast<int>(uch);
					}
					else
					{
						result.put(ch);
					}
			}
		}

		result.put(')');
		return result.str();
	}

	std::string generateHexString(const std::string &str)
	{
		std::ostringstream result;
		result.fill('0');
		result.setf(std::ios_base::hex, std::ios_base::basefield);

		result.put('<');
		for (const auto &ch : str)
		{
			unsigned char byte = static_cast<unsigned char>(ch);
			result.width(2);
			result << static_cast<unsigned>(byte);
		}
		result.put('>');
		return result.str();
	}

	std::string generateFixedNotation(double val)
	{
		// C++ offers no built-in method to use only the number of
		// digits required for a particular float to be represented
		// in decimal. This method will format a double using the
		// maximum number of characters needed and then strip off
		// the trailing zeroes (and the trailing decimal point if
		// needed). Kinda ugly, but, PDF doesn't like scientific
		// notation, so it's a necessary evil.
		std::ostringstream formatter;
		formatter.precision(std::numeric_limits<double>::digits10);
		formatter << std::fixed << val;
		std::string result = formatter.str();
		
		result.erase(result.find_last_not_of('0') + 1, std::string::npos);

		if (result.back() == '.')
			result.pop_back();
		
		return result;
	}

	std::string prettifyName(const std::string &str)
	{
		static char toHex[] = {
			'0', '1', '2', '3', '4', '5', '6', '7',
			'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
		};

		std::string result;
		result.reserve(str.size());
		
		for (const char &ch : str)
		{
			unsigned char uch = static_cast<unsigned char>(ch);
			if (uch >= 0x80 || ::isspace(ch) || LemonPDF::Tokenisation::IsDelimeter(ch) || !::isprint(ch))
			{
				result.push_back('#');
				result.push_back(toHex[(uch >> 4) & 0xF]);
				result.push_back(toHex[(uch >> 0) & 0xF]);
			}
			else
			{
				result.push_back(ch);
			}
		}
		
		return result;
	}

	std::string prettifyString(const std::string &str)
	{
		size_t printableCount = std::count_if(str.begin(), str.end(), isPrintable);

		// if there are >=75% printable, assume it is 
		// really a readable string
		if (printableCount >= str.size() * 3 / 4)
			return generateReadableString(str);
		else
			return generateHexString(str);
	}
}

LemonPDF::Literal::Literal()
  : m_type(kNull), m_value(0.0)
{
}

LemonPDF::Literal::Literal(bool b)
  : m_type(kBoolean), m_value(b)
{
}

LemonPDF::LiteralPtr LemonPDF::Literal::FromBoolean(bool b)
{
	static LiteralPtr s_true = std::shared_ptr<Literal>(new Literal(true));
	static LiteralPtr s_false = std::shared_ptr<Literal>(new Literal(false));
	return b ? s_true : s_false;
}

LemonPDF::LiteralPtr LemonPDF::Literal::FromInteger(long long i)
{
	LiteralPtr result = std::shared_ptr<LemonPDF::Literal>(new LemonPDF::Literal());
	result->m_type = kInteger;
	result->m_value.integer = i;
	return result;
}

LemonPDF::LiteralPtr LemonPDF::Literal::FromObjectReference(const ObjectReference &ref)
{
	LiteralPtr result = std::shared_ptr<LemonPDF::Literal>(new LemonPDF::Literal());
	result->m_type = kReference;
	result->m_reference = ref;
	return result;
}

LemonPDF::LiteralPtr LemonPDF::Literal::FromName(const std::string &name)
{
	LiteralPtr result = std::shared_ptr<LemonPDF::Literal>(new LemonPDF::Literal());
	result->m_type = kName;
	result->m_string = name;
	return result;
}

LemonPDF::LiteralPtr LemonPDF::Literal::FromReal(double r)
{
	LiteralPtr result = std::shared_ptr<LemonPDF::Literal>(new LemonPDF::Literal());
	result->m_type = kReal;
	result->m_value.real = r;
	return result;
}

LemonPDF::LiteralPtr LemonPDF::Literal::FromString(const std::string &str)
{
	LiteralPtr result = std::shared_ptr<LemonPDF::Literal>(new LemonPDF::Literal());
	result->m_type = kString;
	result->m_string = str;
	return result;
}

LemonPDF::LiteralPtr LemonPDF::Literal::Null()
{
	static LiteralPtr s_null = std::shared_ptr<LemonPDF::Literal>(new LemonPDF::Literal());
	return s_null;
}

LemonPDF::LiteralPtr LemonPDF::Literal::Parse(Tokeniser &t)
{
	Token token = t.getNextToken();
	
	switch (token.getType())
	{
		case Token::kTrue:
			return FromBoolean(true);
		
		case Token::kFalse:
			return FromBoolean(false);
		
		case Token::kReal:
			return FromReal(strtod(token.getText().c_str(), nullptr));
		
		case Token::kInteger:
			return FromInteger(strtoull(token.getText().c_str(), nullptr, 10));
		
		case Token::kName:
			return FromName(token.getText().substr(1));
		
		case Token::kString:
			return FromString(token.getText());
		
		case Token::kNull:
			return Null();
		
		case Token::kReference:
			return FromObjectReference(ObjectReference(token.getObjectIDAndGeneration()));
		
		default:
			throw PDFSyntaxException("Not a literal", t.getInputStream());
	}
}

LemonPDF::Literal::~Literal()
{
}

LemonPDF::Literal::operator std::string() const
{
	std::ostringstream ss;
	
	switch (m_type)
	{
		case kBoolean:
			ss << "<Bool " << (m_value.boolean ? "true" : "false") << ">";
			break;
		
		case kInteger:
			ss << "<Integer " << m_value.integer << ">";
			break;
		
		case kName:
			ss << "<Name /" << m_string << ">";
			break;
		
		case kNull:
			ss << "<Null>";
			break;
		
		case kReal:
			ss << "<Real " << m_value.real << ">";
			break;
		
		case kReference:
			ss << "<Ref " << m_reference.getObjectID() << " " << m_reference.getGeneration() << ">";
			break;
		
		case kString:
			ss << "<String \"" << m_string << "\">";
			break;
	}
	
	return ss.str();
}

LemonPDF::ObjectPtr LemonPDF::Literal::copy() const
{
	switch (m_type)
	{
		case kBoolean:
			return Literal::FromBoolean(getBooleanValue());
		
		case kInteger:
			return Literal::FromInteger(getIntegerValue());
		
		case kName:
			return Literal::FromName(getNameValue());
		
		case kNull:
			return Literal::Null();
		
		case kReal:
			return Literal::FromReal(getRealValue());
		
		case kReference:
			return Literal::FromObjectReference(getReference());
		
		case kString:
			return Literal::FromString(getStringValue());
		
		default:
			throw PDFUnimplementedException("Don't know how to copy this literal type");
	}
}

bool LemonPDF::Literal::getBooleanValue() const
{
	if (m_type == kBoolean)
		return m_value.boolean;
	else
		throw PDFInvalidOperationException("Object not a boolean value");
}

long long LemonPDF::Literal::getIntegerValue() const
{
	if (m_type == kInteger)
		return m_value.integer;
	else if (m_type == kReal)
		return (long long) m_value.real;
	else
		throw PDFInvalidOperationException("Object not an integer value");
}

const std::string &LemonPDF::Literal::getNameValue() const
{
	if (m_type == kName)
		return m_string;
	else
		throw PDFInvalidOperationException("Object not a name value");
}

double LemonPDF::Literal::getRealValue() const
{
	if (m_type == kReal)
		return m_value.real;
	else if (m_type == kInteger)
		return static_cast<double>(m_value.integer);
	else
		throw PDFInvalidOperationException("Object not a real value");
}

const LemonPDF::ObjectReference &LemonPDF::Literal::getReference() const
{
	if (m_type == kReference)
		return m_reference;
	else
		throw PDFInvalidOperationException("Object not a reference");
}

const std::string &LemonPDF::Literal::getStringValue() const
{
	if (m_type == kString)
		return m_string;
	else
		throw PDFInvalidOperationException("Object not a string");
}

enum LemonPDF::Literal::Type LemonPDF::Literal::getType() const
{
	return m_type;
}

bool LemonPDF::Literal::isNull() const
{
	return m_type == kNull;
}

void LemonPDF::Literal::write(std::ostream &out, size_t depth) const
{
	(void) depth;

	switch (m_type)
	{
		case kBoolean:
			out << (m_value.boolean ? "true" : "false");
			break;
		
		case kInteger:
			out << m_value.integer;
			break;
		
		case kName:
			out << "/" << prettifyName(m_string);
			break;
		
		case kNull:
			out << "null";
			break;
		
		case kReal:
			out << generateFixedNotation(m_value.real);
			break;
		
		case kReference:
			out << m_reference.getObjectID() << " " << m_reference.getGeneration() << " R";
			break;
		
		case kString:
			out << prettifyString(m_string);
			break;
		
		default:
			throw PDFUnimplementedException("Don't know how to write this type");
	}
}

LemonPDF::Literal::ValueUnion::ValueUnion(double d)
  : real(d)
{
}

LemonPDF::Literal::ValueUnion::ValueUnion(long long i)
  : integer(i)
{
}

LemonPDF::Literal::ValueUnion::ValueUnion(bool b)
  : boolean(b)
{
}
