#ifndef LEMONPDF_PDFOBJECT_H
#define LEMONPDF_PDFOBJECT_H 1

#include <cstddef>
#include <functional>
#include <memory>
#include <string>
#include <unordered_map>
#include <vector>

#include "pdffilter.h"

namespace LemonPDF
{
	class PDFFile;
	class Object;
	class Array;
	class Dictionary;
	class Literal;
	class Stream;
	class Tokeniser;
	
	typedef std::shared_ptr<Object> ObjectPtr;
	typedef std::shared_ptr<Array> ArrayPtr;
	typedef std::shared_ptr<Dictionary> DictionaryPtr;
	typedef std::shared_ptr<Literal> LiteralPtr;
	typedef std::shared_ptr<Stream> StreamPtr;
	
	typedef std::function<ObjectPtr(ObjectPtr)> ResolverFunction;
	
	std::ostream &operator<<(std::ostream &, ObjectPtr);
	
	/* Small class to represent an ObjectID/Generation pair */
	class ObjectReference
	{
	public:
		ObjectReference();
		ObjectReference(size_t id, size_t gen = 0);
		ObjectReference(std::pair<size_t, size_t>);
		virtual ~ObjectReference();

		virtual bool operator<(const ObjectReference &) const;
		virtual bool operator==(const ObjectReference &) const;
		
		virtual size_t getGeneration() const;
		virtual size_t getHash() const;
		virtual size_t getObjectID() const;
		
	protected:
		size_t m_objectID;
		size_t m_generation;
	};
	
	/* Small class to represent an ObjectID/Generation/Offset tuple */
	class CompressedObjectReference : public ObjectReference
	{
	public:
		CompressedObjectReference();
		CompressedObjectReference(size_t id, size_t offset);
		CompressedObjectReference(std::pair<size_t, size_t>);
		virtual ~CompressedObjectReference();

		virtual bool operator<(const CompressedObjectReference &) const;
		virtual bool operator==(const CompressedObjectReference &) const;
		
		virtual size_t getHash() const;
		virtual size_t getIndex() const;
		
	private:
		size_t m_index;
	};
	
	class ObjectReferenceHash
	{
	public:
		size_t operator()(const ObjectReference &) const;
	};
	
	template<typename T>
	using ObjectReferenceMap = std::unordered_map<ObjectReference, T, ObjectReferenceHash>;
	
	
	/**
	 * Base class for all PDF objects.
	 */
	class Object
	{
	public:
		static ObjectPtr Parse(Tokeniser &, ResolverFunction resolver = nullptr);
		explicit virtual operator std::string() const;
		virtual ObjectPtr copy() const;
		virtual void write(std::ostream &out, size_t depth) const;
		virtual ~Object();
	
	protected:
		Object();
	};
	
	/**
	 * Array PDF object. Can contain any PDFObject.
	 */
	class Array : public Object
	{
	public:
		static ArrayPtr Parse(Tokeniser &);
		explicit virtual operator std::string() const override;
		virtual ObjectPtr copy() const override;
		virtual void write(std::ostream &out, size_t depth) const override;
		~Array();
		
		void add(ObjectPtr o);
		void insert(size_t i, ObjectPtr o);
		ObjectPtr get(size_t) const;
		size_t getSize() const;
		void remove(size_t i);
		void truncate(size_t i);
		ArrayPtr shallowCopy() const;
		
		template<typename T>
		inline std::shared_ptr<T> get(size_t i) const
		{
			return std::dynamic_pointer_cast<T>(get(i));
		}
		
		void enumerate(std::function<void(size_t, ObjectPtr)>) const;
		void enumerate(std::function<void(ObjectPtr)>) const;
		
	private:
		std::vector<ObjectPtr> m_objects;
	};
	
	
	/**
	 * Dictionary PDF object. Can contain any PDF object but lookup
	 * must be done using either a string or name literal.
	 */
	class Dictionary : public Object
	{
	public:
		static DictionaryPtr Parse(Tokeniser &);
		explicit virtual operator std::string() const override;
		virtual ObjectPtr copy() const override;
		virtual void write(std::ostream &out, size_t depth) const override;
		~Dictionary();
		
		void add(const std::string &, ObjectPtr o);
		void add(const LiteralPtr, ObjectPtr o);
		bool contains(const std::string &) const;
		bool contains(const LiteralPtr) const;
		ObjectPtr get(const std::string &) const;
		ObjectPtr get(const LiteralPtr) const;
		size_t getSize() const;
		const std::string &getType() const;
		void remove(const std::string &);
		void remove(const LiteralPtr);
		DictionaryPtr shallowCopy() const;
		
		template<typename T>
		inline std::shared_ptr<T> get(const std::string &key) const
		{
			return std::dynamic_pointer_cast<T>(get(key));
		}
		
		template<typename T>
		inline std::shared_ptr<T> get(const LiteralPtr key) const
		{
			return std::dynamic_pointer_cast<T>(get(key));
		}
		
		void enumerate(std::function<void(const std::string &, ObjectPtr)>) const;
		
	private:
		std::unordered_map<std::string, ObjectPtr> m_objects;
	};
	
	
	/**
	 * Literal PDF object. This class describes literal objects
	 * in PDF files, such as names, strings, integers and
	 * references.
	 */
	class Literal : public Object
	{
	public:
		enum Type
		{
			kNull,
			kString,
			kName,
			kBoolean,
			kReal,
			kInteger,
			kReference
		};
		
		static LiteralPtr FromBoolean(bool);
		static LiteralPtr FromInteger(long long);
		static LiteralPtr FromObjectReference(const ObjectReference &);
		static LiteralPtr FromName(const std::string &);
		static LiteralPtr FromReal(double);
		static LiteralPtr FromString(const std::string &);
		static LiteralPtr Null();
		static LiteralPtr Parse(Tokeniser &);
		explicit virtual operator std::string() const override;
		virtual ObjectPtr copy() const override;
		virtual void write(std::ostream &out, size_t depth) const override;
		~Literal();
		
		bool getBooleanValue() const;
		long long getIntegerValue() const;
		const std::string &getNameValue() const;
		double getRealValue() const;
		const ObjectReference &getReference() const;
		const std::string &getStringValue() const;
		enum Type getType() const;
		bool isNull() const;
		
	private:
		Literal();
		Literal(bool);

		enum Type m_type;
		union ValueUnion
		{
			ValueUnion(double);
			ValueUnion(long long);
			ValueUnion(bool);

			double real;
			long long integer;
			bool boolean;
		} m_value;
		std::string m_string;
		ObjectReference m_reference;
	};
	
	
	/**
	 * PDF Stream.
	 */
	class Stream : public Object
	{
	public:
		static StreamPtr Parse(Tokeniser &, ResolverFunction resolver = nullptr, DictionaryPtr dict = nullptr);
		
		/**
		 * Create a Stream based on a std::istream. If filter names are supplied, then the data is first
		 * passed through the filters. If the data is already filtered, then don't supply any filter names
		 * at this point, and instead call setFilterNames() on the object returned from this function
		 */
		static StreamPtr FromIOStream(std::istream &&, const std::vector<std::string> &filterNames = std::vector<std::string> {});

		/**
		 * Create a Stream based on a buffer. If filter names are supplied, then the data is first
		 * passed through the filters. If the data is already filtered, then don't supply any filter names
		 * at this point, and instead call setFilterNames() on the object returned from this function
		 */
		static StreamPtr FromBuffer(const std::vector<unsigned char> &buffer, const std::vector<std::string> &filterNames = std::vector<std::string> {});
		
		explicit virtual operator std::string() const override;
		virtual ObjectPtr copy() const override;
		virtual void write(std::ostream &out, size_t depth) const override;
		Stream();
		~Stream();
		
		void filter(std::ostream &out) const;
		void filterIn(std::istream &in);
		bool filteringSupported() const;
		const std::vector<unsigned char> &getData() const;
		const DictionaryPtr getDictionary() const;
		std::vector<IFilterPtr> getFilters() const;
		std::vector<std::string> getFilterNames() const;
		std::vector<unsigned char> getFilteredData() const;
		const std::string &getType() const;
		void setData(const std::vector<unsigned char> &, const std::vector<IFilterPtr> &filters = std::vector<IFilterPtr>{});
		void setDictionary(DictionaryPtr newDict);
		void setDictionaryValue(const std::string &name, ObjectPtr object);
		void setDictionaryValue(const LiteralPtr name, ObjectPtr object);
		void setFilters(const std::vector<IFilterPtr> &filters);
		void setFilterNames(const std::vector<std::string> &newFilterNames);
		void setType(const std::string &newType);
		
	private:
		bool m_filteringSupported;
		DictionaryPtr m_streamDict;
		std::vector<unsigned char> m_data;
		std::vector<IFilterPtr> m_filters;
	};
}

#endif

