#include <sstream>
#include <libcore/logging.h>
#include "gfxtokeniser.h"
#include "gfxparser.h"
#include "pdfexception.h"
#include "pdffile.h"
#include "pdfpage.h"
#include "pdftypes.h"

namespace
{
	LemonPDF::ArrayPtr arrayFromRectangle(const LemonPDF::Rectangle &r)
	{
		LemonPDF::ArrayPtr array = std::make_shared<LemonPDF::Array>();
		array->add(LemonPDF::Literal::FromReal(r.lowerLeft.x));
		array->add(LemonPDF::Literal::FromReal(r.lowerLeft.y));
		array->add(LemonPDF::Literal::FromReal(r.upperRight.x));
		array->add(LemonPDF::Literal::FromReal(r.upperRight.y));
		return array;
	}
}

LemonPDF::Page::Page(PDFFilePtr pdf, DictionaryPtr object, PageTreeNodePtr parent)
  : PageTreeNode(pdf, object, parent),
    m_lastModified(0),
    m_rotation(0),
    m_userUnit(1.0)
{
	LiteralPtr lastModified = pdf->resolve<Literal>(object->get("LastModified"));
	if (lastModified)
		m_lastModified = LemonPDF::Types::ParseTimestamp(lastModified->getStringValue());
	
	// resources are inheritable
	DictionaryPtr resourceDictionary = pdf->resolve<Dictionary>(lookup("Resources"));
	if (!resourceDictionary)
		throw PDFInvalidOperationException("Page does not have a /Resources dictionary");
	
	m_resources = std::make_shared<ResourceManager>(pdf, resourceDictionary);
	
	// media boxes are inheritable
	ArrayPtr mediaBox = lookup<Array>("MediaBox");
	if (mediaBox)
		m_mediaBox = pdf->convertArrayToRectangle(mediaBox);
	
	// crop boxes are inheritable
	ArrayPtr cropBox = lookup<Array>("CropBox");
	if (cropBox)
		m_cropBox = pdf->convertArrayToRectangle(cropBox);
	else
		m_cropBox = m_mediaBox;
	
	ArrayPtr bleedBox = object->get<Array>("BleedBox");
	if (bleedBox)
		m_bleedBox = pdf->convertArrayToRectangle(bleedBox);
	else
		m_bleedBox = m_cropBox;
	
	ArrayPtr trimBox = object->get<Array>("TrimBox");
	if (trimBox)
		m_trimBox = pdf->convertArrayToRectangle(trimBox);
	else
		m_trimBox = m_cropBox;
	
	ArrayPtr artBox = object->get<Array>("ArtBox");
	if (artBox)
		m_artBox = pdf->convertArrayToRectangle(artBox);
	else
		m_artBox = m_cropBox;
	
	LiteralPtr rotate = lookup<Literal>("Rotate");
	if (rotate)
		m_rotation = rotate->getIntegerValue();
	
	ObjectPtr contentStreams = lookup("Contents");
	if (contentStreams)
	{
		ArrayPtr contentStreamsArray = pdf->resolve<Array>(contentStreams);
		if (contentStreamsArray)
		{
			contentStreamsArray->enumerate([&](ObjectPtr o){
				StreamPtr contentStream = pdf->resolve<Stream>(o);
				if (!contentStream)
					throw PDFInvalidOperationException("Array must contain only streams");
				m_contentStreams.push_back(contentStream);
			});
		}
		else
		{
			StreamPtr contentStream = pdf->resolve<Stream>(contentStreams);
			if (!contentStream)
				throw PDFInvalidOperationException("Content stream must be an array or stream object");
			
			m_contentStreams.push_back(contentStream);
		}
	}
}

LemonPDF::Page::~Page()
{
}

const LemonPDF::Rectangle &LemonPDF::Page::getArtBox() const
{
	return m_artBox;
}

const LemonPDF::Rectangle &LemonPDF::Page::getBleedBox() const
{
	return m_bleedBox;
}

const std::vector<LemonPDF::StreamPtr> &LemonPDF::Page::getContentStreams() const
{
	return m_contentStreams;
}

const LemonPDF::Rectangle &LemonPDF::Page::getCropBox() const
{
	return m_cropBox;
}

time_t LemonPDF::Page::getLastModified() const
{
	return m_lastModified;
}

const LemonPDF::Rectangle &LemonPDF::Page::getMediaBox() const
{
	return m_mediaBox;
}

LemonPDF::ResourceManagerPtr LemonPDF::Page::getResources() const
{
	return m_resources;
}

const LemonPDF::Rectangle &LemonPDF::Page::getTrimBox() const
{
	return m_trimBox;
}

double LemonPDF::Page::getUserUnit() const
{
	return m_userUnit;
}

void LemonPDF::Page::setArtBox(const Rectangle &r)
{
	m_artBox = r;
	m_dictionary->add("ArtBox", arrayFromRectangle(r));
}

void LemonPDF::Page::setBleedBox(const Rectangle &r)
{
	m_bleedBox = r;
	m_dictionary->add("BleedBox", arrayFromRectangle(r));
}

void LemonPDF::Page::setCropBox(const Rectangle &r)
{
	m_cropBox = r;
	m_dictionary->add("CropBox", arrayFromRectangle(r));
}

void LemonPDF::Page::setMediaBox(const Rectangle &r)
{
	m_mediaBox = r;
	m_dictionary->add("MediaBox", arrayFromRectangle(r));
}

void LemonPDF::Page::setTrimBox(const Rectangle &r)
{
	m_trimBox = r;
	m_dictionary->add("TrimBox", arrayFromRectangle(r));
}

void LemonPDF::Page::render(Graphics::IContext &context) const
{
	std::stringstream ss;
	for (StreamPtr s : m_contentStreams)
		s->filter(ss);
	
	Graphics::GFXTokeniser tokeniser(ss);
	Graphics::GFXParser parser(tokeniser, context);
	
	parser.parseAll(m_resources);
}
