#include "pdfexception.h"
#include "pdffile.h"
#include "pdfpagetreenode.h"

LemonPDF::PageTreeNodePtr LemonPDF::PageTreeNode::ConstructTree(PDFFilePtr pdf, DictionaryPtr dict, PageTreeNodePtr parent)
{
	PageTreeNodePtr result;
	
	if (dict->getType() == "Pages")
	{
		result = std::make_shared<PageTreeNode>(pdf, dict, parent);
		
		ArrayPtr kids = dict->get<Array>("Kids");
		if (!kids)
			throw PDFInvalidOperationException("Kids array is empty (but shouldn't be)");
		
		kids->enumerate([=](ObjectPtr o){
			DictionaryPtr nextObject = pdf->resolve<Dictionary>(o);
			if (!nextObject)
				throw PDFInvalidOperationException("Kids elements must be dictionaries");
			
			result->m_children.emplace_back(ConstructTree(pdf, nextObject, result));
		});
	}
	else if (dict->getType() == "Page")
	{
		result = std::make_shared<Page>(pdf, dict, parent);
	}
	else
	{
		throw PDFInvalidOperationException("Expected /Page or /Pages type");
	}
	
	return result;
}

std::vector<LemonPDF::PagePtr> LemonPDF::PageTreeNode::GetPageObjects(PageTreeNodePtr root)
{
	std::vector<PagePtr> result;
	
	for (PageTreeNodePtr node : root->m_children)
	{
		PagePtr page = std::dynamic_pointer_cast<Page>(node);
		if (page)
		{
			result.push_back(page);
		}
		else
		{
			std::vector<PagePtr> childPages = GetPageObjects(node);
			result.insert(result.end(), childPages.begin(), childPages.end());
		}
	}
	
	return result;
}

LemonPDF::PageTreeNode::PageTreeNode(PDFFilePtr pdf, DictionaryPtr object, PageTreeNodePtr parent)
  : m_dictionary(object), m_file(pdf), m_parent(parent)
{

}

LemonPDF::PageTreeNode::~PageTreeNode()
{
}

size_t LemonPDF::PageTreeNode::getChildCount() const
{
	return m_children.size();
}

LemonPDF::PageTreeNodePtr LemonPDF::PageTreeNode::getChild(size_t index) const
{
	return m_children.at(index);
}

LemonPDF::DictionaryPtr LemonPDF::PageTreeNode::getDictionary() const
{
	return m_dictionary;
}

LemonPDF::PDFFilePtr LemonPDF::PageTreeNode::getFile() const
{
	return m_file.lock();
}

LemonPDF::PageTreeNodePtr LemonPDF::PageTreeNode::getParent() const
{
	return m_parent.lock();
}

LemonPDF::ObjectPtr LemonPDF::PageTreeNode::lookup(const std::string &name) const
{
	ObjectPtr result = m_dictionary->get(name);
	if (result)
		return result;
	
	CPageTreeNodePtr node = m_parent.lock();
	while (node)
	{
		result = node->lookup(name);
		if (result)
			return result;
		node = node->m_parent.lock();
	}
	
	return nullptr;
}

LemonPDF::ObjectPtr LemonPDF::PageTreeNode::lookup(LiteralPtr l) const
{
	return lookup(l->getNameValue());
}

void LemonPDF::PageTreeNode::removeChild(size_t index)
{
	m_children.erase(m_children.begin() + index);
	
	PDFFilePtr file = getFile();
	if (!file)
		return;
	
	ArrayPtr kidsArray = file->resolve<Array>(m_dictionary->get("Kids"));
	kidsArray->remove(index);

	LiteralPtr count = file->resolve<Literal>(m_dictionary->get("Count"));
	m_dictionary->add("Count", Literal::FromInteger(count->getIntegerValue() - 1));
}
