#include <sstream>

#include <libcore/logging.h>

#include "pdfexception.h"
#include "pdfresourcemanager.h"
#include "pdffile.h"
#include "gfxparser.h"
#include "gfxtilingpattern.h"
#include "gfxtruetypefont.h"
#include "gfxshadingpattern.h"

namespace
{
	class UnimplementedFont : public LemonPDF::Graphics::Font
	{
	public:
		UnimplementedFont()
			: LemonPDF::Graphics::Font("UnimplementedFont", LemonPDF::Graphics::Font::FontCompleteness::kComplete, LemonPDF::Graphics::Font::FontType::kSymbolic, 400)
		{
		}
	};

	class BuiltInFont : public LemonPDF::Graphics::Font
	{
	public:
		BuiltInFont(const std::string &name)
			: LemonPDF::Graphics::Font(name, LemonPDF::Graphics::Font::FontCompleteness::kComplete, LemonPDF::Graphics::Font::FontType::kScript, 400)
		{
		}

	private:

	};

	// CIDFontType glyph widths are stored in an array, where the sequence of elements takes one
	// of two forms:
	//
	// -- firstChar, [ width, width, width ... ]
	// -- firstChar, lastChar, width
	//
	// in the first form, the widths are specified in the subarray such that the glyph width of
	// (firstChar + subarrayIndex) is subarray[subarrayIndex]
	//
	// in the second form, all characters between firstChar and lastChar inclusive is "width"
	LemonPDF::Graphics::GlyphWidthMap parseCIDFontGlyphWidths(LemonPDF::PDFFilePtr pdf, LemonPDF::DictionaryPtr fontDict)
	{
		LemonPDF::Graphics::GlyphWidthMap glyphWidths;

		LemonPDF::ArrayPtr widthArray = pdf->resolve<LemonPDF::Array>(fontDict->get("W"));
		if (!widthArray)
			return glyphWidths; // nothing to parse
		
		char32_t firstChar = '\0';
		char32_t lastChar = '\0';

		for (size_t i = 0; i < widthArray->getSize(); i++)
		{
			LemonPDF::ObjectPtr thisObject = widthArray->get(i);
			LemonPDF::LiteralPtr charIndex = pdf->resolve<LemonPDF::Literal>(thisObject);
			if (!charIndex)
			{
				LemonPDF::ArrayPtr subGlyphWidths = pdf->resolve<LemonPDF::Array>(thisObject);

				if (!subGlyphWidths)
					throw LemonPDF::PDFInvalidOperationException("Width array for CIDFontType is invalid");
				
				subGlyphWidths->enumerate([&glyphWidths,firstChar,pdf](size_t i, LemonPDF::ObjectPtr o){
					glyphWidths[firstChar + i] = pdf->resolve<LemonPDF::Literal>(o)->getRealValue();
				});
				
				firstChar = '\0';
				continue;
			}

			if (firstChar == '\0')
			{
				firstChar = charIndex->getIntegerValue();
			}
			else if (lastChar == '\0')
			{
				lastChar = charIndex->getIntegerValue();
			}
			else
			{
				for (char32_t ch = firstChar; ch <= lastChar; ch++)
					glyphWidths[ch] = charIndex->getRealValue();
				
				firstChar = '\0';
				lastChar = '\0';
			}
		}

		return glyphWidths;
	}

	LemonPDF::Graphics::GlyphWidthMap parseSimpleFontGlyphWidths(LemonPDF::PDFFilePtr pdf, LemonPDF::DictionaryPtr fontDict)
	{
		LemonPDF::Graphics::GlyphWidthMap glyphWidths;

		LemonPDF::LiteralPtr firstCharLiteral = pdf->resolve<LemonPDF::Literal>(fontDict->get("FirstChar"));
		LemonPDF::ArrayPtr glyphWidthArray = pdf->resolve<LemonPDF::Array>(fontDict->get("Widths"));
		if (!glyphWidthArray || !firstCharLiteral)
			throw LemonPDF::PDFInvalidOperationException("Font dictionary missing required glyph widths or first character");
		
		char32_t firstChar = firstCharLiteral->getIntegerValue();
		glyphWidthArray->enumerate([&glyphWidths,firstChar,pdf](size_t i, LemonPDF::ObjectPtr o) {
			glyphWidths[firstChar + i] = pdf->resolve<LemonPDF::Literal>(o)->getRealValue();
		});

		return glyphWidths;
	}

	LemonPDF::Graphics::GlyphWidthMap parseGlyphWidths(LemonPDF::PDFFilePtr pdf, LemonPDF::DictionaryPtr fontDict)
	{
		LemonPDF::LiteralPtr subtypeLiteral = pdf->resolve<LemonPDF::Literal>(fontDict->get("Subtype"));
		if (!subtypeLiteral)
			throw LemonPDF::PDFInvalidOperationException("Font dictionary is invalid");
		
		std::string subtype = subtypeLiteral->getNameValue();
		if (subtype == "Type0")
		{
			LemonPDF::ArrayPtr descendantFonts = pdf->resolve<LemonPDF::Array>(fontDict->get("DescendantFonts"));
			if (descendantFonts == nullptr || descendantFonts->getSize() != 1)
				throw LemonPDF::PDFInvalidOperationException("No descendant fonts for Type0 font");

			LemonPDF::DictionaryPtr descendantFontDict = pdf->resolve<LemonPDF::Dictionary>(descendantFonts->get(0));
			if (descendantFontDict == nullptr)
				throw LemonPDF::PDFInvalidOperationException("Descendent font is invalid or missing");
			
			return parseGlyphWidths(pdf, descendantFontDict);
		}
		if (subtype == "CIDFontType0" || subtype == "CIDFontType2")
		{
			return parseCIDFontGlyphWidths(pdf, fontDict);
		}
		else
		{
			return parseSimpleFontGlyphWidths(pdf, fontDict);
		}
	}
}

LemonPDF::ResourceManager::ResourceManager(PDFFilePtr pdf, DictionaryPtr resourceDictionary)
  : m_resources(resourceDictionary), m_pdf(pdf)
{
}

LemonPDF::ResourceManager::~ResourceManager()
{
}

void LemonPDF::ResourceManager::addColorSpace(const std::string &name, ObjectPtr object)
{
	PDFFilePtr pdf = getFile();
	DictionaryPtr colorSpaces = pdf->resolve<Dictionary>(m_resources->get("ColorSpace"));
	if (colorSpaces == nullptr)
	{
		colorSpaces = std::make_shared<Dictionary>();
		m_resources->add("ColorSpace", colorSpaces);
	}
	
	colorSpaces->add(name, object);
}

LemonPDF::Graphics::ColorSpacePtr LemonPDF::ResourceManager::getColorSpace(const std::string &name) const
{
	Graphics::ColorSpacePtr cs = Graphics::ColorSpace::GetNamedColorSpace(name);
	if (cs)
	{
		if (!cs->isSupported())
			StripedHorse::Logging::Warning() << cs->getName() << std::endl;
		return cs;
	}
	
	PDFFilePtr pdf = getFile();
	DictionaryPtr colorSpaces = pdf->resolve<Dictionary>(m_resources->get("ColorSpace"));
	if (!colorSpaces)
		throw std::runtime_error("No ColorSpace entry in resource dictionary");
		//return nullptr;
	
	ArrayPtr colorSpaceArray = pdf->resolve<Array>(colorSpaces->get(name));
	if (!colorSpaceArray)
		throw std::runtime_error("No entry in the ColorSpace dictionary for " + name);
	
	LiteralPtr colorSpaceNameLiteral = colorSpaceArray->get<Literal>(0);
	if (!colorSpaceNameLiteral)
		throw std::runtime_error("First element is not a literal");
	
	std::string colorSpaceName = colorSpaceNameLiteral->getNameValue();
	
	ObjectPtr data = pdf->resolve<Object>(colorSpaceArray->get(1));
	
	if (colorSpaceName == "CalRGB")
	{
		cs = Graphics::ColorSpace::CreateCalibratedRGBColorSpace(Graphics::XYZPoint{1.0, 1.0, 1.0}, Graphics::XYZPoint{0.0, 0.0, 0.0}, 1.0, 1.0, 1.0);
	}
	else if (colorSpaceName == "ICCBased")
	{
		size_t numberOfComponents = 3;
		StreamPtr profileStream = std::dynamic_pointer_cast<Stream>(data);
		DictionaryPtr streamDict = profileStream->getDictionary();
		if (streamDict)
		{
			LiteralPtr nLiteral = pdf->resolve<Literal>(streamDict->get("N"));
			if (nLiteral != nullptr)
				numberOfComponents = nLiteral->getIntegerValue();
		}

		cs = Graphics::ColorSpace::CreateICCBasedColorSpace(profileStream->getFilteredData(), numberOfComponents);
	}
	
	return cs;
}

LemonPDF::DictionaryPtr LemonPDF::ResourceManager::getDictionary() const
{
	return m_resources;
}

LemonPDF::PDFFilePtr LemonPDF::ResourceManager::getFile() const
{
	return m_pdf.lock();
}

LemonPDF::Graphics::FontPtr LemonPDF::ResourceManager::getFont(const std::string &name) const
{
	auto cached = m_fontCache.find(name);
	if (cached != m_fontCache.end())
		return cached->second;

	CMap cmap;
	PDFFilePtr pdf = getFile();
	DictionaryPtr fonts = pdf->resolve<Dictionary>(m_resources->get("Font"));
	if (!fonts)
		throw Graphics::GFXParserException("No Font entry in resource dictionary");
	
	DictionaryPtr thisFontDict = pdf->resolve<Dictionary>(fonts->get(name));
	if (!thisFontDict)
		throw Graphics::GFXParserException("No font with name " + name);
	
	LiteralPtr dictType = thisFontDict->get<Literal>("Type");
	if (!dictType || dictType->getNameValue() != "Font")
		throw Graphics::GFXParserException("Supplied dictionary is not a font dictionary");
	
	LiteralPtr dictSubtype = thisFontDict->get<Literal>("Subtype");
	if (!dictSubtype)
		throw Graphics::GFXParserException("Required field /Subtype is missing");

	LiteralPtr baseFontLiteral = pdf->resolve<Literal>(thisFontDict->get("BaseFont"));
	if (!baseFontLiteral)
		throw Graphics::GFXParserException("Required field /BaseFont is missing");

	enum Graphics::Font::FontCompleteness fontCompleteness = Graphics::Font::FontCompleteness::kComplete;
	std::string baseFont = baseFontLiteral->getNameValue();
	if (baseFont.size() > 7)
	{
		bool allAlpha = true;
		for (std::string::size_type i = 0; i < 6 && allAlpha; i++)
			if (!(::isalpha(baseFont[i]) && ::isupper(baseFont[i])))
				allAlpha = false;

		if (allAlpha && baseFont[6] == '+')
		fontCompleteness = Graphics::Font::FontCompleteness::kSubsetted;
	}

	StreamPtr unicodeMap = pdf->resolve<Stream>(thisFontDict->get("ToUnicode"));
	if (unicodeMap != nullptr)
	{
		const std::vector<unsigned char> &cmapData = unicodeMap->getFilteredData();
		std::istringstream dataStream { std::string { cmapData.begin(), cmapData.end() } };
		cmap = CMap::FromPostscriptStyle(dataStream);
	}

	enum Graphics::Font::FontType fontType = Graphics::Font::FontType::kSymbolic;
	int fontWeight = 400;
	DictionaryPtr fontDescriptor = pdf->resolve<Dictionary>(thisFontDict->get("FontDescriptor"));
	if (fontDescriptor)
	{
		LiteralPtr flagsLiteral = pdf->resolve<Literal>(fontDescriptor->get("Flags"));
		if (flagsLiteral)
			fontType = static_cast<Graphics::Font::FontType>(flagsLiteral->getIntegerValue());
		
		LiteralPtr fontWeightLiteral = pdf->resolve<Literal>(thisFontDict->get("Weight"));
		if (fontWeightLiteral)
			fontWeight = static_cast<int>(fontWeightLiteral->getIntegerValue());
	}

	Graphics::FontPtr result;
	if (dictSubtype->getNameValue() == "TrueType")
	{	
		StreamPtr fontFile = pdf->resolve<Stream>(fontDescriptor->get("FontFile2"));
		if (!fontFile)
			throw Graphics::GFXParserException("No font stream");
		
		result = std::make_shared<Graphics::TrueTypeFont>(baseFont, fontCompleteness, fontType, fontWeight, fontFile->getFilteredData());
	}
	else if (dictSubtype->getNameValue() == "Type1")
	{
		result = std::make_shared<Graphics::Type1Font>(baseFont, fontCompleteness, fontType, fontWeight);
	}
	else
	{
		result = std::make_shared<UnimplementedFont>();
	}

	// Override built-in glyph widths with PDF glyph widths (if they are available in the PDF)
	Graphics::GlyphWidthMap glyphWidths;
	try
	{
		glyphWidths = parseGlyphWidths(pdf, thisFontDict);
	}
	catch (PDFException &ex)
	{
		StripedHorse::Logging::Error() << "Font " << baseFont << " cannot be used because an exception occurred parsing the glyph widths: " << ex.what() << std::endl;
	}

	if (glyphWidths.size() > 0)
		result->setGlyphWidths(glyphWidths);
	
	LiteralPtr missingWidthLiteral = pdf->resolve<Literal>(thisFontDict->get("MissingWidth"));
	double missingWidth = missingWidthLiteral ? missingWidthLiteral->getRealValue() : 0.0;
	result->setDefaultGlyphWidth(missingWidth);

	LiteralPtr encodingLiteral = pdf->resolve<Literal>(thisFontDict->get("Encoding"));
	if (encodingLiteral != nullptr)
	{
		std::string encodingName = encodingLiteral->getNameValue();
		result->setEncodingName(encodingName);

		// don't override cmap if there is already one defined
		if (!cmap.hasMappings())
		{
			if (encodingName == "MacRomanEncoding")
				cmap = CMap::MacRomanEncoding();
			else if (encodingName == "WinAnsiEncoding")
				cmap = CMap::WinAnsiEncoding();
			else
				std::cerr << "Warning: [SEVERE] Encoding " << encodingName << " is not known" << std::endl;
		}
	}
	
	if (result && cmap.hasMappings())
		result->setCMap(cmap);
	
	m_fontCache.insert(std::make_pair(name, result));
	return result;
}

LemonPDF::Graphics::PatternPtr LemonPDF::ResourceManager::getPattern(const std::string &name) const
{
	PDFFilePtr pdf = getFile();
	DictionaryPtr patterns = pdf->resolve<Dictionary>(m_resources->get("Pattern"));

	if (patterns == nullptr)
		throw PDFInvalidOperationException("Pattern requested, but no patterns in resource dictionary");
	
	StreamPtr patternStream = pdf->resolve<Stream>(patterns->get(name));
	DictionaryPtr patternDict;
	if (patternStream != nullptr)
		patternDict = patternStream->getDictionary();
	else
		patternDict = pdf->resolve<Dictionary>(patterns->get(name));
	
	if (patternDict == nullptr)
		throw PDFInvalidOperationException("Pattern is invalid, it must be a stream or a dictionary");
	
	LiteralPtr patternTypeLiteral = pdf->resolve<Literal>(patternDict->get("PatternType"));
	long patternType = patternTypeLiteral->getIntegerValue();
	if (patternType == 1)
	{
		if (patternStream == nullptr)
			throw PDFInvalidOperationException("Pattern type 1 must be a stream");
		
		ResourceManagerPtr patternResources = std::make_shared<ResourceManager>(pdf, pdf->resolve<Dictionary>(patternDict->get("Resources")));

		std::stringstream patternData;
		patternStream->filter(patternData);
		patternData.seekg(0);
		Graphics::TilingPatternPtr tilingPattern = std::make_shared<Graphics::TilingPattern>(patternResources, Graphics::GFXTokenStore(patternData));
		
		Rectangle bbox = pdf->convertArrayToRectangle(pdf->resolve<Array>(patternDict->get("BBox")));
		tilingPattern->setBoundingBox(bbox);

		ArrayPtr matrixArray = pdf->resolve<Array>(patternDict->get("Matrix"));
		if (matrixArray)
		{
			TransformMatrix matrix = pdf->convertArrayToTransformMatrix(pdf->resolve<Array>(patternDict->get("Matrix")));
			tilingPattern->setMatrix(matrix);
		}

		LiteralPtr xStepLiteral = pdf->resolve<Literal>(patternDict->get("XStep"));
		LiteralPtr yStepLiteral = pdf->resolve<Literal>(patternDict->get("YStep"));
		
		if (xStepLiteral == nullptr || yStepLiteral == nullptr)
			throw PDFInvalidOperationException("Tiling pattern is missing XStep or YStep");
		
		tilingPattern->setStep(Point { xStepLiteral->getRealValue(), yStepLiteral->getRealValue() });

		return tilingPattern;
	}
	else if (patternType == 2)
	{
		// Unimplemented!
		return std::make_shared<Graphics::ShadingPattern>();
	}
	else
	{
		throw PDFUnimplementedException("Unable to handle pattern types other than 1 and 2");
	}

	return nullptr;
}

LemonPDF::StreamPtr LemonPDF::ResourceManager::getXObject(const std::string &name) const
{
	PDFFilePtr pdf = getFile();
	DictionaryPtr xobjectDict = pdf->resolve<Dictionary>(m_resources->get("XObject"));
	return pdf->resolve<Stream>(xobjectDict->get(name));
}