#include <sstream>
#include <libcore/logging.h>
#include <lemonsqueezy/exceptions.h>

#include "pdfexception.h"
#include "pdffilter.h"
#include "pdfobject.h"
#include "pdftokeniser.h"
#include "pdftypes.h"

#include "pdffilter-ascii85.h"
#include "pdffilter-asciihex.h"
#include "pdffilter-ccitt.h"
#include "pdffilter-dctdecode.h"
#include "pdffilter-flate-lzw.h"
#include "pdffilter-runlengthdecode.h"

namespace
{
	class PassthroughFilter : public LemonPDF::IFilter
	{
	public:
		PassthroughFilter(const std::string &name)
		  : m_name(name)
		{
		}

		virtual std::string getName() const override
		{
			return m_name;
		}

		virtual void decode(std::istream &in, std::ostream &out) override
		{
			while (in)
			{
				char buffer[4096];
				if (in.read(buffer, sizeof buffer))
					out.write(buffer, in.gcount());
			}
		}

		virtual void encode(std::istream &in, std::ostream &out) override
		{
			// the passthrough filter does the same action for both decoding and encoding
			decode(in, out);
		}

	private:
		std::string m_name;
	};

	class StreamDictHelper
	{
	public:
		StreamDictHelper(LemonPDF::DictionaryPtr streamDict, LemonPDF::ResolverFunction resolver)
		  : m_streamDict(streamDict), m_resolver(resolver) {}
		
		/// get a boolean value from the DecodeParms
		bool get(const std::string &name, bool defaultValue)
		{
			if (!m_streamDict)
				return defaultValue;
			
			LemonPDF::LiteralPtr literal = getLiteral(name);
			return literal ? literal->getBooleanValue() : defaultValue;
		}
		
		/// get an integer value from the DecodeParms
		int get(const std::string &name, int defaultValue)
		{
			if (!m_streamDict)
				return defaultValue;
			
			LemonPDF::LiteralPtr literal = getLiteral(name);
			return literal ? literal->getIntegerValue() : defaultValue;
		}

	private:
		LemonPDF::LiteralPtr getLiteral(const std::string &name)
		{
			// if we can resolve, we should try, otherwise, use default value
			if (m_resolver)
				return std::dynamic_pointer_cast<LemonPDF::Literal>(m_resolver(m_streamDict->get(name)));
			else
				return std::dynamic_pointer_cast<LemonPDF::Literal>(m_streamDict->get(name));
		}
		
		LemonPDF::DictionaryPtr m_streamDict;
		LemonPDF::ResolverFunction m_resolver;
	};

	LemonPDF::IFilterPtr getFilter(const std::string &name, LemonPDF::DictionaryPtr decodeParms, LemonPDF::ResolverFunction resolver)
	{
		StreamDictHelper helper(decodeParms, resolver);

		if (name == "ASCII85Decode" || name == "A85")
		{
			return std::make_shared<LemonPDF::ASCII85DecodeFilter>();
		}
		else if (name == "ASCIIHexDecode" || name == "AHx")
		{
			return std::make_shared<LemonPDF::ASCIIHexDecodeFilter>();
		}
		else if (name == "FlateDecode" || name == "Fl" || name == "LZWDecode" || name == "LZW")
		{
			LemonPDF::FlateLZWDecodeFilter::Mode mode = (name[0] == 'F' ? LemonPDF::FlateLZWDecodeFilter::Mode::kFlateDecode : LemonPDF::FlateLZWDecodeFilter::Mode::kLZWDecode);
			unsigned predictor = helper.get("Predictor", 1);
			unsigned colors = helper.get("Colors", 1);
			unsigned bitsPerComponent = helper.get("BitsPerComponent", 8);
			unsigned columns = helper.get("Columns", 1);
			return std::make_shared<LemonPDF::FlateLZWDecodeFilter>(mode, predictor, colors, bitsPerComponent, columns);
		}
		else if (name == "CCITTFaxDecode" || name == "CCF")
		{
			int errorLines = helper.get("K", 0);
			bool encodedByteAlign = helper.get("EncodedByteAlign", false);
			int columns = helper.get("Columns", 1728);
			int rows = helper.get("Rows", 0);
			bool endOfBlock = helper.get("EndOfBlock", true);
			bool blackIs1 = helper.get("BlackIs1", false);

			LemonSqueezy::CCITTCompression::Encoding encoding = LemonSqueezy::CCITTCompression::Encoding::kGroup4_2D;
			if (errorLines == 0)
				encoding = LemonSqueezy::CCITTCompression::Encoding::kGroup3_1D;
			else if (errorLines > 0)
				encoding = LemonSqueezy::CCITTCompression::Encoding::kGroup3_2D;

			return std::make_shared<LemonPDF::CCITTFaxDecodeFilter>(encoding, errorLines, encodedByteAlign, columns, rows, endOfBlock, blackIs1);
		}
		else if (name == "RunLengthDecode")
		{
			return std::make_shared<LemonPDF::RunLengthDecodeFilter>();
		}
		else if (name == "DCTDecode" || name == "DCT" || name == "JPXDecode")
		{
			return std::make_shared<LemonPDF::DCTDecodeFilter>();
		}
		
		return nullptr;
	}

	std::tuple<bool, std::vector<LemonPDF::IFilterPtr>> createFilters(LemonPDF::StreamPtr stream, LemonPDF::ResolverFunction resolver)
	{
		bool successful = true;
		std::vector<LemonPDF::IFilterPtr> result;
		LemonPDF::DictionaryPtr streamDict = stream->getDictionary();
		LemonPDF::ObjectPtr decodeParms = resolver(streamDict->get("DecodeParms"));
		LemonPDF::ObjectPtr filter = resolver(streamDict->get("Filter"));

		LemonPDF::LiteralPtr filterLiteral = std::dynamic_pointer_cast<LemonPDF::Literal>(filter);
		if (filterLiteral)
		{
			LemonPDF::IFilterPtr thisFilter = getFilter(filterLiteral->getNameValue(), std::dynamic_pointer_cast<LemonPDF::Dictionary>(decodeParms), resolver);
			if (thisFilter == nullptr)
			{
				successful = false;
				thisFilter = std::make_shared<PassthroughFilter>(filterLiteral->getNameValue());
			}
			
			result.emplace_back(thisFilter);
		}
		else
		{
			LemonPDF::ArrayPtr filterArray = std::dynamic_pointer_cast<LemonPDF::Array>(filter);
			if (!filterArray)
				return std::make_tuple(successful, result);
			
			LemonPDF::ArrayPtr decodeParmsArray = std::dynamic_pointer_cast<LemonPDF::Array>(decodeParms);

			for (size_t i = 0; i < filterArray->getSize() && successful; i++)
			{
				LemonPDF::DictionaryPtr thisDecodeParms;
				LemonPDF::LiteralPtr thisFilterLiteral = std::dynamic_pointer_cast<LemonPDF::Literal>(resolver(filterArray->get(i)));
				
				if (decodeParmsArray && i < decodeParmsArray->getSize())
					thisDecodeParms = std::dynamic_pointer_cast<LemonPDF::Dictionary>(resolver(decodeParmsArray->get(i)));
				
				if (!thisFilterLiteral)
					throw LemonPDF::PDFInvalidOperationException("Unable to determine filter name");
				
				LemonPDF::IFilterPtr thisFilter = getFilter(thisFilterLiteral->getNameValue(), thisDecodeParms, resolver);
				if (thisFilter == nullptr)
				{
					successful = false;
					thisFilter = std::make_shared<PassthroughFilter>(thisFilterLiteral->getNameValue());
				}
				
				result.emplace_back(thisFilter);
			}
		}

		return std::make_tuple(successful, result);
	}
}

LemonPDF::StreamPtr LemonPDF::Stream::Parse(Tokeniser &tokeniser, ResolverFunction resolver, DictionaryPtr dict)
{
	StreamPtr result = std::make_shared<Stream>();
	
	result->m_streamDict = dict;
	
	Token t = tokeniser.getNextToken();
	if (t.getType() != Token::kStreamStart)
		throw PDFSyntaxException("Expected 'stream'", tokeniser.getInputStream());
	
	LiteralPtr length = std::dynamic_pointer_cast<Literal>(result->m_streamDict->get("Length"));
	if (resolver)
		length = std::dynamic_pointer_cast<Literal>(resolver(length));
	
	if (length == nullptr)
		throw PDFSyntaxException("Length value missing or invalid", tokeniser.getInputStream());
	
	result->m_data = tokeniser.read(length->getIntegerValue());
	t = tokeniser.getNextToken();
	if (t.getType() != Token::kStreamEnd)
		throw PDFSyntaxException("Expected 'endstream'", tokeniser.getInputStream());
	
	std::tie(result->m_filteringSupported, result->m_filters) = createFilters(result, resolver);

	return result;
}

LemonPDF::StreamPtr LemonPDF::Stream::FromBuffer(const std::vector<unsigned char> &buffer, const std::vector<std::string> &filterNames)
{
	InputVectorStreamBuffer inputBuffer(buffer);
	return FromIOStream(std::istream(&inputBuffer), filterNames);
}

LemonPDF::StreamPtr LemonPDF::Stream::FromIOStream(std::istream &&in, const std::vector<std::string> &filterNames)
{
	if (!in)
		throw PDFInvalidOperationException("Cannot create PDF stream from invalid iostream");

	StreamPtr stream = std::make_shared<Stream>();
	stream->m_streamDict = std::make_shared<Dictionary>();
	stream->setFilterNames(filterNames);
	
	OutputVectorStreamBuffer buf(stream->m_data);
	std::ostream output(&buf);

	if (stream->m_filters.size() == 0)
	{
		output << in.rdbuf();
	}
	else if (stream->m_filters.size() == 1)
	{
		std::stringstream filtered;
		stream->m_filters.front()->encode(in, filtered);
		output << filtered.rdbuf();
	}
	else
	{
		std::stringstream nextInput;
		std::stringstream nextOutput;

		stream->m_filters.front()->encode(in, nextInput);

		for (size_t i = 1; i < stream->m_filters.size() - 1; i++)
		{
			stream->m_filters.at(i)->encode(nextInput, nextOutput);
			std::swap(nextInput, nextOutput);
			nextOutput.str("");
		}

		stream->m_filters.back()->encode(nextInput, output);
	}
	
	return stream;
}

LemonPDF::Stream::Stream()
  : m_filteringSupported(true), m_streamDict(std::make_shared<Dictionary>())
{
}

LemonPDF::Stream::~Stream()
{
}

LemonPDF::Stream::operator std::string() const
{
	std::ostringstream ss;
	ss << "<Stream, ";
	ss << m_data.size();
	ss << " bytes>";
	return ss.str();	
}

LemonPDF::ObjectPtr LemonPDF::Stream::copy() const
{
	StreamPtr result = std::make_shared<Stream>();

	result->m_filteringSupported = m_filteringSupported;
	result->m_streamDict = std::dynamic_pointer_cast<Dictionary>(m_streamDict->copy());
	result->m_data = m_data;
	result->m_filters = m_filters;

	return result;
}

void LemonPDF::Stream::filter(std::ostream &out) const
{
	try
	{
		if (m_filters.size() == 0)
		{
			if (m_data.size() > 0)
				out.write(reinterpret_cast<const char *>(&m_data[0]), m_data.size());
		}
		else if (m_filters.size() == 1)
		{
			InputVectorStreamBuffer buf(m_data);
			std::istream in(&buf);
			m_filters.front()->decode(in, out);
		}
		else
		{
			InputVectorStreamBuffer buf(m_data);
			std::istream firstInput(&buf);

			std::stringstream nextInput;
			std::stringstream nextOutput;
			
			// prime the filter pipeline with firstInput
			m_filters.front()->decode(firstInput, nextInput);

			for (size_t i = 1; i < m_filters.size() - 1; i++)
			{
				m_filters.at(i)->decode(nextInput, nextOutput);
				std::swap(nextInput, nextOutput);
				nextOutput.str("");
			}

			m_filters.back()->decode(nextInput, out);
		}
	}
	catch (LemonSqueezy::IOException &)
	{
		StripedHorse::Logging::Warning() << "Stream ended prematurely" << std::endl;
	}
}

void LemonPDF::Stream::filterIn(std::istream &in)
{
	m_data.clear();

	try
	{
		if (m_filters.size() == 0)
		{
			while (in)
			{
				unsigned char buffer[1024] = {};
				in.read(reinterpret_cast<char *>(buffer), sizeof buffer);
				if (in.gcount() > 0)
					m_data.insert(m_data.end(), buffer, buffer + in.gcount());
			}
		}
		else if (m_filters.size() == 1)
		{
			OutputVectorStreamBuffer buf(m_data);
			std::ostream out(&buf);
			m_filters.front()->encode(in, out);
		}
		else
		{
			// We have to encode the data in reverse (e.g. start
			// at the last filter and work our way to the first)
			OutputVectorStreamBuffer buf(m_data);
			std::ostream lastOutput(&buf);

			std::stringstream nextInput;
			std::stringstream nextOutput;
			
			// prime the filter pipeline with the input stream
			m_filters.back()->encode(in, nextInput);

			// start at the second to last filter (index = size - 2) and
			// continue until the second filter (index 1). If size is 3,
			// then the second to last and second filter are the same,
			// and if size is 2, this loop is unused
			for (size_t i = m_filters.size() - 2; i >= 1; i--)
			{
				m_filters.at(i)->encode(nextInput, nextOutput);
				std::swap(nextInput, nextOutput);
				nextOutput.str("");
			}

			// send the nextInput into the lastOutput
			m_filters.front()->encode(nextInput, lastOutput);
		}
	}
	catch (LemonSqueezy::IOException &)
	{
		StripedHorse::Logging::Warning() << "Stream ended prematurely" << std::endl;
	}
}

bool LemonPDF::Stream::filteringSupported() const
{
	return m_filteringSupported;
}

const LemonPDF::DictionaryPtr LemonPDF::Stream::getDictionary() const
{
	return m_streamDict;
}

const std::vector<unsigned char> &LemonPDF::Stream::getData() const
{
	return m_data;
}

std::vector<LemonPDF::IFilterPtr> LemonPDF::Stream::getFilters() const
{
	return m_filters;
}

std::vector<std::string> LemonPDF::Stream::getFilterNames() const
{
	std::vector<std::string> result;
	result.reserve(m_filters.size());
	for (auto f : m_filters)
		result.emplace_back(f->getName());
	return result;
}

std::vector<unsigned char> LemonPDF::Stream::getFilteredData() const
{
	std::vector<unsigned char> data;
	OutputVectorStreamBuffer buffer(data);
	std::ostream stream(&buffer);
	filter(stream);
	return data;
}

const std::string &LemonPDF::Stream::getType() const
{
	LiteralPtr typeName = m_streamDict->get<Literal>("Type");
	if (typeName == nullptr)
		throw PDFInvalidOperationException("Stream dictionary does not have a type");
	else
		return typeName->getNameValue();
}

void LemonPDF::Stream::setData(const std::vector<unsigned char> &data, const std::vector<IFilterPtr> &filters)
{
	m_filteringSupported = true;
	m_data = data;
	m_streamDict->add("Length", Literal::FromInteger(data.size()));
	m_filters = filters;
}

void LemonPDF::Stream::setDictionary(DictionaryPtr newDict)
{
	m_streamDict = newDict;
	m_streamDict->add("Length", Literal::FromInteger(m_data.size()));
}

void LemonPDF::Stream::setDictionaryValue(const std::string &name, ObjectPtr value)
{
	setDictionaryValue(Literal::FromName(name), value);
}

void LemonPDF::Stream::setDictionaryValue(const LiteralPtr name, ObjectPtr value)
{
	m_streamDict->add(name, value);
}

void LemonPDF::Stream::setFilterNames(const std::vector<std::string> &filterNames)
{
	m_filters.clear();
	
	if (filterNames.size() == 1)
	{
		// If this doesn't exist, decodeParms will be nullptr (which is OK)
		DictionaryPtr decodeParms = m_streamDict->get<Dictionary>("DecodeParms");
		m_filters.emplace_back(getFilter(filterNames.front(), decodeParms, nullptr));
	}
	else if (filterNames.size() > 1)
	{
		ArrayPtr decodeParmsArray = m_streamDict->get<Array>("DecodeParms");
		
		for (size_t i = 0; i < filterNames.size(); i++)
		{
			DictionaryPtr currentDecodeParms = nullptr;
			if (decodeParmsArray != nullptr)
				currentDecodeParms = decodeParmsArray->get<Dictionary>(i);
			m_filters.emplace_back(getFilter(filterNames[i], currentDecodeParms, nullptr));
		}
	}
}

void LemonPDF::Stream::setFilters(const std::vector<IFilterPtr> &filters)
{
	m_filters = filters;
}

void LemonPDF::Stream::setType(const std::string &newType)
{
	setDictionaryValue("Type", Literal::FromName(newType));
}

void LemonPDF::Stream::write(std::ostream &out, size_t depth) const
{
	if (depth > 1)
		throw PDFInvalidOperationException("Stream objects can only be at the root level");
	
	DictionaryPtr newStreamDict = m_streamDict->shallowCopy();
	newStreamDict->add("Length", Literal::FromInteger(m_data.size()));
	
	if (m_filters.size() == 0)
	{
		newStreamDict->remove("Filter");
	}
	else if (m_filters.size() == 1)
	{
		newStreamDict->add("Filter", Literal::FromName(m_filters[0]->getName()));
	}
	else
	{
		ArrayPtr filterArray = std::make_shared<Array>();
		for (const auto &filter : m_filters)
			filterArray->add(Literal::FromName(filter->getName()));
		newStreamDict->add("Filter", filterArray);
	}
	
	newStreamDict->write(out, depth);
	
	if (depth > 0)
		out << "\n";
	out << "stream\n";
	out.write(reinterpret_cast<const char *>(&m_data[0]), m_data.size());
	out << "\nendstream";
}
