#include <cstdlib>
#include <sstream>
#include <tuple>
#include "libtokenisation.h"
#include "pdfexception.h"
#include "pdftokeniser.h"

namespace
{
	std::tuple<LemonPDF::Token::TokenType, std::string> parseKeyword(std::istream &input)
	{
		using LemonPDF::Token;
		std::string keyword;
		char ch;
		
		while (input.get(ch))
		{
			if (!::isalpha(ch))
			{
				input.putback(ch);
				break;
			}

			keyword.push_back(ch);
		}
		
		if (keyword == "true")
			return std::make_tuple(Token::kTrue, keyword);
		else if (keyword == "false")
			return std::make_tuple(Token::kFalse, keyword);
		else if (keyword == "endobj")
			return std::make_tuple(Token::kObjectEnd, keyword);
		else if (keyword == "stream")
			return std::make_tuple(Token::kStreamStart, keyword);
		else if (keyword == "endstream")
			return std::make_tuple(Token::kStreamEnd, keyword);
		else if (keyword == "trailer")
			return std::make_tuple(Token::kTrailer, keyword);
		else if (keyword == "xref")
			return std::make_tuple(Token::kXref, keyword);
		else if (keyword == "startxref")
			return std::make_tuple(Token::kStartXref, keyword);
		else if (keyword == "null")
			return std::make_tuple(Token::kNull, keyword);
		else if (keyword == "obj")
			return std::make_tuple(Token::kObjectStart, keyword);
		else if (keyword == "R")
			return std::make_tuple(Token::kReference, keyword);
		
		throw LemonPDF::PDFSyntaxException("Unknown keyword", input);
	}
	
	void tokenise(std::istream &stream, std::deque<LemonPDF::Token> &tokens)
	{
		using namespace LemonPDF::Tokenisation;
		using LemonPDF::Token;
		
		while (1)
		{
			size_t offset = stream.tellg();
			char ch;
			if (!stream.get(ch))
			{
				tokens.emplace_back(offset, Token::kEndOfFile);
				break;
			}
			
			switch (ch)
			{
				case '<':
				{
					char other;
					if (!stream.get(other))
					{
						tokens.emplace_back(offset, Token::kEndOfFile);
						break;
					}
					
					if (other == '<')
					{
						tokens.emplace_back(offset, Token::kDictStart);
						break;
					}
					else
					{
						stream.putback(other);
						tokens.emplace_back(offset, Token::kString, TokeniseHexString(stream));
						break;
					}
				}
				
				case '>':
				{
					char other;
					if (!stream.get(other))
					{
						tokens.emplace_back(offset, Token::kEndOfFile);
						break;
					}
					
					if (other == '>')
					{
						tokens.emplace_back(offset, Token::kDictEnd);
						break;
					}
					
					throw LemonPDF::PDFSyntaxException("Unexpected token", stream);
				}
				
				case '+':
				case '-':
				case '.':
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
				{
					stream.putback(ch);
					std::pair<bool, std::string> t = TokeniseNumber(stream);
					Token::TokenType tt = t.first ? Token::kInteger : Token::kReal;
					tokens.emplace_back(offset, tt, t.second);
					if (tt == Token::kInteger && tokens.size() < 3)
						continue;
					break;
				}
				
				case '/':
				{
					stream.putback(ch);
					tokens.emplace_back(offset, Token::kName, TokeniseName(stream));
					break;
				}
				
				case '(':
				{
					tokens.emplace_back(offset, Token::kString, TokeniseString(stream));
					break;
				}
				
				case '[':
				{
					tokens.emplace_back(offset, Token::kArrayStart);
					break;
				}
				
				case ']':
				{
					tokens.emplace_back(offset, Token::kArrayEnd);
					break;
				}
				
				case '%':
				{
					std::string line;
					if (!ReadLine(stream, line))
						throw LemonPDF::PDFIOException("Read error");
					tokens.emplace_back(offset, Token::kComment, line);
					break;
				}
				
				case ' ':
				case '\r':
				case '\n':
				case '\t':
				{
					continue;
				}
				
				/**
				 * try parsing a keyword
				 */
				default:
				{
					stream.putback(ch);

					Token::TokenType t;
					std::string keyword;
					std::tie(t, keyword) = parseKeyword(stream);

					tokens.emplace_back(offset, t, keyword);
					
					if (t == Token::kReference || t == Token::kObjectStart)
					{
						Token idToken = *(tokens.rbegin() + 2);
						Token genToken = *(tokens.rbegin() + 1);
						
						if (idToken.getType() != Token::kInteger)
							throw LemonPDF::PDFSyntaxException("Invalid ID in reference", stream);
						if (genToken.getType() != Token::kInteger)
							throw LemonPDF::PDFSyntaxException("Invalid generation in reference", stream);
						
						tokens.pop_back();
						tokens.pop_back();
						tokens.pop_back();
						
						tokens.emplace_back(idToken.getOffset(), t, idToken.getText() + " " + genToken.getText());
					}
				}
			}
			
			break;
		}
	}
}

LemonPDF::Token::Token(size_t offset, enum TokenType type, const std::string &text)
  : m_offset(offset), m_type(type), m_text(text)
{
}

std::pair<size_t, size_t> LemonPDF::Token::getObjectIDAndGeneration() const
{
	if (m_type != kReference && m_type != kObjectStart)
		throw PDFInvalidOperationException("Token is not an object definition or reference");
	
	std::pair<size_t, size_t> result;
	std::istringstream ss(m_text);
	if (!(ss >> result.first >> result.second))
		throw PDFInvalidOperationException("Token text is not parsable as an object definition or reference");
	
	return result;
}

size_t LemonPDF::Token::getOffset() const
{
	return m_offset;
}

LemonPDF::Token::TokenType LemonPDF::Token::getType() const
{
	return m_type;
}

const std::string &LemonPDF::Token::getText() const
{
	return m_text;
}

LemonPDF::Tokeniser::Tokeniser(std::istream &input)
  : m_stream(input)
{
	
}

void LemonPDF::Tokeniser::consume()
{
	m_tokens.pop_front();
}

std::istream &LemonPDF::Tokeniser::getInputStream()
{
	return m_stream;
}

LemonPDF::Token LemonPDF::Tokeniser::getNextToken()
{
	if (shouldTokeniseMore())
		tokenise(m_stream, m_tokens);
	
	Token result = m_tokens.front();
	m_tokens.pop_front();
	return result;
}

LemonPDF::Token LemonPDF::Tokeniser::peekNextToken()
{
	if (shouldTokeniseMore())
		tokenise(m_stream, m_tokens);
	
	return m_tokens.front();
}

std::vector<unsigned char> LemonPDF::Tokeniser::read(size_t count)
{
	std::vector<unsigned char> buffer(count);
	
	char ch;
	while (m_stream.get(ch))
		if (ch == '\n')
			break;
	
	if (!m_stream.read(reinterpret_cast<char *>(&buffer[0]), count))
		throw PDFIOException("Read error");
	
	return buffer;
}

bool LemonPDF::Tokeniser::shouldTokeniseMore() const
{
	if (m_tokens.size() == 0)
		return true;
	
	if (m_tokens.size() == 1 && m_tokens.at(0).getType() == Token::TokenType::kInteger)
		return true;
	
	if (m_tokens.size() == 2 && m_tokens.at(0).getType() == Token::TokenType::kInteger && m_tokens.at(1).getType() == Token::TokenType::kInteger)
		return true;
	
	return false;
}
