#include <regex>
#include "libtokenisation.h"
#include "pjlreader.h"

namespace
{
	struct PJLSetting
	{
		PJLSetting(const std::string &n, const std::string &v)
		  : name(n), value(v)
		{
		}

		std::string name;
		std::string value;
	};

	struct PJLCommand
	{
		std::string name;
		std::vector<PJLSetting> settings;
	};

	bool parseSetting(const std::string &str, PJLCommand &command)
	{
		static const std::regex settingRegex { R"((([IL]PARM)\s*:\s*([A-Za-z]\w*)\s*)?([A-Za-z]\w*)\s*=\s*(\S+|("[^"]*?")))" };
		
		std::sregex_iterator start = std::sregex_iterator(str.begin(), str.end(), settingRegex);
		std::sregex_iterator end;

		for (auto i = start; i != end; i++)
			command.settings.emplace_back(i->str(4), i->str(5));
		
		return true;
	}

	bool parseToshibaCommand(const std::string &str, PJLCommand &cmd)
	{
		std::string::size_type equals = str.find('=');
		if (equals == std::string::npos)
			return false;
		
		cmd.settings.emplace_back(str.substr(0, equals), str.substr(equals + 1));
		return true;
	}

	bool parsePJLLine(const std::string &line, PJLCommand &command)
	{
		// PJL Reference manual describes the 4 formats of PJL commands. Format 1 is just the UEL (\x1B%-12345X)
		static const std::regex format2 { R"(^@PJL\s*$)" };
		static const std::regex format3 { R"(^@PJL\s+(COMMENT|ECHO)\s+(.+)$)" };
		static const std::regex format4 { R"(^@PJL\s+(JOB|EOJ|ENTER|DEFAULT|SET|INITIALIZE|RESET|D?INQUIRE|INFO|USTATUS(OFF)?|(RDY|OP|ST)MSG|FS(APPEND|DELETE|DIRLIST|DOWNLOAD|INIT|MKDIR|QUERY|UPLOAD))(\s+(.+))$)" };

		std::smatch result;
		if (std::regex_match(line, result, format2))
		{
			// no-op
			return true;
		}
		else if (std::regex_match(line, result, format3))
		{
			if (result.str(2).substr(0, 10) == "DSSC PRINT")
			{
				command.name = "TOSHIBA";
				return parseToshibaCommand(result.str(2).substr(11), command);
			}
			
			return true;
		}
		else if (std::regex_match(line, result, format4))
		{
			command.name = result.str(1);
			return parseSetting(result.str(6), command);
		}

		return false;
	}

	bool readUntilMarker(std::istream &in, const std::string &marker, std::ostream *out, std::streamsize limit)
	{
		std::streamsize bytesRead = 0;
		std::string::size_type currentMatch = 0;
		int ch;
		
		if (marker.size() == 0)
			return false;

		while (in && (limit == 0 || bytesRead < limit))
		{
			ch = in.get();
			if (ch == std::istream::traits_type::eof())
				break;
			
			bytesRead++;
			
			if (ch == marker[currentMatch])
			{
				currentMatch++;
			}
			else
			{
				if (out != nullptr)
				{
					if (currentMatch > 0)
						out->write(&marker[0], currentMatch);
					out->put(ch);
				}

				currentMatch = 0;
			}
			
			if (currentMatch == marker.size())
				return true;
		}

		return false;
	}

	const char g_pjlEscapeMarker[] = "\x1B%-12345X";
}

LemonPDF::PJLReader::PJLReader(std::istream &in)
  : m_input(in), m_printDataOffset(0)
{
	if (!readUntilMarker(in, g_pjlEscapeMarker, nullptr, 4096))
	{
		in.clear();
		in.seekg(0, std::ios::beg);
		return;
	}

	bool stop = false;
	std::string line;
	while (!stop && Tokenisation::ReadLine(m_input, line))
	{
		PJLCommand cmd;
		std::streampos restorePoint = m_input.tellg();
		if (parsePJLLine(line, cmd))
		{
			auto iter = m_settings.emplace(std::make_pair(cmd.name, std::unordered_map<std::string, std::string> {})).first;
			for (auto setting : cmd.settings)
			{
				(iter->second)[setting.name] = setting.value;
				if (cmd.name == "ENTER" && setting.name == "LANGUAGE")
					stop = true;
			}
		}
		else
		{
			m_input.seekg(restorePoint, std::ios::beg);
			stop = true;
		}
	}

	m_printDataOffset = in.tellg();
}

bool LemonPDF::PJLReader::extractPrintData(std::ostream &out)
{
	return readUntilMarker(m_input, g_pjlEscapeMarker, &out, 0);
}

std::streampos LemonPDF::PJLReader::getPrintDataOffset() const
{
	return m_printDataOffset;
}
