#include <cctype>
#include <cstring>
#include "ascii85.h"
#include "exceptions.h"

namespace
{
	void decodeASCII85(uint32_t val, size_t count, std::ostream &out)
	{
		if (count == 0)
			return;
		
		if (count < 2)
			throw LemonSqueezy::DecompressionException("Buffer too short to decode");

		for (size_t i = 0; i < count - 1; i++)
			out.put(static_cast<char>(val >> ((3 - i) * 8)));
	}
}

LemonSqueezy::ASCII85Encoding::ASCII85Encoding(size_t lineLimit, size_t initialCharPosition)
  : LineLengthLimitedEncoding(lineLimit, initialCharPosition)
{
}

void LemonSqueezy::ASCII85Encoding::read(std::istream &in, std::ostream &out)
{
	size_t count = 0;
	const static uint32_t placeValues[5] = { 85 * 85 * 85 * 85, 85 * 85 * 85, 85 * 85, 85, 1 };
	uint32_t value = 0;
	char ch;
	bool trailerFound = false;

	// empty stream
	if (!in.get(ch))
		return;
	
	// header is optional
	if (ch == '<')
	{
		if (!in.get(ch))
			throw DecompressionException("EOF encountered early in ASCII85 stream");
		
		if (ch != '~')
		{
			if (ch < 33 || ch > 117)
				throw DecompressionException("Invalid token in ASCII85 stream");
			
			value += ('<' - 33) * placeValues[count++];
			value += (ch - 33) * placeValues[count++];
		}
	}
	else
	{
		in.unget();
	}

	while (in.get(ch))
	{
		if (::isspace(ch))
		{
			continue;
		}
		else if (ch == 'z' && count == 0)
		{
			out.write("\0\0\0", 4);
			continue;
		}
		else if (ch == '~')
		{
			if (!in.get(ch) || ch != '>')
				throw DecompressionException("Invalid token in ASCII85 stream");
			
			for (size_t i = count; i < 5; i++)
				value += (117 - 33) * placeValues[i];
			
			trailerFound = true;
		}
		else if (ch < 33 || ch > 117)
		{
			throw DecompressionException("Invalid token in ASCII85 stream");
		}
		else
		{
			value += (ch - 33) * placeValues[count++];
			if (count < 5)
				continue;
		}
		
		decodeASCII85(value, count, out);
		value = 0;
		count = 0;
	}

	if (in.eof() && !trailerFound)
		throw DecompressionException("ASCII85 stream ended prematurely");
}

void LemonSqueezy::ASCII85Encoding::write(std::istream &in, std::ostream &out)
{
	put('<', out);
	put('~', out);

	unsigned char inBuffer[4];
	while (true)
	{
		in.read(reinterpret_cast<char *>(inBuffer), sizeof inBuffer);
		size_t bytes = in.gcount();
		
		if (bytes == 0)
			break;

		for (size_t i = bytes; i < 4; i++)
			inBuffer[i] = 0;

		uint32_t val = (inBuffer[0] << 24U) | (inBuffer[1] << 16U) | (inBuffer[2] << 8U) | (inBuffer[3]);

		if (val == 0 && bytes == 4)
		{
			put('z', out);
		}
		else
		{
			char outBuffer[5] = { '\0' };
			for (size_t i = 0; i < sizeof outBuffer; i++)
			{
				outBuffer[i] = (val % 85) + 33;
				val /= 85;
			}

			// output is always one more byte than input
			for (size_t i = 0; i < bytes + 1; i++)
				put(outBuffer[4 - i], out);
		}
	}

	put('~', out);
	put('>', out);
}
