#include <algorithm>
#include <cstdint>
#include <iostream>
#include <unordered_map>
#include <vector>
#include "bitsequence.h"
#include "ccitt.h"
#include "exceptions.h"
#include "streambitreader.h"
#include "streambitwriter.h"

namespace
{
	using CCITTCodeword = LemonSqueezy::BitSequence;

	class CCITTCodewordHash
	{
	public:
		uint32_t operator()(const CCITTCodeword &codeword) const
		{
			// codeword.code is 32-bits in length, but the largest CCITT code is 24 bits
			// the value "24" itself needs only 5 bits, so the length can be pushed
			// to the most significant 5 bits
			return (codeword.length << 26) | codeword.code;
		}
	};

	enum class CCITT2DCodingMode : int
	{
		kVerticalL1 = -1,
		kVerticalL2 = -2,
		kVerticalL3 = -3,
		kVertical0 = 0,
		kVerticalR1 = 1,
		kVerticalR2 = 2,
		kVerticalR3 = 3,
		kPass,
		kHorizontal,
		kEndOfFacsimileBlock
	};

	const std::unordered_map<CCITTCodeword, CCITT2DCodingMode, CCITTCodewordHash> g_codingModes {
		{ {                     0b0001,  4 }, CCITT2DCodingMode::kPass },
		{ {                      0b001,  3 }, CCITT2DCodingMode::kHorizontal },
		{ {                        0b1,  1 }, CCITT2DCodingMode::kVertical0 },
		{ {                      0b011,  3 }, CCITT2DCodingMode::kVerticalR1 },
		{ {                   0b000011,  6 }, CCITT2DCodingMode::kVerticalR2 },
		{ {                  0b0000011,  7 }, CCITT2DCodingMode::kVerticalR3 },
		{ {                      0b010,  3 }, CCITT2DCodingMode::kVerticalL1 },
		{ {                   0b000010,  6 }, CCITT2DCodingMode::kVerticalL2 },
		{ {                  0b0000010,  7 }, CCITT2DCodingMode::kVerticalL3 },
		{ { 0b000000000001000000000001, 24 }, CCITT2DCodingMode::kEndOfFacsimileBlock }
	};

	using RunLengthDecodingTable = std::unordered_map<CCITTCodeword, uint16_t, CCITTCodewordHash>;
	using RunLengthEncodingTable = std::unordered_map<uint16_t, CCITTCodeword>;

	RunLengthEncodingTable reverseTable(const RunLengthDecodingTable &decodeTable)
	{
		RunLengthEncodingTable result;
		for (auto kv : decodeTable)
			result[kv.second] = kv.first;
		return result;
	}

	const RunLengthDecodingTable g_whiteRunLengths {
		{ { 0b00110101, 8 }, 0  },
		{ {   0b000111, 6 }, 1  },
		{ {     0b0111, 4 }, 2  },
		{ {     0b1000, 4 }, 3  },
		{ {     0b1011, 4 }, 4  },
		{ {     0b1100, 4 }, 5  },
		{ {     0b1110, 4 }, 6  },
		{ {     0b1111, 4 }, 7  },
		{ {    0b10011, 5 }, 8  },
		{ {    0b10100, 5 }, 9  },
		{ {    0b00111, 5 }, 10 },
		{ {    0b01000, 5 }, 11 },
		{ {   0b001000, 6 }, 12 },
		{ {   0b000011, 6 }, 13 },
		{ {   0b110100, 6 }, 14 },
		{ {   0b110101, 6 }, 15 },
		{ {   0b101010, 6 }, 16 },
		{ {   0b101011, 6 }, 17 },
		{ {  0b0100111, 7 }, 18 },
		{ {  0b0001100, 7 }, 19 },
		{ {  0b0001000, 7 }, 20 },
		{ {  0b0010111, 7 }, 21 },
		{ {  0b0000011, 7 }, 22 },
		{ {  0b0000100, 7 }, 23 },
		{ {  0b0101000, 7 }, 24 },
		{ {  0b0101011, 7 }, 25 },
		{ {  0b0010011, 7 }, 26 },
		{ {  0b0100100, 7 }, 27 },
		{ {  0b0011000, 7 }, 28 },
		{ { 0b00000010, 8 }, 29 },
		{ { 0b00000011, 8 }, 30 },
		{ { 0b00011010, 8 }, 31 },
		{ { 0b00011011, 8 }, 32 },
		{ { 0b00010010, 8 }, 33 },
		{ { 0b00010011, 8 }, 34 },
		{ { 0b00010100, 8 }, 35 },
		{ { 0b00010101, 8 }, 36 },
		{ { 0b00010110, 8 }, 37 },
		{ { 0b00010111, 8 }, 38 },
		{ { 0b00101000, 8 }, 39 },
		{ { 0b00101001, 8 }, 40 },
		{ { 0b00101010, 8 }, 41 },
		{ { 0b00101011, 8 }, 42 },
		{ { 0b00101100, 8 }, 43 },
		{ { 0b00101101, 8 }, 44 },
		{ { 0b00000100, 8 }, 45 },
		{ { 0b00000101, 8 }, 46 },
		{ { 0b00001010, 8 }, 47 },
		{ { 0b00001011, 8 }, 48 },
		{ { 0b01010010, 8 }, 49 },
		{ { 0b01010011, 8 }, 50 },
		{ { 0b01010100, 8 }, 51 },
		{ { 0b01010101, 8 }, 52 },
		{ { 0b00100100, 8 }, 53 },
		{ { 0b00100101, 8 }, 54 },
		{ { 0b01011000, 8 }, 55 },
		{ { 0b01011001, 8 }, 56 },
		{ { 0b01011010, 8 }, 57 },
		{ { 0b01011011, 8 }, 58 },
		{ { 0b01001010, 8 }, 59 },
		{ { 0b01001011, 8 }, 60 },
		{ { 0b00110010, 8 }, 61 },
		{ { 0b00110011, 8 }, 62 },
		{ { 0b00110100, 8 }, 63 },

		{ {     0b11011, 5 },  64  },
		{ {     0b10010, 5 },  128 },
		{ {    0b010111, 6 },  192 },
		{ {   0b0110111, 7 },  256 },
		{ {  0b00110110, 8 },  320 },
		{ {  0b00110111, 8 },  384 },
		{ {  0b01100100, 8 },  448 },
		{ {  0b01100101, 8 },  512 },
		{ {  0b01101000, 8 },  576 },
		{ {  0b01100111, 8 },  640 },
		{ { 0b011001100, 9 },  704 },
		{ { 0b011001101, 9 },  768 },
		{ { 0b011010010, 9 },  832 },
		{ { 0b011010011, 9 },  896 },
		{ { 0b011010100, 9 },  960 },
		{ { 0b011010101, 9 }, 1024 },
		{ { 0b011010110, 9 }, 1088 },
		{ { 0b011010111, 9 }, 1152 },
		{ { 0b011011000, 9 }, 1216 },
		{ { 0b011011001, 9 }, 1280 },
		{ { 0b011011010, 9 }, 1344 },
		{ { 0b011011011, 9 }, 1408 },
		{ { 0b010011000, 9 }, 1472 },
		{ { 0b010011001, 9 }, 1536 },
		{ { 0b010011010, 9 }, 1600 },
		{ {    0b011000, 6 }, 1664 },
		{ { 0b010011011, 9 }, 1728 },

		{ {  0b00000001000, 11}, 1792 },
		{ {  0b00000001100, 11}, 1856 },
		{ {  0b00000001101, 11}, 1920 },
		{ { 0b000000010010, 12}, 1984 },
		{ { 0b000000010011, 12}, 2048 },
		{ { 0b000000010100, 12}, 2112 },
		{ { 0b000000010101, 12}, 2176 },
		{ { 0b000000010110, 12}, 2240 },
		{ { 0b000000010111, 12}, 2304 },
		{ { 0b000000011100, 12}, 2368 },
		{ { 0b000000011101, 12}, 2432 },
		{ { 0b000000011110, 12}, 2496 },
		{ { 0b000000011111, 12}, 2560 }
	};

	const RunLengthDecodingTable g_blackRunLengths {
		{ {   0b0000110111, 10 }, 0  },
		{ {          0b010,  3 }, 1  },
		{ {           0b11,  2 }, 2  },
		{ {           0b10,  2 }, 3  },
		{ {          0b011,  3 }, 4  },
		{ {         0b0011,  4 }, 5  },
		{ {         0b0010,  4 }, 6  },
		{ {        0b00011,  5 }, 7  },
		{ {       0b000101,  6 }, 8  },
		{ {       0b000100,  6 }, 9  },
		{ {      0b0000100,  7 }, 10 },
		{ {      0b0000101,  7 }, 11 },
		{ {      0b0000111,  7 }, 12 },
		{ {     0b00000100,  8 }, 13 },
		{ {     0b00000111,  8 }, 14 },
		{ {    0b000011000,  9 }, 15 },
		{ {   0b0000010111, 10 }, 16 },
		{ {   0b0000011000, 10 }, 17 },
		{ {   0b0000001000, 10 }, 18 },
		{ {  0b00001100111, 11 }, 19 },
		{ {  0b00001101000, 11 }, 20 },
		{ {  0b00001101100, 11 }, 21 },
		{ {  0b00000110111, 11 }, 22 },
		{ {  0b00000101000, 11 }, 23 },
		{ {  0b00000010111, 11 }, 24 },
		{ {  0b00000011000, 11 }, 25 },
		{ { 0b000011001010, 12 }, 26 },
		{ { 0b000011001011, 12 }, 27 },
		{ { 0b000011001100, 12 }, 28 },
		{ { 0b000011001101, 12 }, 29 },
		{ { 0b000001101000, 12 }, 30 },
		{ { 0b000001101001, 12 }, 31 },
		{ { 0b000001101010, 12 }, 32 },
		{ { 0b000001101011, 12 }, 33 },
		{ { 0b000011010010, 12 }, 34 },
		{ { 0b000011010011, 12 }, 35 },
		{ { 0b000011010100, 12 }, 36 },
		{ { 0b000011010101, 12 }, 37 },
		{ { 0b000011010110, 12 }, 38 },
		{ { 0b000011010111, 12 }, 39 },
		{ { 0b000001101100, 12 }, 40 },
		{ { 0b000001101101, 12 }, 41 },
		{ { 0b000011011010, 12 }, 42 },
		{ { 0b000011011011, 12 }, 43 },
		{ { 0b000001010100, 12 }, 44 },
		{ { 0b000001010101, 12 }, 45 },
		{ { 0b000001010110, 12 }, 46 },
		{ { 0b000001010111, 12 }, 47 },
		{ { 0b000001100100, 12 }, 48 },
		{ { 0b000001100101, 12 }, 49 },
		{ { 0b000001010010, 12 }, 50 },
		{ { 0b000001010011, 12 }, 51 },
		{ { 0b000000100100, 12 }, 52 },
		{ { 0b000000110111, 12 }, 53 },
		{ { 0b000000111000, 12 }, 54 },
		{ { 0b000000100111, 12 }, 55 },
		{ { 0b000000101000, 12 }, 56 },
		{ { 0b000001011000, 12 }, 57 },
		{ { 0b000001011001, 12 }, 58 },
		{ { 0b000000101011, 12 }, 59 },
		{ { 0b000000101100, 12 }, 60 },
		{ { 0b000001011010, 12 }, 61 },
		{ { 0b000001100110, 12 }, 62 },
		{ { 0b000001100111, 12 }, 63 },

		{ {    0b0000001111, 10 },  64  },
		{ {  0b000011001000, 12 },  128 },
		{ {  0b000011001001, 12 },  192 },
		{ {  0b000001011011, 12 },  256 },
		{ {  0b000000110011, 12 },  320 },
		{ {  0b000000110100, 12 },  384 },
		{ {  0b000000110101, 12 },  448 },
		{ { 0b0000001101100, 13 },  512 },
		{ { 0b0000001101101, 13 },  576 },
		{ { 0b0000001001010, 13 },  640 },
		{ { 0b0000001001011, 13 },  704 },
		{ { 0b0000001001100, 13 },  768 },
		{ { 0b0000001001101, 13 },  832 },
		{ { 0b0000001110010, 13 },  896 },
		{ { 0b0000001110011, 13 },  960 },
		{ { 0b0000001110100, 13 }, 1024 },
		{ { 0b0000001110101, 13 }, 1088 },
		{ { 0b0000001110110, 13 }, 1152 },
		{ { 0b0000001110111, 13 }, 1216 },
		{ { 0b0000001010010, 13 }, 1280 },
		{ { 0b0000001010011, 13 }, 1344 },
		{ { 0b0000001010100, 13 }, 1408 },
		{ { 0b0000001010101, 13 }, 1472 },
		{ { 0b0000001011010, 13 }, 1536 },
		{ { 0b0000001011011, 13 }, 1600 },
		{ { 0b0000001100100, 13 }, 1664 },
		{ { 0b0000001100101, 13 }, 1728 },

		{ {  0b00000001000, 11}, 1792 },
		{ {  0b00000001100, 11}, 1856 },
		{ {  0b00000001101, 11}, 1920 },
		{ { 0b000000010010, 12}, 1984 },
		{ { 0b000000010011, 12}, 2048 },
		{ { 0b000000010100, 12}, 2112 },
		{ { 0b000000010101, 12}, 2176 },
		{ { 0b000000010110, 12}, 2240 },
		{ { 0b000000010111, 12}, 2304 },
		{ { 0b000000011100, 12}, 2368 },
		{ { 0b000000011101, 12}, 2432 },
		{ { 0b000000011110, 12}, 2496 },
		{ { 0b000000011111, 12}, 2560 }
	};

	RunLengthEncodingTable g_whiteRunEncodings = reverseTable(g_whiteRunLengths);
	RunLengthEncodingTable g_blackRunEncodings = reverseTable(g_blackRunLengths);

	template<typename K, typename T, typename H>
	bool tryReadBitsUntilMatchInTable(LemonSqueezy::IBitReader &bitReader, const std::unordered_map<K, T, H> &table, T &out)
	{
		CCITTCodeword cw;
		while (bitReader.readBitsIntoBitSequence(1, cw))
		{
			typename std::unordered_map<K, T, H>::const_iterator it = table.find(cw);
			if (it == table.end())
				continue;

			out = it->second;
			return true;
		}
		
		throw LemonSqueezy::IOException("Unexpected end of stream");
	}

	template<typename K, typename T, typename H>
	unsigned readBitsUntilMatchInRunLengthTable(LemonSqueezy::IBitReader &bitReader, const std::unordered_map<K, T, H> &table)
	{
		unsigned currentRun = 0;
		uint16_t value = 0;
		
		while (tryReadBitsUntilMatchInTable(bitReader, table, value))
		{
			currentRun += value;
			if (value <= 63)
				return currentRun;
		}

		throw LemonSqueezy::IOException("Unexpected end of stream");
	}

	template<typename K, typename T>
	void writeBitSequenceForRunLength(size_t length, LemonSqueezy::IBitWriter &bitWriter, const std::unordered_map<K, T> &table)
	{
		// we require potentially multiple of these if the length is huge
		if (length > 2560)
		{
			auto bitSequence = table.find(2560);
			if (bitSequence == table.end())
				throw LemonSqueezy::CompressionException("Unable to find appropriate make-up code for 2560 run length");
				
			while (length > 2560)
			{
				bitWriter.writeBits(bitSequence->second);
				length -= 2560;
			}
		}

		// but we require at most one of these (when length is between 64 and 2560)
		if (length >= 64)
		{
			auto bitSequence = table.find(length - (length % 64));
			if (bitSequence == table.end())
				throw LemonSqueezy::CompressionException("Unable to find appropriate make-up code for 64 <= n < 2560 run length");
			
			bitWriter.writeBits(bitSequence->second);
			length -= bitSequence->first;
		}

		// and at most one of these (when length is less than 64)
		auto finalBitSequence = table.find(length);
		if (finalBitSequence == table.end())
			throw LemonSqueezy::CompressionException("Unable to find appropriate code for 0 <= n < 64 run length");
		bitWriter.writeBits(finalBitSequence->second);
	}

	inline int valueOfColumn(const unsigned char *data, int column)
	{
		return (data[column / 8] & (0x80 >> (column % 8))) > 0;
	}

	inline int findNextBitOfOppositeValue(const unsigned char *data, int start, int max, int current)
	{
		while (start < max && valueOfColumn(data, start) == current)
		{
			start++;

			// if we are on a byte boundary, we can check 8-bits at once
			if (start % 8 == 0)
				while (start + 8 < max && data[start / 8] == (current ? 0xFF : 0x00))
					start += 8;
		}
		
		return start;
	}

	inline int findNextBitOfSameValue(const unsigned char *data, int start, int max, int current)
	{
		while (start < max && valueOfColumn(data, start) != current)
		{
			start++;

			// if we are on a byte boundary, we can check 8-bits at once
			if (start % 8 == 0)
				while (start + 8 < max && data[start / 8] == (current ? 0x00 : 0xFF))
					start += 8;
		}
		
		return start;
	}
}

LemonSqueezy::CCITTCompression::DecodingLine::DecodingLine(size_t columns)
  : m_columns(columns), m_a0(0), m_penDown(false), m_penDownPosition(0)
{
}

LemonSqueezy::CCITTCompression::~CCITTCompression()
{
}

void LemonSqueezy::CCITTCompression::DecodingLine::decodeLineIntoBuffer(std::vector<unsigned char> &buffer) const
{
	for (const BlackRange &br : m_blackRanges)
	{
		for (size_t i = br.index; i < br.index + br.length; i++)
		{
			if (i > m_columns)
				break;
			
			size_t byteIndexOfBlock = i / 8;
			size_t bitIndex = i % 8;
			buffer.at(byteIndexOfBlock) |= (1 << (7 - bitIndex));
		}
	}
}

size_t LemonSqueezy::CCITTCompression::DecodingLine::getB1(const DecodingLine &referenceLine) const
{
	auto br = std::lower_bound(referenceLine.m_blackRanges.begin(), referenceLine.m_blackRanges.end(), m_a0, [=](const BlackRange &br, size_t index) {
		return (m_penDown ? (br.index + br.length) : br.index) < index;
	});

	if (br == referenceLine.m_blackRanges.end())
		return m_columns;

	size_t b1 = m_penDown ? (br->index + br->length) : br->index;
	
	// if b1 is calculated to be the same as a0, then we need to obtain b1 from
	// the next black range instead (but only if a0 has been set)
	if ((m_a0 != 0 || m_penDown) && b1 == m_a0)
	{
		std::advance(br, 1);
		if (br != referenceLine.m_blackRanges.end())
			b1 = m_penDown ? (br->index + br->length) : br->index;
		else
			b1 = m_columns;
	}

	return b1;
}

size_t LemonSqueezy::CCITTCompression::DecodingLine::getB2(const DecodingLine &referenceLine) const
{
	auto br = std::lower_bound(referenceLine.m_blackRanges.begin(), referenceLine.m_blackRanges.end(), getB1(referenceLine), [=](const BlackRange &br, size_t index) {
		return (m_penDown ? br.index : (br.index + br.length)) < index;
	});

	if (br != referenceLine.m_blackRanges.end())
		return m_penDown ? br->index : (br->index + br->length);
	
	return m_columns;
}
		
bool LemonSqueezy::CCITTCompression::DecodingLine::isAtEnd() const
{
	return m_a0 >= m_columns;
}

bool LemonSqueezy::CCITTCompression::DecodingLine::isPenDown() const
{
	return m_penDown;
}
		
void LemonSqueezy::CCITTCompression::DecodingLine::movePen(size_t p)
{
	m_a0 = p;
}

size_t LemonSqueezy::CCITTCompression::DecodingLine::offsetPen(size_t p)
{
	if (m_a0 + p > m_columns)
	{
		m_a0 = m_columns;
	}
	else
	{
		m_a0 += p;
	}
	return m_a0;
}

void LemonSqueezy::CCITTCompression::DecodingLine::togglePen()
{
	m_penDown = !m_penDown;
	if (m_penDown)
		m_penDownPosition = m_a0;
	else if (m_penDownPosition < m_a0)
		m_blackRanges.emplace_back(m_penDownPosition, m_a0 - m_penDownPosition);
}

LemonSqueezy::CCITTCompression::EncodingLine::EncodingLine(int columns)
  : m_columns(columns), m_a0(-1), m_data((m_columns + 7) / 8)
{

}

int LemonSqueezy::CCITTCompression::EncodingLine::getA0() const
{
	// a0 is the current coding position
	return m_a0;
}

int LemonSqueezy::CCITTCompression::EncodingLine::getA1() const
{
	// a1 is the next bit of opposite color to a0
	// a0 should never be set to m_columns when this function is called
	int current = (m_a0 < 0 ? 0 : valueOfColumn(m_data.data(), m_a0));
	return findNextBitOfOppositeValue(m_data.data(), m_a0 + 1, m_columns, current);
}

int LemonSqueezy::CCITTCompression::EncodingLine::getA2(int a1) const
{
	// a2 is the next bit of opposite color to a1
	if (a1 == m_columns)
		return m_columns;
	
	int current = valueOfColumn(m_data.data(), a1);
	return findNextBitOfOppositeValue(m_data.data(), a1 + 1, m_columns, current);
}

int LemonSqueezy::CCITTCompression::EncodingLine::getB1(const EncodingLine &referenceLine) const
{
	// b1 is the column of the next change INTO the the opposite color to a0 in the reference line
	// meaning if a0 is black, and the reference line is black in the same column
	int current = valueOfColumn(m_data.data(), m_a0);
	int firstBitSame = findNextBitOfSameValue(referenceLine.m_data.data(), m_a0, m_columns, current);
	if (firstBitSame == m_columns)
		return m_columns;
	return findNextBitOfOppositeValue(referenceLine.m_data.data(), firstBitSame + 1, m_columns, current);
}

int LemonSqueezy::CCITTCompression::EncodingLine::getB2(const EncodingLine &referenceLine, int b1) const
{
	int current = valueOfColumn(referenceLine.m_data.data(), b1);
	return findNextBitOfOppositeValue(referenceLine.m_data.data(), b1, m_columns, current);
}

LemonSqueezy::CCITTCompression::CCITTCompression(enum Encoding encoding, int errorLines, bool encodedByteAlign, size_t columns, size_t rows, bool endOfBlock, bool blackIs1)
  : m_encodedByteAlign(encodedByteAlign), m_columns(columns), m_rows(rows), m_endOfBlock(endOfBlock), m_blackIs1(blackIs1), m_referenceLine(columns), m_codingLine(columns)
{
	(void) errorLines; // used for CCITT Group 3 encoding
	(void) m_rows; // not used with this implementation

	if (encoding != Encoding::kGroup4_2D)
		throw UnimplementedException("Only CCITT Group 4 encoding is supported");
	
	if (m_columns < 1)
		throw DecompressionException("Invalid number of columns");
}

void LemonSqueezy::CCITTCompression::read(std::istream &in, std::ostream &out)
{
	size_t rowCount = 0;
	StreamBitReader bitReader(in);

	// reuse the same line buffer to avoid constant deallocation/reallocation
	std::vector<unsigned char> lineBuffer((m_columns - 1) / CHAR_BIT + 1);
			
	enum CCITT2DCodingMode codingMode;
	while (tryReadBitsUntilMatchInTable(bitReader, g_codingModes, codingMode))
	{
		switch (codingMode)
		{
			case CCITT2DCodingMode::kEndOfFacsimileBlock:
			{
				if (!m_endOfBlock)
					throw IOException("End of block unexpected");
				
				return;
			}

			case CCITT2DCodingMode::kPass:
			{
				size_t b2 = m_codingLine.getB2(m_referenceLine);
				m_codingLine.movePen(b2);
				break;
			}
			
			case CCITT2DCodingMode::kHorizontal:
			{
				bool penDown = m_codingLine.isPenDown();
				unsigned moveStep1 = readBitsUntilMatchInRunLengthTable(bitReader, penDown ? g_blackRunLengths : g_whiteRunLengths);
				unsigned moveStep2 = readBitsUntilMatchInRunLengthTable(bitReader, penDown ? g_whiteRunLengths : g_blackRunLengths);
				m_codingLine.offsetPen(moveStep1);
				m_codingLine.togglePen();
				m_codingLine.offsetPen(moveStep2);
				m_codingLine.togglePen();
				break;
			}
			
			case CCITT2DCodingMode::kVerticalL1:
			case CCITT2DCodingMode::kVerticalL2:
			case CCITT2DCodingMode::kVerticalL3:
			case CCITT2DCodingMode::kVertical0:
			case CCITT2DCodingMode::kVerticalR1:
			case CCITT2DCodingMode::kVerticalR2:
			case CCITT2DCodingMode::kVerticalR3:
			{
				size_t b1 = m_codingLine.getB1(m_referenceLine);
				int localOffset = static_cast<int>(codingMode);
				m_codingLine.movePen(b1 + localOffset);
				m_codingLine.togglePen();
				break;
			}
		}

		if (m_codingLine.isAtEnd())
		{
			m_referenceLine = m_codingLine;
			m_codingLine.decodeLineIntoBuffer(lineBuffer);
			m_codingLine = DecodingLine(m_columns);
			if (m_encodedByteAlign)
				bitReader.alignToByteBoundary();
			
			for (unsigned char &ch : lineBuffer)
			{
				out.put(static_cast<char>(m_blackIs1 ? ch : ~ch));
				ch = 0;
			}

			rowCount++;
		}
	}

	if (m_endOfBlock)
		throw IOException("Expected EOFB");
}

void LemonSqueezy::CCITTCompression::write(std::istream &in, std::ostream &out)
{
	size_t rowCount = 0;
	StreamBitWriter bitWriter(out);

	const static BitSequence verticalCodingSequences[]
	{
		{ 0b0000010, 7 },
		{ 0b000010, 6},
		{ 0b010, 3 },
		{ 0b1, 1 },
		{ 0b011, 3 },
		{ 0b000011, 6 },
		{ 0b0000011, 7 }
	};

	EncodingLine encodingLine(m_columns);
	EncodingLine referenceLine(m_columns);

	while (in.read(reinterpret_cast<char *>(encodingLine.getData().data()), encodingLine.getData().size()))
	{
		if (static_cast<size_t>(in.gcount()) != encodingLine.getData().size())
			throw IOException("Unexpected end of data");

		while (!encodingLine.isAtEnd())
		{
			int a1 = encodingLine.getA1();
			int b1 = encodingLine.getB1(referenceLine);
			int b2 = encodingLine.getB2(referenceLine, b1);

			int difference = a1 - b1;

			if (b2 < a1)
			{
				// pass mode
				bitWriter.writeBits({0b0001, 4});
				encodingLine.setA0(b2);
			}
			else if (difference >= -3 && difference <= 3)
			{
				// vertical mode
				bitWriter.writeBits(verticalCodingSequences[difference + 3]);
				encodingLine.setA0(a1);
			}
			else
			{
				// horizontal mode
				bitWriter.writeBits({0b001, 3});
				int a0 = encodingLine.getA0();
				if (a0 < 0)
					a0 = 0;
				
				int a2 = encodingLine.getA2(a1);
				int moveStep1 = a1 - a0;
				int moveStep2 = a2 - a1;

				writeBitSequenceForRunLength(moveStep1, bitWriter, encodingLine.isPenDown() ? g_blackRunEncodings : g_whiteRunEncodings);
				writeBitSequenceForRunLength(moveStep2, bitWriter, encodingLine.isPenDown() ? g_whiteRunEncodings : g_blackRunEncodings);

				encodingLine.setA0(a2);
			}
		}

		referenceLine = encodingLine;
		encodingLine = EncodingLine(m_columns);
		rowCount++;
	}

	if (m_endOfBlock)
		bitWriter.writeBits({0b000000000001000000000001, 24});
}
