#include <algorithm>
#include <array>
#include <climits>
#include <cstring>
#include <tuple>
#include <unordered_map>
#include <vector>
#include "exceptions.h"
#include "deflate.h"
#include "deflatestreambitreader.h"
#include "deflatestreambitwriter.h"
#include "rollingoutputbuffer.h"

#ifndef USE_STRIPED_HORSE_DEFLATE_COMPRESSION
#include <zlib.h>
#endif

namespace
{
	enum class BlockType
	{
		kNoCompression =  0b00,
		kFixedHuffman =   0b01,
		kDynamicHuffman = 0b10,
		kReserved =       0b11
	};

	using HuffmanCode = LemonSqueezy::BitSequence;

	struct HuffmanCodeRange
	{
		HuffmanCodeRange(const HuffmanCode &firstCode, HuffmanCode::code_type value, HuffmanCode::size_type length)
		  : firstCode(firstCode), value(value), length(length)
		{
		}

		HuffmanCode firstCode;
		HuffmanCode::code_type value;
		HuffmanCode::size_type length;
	};

	struct CompressedInteger
	{
		CompressedInteger(size_t base, size_t extraBitsNeeded)
		  : base(base), extraBitsNeeded(extraBitsNeeded)
		{
		}

		size_t base;
		size_t extraBitsNeeded;
	};

	// hugely inefficient from a memory point of view, but hugely efficient
	// from a lookup point of view (guaranteed constant rather than simply
	// amortised constant). Uses potentially 1MB of space for the table but
	// will outperform any hash table.
	class FastHuffmanCodeMap
	{
	public:
		using Tree = std::vector<short>;

		FastHuffmanCodeMap(HuffmanCode::size_type maxCodeLength)
		  : m_tree(2U << maxCodeLength), m_reverse(289), m_maxCodeLength(maxCodeLength)
		{
		}

		FastHuffmanCodeMap(const FastHuffmanCodeMap &) = delete;

		FastHuffmanCodeMap(FastHuffmanCodeMap &&other)
		  : m_tree(std::move(other.m_tree)), m_reverse(std::move(other.m_reverse)), m_maxCodeLength(other.m_maxCodeLength)
		{
		}

		bool insert(const HuffmanCode &code, HuffmanCode::code_type value)
		{
			Tree::size_type offset = offsetForCode(code);
			if (m_tree[offset])
				return false;
			m_tree[offset] = value + 1;
			m_reverse[value] = code;
			return true;
		}

		bool search(const HuffmanCode &code, HuffmanCode::code_type &result) const
		{
			Tree::size_type offset = offsetForCode(code);
			if (m_tree[offset])
			{
				result = m_tree[offset] - 1;
				return true;
			}
			return false;
		}

		HuffmanCode reverseSearch(const HuffmanCode::code_type value) const
		{
			if (value >= m_reverse.size())
				throw LemonSqueezy::CompressionException("Fixed Huffman code is invalid");
			
			return m_reverse[value];
		}

	private:
		Tree::size_type offsetForCode(const HuffmanCode &code) const
		{
			if (code.length > m_maxCodeLength)
				throw LemonSqueezy::DecompressionException("Code too large for this table");
			return ((1U << code.length) - 2) + code.code;
		}

		Tree m_tree;
		std::vector<HuffmanCode> m_reverse;
		HuffmanCode::size_type m_maxCodeLength;
	};

	HuffmanCode::size_type findLongestCode(const std::vector<HuffmanCodeRange> &codeRanges)
	{
		HuffmanCode::size_type longest = 0;
		for (auto codeRange : codeRanges)
			longest = std::max(codeRange.firstCode.length, longest);
		return longest;
	}

	HuffmanCode::size_type findLongestCode(const std::vector<HuffmanCode::code_type> &lengths)
	{
		HuffmanCode::size_type longest = 0;
		for (auto length : lengths)
			longest = std::max(static_cast<HuffmanCode::size_type>(length), longest);
		return longest;
	}

	FastHuffmanCodeMap generateHuffmanCodeTableFromRanges(const std::vector<HuffmanCodeRange> &codeRanges)
	{
		FastHuffmanCodeMap result(findLongestCode(codeRanges));

		// cycle through all ranges and generate a Huffman code for each value
		for (const auto &r : codeRanges)
		{
			for (HuffmanCode::size_type i = 0; i < r.length; i++)
			{
				HuffmanCode hc = r.firstCode;
				hc.code += i;
				if (!result.insert(hc, r.value + i))
					throw LemonSqueezy::DecompressionException("Duplicate code in range");
			}
		}

		return result;
	}

	const FastHuffmanCodeMap g_fixedHuffmanLiteralTable = generateHuffmanCodeTableFromRanges({
		{ {  0b00110000, 8 },   0, 144 },
		{ { 0b110010000, 9 }, 144, 112 },
		{ {   0b0000000, 7 }, 256,  24 },
		{ {  0b11000000, 8 }, 280,   8 }
	});

	const FastHuffmanCodeMap g_fixedHuffmanDistanceTable = generateHuffmanCodeTableFromRanges({
		{ { 0b00000, 5 }, 0, 30 }
	});

	// This function utilises the algorithm from RFC1951. Basically, a Huffman tree can be
	// constructed purely based on the length of the codes of each symbol.
	FastHuffmanCodeMap generateFastHuffmanCodeTableFromLengths(const std::vector<HuffmanCode::code_type> &lengths)
	{
		FastHuffmanCodeMap table(findLongestCode(lengths));

		std::vector<HuffmanCode::size_type> lengthCounts(sizeof(HuffmanCode::code_type) * CHAR_BIT + 1);
		for (HuffmanCode::code_type l : lengths)
			if (l <= lengthCounts.size() && l > 0)
				lengthCounts[l]++;

		HuffmanCode::code_type code = 0;
		std::vector<HuffmanCode::code_type> nextCodes(lengthCounts.size() + 1);
		for (HuffmanCode::size_type l = 1; l <= lengthCounts.size(); l++)
		{
			code = (code + lengthCounts[l - 1]) << 1;
			nextCodes[l] = code;
		}

		for (std::vector<HuffmanCode::code_type>::size_type l = 0; l < lengths.size(); l++)
		{
			if (lengths[l] == 0)
				continue;
			
			bool didInsert = table.insert(HuffmanCode(nextCodes[lengths[l]]++, lengths[l]), static_cast<unsigned>(l));
			if (!didInsert)
				throw LemonSqueezy::DecompressionException("Duplicate code!");
		}

		return table;
	}

	bool readNextCode(LemonSqueezy::IBitReader &bitReader, const FastHuffmanCodeMap &table, HuffmanCode::code_type &value)
	{
		HuffmanCode hc;
		value = 0;

		while (hc.length < (sizeof hc.code) * CHAR_BIT)
		{
			// read bits 1-by-1, this causes the code to be read in reversed order
			// which is how it is stored in the stream
			if (!bitReader.appendNextBit(hc))
				throw LemonSqueezy::IOException("Unexpected end of stream reading next Huffman code");

			if (table.search(hc, value))
				return true;
		}

		return false;
	}

	const std::unordered_map<size_t, CompressedInteger> g_lengthMap {
		{ 257, {   3, 0 } },
		{ 258, {   4, 0 } },
		{ 259, {   5, 0 } },
		{ 260, {   6, 0 } },
		{ 261, {   7, 0 } },
		{ 262, {   8, 0 } },
		{ 263, {   9, 0 } },
		{ 264, {  10, 0 } },
		{ 265, {  11, 1 } },
		{ 266, {  13, 1 } },
		{ 267, {  15, 1 } },
		{ 268, {  17, 1 } },
		{ 269, {  19, 2 } },
		{ 270, {  23, 2 } },
		{ 271, {  27, 2 } },
		{ 272, {  31, 2 } },
		{ 273, {  35, 3 } },
		{ 274, {  43, 3 } },
		{ 275, {  51, 3 } },
		{ 276, {  59, 3 } },
		{ 277, {  67, 4 } },
		{ 278, {  83, 4 } },
		{ 279, {  99, 4 } },
		{ 280, { 115, 4 } },
		{ 281, { 131, 5 } },
		{ 282, { 163, 5 } },
		{ 283, { 195, 5 } },
		{ 284, { 227, 5 } },
		{ 285, { 258, 0 } }
	};

	const std::unordered_map<size_t, CompressedInteger> g_distanceMap {
		{  0, {     1,  0 } },
		{  1, {     2,  0 } },
		{  2, {     3,  0 } },
		{  3, {     4,  0 } },
		{  4, {     5,  1 } },
		{  5, {     7,  1 } },
		{  6, {     9,  2 } },
		{  7, {    13,  2 } },
		{  8, {    17,  3 } },
		{  9, {    25,  3 } },
		{ 10, {    33,  4 } },
		{ 11, {    49,  4 } },
		{ 12, {    65,  5 } },
		{ 13, {    97,  5 } },
		{ 14, {   129,  6 } },
		{ 15, {   193,  6 } },
		{ 16, {   257,  7 } },
		{ 17, {   385,  7 } },
		{ 18, {   513,  8 } },
		{ 19, {   769,  8 } },
		{ 20, {  1025,  9 } },
		{ 21, {  1537,  9 } },
		{ 22, {  2049, 10 } },
		{ 23, {  3073, 10 } },
		{ 24, {  4097, 11 } },
		{ 25, {  6145, 11 } },
		{ 26, {  8193, 12 } },
		{ 27, { 12289, 12 } },
		{ 28, { 16385, 13 } },
		{ 29, { 24577, 13 } }
	};

	std::tuple<size_t, size_t> readLengthAndDistance(LemonSqueezy::IBitReader &bitReader, HuffmanCode::code_type lengthCode, const FastHuffmanCodeMap &distanceTable)
	{
		const auto lengthEntry = g_lengthMap.find(lengthCode);
		if (lengthEntry == g_lengthMap.end())
			throw LemonSqueezy::DecompressionException("Unknown length code");

		HuffmanCode extraLengthBits;
		if (!bitReader.readBitsIntoBitSequence(lengthEntry->second.extraBitsNeeded, extraLengthBits))
			throw LemonSqueezy::DecompressionException("Unexpected end of stream while reading extra length bits");
		
		HuffmanCode::code_type distanceCode;
		if (!readNextCode(bitReader, distanceTable, distanceCode))
			throw LemonSqueezy::DecompressionException("Unable to read distance code");
		
		const auto distanceEntry = g_distanceMap.find(distanceCode);
		if (distanceEntry == g_distanceMap.end())
			throw LemonSqueezy::DecompressionException("Unknown distance code");
		
		HuffmanCode extraDistanceBits;
		if (!bitReader.readBitsIntoBitSequence(distanceEntry->second.extraBitsNeeded, extraDistanceBits))
			throw LemonSqueezy::DecompressionException("Unexpected end of stream while reading extra distance bits");
		
		return std::make_tuple<size_t, size_t>(lengthEntry->second.base + extraLengthBits.code, distanceEntry->second.base + extraDistanceBits.code);
	}

	void decodeCodeLengthTable(LemonSqueezy::IBitReader &bitReader, std::vector<HuffmanCode::code_type> &codeLengths, const FastHuffmanCodeMap &codeLengthTable)
	{
		size_t i = 0;
		while (i < codeLengths.size())
		{
			HuffmanCode::code_type code;
			if (!readNextCode(bitReader, codeLengthTable, code))
				throw LemonSqueezy::DecompressionException("Unable to read code length");
			
			// if code is equal to or less than 15, then it is a literal
			// code length
			if (code <= 15)
			{
				codeLengths[i++] = code;
				continue;
			}
			
			LemonSqueezy::BitSequence bs;
			size_t j = 0;
			
			if (code == 16)
			{
				// when code = 16, it means to repeat the previous code 3-6 times, depending on the next two bits
				if (i == 0)
					throw LemonSqueezy::DecompressionException("No codes to repeat");
				
				if (!bitReader.readBitsIntoBitSequence(2, bs))
					throw LemonSqueezy::DecompressionException("Unable to read repeat count for code length table");
				
				if (i + bs.code + 3 > codeLengths.size())
					throw LemonSqueezy::DecompressionException("Repetition goes past the end of the table!");

				for (j = 0; j < bs.code + 3; j++)
					codeLengths[i + j] = codeLengths[i - 1];
			}
			else if (code == 17)
			{
				// when code = 17, it means to insert 3-10 "zeros", depending on the next 3 bits
				if (!bitReader.readBitsIntoBitSequence(3, bs))
					throw LemonSqueezy::DecompressionException("Unable to read repeat count for code length table");
				
				for (j = 0; j < bs.code + 3 && i + j < codeLengths.size(); j++)
					codeLengths[i + j] = 0;
			}
			else if (code == 18)
			{
				// when code = 18, it means to insert 11-138 "zeroes", depending on the next 7 bits
				if (!bitReader.readBitsIntoBitSequence(7, bs))
					throw LemonSqueezy::DecompressionException("Unable to read repeat count for code length table");
				
				for (j = 0; j < bs.code + 11; j++)
					codeLengths[i + j] = 0;
			}
			
			i += j;
		}
	}

	std::pair<FastHuffmanCodeMap, FastHuffmanCodeMap> readDynamicHuffmanDictionaries(LemonSqueezy::IBitReader &bitReader)
	{
		LemonSqueezy::BitSequence literalCodeCount;
		LemonSqueezy::BitSequence distanceCodeCount;
		LemonSqueezy::BitSequence codeLengthCodeCount;

		if (!bitReader.readBitsIntoBitSequence(5, literalCodeCount))
			throw LemonSqueezy::DecompressionException("Unable to read number of literal codes");
		
		if (!bitReader.readBitsIntoBitSequence(5, distanceCodeCount))
			throw LemonSqueezy::DecompressionException("Unable to read number of distance codes");
		
		if (!bitReader.readBitsIntoBitSequence(4, codeLengthCodeCount))
			throw LemonSqueezy::DecompressionException("Unable to read number of code length codes");
		
		unsigned codeLengthPermutation[19] = { 16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15 };
		std::vector<HuffmanCode::code_type> codeLengths(19);
		for (size_t i = 0; i < codeLengthCodeCount.code + 4; i++)
		{
			LemonSqueezy::BitSequence length;
			if (!bitReader.readBitsIntoBitSequence(3, length))
				throw LemonSqueezy::DecompressionException("Cannot read code length");
			codeLengths[codeLengthPermutation[i]] = length.code;
		}

		FastHuffmanCodeMap codeLengthTable = generateFastHuffmanCodeTableFromLengths(codeLengths);
		
		std::vector<HuffmanCode::code_type> allCodeLengths(literalCodeCount.code + 257 + distanceCodeCount.code + 1);
		decodeCodeLengthTable(bitReader, allCodeLengths, codeLengthTable);

		FastHuffmanCodeMap lengthMap = generateFastHuffmanCodeTableFromLengths({ allCodeLengths.begin(), allCodeLengths.begin() + literalCodeCount.code + 257 });
		FastHuffmanCodeMap distanceMap = generateFastHuffmanCodeTableFromLengths({ allCodeLengths.begin() + literalCodeCount.code + 257, allCodeLengths.end() });
		return std::make_pair(std::move(lengthMap), std::move(distanceMap));
	}

	void readBlock(LemonSqueezy::IBitReader &bitReader, const FastHuffmanCodeMap &literalTable, const FastHuffmanCodeMap &distanceTable, LemonSqueezy::RollingOutputBuffer &out)
	{
		bool endOfBlockFound = false;
		while (!endOfBlockFound)
		{
			HuffmanCode::code_type code;
			if (!readNextCode(bitReader, literalTable, code))
				throw LemonSqueezy::DecompressionException("Encountered end of stream too early");

			if (code < 256)
			{
				// literal value
				out.put(static_cast<char>(code));
			}
			else if (code == 256)
			{
				// end of block marker
				endOfBlockFound = true;
			}
			else if (code >= 257 && code <= 285)
			{
				// distance length pair
				size_t length, distance;
				std::tie(length, distance) = readLengthAndDistance(bitReader, code, distanceTable);
				out.repeatBytes(distance, length);
			}
		}
	}

	/**
	 * Since both lengths and distances can be represented using compressed integers,
	 * this function unifies the work required for generating either a length code or
	 * a distance code, and returns a tuple of two HuffmanCodes. The first element in
	 * the tuple is the Huffman code itself, and the second one represents the extra
	 * bits that need to be added to complete the compressed integer (strictly speaking
	 * the extra bits are not a Huffman code but the data wraps nicely inside that object
	 * and makes it easier to output using the IBitWriter)
	 */
	template<typename T, size_t N>
	auto makeHuffmanCodesFromMap(const std::array<T, N> &table, T val, const std::unordered_map<size_t, CompressedInteger> &extraBitsMap, const FastHuffmanCodeMap &map, HuffmanCode::code_type codeOffset = 0)
	{
		if (val < table.front())
			throw LemonSqueezy::CompressionException("Value is too small for this table");
		
		if (val >= table.back())
			throw LemonSqueezy::CompressionException("Value is too large for this table");

		auto it = std::upper_bound(table.begin(), table.end(), val);
		std::advance(it, -1);
		
		// get the length code from the provided code map
		HuffmanCode::code_type codeValue = codeOffset + (it - table.begin());
		HuffmanCode code = map.reverseSearch(codeValue);

		// create a huffman code for the extra bits of the compressed integer
		auto extra = extraBitsMap.find(codeValue);
		if (extra == extraBitsMap.end())
			throw LemonSqueezy::CompressionException("Internal error: extra bits map is invalid");
		
		HuffmanCode remainder(val - *it, extra->second.extraBitsNeeded);
		return std::make_tuple(code, remainder);
	}

#ifdef USE_STRIPED_HORSE_DEFLATE_COMPRESSION
	std::tuple<HuffmanCode, HuffmanCode> getHuffmanCodeForLength(const FastHuffmanCodeMap &map, uint16_t length)
	{
		// This table represents the lower bound of the codeable lengths
		// in RFC1951. The last value in this table is not specified in
		// the RFC, but it is just used to aid in range validation
		const static std::array<uint16_t, 30> lengths = { {
			3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
			35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 259
		} };

		return makeHuffmanCodesFromMap(lengths, length, g_lengthMap, map, 257);
	}

	std::tuple<HuffmanCode, HuffmanCode> getHuffmanCodeForDistance(const FastHuffmanCodeMap &map, uint16_t distance)
	{
		// This table represents the lower bound of the codeable distances
		// in RFC1951. The last value in this table is not specified in
		// the RFC, but it is just used to aid in range validation
		const static std::array<uint16_t, 31> distances = { {
			1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
			257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
			8193, 12289, 16385, 24577, 32769
		} };

		return makeHuffmanCodesFromMap(distances, distance, g_distanceMap, map, 0);
	}
#endif
}

LemonSqueezy::DeflateCompression::DeflateCompression()
{
}

LemonSqueezy::DeflateCompression::~DeflateCompression()
{
}

void LemonSqueezy::DeflateCompression::read(std::istream &in, std::ostream &out)
{
	DeflateStreamBitReader bitReader(in);
	RollingOutputBuffer rollingBuffer(32768, out); // maximum supported backwards distance
	
	BitSequence finalBlock;
	while (finalBlock.code == 0 && bitReader.readBitsIntoBitSequence(1, finalBlock))
	{
		finalBlock.length = 0; // reset length but keep code

		BitSequence blockTypeCode;
		if (!bitReader.readBitsIntoBitSequence(2, blockTypeCode))
			throw DecompressionException("Unable to read block type");

		BlockType blockType = static_cast<BlockType>(blockTypeCode.code);
		switch (blockType)
		{
			case BlockType::kNoCompression:
			{
				bitReader.alignToByteBoundary();
				BitSequence len, nlen;

				if (!(bitReader.readBitsIntoBitSequence(16, len) && bitReader.readBitsIntoBitSequence(16, nlen)))
					throw DecompressionException("Unexpected end of stream reading uncompressed data lengths");

				if ((len.code & 0xFFFF) != (~nlen.code & 0xFFFF))
					throw DecompressionException("Error reading length of uncompressed data");
				
				for (BitSequence::code_type i = 0; i < len.code; i++)
				{
					char byte;
					if (!in.get(byte))
						throw DecompressionException("Stream ended early, read " + std::to_string(i) + " of " + std::to_string(len.code) + " bytes");
					
					rollingBuffer.put(byte);
				}
				
				break;
			}
			
			case BlockType::kFixedHuffman:
			{
				readBlock(bitReader, g_fixedHuffmanLiteralTable, g_fixedHuffmanDistanceTable, rollingBuffer);
				break;
			}
			
			case BlockType::kDynamicHuffman:
			{
				// first is lengths, second is distances
				std::pair<FastHuffmanCodeMap, FastHuffmanCodeMap> dictionaries = readDynamicHuffmanDictionaries(bitReader);
				readBlock(bitReader, dictionaries.first, dictionaries.second, rollingBuffer);
				break;
			}
			
			case BlockType::kReserved:
				throw DecompressionException("Encountered reserved block");
		}
	}
}

#ifndef USE_STRIPED_HORSE_DEFLATE_COMPRESSION
// LemonSqueezy's deflate is being outperformed by the standard ZLib deflate
// by a noticeable factor (up to 20% better in some cases) so until Striped Horse's
// deflate algorithm is improved, we should use the ZLib deflate for now since it
// virtually always gives us better results

void LemonSqueezy::DeflateCompression::write(std::istream &in, std::ostream &out)
{
	std::vector<unsigned char> inBuffer(4096);
	std::vector<unsigned char> outBuffer(4096);

	z_stream zs;
	memset(&zs, 0, sizeof zs);

	zs.next_out = &outBuffer[0];
	zs.avail_out = outBuffer.size();

	int flush;
	int result = deflateInit2(&zs, 9, Z_DEFLATED, -15, 9, Z_DEFAULT_STRATEGY);
	
	do
	{
		flush = in.eof() ? Z_FINISH : Z_NO_FLUSH;

		in.read(reinterpret_cast<char *>(&inBuffer[0]), inBuffer.size());
		zs.next_in = &inBuffer[0];
		zs.avail_in = in.gcount();

		do
		{
			zs.next_out = &outBuffer[0];
			zs.avail_out = outBuffer.size();
			
			result = deflate(&zs, flush);
			if (result == Z_STREAM_ERROR)
				throw LemonSqueezy::CompressionException("Unknown stream exception occurred");
			
			out.write(reinterpret_cast<const char *>(&outBuffer[0]), outBuffer.size() - zs.avail_out);
		}
		while (zs.avail_out == 0);
	}
	while (flush != Z_FINISH);

	deflateEnd(&zs);
}

#else // don't use ZLIB compression

void LemonSqueezy::DeflateCompression::write(std::istream &in, std::ostream &out)
{
	DeflateStreamBitWriter bitWriter(out);
	HuffmanCode currentCode;
		
	// make this the final block (we are not yet dividing input into
	// multiple blocks)
	bitWriter.writeBits({0b1, 1});

	// use fixed Huffman code format for the entire block
	bitWriter.writeBits({0b01, 2});

	std::streambuf *buf = in.rdbuf();
	while (in)
	{
		size_t occurrences = 1;

		std::istream::int_type ch = buf->sbumpc();
		if (ch == std::istream::traits_type::eof())
			break;
		
		std::istream::int_type next = buf->sgetc();

		// max number of occurrences is 258 so we stop
		// if we reach that number
		while (ch == next && occurrences < 257)
		{
			next = buf->snextc();
			occurrences++;
		}
		
		currentCode = g_fixedHuffmanLiteralTable.reverseSearch(static_cast<unsigned char>(ch));
		bitWriter.writeBitsBackwards(currentCode);

		if (occurrences <= 3)
		{
			// no shortcut for just two or three occurrences, we just output the code a
			// second/third time (it's less bits than if we use a length/distance pair
			// anyway)
			while (occurrences > 1)
			{
				bitWriter.writeBitsBackwards(currentCode);
				occurrences--;
			}
		}
		else
		{
			// work out what code to use for the given number of occurrences
			HuffmanCode distanceCode, lengthCode, extraBits;

			// we find the code for the number of occurrences minus one (because the first
			// one was already added to the stream)
			std::tie(lengthCode, extraBits) = getHuffmanCodeForLength(g_fixedHuffmanLiteralTable, occurrences - 1);
			bitWriter.writeBitsBackwards(lengthCode);
			if (extraBits.length > 0)
				bitWriter.writeBits(extraBits); // these appear in "machine order"

			std::tie(distanceCode, extraBits) = getHuffmanCodeForDistance(g_fixedHuffmanDistanceTable, 1);
			bitWriter.writeBitsBackwards(distanceCode);
			if (extraBits.length > 0)
				bitWriter.writeBits(extraBits);
		}
	}

	// output end of block (code 256 indicates end of block)
	currentCode = g_fixedHuffmanLiteralTable.reverseSearch(256);
	bitWriter.writeBitsBackwards(currentCode);
}

#endif
