#include "exceptions.h"
#include "jpeg-dct.h"

#ifdef __APPLE__
#include <CoreFoundation/CoreFoundation.h>
#include <CoreGraphics/CoreGraphics.h>

namespace
{
	struct DataProviderContext
	{
		DataProviderContext(std::istream &in)
		  : input(in)
		{
		}

		static size_t GetBytes(void *info, void *buffer, size_t count)
		{
			DataProviderContext *context = static_cast<DataProviderContext *>(info);
			context->input.read(reinterpret_cast<char *>(buffer), count);
			return context->input.gcount();
		}

		static off_t SkipBytes(void *info, off_t count)
		{
			DataProviderContext *context = static_cast<DataProviderContext *>(info);
			context->input.seekg(count, std::ios_base::cur);
			return count;
		}

		static void Rewind(void *info)
		{
			DataProviderContext *context = static_cast<DataProviderContext *>(info);
			context->input.seekg(0, std::ios_base::beg);
		}

		std::istream &input;
	};

	std::tuple<CFDataRef, bool> createBufferAndDrawImage(CGImageRef image, size_t &samplesPerPixel)
	{
		bool hasExtraChannel = false;
		size_t width = CGImageGetWidth(image);
		size_t height = CGImageGetHeight(image);
		size_t bytesPerRow = CGImageGetBytesPerRow(image);
		uint32_t flags = 0;
		CGColorSpaceRef colorSpace = CGImageGetColorSpace(image);

		CFMutableDataRef data = CFDataCreateMutable(kCFAllocatorDefault, height * bytesPerRow);
		CFDataSetLength(data, height * bytesPerRow);

		// CGContext does not support 24-bit color values, so we have to specify 32-bit with no alpha channel
		// and subsequently strip it out
		if (CGColorSpaceGetModel(colorSpace) == kCGColorSpaceModelRGB)
		{
			flags = kCGImageAlphaNoneSkipLast;
			hasExtraChannel = true;
			samplesPerPixel = 3;
		}
		else
		{
			samplesPerPixel = CGColorSpaceGetNumberOfComponents(colorSpace);
		}
		
		CGContextRef context = CGBitmapContextCreate(CFDataGetMutableBytePtr(data), width, height, 8, bytesPerRow, colorSpace, flags);

		CGContextDrawImage(context, CGRectMake(0, 0, width, height), image);
		CGContextRelease(context);

		return std::make_tuple(data, hasExtraChannel);
	}

	void filterJPEG(std::istream &in, std::ostream *out, LemonSqueezy::JPEGCompression::ScanlineCallback callback, size_t &samplesPerPixel)
	{
		// not used on macOS
		(void) callback;

		DataProviderContext providerContext(in);

		CGDataProviderSequentialCallbacks providerCallbacks;
		providerCallbacks.version = 0;
		providerCallbacks.getBytes = DataProviderContext::GetBytes;
		providerCallbacks.skipForward = DataProviderContext::SkipBytes;
		providerCallbacks.rewind = DataProviderContext::Rewind;
		providerCallbacks.releaseInfo = nullptr;

		CGDataProviderRef dataProvider = CGDataProviderCreateSequential(&providerContext, &providerCallbacks);
		CGImageRef image = CGImageCreateWithJPEGDataProvider(dataProvider, nullptr, true, kCGRenderingIntentDefault);
		CFDataRef imageBuffer;
		bool hasExtraChannel;
		std::tie(imageBuffer, hasExtraChannel) = createBufferAndDrawImage(image, /*out*/ samplesPerPixel);

		CFIndex dataSize = CFDataGetLength(imageBuffer);

		const unsigned char *bytes = static_cast<const unsigned char *>(CFDataGetBytePtr(imageBuffer));
		size_t width = CGImageGetWidth(image);
		size_t height = CGImageGetHeight(image);
		size_t bytesPerRow = CGImageGetBytesPerRow(image);
			
		if (hasExtraChannel)
		{
			// strip out alpha channel (4th component)
			std::vector<unsigned char> row(CGImageGetWidth(image) * 3);

			for (size_t y = 0; y < height; y++)
			{
				for (size_t x = 0; x < width; x++)
				{
					row[x * 3 + 0] = bytes[(y * width + x) * 4 + 0];
					row[x * 3 + 1] = bytes[(y * width + x) * 4 + 1];
					row[x * 3 + 2] = bytes[(y * width + x) * 4 + 2];
				}
				
				if (out)
					out->write(reinterpret_cast<const char *>(row.data()), row.size());
				
				if (callback)
					callback(row.data(), row.size());
			}
		}
		else
		{
			if (out)
				out->write(reinterpret_cast<const char *>(&bytes[0]), dataSize);
			
			if (callback)
				for (size_t y = 0; y < height; y++)
					callback(&bytes[y * bytesPerRow], bytesPerRow);
		}

		CFRelease(imageBuffer);
		CGImageRelease(image);
		CGDataProviderRelease(dataProvider);
	}
}
#else
#include <csetjmp>
#include <vector>
#include <jpeglib.h>

namespace
{
	class InternalJPEGError : public std::runtime_error
	{
	public:
		InternalJPEGError(const std::string &msg) : std::runtime_error(msg) {}
	};
	
	// istream in this case is a pointer to keep this struct POD (i.e. C-compatible)
	struct istream_source_mgr
	{
		jpeg_source_mgr pub;
		char buffer[4096];
		std::istream *input;
	};

	void initSource(j_decompress_ptr cinfo)
	{
		(void) cinfo;
		// nothing to do
	}

	boolean fillInputBuffer(j_decompress_ptr cinfo)
	{
		istream_source_mgr *src = reinterpret_cast<istream_source_mgr *>(cinfo->src);
		src->input->read(src->buffer, sizeof src->buffer);
		src->pub.bytes_in_buffer = src->input->gcount();
		src->pub.next_input_byte = reinterpret_cast<JOCTET *>(src->buffer);
		return TRUE;
	}

	void skipInputData(j_decompress_ptr cinfo, long num_bytes)
	{
		istream_source_mgr *src = reinterpret_cast<istream_source_mgr *>(cinfo->src);
		src->input->seekg(num_bytes, std::ios_base::cur);
	}

	boolean resyncToRestart(j_decompress_ptr cinfo, int desired)
	{
		(void) cinfo;
		(void) desired;
		// not sure what this function is for
		return TRUE;
	}

	void termSource(j_decompress_ptr cinfo)
	{
		(void) cinfo;
		// nothing to do
	}

	void errorExit(j_common_ptr cinfo)
	{
		char buffer[JMSG_LENGTH_MAX];
		cinfo->err->format_message(cinfo, buffer);
		throw InternalJPEGError(buffer);
	}

	class JPEGDecompressorHandle
	{
	public:
		JPEGDecompressorHandle(const JPEGDecompressorHandle &) = delete;
		JPEGDecompressorHandle(JPEGDecompressorHandle &&) = delete;
		JPEGDecompressorHandle(std::istream &in)
		{
			m_src.pub.init_source = initSource;
			m_src.pub.fill_input_buffer = fillInputBuffer;
			m_src.pub.skip_input_data = skipInputData;
			m_src.pub.resync_to_restart = resyncToRestart;
			m_src.pub.term_source = termSource;
			m_src.pub.bytes_in_buffer = 0;
			m_src.pub.next_input_byte = reinterpret_cast<JOCTET *>(m_src.buffer);
			m_src.input = &in;

			jpeg_std_error(&m_err);

			m_err.error_exit = errorExit;

			jpeg_create_decompress(&m_cinfo);
		
			m_cinfo.err = &m_err;
			m_cinfo.client_data = nullptr;
			m_cinfo.src = reinterpret_cast<jpeg_source_mgr *>(&m_src);
		}

		~JPEGDecompressorHandle()
		{
			jpeg_destroy_decompress(&m_cinfo);
		}

		jpeg_decompress_struct *get()
		{
			return &m_cinfo;
		}

	private:
		jpeg_decompress_struct m_cinfo;
		istream_source_mgr m_src;
		struct jpeg_error_mgr m_err;
	};

	void filterJPEG(std::istream &in, std::ostream *out, LemonSqueezy::JPEGCompression::ScanlineCallback callback, size_t &samplesPerPixel)
	{
		JSAMPARRAY buffer;
		JPEGDecompressorHandle handle(in);

		try
		{
			jpeg_decompress_struct *cinfo = handle.get();
			jpeg_read_header(cinfo, (boolean)1);
			jpeg_start_decompress(cinfo);

			samplesPerPixel = cinfo->output_components;
			size_t bytesPerRow = cinfo->output_width * samplesPerPixel;
			
			buffer = cinfo->mem->alloc_sarray(reinterpret_cast<j_common_ptr>(cinfo), JPOOL_IMAGE, bytesPerRow, 1);

			while (cinfo->output_scanline < cinfo->output_height)
			{
				jpeg_read_scanlines(cinfo, buffer, 1);
				if (out)
					out->write(reinterpret_cast<const char *>(buffer[0]), bytesPerRow);
				if (callback)
					callback(buffer[0], bytesPerRow);
			}

			jpeg_finish_decompress(cinfo);
		}
		catch (const InternalJPEGError &ex)
		{
			throw LemonSqueezy::DecompressionException(ex.what());
		}
	}
}
#endif

LemonSqueezy::JPEGCompression::JPEGCompression()
  : m_samplesPerPixel(0)
{
}

void LemonSqueezy::JPEGCompression::read(std::istream &in, ScanlineCallback callback)
{
	filterJPEG(in, nullptr, callback, /* out */ m_samplesPerPixel);
}

void LemonSqueezy::JPEGCompression::read(std::istream &in, std::ostream &out)
{
	filterJPEG(in, &out, nullptr, /* out */ m_samplesPerPixel);
}
