#include <string>
#include <vector>
#include <unordered_map>

#include "streambitreader.h"
#include "streambitwriter.h"
#include "exceptions.h"
#include "lzw.h"

namespace
{
	using LZWCode = LemonSqueezy::BitSequence;

	class LZWTable
	{
	public:
		using TablingSystem = LemonSqueezy::LZWCompression::TablingSystem;

		enum
		{
			kMaxLZWCode = 4095
		};

		LZWTable(TablingSystem tablingSystem)
		  : m_codeSize(9), m_nextCode(258), m_tablingSystem(tablingSystem)
		{
			m_dictionary.resize(kMaxLZWCode + 1);
			for (size_t i = 0; i < 256; i++)
			{
				std::string value = std::string { static_cast<char>(i) };
				m_dictionary[i] = value;
			}
			
			reset();
		}

		LZWCode add(const std::string &str)
		{
			if (m_nextCode > kMaxLZWCode)
				throw LemonSqueezy::DecompressionException("LZW table is full!");
			
			LZWCode result = LZWCode(m_nextCode, m_codeSize);
			m_dictionary[m_nextCode] = str;
			m_reverseDictionary[str] = m_nextCode;
			m_nextCode++;

			size_t nextCodeOffset = 0;
			if (m_tablingSystem == TablingSystem::kOffByOne)
				nextCodeOffset = 1;

			if (m_nextCode + nextCodeOffset >= (1U << m_codeSize))
				m_codeSize++;
			
			return result;
		}

		void add(const LZWCode &code, const std::string &str)
		{
			if (code.length > m_codeSize)
				throw LemonSqueezy::DecompressionException("Invalid LZW code");

			m_dictionary[code.code] = str;
		}

		size_t getCodeSize() const
		{
			return m_codeSize;
		}

		size_t getNextCode() const
		{
			return m_nextCode;
		}
		
		bool lookupCode(const LZWCode &code, std::string &result) const
		{
			if (code.length > m_codeSize)
				throw LemonSqueezy::DecompressionException("LZW code is too large!");
			
			if (code.code >= m_nextCode)
				return false;

			result = m_dictionary[code.code];
			return true;
		}

		bool lookupString(const std::string &str, LZWCode &code) const
		{
			if (str.size() == 1)
			{
				code.code = static_cast<unsigned char>(str[0]);
				code.length = m_codeSize;
				return true;
			}
			
			const auto it = m_reverseDictionary.find(str);
			if (it == m_reverseDictionary.end())
				return false;
			
			code.code = it->second;
			code.length = m_codeSize;
			return true;
		}

		void reset()
		{
			m_nextCode = 258;
			m_codeSize = 9;
			m_reverseDictionary.clear();
		}

	private:
		size_t m_codeSize;
		size_t m_nextCode;
		LemonSqueezy::LZWCompression::TablingSystem m_tablingSystem;
		std::vector<std::string> m_dictionary;
		std::unordered_map<std::string, size_t> m_reverseDictionary;
	};
}


LemonSqueezy::LZWCompression::LZWCompression(TablingSystem tablingSystem)
  : m_tablingSystem(tablingSystem)
{
}


void LemonSqueezy::LZWCompression::read(std::istream &in, std::ostream &out)
{
	StreamBitReader bitReader(in);
	bool endCodeEncountered = false;

	LZWTable table(m_tablingSystem);
	LZWCode lzwCode;
	LZWCode oldCode;
	std::string previousString;

	while (!endCodeEncountered && bitReader.readBitsIntoBitSequence(table.getCodeSize(), lzwCode))
	{
		std::string s;

		switch (lzwCode.code)
		{
			case 256:
			{
				table.reset();

				while (lzwCode.code == 256)
				{
					lzwCode = LZWCode();
					if (!bitReader.readBitsIntoBitSequence(table.getCodeSize(), lzwCode))
						throw DecompressionException("Unexpected end of LZW stream");
				}
				
				if (lzwCode.code == 257)
				{
					endCodeEncountered = true;
					break;
				}
				
				if (!table.lookupCode(lzwCode, s))
					throw DecompressionException("Invalid code in stream");
				
				out << s;
				oldCode = lzwCode;		

				break;
			}
			
			case 257:
			{
				endCodeEncountered = true;
				break;
			}
			
			default:
			{
				std::string previousString;
				if (!table.lookupCode(oldCode, previousString))
					throw DecompressionException("Internal inconsistency in LZW decompression");

				if (table.lookupCode(lzwCode, s))
				{
					if (m_tablingSystem == TablingSystem::kOffByOne || previousString.size() > 0)
						table.add(previousString + s[0]);
				}
				else
				{
					if (previousString.size() == 0)
						throw DecompressionException("LZW code is invalid");
					
					s = previousString + previousString[0];
					table.add(s);
				}

				out << s;
				oldCode = lzwCode;
			}
		}

		lzwCode = LZWCode();
	}

	if (!endCodeEncountered)
		throw DecompressionException("Unexpected end of stream");
	
	if (!in.eof() && in.peek() != std::istream::traits_type::eof())
		std::cerr << "Warning, not all data read" << std::endl;
}

void LemonSqueezy::LZWCompression::write(std::istream &in, std::ostream &out)
{
	size_t maxCode = LZWTable::kMaxLZWCode - (m_tablingSystem == TablingSystem::kOffByOne ? 1 : 0); 
	StreamBitWriter bitWriter(out);
	LZWTable table(m_tablingSystem);
	std::string currentString;

	// make a big buffer to prevent lots of reallocations
	currentString.reserve(1024);

	// write initialisation code to output
	bitWriter.writeBits({256, 9});

	// lastCode refers to either the code most recently found in the table
	// or the code of the most recently output character after a table reset
	LZWCode lastCode;
	while (true)
	{
		int nextChar = in.rdbuf()->sbumpc();
		if (nextChar == std::istream::traits_type::eof())
			break;
		
		char ch = static_cast<char>(nextChar);
		currentString.push_back(ch);

		// lookupString only updates lastCode if currentString was found
		if (table.lookupString(currentString, lastCode) == false)
		{
			// the longest match was found (and is represented by lastCode)
			bitWriter.writeBits(lastCode);
			
			if (table.getNextCode() == maxCode)
			{
				// reset the table (code 256), we've got too many codes
				bitWriter.writeBits({ 256, lastCode.length });
				table.reset();
			}
			else
			{
				table.add(currentString);
			}

			// reset what we're looking for
			currentString.resize(1);
			currentString[0] = ch;
			lastCode = LZWCode(static_cast<unsigned char>(ch), table.getCodeSize());
		}
	}

	// lastCode has not been written yet
	bitWriter.writeBits(lastCode);

	// 257 is the end-of-input marker
	bitWriter.writeBits({ 257, table.getCodeSize() });
}
