#include <algorithm>
#include <climits>
#include "exceptions.h"
#include "pngprediction.h"

namespace
{
	enum class PNGPredictionType : unsigned char
	{
		kSubtract = 0x01,
		kUp = 0x02,
		kAverage = 0x03,
		kPaeth = 0x04
	};
}

LemonSqueezy::PNGPrediction::PNGPrediction(unsigned predictor, unsigned colors, unsigned bitsPerComponent, unsigned columns)
  : m_predictor(predictor), m_colors(colors), m_bitsPerComponent(bitsPerComponent), m_columns(columns)
{
	if (m_predictor != 1 && m_predictor != 2 && !(m_predictor >= 10 && m_predictor <= 15))
		throw UnimplementedException("Unknown predictor value for deflate decompression");
	
	if (m_colors == 0)
		throw UnimplementedException("Colors must be greater than 0");
	
	if (m_bitsPerComponent != 8)
		throw UnimplementedException("Bits-per-component must be 8");

	if (m_columns == 0)
		throw UnimplementedException("Columns must be greater than 0");
}

void LemonSqueezy::PNGPrediction::read(std::istream &in, std::ostream &out)
{
	const size_t bytesPerComponent = m_colors * m_bitsPerComponent / CHAR_BIT;
	const size_t bytesPerRow = m_columns * bytesPerComponent;

	// must persist
	std::vector<unsigned char> previousRow(bytesPerRow);

	while (true)
	{
		char ch;
		if (!in.get(ch))
			break;
		
		enum PNGPredictionType prediction = static_cast<enum PNGPredictionType>(ch);

		std::vector<unsigned char> previousSample(m_colors);
		for (size_t i = 0; i < m_columns; i++)
		{
			for (size_t j = 0; j < m_colors; j++)
			{
				if (!in.get(ch))
					throw DecompressionException("Unexpected end of stream");

				unsigned char thisByte = static_cast<unsigned char>(ch);
				switch (prediction)
				{
					case PNGPredictionType::kSubtract:
						thisByte += previousSample[j];
						break;

					case PNGPredictionType::kAverage:
						thisByte += (previousSample[j] + previousRow[i * bytesPerComponent + j]) / 2;
						break;
					
					case PNGPredictionType::kUp:
						thisByte += previousRow[i * bytesPerComponent + j];
						break;
					
					case PNGPredictionType::kPaeth:
						throw UnimplementedException("Paeth prediction not yet implemented");
					
					default:
						throw DecompressionException("Prediction type is invalid");
				}
				
				out.put(thisByte);
				previousRow[i * bytesPerComponent + j] = thisByte;
				previousSample[j] = thisByte;
			}
		}
	}
}

void LemonSqueezy::PNGPrediction::write(std::istream &in, std::ostream &out)
{
	const size_t bytesPerComponent = m_colors * m_bitsPerComponent / CHAR_BIT;
	const size_t bytesPerRow = m_columns * bytesPerComponent;

	std::vector<unsigned char> thisRow(bytesPerRow);
	std::vector<unsigned char> previousRow(bytesPerRow);

	while (in.read(reinterpret_cast<char *>(thisRow.data()), thisRow.size()))
	{
		out.put(static_cast<char>(m_predictor));

		switch (static_cast<PNGPredictionType>(m_predictor))
		{
			case PNGPredictionType::kSubtract:
			{
				for (size_t column = 0; column < bytesPerRow; column += bytesPerComponent)
					for (size_t component = 0; component < bytesPerComponent; component++)
						out.put(thisRow[column] - ((column + component < bytesPerComponent) ? 0 : thisRow[column + component - bytesPerComponent]));
				
				break;
			}
			
			case PNGPredictionType::kUp:	
			{
				for (size_t column = 0; column < thisRow.size(); column++)
					out.put(static_cast<char>(thisRow[column] - previousRow[column]));
				break;
			}

			case PNGPredictionType::kAverage:
			{
				for (size_t column = 0; column < thisRow.size(); column++)
				{
					for (size_t component = 0; component < bytesPerComponent; component++)
					{
						unsigned char up = previousRow[column + component];
						unsigned char left = ((column + component < bytesPerComponent) ? 0 : thisRow[column + component - bytesPerComponent]);
						out.put(static_cast<char>(thisRow[column + component] - (left + up) / 2));
					}
				}
				break;
			}
			
			case PNGPredictionType::kPaeth:
				throw CompressionException("Paeth prediction not yet implemented");

			default:
				throw CompressionException("Prediction type is invalid");
		}

		previousRow.swap(thisRow);
	}

	if (in.gcount() != 0)
		throw CompressionException("Unexpected end of input");
}
