#include <algorithm>
#include <climits>
#include "bit-utilities.h"
#include "exceptions.h"
#include "streambitreader.h"

LemonSqueezy::StreamBitReader::StreamBitReader(std::istream &in)
  : m_bitPosition(CHAR_BIT), m_currentByte(0), m_in(in)
{
}

bool LemonSqueezy::StreamBitReader::appendNextBit(BitSequence &seq)
{
	return readBitsIntoBitSequence(1, seq);
}

void LemonSqueezy::StreamBitReader::alignToByteBoundary()
{
	if (m_bitPosition != 0)
		readNextByte();
}

bool LemonSqueezy::StreamBitReader::getByte(unsigned char &byte)
{
	if (m_bitPosition % 8 != 0)
		throw IOException("Not aligned at byte boundary");

	if (m_bitPosition == 8 && !readNextByte())
		return false;
	
	byte = m_currentByte;
	m_bitPosition = 8; // consider all of the bits of the byte read
	return true;
}

bool LemonSqueezy::StreamBitReader::isAtEndOfStream() const
{
	return m_in.eof();
}

bool LemonSqueezy::StreamBitReader::readBitsIntoBitSequence(BitSequence::size_type count, BitSequence &seq)
{
	if (count == 0)
		return true;

	if (isAtEndOfStream())
		return false;

	if (seq.length + count > sizeof(seq.code) * CHAR_BIT)
		throw std::out_of_range("count too large");

	seq.code <<= count;
	
	size_t bitsLeft = count;
	while (bitsLeft > 0 && m_in)
	{
		if (m_bitPosition == CHAR_BIT && !readNextByte())
			throw IOException("Unexpected end of stream");

		size_t bitsFromThisByte = std::min(CHAR_BIT - m_bitPosition, bitsLeft);
		size_t mask = (1U << bitsFromThisByte) - 1;
		seq.code |= ((m_currentByte >> (CHAR_BIT - bitsFromThisByte - m_bitPosition)) & mask) << (bitsLeft - bitsFromThisByte);
		
		bitsLeft -= bitsFromThisByte;
		m_bitPosition += bitsFromThisByte;
	}

	seq.length += count;
	return static_cast<bool>(m_in);
}

bool LemonSqueezy::StreamBitReader::readNextByte()
{
	if (m_in.rdbuf()->sgetc() == std::istream::traits_type::eof())
	{
		m_currentByte = 0;
		m_bitPosition = 8;
		return false;
	}

	m_currentByte = static_cast<unsigned char>(m_in.rdbuf()->sbumpc());
	m_bitPosition = 0;
	return true;
}
