#include <array>
#include <climits>
#include "exceptions.h"
#include "tiffprediction.h"

namespace
{
	bool tryReadSample(std::istream &in, uint8_t &uch)
	{
		char ch;
		if (in.get(ch))
		{
			uch = static_cast<uint8_t>(ch);
			return true;
		}

		return false;
	}

	bool tryReadSample(std::istream &in, uint16_t &val)
	{
		uint8_t low;
		uint8_t high;

		if (tryReadSample(in, high) && tryReadSample(in, low))
		{
			val = static_cast<uint16_t>(high) << 16 | low;
			return true;
		}

		return false;
	}

	template<typename T>
	void predict(std::istream &in, size_t columns, size_t colors, std::ostream &out)
	{
		std::array<T, LemonSqueezy::TIFFPrediction::kMaxSupportedColors> lastSample { { 0, 0, 0, 0 } };

		while (in)
		{
			for (size_t i = 0; i < columns; i++)
			{
				for (size_t j = 0; j < colors; j++)
				{
					T value;
					if (!tryReadSample(in, value))
					{
						if (i > 0 && j > 0)
							throw LemonSqueezy::DecompressionException("Row data is incomplete");
						break;
					}

					lastSample[j] += value;
					out.put(lastSample[j]);
				}
			}
		}
	}
}

LemonSqueezy::TIFFPrediction::TIFFPrediction(unsigned colors, unsigned bitsPerComponent, unsigned columns)
  : m_colors(colors), m_bitsPerComponent(bitsPerComponent), m_columns(columns)
{
	if (m_colors > kMaxSupportedColors || m_colors == 0)
		throw UnimplementedException("Unsupported number of colors for TIFF prediction");
	
	if (m_bitsPerComponent != 8 && m_bitsPerComponent != 16)
		throw UnimplementedException("Bits per component must be 8 or 16");
	
	if (m_columns == 0)
		throw DecompressionException("Unable to perform TIFF prediction with no column data");
}

void LemonSqueezy::TIFFPrediction::read(std::istream &in, std::ostream &out)
{
	if (m_bitsPerComponent == 8)
		predict<uint8_t>(in, m_columns, m_colors, out);
	else if (m_bitsPerComponent == 16)
		predict<uint16_t>(in, m_columns, m_colors, out);
}

void LemonSqueezy::TIFFPrediction::write(std::istream &in, std::ostream &out)
{
	const size_t bytesPerComponent = m_colors * m_bitsPerComponent / CHAR_BIT;
	const size_t bytesPerRow = bytesPerComponent * m_columns;

	std::vector<unsigned char> row(bytesPerRow);
	while (in.read(reinterpret_cast<char *>(row.data()), row.size()))
	{
		std::vector<unsigned char> lastComponent(bytesPerComponent);
		for (size_t i = 0; i < bytesPerRow; i += bytesPerComponent)
		{
			for (size_t j = 0; j < bytesPerComponent; j++)
			{
				unsigned char thisByte = row[i + j];
				out.put(thisByte - lastComponent[j]);
				lastComponent[j] = thisByte;
			}
		}
	}

	if (in.gcount() != 0)
		throw CompressionException("Unexpected end of input");
}
