#include <functional>
#include <fstream>
#include <sstream>
#include <tuple>
#include <unordered_map>
#include "crc32.h"
#include "deflate.h"
#include "zip.h"

#ifndef _WIN32
#include <arpa/inet.h>
#else
#include <winsock2.h>
#endif

namespace
{
	LemonSqueezy::ZipArchiveEntry::dostime getdostime(std::time_t t)
	{
		struct tm lt;
#ifndef _WIN32
		localtime_r(&t, &lt);
#else
		localtime_s(&lt, &t);
#endif

		LemonSqueezy::ZipArchiveEntry::dostime result = 0;

		result |= (lt.tm_year - 80) << 25;
		result |= (lt.tm_mon + 1) << 21;
		result |= lt.tm_mday << 16;
		result |= lt.tm_hour << 11;
		result |= lt.tm_min << 5;
		result |= lt.tm_sec >> 1;
		
		return result;
	}

	std::vector<unsigned char> readStreamToBuffer(std::istream &in)
	{
		std::vector<unsigned char> buffer;
		size_t bytesRead = 0;

		do
		{
			unsigned char rb[4096] = {};
			in.read(reinterpret_cast<char *>(rb), sizeof rb);			
			bytesRead = in.gcount();

			if (bytesRead > 0)
				buffer.insert(buffer.end(), rb, rb + bytesRead);
		}
		while (bytesRead > 0);
		
		return buffer;
	}

	template<typename T>
	inline std::istream &readFromStream(std::istream &in, T &val)
	{
		in.read(reinterpret_cast<char *>(&val), sizeof val);
		return in;
	}
	
	inline std::ostream &writeToStream(std::ostream &out, uint16_t val)
	{
		out.put((val >> 0) & 0xFF);
		out.put((val >> 8) & 0xFF);
		return out;
	}
	
	inline std::ostream &writeToStream(std::ostream &out, uint32_t val)
	{
		writeToStream(out, static_cast<uint16_t>((val >>  0) & 0xFFFF));
		writeToStream(out, static_cast<uint16_t>((val >> 16) & 0xFFFF));
		return out;
	}

	inline std::ostream &writeToStream(std::ostream &out, const std::vector<unsigned char> &buffer)
	{
		out.write(reinterpret_cast<const char *>(buffer.data()), buffer.size());
		return out;
	}

	inline std::streampos writeZipArchiveEntryToZip(std::ostream &out, const LemonSqueezy::ZipArchiveEntry &zae)
	{
		std::streampos currentPosition = out.tellp();
		
		// magic
		writeToStream(out, static_cast<uint32_t>(0x04034B50));
		
		// version needed to extract (1.0 => 10 => 0x0A, 2.0 => 20 => 0x14)
		writeToStream(out, static_cast<uint16_t>(0x0014));
		
		// general purpose bit flag
		writeToStream(out, static_cast<uint16_t>(0x0000));
		
		// compression method (0 == store)
		writeToStream(out, static_cast<uint16_t>(zae.getCompressionMethod()));
		
		// time/date 
		writeToStream(out, getdostime(zae.getModifiedTime())); 
		
		// crc
		writeToStream(out, zae.getCRC32());
		
		// compressed size
		writeToStream(out, static_cast<uint32_t>(zae.getSize()));
		
		// uncompressed size
		writeToStream(out, static_cast<uint32_t>(zae.getDecompressedSize()));
		
		// filename length
		const std::string &filename = zae.getFilename();
		writeToStream(out, static_cast<uint16_t>(filename.size()));
		
		// extra header size
		writeToStream(out, static_cast<uint16_t>(0));

		out << filename;

		writeToStream(out, zae.getBuffer());

		return currentPosition;
	}

	void writeCentralDirectoryEntryToZip(std::ostream &out, const LemonSqueezy::ZipArchiveEntry &zae, std::streampos offset)
	{
		// central header magic
		writeToStream(out, static_cast<uint32_t>(0x02014B50));
		
		// version made by (03 == UNIX, 1E == ZIP v3.0)
		writeToStream(out, static_cast<uint16_t>(0x031E));
		
		// version needed to extract (0x0014 == 2.0)
		writeToStream(out, static_cast<uint16_t>(0x0014));
		
		// general purpose flags register
		writeToStream(out, static_cast<uint16_t>(0x0000));
		
		// compression method (0x00 == store)
		writeToStream(out, static_cast<uint16_t>(zae.getCompressionMethod()));
		
		// time/date 
		writeToStream(out, getdostime(zae.getModifiedTime())); 
		
		// crc
		writeToStream(out, zae.getCRC32());
		
		// compressed size
		writeToStream(out, static_cast<uint32_t>(zae.getSize()));
		
		// uncompressed size
		writeToStream(out, static_cast<uint32_t>(zae.getDecompressedSize()));
		
		// filename length
		const std::string &filename = zae.getFilename();
		writeToStream(out, static_cast<uint16_t>(filename.size()));
		
		// extra field length
		writeToStream(out, static_cast<uint16_t>(0x0000));
		
		// file comment length
		writeToStream(out, static_cast<uint16_t>(0x0000));
		
		// disk number start
		writeToStream(out, static_cast<uint16_t>(0x0000));
		
		// internal file attributes
		writeToStream(out, static_cast<uint16_t>(0x0000));
		
		// external file attributes
		if (zae.getBuffer().size() == 0) // directory
			writeToStream(out, static_cast<uint32_t>(0x01FD0000));
		else // file
			writeToStream(out, static_cast<uint32_t>(0x01B60000));
		
		// relative offset of local header
		writeToStream(out, static_cast<uint32_t>(offset));

		// filename
		out << filename;
	}

	std::vector<std::string> getDirectoryNamesFromFilename(const std::string &filename)
	{
		std::vector<std::string> result;
		std::string::size_type start = 0;
		
		while (true)
		{
			std::string::size_type slash = filename.find('/', start);
			if (slash == std::string::npos)
				break;
			
			result.push_back(filename.substr(0, slash + 1));
			start = slash + 1;
		}
		
		return result;
	}

	std::vector<unsigned char> decompressData(const std::vector<unsigned char> &buffer)
	{
		std::stringstream compressed;
		std::stringstream decompressed;
		LemonSqueezy::DeflateCompression deflate;

		writeToStream(compressed, buffer);
		deflate.read(compressed, decompressed);
		return readStreamToBuffer(decompressed);
	}

	std::vector<unsigned char> compressData(std::istream &in, enum LemonSqueezy::ZipArchiveEntry::CompressionMethod cm)
	{
		switch (cm)
		{
			case LemonSqueezy::ZipArchiveEntry::CompressionMethod::kStore:
			{
				return readStreamToBuffer(in);
			}
		
			case LemonSqueezy::ZipArchiveEntry::CompressionMethod::kDeflate:
			{
				LemonSqueezy::DeflateCompression deflate;
				std::stringstream out;
				deflate.write(in, out);
				return readStreamToBuffer(out);
			}

			default:
				throw LemonSqueezy::ZipArchiveException("Invalid compression method");
		}
	}
}

LemonSqueezy::ZipArchiveEntry::ZipArchiveEntry(std::istream &in, enum ZipArchiveEntry::CompressionMethod cm)
  : m_modifiedTime(::time(nullptr)), m_decompressedSize(0), m_compressionMethod(cm)
{
	std::vector<unsigned char> decompressed = readStreamToBuffer(in);
	
	CRC32Checksum checksum;
	checksum.update(decompressed);
	m_crc32 = checksum.getChecksum();
	m_decompressedSize = checksum.getLength();

	in.clear();
	in.seekg(0);
	m_buffer = compressData(in, cm);
}

LemonSqueezy::ZipArchiveEntry::ZipArchiveEntry(const ZipArchiveEntry &other)
  : m_buffer(other.m_buffer), m_filename(other.m_filename),
    m_modifiedTime(other.m_modifiedTime), m_crc32(other.m_crc32),
	m_decompressedSize(other.m_decompressedSize),
	m_compressionMethod(other.m_compressionMethod)
{
}

LemonSqueezy::ZipArchiveEntry::ZipArchiveEntry(ZipArchiveEntry &&other)
  : m_buffer(std::move(other.m_buffer)), m_filename(std::move(other.m_filename)),
    m_modifiedTime(other.m_modifiedTime), m_crc32(other.m_crc32),
	m_decompressedSize(other.m_decompressedSize),
	m_compressionMethod(other.m_compressionMethod)
{
}

LemonSqueezy::ZipArchiveEntry::ZipArchiveEntry(const std::string &filename)
  : m_filename(filename), m_modifiedTime(::time(nullptr)), m_crc32(0), m_decompressedSize(0), m_compressionMethod(kStore)
{
}

LemonSqueezy::ZipArchiveEntry::ZipArchiveEntry(const std::vector<unsigned char> &buffer, enum CompressionMethod cm)
  : m_modifiedTime(::time(nullptr)), m_compressionMethod(cm)
{
	CRC32Checksum checksum;
	checksum.update(buffer);
	m_decompressedSize = buffer.size();
	m_crc32 = checksum.getChecksum();

	std::stringstream decompressed;
	writeToStream(decompressed, buffer);
	m_buffer = compressData(decompressed, cm);
}

LemonSqueezy::ZipArchiveEntry::ZipArchiveEntry()
  : m_modifiedTime(::time(nullptr)), m_decompressedSize(0), m_compressionMethod(kStore)
{
}

LemonSqueezy::ZipArchiveEntry LemonSqueezy::ZipArchiveEntry::CreateFromCompressedData(const std::vector<unsigned char> &data, uint32_t crc32, uint32_t decompressedSize, enum CompressionMethod cm)
{
	LemonSqueezy::ZipArchiveEntry result;
	result.m_buffer = data;
	result.m_crc32 = crc32;
	result.m_compressionMethod = cm;
	result.m_decompressedSize = decompressedSize;
	return result;
}

std::vector<unsigned char> LemonSqueezy::ZipArchiveEntry::decompress() const
{
	if (m_compressionMethod == CompressionMethod::kStore)
		return m_buffer;
	else
		return decompressData(m_buffer);
}

const std::vector<unsigned char> &LemonSqueezy::ZipArchiveEntry::getBuffer() const
{
	return m_buffer;
}

uint16_t LemonSqueezy::ZipArchiveEntry::getCompressionMethod() const
{
	return m_compressionMethod;
}

uint32_t LemonSqueezy::ZipArchiveEntry::getCRC32() const
{
	return m_crc32;
}

uint32_t LemonSqueezy::ZipArchiveEntry::getDecompressedSize() const
{
	return m_decompressedSize;
}

const std::string &LemonSqueezy::ZipArchiveEntry::getFilename() const
{
	return m_filename;
}

std::time_t LemonSqueezy::ZipArchiveEntry::getModifiedTime() const
{
	return m_modifiedTime;
}

size_t LemonSqueezy::ZipArchiveEntry::getSize() const
{
	return m_buffer.size();
}

void LemonSqueezy::ZipArchiveEntry::setFilename(const std::string &filename)
{
	m_filename = filename;
}

void LemonSqueezy::ZipArchiveEntry::setModifiedTime(std::time_t modifiedTime)
{
	m_modifiedTime = modifiedTime;
}

void LemonSqueezy::ZipArchiveEntry::setModifiedTime(LemonSqueezy::ZipArchiveEntry::dostime modifiedTime)
{
	m_modifiedTime = modifiedTime;
}

void LemonSqueezy::ZipArchiveEntry::setSize(size_t size)
{
	m_buffer.resize(size);
}

bool LemonSqueezy::ZipArchiveEntry::operator<(const ZipArchiveEntry &other) const
{
	return m_filename < other.m_filename;
}

LemonSqueezy::ZipArchive::ZipArchive(std::istream &in)
{
	if (htonl(1) == 1)
		throw ZipArchiveException("Big-endian machines currently unimplemented");

	while (in)
	{
		uint32_t magic;
		if (!readFromStream(in, magic))
			throw ZipArchiveException("Unable to read from input stream");
		
		if (magic == 0x02014B50)
			break; // don't bother parsing central directory header

		if (magic != 0x04034B50)
			throw ZipArchiveException("Unknown magic number: " + std::to_string(magic));

		uint16_t versionNeededToExtract;
		uint16_t generalPurposeFlags;
		uint16_t compressionMethod;
		ZipArchiveEntry::dostime timestamp;
		uint32_t crc32;
		uint32_t compressedSize;
		uint32_t decompressedSize;
		uint16_t filenameLength;
		uint16_t extraFieldLength;

		readFromStream(in, versionNeededToExtract);
		readFromStream(in, generalPurposeFlags);
		readFromStream(in, compressionMethod);
		readFromStream(in, timestamp);
		readFromStream(in, crc32);
		readFromStream(in, compressedSize);
		readFromStream(in, decompressedSize);
		readFromStream(in, filenameLength);
		readFromStream(in, extraFieldLength);

		std::vector<char> filename(filenameLength);
		in.read(&filename.at(0), filenameLength);
		in.ignore(extraFieldLength, EOF);
		
		std::vector<unsigned char> compressedData(compressedSize);
		if (compressedSize > 0)
			in.read(reinterpret_cast<char *>(&compressedData.at(0)), compressedData.size());
		
		ZipArchiveEntry zae = ZipArchiveEntry::CreateFromCompressedData(compressedData, crc32, decompressedSize, static_cast<enum ZipArchiveEntry::CompressionMethod>(compressionMethod));
		zae.setFilename(std::string { filename.begin(), filename.end() });
		zae.setModifiedTime(timestamp);

		m_entries.emplace(zae);
	}
}

LemonSqueezy::ZipArchive::ZipArchive()
  : m_defaultCompressionMethod(ZipArchiveEntry::CompressionMethod::kStore)
{
}

void LemonSqueezy::ZipArchive::addEntry(const ZipArchiveEntry &entry)
{
	m_entries.emplace(entry);
}

void LemonSqueezy::ZipArchive::addEntry(const std::string &filename, std::istream &in, enum ZipArchiveEntry::CompressionMethod cm)
{
	if (cm == ZipArchiveEntry::CompressionMethod::kDefault)
		cm = m_defaultCompressionMethod;
	
	ZipArchiveEntry zae { in, cm };
	zae.setFilename(filename);
	m_entries.emplace(zae);
}

void LemonSqueezy::ZipArchive::addEntry(const std::string &filename, const std::vector<unsigned char> &buffer, enum ZipArchiveEntry::CompressionMethod cm)
{
	if (cm == ZipArchiveEntry::CompressionMethod::kDefault)
		cm = m_defaultCompressionMethod;
	
	ZipArchiveEntry zae { buffer, cm };
	zae.setFilename(filename);
	m_entries.emplace(zae);
}

LemonSqueezy::ZipArchive::const_iterator LemonSqueezy::ZipArchive::begin() const
{
	return m_entries.begin();
}

LemonSqueezy::ZipArchive::const_iterator LemonSqueezy::ZipArchive::end() const
{
	return m_entries.end();
}

LemonSqueezy::ZipArchive::const_iterator LemonSqueezy::ZipArchive::find(const std::string &filename) const
{
	if (filename.size() == 0)
		return m_entries.end();
	else if (filename[0] == '/')
		return m_entries.find(filename.substr(1));
	else 
		return m_entries.find(filename);
}

void LemonSqueezy::ZipArchive::setDefaultCompressionMethod(enum ZipArchiveEntry::CompressionMethod cm)
{
	if (cm == ZipArchiveEntry::CompressionMethod::kDefault)
		throw ZipArchiveException("Invalid default compression method");
	m_defaultCompressionMethod = cm;
}

void LemonSqueezy::ZipArchive::write(std::ostream &out) const
{
	std::set<ZipArchiveEntry> entriesToWrite;

	for (const ZipArchiveEntry &zae : m_entries)
	{
		for (const std::string &dirName : getDirectoryNamesFromFilename(zae.getFilename()))
			entriesToWrite.emplace(dirName);

		entriesToWrite.emplace(zae);
	}
	
	std::unordered_map<std::string, std::streampos> offsets;
	offsets.reserve(entriesToWrite.size());
	
	for (const ZipArchiveEntry &zae : entriesToWrite)
		offsets[zae.getFilename()] = writeZipArchiveEntryToZip(out, zae);
	
	std::streampos startOfCentralDirectory = out.tellp();
	
	for (const ZipArchiveEntry &zae : entriesToWrite)
		writeCentralDirectoryEntryToZip(out, zae, offsets[zae.getFilename()]);
	
	std::streampos endOfCentralDirectory = out.tellp();
	
	// End of central directory magic
	writeToStream(out, static_cast<uint32_t>(0x06054B50));

	// number of this disk
	writeToStream(out, static_cast<uint16_t>(0x0000));
	
	// number of the disk with the start of the central directory
	writeToStream(out, static_cast<uint16_t>(0x0000));
	
	// total number of entries in the central directory for this disk
	writeToStream(out, static_cast<uint16_t>(entriesToWrite.size()));
	
	// total number of entries in the central directory
	writeToStream(out, static_cast<uint16_t>(entriesToWrite.size()));
	
	// size of central directory
	writeToStream(out, static_cast<uint32_t>(endOfCentralDirectory - startOfCentralDirectory));
	
	// offset of central directory
	writeToStream(out, static_cast<uint32_t>(startOfCentralDirectory));
	
	// zip file comment length
	writeToStream(out, static_cast<uint16_t>(0x0000));
	
	// no comment
}

