#ifndef LEMONSQUEEZY_ZIP_H
#define LEMONSQUEEZY_ZIP_H

#include <ctime>
#include <iostream>
#include <set>
#include <sstream>
#include <string>
#include <vector>

#include "exceptions.h"

namespace LemonSqueezy
{
	class ZipArchiveEntry
	{
	public:
		typedef uint32_t dostime;
	
		enum CompressionMethod
		{
			kDefault = -1,
			kStore = 0,
			kDeflate = 8
		};

		static ZipArchiveEntry CreateFromCompressedData(const std::vector<unsigned char> &data, uint32_t crc, uint32_t decompressedSize, enum CompressionMethod cm);

		ZipArchiveEntry(std::istream &, enum CompressionMethod cm);
		ZipArchiveEntry(const ZipArchiveEntry &);
		ZipArchiveEntry(ZipArchiveEntry &&);
		ZipArchiveEntry(const std::string &filename);
		ZipArchiveEntry(const std::vector<unsigned char> &buffer, enum CompressionMethod cm);
		ZipArchiveEntry();

		std::vector<unsigned char> decompress() const;
		const std::vector<unsigned char> &getBuffer() const;
		uint16_t getCompressionMethod() const;
		uint32_t getCRC32() const;
		uint32_t getDecompressedSize() const;
		const std::string &getFilename() const;
		std::time_t getModifiedTime() const;
		size_t getSize() const;
		void setFilename(const std::string &name);
		void setModifiedTime(std::time_t t);
		void setModifiedTime(dostime t);
		void setSize(size_t newSize);

		bool operator<(const ZipArchiveEntry &) const;
		
	private:
		std::vector<unsigned char> m_buffer;
		std::string m_filename;
		std::time_t m_modifiedTime;
		uint32_t m_crc32;
		uint32_t m_decompressedSize;
		uint16_t m_compressionMethod;
	};

	class ZipArchive
	{
	public:
		ZipArchive(std::istream &in);
		ZipArchive();

		typedef std::set<ZipArchiveEntry>::const_iterator const_iterator;

		void addEntry(const ZipArchiveEntry &entry);
		void addEntry(const std::string &filename, std::istream &contents, enum ZipArchiveEntry::CompressionMethod cm = ZipArchiveEntry::CompressionMethod::kDefault);
		void addEntry(const std::string &filename, const std::vector<unsigned char> &buffer, enum ZipArchiveEntry::CompressionMethod cm = ZipArchiveEntry::CompressionMethod::kDefault);
		const_iterator begin() const;
		const_iterator end() const;
		const_iterator find(const std::string &filename) const;
		void setDefaultCompressionMethod(enum ZipArchiveEntry::CompressionMethod cm);
		void write(std::ostream &out) const;
		
	private:
		enum ZipArchiveEntry::CompressionMethod m_defaultCompressionMethod;
		std::set<ZipArchiveEntry> m_entries;
	};

	class ZipArchiveException : public Exception
	{
	public:
		ZipArchiveException(const std::string &msg) : Exception(msg) {}
	};
}

#endif

