#include "exceptions.h"
#include "deflate.h"
#include "streambitreader.h"
#include "streambitwriter.h"
#include "zlib.h"

namespace
{
	enum class CompressionMethod
	{
		kDeflate = 8,
		kReserved = 15
	};

	enum class CompressionLevel
	{
		kFastest,
		kFast,
		kDefault,
		kMaximum
	};

	struct ZLibHeader
	{
		enum CompressionMethod compressionMethod = CompressionMethod::kDeflate;
		size_t windowSize = 32768;
		bool presetDictionary = false;
		enum CompressionLevel compressionLevel = CompressionLevel::kDefault;
	};

	/// Not that efficient, but unlikely to matter given it is not used much
	template<typename T, typename = std::enable_if_t<std::is_unsigned<T>::value>>
	constexpr size_t naiveLogBase2(T val)
	{
		size_t count = 0;
		while (val >>= 1)
			count++;
		return count;
	}

	bool readZLibHeader(LemonSqueezy::IBitReader &bitReader, ZLibHeader &header)
	{
		LemonSqueezy::BitSequence bs;

		if (!bitReader.readBitsIntoBitSequence(16, bs))
			return false;
		
		if (bs.code % 31 != 0)
			return false;

		header.compressionMethod = static_cast<enum CompressionMethod>((bs.code & 0x0F00) >> 8);
		if (header.compressionMethod == CompressionMethod::kDeflate)
		{
			header.windowSize = 1U << (((bs.code & 0xF000) >> 12) + 8);
			header.compressionLevel = static_cast<enum CompressionLevel>((bs.code & 0x00C0) >> 6);
			header.presetDictionary = static_cast<bool>(bs.code & 0x0020);
		}

		return true;
	}

	bool writeZLibHeader(std::ostream &out, const ZLibHeader &header)
	{
		LemonSqueezy::StreamBitWriter bitWriter(out);
		LemonSqueezy::BitSequence bs;
		
		if (header.compressionMethod != CompressionMethod::kDeflate)
			return false;
		
		bs.length = 16;

		bs.code |= (naiveLogBase2(header.windowSize) - 8) << 12U;
		bs.code |= 0x0800; // deflate compression method
		bs.code |= (static_cast<unsigned int>(header.compressionLevel) & 0x03) << 6U;
		if (header.presetDictionary)
			bs.code |= 0x0020;
		
		bs.code += 31 - (bs.code % 31);
		
		bitWriter.writeBits(bs);
		return true;
	}
}

LemonSqueezy::ZLibCompression::ZLibCompression()
{
}

LemonSqueezy::ZLibCompression::~ZLibCompression()
{
}

void LemonSqueezy::ZLibCompression::read(std::istream &in, std::ostream &out)
{
	StreamBitReader bitReader(in);

	ZLibHeader header;
	if (!readZLibHeader(bitReader, header))
		throw DecompressionException("Unable to parse ZLib header, it is possibly invalid or not a ZLib stream");
	
	if (header.compressionMethod != CompressionMethod::kDeflate)
		throw DecompressionException("Unable to decompress this ZLib stream");

	if (header.presetDictionary)
	{
		BitSequence presetDictionary;
		if (!bitReader.readBitsIntoBitSequence(32, presetDictionary))
			throw DecompressionException("Unable to read preset dictionary");
		
		throw DecompressionException("Cannot decompress because a dictionary is required");
	}

	DeflateCompression deflate;
	deflate.read(in, out);
}

void LemonSqueezy::ZLibCompression::write(std::istream &in, std::ostream &out)
{
	// write default header
	if (!writeZLibHeader(out, ZLibHeader()))
		throw CompressionException("Unable to write ZLib header");

	DeflateCompression deflate;
	deflate.write(in, out);
}
