#include <libxml/tree.h>
#include "libxmlhelper.h"
#include "document.h"
#include "filepath-utilities.h"

namespace
{
	std::vector<std::string> getPageFilenames(xmlDocPtr fixedDoc)
	{
		std::vector<std::string> result;
		xmlNodePtr root = xmlDocGetRootElement(fixedDoc);
		
		for (xmlNodePtr pageContent = LemonXPS::XML::GetFirstElementChild(root); pageContent != nullptr; pageContent = LemonXPS::XML::GetNextElementSibling(pageContent))
		{
			if (LemonXPS::XML::GetName(pageContent) == "PageContent")
			{
				std::string source;
				if (LemonXPS::XML::TryGetProperty(pageContent, "Source", source))
					result.push_back(source);
			}
		}

		return result;
	}
}

LemonXPS::DocumentPtr LemonXPS::Document::Load(const LemonSqueezy::ZipArchive &zip, const std::string &source)
{
	LemonXPS::DocumentPtr document = std::shared_ptr<Document>(new Document(zip, source));
	LemonSqueezy::ZipArchive::const_iterator fixedDocEntry = zip.find(source);
	if (fixedDocEntry == zip.end())
		throw XPSDocumentException("Unable to find document in archive");
	
	xmlDocPtr fixedDoc = LemonXPS::XML::CreateXMLDocFromZipArchiveEntry(*fixedDocEntry);
	if (fixedDoc == nullptr)
		throw XPSDocumentException("Document was unable to be loaded, the XML could be corrupted");
	
	std::vector<std::string> pageFilenames = getPageFilenames(fixedDoc);
	xmlFreeDoc(fixedDoc);

	std::string root = LemonXPS::Path::GetRoot(source);
	for (const std::string &pageFilename : pageFilenames)
		document->m_pages.emplace_back(Page::Load(document, root, pageFilename));
	
	return document;
}

LemonXPS::Document::Document(const LemonSqueezy::ZipArchive &zip, const std::string &source)
  : m_archive(zip), m_source(source)
{
	
}

LemonXPS::Document::~Document()
{
}

const LemonSqueezy::ZipArchive &LemonXPS::Document::getArchive() const
{
	return m_archive;
}

const std::vector<LemonXPS::PagePtr> &LemonXPS::Document::getPages() const
{
	return m_pages;
}

const std::string &LemonXPS::Document::getSource() const
{
	return m_source;
}

void LemonXPS::Document::render(IPageElementContext &ctx) const
{
	for (PagePtr page : m_pages)
		page->render(ctx);
}
