#include <iostream>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <lemonsqueezy/zip.h>
#include "file.h"
#include "libxmlhelper.h"

namespace
{
	std::vector<std::string> getFixedDocumentSequenceFilenames(xmlDocPtr relationships)
	{
		using namespace LemonXPS;
		
		std::vector<std::string> filenames;
		xmlNodePtr root = xmlDocGetRootElement(relationships);
		
		for (xmlNodePtr relationship = LemonXPS::XML::GetFirstElementChild(root); relationship != nullptr; relationship = LemonXPS::XML::GetNextElementSibling(relationship))
		{
			if (XML::GetName(relationship) != "Relationship")
				continue;
			
			if (XML::GetProperty(relationship, "Type") == "http://schemas.microsoft.com/xps/2005/06/fixedrepresentation")
			{
				std::string target;
				if (XML::TryGetProperty(relationship, "Target", target))
					filenames.emplace_back(target);
			}
		}

		return filenames;
	}

	std::vector<std::string> getFixedDocumentFilenames(xmlDocPtr fixedDocSeq)
	{
		using namespace LemonXPS;

		std::vector<std::string> filenames;
		xmlNodePtr root = xmlDocGetRootElement(fixedDocSeq);
		
		for (xmlNodePtr documentRef = LemonXPS::XML::GetFirstElementChild(root); documentRef != nullptr; documentRef = LemonXPS::XML::GetNextElementSibling(documentRef))
		{
			if (XML::GetName(documentRef) == "DocumentReference")
			{
				std::string source;
				if (XML::TryGetProperty(documentRef, "Source", source))
					filenames.emplace_back(source);
			}
		}

		return filenames;
	}

	LemonXPS::DocumentSequence getDocumentSequenceFromArchive(const LemonSqueezy::ZipArchive &zip, const std::string &fixedDocSeqFilename)
	{
		LemonSqueezy::ZipArchive::const_iterator fixedDocSeqEntry = zip.find(fixedDocSeqFilename);
		if (fixedDocSeqEntry == zip.end())
			throw LemonXPS::XPSFileException("Missing fixed document sequence file " + fixedDocSeqFilename);
		
		LemonXPS::DocumentSequence result;
		xmlDocPtr fixedDocSeq = LemonXPS::XML::CreateXMLDocFromZipArchiveEntry(*fixedDocSeqEntry);
		for (const std::string &filename : getFixedDocumentFilenames(fixedDocSeq))
			result.emplace_back(LemonXPS::Document::Load(zip, filename));

		xmlFreeDoc(fixedDocSeq);

		return result;
	}

	std::vector<LemonXPS::DocumentSequence> getDocumentSequencesFromStream(std::istream &in)
	{
		LemonSqueezy::ZipArchive zip { in };

		LemonSqueezy::ZipArchive::const_iterator relsEntry = zip.find("_rels/.rels");
		if (relsEntry == zip.end())
			throw LemonXPS::XPSFileException("Missing relationship file");
		
		xmlDocPtr relationships = LemonXPS::XML::CreateXMLDocFromZipArchiveEntry(*relsEntry);
		std::vector<std::string> fixedDocumentSequenceFilenames = getFixedDocumentSequenceFilenames(relationships);
		xmlFreeDoc(relationships);

		xmlCleanupParser();
		
		std::vector<LemonXPS::DocumentSequence> documentSequences;
		for (const std::string &fds : fixedDocumentSequenceFilenames)
			documentSequences.emplace_back(getDocumentSequenceFromArchive(zip, fds));

		return documentSequences;
	}
}

LemonXPS::File::File(std::istream &in)
  : m_documentSequences { getDocumentSequencesFromStream(in) }
{

}

LemonXPS::File::~File()
{
}

void LemonXPS::File::addDocumentSequence(const DocumentSequence &docSeq)
{
	m_documentSequences.push_back(docSeq);
}

const std::vector<LemonXPS::DocumentSequence> &LemonXPS::File::getDocumentSequences() const
{
	return m_documentSequences;
}
