#include "document.h"
#include "libxmlhelper.h"
#include "page.h"
#include "filepath-utilities.h"

namespace
{
	void recursePageElements(LemonXPS::PageElementCPtr root, LemonXPS::IPageElementContext &ctx)
	{
		ctx.beginPageElement(root);
		for (LemonXPS::PageElementCPtr e : root->getChildren())
			recursePageElements(e, ctx);
		ctx.endPageElement(root);
	}

	LemonXPS::PageElementPtr getPageElementFromXmlElement(LemonXPS::PagePtr parentPage, xmlNodePtr node)
	{
		std::string name = LemonXPS::XML::GetName(node);
		std::string ns = LemonXPS::XML::GetNamespace(node);
		LemonXPS::Attributes attributes;

		LemonXPS::XML::EnumerateProperties(node, [&attributes](const std::string &name, const std::string &value){
			attributes.emplace(name, value);
		});

		return std::make_shared<LemonXPS::PageElement>(parentPage, name, ns, attributes);
	}

	LemonXPS::PageElementPtr generatePageElementTree(LemonXPS::PagePtr parentPage, xmlNodePtr node)
	{
		LemonXPS::PageElementPtr result = getPageElementFromXmlElement(parentPage, node);

		for (xmlNodePtr child = LemonXPS::XML::GetFirstElementChild(node); child != nullptr; child = LemonXPS::XML::GetNextElementSibling(child))
			result->addChild(generatePageElementTree(parentPage, child));
		
		return result;
	}
}

LemonXPS::PagePtr LemonXPS::Page::Load(LemonXPS::DocumentPtr document, const std::string &root, const std::string &source)
{
	PagePtr page = std::shared_ptr<Page>(new Page(document, root, source));
	
	std::string fullPath = LemonXPS::Path::Append(root, source);
	const LemonSqueezy::ZipArchive &archive = document->getArchive();
	LemonSqueezy::ZipArchive::const_iterator pageEntry = archive.find(fullPath);
	if (pageEntry == archive.end())
		throw XPSPageException("Unable to find page in archive");
	
	xmlDocPtr fixedPageDoc = LemonXPS::XML::CreateXMLDocFromZipArchiveEntry(*pageEntry);
	if (fixedPageDoc == nullptr)
		throw XPSPageException("Unable to load page, the XML could be corrupt");
	
	page->m_rootElement = generatePageElementTree(page, xmlDocGetRootElement(fixedPageDoc));
	xmlFreeDoc(fixedPageDoc);

	return page;
}

LemonXPS::Page::Page(LemonXPS::DocumentPtr document, const std::string &root, const std::string &source)
  : m_document(document), m_root(root), m_source(source)
{
	
}

LemonXPS::Page::~Page()
{
}

std::weak_ptr<LemonXPS::Document> LemonXPS::Page::getDocument() const
{
	return m_document;
}

double LemonXPS::Page::getHeight() const
{
	return std::stod(m_rootElement->getAttribute("Height"));
}

LemonXPS::PageElementPtr LemonXPS::Page::getRoot()
{
	return m_rootElement;
}

double LemonXPS::Page::getWidth() const
{
	return std::stod(m_rootElement->getAttribute("Width"));
}

void LemonXPS::Page::render(IPageElementContext &ctx) const
{
	recursePageElements(m_rootElement, ctx);
}
