#ifndef STRIPEDHORSE_ENDIAN_H
#define STRIPEDHORSE_ENDIAN_H

#include <cstdlib>

namespace StripedHorse
{
	enum Endian
	{
		kLittleEndian,
		kBigEndian
	};
	
	Endian GetHostEndianness();
	
	/**
	 Swaps bytes regardless of host or source endian
	 @param[in] in the value to swap
	 @return the value with its bytes in reverse order
	 */
	template<typename T>
	T SwapBytes(const T &in)
	{
		union
		{
			T val;
			unsigned char buf[sizeof(T)]; 
		} original, swapped;
		original.val = in;
		for (size_t i = 0; i < sizeof(T); i++)
			swapped.buf[i] = original.buf[sizeof(T) - i - 1];
		return swapped.val;
	}
	
	/**
	 Converts the value to big endian. A no-op on big-endian architectures.
	 @param[in] v the value to convert to big endian
	 @return on little endian architectures, the bytes are swapped, otherwise
	         the value is return unchanged.
	 */
	template<typename T>
	T ToBigEndian(const T &v)
	{
		if (GetHostEndianness() == kBigEndian)
			return v;
		else
			return SwapBytes(v);
	}
	
	/**
	 Converts the value to little endian. A no-op on little-endian
	 architectures.
	 @param[in] v the value to convert to little endian
	 @return on big endian architectures, the bytes are swapped, otherwise the
	         input is returned unchanged.
	 */
	template<typename T>
	T ToLittleEndian(const T &v)
	{
		if (GetHostEndianness() == kLittleEndian)
			return v;
		else
			return SwapBytes(v);
	}
	
	/**
	 Convers a value from big endian format to the host's endian format
	 @param[in] v the value to convert from big endian
	 @return if the host's architecture is also big endian, the value is
	         returned unchanged, otherwise the bytes are swapped.
	 */
	template<typename T>
	T FromBigEndian(const T &v)
	{
		if (GetHostEndianness() == kLittleEndian)
			return SwapBytes(v);
		else
			return v;
	}
	
	/**
	 Converts a value from little endian format to the host's endian format
	 @param[in] v the value to convert from little endian format
	 @return if the host architecture is also little endian, the value is
	         returned unchanged, otherwise the bytes are swapped.
	 */
	template<typename T>
	T FromLittleEndian(const T &v)
	{
		if (GetHostEndianness() == kBigEndian)
			return SwapBytes(v);
		else
			return v;
	}
	
	/**
	 Converts a value from a specified endian format
	 @param[in] e the source endian format for the value
	 @param[in] v the value to convert
	 @return if the specified endian format matches the host architecture's
	         endian format, then the value is returned unchanged, otherwise
	         the bytes are swapped.
	 */
	template<typename T>
	T FromEndian(enum Endian e, const T &v)
	{
		if (e == kBigEndian)
			return FromBigEndian(v);
		else
			return FromLittleEndian(v);
	}
}

#endif

