#ifndef STRIPEDHORSE_TYPES_H
#define STRIPEDHORSE_TYPES_H 1

#include <cmath>
#include <iostream>
#include <map>
#include <string>
#include <unordered_map>
#include <vector>

#include "exceptions.h"

namespace StripedHorse
{
	typedef std::vector<unsigned char> Buffer;
	
	typedef std::unordered_map<std::string, std::string> Dictionary;
	
	template<typename T>
	struct BasicPoint
	{
		using type = T;

		T x;
		T y;
		
		BasicPoint(T x = T(), T y = T()) : x(x), y(y) {}

		template<typename U>
		explicit operator BasicPoint<U>() const
		{
			return BasicPoint<U>{
				static_cast<U>(x),
				static_cast<U>(y)
			};
		}

		bool operator==(const BasicPoint<T> &other) const
		{
			return x == other.x && y == other.y;
		}

		bool operator!=(const BasicPoint<T> &other) const
		{
			return x != other.x || y != other.y;
		}
	};
	
	template<typename T>
	inline std::ostream &operator<<(std::ostream &o, const BasicPoint<T> &p)
	{
		return o << "(" << p.x << ", " << p.y << ")";
	}

	using PointF = BasicPoint<double>;
	using Point = BasicPoint<long>;
	using PointU = BasicPoint<unsigned long>;
	
	template<typename T>
	struct BasicRectangle
	{
		using type = T;

		BasicPoint<T> lowerLeft;
		BasicPoint<T> upperRight;
		
		BasicRectangle(T l = T(), T b = T(), T r = T(), T t = T()) : lowerLeft(l, b), upperRight(r, t) {}
		BasicRectangle(const BasicPoint<T> &ll, const BasicPoint<T> &ur) : lowerLeft(ll), upperRight(ur) {}

		bool contains(const BasicPoint<T> &p) const
		{
			return lowerLeft.x <= p.x && lowerLeft.y <= p.y && upperRight.x >= p.x && upperRight.y >= p.y;
		}

		T getHeight() const
		{
			return upperRight.y - lowerLeft.y;
		}

		T getWidth() const
		{
			return upperRight.x - lowerLeft.x;
		}

		template<typename U>
		explicit operator BasicRectangle<U>() const
		{
			return BasicRectangle<U>{
				static_cast<BasicPoint<U>>(lowerLeft),
				static_cast<BasicPoint<U>>(upperRight)
			};
		}

		bool operator==(const BasicRectangle<T> &other) const
		{
			return lowerLeft == other.lowerLeft && upperRight == other.upperRight;
		}

		bool operator!=(const BasicRectangle<T> &other) const
		{
			return lowerLeft != other.lowerLeft || upperRight != other.upperRight;
		}
	};

	using RectangleF = BasicRectangle<double>;
	using Rectangle = BasicRectangle<long>;
	using RectangleU = BasicRectangle<unsigned long>;
	
	template<typename T>
	inline std::ostream &operator<<(std::ostream &o, const BasicRectangle<T> &r)
	{
		return o << "(" << r.lowerLeft.x << ", " << r.lowerLeft.y << "), (" << r.upperRight.x << ", " << r.upperRight.y << ")";
	}

	template<typename T>
	struct BasicTransformMatrix
	{
		using type = T;

		BasicTransformMatrix()
		  : BasicTransformMatrix(1, 0, 0, 1, 0, 0)
		{
		}

		BasicTransformMatrix(const BasicTransformMatrix &other)
		{
			this->operator=(other);
		}

		BasicTransformMatrix(BasicTransformMatrix &&other)
		{
			this->operator=(other);
		}

		BasicTransformMatrix(T a, T b, T c, T d, T e, T f)
		  : m { { a, c, e }, { b, d, f }, { 0, 0, 1 } }
		{
		}
		
		BasicTransformMatrix<T> &operator=(const BasicTransformMatrix<T> &other)
		{
			for (size_t i = 0; i < 3; i++)
				for (size_t j = 0; j < 3; j++)
					m[i][j] = other.m[i][j];
			return *this;
		}

		BasicTransformMatrix<T> &operator*=(const BasicTransformMatrix<T> &other)
		{
			return this->operator=(this->operator*(other));
		}

		BasicTransformMatrix<T> operator*(const BasicTransformMatrix<T> &other)
		{
			BasicTransformMatrix<T> result;
	
			for (size_t i = 0; i < 3; i++)
				for (size_t j = 0; j < 3; j++)
					result.m[i][j] = m[i][0] * other.m[0][j] + m[i][1] * other.m[1][j] + m[i][2] * other.m[2][j];
			
			return result;
		}
		
		BasicTransformMatrix<T> &concat(const BasicTransformMatrix<T> &other)
		{
			this->operator*=(other);
			return *this;
		}

		BasicTransformMatrix<T> &rotate(double rads)
		{
			return this->operator*=(BasicTransformMatrix<T> { cos(rads), sin(rads), -sin(rads), cos(rads), T(), T() });
		}

		BasicTransformMatrix<T> &scale(T x, T y)
		{
			return this->operator*=(BasicTransformMatrix<T>(x, T(), T(), y, T(), T()));
		}

		BasicTransformMatrix<T> &translate(T x, T y)
		{
			return this->operator*=(BasicTransformMatrix<T> { static_cast<T>(1), T(), T(), static_cast<T>(1), x ,y });
		}

		T m[3][3];
	};

	using TransformMatrix = BasicTransformMatrix<double>;
	
	template<typename T>
	inline std::ostream &operator<<(std::ostream &o, const BasicTransformMatrix<T> &tm)
	{
		return o << "(" << tm.m[0][0] << ", " << tm.m[0][1] << ", " << tm.m[0][2] << "), "
		         << "(" << tm.m[1][0] << ", " << tm.m[1][1] << ", " << tm.m[1][2] << "), "
		         << "(" << tm.m[2][0] << ", " << tm.m[2][1] << ", " << tm.m[2][2] << ")";
	}

	template<typename T>
	BasicPoint<T> operator*(const BasicTransformMatrix<T> &matrix, const BasicPoint<T> &p)
	{
		return BasicPoint<T> {
			matrix.m[0][0] * p.x + matrix.m[0][1] * p.y + matrix.m[0][2],
  			matrix.m[1][0] * p.x + matrix.m[1][1] * p.y + matrix.m[1][2]
		};
	}

	template<typename T>
	BasicRectangle<T> operator*(const BasicTransformMatrix<T> &matrix, const BasicRectangle<T> &r)
	{
		return BasicRectangle<T> {
			matrix * r.lowerLeft,
			matrix * r.upperRight
		};
	}
}

#endif

