#include <cctype>
#include <iomanip>
#include <regex>
#include <sstream>
#include <unordered_set>
#include "uri.h"

namespace StripedHorse
{
	class URIDecodeException : public Exception
	{
	public:
		URIDecodeException(const std::string &msg) : Exception(msg) {}
	};
}

namespace
{
	[[noreturn]] void invalidDecode()
	{
		throw StripedHorse::URIDecodeException("Invalid or incomplete escape sequence");
	}

	std::string uriDecode(const std::string &encoded, bool isFormData)
	{
		std::ostringstream decoded;

		for (std::string::size_type i = 0; i < encoded.size(); i++)
		{
			if (encoded[i] == '%')
			{
				if (i + 2 >= encoded.size())
					invalidDecode();

				char xdigits[3] = { encoded[i + 1], encoded[i + 2] };
				char *endptr = nullptr;
				unsigned long value = std::strtoul(xdigits, &endptr, 16);
				if (!endptr || *endptr)
					invalidDecode();

				decoded.put(static_cast<char>(value));
				i += 2;
			}
			else if (encoded[i] == '+' && isFormData)
			{
				decoded.put(' ');
			}
			else
			{
				decoded.put(encoded[i]);
			}
		}

		return decoded.str();
	}

	std::string uriEncode(const std::string &unencoded)
	{
		// the following characters are reserved re RFC3986
		static const std::unordered_set<char> reserved {
			'!', '*', '\'', '{', '}', ';', ':', '@', '&', '=',
			'+', '$', ',', '/' , '?', '#', '[', ']' ,'%'
		};
		
		std::ostringstream encoded;
		encoded.fill('0');
		encoded.setf(std::ios_base::hex, std::ios_base::basefield);
		for (const char &ch : unencoded)
		{
			if (!::isprint(ch) || ::isspace(ch) || (reserved.find(ch) != reserved.end()))
			{
				unsigned char value = static_cast<unsigned char>(ch);
				encoded.put('%');
				encoded << std::setw(2) << static_cast<unsigned>(value);
			}
			else
			{
				encoded.put(ch);
			}
		}

		return encoded.str();
	}

	StripedHorse::URI::Authority parseAuthority(const std::string &authorityString, long defaultPort)
	{
		StripedHorse::URI::Authority result;

		std::string::size_type hostBegin = 0;
		std::string::size_type at = authorityString.find_first_of('@');
		std::string::size_type colon = authorityString.find_first_of(':');
		
		if (at != std::string::npos)
		{
			hostBegin = at + 1;
			
			if (colon > at)
			{
				result.username = uriDecode(authorityString.substr(0, at), false);
			}
			else
			{
				result.username = uriDecode(authorityString.substr(0, colon), false);
				result.password = uriDecode(authorityString.substr(colon + 1, at - colon - 1), false);
				colon = authorityString.find_first_of(':', at);
			}
		}

		if (colon != std::string::npos)
		{
			result.host = authorityString.substr(hostBegin, colon - hostBegin);
			result.port = std::stoul(authorityString.substr(colon + 1));
		}
		else
		{
			result.host = authorityString.substr(hostBegin);
			result.port = defaultPort;
		}

		return result;
	}

	std::vector<std::string> parsePath(const std::string &pathString, char delim)
	{
		std::vector<std::string> pathComponents;
		
		std::string::size_type prevDelim = 0;
		std::string::size_type nextDelim = pathString.find(delim);
		while (nextDelim != std::string::npos)
		{
			pathComponents.push_back(uriDecode(pathString.substr(prevDelim, nextDelim - prevDelim), false));
			prevDelim = nextDelim + 1;
			nextDelim = pathString.find(delim, prevDelim);
		}

		pathComponents.push_back(uriDecode(pathString.substr(prevDelim), false));
		return pathComponents;
	}

	std::vector<StripedHorse::URI::QueryValue> parseQueryValues(const std::string &queryString)
	{
		std::vector<StripedHorse::URI::QueryValue> result;

		std::istringstream queryStream(queryString);
		std::string queryComponent;
		while (std::getline(queryStream, queryComponent, '&'))
		{
			StripedHorse::URI::QueryValue qv;

			std::string::size_type equals = queryComponent.find('=');
			if (equals != std::string::npos)
			{
				qv.key = uriDecode(queryComponent.substr(0, equals), true);
				qv.value = uriDecode(queryComponent.substr(equals + 1), true);
			}
			else
			{
				qv.key = queryComponent;
			}

			result.emplace_back(std::move(qv));
		}

		return result;
	}
}

bool StripedHorse::URI::QueryValue::operator==(const URI::QueryValue &other) const
{
	return key == other.key && value == other.value;
}

StripedHorse::URI::Authority::Authority()
  : port(-1)
{
}

bool StripedHorse::URI::Authority::operator==(const URI::Authority &other) const
{
	return username == other.username && password == other.password && host == other.host && port == other.port;
}

bool StripedHorse::URI::Authority::isEmpty() const
{
	return host.empty() && username.empty() && password.empty();
}

StripedHorse::URI::URI()
{
}

StripedHorse::URI::URI(const std::string &uriString, long defaultPort)
{
	std::string::size_type colon = uriString.find(':');

	if (colon == std::string::npos)
		throw URIException("Invalid URI: " + uriString);

	m_scheme = uriString.substr(0, colon);

	std::string::size_type fragmentBegin = uriString.rfind('#');
	if (fragmentBegin != std::string::npos)
		m_fragment = uriDecode(uriString.substr(fragmentBegin + 1), false);
	else
		fragmentBegin = uriString.size();
	
	std::string leftover = uriString.substr(colon + 1, fragmentBegin - (colon + 1));

	if (leftover.substr(0, 2) == "//")
	{
		std::string::size_type pathBegin = leftover.find_first_of('/', 2);

		if (pathBegin != std::string::npos)
		{
			m_authority = parseAuthority(leftover.substr(2, pathBegin - 2), defaultPort);
			std::string::size_type queryBegin = leftover.find_first_of('?', pathBegin);
			if (queryBegin != std::string::npos)
			{
				m_path = parsePath(leftover.substr(pathBegin + 1, queryBegin - pathBegin - 1), '/');
				m_query = parseQueryValues(leftover.substr(queryBegin + 1));
			}
			else
			{
				m_path = parsePath(leftover.substr(pathBegin + 1), '/');
			}
		}
		else
		{
			m_authority = parseAuthority(leftover.substr(2), defaultPort);
		}
	}
	else
	{
		m_path = parsePath(leftover, ':');
	}
}

StripedHorse::URI::URI(const URI &other)
{
	(void) this->operator=(other);
}

StripedHorse::URI::URI(URI &&other)
  : m_scheme(std::move(other.m_scheme)), m_authority(std::move(other.m_authority)),
    m_path(std::move(other.m_path)), m_query(std::move(other.m_query)),
	m_fragment(std::move(other.m_fragment))
{
}

StripedHorse::URI::operator std::string() const
{
	std::ostringstream result;
	result << *this;
	return result.str();
}

StripedHorse::URI &StripedHorse::URI::operator=(const URI &other)
{
	m_scheme = other.m_scheme;
	m_authority = other.m_authority;
	m_path = other.m_path;
	m_query = other.m_query;
	m_fragment = other.m_fragment;
	return *this;
}

bool StripedHorse::URI::operator==(const URI &other) const
{
	return m_scheme == other.m_scheme && m_authority == other.m_authority && m_path == other.m_path && m_query == other.m_query && m_fragment == other.m_fragment;
}

void StripedHorse::URI::appendPathComponent(const std::string &pathComponent)
{
	m_path.push_back(pathComponent);
}

void StripedHorse::URI::clear()
{
	m_scheme.clear();
	m_authority.username.clear();
	m_authority.password.clear();
	m_authority.host.clear();
	m_authority.port = -1;
	m_path.clear();
	m_query.clear();
	m_fragment.clear();
}

const std::string &StripedHorse::URI::getFragment() const
{
	return m_fragment;
}

const std::string &StripedHorse::URI::getScheme() const
{
	return m_scheme;
}

const std::string &StripedHorse::URI::getHost() const
{
	return m_authority.host;
}

long StripedHorse::URI::getPort() const
{
	return m_authority.port;
}

std::string StripedHorse::URI::getPath() const
{
	if (m_path.size() == 0)
		return m_authority.isEmpty() ? "" : "/";

	std::ostringstream fullPath;

	if (m_authority.isEmpty())
	{
		for (std::vector<std::string>::size_type i = 0; i + 1 < m_path.size(); i++)
			fullPath << uriEncode(m_path[i]) << ':';
		fullPath << m_path.back();
	}
	else
	{
		fullPath << '/';
		for (std::vector<std::string>::size_type i = 0; i + 1 < m_path.size(); i++)
			fullPath << uriEncode(m_path[i]) << '/';
		fullPath << m_path.back();
	}

	return fullPath.str();
}

const std::vector<std::string> &StripedHorse::URI::getPathComponents() const
{
	return m_path;
}

bool StripedHorse::URI::isEmpty() const
{
	return m_scheme.empty() && m_authority.isEmpty() && m_path.empty() && m_query.empty() && m_fragment.empty();
}

void StripedHorse::URI::setHost(const std::string &host)
{
	m_authority.host = host;
}

void StripedHorse::URI::setPort(long port)
{
	m_authority.port = port;
}

void StripedHorse::URI::setScheme(const std::string &scheme)
{
	m_scheme = scheme;
}

std::ostream &StripedHorse::operator<<(std::ostream &out, const StripedHorse::URI &uri)
{
	if (uri.m_scheme.empty())
		return out << "<Invalid URI>";

	out << uri.m_scheme << ":";
	if (!uri.m_authority.isEmpty())
	{
		out << "//";
		if (!uri.m_authority.username.empty())
		{
			out << uriEncode(uri.m_authority.username);
			if (!uri.m_authority.password.empty())
			{
				out << ":";
				out << uriEncode(uri.m_authority.password);
			}
			out << "@";
		}

		out << uri.m_authority.host;
		
		if (uri.m_authority.port >= 0)
			out << ":" << std::to_string(uri.m_authority.port);
	}

	// construct path
	out << uri.getPath();

	if (uri.m_query.size() > 0)
	{
		out << "?";
		for (size_t i = 0; i < uri.m_query.size() - 1; i++)
			out << uri.m_query[i] << '&';
		out << uri.m_query.back();
	}
	
	if (!uri.m_fragment.empty())
		out << "#" << uriEncode(uri.m_fragment);

	return out;
}

std::ostream &StripedHorse::operator<<(std::ostream &out, const StripedHorse::URI::QueryValue &qv)
{
	out << uriEncode(qv.key);
	if (!qv.value.empty())
		out << "=" << uriEncode(qv.value);
	return out;
}
