#ifndef STRIPEDHORSE_URI_H
#define STRIPEDHORSE_URI_H

#include <iostream>
#include <string>
#include <vector>
#include "exceptions.h"

namespace StripedHorse
{
	/**
	 A URI class to make it easier to deal with the various components of a URI
	 */
	class URI
	{
	public:
		/**
		 A struct representing the key/value pairs of the
		 query part of the URI
		*/
		struct QueryValue
		{
			std::string key;
			std::string value;

			bool operator==(const QueryValue &other) const;
		};

		/**
		 A struct that represents the authority part of a
		 URI (not all URIs have an authority)
		*/
		struct Authority
		{
			std::string username;
			std::string password;
			std::string host;
			long port;

			Authority();
			bool operator==(const Authority &other) const;
			bool isEmpty() const;
		};

		/**
		 Constructs an empty URI
		*/
		URI();

		/**
		 Constructs a URI object
		 @param[in] uriString a string representing the URI
		 */
		URI(const std::string &uriString, long defaultPort = -1);
		
		/**
		 Copies an existing URL object
		 */
		URI(const URI &);
		
		/**
		 Moves an existing URL object
		 */
		URI(URI &&);

		/**
		 Converts the URI to a string
		 */
		explicit operator std::string() const;
		
		/**
		 Copies an existing URL object
		 */
		URI &operator=(const URI &);

		/**
		 Compares two URIs for equality
		 */
		bool operator==(const URI &other) const;

		/**
		 Adds a path component.
		 */
		void appendPathComponent(const std::string &pathComponent);

		/**
		 Clears all the URI components
		 */
		void clear();

		/**
		 Gets the fragment of the URI (the bit following
		 the '#' character)
		 */
		const std::string &getFragment() const;
		
		/**
		 Gets the host portion of the URL
		 @return the host portion of the URL
		 */
		const std::string &getHost() const;

		/**
		 Gets the port of the URL
		 @return the port parsed from the URL. If no port was present, the
		         result is the default port specified during construction. If
		         no port was specified during construction, the result is -1.
		 */
		long getPort() const;
		
		/**
		 Gets the path component of the URL by first encoding and then
		 concatenating each component in the path
		 @return the path component of the URL (including the leading slash).
		 */
		std::string getPath() const;

		/**
		 Gets the decoded components of the path as a vector
		 */
		const std::vector<std::string> &getPathComponents() const;
		
		/**
		 Gets the scheme of the URL
		 @return the scheme of the URL (e.g. "http")
		 */
		const std::string &getScheme() const;
		
		/**
		 Determines whether the URL is empty (has no components)
		 */
		bool isEmpty() const;

		/**
		 Changes the host of the URI's authority
		 */
		void setHost(const std::string &host);

		/**
		 Changes the port of the URI's authority.
		 If the port is negative, it will not be considered significant
		 and will not be written in the output when converted to a
		 string
		 */
		void setPort(long);

		/**
		 Changes the scheme of the URI
		 */
		void setScheme(const std::string &scheme);
		
	private:
		std::string m_scheme;
		Authority m_authority;
		std::vector<std::string> m_path;
		std::vector<QueryValue> m_query;
		std::string m_fragment;
	
	friend std::ostream &operator<<(std::ostream &, const URI &);
	};
	
	std::ostream &operator<<(std::ostream &, const URI &);
	std::ostream &operator<<(std::ostream &, const URI::QueryValue &);
	
	class URIException : public Exception
	{
	public:
		URIException(const std::string &msg) : Exception(msg) {}
	};
}

#endif