#include <algorithm>
#include <cmath>
#include <iomanip>
#include <regex>
#include <sstream>
#include "commands.h"

namespace
{
	template<typename T>
	std::string formatInteger(const std::string &format, const T &integer)
	{
		std::string::size_type slash = format.find('/');
		if (slash != std::string::npos)
			throw StripedHorse::CommandException("Unsupported format");
		
		if (format.size() == 0)
			return "";
		
		std::ostringstream ss;

		if (format.back() == '*')
		{
			// do binary output
			for (int i = sizeof integer - 1; i >= 0; i--)
			{
				char ch = (integer >> (i * 8)) & 0xFF;
				ss.put(ch);
			}
			
			return ss.str().substr(sizeof integer - (format.size() - 1));
		}
		else if (format.back() == '$')
		{
			// do hexadecimal output
			ss.width(format.size() - 1);
			ss << std::setfill('0') << std::hex << std::uppercase << integer;
			return ss.str().substr(0, format.size() - 1);
		}
		else
		{
			// do decimal output
			if (format[0] == '+')
				ss << std::internal << std::showpos;
			
			ss.width(format.size());
			ss << std::setfill('0') << integer;
			return ss.str().substr(0, format.size());
		}
	}
	
	size_t getMaxLengthFromFormat(const std::string &format)
	{
		std::string::size_type slash = format.find('/');
		
		if (slash != std::string::npos)
			return std::stoull(format.substr(slash + 1));
		else
			return format.size();
	}

	std::regex g_cmdRegex("\\+?(([a-z])\\2*)(\\*|\\$|/[0-9]+)?");
}

bool StripedHorse::TPCLCommands::WriteCommandArgs(std::ostream &out, const std::string &format, const std::vector<ArgumentPtr> &args)
{
	auto cmdArgsStart = std::sregex_iterator(format.begin(), format.end(), g_cmdRegex);
	auto cmdArgsEnd = std::sregex_iterator();
	
	size_t argCount = std::distance(cmdArgsStart, cmdArgsEnd);
	if (args.size() != argCount)
		throw CommandException("Wrong number of arguments for command: " + format);
	
	auto cmdArg = cmdArgsStart;
	size_t last = 0;
	for (size_t i = 0; i < argCount; i++)
	{
		std::smatch m = *(cmdArg++);
		out << format.substr(last, m.position() - last);
		if (!args[i]->write(out, m[0].str()))
			break;
		last = m.position() + m.length();
	}
	
	out << format.substr(last);
	
	return static_cast<bool>(out);
}

bool StripedHorse::TPCLCommands::IntegerArgument::write(std::ostream &out, const std::string &format) const
{
	return static_cast<bool>(out << formatInteger(format, m_int));
}

bool StripedHorse::TPCLCommands::USmallIntegerArgument::write(std::ostream &out, const std::string &format) const
{
	return static_cast<bool>(out << formatInteger(format, m_usmallint));
}

bool StripedHorse::TPCLCommands::UIntegerArgument::write(std::ostream &out, const std::string &format) const
{
	return static_cast<bool>(out << formatInteger(format, m_uint));
}

bool StripedHorse::TPCLCommands::ULargeIntegerArgument::write(std::ostream &out, const std::string &format) const
{
	return static_cast<bool>(out << formatInteger(format, m_ulargeint));
}

bool StripedHorse::TPCLCommands::NegativeZeroArgument::write(std::ostream &out, const std::string &format) const
{
	if (format.size() == 0)
		return false;

	// we take one away from this because we always print a negative symbol
	// regardless of whether this format is prefixed with "+" or not (because
	// in the case of negative zero it makes no difference, it is always
	// negative).
	size_t numberOfZeroes = format.size() - 1;

	// ignore binary marker (because it makes no sense here) and ignore
	// hexadecimal marker because we are printing zero
	if (numberOfZeroes > 1 && (format.back() == '*' || format.back() == '$'))
		numberOfZeroes--;

	out.put('-');
	for (size_t i = 0; i < numberOfZeroes; i++)
		out.put('0');
	
	return static_cast<bool>(out);
}

bool StripedHorse::TPCLCommands::StringArgument::write(std::ostream &out, const std::string &format) const
{
	size_t maxLen = getMaxLengthFromFormat(format);
	
	if (maxLen == 0)
		return static_cast<bool>(out << m_str);
	else if (m_str.size())
		return static_cast<bool>(out << m_str.substr(0, maxLen));
	else
		return true;
}

bool StripedHorse::TPCLCommands::BufferArgument::write(std::ostream &out, const std::string &format) const
{
	if (m_buffer.size() == 0)
	{
		return true;
	}
	else
	{
		size_t maxLen = getMaxLengthFromFormat(format);
		if (maxLen == 0 || maxLen > m_buffer.size())
			maxLen = m_buffer.size();
		return static_cast<bool>(out.write(reinterpret_cast<const char *>(&m_buffer[0]), maxLen));
	}
}

bool StripedHorse::TPCLCommands::CharacterArgument::write(std::ostream &out, const std::string &fmt) const
{
	(void) fmt;
	return static_cast<bool>(out.put(m_char));
}

