#ifndef STRIPEDHORSE_COMMANDS_H
#define STRIPEDHORSE_COMMANDS_H 1

#include <iostream>
#include <memory>
#include <string>
#include <vector>
#include <libcore/exceptions.h>
#include <libcore/types.h>

namespace StripedHorse
{
	/**
	 The functions in this namespace provide a way to use the TPCL command
	 format found in various TOSHIBA TEC barcode printer "external equipment
	 interface" manuals.
	 
	 For example, the "label size set" command is shown in the manual as
	 `Daaaa,bbbb,cccc` (ignoring optional arguments and escape characters).
	 Therefore, one can use the following in code:
	 
	 \code{cpp}
	 WriteCommand("Daaaa,bbbb,cccc", 220, 400, 200);
	 \endcode
	*/
	namespace TPCLCommands
	{
		class NegativeZero {};

		/**
		 An abstract class representing an argument of no particular type
		 */
		class Argument
		{
		public:
			/**
			 Serialises the argument value to the output stream based on the
			 supplied format
			 @param[in] o the output stream to write to
			 @param[in] fmt the format to use
			 @return true if the argument value was successfully serialised
			*/
			virtual bool write(std::ostream &o, const std::string &fmt) const = 0;
			virtual ~Argument() {};
			
		protected:
			Argument() {};
		};
		
		/**
		 A class representing a 32-bit integer argument
		 */
		class IntegerArgument : public Argument
		{
		public:
			IntegerArgument(int32_t i) : m_int(i) {}
			virtual bool write(std::ostream &, const std::string &fmt) const;
			virtual ~IntegerArgument() {}
			
		private:
			int32_t m_int;
		};
		
		/**
		 A class representing an 8-bit unsigned integer argument
		 */
		class USmallIntegerArgument : public Argument
		{
		public:
			USmallIntegerArgument(uint8_t i) : m_usmallint(i) {}
			virtual bool write(std::ostream &, const std::string &fmt) const;
			virtual ~USmallIntegerArgument() {}
		
		private:
			uint8_t m_usmallint;
		};
		
		/**
		 A class representing a 32-bit unsigned integer argument
		 */
		class UIntegerArgument : public Argument
		{
		public:
			UIntegerArgument(uint32_t i) : m_uint(i) {}
			virtual bool write(std::ostream &, const std::string &fmt) const;
			virtual ~UIntegerArgument() {}
			
		private:
			uint32_t m_uint;
		};
		
		/**
		 A class representing a 64-bit unsigned integer argument
		 */
		class ULargeIntegerArgument : public Argument
		{
		public:
			ULargeIntegerArgument(uint64_t i) : m_ulargeint(i) {}
			virtual bool write(std::ostream &, const std::string &fmt) const;
			virtual ~ULargeIntegerArgument() {}
			
		private:
			uint64_t m_ulargeint;
		};

		/**
		 A class representing string argument
		 */
		class StringArgument : public Argument
		{
		public:
			StringArgument(const std::string &s) : m_str(s) {}
			virtual bool write(std::ostream &, const std::string &fmt) const;
			virtual ~StringArgument() {}
			
		private:
			std::string m_str;
		};
		
		/**
		 A class representing a buffer argument
		 */
		class BufferArgument : public Argument
		{
		public:
			BufferArgument(const Buffer &b) : m_buffer(b) {}
			virtual bool write(std::ostream &, const std::string &fmt) const;
			virtual ~BufferArgument() {}
		
		private:
			Buffer m_buffer;
		};
		
		/**
		 A class representing a char argument
		 */
		class CharacterArgument : public Argument
		{
		public:
			CharacterArgument(char c) : m_char(c) {}
			virtual bool write(std::ostream &out, const std::string &fmt) const;
			virtual ~CharacterArgument() {}
		private:
			char m_char;
		};

		/**
		 A class representing negative zero argument
		*/
		class NegativeZeroArgument : public Argument
		{
		public:
			NegativeZeroArgument() {}
			virtual bool write(std::ostream &out, const std::string &fmt) const;
			virtual ~NegativeZeroArgument() {}
		};
		
		typedef std::shared_ptr<Argument> ArgumentPtr;
		
		template<typename T>
		inline ArgumentPtr MakeArgument(T t);

		template<>
		inline ArgumentPtr MakeArgument(int32_t t)
		{
			return std::shared_ptr<IntegerArgument>(new IntegerArgument(t));
		}
		
		template<>
		inline ArgumentPtr MakeArgument(uint8_t t)
		{
			return std::shared_ptr<USmallIntegerArgument>(new USmallIntegerArgument(t));
		}
		
		template<>
		inline ArgumentPtr MakeArgument(uint32_t t)
		{
			return std::shared_ptr<UIntegerArgument>(new UIntegerArgument(t));
		}
		
		template<>
		inline ArgumentPtr MakeArgument(uint64_t t)
		{
			return std::shared_ptr<ULargeIntegerArgument>(new ULargeIntegerArgument(t));
		}
		
		template<>
		inline ArgumentPtr MakeArgument(std::string t)
		{
			return std::shared_ptr<StringArgument>(new StringArgument(t));
		}
		
		template<>
		inline ArgumentPtr MakeArgument(const char *t)
		{
			return std::shared_ptr<StringArgument>(new StringArgument(std::string(t)));
		}
		
		template<>
		inline ArgumentPtr MakeArgument(Buffer b)
		{
			return std::shared_ptr<BufferArgument>(new BufferArgument(b));
		}
		
		template<>
		inline ArgumentPtr MakeArgument(char c)
		{
			return std::shared_ptr<CharacterArgument>(new CharacterArgument(c));
		}

		template<>
		inline ArgumentPtr MakeArgument(NegativeZero nz)
		{
			(void) nz;
			return std::shared_ptr<NegativeZeroArgument>(new NegativeZeroArgument());
		}
		
		/**
		 Writes the command to the output stream with a predefined list of arguments
		 @param[in] out the stream to write to
		 @param[in] cmd the TPCL command to write specifying its accepted
		            arguments
		 @oaram[in] args the arguments to the command
		 @return true if the command and all arguments were successfully
		         written
		 */
		bool WriteCommandArgs(std::ostream &out, const std::string &cmd, const std::vector<ArgumentPtr> &args);
		
		/**
		 Writes the command to the output stream. This function is similar
		 WriteCommandArgs but this function takes an unspecified number of
		 arguments which are automatically put into a vector.
		 @param[in] out the stream to write to
		 @param[in] cmd the TPCL command to write specifying its accepted
		            arguments
		 @param[in] args a variadic list of arguments
		 */
		template<typename ... Args>
		inline bool WriteCommand(std::ostream &out, const std::string &cmd, const Args & ... args)
		{
			return WriteCommandArgs(out, cmd, std::vector<ArgumentPtr> { MakeArgument(args)... });
		}
	}
	
	class CommandException : public Exception
	{
	public:
		CommandException(const std::string &msg) : Exception(msg) {}
	};
}

#endif
