#include <climits>
#include <sstream>
#ifdef _WIN32
#include <BaseTsd.h>
#include <windows.h>
#include <winsock.h>
#include "winsock-helper.h"
#else
#include <unistd.h>
#endif
#include "dbea4dprinter.h"

namespace
{
	void sendCommandToPrinter(int fd, const std::string &cmd, StripedHorse::Buffer &buffer)
	{
		ssize_t bytesWritten = write(fd, cmd.data(), cmd.length());
		if (bytesWritten < 0)
			throw StripedHorse::DBEA4DPrinterException("An error occurred while sending the command to the printer");
		else if (static_cast<size_t>(bytesWritten) != cmd.length())
			throw StripedHorse::DBEA4DPrinterException("Unable to send complete command to printer");
		
		ssize_t bytesRead = read(fd, &buffer[0], buffer.size());
		if (bytesRead < 0)
			throw StripedHorse::DBEA4DPrinterException("An error occurred while receiving the response from the printer");
		else if (static_cast<size_t>(bytesRead) != buffer.size())
			throw StripedHorse::DBEA4DPrinterException("Unexpected response from printer");
	}

	[[noreturn]] void parameterIsOutsideExpectedRange(const std::string &arg)
	{
		std::ostringstream message;
		message << "Parameter #" << arg << " of the status response is outside the expected range";
		throw StripedHorse::InvalidDBEA4DPrinterResponse(message.str());
	}

	int byteToSignedInt(unsigned char b)
	{
		// There's some tricky rules around signed types that can
		// lead to implementation-defined behaviour. We can't assign
		// directly to char, because if b is greater than the maximum
		// value represented by char then the conversion is implementation
		// defined (in almost all case it will probably work as expected,
		// but better to ensure the result by using behaviour that is
		// documented to work.
		if (b > SCHAR_MAX)
			return static_cast<int>(b) - UCHAR_MAX - 1;
		else
			return static_cast<int>(b);
	}

	StripedHorse::Buffer getPrinterConfiguration(int fd)
	{
		StripedHorse::Buffer buffer(16);
		sendCommandToPrinter(fd, "\x1D\x30\x05\x00", buffer);
		return buffer;
	}

	StripedHorse::Buffer getPrinterStatus(int fd)
	{
		StripedHorse::Buffer buffer(4);
		sendCommandToPrinter(fd, "\x14\x01\x00", buffer);
		return buffer;
	}

	StripedHorse::DBEA4DPrinter::Configuration parseConfigurationResponse(const StripedHorse::Buffer &buffer)
	{
		// layout according to spec (16 bytes total):
		// byte 0  - paper type (0 to 4)
		// byte 1  - document length low byte
		// byte 2  - document length high byte
		// byte 3  - front head print density (-15 to +15)
		// byte 4  - back head print density (-15 to +15)
		// byte 5  - power control (0 or 1)
		// byte 6  - reserved
		// byte 7  - print speed (0 to 4)
		// byte 8  - error page recover enabled (0/off or 1/on)
		// byte 9  - reserved
		// byte 10 - black mark cut position offset (-59 to +59)
		// byte 11 - paper width (0 to 4)
		// byte 12 - max variable print speed (0 to 3)
		// byte 13 - reserved
		// byte 14 - reserved
		// byte 15 - manual loading
		using namespace StripedHorse;

		DBEA4DPrinter::Configuration config;
		if (buffer.size() != 16)
			throw InvalidDBEA4DPrinterResponse("The configuration response must be 16 bytes in length");

		if (buffer[0] > 4)
			parameterIsOutsideExpectedRange("paper type");
		
		if (buffer[7] > 1)
			parameterIsOutsideExpectedRange("power control");

		config.paperType = static_cast<DBEA4DPrinter::PaperType>(buffer[0]);
		config.documentLength = buffer[2] << 8 | buffer[1];
		config.frontSidePrintDensity = byteToSignedInt(buffer[3]);
		config.backSidePrintDensity = byteToSignedInt(buffer[4]);
		config.powerControl = static_cast<DBEA4DPrinter::PowerControl>(buffer[5]);
		config.printSpeed = static_cast<DBEA4DPrinter::PrintSpeed>(buffer[7]);
		config.errorPageRecoveryEnabled = static_cast<bool>(buffer[8]);
		config.blackMarkCutPositionOffset = byteToSignedInt(buffer[10]);
		config.paperWidth = static_cast<DBEA4DPrinter::PaperWidth>(buffer[11]);
		config.maximumSpeedForVariablePrintSpeed = static_cast<DBEA4DPrinter::PrintSpeed>(buffer[12] + 1);
		config.manualPaperLoading = static_cast<bool>(buffer[15]);

		return config;
	}

	StripedHorse::DBEA4DPrinter::Status parseStatusResponse(const StripedHorse::Buffer &buffer)
	{
		using namespace StripedHorse;
		using Status = DBEA4DPrinter::Status;

		Status status;
		if (buffer.size() != 6)
			throw StripedHorse::InvalidDBEA4DPrinterResponse("The status response must be 6 bytes in length");
		
		// byte #1 - sensor status 1 (bits 0 and 7 = undefined)
		status.paperEndSensor    = buffer[0] & 0b01000000;
		status.blackMarkSensor   = buffer[0] & 0b00100000;
		status.coverOpen         = buffer[0] & 0b00010000;
		status.cutterSensor      = buffer[0] & 0b00001000 ? DBEA4DPrinter::SensorPosition::kNotHomePosition : DBEA4DPrinter::SensorPosition::kHomePosition;
		status.feedSwitchSensor  = buffer[0] & 0b00000100 ? Status::FeedSwitchSensor::kPushed : Status::FeedSwitchSensor::kReleased;
		status.paperJammedSensor = buffer[0] & 0b00000010;

		// byte #2 - sensor status 2 (bits 4 to 7 = undefined)
		status.timeOfFlightSensor = buffer[1] & 0b10000000 ? Status::TimeOfFlightSensor::kPaperOn : Status::TimeOfFlightSensor::kPaperOff;
		status.camMotor1          = buffer[1] & 0b01000000 ? DBEA4DPrinter::SensorPosition::kNotHomePosition : DBEA4DPrinter::SensorPosition::kHomePosition;
		status.labelSensor        = buffer[1] & 0b00100000 ? Status::LabelSensor::kLabel : Status::LabelSensor::kLiner;
		status.camMotor2          = buffer[1] & 0b00010000 ? DBEA4DPrinter::SensorPosition::kNotHomePosition : DBEA4DPrinter::SensorPosition::kHomePosition;

		// byte #3 - printer status (bit 5 = undefined)
		status.isPrinting       = buffer[2] & 0b10000000;
		status.isCutting        = buffer[2] & 0b01000000;
		status.isReceiving      = buffer[2] & 0b00100000;
		status.isBusy           = buffer[2] & 0b00010000;
		status.isPaused         = buffer[2] & 0b00001000;
		status.isInWarningState = buffer[2] & 0b00000010;
		status.isInErrorState   = buffer[2] & 0b00000001;

		// byte #4 - error status 1
		status.hasPaperJamError               = buffer[3] & 0b10000000;
		status.hasCutterJamError              = buffer[3] & 0b01000000;
		status.hasThermalHeadTemperatureError = buffer[3] & 0b00100000;
		status.hasBlackMarkDetectionError     = buffer[3] & 0b00010000;
		status.hasPaperEmptyError             = buffer[3] & 0b00001000;
		status.hasCoverOpenError              = buffer[3] & 0b00000100;
		status.hasPaperLoadError              = buffer[3] & 0b00000010;
		status.hasCamMotorJamError            = buffer[3] & 0b00000001;

		// byte #5 - error status 2 (bits 6 and 7 = undefined)
		status.hasLabelModeError             = buffer[4] & 0b10000000;
		status.hasOverLengthLabelError       = buffer[4] & 0b01000000;
		status.hasMemoryError                = buffer[4] & 0b00100000;
		status.hasPerforationDetectionError  = buffer[4] & 0b00010000;
		status.hasOverLengthPerforationError = buffer[4] & 0b00001000;
		status.hasThermalHeadError           = buffer[4] & 0b00000100;

		// byte #6 - warning status (bit 0 and 2 = undefined, bit 5 = unknown)
		status.has24VoltAnomalyWarning    = buffer[5] & 0b01000000;
		status.hasOverHeatedWarning       = buffer[5] & 0b00010000;
		status.hasMemoryChecksumWarning   = buffer[5] & 0b00001000;
		status.hasSensorAdjustmentWarning = buffer[5] & 0b00000010;
		status.hasThermalHeadWarning      = buffer[5] & 0b00000001;

		return status;
	}
}

StripedHorse::DBEA4DPrinter StripedHorse::DBEA4DPrinter::ReadPrinterConfigurationAndStatus(int fd)
{
	DBEA4DPrinter printer;
	printer.m_configuration = parseConfigurationResponse(getPrinterConfiguration(fd));
	printer.m_status = parseStatusResponse(getPrinterStatus(fd));
	return printer;
}

StripedHorse::DBEA4DPrinter::DBEA4DPrinter()
{
}

StripedHorse::DBEA4DPrinter::DBEA4DPrinter(const DBEA4DPrinter &other)
  : m_configuration(other.m_configuration), m_status(other.m_status)
{
}

StripedHorse::DBEA4DPrinter::DBEA4DPrinter(DBEA4DPrinter &&other)
  : m_configuration(std::move(other.m_configuration)), m_status(std::move(other.m_status))
{
}

StripedHorse::DBEA4DPrinter::DBEA4DPrinter(const Buffer &configurationBuffer, const Buffer &statusBuffer)
  : m_configuration(parseConfigurationResponse(configurationBuffer)), m_status(parseStatusResponse(statusBuffer))
{
}

StripedHorse::DBEA4DPrinter &StripedHorse::DBEA4DPrinter::operator=(const DBEA4DPrinter &other)
{
	m_configuration = other.m_configuration;
	m_status = other.m_status;
	return *this;
}

const StripedHorse::DBEA4DPrinter::Configuration &StripedHorse::DBEA4DPrinter::getConfiguration() const
{
	return m_configuration;
}

const StripedHorse::DBEA4DPrinter::Status &StripedHorse::DBEA4DPrinter::getStatus() const
{
	return m_status;
}

int StripedHorse::DBEA4DPrinter::getStatusCode() const
{
	if (m_status.isInErrorState)
	{
		     if (m_status.hasPaperJamError)               return 1;
		else if (m_status.hasCutterJamError)              return 2;
		else if (m_status.hasThermalHeadTemperatureError) return 3;
		else if (m_status.hasBlackMarkDetectionError)     return 4;
		else if (m_status.hasPaperEmptyError)             return 5;
		else if (m_status.hasCoverOpenError)              return 6;
		else if (m_status.hasPaperLoadError)              return 7;
		else if (m_status.hasCamMotorJamError)            return 8;
		else if (m_status.hasLabelModeError)              return 9;
		else if (m_status.hasOverLengthLabelError)        return 10;
		else if (m_status.hasMemoryError)                 return 11;
		else if (m_status.hasPerforationDetectionError)   return 12;
		else if (m_status.hasOverLengthPerforationError)  return 13;
		else if (m_status.hasThermalHeadError)            return 14;
		else                                              return 99; // generic error
	}
	else if (m_status.isInWarningState)
	{
		     if (m_status.has24VoltAnomalyWarning)    return 100;
		else if (m_status.hasOverHeatedWarning)       return 101;
		else if (m_status.hasMemoryChecksumWarning)   return 102;
		else if (m_status.hasSensorAdjustmentWarning) return 103;
		else if (m_status.hasThermalHeadWarning)      return 104;
		else                                          return 199; // generic warning
	}
	else
	{
		// oll klear (OK)
		return 0;
	}
}
