#ifndef STRIPEDHORSE_DBEA4DPRINTER_H
#define STRIPEDHORSE_DBEA4DPRINTER_H

#include <libcore/exceptions.h>
#include <libcore/types.h>
#include "cups-utilities.h"

namespace StripedHorse
{
	class DBEA4DPrinter
	{
	public:
		enum class PaperType
		{
			kBM,
			kDocumentLength,
			kLabel,
			kPerforation,
			kCutSheet
		};

		enum class PowerControl
		{
			kLow,
			kHigh
		};

		enum class PrintSpeed
		{
			kVariable,
			k6inchesPerSecond,
			k5inchesPerSecond,
			k4inchesPerSecond,
			k3inchesPerSecond
		};

		enum class PaperWidth
		{
			k58mm,
			k80mm,
			k102mm,
			k130mm,
			kReserved
		};

		struct Configuration
		{
			PaperType paperType = PaperType::kLabel;
			unsigned documentLength = 0;
			int frontSidePrintDensity = 0;
			int backSidePrintDensity = 0;
			PowerControl powerControl = PowerControl::kLow;
			PrintSpeed printSpeed = PrintSpeed::kVariable;
			bool errorPageRecoveryEnabled;
			int blackMarkCutPositionOffset;
			PaperWidth paperWidth = PaperWidth::k102mm;
			PrintSpeed maximumSpeedForVariablePrintSpeed = PrintSpeed::k6inchesPerSecond;
			bool manualPaperLoading = false;
		};

		enum class SensorPosition
		{
			kHomePosition,
			kNotHomePosition
		};

		struct Status
		{
			// byte #1 (sensor status 1)
			bool paperEndSensor = false;
			bool blackMarkSensor = false;
			bool coverOpen = false;
			SensorPosition cutterSensor = SensorPosition::kHomePosition;
			enum class FeedSwitchSensor { kReleased, kPushed } feedSwitchSensor = FeedSwitchSensor::kReleased;
			bool paperJammedSensor = false;

			// byte #2 (sensor status 2)
			enum class TimeOfFlightSensor { kPaperOff, kPaperOn } timeOfFlightSensor = TimeOfFlightSensor::kPaperOff;
			SensorPosition camMotor1;
			enum class LabelSensor { kLiner, kLabel } labelSensor = LabelSensor::kLiner;
			SensorPosition camMotor2;

			// byte #3 (printer status)
			bool isPrinting = false;
			bool isCutting = false;
			bool isReceiving = false;
			bool isBusy = false;
			bool isPaused = false;
			bool isInWarningState = false;
			bool isInErrorState = false;

			// byte #4 (error status 1)
			bool hasPaperJamError = false;
			bool hasCutterJamError = false;
			bool hasThermalHeadTemperatureError = false;
			bool hasBlackMarkDetectionError = false;
			bool hasPaperEmptyError = false;
			bool hasCoverOpenError = false;
			bool hasPaperLoadError = false;
			bool hasCamMotorJamError = false;

			// byte #5 (error status 2)
			bool hasLabelModeError = false;
			bool hasOverLengthLabelError = false;
			bool hasMemoryError = false;
			bool hasPerforationDetectionError = false;
			bool hasOverLengthPerforationError = false;
			bool hasThermalHeadError = false;

			// byte #6
			bool has24VoltAnomalyWarning = false;
			bool hasOverHeatedWarning = false;
			bool hasMemoryChecksumWarning = false;
			bool hasSensorAdjustmentWarning = false;
			bool hasThermalHeadWarning = false;
		};

		/**
		 Returns a DBEA4DPrinter object after communicating
		 with the printer via the specified file descriptor
		 @param[in] fd the file descriptor to communicate with (defaults
		     to the regular file descriptor that CUPS uses for
			 backchannel communication (which is 3)
		*/
		static DBEA4DPrinter ReadPrinterConfigurationAndStatus(int fd = static_cast<int>(CUPSCommunicationFileDescriptor::kBackChannel));

		DBEA4DPrinter(const DBEA4DPrinter &);
		DBEA4DPrinter(DBEA4DPrinter &&);

		/**
		 Constructs a DBEA4DPrinter object based on data
		 that was already obtained via other communications
		 channels
		 @param[in] buffer the data returned from the printer
		*/
		DBEA4DPrinter(const Buffer &configurationBuffer, const Buffer &statusBuffer);

		DBEA4DPrinter &operator=(const DBEA4DPrinter &other);

		const Configuration &getConfiguration() const;
		const Status &getStatus() const;
		int getStatusCode() const;

	private:
		DBEA4DPrinter();
		
		Configuration m_configuration;
		Status m_status;
	};

	class DBEA4DPrinterException : public Exception
	{
	public:
		DBEA4DPrinterException(const std::string &msg) : Exception(msg) {}
	};

	class InvalidDBEA4DPrinterResponse : public DBEA4DPrinterException
	{
	public:
		InvalidDBEA4DPrinterResponse(const std::string &msg) : DBEA4DPrinterException(msg) {}
	};
}

#endif
