#ifdef _WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <windows.h>
#include "winsock-helper.h"
#else
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#endif

#include <cstring>
#include <iostream>
#include <vector>
#include "network.h"

StripedHorse::Network::Address::Address(const std::string &name)
  : m_addr(nullptr), m_len(0), m_family(0)
{
	std::string::size_type colon = name.find(':');
	if (colon == std::string::npos)
		throw AddressException("Expected \"host:port\" format");
	
	std::string hostName = name.substr(0, colon);
	std::string port = name.substr(colon + 1);
	
	struct addrinfo hints;
	struct addrinfo *result;
	
	memset(&hints, 0, sizeof hints);
	
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	
	int status = getaddrinfo(hostName.c_str(), port.c_str(), &hints, &result);
	if (status != 0)
	{
		std::cerr << "Unable to resolve " << name << ": ";
		std::cerr << gai_strerror(status) << std::endl;
		return;
	}
	
	m_len = result->ai_addrlen;
	m_family = result->ai_family;
	m_addr = new char[m_len];
	
	memcpy(m_addr, result->ai_addr, m_len);
	
	freeaddrinfo(result);
}

StripedHorse::Network::Address::Address(const Address &addr)
{
	(void) this->operator=(addr);
}

StripedHorse::Network::Address::Address(Address &&addr)
  : m_addr(addr.m_addr), m_len(addr.m_len), m_family(addr.m_family)
{
	addr.m_addr = nullptr;
	addr.m_len = 0;
}

StripedHorse::Network::Address &StripedHorse::Network::Address::operator=(const Address &addr)
{
	m_len = addr.m_len;
	m_family = addr.m_family;
	if (m_len > 0 && addr.m_addr != nullptr)
	{
		m_addr = new char[addr.m_len];
		memcpy(m_addr, addr.m_addr, addr.m_len);
	}
	else
	{
		m_addr = nullptr;
	}
	
	return *this;
}

StripedHorse::Network::Address::~Address()
{
	delete[] static_cast<char *>(m_addr);
}

int StripedHorse::Network::Address::connect() const
{
	int s = socket(m_family, SOCK_STREAM, 0);
	if (s < 0)
		throw AddressException("Unable to create socket");
	
	if (::connect(s, static_cast<struct sockaddr *>(m_addr), m_len) != 0)
	{
		close(s);
		throw AddressException("Unable to connect");
	}
	
	return s;
}

const void *StripedHorse::Network::Address::getAddressData() const
{
	return m_addr;
}

int StripedHorse::Network::Address::getFamily() const
{
	return m_family;
}

size_t StripedHorse::Network::Address::getLength() const
{
	return m_len;
}

std::string StripedHorse::Network::Address::getPresentation() const
{
	std::vector<char> buffer(128);
	const char *status = inet_ntop(m_family, m_addr, &buffer[0], m_len);
	
	if (status == nullptr)
		return "unknown";
	else
		return std::string(&buffer[0]);
}

bool StripedHorse::Network::Address::isValid() const
{
	return m_addr != nullptr && m_len > 0;
}
