#ifndef STRIPEDHORSE_NETWORK_H
#define STRIPEDHORSE_NETWORK_H

#include <cstdint>
#include <string>

#include <libcore/exceptions.h>

namespace StripedHorse
{
	namespace Network
	{
		class Address
		{
		public:
			/**
			 Constructs a network address object.
			 @param[in] name the hostname or IP address
			 */
			Address(const std::string &name);
			
			/**
			 Copies a network address object from another one
			 */
			Address(const Address &);
			
			/**
			 Moves a network address object from another one
			 */
			Address(Address &&);
			
			/**
			 Copies a network address from another one
			 @param[in] other the other address (to copy from)
			 */
			Address &operator=(const Address &other);
			~Address();
			
			/**
			 Create a socket establishing a TCP connection to this address
			 @return if the socket could be created, the result is a file
			         descriptor which can be used for communication. If the
			         socket could not be created, the result is negative.
			 */
			int connect() const;
			
			/**
			 Gets a pointer to the arbitrary address data
			 @return if the address object is valid, the return value is
			         a pointer pointing to the socket address structure. It is
			         a pointer to void because it may represent either an IPv4
			         or an IPv6 address. If the address object is invalid, the
			         result is `nullptr`.
			 */
			const void *getAddressData() const;
			
			/**
			 Gets the address family
			 @return returns either `AF_INET` or `AF_INET6`
			 */
			int getFamily() const;
			
			/**
			 Gets the length (in bytes) of the socket address data
			 @return if this address object is valid, the result is the length
			         of the socket address data. If this address object is
			         invalid, the result is 0.
			 */
			size_t getLength() const;
			
			/**
			 Gets the textual representation of the address
			 @return if this address object is valid, the result is the textual
			         representation of the address represented by this address
			         object. For example, if this address object represents an
			         IPv4 address, the result may be "10.0.0.1". If this
			         address object is invalid, the result is "unknown".
			 */
			std::string getPresentation() const;
			
			/**
			 Determines whether this address object represents a valid socket
			 address
			 @return if this address object was constructed using a hostname
			         and the hostname was successfully resolved, then this
			         address object should be valid. If the address object
			         was constructed using a valid textual representation of
			         an IPv4 or IPv6 address, then this address object should
			         be valid. Otherwise, the address object may be invalid.
			 */
			bool isValid() const;
				
		private:
			void *m_addr;
			size_t m_len;
			int m_family;
		};
		
		class AddressException : public Exception
		{
		public:
			AddressException(const std::string &m) : Exception(m) {}
		};
	}
}

#endif

