#include <cmath>
#include <sstream>
#include <string>
#include <regex>
#include <unordered_map>
#include "options-utilities.h"

namespace
{
	/// An enum representing one of the possible types of parameters for custom options in a PPD
	enum class ParameterType
	{
		kCurve,
		kInteger,
		kInverseCurve,
		kPasscode,
		kPassword,
		kPoints,
		kReal,
		kString
	};

	std::unordered_map<std::string, enum ParameterType> g_parameterNameMap {
		{ "curve", ParameterType::kCurve },
		{ "int", ParameterType::kInteger },
		{ "invcurve", ParameterType::kInverseCurve },
		{ "passcode", ParameterType::kPasscode },
		{ "password", ParameterType::kPassword },
		{ "points", ParameterType::kPoints },
		{ "real", ParameterType::kReal },
		{ "string", ParameterType::kString }
	};

	/// A struct representing the parameters of a custom option in a PPD
	struct CustomParameters
	{
		int order;
		enum ParameterType parameterType;
		
		union
		{
			int integer;
			double real;
		} minimum;

		union
		{
			int integer;
			double real;
		} maximum;
	};

	bool tryParseCustomParameters(const std::string &cpString, CustomParameters &cp)
	{
		std::istringstream inStream(cpString);
		
		if (!(inStream >> cp.order))
			return false;
		
		std::string paramTypeName;
		if (!(inStream >> paramTypeName))
			return false;
		
		auto paramIter = g_parameterNameMap.find(paramTypeName);
		if (paramIter == g_parameterNameMap.end())
			return false;
		
		cp.parameterType = paramIter->second;

		// several parameter types accept real numbers, while others
		// are integers
		switch (cp.parameterType)
		{
			case ParameterType::kCurve:
			case ParameterType::kInverseCurve:
			case ParameterType::kPoints:
			case ParameterType::kReal:
				if (!((inStream >> cp.minimum.real) && (inStream >> cp.maximum.real)))
					return false;
				break;

			default:
				if (!((inStream >> cp.minimum.integer) && (inStream >> cp.maximum.integer)))
					return false;
		}

		return true;
	}
}

bool StripedHorse::OptionsUtilities::TryGetIntegerRangeForOption(const Options &options, const std::string &optionName, IntegerValueRange &valueRange)
{
	const std::string minimumName = optionName + ".Minimum";
	const std::string maximumName = optionName + ".Maximum";
	
	if (!(options.hasOption(minimumName) && options.hasOption(maximumName)))
		return false;
	
	int32_t minValue, maxValue;
	if (!(options.tryGetInt32(minimumName, minValue) && options.tryGetInt32(maximumName, maxValue)))
		return false;
	
	valueRange.minimum = minValue;
	valueRange.maximum = maxValue;
	return true;
}

bool StripedHorse::OptionsUtilities::TryGetRealRangeForOption(const Options &options, const std::string &optionName, RealValueRange &valueRange)
{
	const std::string minimumName = optionName + ".Minimum";
	const std::string maximumName = optionName + ".Maximum";
	
	if (!(options.hasOption(minimumName) && options.hasOption(maximumName)))
		return false;
	
	double minValue, maxValue;
	if (!(options.tryGetDouble(minimumName, minValue) && options.tryGetDouble(maximumName, maxValue)))
		return false;
	
	valueRange.minimum = minValue;
	valueRange.maximum = maxValue;
	return true;
}

bool StripedHorse::OptionsUtilities::ParseDefaultValuesFromPPD(Options &options, const PPDFile &ppd)
{
	std::regex defaultRegex { "^Default(.+)$" };

	bool parsedOptions = false;
	for (auto it = ppd.begin(); it != ppd.end(); it++)
	{
		if (it->second.getValue().getType() != PPDValue::kStringValue)
			continue;
		
		std::smatch result;
		if (std::regex_match(it->first, result, defaultRegex))
		{
			std::string name = result.str(1);
			std::string value = static_cast<std::string>(it->second.getValue());
			
			options.set(name, value);
			parsedOptions = true;
		}
	}

	return parsedOptions;
}

bool StripedHorse::OptionsUtilities::ParseMinAndMaxValuesFromPPD(Options &options, const PPDFile &ppd)
{
	std::regex paramRegex { "^ParamCustom(.+)$" };

	bool parsedMinAndMaxValues = false;
	for (auto it = ppd.begin(); it != ppd.end(); it++)
	{
		if (it->second.getValue().getType() != PPDValue::kStringValue)
			continue;
		
		std::smatch result;
		if (std::regex_match(it->first, result, paramRegex))
		{
			std::string name = result.str(1);
			std::string value = static_cast<std::string>(it->second.getValue());

			CustomParameters cp;
			if (tryParseCustomParameters(value, cp))
			{
				switch (cp.parameterType)
				{
					case ParameterType::kCurve:
					case ParameterType::kInverseCurve:
					case ParameterType::kPoints:
					case ParameterType::kReal:
						options.set(name + ".Minimum", std::to_string(cp.minimum.real));
						options.set(name + ".Maximum", std::to_string(cp.maximum.real));
						break;
					
					default:
						options.set(name + ".Minimum", std::to_string(cp.minimum.integer));
						options.set(name + ".Maximum", std::to_string(cp.maximum.integer));
						break;
				}

				parsedMinAndMaxValues = true;
			}
		}
	}

	return parsedMinAndMaxValues;
}

int32_t StripedHorse::OptionsUtilities::GetValue(const Options &options, const std::string &optionName, IntegerValueRange range, int32_t defaultValue)
{
	int32_t result = defaultValue;

	if (options.hasOption(optionName))
		result = options.getInt32(optionName, defaultValue);

	if (result < range.minimum)
		result = range.minimum;
	else if (result > range.maximum)
		result = range.maximum;
	
	return result;
}

double StripedHorse::OptionsUtilities::GetValue(const Options &options, const std::string &optionName, RealValueRange range, double defaultValue)
{
	double result = defaultValue;

	if (options.hasOption(optionName))
		result = options.getDouble(optionName, defaultValue);

	if (result < range.minimum)
		result = range.minimum;
	else if (result > range.maximum)
		result = range.maximum;

	return result;
}

bool StripedHorse::OptionsUtilities::TryGetClampedValueIn10dpmm(const Options &options, const std::string &optionName, int32_t &output, double defaultValue)
{
	if (!options.hasOption(optionName))
		return false;

	RealValueRange realRange;
	if (!TryGetRealRangeForOption(options, optionName, realRange))
	{
		// there are no suitable values that can be used to clamp the output
		return false;
	}

	output = GetValueIn10dpmm(options, optionName, realRange, defaultValue);
	return true;
}

int32_t StripedHorse::OptionsUtilities::GetValueIn10dpmm(const Options &options, const std::string &optionName, RealValueRange range, double defaultValue)
{
	double result;
	if (!options.tryGetDouble(optionName, result))
		result = defaultValue;

	if (result < range.minimum)
		result = range.minimum;
	else if (result > range.maximum)
		result = range.maximum;

	// Custom values and PPD options are represented in millimetres so we multiply
	// by 10 to get it into 10dpmm format. By default, conversion to integer results
	// in truncation, but in this case, we want to round to the nearest integer instead.
	return std::round(result * 10.0);
}
