#ifndef STRIPEDHORSE_OPTIONS_UTILITIES_H
#define STRIPEDHORSE_OPTIONS_UTILITIES_H

#include "options.h"
#include "ppdfile.h"

namespace StripedHorse
{
	namespace OptionsUtilities
	{
		struct IntegerValueRange
		{
			int32_t minimum;
			int32_t maximum;
		};

		struct RealValueRange
		{
			double minimum;
			double maximum;
		};

		/**
		 Checks the Options object to see whether the minimum and maximum values are available. If
		 the Options object contains the minimum and maximum values and were successfully parsed as
		 integers, this function sets the minimum and maximum values in the valueRange argument and
		 returns true, otherwise returns false.
		 @param[in] options the Options object containing the potential minimum and maximum values
		 @param[in] optionName the name of the option to check
		 @param[out] valueRange will contain the minimum and maximum value of the option if they were
		             both in the Options object. These values are represented in 10dpmm units.
		*/
		bool TryGetIntegerRangeForOption(const Options &options, const std::string &optionName, IntegerValueRange &valueRange);

		/**
		 Checks the Options object to see whether the minimum and maximum values are available. If
		 the Options object contains the minimum and maximum values and were successfully parsed as
		 real numbers, this function sets the minimum and maximum values in the valueRange argument and
		 returns true, otherwise returns false.
		 @param[in] options the Options object containing the potential minimum and maximum values
		 @param[in] optionName the name of the option to check
		 @param[out] valueRange will contain the minimum and maximum value of the option if they were
		             both in the Options object. These values are represented in 10dpmm units.
		*/
		bool TryGetRealRangeForOption(const Options &options, const std::string &optionName, RealValueRange &valueRange);

		/**
		 Parses a PPD file for default values and sets the corresponding value
		 in an Options object.
		 @param[out] options the Options object to manipulate with settings from the PPD
		 @param[in] ppd the PPD file to obtain the default options from
		*/
		bool ParseDefaultValuesFromPPD(Options &options, const PPDFile &ppd);

		/**
		 Parses a PPD for customisable options (those with a "Param" prefix) and determines
		 the minimum and maximum values for those options and stores them in the Options
		 object as "OptionName.Minimum" and "OptionName.Maximum"
		 @param[out] options the Options object to manipulate
		 @param[in] ppd the PPD file to obtain the minimum and maximum values from
		 */
		bool ParseMinAndMaxValuesFromPPD(Options &options, const PPDFile &ppd);

		/**
		 Some options are customizable but should be restricted to a range. The
		 IntegerValueRange represents the actual range of acceptable values (as
		 integers). For example, if an option allows a range of -20 to 20, then
		 range will be { -20, 20 }.
		 */
		int32_t GetValue(const Options &options, const std::string &optionName, IntegerValueRange range, int32_t defaultValue = 0);

		/**
		 Some options are customizable but should be restricted to a range. The
		 RealValueRange represents the actual range of acceptable values (as
		 real numbers as opposed to integers). For example, if an option allows
		 a range of -9.9mm to 9.9mm, then range will be { -9.9, 9.9 }.
		 */
		double GetValue(const Options &options, const std::string &optionName, RealValueRange range, double defaultValue = 0);

		/**
		 Since options like BackFeedAdjustment are represented in TPCL as 10dpmm but
		 custom values and PPD options are represented as just millimetres, this
		 function obtains the value and automatically ensures that the value falls
		 within the specified minimum and maximum values (inclusive), and if the value
		 is valid, it is then converted into 10dpmm units by multiplying by 10 and
		 rounding to the nearest integer.
		*/
		int32_t GetValueIn10dpmm(const Options &options, const std::string &optionName, RealValueRange range, double defaultValue = 0.0);

		/**
		 This is similar to the GetValueIn10dpmm function but it attempts to use the
		 minimum and maximum values specified in the Options object to determine how
		 to clamp the output value. If this function returns false, "value" not
		 updated.
		*/
		bool TryGetClampedValueIn10dpmm(const Options &options, const std::string &optionName, int32_t &value, double defaultValue = 0.0);
	}
}

#endif
