#include <cstdlib>
#include "options.h"
#include "string-escape.h"

StripedHorse::Options::Options()
{
}

StripedHorse::Options::Options(const std::string &opts)
{
	parse(opts);
}

StripedHorse::Options::Options(const Options &other)
{
	(void) this->operator=(other);
}

StripedHorse::Options::Options(Options &&other)
  : m_optionMap(std::move(other.m_optionMap)),
    m_customOptions(std::move(other.m_customOptions))
{
}

StripedHorse::Options &StripedHorse::Options::operator=(const StripedHorse::Options &other)
{
	m_optionMap = other.m_optionMap;
	m_customOptions = other.m_customOptions;
	return *this;
}

double StripedHorse::Options::getDouble(const std::string &name, double def) const
{
	OptionMap::const_iterator it = m_optionMap.find(name);
	if (it == m_optionMap.end())
		return def;
	
	size_t pos = 0;
	double result = std::stod(it->second, &pos);
	if (pos < it->second.size())
		return def;
	else
		return result;
}

bool StripedHorse::Options::getBoolean(const std::string &name, bool def) const
{
	OptionMap::const_iterator it = m_optionMap.find(name);
	if (it == m_optionMap.end())
		return def;
	
	if (it->second == "True" || it->second == "true" || it->second == "Enabled" || it->second == "enabled")
		return true;
	else if (it->second == "False" || it->second == "false" || it->second == "Disabled" || it->second == "disabled")
		return false;
	else
		return def;
}

float StripedHorse::Options::getFloat(const std::string &name, float def) const
{
	OptionMap::const_iterator it = m_optionMap.find(name);
	if (it == m_optionMap.end())
		return def;
	
	size_t pos = 0;
	float result = std::stof(it->second, &pos);
	if (pos < it->second.size())
		return def;
	else
		return result;
}

int32_t StripedHorse::Options::getInt32(const std::string &name, int32_t def) const
{
	OptionMap::const_iterator it = m_optionMap.find(name);
	if (it == m_optionMap.end())
		return def;
	
	try
	{
		size_t pos = 0;
		long result = std::stol(it->second, &pos, 10);
		if (pos < it->second.size())
			return def;
		else
			return static_cast<int32_t>(result);
	}
	catch (std::invalid_argument &)
	{
		return def;
	}
	catch (std::out_of_range &)
	{
		return def;
	}
}

uint32_t StripedHorse::Options::getUInt32(const std::string &name, uint32_t def) const
{
	OptionMap::const_iterator it = m_optionMap.find(name);
	if (it == m_optionMap.end())
		return def;
	
	try
	{
		size_t pos = 0;
		long long result = std::stoll(it->second, &pos, 10);
		if (pos < it->second.size())
			return def;
		else
			return static_cast<uint32_t>(result);
	}
	catch (std::invalid_argument &)
	{
		return def;
	}
	catch (std::out_of_range &)
	{
		return def;
	}
}

std::string StripedHorse::Options::getString(const std::string &name, const std::string &def) const
{
	OptionMap::const_iterator it = m_optionMap.find(name);
	if (it == m_optionMap.end())
		return def;
	else
		return it->second;
}

bool StripedHorse::Options::hasOption(const std::string &name) const
{
	return m_optionMap.find(name) != m_optionMap.end();
}

bool StripedHorse::Options::isCustom(const std::string &name) const
{
	return m_customOptions.find(name) != m_customOptions.end();
}

void StripedHorse::Options::parse(const std::string &str)
{
	bool escaped = false;
	char current = '\0';
	
	std::string::const_iterator parseStart = str.begin();
	std::string::const_iterator parseEnd = parseStart;
	std::string::const_iterator parseEquals = str.end();
	
	while (parseEnd <= str.end())
	{
		if (parseEnd == str.end())
			current = ' ';
		else
			current = *parseEnd;
		
		if (escaped)
		{
			escaped = false;
			parseEnd++;
			continue;
		}
		
		switch (current)
		{
			case '\\':
			{
				escaped = true;
				break;
			}
			
			case '=':
			{
				parseEquals = parseEnd;
				break;
			}
				
			case ' ':
			{
				if (parseEquals == str.end())
				{
					std::string keyName(parseStart, parseEnd);
					if (keyName.size() > 2 && keyName.substr(0, 2) == "no")
						m_optionMap[keyName.substr(2)] = "False";
					else
						m_optionMap[keyName] = "True";
				}
				else
				{
					std::string keyName(parseStart, parseEquals);
					std::string rawValueName(parseEquals + 1, parseEnd);
					std::string valueName;
					
					valueName.reserve(rawValueName.size());
					for (auto vch : rawValueName)
					{
						if (vch == '\\')
							continue;
						valueName.push_back(vch);
					}
					
					if (valueName.substr(0, 7) == "Custom.")
					{
						m_optionMap[keyName] = valueName.substr(7);
						m_customOptions.insert(keyName);
					}
					else
					{
						m_optionMap[keyName] = valueName;
					}
				}
				
				parseEquals = str.end();
				if (parseEnd == str.end())
					break;
				parseStart = parseEnd + 1;
			}
		}
		
		parseEnd++;
	}
}

void StripedHorse::Options::remove(const std::string &name)
{
	auto optionMapIt = m_optionMap.find(name);
	if (optionMapIt != m_optionMap.end())
		m_optionMap.erase(optionMapIt);
	
	auto customOptionsIt = m_customOptions.find(name);
	if (customOptionsIt != m_customOptions.end())
		m_customOptions.erase(customOptionsIt);
}

void StripedHorse::Options::set(const std::string &name, const std::string &value)
{
	m_optionMap[name] = value;
}

bool StripedHorse::Options::tryGetDouble(const std::string &name, double &val) const
{
	auto optionIter = m_optionMap.find(name);
	if (optionIter == m_optionMap.end())
		return false;

	try
	{
		size_t pos;
		double result = std::stod(optionIter->second, &pos);
		if (pos != optionIter->second.size())
			return false;
	
		val = result;
		return true;
	}
	catch (std::invalid_argument &)
	{
		return false;
	}
	catch (std::out_of_range &)
	{
		return false;
	}
}

bool StripedHorse::Options::tryGetInt32(const std::string &name, int32_t &val) const
{
	auto optionIter = m_optionMap.find(name);
	if (optionIter == m_optionMap.end())
		return false;

	try
	{
		size_t pos;
		long result = std::stol(optionIter->second, &pos, 10);
		if (pos != optionIter->second.size())
			return false;
		
		val = static_cast<int32_t>(result);
		return true;
	}
	catch (std::invalid_argument &)
	{
		return false;
	}
	catch (std::out_of_range &)
	{
		return false;
	}
}

std::ostream &StripedHorse::operator<<(std::ostream &out, const Options &options)
{
	for (auto opt : options.m_optionMap)
	{
		if (options.m_customOptions.find(opt.first) != options.m_customOptions.end())
			out << "Custom.";
		
		out << EscapeString(opt.first) << "=" << EscapeString(opt.second) << " ";
	}

	return out;
}
