#ifndef STRIPEDHORSE_OPTIONS_H
#define STRIPEDHORSE_OPTIONS_H 1

#include <cstdint>
#include <iostream>
#include <string>
#include <unordered_map>
#include <unordered_set>

namespace StripedHorse
{
	typedef std::unordered_map<std::string, std::string> OptionMap;
	class Options
	{
	public:
		/**
		 Constructs an Options object.
		 @param[in] str a string representation of the options in key=value
		            format.
		 */
		Options();
		Options(const std::string &str);
		Options(const Options &);
		Options(Options &&);
		
		/**
		 Copies another options object
		 */
		Options &operator=(const Options &);
		
		/**
		 Finds the option with the specified name and returns its boolean value
		 @param[in] name the name of the option
		 @param[in] def the default value to return
		 @return if the option was found, the value is converted to a boolean
		         value (e.g. if it is non-zero, it is converted to true, or if
		         the value is the word "true" it is converted to true etc). If
		         the option was not found or the value was not understood as
		         a boolean value, the default value is returned.
		 */
		bool getBoolean(const std::string &name, bool def = false) const;
		
		/**
		 Finds the option with the specified name and returns the value as a
		 double-precision floating point value
		 @param[in] name the name of the option
		 @param[in] def the default value to return
		 @return if the option was found, the value is converted to a floating
		         point value and returned. If the option was not found or the
		         value does not represent a floating point value, the default
		         value is returned.
		 */
		double getDouble(const std::string &name, double def = 0) const;
		
		/**
		 Finds the option with the specified name and returns its integer value
		 @param[in] name the name of the option
		 @param[in] def the default value to return
		 @return if the option was found, the value is converted to an integer
		         and returned. If the option was not found or the value does not
		         represent an integer, the default value is returned.
		 */
		int32_t getInt32(const std::string &name, int32_t def = 0) const;
		
		/**
		 Finds the option with the specified name and returns its unsigned
		 integer value
		 @param[in] name the name of the option
		 @param[in] def the default value to return
		 @return if the option was found, the value is converted to an unsigned
		         integer and returned. If the option was not found or the value
		         does not represent an unsigned integer (e.g. it is negative),
		         the default value is returned.
		 */
		uint32_t getUInt32(const std::string &name, uint32_t def = 0) const;
		
		/**
		 Finds the option with the specified name and returns the value as a
		 floating point value
		 @param[in] name the name of the option
		 @param[in] def the default value to return
		 @return if the option was found, the value is converted to a floating
		         point value and returned. If the option was not found or the
		         value does not represent a floating point value, the default
		         value is returned.
		 */
		float getFloat(const std::string &name, float def = 0) const;
		
		/**
		 Finds the option with the specified name and returns the value as a
		 string
		 @param[in] name the name of the option
		 @param[in] def the default value to return
		 @return if the option was found, the value is returned as a string. If
		         the value was not found, the default value is returned.
		 */
		std::string getString(const std::string &name, const std::string &def = "") const;
		
		/**
		 Determines whether the option exists
		 @param[in] name the name of the option
		 @return if the value exists, the result is true, otherwise false
		 */
		bool hasOption(const std::string &name) const;
		
		/**
		 Determines whether the option exists and whether it has a custom value
		 @param[in] name the name of the option
		 @return if the option has a custom value, the result is true. CUPS
		         allows PPDs to specify that they accept custom values
		         alongside a set of predefined values. If the user has entered
		         a custom value instead of choosing a predfined value, this
		         method should return true.
		 */
		bool isCustom(const std::string &name) const;

		/**
		 Parses a string containing options in the form "key=value key=value..."
		 @param[in] opts the string containing the options in key=value form
		*/
		void parse(const std::string &opts);
		
		/**
		 Removes an option
		 @param[in] option the name of the option to remove
		*/
		void remove(const std::string &option);

		/**
		 Sets an option
		 @param[in] name the name of the option
		 @param[in] value the value of the option
		 */
		void set(const std::string &name, const std::string &value);

		/**
		 Tries to get the value of an option as a double floating point. If the
		 value does not exist or represent a double, this function returns false,
		 otherwise, the parsed value is saved into the argument "val" and this
		 function returns true. If this function returns false, "val" is untouched.
		 @param[in] name the name of the option
		 @param[out] val the value of the option as a double, if successfully parsed
		 */
		bool tryGetDouble(const std::string &name, double &val) const;

		/**
		 Tries to get the value of an option as a 32-bit signed integer. If the
		 value does not exist or represent an integer, this function returns false,
		 otherwise, the parsed value is saved into the argument "val" and this
		 function returns true. If this function returns false, "val" is untouched.
		 @param[in] name the name of the option
		 @param[out] val the value of the option as an integer, if successfully parsed
		 */
		bool tryGetInt32(const std::string &name, int32_t &val) const;
		
	private:
		OptionMap m_optionMap;
		std::unordered_set<std::string> m_customOptions;
	
	friend std::ostream &operator<<(std::ostream &out, const Options &options);
	};

	std::ostream &operator<<(std::ostream &out, const Options &options);
}

#endif

