#include <algorithm>
#include <fstream>
#include <iterator>
#include "ppdfile.h"

namespace
{
	class Parser
	{
	public:
		Parser(std::istream &in)
		  : m_in(in), m_lineNumber(0)
		{
		}
		
		bool ended() const
		{
			return m_in.eof();
		}
		
		bool getLine(std::string &out, size_t max = 255)
		{
			std::string result;
			char ch;
			bool readOK = false;
			
			out.erase();
			result.reserve(max);
			
			while (m_in.get(ch))
			{
				result.push_back(ch);
				if (ch == '\r' && m_in.peek() == '\n')
					continue;
				
				if (result.length() == max && ch != '\n')
				{
					break;
				}
				else if (ch == '\n')
				{
					readOK = true;
					m_lineNumber++;
					break;
				}
			}
			
			out.swap(result);
			return readOK;
		}
		
		size_t getLineNumber() const
		{
			return m_lineNumber;
		}
	
	private:
		std::istream &m_in;
		size_t m_lineNumber;
	};
	
	int fromHex(char c)
	{
		if (c >= '0' && c <= '9')
			return c - '0';
		else if (c >= 'A' && c <= 'F')
			return c - 'A';
		else if (c >= 'a' && c <= 'f')
			return c - 'a';
		else
			throw std::invalid_argument("Not hexadecimal character");
	}
	
	std::string decodeHexString(const std::string &in)
	{
		if (in.size() % 2 != 0)
			throw std::invalid_argument("Hex string is not multiple of 2 characters");
		
		std::string result;
		result.reserve(in.size() / 2);
		
		for (size_t i = 0; i < in.size(); i += 2)
			result.push_back((fromHex(in[i]) << 4) | fromHex(in[i + 1]));
		
		return result;
	}
	
	std::string processRawQuotedValue(const std::string &in)
	{
		std::string result;
		std::string::size_type last = 0;
		
		while (last < in.size())
		{
			std::string::size_type i = in.find('<', last);
			if (i == std::string::npos)
			{
				result.append(in.substr(last));
				break;
			}
			else
			{
				last = in.find('>', i);
				if (last == std::string::npos)
					throw std::invalid_argument("Unterminated hex string");
				
				result.append(decodeHexString(in.substr(i + 1, last - 1)));
				last++;
			}
		}
		
		return result;
	}
	
	StripedHorse::PPDValue parseValue(const std::string &line, size_t offset, Parser &parser, bool expectInvocationValue)
	{
		using namespace StripedHorse;
		
		if (line[offset] == '"')
		{
			enum PPDValue::ValueType valueType = expectInvocationValue ? PPDValue::kInvocationValue : PPDValue::kQuotedValue;
			std::string::size_type end = line.find('"', offset + 1);
			if (end != std::string::npos)
			{
				// value fits on a single line
				std::string value = line.substr(offset + 1, end - offset - 1);
				if (valueType == PPDValue::kQuotedValue)
				{
					try
					{
						value = processRawQuotedValue(value);
					}
					catch (std::invalid_argument &ex)
					{
						throw PPDSyntaxException(parser.getLineNumber(), ex.what());
					}
				}
				
				return StripedHorse::PPDValue(valueType, value);
			}
			else if (expectInvocationValue)
			{
				// value spans multiple lines
				size_t originalLineNumber = parser.getLineNumber();
				std::string value = line.substr(offset + 1);
				std::string line;
				bool parseOK = false;
				
				while (parser.getLine(line))
				{
					std::string::size_type quote = line.find('"');
					if (quote != std::string::npos)
					{
						value.append(line.substr(0, quote));
						parseOK = true;
						break;
					}
					
					value.append(line);
				}
				
				if (!parseOK)
					throw PPDSyntaxException(originalLineNumber, "Unexpected end of file encountered while parsing invocation value");
				
				return PPDValue(valueType, value);
			}
			else
			{
				throw PPDSyntaxException(parser.getLineNumber(), "Unexpected end of line encountered while parsing quoted value");
			}
		}
		else if (line[offset] == '^')
		{
			std::string::size_type i;
			for (i = offset + 1; i < line.size() - 1; i++)
				if (!isprint(line[i]))
					break;
			return StripedHorse::PPDValue(StripedHorse::PPDValue::kSymbolValue, line.substr(offset, i));
		}
		else
		{
			std::string::size_type end = line.find_last_not_of(" \t\r\n");
			return StripedHorse::PPDValue(StripedHorse::PPDValue::kStringValue, line.substr(offset, end - offset + 1));
		}
	}
	
	StripedHorse::PPDStatement parseStatement(const std::string &line, size_t offset, Parser &parser)
	{
		using namespace StripedHorse;
		
		std::string mainKeyword;
		mainKeyword.reserve(40);
		
		for (size_t i = offset; i < line.size(); i++)
		{
			char ch = line[i];
			if (ch >= 33 && ch <= 126 && ch != ':' && ch != '/' && isprint(ch))
				mainKeyword.push_back(ch);
			else
				break;
		}
		
		if (mainKeyword.size() == 0)
			throw PPDSyntaxException(parser.getLineNumber(), "No main keyword");

		if (mainKeyword.length() > 40)
			throw PPDSyntaxException(parser.getLineNumber(), "Main keyword exceeds 40 characters");
		
		offset = line.find_first_not_of(" \t", offset + mainKeyword.size());
		if (line[offset] == '\n')
		{
			// *MainKeyword
			return PPDStatement(parser.getLineNumber(), mainKeyword);
		}
		
		if (line[offset] == ':')
		{
			bool isQueryString = mainKeyword.front() == '?';
			// *MainKeyword: Value
			offset = line.find_first_not_of(" \t", offset + 1);
			return PPDStatement(parser.getLineNumber(), mainKeyword, parseValue(line, offset, parser, isQueryString));
		}
		
		// *MainKeyword Option: Value
		std::string::size_type colon = line.find(':', offset + 1);
		if (colon == std::string::npos)
			throw PPDSyntaxException(parser.getLineNumber(), "Syntax error parsing option keyword");
		
		std::string optionKeyword = line.substr(offset, colon - offset);
		std::string translation;
		std::string::size_type slash = optionKeyword.find('/');
		if (slash != std::string::npos)
		{
			translation = optionKeyword.substr(slash + 1);
			optionKeyword.erase(slash);
		}
		
		offset = line.find_first_not_of(" \t", colon + 1);
		if (offset == std::string::npos)
			throw PPDSyntaxException(parser.getLineNumber(), "Unexpected end of line looking for value");
		
		if (translation.length() > 0)
			return PPDStatement(parser.getLineNumber(), mainKeyword, optionKeyword, translation, parseValue(line, offset, parser, true));
		else
			return PPDStatement(parser.getLineNumber(), mainKeyword, optionKeyword, parseValue(line, offset, parser, true));
	}
	
	bool isWhitespace(const std::string &str)
	{
		return str.find_first_not_of(" \t\r\n") == std::string::npos;
	}
	
	void normaliseLineEnding(std::string &str)
	{
		if (str.length() >= 2 && str.substr(str.length() - 2) == "\r\n")
			str.erase(str.length() - 2);
	}
}

using namespace StripedHorse;

PPDFile::PPDFile()
{
}

PPDFile::PPDFile(std::istream &in)
{
	load(in);
}

PPDFile::PPDFile(const std::string &filename)
{
	std::ifstream in(filename, std::ios::in | std::ios::binary);
	load(in);
}

PPDFile::PPDFile(const PPDFile &other)
{
	this->operator=(other);
}

PPDFile::PPDFile(PPDFile &&other)
  : m_statements(std::move(other.m_statements))
{
}

PPDFile &PPDFile::operator=(const PPDFile &other)
{
	m_statements = other.m_statements;
	return *this;
}

PPDFile::PPDStatementMap::const_iterator PPDFile::begin() const
{
	return m_statements.begin();
}

size_t PPDFile::countStatementsWithMainKeyword(const std::string &kw) const
{
	return m_statements.count(kw);
}

PPDFile::PPDStatementMap::const_iterator PPDFile::end() const
{
	return m_statements.end();
}

const PPDStatement &PPDFile::getAnyStatementWithMainKeyword(const std::string &kw) const
{
	auto it = m_statements.find(kw);
	if (it == m_statements.end())
		throw PPDKeywordNotFoundException("Unknown main keyword");
	return it->second;
}

std::vector<PPDStatement> PPDFile::getAllStatementsWithMainKeyword(const std::string &kw) const
{
	auto range = m_statements.equal_range(kw);
	std::vector<PPDStatement> result;
	
	std::transform(range.first, range.second, std::back_inserter(result), [](const std::pair<std::string, PPDStatement> &p) { return p.second; });
	
	return result;
}

std::string PPDFile::getDefaultChoiceForOption(const std::string &optionName) const
{
	std::string fullMainKeywordName = "Default" + optionName;

	auto range = m_statements.equal_range(fullMainKeywordName);
	if (range.first == range.second)
		throw PPDKeywordNotFoundException(fullMainKeywordName);
	
	return static_cast<std::string>(range.first->second.getValue());
}

bool PPDFile::hasOption(const std::string &optionName) const
{
	std::string fullOptionName = "*" + optionName;
	auto range = m_statements.equal_range("OpenUI");
	
	for (PPDStatementMap::const_iterator it = range.first; it != range.second; it++)
	{
		if (it->second.getOptionKeyword() == fullOptionName)
			return true;
	}

	return false;
}

bool PPDFile::hasOptionWithChoice(const std::string &optionName, const std::string &choiceName) const
{
	if (!hasOption(optionName))
		return false;
	
	auto range = m_statements.equal_range(optionName);
	
	for (PPDStatementMap::const_iterator it = range.first; it != range.second; it++)
	{
		if (it->second.getOptionKeyword() == choiceName)
			return true;
	}

	return false;
}

bool PPDFile::hasStatementWithMainKeyword(const std::string &kw) const
{
	return m_statements.find(kw) != m_statements.end();
}

void PPDFile::load(std::istream &in)
{
	Parser parser(in);
	std::string line;
	
	while (parser.getLine(line))
	{
		if (isWhitespace(line))
			continue;
		
		normaliseLineEnding(line);
		
		if (line.length() < 2) // no valid PPD line is 1 character (excluding linefeed)
			throw PPDSyntaxException(parser.getLineNumber(), "Line too short");
		
		if (line[0] != '*')
			throw PPDSyntaxException(parser.getLineNumber(), "Invalid statement, missing preceding '*'");
		
		if (line[1] == '%') // comment
			continue;
		
		PPDStatement statement = parseStatement(line, 1, parser);
		m_statements.emplace(statement.getMainKeyword(), std::move(statement));
	}
	
	if (line.length() != 0)
		throw PPDSyntaxException(parser.getLineNumber(), "Line exceeds 255 characters");
}

bool PPDFile::tryGetAnyStatementWithMainKeyword(const std::string &kw, PPDStatement &st) const
{
	auto it = m_statements.find(kw);
	if (it == m_statements.end())
	{
		st = PPDStatement();
		return false;
	}
	else
	{
		st = it->second;
		return true;
	}
}
