#ifndef RUNTIME_VALIDATOR_INCLUDED
#define RUNTIME_VALIDATOR_INCLUDED

#include <limits>
#include <cstdlib>
#include <iostream>
#include "version.h"

#if LIBTECFILTER_EXPIRY > 0

// Functions inside anonymous namespaces should have their names optimised out
namespace
{
	inline void checkTime()
	{
		time_t t = time(nullptr);
		if (t >= StripedHorse::g_expiry)
		{
			std::cerr << "ERROR: Evaluation period has elapsed" << std::endl;
			std::abort();
		}
	}
}

#ifdef __APPLE__
// Do macOS-specific checking. This is a relatively good approach. Security
// Framework on macOS is unmodifiable when System Integrity Protection is
// enabled. By calling "setuid()", we instruct DYLD_INSERT_LIBRARIES from
// being used (dyld will not honour environment variables if this function
// is used as a security meausre).
//
// Just as a backup, executables that have this compiled in should be
// codesigned with the "library" and "kill" options, which ensure that this
// binary only links against system libraries and will kill itself if it
// is dynamically modified in any way.
#include <stdexcept>
#include <CoreFoundation/CoreFoundation.h>
#include <Security/Security.h>

namespace
{
	inline void checkSignature()
	{
		// Error codes here are cryptic to avoid meaningful information landing
		// in the hands of someone trying to break this. This is a little bit
		// of "security through obscurity" but there's still no need to tell
		// people why this validation is failing.
		SecCodeRef self = nullptr;
		OSStatus result = SecCodeCopySelf(kSecCSDefaultFlags, &self);
		if (result != noErr)
			throw std::runtime_error("Invalid binary (M)"); // cannot obtain "Myself"
		
		CFDictionaryRef info = nullptr;
		result = SecCodeCopySigningInformation(self, kSecCSDefaultFlags | kSecCSSigningInformation, &info);
		if (result != noErr)
			throw std::runtime_error("Invalid binary (I)"); // cannot obtain "Information"
		
		SecRequirementRef requirement = nullptr;
		result = SecRequirementCreateWithString(CFSTR("anchor trusted"), kSecCSDefaultFlags, &requirement);
		if (result != noErr)
			throw std::runtime_error("Invalid binary (R)"); // cannot create "Requirements"
		
		result = SecCodeCheckValidity(self, kSecCSDefaultFlags, requirement);
		if (result != noErr)
			throw std::runtime_error("Invalid binary (V)"); // cannot "Verify"
		CFRelease(requirement);
		
		CFArrayRef signers = static_cast<CFArrayRef>(CFDictionaryGetValue(info, kSecCodeInfoCertificates));
		if (signers == nullptr)
			throw std::runtime_error("Invalid binary (S)"); // cannot obtain "Signing" certificates
		
		CFIndex count = CFArrayGetCount(signers);
		if (count == 0)
			throw std::runtime_error("Invalid binary (N)"); // invalid "Number" of certificates
		
		CFIndex toshibaCert = kCFNotFound;
		for (CFIndex i = 0; i < count; i++)
		{
			CFStringRef commonName = nullptr;
			SecCertificateRef cert = (SecCertificateRef) CFArrayGetValueAtIndex(signers, i);
			result = SecCertificateCopyCommonName(cert, &commonName);
			if (result != noErr)
				throw std::runtime_error("Invalid binary (C)"); // cannot obtain "Common" name
			
			CFRange certName = CFStringFind(commonName, CFSTR("Toshiba (Australia) Pty Limited"), 0);
			if (certName.location != kCFNotFound)
				toshibaCert = i;
			
			CFRelease(commonName);
		}
		
		if (toshibaCert == kCFNotFound)
			throw std::runtime_error("Invalid binary (T)"); // "Toshiba" certificate not found
		
		CFRelease(info);
		CFRelease(self);
	}
	
	inline void validateThisExecution()
	{
		try
		{
			uid_t uid = getuid();
			setuid(uid);
			checkSignature();
			checkTime();
		}
		catch (std::runtime_error &ex)
		{
			std::cerr << "ERROR: " << ex.what() << std::endl;
			std::abort();
		}
	}
}
#else
// Non-Apple implementation
//
// Do naïve check - this is ridiculously circumventable so a more
// system-specific implementation should be used if possible. It is
// implemented only as an example, if anything.
namespace
{
	inline void validateThisExecution()
	{
		checkTime();
	}
}
#endif // __APPLE__

// This will be run before main() is called. Declaring it this way means that
// this file only needs to be included (and an expiry set) in order to take
// advantage of the time-bomb
namespace
{
	struct RuntimeValidator
	{
		RuntimeValidator()
		{
			validateThisExecution();
		}
	};
	
	RuntimeValidator g_validator;
}

#endif // LIBTECFILTER_EXPIRY

#endif // RUNTIME_VALIDATOR_INCLUDED
