#include <algorithm>
#include <climits>
#include <cstdlib>
#include <fstream>
#include <limits>
#include <sstream>

#ifdef _WIN32
#include <BaseTsd.h>
#include <windows.h>
#include "winsock-helper.h"
#undef max
#undef min
#else
#include <unistd.h>
#endif

#include "tpclprinterstatus.h"

namespace
{
	void readIntoBuffer(int fd, StripedHorse::Buffer &buf)
	{
		if (buf.size() > static_cast<size_t>(std::numeric_limits<ssize_t>::max()))
			throw StripedHorse::InternalInconsistencyException("Cannot read this amount of data at once");
		
		ssize_t bytesRead = read(fd, &buf[0], buf.size());
		
		if (bytesRead < 0)
			throw StripedHorse::TPCLPrinterStatusException("Unable to read any data from the printer");
		
		if (static_cast<StripedHorse::Buffer::size_type>(bytesRead) != buf.size())
			throw StripedHorse::TPCLPrinterStatusException("Unable to read the requested number of bytes from the printer");
	}
}

StripedHorse::TPCLPrinterStatus StripedHorse::TPCLPrinterStatus::ReadPrinterStatus(int fd)
{
	fd_set fds;
	struct timeval timeout = {0, 0};
	
	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	
	if (select(fd + 1, &fds, nullptr, nullptr, &timeout) != 1)
		throw EmptyTPCLPrinterStatusException();
	
	Buffer buf(2);
	readIntoBuffer(fd, buf);
	
	if (buf[0] != '\x01' || buf[1] != '\x02')
		throw InvalidTPCLPrinterStatusException();
	
	buf.resize(7);
	readIntoBuffer(fd, buf);
	
	TPCLPrinterStatus result;
	
	result.m_statusCode = std::stoi(std::string(buf.begin(), buf.begin() + 2));
	result.m_remainingCount = std::stoi(std::string(buf.begin() + 3, buf.end()));
	
	buf.resize(4);
	readIntoBuffer(fd, buf);
	
	if (std::string(buf.begin(), buf.end()) != "\x03\x04\x0A\x0D")
		throw InvalidTPCLPrinterStatusException();
	
	return result;
}

StripedHorse::TPCLPrinterStatus::TPCLPrinterStatus()
  : m_statusCode(0), m_remainingCount(0)
{
}

StripedHorse::TPCLPrinterStatus::TPCLPrinterStatus(const TPCLPrinterStatus &ps)
  : m_statusCode(ps.m_statusCode), m_remainingCount(ps.m_remainingCount)
{
}

StripedHorse::TPCLPrinterStatus::TPCLPrinterStatus(TPCLPrinterStatus &&ps)
   : m_statusCode(ps.m_statusCode), m_remainingCount(ps.m_remainingCount)
{
}

StripedHorse::TPCLPrinterStatus &StripedHorse::TPCLPrinterStatus::operator=(const TPCLPrinterStatus &ps)
{
	m_statusCode = ps.m_statusCode;
	m_remainingCount = ps.m_remainingCount;
	return *this;
}

StripedHorse::TPCLPrinterStatus::TPCLPrinterStatus(const Buffer &b)
{
	// empty response is OK
	if (b.size() == 0)
		return;
	
	// a response of less than 13 bytes is not
	if (b.size() < 13)
		throw TPCLPrinterStatusException("Response from printer is too small");
	
	if (b[0] != '\x01' || b[1] != '\x02')
		throw TPCLPrinterStatusException("Unexpected response from printer");
	
	std::string statusCodeStr(&b[2], &b[4]);
	std::string remainingCountStr(&b[4], &b[8]);
	
	m_statusCode = std::stoi(statusCodeStr);
	m_remainingCount = std::stoi(remainingCountStr);
}

uint16_t StripedHorse::TPCLPrinterStatus::getRemainingCount() const
{
	return m_remainingCount;
}

uint16_t StripedHorse::TPCLPrinterStatus::getStatusCode() const
{
	return m_statusCode;
}
