#ifndef STRIPEDHORSE_STATUS_H
#define STRIPEDHORSE_STATUS_H

#include <cstdint>
#include <string>
#include <libcore/exceptions.h>
#include <libcore/types.h>
#include "cups-utilities.h"

namespace StripedHorse
{
	/**
	 A class that reads the status from a printer
	 */
	class TPCLPrinterStatus
	{
	public:
		/**
		 Gets a TPCLPrinterStatus object by reading the status from the specified
		 file descriptor.
		 @param[in] fd the file descriptor to read from (default value is 3,
		            which is the default file descriptor for back-channel
		            communication within CUPS.
		 @return a TPCLPrinterStatus object. If the printer status could not be
		         read, a TPCLPrinterStatusException object is thrown instead.
		 */
		static TPCLPrinterStatus ReadPrinterStatus(int fd = static_cast<int>(CUPSCommunicationFileDescriptor::kBackChannel));
		
		/**
		 Constructs a TPCLPrinterStatus object based on a pre-read buffer.
		 @param[in] b the pre-read buffer
		 */
		TPCLPrinterStatus(const Buffer &b);
		
		/**
		 Copies an existing TPCLPrinterStatus object
		 */
		TPCLPrinterStatus(const TPCLPrinterStatus &);
		
		/**
		 Moves an existing TPCLPrinterStatus object
		 */
		TPCLPrinterStatus(TPCLPrinterStatus &&);
		
		/**
		 Copies an existing TPCLPrinterStatus object
		 @param[in] ps the existing TPCLPrinterStatus object.
		 @return the result is this TPCLPrinterStatus object after copying from
		         the existing TPCLPrinterStatus object.
		 */
		TPCLPrinterStatus &operator=(const TPCLPrinterStatus &ps);
		
		/**
		 Gets the number of labels remaining
		 @return if printing was interrupted (label jam etc) or status is 
		         checked between label issues, this returns the number of
		         labels left to print. This may only return the number of
		         labels remaining if the copy count for the label issue command
		         was more than one.
		 */
		uint16_t getRemainingCount() const;
		
		/**
		 Gets the status code
		 @return the meaning of this code depends on the printer model. It may
		         mean everything is OK or may indicate an error with a specific
		         module, such as the cutter, RFID encoder etc.
		 */
		uint16_t getStatusCode() const;
		
	private:
		TPCLPrinterStatus();
		uint16_t m_statusCode;
		uint16_t m_remainingCount;
	};
	
	class TPCLPrinterStatusException : public Exception
	{
	public:
		TPCLPrinterStatusException(const std::string &msg) : Exception(msg) {}
	};
	
	class EmptyTPCLPrinterStatusException : public TPCLPrinterStatusException
	{
	public:
		EmptyTPCLPrinterStatusException() : TPCLPrinterStatusException("There was no printer status to read") {}
	};
	
	class InvalidTPCLPrinterStatusException : public TPCLPrinterStatusException
	{
	public:
		InvalidTPCLPrinterStatusException() : TPCLPrinterStatusException("The response from the printer was invalid") {}
	};
}

#endif

