#include <algorithm>
#include <tuple>
#include <libcore/logging.h>
#include "options-utilities.h"
#include "tpclwriter.h"

namespace
{
	template<class RangeType>
	struct Adjustment
	{
		std::string optionName;
		std::string format;
		RangeType range;
	};
}

StripedHorse::TPCLWriter::TPCLWriter(std::ostream &out)
  : m_out(out), m_commandStart("{"), m_commandEnd("|}"), m_putLineFeed(true)
{
}

bool StripedHorse::TPCLWriter::doesPutLineFeed() const
{
	return m_putLineFeed;
}

const std::string &StripedHorse::TPCLWriter::getCommandEnd() const
{
	return m_commandEnd;
}

const std::string &StripedHorse::TPCLWriter::getCommandStart() const
{
	return m_commandStart;
}

std::ostream &StripedHorse::TPCLWriter::getStream()
{
	return m_out;
}

void StripedHorse::TPCLWriter::setOutputStyle(const std::string &commandStart, const std::string &commandEnd, bool putLineFeed)
{
	m_commandStart = commandStart;
	m_commandEnd = commandEnd;
	m_putLineFeed = putLineFeed;
}

void StripedHorse::TPCLWriter::writeAdjustmentsFromOptions(const Options &options)
{
	using OptionsUtilities::RealValueRange;
	using OptionsUtilities::IntegerValueRange;

	// There are a few complicated rules surrounding printing these adjustments
	// because apparently negative-zero has a special meaning.
	std::vector<Adjustment<RealValueRange>> feedAdjustments = {
		{ "FeedAdjustment", "+bbb,", { -50.0, 50.0 } },
		{ "CutStripAdjustment", "+ddd,", { -50.0, 50.0 } },
		{ "BackFeedAdjustment", "+ff", { -9.9, 9.9 } }
	};

	bool hasFeedAdjustments = std::any_of(feedAdjustments.begin(), feedAdjustments.end(), [&](const auto &adjustment){
		double value;
		return options.getString(adjustment.optionName, "None") != "None" && options.tryGetDouble(adjustment.optionName, value);
	});

	// Only output AX command if there are adjustments to be made
	if (hasFeedAdjustments)
	{
		writeCommandStart("AX;");

		// Loop over feed adjustment, cut-strip adjustment and backfeed adjustment
		for (auto &adjustment : feedAdjustments)
		{
			int32_t value;
			if (!OptionsUtilities::TryGetClampedValueIn10dpmm(options, adjustment.optionName, value))
			{
				Logging::Debug() << "Model-specific range for " << adjustment.optionName << " could not be determined - using default (" <<
					adjustment.range.minimum << ", " << adjustment.range.maximum << ")" << std::endl;
				value = OptionsUtilities::GetValueIn10dpmm(options, adjustment.optionName, adjustment.range);
			}

			std::string strValue = options.getString(adjustment.optionName);
			if (strValue == "-0" || strValue == "-0.0")
			{
				writeFormat(adjustment.format, TPCLCommands::NegativeZero());
			}
			else
			{
				writeFormat(adjustment.format, value);
			}
		}

		writeCommandEnd();
	}

	// These ones are less complicated because they are not represented in a different unit
	std::vector<Adjustment<IntegerValueRange>> toneAdjustments = {
		{ "ToneAdjustmentThermalTransfer", "AY;+aa,0", { -10, 10 } },
		{ "ToneAdjustmentDirectThermal", "AY;+aa,1", { -10, 10 } }
	};

	for (auto &adjustment : toneAdjustments)
	{
		if (!options.hasOption(adjustment.optionName))
			continue;
		
		if (!OptionsUtilities::TryGetIntegerRangeForOption(options, adjustment.optionName, adjustment.range))
		{
			Logging::Debug() << "Model-specific range for " << adjustment.optionName << " could not be determined - using default (" <<
				adjustment.range.minimum << ", " << adjustment.range.maximum << ")" << std::endl;
		}
		
		const std::string &strValue = options.getString(adjustment.optionName);
		int32_t value;
		if (strValue == "-0" || strValue == "-0.0")
		{
			writeCommand(adjustment.format, TPCLCommands::NegativeZero());
		}
		else if (strValue != "None" && options.tryGetInt32(adjustment.optionName, value))
		{
			if (value < adjustment.range.minimum)
				value = adjustment.range.minimum;
			else if (value > adjustment.range.maximum)
				value = adjustment.range.maximum;
			writeCommand(adjustment.format, value);
		}
	}
}

void StripedHorse::TPCLWriter::writeCommandStart(const std::string &cmdName)
{
	for (const auto &ch : m_commandStart)
		m_out.put(ch);
	
	m_out << cmdName;
}

void StripedHorse::TPCLWriter::writeCommandEnd()
{
	for (const auto &ch : m_commandEnd)
		m_out.put(ch);
	
	if (m_putLineFeed)
		m_out.put('\n');
}

void StripedHorse::TPCLWriter::writeString(const std::string &str)
{
	m_out << str;
}
