#include <algorithm>
#include <cmath>
#include <libcore/logging.h>
#include "detectedshape.h"

namespace
{
	using PathNodeType = LemonPDF::Graphics::Path::PathNodeType;

	const std::vector<PathNodeType> g_rectanglePathNodeTypes {
		PathNodeType::kNewPath,
		PathNodeType::kLine,
		PathNodeType::kLine,
		PathNodeType::kLine,
		PathNodeType::kClosePath
	};

	template<class T, class Compare>
	constexpr const T &clamp(const T &val, const T &low, const T &high, Compare comp)
	{
		return comp(val, low) ? low : (comp(high, val) ? high : val);
	}
	
	template<class T>
	constexpr const T &clamp(const T &val, const T &low, const T &high)
	{
		return ::clamp(val, low, high, std::less<T>());
	}

	bool pathNodePatternMatch(const std::vector<LemonPDF::Graphics::Path::PathNode> &nodes, const std::vector<PathNodeType> &nodeTypes)
	{
		if (nodes.size() != nodeTypes.size())
			return false;
		
		for (std::vector<PathNodeType>::size_type i = 0; i < nodeTypes.size(); i++)
			if (nodes[i].getPathNodeType() != nodeTypes[i])
				return false;
		
		return true;
	}

	LemonPDF::Rectangle getBounds(const LemonPDF::Graphics::Path &p)
	{
		bool first = true;
		LemonPDF::Point lowerLeft;
		LemonPDF::Point upperRight;
		
		const std::vector<LemonPDF::Graphics::Path::PathNode> &pathNodes = p.getPathNodes();
		
		if (pathNodes.size() == 0)
			throw StripedHorse::InvalidArgumentException("Invalid number of points in path");
		
		for (const LemonPDF::Graphics::Path::PathNode &pathNode : pathNodes)
		{
			for (const LemonPDF::Point &p : pathNode.getPoints())
			{
				if (first)
				{
					lowerLeft = p;
					upperRight = p;
					first = false;
				}
				else
				{
					lowerLeft.x = std::min(lowerLeft.x, p.x);
					lowerLeft.y = std::min(lowerLeft.y, p.y);
					upperRight.x = std::max(upperRight.x, p.x);
					upperRight.y = std::max(upperRight.y, p.y);
				}
			}
		}
	
		return LemonPDF::Rectangle { lowerLeft, upperRight };
	}

	class DetectedRectangle : public StripedHorse::DetectedShape
	{
	public:
		static bool IsRectangle(const LemonPDF::Graphics::Path &p)
		{
			const std::vector<LemonPDF::Graphics::Path::PathNode> &pathNodes = p.getPathNodes();
			
			if (!pathNodePatternMatch(pathNodes, g_rectanglePathNodeTypes))
				return false;
			
			std::vector<LemonPDF::Point> allPoints;
			for (std::vector<LemonPDF::Graphics::Path::PathNode>::size_type i = 0; i < 4; i++)
				allPoints.emplace_back(pathNodes.at(i).getPoints().at(0));
			
			const double epsilon = 0.01;
			double hypotenuse1 = std::sqrt(std::pow(allPoints[2].x - allPoints[0].x, 2.0) + std::pow(allPoints[2].y - allPoints[0].y, 2.0));
			double hypotenuse2 = std::sqrt(std::pow(allPoints[1].x - allPoints[3].x, 2.0) + std::pow(allPoints[1].y - allPoints[3].y, 2.0));
		
			return std::abs(std::abs(hypotenuse1) - std::abs(hypotenuse2)) < epsilon;
		}

		DetectedRectangle(enum StripedHorse::FixedColor color, const LemonPDF::Rectangle &r)
		  : DetectedShape(color), m_region(r)
		{
		}

		virtual bool appendData(const LemonPDF::Point &p, const std::string &data) override
		{
			if (!pointIsInShape(p))
				return false;
			
			m_data += data;
			return true;
		}
		
		virtual const std::string &getData() const override
		{
			return m_data;
		}

		virtual std::string getName() const override
		{
			return "Rectangle";
		}

		virtual LemonPDF::Rectangle getRegion() const override
		{
			return m_region;
		}

		virtual bool pointIsInShape(const LemonPDF::Point &p) const override
		{
			return m_region.contains(p);
		}

	private:
		std::string m_data;
		LemonPDF::Rectangle m_region;
	};

	enum StripedHorse::FixedColor getFixedColor(LemonPDF::Graphics::ColorPtr c, double hueThresholdDegrees, double saturationThreshold, double valueThreshold)
	{
		double hue = NAN, saturation = 0.0, value = 0.0;

		try
		{
			std::tie(hue, saturation, value) = c->getHueSaturationValue();
		}
		catch (LemonPDF::PDFUnimplementedException &)
		{
			StripedHorse::Logging::Warning() << "Unable to obtain HSV from colour" << std::endl;
			hue = NAN;
		}
		
		if (std::isnan(hue))
			return StripedHorse::FixedColor::kOther;

		// max 30 degree threshold
		hueThresholdDegrees = clamp(hueThresholdDegrees, 0.0, 30.0);
		saturationThreshold = clamp(saturationThreshold, 0.0, 1.0);
		valueThreshold = clamp(valueThreshold, 0.0, 1.0);

		// thresholds of 0.0 mean saturation and value must be exactly 1.0
		if (saturation < (1.0 - saturationThreshold) || value < (1.0 - valueThreshold))
			return StripedHorse::FixedColor::kOther;
		
		// 000d = red
		// 060d = yellow
		// 120d = green
		// 180d = cyan
		// 240d = blue
		// 300d = magenta

		// red is special because it's right on the boundary
		if (hue >= 360.0 - hueThresholdDegrees || hue <= hueThresholdDegrees)
			return StripedHorse::FixedColor::kRed;
		else if (hue >= 60.0 - hueThresholdDegrees && hue <= 60.0 + hueThresholdDegrees)
			return StripedHorse::FixedColor::kYellow;
		else if (hue >= 120.0 - hueThresholdDegrees && hue <= 120.0 + hueThresholdDegrees)
			return StripedHorse::FixedColor::kGreen;
		else if (hue >= 180.0 - hueThresholdDegrees && hue <= 180.0 + hueThresholdDegrees)
			return StripedHorse::FixedColor::kCyan;
		else if (hue >= 240.0 - hueThresholdDegrees && hue <= 240.0 + hueThresholdDegrees)
			return StripedHorse::FixedColor::kBlue;
		else if (hue >= 300.0 - hueThresholdDegrees && hue <= 300.0 + hueThresholdDegrees)
			return StripedHorse::FixedColor::kMagenta;
		else
			return StripedHorse::FixedColor::kOther;
	}
}

StripedHorse::DetectedShapePtr StripedHorse::DetectedShape::DetectFromPath(const LemonPDF::Graphics::Path &p, LemonPDF::Graphics::ColorPtr c, double hueThresholdDegrees, double saturationThreshold, double valueThreshold)
{
	if (!DetectedRectangle::IsRectangle(p))
		return nullptr;
	
	FixedColor fc = FixedColor::kOther;
	if (c != nullptr)
	{
		fc = getFixedColor(c, hueThresholdDegrees, saturationThreshold, valueThreshold);
		if (fc == FixedColor::kOther)
			return nullptr;
	}

	return std::make_shared<DetectedRectangle>(fc, getBounds(p));
}

StripedHorse::DetectedShape::DetectedShape(FixedColor fc)
  : m_color(fc), m_angle(0.0)
{
}

bool StripedHorse::DetectedShape::appendData(const LemonPDF::Point &p, const std::string &data)
{
	(void) p;
	(void) data;
	return false;
}

double StripedHorse::DetectedShape::getAngle() const
{
	return m_angle;
}

enum StripedHorse::FixedColor StripedHorse::DetectedShape::getColor() const
{
	return m_color;
}

std::string StripedHorse::DetectedShape::getColorName() const
{
	switch (m_color)
	{
		case FixedColor::kRed:
			return "Red";
		case FixedColor::kYellow:
			return "Yellow";
		case FixedColor::kGreen:
			return "Green";
		case FixedColor::kCyan:
			return "Cyan";
		case FixedColor::kBlue:
			return "Blue";
		case FixedColor::kMagenta:
			return "Magenta";
		default:
			return "Other";
	}
}

const std::string &StripedHorse::DetectedShape::getData() const
{
	throw StripedHorse::UnimplementedException("Cannot get data from unknown shape");
}

std::string StripedHorse::DetectedShape::getName() const
{
	return "Unknown";
}

LemonPDF::Rectangle StripedHorse::DetectedShape::getRegion() const
{
	throw StripedHorse::UnimplementedException("Cannot get region from unknown shape");
}

bool StripedHorse::DetectedShape::pointIsInShape(const LemonPDF::Point &p) const
{
	(void) p;
	return false;
}

void StripedHorse::DetectedShape::setAngle(double angle)
{
	m_angle = angle;
}
