#include <cstdlib>
#include <cstring>
#include <iostream>
#include <memory>
#include <string>
#include <lemonpdf/pdffile.h>
#include <libcore/logging.h>
#include <libtecfilter/argumentparser.h>
#include <libtecfilter/files.h>
#include <libtecfilter/managedfile.h>
#include <libtecfilter/options.h>
#include <libtecfilter/options-utilities.h>
#include <libtecfilter/ppdfile.h>
#include <libtecfilter/version.h>

#ifndef _WIN32
#include <unistd.h>
#else
#include <windows.h>
#define unlink _unlink
#endif

#include "xpscontext.h"

#include "libtecfilter/runtimevalidator.inc.cpp"

namespace
{
	std::unique_ptr<std::ofstream> g_logFile;
}

int main(int argc, char *argv[])
{
	using namespace LemonPDF;

	int filenameArg = argc - 1;
	bool debugMode = false;
	bool cupsInvocation = false;
	char *softwareEnv = ::getenv("SOFTWARE");
	if (::getenv("SOFTWARE") != nullptr)
		cupsInvocation = strstr(softwareEnv, "CUPS/") == softwareEnv;

	if (::getenv("STRIPEDHORSE_DEBUG") != nullptr)
		debugMode = true;

	StripedHorse::ManagedFile infile;
	std::string outfilename = "-";

	StripedHorse::XPSContext::FontConversionMode fontConversionMode = StripedHorse::XPSContext::FontConversionMode::kConvertToPathsIfUnicodeUnavailable;
	bool useColouredRegionDetection = false;
	bool standardXPS = false;
	bool force = false;
	size_t startPage = 0;
	size_t endPage = 0;

	if (cupsInvocation)
	{
		// argv[1] = job ID
		// argv[2] = username
		// argv[3] = job title
		// argv[4] = number of copies
		// argv[5] = options
		// argv[6] = filename (nullptr = read from stdin)

		StripedHorse::Logging::SetMessagePrefixes("DEBUG", "INFO", "WARNING", "ERROR", "CRIT");

		if (argc < 5)
		{
			std::cerr << "Invalid invocation:" << std::endl;
			std::cerr << argv[0] << " job user title num-copies options [ filename ] " << std::endl;
			::exit(EXIT_FAILURE);
		}

		StripedHorse::Options options;

		char *ppdFilename = ::getenv("PPD");
		if (ppdFilename != nullptr)
		{
			try
			{
				StripedHorse::PPDFile ppd;
				std::ifstream ppdFile(ppdFilename);
				ppd.load(ppdFile);
				if (!StripedHorse::OptionsUtilities::ParseDefaultValuesFromPPD(options, ppd))
					StripedHorse::Logging::Error() << "PPD file was specified but no values were parsed" << std::endl;
			}
			catch (StripedHorse::Exception &ex)
			{
				StripedHorse::Logging::Warning() << "Unable to obtain default values from PPD (" << ex.what() << ")" << std::endl;
			}
		}
		
		options.parse(argv[5]);

		useColouredRegionDetection = options.getBoolean("ColouredRegionDetection");
		debugMode = debugMode || options.getInt32("StripedHorseDebug") > 0;

		if (options.getBoolean("ConvertAllTextToPaths"))
			fontConversionMode = StripedHorse::XPSContext::FontConversionMode::kConvertToPathsOnly;
		else if (options.getBoolean("DoNotConvertAnyTextToPaths"))
			fontConversionMode = StripedHorse::XPSContext::FontConversionMode::kConvertToUnicodeStringsOnly;

		filenameArg = 6;
	}
	else
	{
		if (argc < 2)
		{
			std::cerr << "This utility is not intended to be run manually" << std::endl;
			std::cerr << argv[0] << " job-id username title num-copies options [ filename ]" << std::endl;
			std::cerr << argv[0] << " --version" << std::endl;
			return 0;
		}
		else if (argc >= 2 && strcmp(argv[1], "--version") == 0)
		{
			std::cerr << StripedHorse::g_version << std::endl;
			return 0;
		}
		
		StripedHorse::ArgumentParser ap(argc, argv);
		
		try
		{
			ap.parse([&](char c){
				switch (c)
				{
					case 'c':
						useColouredRegionDetection = true;
						break;
					
					case 'f':
						force = true;
						break;
					
					case 'e':
						endPage = std::stoul(ap.getArgumentValue());
						break;
						
					case 'l':
						g_logFile = std::make_unique<std::ofstream>(ap.getArgumentValue(), std::ios_base::out | std::ios_base::app);
						break;

					case 'L':
						g_logFile = std::make_unique<std::ofstream>(ap.getArgumentValue(), std::ios_base::out | std::ios_base::trunc);
						break;

					case 'o':
						outfilename = ap.getArgumentValue();
						break;

					case 's':
						startPage = std::stoul(ap.getArgumentValue());
						break;
					
					case 't':
						fontConversionMode = StripedHorse::XPSContext::FontConversionMode::kConvertToPathsOnly;
						break;
					
					case 'X':
						standardXPS = true;
						break;
					
					default:
						return false;
				}

				return true;
			});
		}
		catch (std::exception &ex)
		{
			std::cerr << "Error parsing command-line arguments: " << ex.what() << std::endl;
			::exit(EXIT_FAILURE);
		}

		filenameArg = ap.getCurrentPosition();

		if (g_logFile)
			StripedHorse::Logging::SetOutputStream(g_logFile.get());
	}

	if (argv[filenameArg] != nullptr)
	{
		infile.setFilename(argv[filenameArg]);
	}
	else
	{
		StripedHorse::ManagedFile::SetManagedFileToStandardInput(infile, "pdftoxps");
		if (debugMode)
			infile.setDeleteWhenDone(false);
	}

	try
	{
		PDFFilePtr pdf = PDFFile::FromFile(infile.getFilename());
		StripedHorse::XPSContext ctx;

		ctx.setFontConversionMode(fontConversionMode);
		ctx.setShapeDetectionState(useColouredRegionDetection);
		ctx.setStandardXPSOutputMode(standardXPS);

		ctx.beginDocument();
		std::vector<PagePtr> pages = pdf->getPages();

		if (startPage == 0)
			startPage = 1;
		
		if (startPage > pages.size())
			startPage = pages.size();
		
		if (endPage == 0 || endPage > pages.size())
			endPage = pages.size();
		
		for (size_t i = startPage; i <= endPage; i++)
		{
			PagePtr p = pages.at(i - 1);
			ctx.beginPage(p->getMediaBox());
			p->render(ctx);
			ctx.endPage();
		}
		ctx.endDocument();

		if (outfilename == "-")
		{
			StripedHorse::Files::SetStandardOutputToBinaryMode();
			std::string tempfilename = StripedHorse::Files::GetTemporaryFilename("pdftoxps");
			std::fstream output(tempfilename, std::ios_base::binary | std::ios_base::out | std::ios_base::in);
			ctx.write(output);
			output.clear();
			output.seekg(0);

			std::cout << output.rdbuf();
			unlink(tempfilename.c_str());
		}
		else
		{
			std::ofstream out(outfilename, std::ios_base::binary | std::ios_base::out);
			if (!out.is_open())
				throw StripedHorse::IOException("Unable to write to " + outfilename);
			ctx.write(out);
		}
	}
	catch (PDFException &e)
	{
		StripedHorse::Logging::Critical() << "The operation failed due to a PDF-related issue: " << e.what() << std::endl;
		::exit(EXIT_FAILURE);
	}
	catch (StripedHorse::Exception &e)
	{
		StripedHorse::Logging::Critical() << "The operation failed: " << e.what() << std::endl;
		::exit(EXIT_FAILURE);
	}
	catch (LemonSqueezy::Exception &e)
	{
		StripedHorse::Logging::Critical() << "The operation failed due to a compression-related issue: " << e.what() << std::endl;
		::exit(EXIT_FAILURE);
	}
}
