#include <algorithm>
#include <cctype>
#include <cmath>
#include <exception>
#include <fstream>
#include <lemonpdf/gfxtruetypefont.h>
#include <lemonsqueezy/zip.h>
#include <libcore/logging.h>
#include <libcore/rotation.h>
#include <libcore/uuid.h>
#include <libtecfilter/stream-utilities.h>
#include "xpscontext.h"

namespace
{
	template<size_t N>
	inline const xmlChar *xStr(const char (&buffer)[N])
	{
		return reinterpret_cast<const xmlChar *>(buffer);
	}

	inline std::tuple<std::string, std::string, double> generateGlyphsAttributesAndGetStringWidth(const std::vector<LemonPDF::Graphics::SpacedText> &spacedTexts, const LemonPDF::Graphics::TextState &ts)
	{
		std::ostringstream unicodeString;
		std::ostringstream spacings;

		double grandTotalWidth = 0.0;
		char32_t spaceGlyph;
		double spaceWidth = HUGE_VAL;
		if (ts.font->getCMap().tryGetGlyphIndexForString(" ", spaceGlyph))
			spaceWidth = ts.font->getWidthsForGlyphIndices({ static_cast<char16_t>(spaceGlyph) }, nullptr);

		for (const LemonPDF::Graphics::SpacedText &st : spacedTexts)
		{
			LemonPDF::Graphics::GlyphIndices indices = ts.font->getGlyphIndicesForData(st.glyphData);
			std::vector<double> widths;

			std::string decoded = ts.font->decodeText(indices);
			double runningTotalWidth = ts.font->getWidthsForGlyphIndices(indices, &widths);

			grandTotalWidth += runningTotalWidth;
			unicodeString << decoded;

			// if the spacing is greater than the width of space, just use a space
			if (st.spacing > spaceWidth)
			{
				unicodeString << ' ';
				runningTotalWidth += spaceWidth;
				spacings << (decoded.size() + 1) << " " << runningTotalWidth << " " << (st.spacing - spaceWidth);
			}
			else
			{
				spacings << decoded.size() << " " << runningTotalWidth << " " << st.spacing;
			}
			
			spacings << "; ";
		}

		return std::make_tuple(unicodeString.str(), spacings.str(), grandTotalWidth);
	}
	
	inline int streamWriter(void *context, const char *buffer, int len)
	{
		std::ostream *output = reinterpret_cast<std::ostringstream *>(context);
		return static_cast<int>(output->rdbuf()->sputn(buffer, len));
	}

	inline int streamCloser(void *context)
	{
		(void) context;
		return 0;
	}

	void writeXMLDocToStream(xmlDocPtr doc, std::ostream &out)
	{
		xmlOutputBufferPtr outputBuffer = xmlOutputBufferCreateIO(streamWriter, streamCloser, &out, nullptr);
		int len = xmlSaveFormatFileTo(outputBuffer, doc, "UTF-8", 1);

		if (len == 0)
			throw std::runtime_error("0 bytes written!");

		if (len == -1)
			throw std::runtime_error("error writing XML to stream");
	}

	void addXMLDocToZip(xmlDocPtr doc, LemonSqueezy::ZipArchive &zip, const std::string &filename)
	{
		std::stringstream xmlData;
		writeXMLDocToStream(doc, xmlData);
		zip.addEntry(filename, xmlData);
	}

	std::string to_string(const LemonPDF::Point &p)
	{
		std::ostringstream result;
		result << p.x << "," << p.y;
		return result.str();
	}

	std::string to_string(const LemonPDF::Rectangle &r)
	{
		std::ostringstream result;
		result << r.lowerLeft.x << "," << r.lowerLeft.y << "," << r.getWidth() << "," << r.getHeight();
		return result.str();
	}

	std::string to_string(const LemonPDF::TransformMatrix &m)
	{
		std::ostringstream result;
		result << m.m[0][0] << ',';
		result << m.m[1][0] << ',';
		result << m.m[0][1] << ',';
		result << m.m[1][1] << ',';
		result << m.m[0][2] << ',';
		result << m.m[1][2];
		return result.str();
	}

	std::string abbreviatePathGeometry(const LemonPDF::Graphics::Path &path)
	{
		using namespace LemonPDF;
		using namespace LemonPDF::Graphics;

		std::ostringstream result;

		if (path.getPathNodes().size() == 0)
			StripedHorse::Logging::Info() << "Path is empty" << std::endl;
		
		for (const Path::PathNode &node : path.getPathNodes())
		{
			switch (node.getPathNodeType())
			{
				case Path::PathNodeType::kNewPath:
				{
					result << "M" << to_string(node.getPoints().at(0)) << " ";
					break;
				}

				case Path::PathNodeType::kLine:
				{
					result << "L";
					result << to_string(node.getPoints().at(0)) << " ";
					break;
				}

				case Path::PathNodeType::kBezierCurve:
				{
					result << "C";
					result << to_string(node.getPoints().at(0)) << " ";
					result << to_string(node.getPoints().at(1)) << " ";
					result << to_string(node.getPoints().at(2)) << " ";
					break;
				}

				case Path::PathNodeType::kTerminal:
				case Path::PathNodeType::kClosePath:
				{
					result << "Z";
					break;
				}
			}
		}

		return result.str();
	}

	void addStrokeDataToPathElement(xmlNodePtr path, const LemonPDF::Graphics::State &state)
	{
		if (state.strokeColor)
			xmlSetProp(path, xStr("Stroke"), BAD_CAST state.strokeColor->toRGBString().c_str());
		else
			xmlSetProp(path, xStr("Stroke"), xStr("#000000"));	
		xmlSetProp(path, xStr("StrokeThickness"), BAD_CAST std::to_string(state.lineWidth).c_str());

		switch (state.lineJoinStyle)
		{
			case LemonPDF::Graphics::LineJoinStyle::kMiterJoin:
			{
				if (state.miterLimit != 10.0)
					xmlSetProp(path, xStr("StrokeMiterLimit"), BAD_CAST std::to_string(state.miterLimit).c_str());
				break;
			}

			case LemonPDF::Graphics::LineJoinStyle::kRoundJoin:
			{
				xmlSetProp(path, xStr("StrokeLineJoin"), xStr("Round"));
				break;
			}

			case LemonPDF::Graphics::LineJoinStyle::kBevelJoin:
			{
				xmlSetProp(path, xStr("StrokeLineJoin"), xStr("Bevel"));
				break;
			}
		}

		switch (state.lineCapStyle)
		{
			case LemonPDF::Graphics::LineCapStyle::kButtCap:
			{
				// do nothing
				break;
			}

			case LemonPDF::Graphics::LineCapStyle::kRoundCap:
			{
				xmlSetProp(path, xStr("StrokeEndLineCap"), xStr("Round"));
				xmlSetProp(path, xStr("StrokeStartLineCap"), xStr("Round"));
				break;
			}

			case LemonPDF::Graphics::LineCapStyle::kProjectingSquareCap:
			{
				xmlSetProp(path, xStr("StrokeEndLineCap"), xStr("Square"));
				xmlSetProp(path, xStr("StrokeStartLineCap"), xStr("Square"));
				break;
			}
		}
	}

	bool isWhite(LemonPDF::Graphics::ColorPtr color)
	{
		// It's white if all components are close to 1
		const double epsilon = 0.0001;
		const std::vector<double> &comps = color->getComponents();
		for (double i : comps)
			if (std::abs(i - 1.0) > epsilon)
				return false;
		return true;
	}
}

StripedHorse::XPSContext::XPSContextState::XPSContextState()
  : canvas(nullptr), lastPathElement(nullptr)
{
}

StripedHorse::XPSContext::XPSContext()
  : m_documentCount(0), m_pageCount(0), m_relationshipCount(0),
    m_pageRelationships(nullptr), m_relationshipsRoot(nullptr),
    m_pageDoc(nullptr), m_fontConversionMode(FontConversionMode::kConvertToPathsIfUnicodeUnavailable),
	m_detectShapes(true), m_currentPageIsDirty(false)
{
	LIBXML_TEST_VERSION;

	m_xpsDoc.setDefaultCompressionMethod(LemonSqueezy::ZipArchiveEntry::kDeflate);

	// no need to use XML writer for such a simple file
	std::stringstream fixedDocSeq;
	fixedDocSeq << "<?xml version=\"1.0\"?>" << std::endl;
	fixedDocSeq << "<FixedDocumentSequence xmlns=\"http://schemas.microsoft.com/xps/2005/06\">" << std::endl;
	fixedDocSeq << "  <DocumentReference Source=\"Documents/1/FixedDocument.fdoc\" />" << std::endl;
	fixedDocSeq << "</FixedDocumentSequence>" << std::endl;
	
	m_xpsDoc.addEntry("FixedDocSeq.fdseq", fixedDocSeq);
	
	std::stringstream rel;
	rel << "<?xml version=\"1.0\"?>" << std::endl;
	rel << "<Relationships xmlns=\"http://schemas.openxmlformats.org/package/2006/relationships\">" << std::endl;
	rel << "  <Relationship Id=\"R0\" Type=\"http://schemas.microsoft.com/xps/2005/06/fixedrepresentation\" Target=\"FixedDocSeq.fdseq\" />" << std::endl;
	rel << "</Relationships>" << std::endl;

	m_xpsDoc.addEntry("_rels/.rels", rel);

	std::stringstream contentTypes;
	contentTypes << "<?xml version=\"1.0\"?>" << std::endl;
	contentTypes << "<Types xmlns=\"http://schemas.openxmlformats.org/package/2006/content-types\">" << std::endl;
	contentTypes << "  <Default Extension=\"rels\" ContentType=\"application/vnd.openxmlformats-package.relationships+xml\" />" << std::endl;
	contentTypes << "  <Default Extension=\"fdseq\" ContentType=\"application/vnd.ms-package.xps-fixeddocumentsequence+xml\" />" << std::endl;
	contentTypes << "  <Default Extension=\"fpage\" ContentType=\"application/vnd.ms-package.xps-fixedpage+xml\" />" << std::endl;
	contentTypes << "  <Default Extension=\"fdoc\" ContentType=\"application/vnd.ms-package.xps-fixeddocument+xml\" />" << std::endl;
	contentTypes << "  <Default Extension=\"odttf\" ContentType=\"application/vnd.ms-package.obfuscated-opentype\" />" << std::endl;
	contentTypes << "  <Default Extension=\"tif\" ContentType=\"image/tiff\" />" << std::endl;
	contentTypes << "</Types>" << std::endl;
	m_xpsDoc.addEntry("[Content_Types].xml", contentTypes);
}

StripedHorse::XPSContext::~XPSContext()
{
	if (m_pageDoc != nullptr)
		xmlFreeDoc(m_pageDoc);
}

std::string StripedHorse::XPSContext::addFontToDocument(LemonPDF::Graphics::FontPtr f)
{
	std::shared_ptr<LemonPDF::Graphics::TrueTypeFont> ttf = std::dynamic_pointer_cast<LemonPDF::Graphics::TrueTypeFont>(f);

	// embed the font even if substitution is done
	if (ttf)
	{
		std::stringstream fontDataStream;
		ttf->enableEmbeddability();
		ttf->write(fontDataStream);
		std::vector<unsigned char> fontData = StripedHorse::StreamUtilities::ReadStreamToBuffer(fontDataStream);
		
		// not allowed by XPS but generating a new UUID every time seems like it could reduce testability
		UUID uuid = UUID::Derive(fontData);
		
		if (fontData.size() < 32)
			throw XPSContextException("Font data is invalid");

		size_t order[] = { 15, 14, 13, 12, 11, 10, 9, 8, 6, 7, 4, 5, 0, 1, 2, 3 };
		for (size_t i = 0; i < 32; i++)
			fontData[i] ^= uuid.getBytes()[order[i % 16]];
		
		std::string name = static_cast<std::string>(uuid) + ".odttf";
		
		std::ostringstream fullURI;
		fullURI << "/Documents/" << m_documentCount;
		fullURI << "/Resources/Fonts/" << name;
		
		m_xpsDoc.addEntry(fullURI.str().substr(1), fontData);
		addPageRelationship(fullURI.str());

		if (m_standardXPS)
			return fullURI.str();
	}
	else if (m_standardXPS)
	{
		Logging::Error() << "Unsupported font type used in standard XPS mode" << std::endl;
		return "";
	}

	std::ostringstream fontName;
	fontName << "#substitute:";

	if (f->getFontType() & LemonPDF::Graphics::Font::FontType::kFixedPitch || f->hasFixedWidthGlyphs())
		fontName << "fixed-width:";
	else if (f->getFontType() & LemonPDF::Graphics::Font::FontType::kSerif)
		fontName << "serif:";
	else
		fontName << "sans-serif:";
	
	fontName << f->getName() << ":";
	
	if (f->getFontWeight() >= 700)
		fontName << "bold";
	if (f->getFontType() & LemonPDF::Graphics::Font::FontType::kItalic)
		fontName << "italic";
	
	return fontName.str();
}

std::string StripedHorse::XPSContext::addImageToDocument(const LemonGFX::Image &image)
{
	std::stringstream data;
	image.write(LemonGFX::Image::Format::kTIFF, data);

	std::ostringstream imageURI;
	imageURI << "Documents/" << m_documentCount;
	imageURI << "/Resources/Images/img_" << m_relationshipCount << ".tif";

	m_xpsDoc.addEntry(imageURI.str(), data);
	addPageRelationship("/" + imageURI.str());

	return imageURI.str();
}

bool StripedHorse::XPSContext::addPageRelationship(const std::string &uri, const std::string &type)
{
	if (m_allPageRelationships.insert(uri).second == false)
		return false;
	
	if (m_pageRelationships == nullptr)
	{
		m_pageRelationships = xmlNewDoc(xStr("1.0"));
		m_relationshipsRoot = xmlNewNode(nullptr, xStr("Relationships"));
		xmlNewNs(m_relationshipsRoot, xStr("http://schemas.openxmlformats.org/package/2006/relationships"), nullptr);
		xmlDocSetRootElement(m_pageRelationships, m_relationshipsRoot);
	}

	xmlNodePtr relationship = xmlNewChild(m_relationshipsRoot, nullptr, xStr("Relationship"), nullptr);
	xmlSetProp(relationship, xStr("Id"), BAD_CAST ("R" + std::to_string(m_relationshipCount)).c_str());
	xmlSetProp(relationship, xStr("Target"), BAD_CAST uri.c_str());
	xmlSetProp(relationship, xStr("Type"), BAD_CAST type.c_str());

	m_relationshipCount++;

	return true;
}

void StripedHorse::XPSContext::beginDocument()
{
	m_documentCount++;
	m_pageCount = 0;
}

void StripedHorse::XPSContext::beginPage(const LemonPDF::Rectangle &mbox)
{
	m_pageCount++;
	
	m_pageDoc = xmlNewDoc(xStr("1.0"));
	xmlNodePtr pageRoot = xmlNewNode(nullptr, xStr("FixedPage"));
	xmlNewNs(pageRoot, xStr("http://schemas.microsoft.com/xps/2005/06"), nullptr);
	m_stripedHorseNamespace = xmlNewNs(pageRoot, xStr("http://www.toshibatec.co.jp/striped-horse/2017/10"), xStr("sh"));
	
	xmlSetProp(pageRoot, xStr("Width"), BAD_CAST std::to_string(mbox.getWidth() * 4 / 3).c_str());
	xmlSetProp(pageRoot, xStr("Height"), BAD_CAST std::to_string(mbox.getHeight() * 4 / 3).c_str());
	xmlSetProp(pageRoot, xStr("xml:lang"), xStr("en-US"));
	
	xmlDocSetRootElement(m_pageDoc, pageRoot);

	m_states.clear();
	m_states.emplace_back();
	topXPSContextState().canvas = xmlNewChild(pageRoot, nullptr, xStr("Canvas"), nullptr);

	LemonPDF::TransformMatrix matrix = topState().ctm;
	matrix.translate(0.0, mbox.getHeight() * 4 / 3);
	matrix.scale(4.0 / 3.0, -4.0 / 3.0);
	xmlSetProp(topXPSContextState().canvas, xStr("RenderTransform"), BAD_CAST to_string(matrix).c_str());

	m_currentPageIsDirty = false;

	saveState();
}

void StripedHorse::XPSContext::beginText()
{
	//StandardManagedContext::beginText();
}

void StripedHorse::XPSContext::clearAllPathData()
{
	topXPSContextState().lastPathElement = nullptr;
	topXPSContextState().lastPath = topState().path;
	topState().path.clear();
}

void StripedHorse::XPSContext::concat(const LemonPDF::TransformMatrix &m)
{
	StandardManagedContext::concat(m);

	xmlNodePtr canvas = xmlNewChild(topXPSContextState().canvas, nullptr, xStr("Canvas"), nullptr);
	xmlSetProp(canvas, xStr("RenderTransform"), BAD_CAST to_string(m).c_str());
	topXPSContextState().canvas = canvas;
}

void StripedHorse::XPSContext::endDocument()
{
	// No need to use XML writer for such a simple file
	std::stringstream fdoc;
	
	fdoc << "<?xml version=\"1.0\"?>" << std::endl;
	fdoc << "<FixedDocument xmlns=\"http://schemas.microsoft.com/xps/2005/06\">" << std::endl;
	
	for (size_t i = 1; i <= m_pageCount; i++)
	{
		fdoc << "  <PageContent Source=\"Pages/" << i << ".fpage\" />" << std::endl;
	}
	
	fdoc << "</FixedDocument>" << std::endl;
	
	std::ostringstream filename;
	filename << "Documents/" << m_documentCount << "/FixedDocument.fdoc";
	m_xpsDoc.addEntry(filename.str(), fdoc);
}

void StripedHorse::XPSContext::endPage()
{
	for (const auto &detectedShapePair : m_detectedShapes)
	{
		xmlSetProp(detectedShapePair.second, xStr("Angle"), BAD_CAST std::to_string(detectedShapePair.first->getAngle()).c_str());
		xmlSetProp(detectedShapePair.second, xStr("Data"), BAD_CAST detectedShapePair.first->getData().c_str());
	}
	m_detectedShapes.clear();

	std::ostringstream fixedPageFilename;
	fixedPageFilename << "Documents/" << m_documentCount << "/Pages/" << m_pageCount << ".fpage";
	addXMLDocToZip(m_pageDoc, m_xpsDoc, fixedPageFilename.str());
	xmlFreeDoc(m_pageDoc);
	m_pageDoc = nullptr;
	
	if (m_pageRelationships != nullptr)
	{
		std::ostringstream pageRelationshipsFilename;
		pageRelationshipsFilename << "Documents/" << m_documentCount << "/Pages/_rels/" << m_pageCount << ".fpage.rels";
		addXMLDocToZip(m_pageRelationships, m_xpsDoc, pageRelationshipsFilename.str());
		xmlFreeDoc(m_pageRelationships);
		m_pageRelationships = nullptr;
		m_relationshipsRoot = nullptr;

		m_allPageRelationships.clear();
	}
	
	m_stripedHorseNamespace = nullptr;

	restoreState();
}

void StripedHorse::XPSContext::endText()
{
	//StandardManagedContext::endText();
}

void StripedHorse::XPSContext::fill(enum LemonPDF::Graphics::FillType ft)
{
	topState().path.simplify();
	outputCurrentPath(false, true, ft);
}

void StripedHorse::XPSContext::fillAndStroke(enum LemonPDF::Graphics::FillType ft)
{
	topState().path.simplify();
	outputCurrentPath(true, true, ft);
}

StripedHorse::XPSContext::FontConversionMode StripedHorse::XPSContext::getFontConversionMode() const
{
	return m_fontConversionMode;
}

bool StripedHorse::XPSContext::getShapeDetectionState() const
{
	return m_detectShapes;
}

bool StripedHorse::XPSContext::getStandardXPSOutputMode() const
{
	return m_standardXPS;
}

void StripedHorse::XPSContext::intersectClippingPath(enum LemonPDF::Graphics::FillType ft)
{
	DetectedShapePtr shape = DetectedShape::DetectFromPath(topState().path, nullptr);
	if ((!shape || shape->getName() != "Rectangle") && ft == LemonPDF::Graphics::FillType::kNonzeroWindingFill)
		Logging::Warning() << "Nonzero-winding clipping path has been substituted for even-odd" << std::endl;
	
	std::string pathString = abbreviatePathGeometry(topState().path);
	xmlNodePtr currentCanvas = topXPSContextState().canvas;
	
	if (xmlHasProp(currentCanvas, xStr("Clip")))
	{
		xmlNodePtr subCanvas = xmlNewChild(topXPSContextState().canvas, nullptr, xStr("Canvas"), nullptr);
		topXPSContextState().canvas = subCanvas;
	}
	
	xmlSetProp(topXPSContextState().canvas, xStr("Clip"), BAD_CAST pathString.c_str());
	topState().path.clear();
}

void StripedHorse::XPSContext::outputCurrentPath(bool stroke, bool fill, enum LemonPDF::Graphics::FillType ft)
{
	using namespace LemonPDF;
	using namespace LemonPDF::Graphics;
	
	if (!topState().path.isDrawable())
	{
		clearAllPathData();
		return;
	}

	if (fill && topState().fillColorSpace->getName() == "Pattern")
	{
		// the page is not explicitly marked dirty here because the pattern
		// may be a no-op.
		PatternPtr pattern = std::dynamic_pointer_cast<Pattern>(topState().fillColor);
		pattern->paint(*this);
		return;
	}

	if (!m_currentPageIsDirty && fill && isWhite(topState().fillColor) && (!stroke || isWhite(topState().strokeColor)))
	{
		clearAllPathData();
		return;
	}

	if (!m_standardXPS && m_detectShapes && fill && topState().fillColor != nullptr)
	{
		DetectedShapePtr detectedShape = DetectedShape::DetectFromPath(topState().path, topState().fillColor);
		if (detectedShape)
		{
			LemonPDF::Rectangle rect = detectedShape->getRegion();
			xmlNodePtr shape = xmlNewChild(topXPSContextState().canvas, m_stripedHorseNamespace, xStr("DetectedShape"), nullptr);
			xmlSetProp(shape, xStr("Color"), BAD_CAST detectedShape->getColorName().c_str());
			xmlSetProp(shape, xStr("Shape"), BAD_CAST detectedShape->getName().c_str());
			xmlSetProp(shape, xStr("OriginX"), BAD_CAST std::to_string(rect.lowerLeft.x).c_str());
			xmlSetProp(shape, xStr("OriginY"), BAD_CAST std::to_string(rect.lowerLeft.y).c_str());
			xmlSetProp(shape, xStr("Width"), BAD_CAST std::to_string(rect.getWidth()).c_str());
			xmlSetProp(shape, xStr("Height"), BAD_CAST std::to_string(rect.getHeight()).c_str());

			m_detectedShapes.emplace_back(detectedShape, shape);

			clearAllPathData();
			return;
		}
	}

	xmlNodePtr path = xmlNewChild(topXPSContextState().canvas, nullptr, xStr("Path"), nullptr);

	if (stroke)
	{
		addStrokeDataToPathElement(path, topState());
	}

	if (fill)
	{
		if (topState().fillColor)
			xmlSetProp(path, xStr("Fill"), BAD_CAST topState().fillColor->toRGBString().c_str());
		else
			xmlSetProp(path, xStr("Fill"), xStr("#000000"));
	}
	
	xmlNodePtr pathData = xmlNewChild(path, nullptr, xStr("Path.Data"), nullptr);
	xmlNodePtr pathGeometry = xmlNewChild(pathData, nullptr, xStr("PathGeometry"), nullptr);

	if (fill)
	{
		switch (ft)
		{
			case FillType::kEvenOddFill:
				xmlSetProp(pathGeometry, xStr("FillRule"), xStr("EvenOdd"));
				break;

			case FillType::kNonzeroWindingFill:
				xmlSetProp(pathGeometry, xStr("FillRule"), xStr("NonZero"));
				break;
		}
	}
	
	xmlNodePtr currentPath = nullptr;
	for (const Path::PathNode &node : topState().path)
	{
		switch (node.getPathNodeType())
		{
			case Path::PathNodeType::kNewPath:
			{
				const LemonPDF::Point &p = node.getPoints()[0];
				currentPath = xmlNewChild(pathGeometry, nullptr, xStr("PathFigure"), nullptr);

				if (fill)
					xmlSetProp(currentPath, xStr("IsFilled"), xStr("true"));
				
				xmlSetProp(currentPath, xStr("StartPoint"), BAD_CAST to_string(p).c_str());
				break;
			}

			case Path::PathNodeType::kLine:
			{
				const LemonPDF::Point &p = node.getPoints()[0];

				xmlNodePtr polyLineSegment = xmlNewChild(currentPath, nullptr, xStr("PolyLineSegment"), nullptr);
				xmlSetProp(polyLineSegment, xStr("Points"), BAD_CAST to_string(p).c_str());
				break;
			}

			case Path::PathNodeType::kBezierCurve:
			{
				xmlNodePtr polyBezierSegment = xmlNewChild(currentPath, nullptr, xStr("PolyBezierSegment"), nullptr);
				std::ostringstream points;
				points << to_string(node.getPoints()[0]) << " ";
				points << to_string(node.getPoints()[1]) << " ";
				points << to_string(node.getPoints()[2]);
				xmlSetProp(polyBezierSegment, xStr("Points"), BAD_CAST points.str().c_str());
				break;
			}

			case Path::PathNodeType::kClosePath:
			{
				xmlSetProp(currentPath, xStr("IsClosed"), xStr("true"));
				break;
			}

			case Path::PathNodeType::kTerminal:
			{
				break;
			}
		}
	}

	topXPSContextState().lastPathElement = path;
	topXPSContextState().lastPath = topState().path;
	topState().path.clear();
	m_currentPageIsDirty = true;
}

void StripedHorse::XPSContext::putImage(const LemonGFX::Image &image)
{
	putMaskedImage(image, LemonGFX::Image());
}

void StripedHorse::XPSContext::putMaskedImage(const LemonGFX::Image &image, const LemonGFX::Image &mask)
{
	if (m_standardXPS && mask.isValidSize())
	{
		Logging::Warning() << "Standards-compliant XPS mode does not support masked images" << std::endl;
		return;
	}

	xmlNsPtr imageBrushNamespace = nullptr;
	std::string imageURI = addImageToDocument(image);
	std::string maskURI;

	if (mask.isValidSize())
	{
		if (m_standardXPS)
		{
			Logging::Warning() << "Standards-compliant XPS mode does not support masked images" << std::endl;
		}
		else
		{
			maskURI = addImageToDocument(mask);
			imageBrushNamespace = m_stripedHorseNamespace;
		}
	}

	LemonPDF::Rectangle bounds;
	bounds.upperRight.x = image.getWidth();
	bounds.upperRight.y = image.getHeight();

	xmlNodePtr path = xmlNewChild(topXPSContextState().canvas, nullptr, xStr("Path"), nullptr);
	xmlSetProp(path, xStr("Data"), xStr("M0,0 L0,1 L1,1 L1,0 Z"));

	xmlNodePtr pathFill = xmlNewChild(path, nullptr, xStr("Path.Fill"), nullptr);

	xmlNodePtr imageBrush = xmlNewChild(pathFill, imageBrushNamespace, xStr("ImageBrush"), nullptr);
	xmlSetProp(imageBrush, xStr("ImageSource"), BAD_CAST(("/" + imageURI).c_str()));
	xmlSetProp(imageBrush, xStr("TileMode"), xStr("None"));
	xmlSetProp(imageBrush, xStr("Viewbox"), BAD_CAST to_string(bounds).c_str());
	xmlSetProp(imageBrush, xStr("ViewboxUnits"), xStr("Absolute"));
	xmlSetProp(imageBrush, xStr("Viewport"), xStr("0,-1,1,1"));
	xmlSetProp(imageBrush, xStr("ViewportUnits"), xStr("Absolute"));
	xmlSetProp(imageBrush, xStr("Transform"), xStr("1,0,0,-1,0,0"));

	if (maskURI.size() > 0)
		xmlSetProp(imageBrush, xStr("MaskSource"), BAD_CAST(("/" + maskURI).c_str()));

	m_currentPageIsDirty = true;
}

void StripedHorse::XPSContext::restoreState()
{
	StandardManagedContext::restoreState();
	m_states.pop_back();
}

void StripedHorse::XPSContext::saveState()
{
	StandardManagedContext::saveState();
	m_states.emplace_back(m_states.back());
}

void StripedHorse::XPSContext::setFontConversionMode(FontConversionMode fcm)
{
	m_fontConversionMode = fcm;
}

void StripedHorse::XPSContext::setShapeDetectionState(bool b)
{
	m_detectShapes = b;
}

void StripedHorse::XPSContext::setStandardXPSOutputMode(bool b)
{
	m_standardXPS = b;
}

void StripedHorse::XPSContext::setTextFont(LemonPDF::Graphics::FontPtr f, double fontSize)
{
	StandardManagedContext::setTextFont(f, fontSize);
	topXPSContextState().fontURI = addFontToDocument(topState().textState.font);

	LemonPDF::CMap cmap = f->getCMap();
	if (f->canConvertGlyphIndicesToText() && !cmap.hasMappings())
	{
		if (f->getEncodingName().size())
		{
			Logging::Error() << "Just making up encoding data..." << std::endl;
			for (char32_t ch = 0; ch < 256; ch++)
				if (::isprint(ch))
					cmap.addMapping(ch, std::string { static_cast<char>(ch) });
			f->setCMap(cmap);
		}
		else
		{
			f->setCanConvertGlyphIndicesToText(false);
			Logging::Error() << "No CMap for font, unable to determine textual content" << std::endl;
		}
	}
}

void StripedHorse::XPSContext::showText(const LemonPDF::Graphics::GlyphData &glyphData)
{
	showText(std::vector<LemonPDF::Graphics::SpacedText> { { 0.0, glyphData } });
}

void StripedHorse::XPSContext::showText(const std::vector<LemonPDF::Graphics::SpacedText> &spacedTexts)
{
	LemonPDF::Graphics::FontPtr font = topState().textState.font;
	if (font == nullptr || topXPSContextState().fontURI.empty())
	{
		// no point creating an empty Glyphs element
		Logging::Error() << "Text discarded because no font is set" << std::endl;
		return;
	}
	
	bool shouldFill = false;
	bool shouldStroke = false;

	switch (topState().textState.renderingMode)
	{
		case LemonPDF::Graphics::TextRenderingMode::kFillTextRenderingMode:
			shouldFill = true;
			break;

		case LemonPDF::Graphics::TextRenderingMode::kStrokeTextRenderingMode:
			shouldStroke = true;
			break;

		case LemonPDF::Graphics::TextRenderingMode::kFillAndStrokeTextRenderingMode:
			shouldFill = true;
			shouldStroke = true;
			break;

		default:
			throw XPSContextException("Unsupported text rendering mode");
	}

	if (font->canConvertGlyphIndicesToPaths() && (m_fontConversionMode == FontConversionMode::kConvertToPathsOnly || (!font->canConvertGlyphIndicesToText() && m_fontConversionMode == FontConversionMode::kConvertToPathsIfUnicodeUnavailable)))
	{
		// PDF text units are 1000 times smaller than regular page units
		
		const double fontSize = topState().textState.fontSize;
		const double fontScale = fontSize / 1000.0;

		for (auto st : spacedTexts)
		{
			double width = 0.0;

			saveState();
			
			TransformMatrix ctm = topState().textState.ctm;
			ctm.scale(fontScale, fontScale);
			concat(ctm);

			try
			{
				std::vector<char16_t> glyphIndices = font->getGlyphIndicesForData(st.glyphData);
				topState().path = font->getPathForGlyphIndices(glyphIndices, &width);
			}
			catch (LemonPDF::PDFException &ex)
			{
				StripedHorse::Logging::Error() << "Cannot convert text to paths: " << ex.what() << std::endl;
				
				if (m_fontConversionMode == FontConversionMode::kConvertToPathsOnly)
					return;
			}
			
			outputCurrentPath(shouldStroke, shouldFill, LemonPDF::Graphics::FillType::kNonzeroWindingFill);
			restoreState();
			topState().textState.ctm.translate((width * fontSize + st.spacing) / 1000.0, 0);
		}

		return;
	}

	if (!font->canConvertGlyphIndicesToText())
		return;

	std::string unicodeString, spacings;
	double stringWidth;
	std::tie(unicodeString, spacings, stringWidth) = generateGlyphsAttributesAndGetStringWidth(spacedTexts, topState().textState);

	xmlNodePtr glyphs;
	if (m_standardXPS)
	{
		glyphs = xmlNewChild(topXPSContextState().canvas, nullptr, xStr("Glyphs"), nullptr);

		std::ostringstream indicesAttribute;
		for (auto it = spacedTexts.begin(); it < spacedTexts.end(); it++)
		{
			LemonPDF::Graphics::GlyphIndices glyphIndices = topState().textState.font->getGlyphIndicesForData(it->glyphData);

			for (auto ch = glyphIndices.begin(); ch < glyphIndices.end(); ch++)
			{
				indicesAttribute << *ch;
				if (ch < glyphIndices.end() - 1)
					indicesAttribute << ";";
			}

			if (it < spacedTexts.end() - 1)
				indicesAttribute << ";";
		}

		xmlSetProp(glyphs, xStr("Indices"), BAD_CAST indicesAttribute.str().c_str());
	}
	else
	{
		PointF currentTextPoint = topState().textState.ctm * PointF { 0, 0 };
		for (DetectedShapes::const_reverse_iterator it = m_detectedShapes.rbegin(); it != m_detectedShapes.rend(); it++)
		{
			if (it->first->appendData(currentTextPoint, unicodeString))
			{
				it->first->setAngle(StripedHorse::DetectAngle(topState().textState.ctm));
				topState().textState.ctm.translate(stringWidth / 1000.0, 0);
				return;
			}
		}

		glyphs = xmlNewChild(topXPSContextState().canvas, m_stripedHorseNamespace, xStr("Glyphs"), nullptr);

		if (unicodeString.size() > 0)
			xmlSetProp(glyphs, xStr("UnicodeString"), BAD_CAST unicodeString.c_str());
		
		if (spacings.size() > 0 && !m_standardXPS)
			xmlSetProp(glyphs, xStr("Spacings"), BAD_CAST spacings.c_str());
		
		xmlSetProp(glyphs, xStr("CharacterSpacing"), BAD_CAST std::to_string(topState().textState.characterSpacing).c_str());
	}
	
	if (shouldFill)
	{	
		std::string fillColor = "#000000";
		if (topState().fillColor)
			fillColor = topState().fillColor->toRGBString();
		xmlSetProp(glyphs, xStr("Fill"), BAD_CAST fillColor.c_str());
	}

	if (shouldStroke)
		Logging::Warning() << "Stroking text is unsupported" << std::endl;
	
	LemonPDF::TransformMatrix matrix = topState().textState.ctm;
	matrix.m[1][1] *= -1.0; // because all coordinates are flipped except for text
	
	xmlSetProp(glyphs, xStr("FontRenderingEmSize"), BAD_CAST std::to_string(topState().textState.fontSize).c_str());
	xmlSetProp(glyphs, xStr("FontUri"), BAD_CAST topXPSContextState().fontURI.c_str());
	xmlSetProp(glyphs, xStr("OriginX"), xStr("0"));
	xmlSetProp(glyphs, xStr("OriginY"), xStr("0"));
	xmlSetProp(glyphs, xStr("RenderTransform"), BAD_CAST to_string(matrix).c_str());
	
	topState().textState.ctm.translate(stringWidth / 1000.0, 0);
	m_currentPageIsDirty = true;
}

void StripedHorse::XPSContext::stroke()
{
	topState().path.simplify();
	if (topXPSContextState().lastPath == topState().path)
	{
		if (topXPSContextState().lastPathElement != nullptr)
			addStrokeDataToPathElement(topXPSContextState().lastPathElement, topState());
		topState().path.clear();
	}
	else
	{
		outputCurrentPath(true, false, LemonPDF::Graphics::FillType::kEvenOddFill);
	}
}

StripedHorse::XPSContext::XPSContextState &StripedHorse::XPSContext::topXPSContextState()
{
	return m_states.back();
}

void StripedHorse::XPSContext::write(std::ostream &out)
{
	m_xpsDoc.write(out);
}

