#ifndef STRIPEDHORSE_XPSCONTEXT_H
#define STRIPEDHORSE_XPSCONTEXT_H

#include <memory>
#include <set>
#include <libxml/tree.h>
#include <lemonpdf/gfxstandardmanagedcontext.h>
#include <lemonsqueezy/zip.h>
#include "detectedshape.h"

namespace StripedHorse
{
	class XPSContext : public LemonPDF::Graphics::StandardManagedContext
	{
	public:
		enum FontConversionMode
		{
			kConvertToPathsIfUnicodeUnavailable,
			kConvertToUnicodeStringsOnly,
			kConvertToPathsOnly
		};

		XPSContext();
		~XPSContext();

		// XPS Context options
		FontConversionMode getFontConversionMode() const;
		bool getShapeDetectionState() const;
		bool getStandardXPSOutputMode() const;
		void setFontConversionMode(FontConversionMode m);
		void setShapeDetectionState(bool b);
		void setStandardXPSOutputMode(bool b);

		// XPS Context specific methods
		void beginDocument();
		void beginPage(const LemonPDF::Rectangle &size);
		void endDocument();
		void endPage();

		// LemonPDF::IContext methods
		void beginText() override;
		void concat(const LemonPDF::TransformMatrix &ctm) override;
		void endText() override;
		void fill(enum LemonPDF::Graphics::FillType) override;
		void fillAndStroke(enum LemonPDF::Graphics::FillType) override;
		void intersectClippingPath(enum LemonPDF::Graphics::FillType) override;
		void putImage(const LemonGFX::Image &) override;
		void putMaskedImage(const LemonGFX::Image &, const LemonGFX::Image &) override;
		void restoreState() override;
		void saveState() override;
		void setTextFont(LemonPDF::Graphics::FontPtr font, double fontSize) override;
		void showText(const LemonPDF::Graphics::GlyphData &glyphData) override;
		void showText(const std::vector<LemonPDF::Graphics::SpacedText> &spacedText) override;
		void stroke() override;

		void write(std::ostream &);
	
	private:
		using DetectedShapePair = std::pair<DetectedShapePtr, xmlNodePtr>;
		using DetectedShapes = std::vector<DetectedShapePair>;

		struct XPSContextState
		{
			XPSContextState();

			std::string fontURI;
			xmlNodePtr canvas;
			xmlNodePtr lastPathElement;
			LemonPDF::Graphics::Path lastPath;
		};

		bool addDocumentRelationship(const std::string &uri, const std::string &type = "http://schemas.microsoft.com/xps/2005/06/required-resource");
		std::string addFontToDocument(LemonPDF::Graphics::FontPtr f);
		std::string addImageToDocument(const LemonGFX::Image &img);
		bool addPageRelationship(const std::string &uri, const std::string &type = "http://schemas.microsoft.com/xps/2005/06/required-resource");
		void clearAllPathData();
		DetectedShapes::const_iterator findDetectedShapeUnderPoint(const LemonPDF::Point &p) const;
		void outputCurrentPath(bool stroke, bool fill, enum LemonPDF::Graphics::FillType);
		XPSContextState &topXPSContextState();

		size_t m_documentCount;
		size_t m_pageCount;
		size_t m_relationshipCount;

		xmlDocPtr m_pageRelationships;
		xmlNodePtr m_relationshipsRoot;
		xmlNsPtr m_stripedHorseNamespace;
		std::set<std::string> m_allPageRelationships;
		
		xmlDocPtr m_pageDoc;
		std::vector<XPSContextState> m_states;

		FontConversionMode m_fontConversionMode;

		bool m_detectShapes;
		DetectedShapes m_detectedShapes;

		bool m_standardXPS;

		bool m_currentPageIsDirty;

		LemonSqueezy::ZipArchive m_xpsDoc;
	};

	class XPSContextException : public Exception
	{
	public:
		XPSContextException(const std::string &msg) : Exception(msg) {}	
	};
}

#endif

