#include <libusb-1.0/libusb.h>

namespace
{
	bool endpointIsSuitableForWriting(const struct libusb_endpoint_descriptor &endpointDescriptor)
	{
		return ((endpointDescriptor.bmAttributes & LIBUSB_TRANSFER_TYPE_MASK) == LIBUSB_TRANSFER_TYPE_BULK)
		    && ((endpointDescriptor.bEndpointAddress & LIBUSB_ENDPOINT_DIR_MASK) == LIBUSB_ENDPOINT_OUT);
	}

	USBPortPath getPortPath(libusb_device *device)
	{
		USBPortPath portPath { 0, 0, 0, 0, 0, 0, 0, 0 };
		
		int len = libusb_get_port_numbers(device, &portPath[1], portPath.size() - 1);

		if (len < 0 || len > 7)
			throw std::runtime_error("USB port path is not a valid length (must be between 1 and 7 inclusive)");
		portPath[0] = len;

		for (size_t i = len + 1; i < portPath.size(); i++)
			portPath[i] = 0;
		
		return portPath;
	}

	class LibUSBDevice
	{
	public:
		LibUSBDevice(libusb_device *device)
		  : m_device(libusb_ref_device(device)), m_handle(nullptr), m_reattach(false)
		{
			libusb_get_device_descriptor(m_device, &m_descriptor);	
		}

		LibUSBDevice(const LibUSBDevice &other) = delete;

		LibUSBDevice(LibUSBDevice &&other)
		{
			m_device = other.m_device;
			m_handle = other.m_handle;
			m_descriptor = other.m_descriptor;
			m_reattach = other.m_reattach;

			other.m_device = nullptr;
			other.m_handle = nullptr;
			other.m_reattach = false;
		}

		void close()
		{
			if (m_handle)
			{
				if (m_reattach)
					libusb_attach_kernel_driver(m_handle, 0);
			
				libusb_close(m_handle);
			}

			m_handle = nullptr;
		}

		std::string getDescription()
		{
			auto parsed = parseIEEE1284DeviceID(getIEEE1284DeviceID());

			std::ostringstream description;
			description << "TOSHIBA ";

			auto modelIter = parsed.find("MDL");
			if (modelIter == parsed.end())
				modelIter = parsed.find("MODEL");
			
			if (modelIter == parsed.end())
				description << "Barcode Printer";
			else
				description << modelIter->second;
			
			return description.str();
		}

		USBDeviceID getDeviceID() const
		{
			return { m_descriptor.idVendor, m_descriptor.idProduct };
		}

		std::string getIEEE1284DeviceID()
		{
			const static std::string &defaultDeviceID = "MFG:TOSHIBA;MDL:Barcode Printer;CMD:TPCL;";
			int result = open();
			if (result != 0)
				return defaultDeviceID;
			
			uint8_t response[1024] = {0};
			uint8_t bmRequestType = LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_INTERFACE | LIBUSB_ENDPOINT_IN;
			result = libusb_control_transfer(m_handle, bmRequestType, 0, 0, 0, response, sizeof response, 10000);

			uint16_t length = response[0] << 8 | response[1];

			// IEEE1284 says that the length includes the two length bytes itself
			if (length < 2)
				return defaultDeviceID;
			length -= 2;

			const char *buf = reinterpret_cast<char *>(response + 2);
			return preprocessIEEE1284DeviceID(std::string { buf, buf + length - 2 });
		}

		uint8_t getWriteEndpoint()
		{
			uint8_t endpoint = 0;
			
			for (int c = 0; c < m_descriptor.bNumConfigurations; c++)
			{
				struct libusb_config_descriptor *configuration = nullptr;
				int result = libusb_get_config_descriptor(m_device, c, &configuration);
				if (result < 0)
					continue;
				
				for (int i = 0; i < configuration->bNumInterfaces; i++)
				{
					const struct libusb_interface *interface = &configuration->interface[i];
					for (int altSetting = 0; altSetting < interface->num_altsetting; altSetting++)
					{
						const struct libusb_interface_descriptor *interfaceDescriptor = &interface->altsetting[altSetting];
						for (uint8_t ep = 0; ep < interfaceDescriptor->bNumEndpoints; ep++)
							if (endpointIsSuitableForWriting(interfaceDescriptor->endpoint[ep]))
								endpoint = interfaceDescriptor->endpoint[ep].bEndpointAddress;
					}
				}

				libusb_free_config_descriptor(configuration);
			}

			return endpoint;
		}

		int open()
		{
			if (m_handle)
				return 0;

			int status = libusb_open(m_device, &m_handle);
			if (status != 0)
			{
				StripedHorse::Logging::Error() << "Unable to query device: " << libusb_strerror(static_cast<libusb_error>(status)) << std::endl;
				return status;
			}

			if (libusb_kernel_driver_active(m_handle, 0))
			{
				StripedHorse::Logging::Debug() << "Kernel driver is attached, attempting to detach" << std::endl;
				status = libusb_detach_kernel_driver(m_handle, 0);
				if (status != 0)
				{
					StripedHorse::Logging::Error() << "Unable to detach kernel driver" << std::endl;
					return status;
				}
				
				m_reattach = true;
				StripedHorse::Logging::Debug() << "Kernel driver successfully detached" << std::endl;
			}

			status = libusb_set_interface_alt_setting(m_handle, 0, 0);

			return status;
		}

		int setConfiguration(uint8_t configNumber = 1)
		{
			int status = open();
			if (status != 0)
				return status;
			
			status = libusb_set_configuration(m_handle, configNumber);
			if (status < 0)
				StripedHorse::Logging::Error() << "Unable to set device configuration" << std::endl;

			return status;
		}

		bool tryConstructURI(StripedHorse::URI &uri)
		{
			USBPortPath portPath = getPortPath(m_device);

			uint64_t uid = 0;
			for (size_t i = 0; i < portPath.size(); i++)
				uid |= static_cast<uint64_t>(portPath[i]) << ((7 - i) * 8);

			std::ostringstream uriString;
			uriString << "tec://" << uid << "/?p=" << m_descriptor.idProduct;
			uri = StripedHorse::URI(uriString.str());
			return true;
		}

		bool tryGetString(uint8_t id, std::string &out)
		{
			if (!open())
				return false;
			
			// some devices apparently choke on strings larger than 255
			std::vector<unsigned char> buffer(255);
			int status = libusb_get_string_descriptor_ascii(m_handle, id, buffer.data(), buffer.size());
			if (static_cast<size_t>(status) <= buffer.size())
			{
				out = std::string { buffer.begin(), buffer.begin() + status };
				return true;
			}

			return false;   
		}

		std::tuple<size_t, size_t> write(std::istream &in)
		{
			int result = open();
			if (result != 0)
			{
				StripedHorse::Logging::Critical() << "Unable to open USB device:" << std::endl;
				StripedHorse::Logging::Critical() << libusb_strerror(static_cast<libusb_error>(result)) << std::endl;
				return std::make_tuple(0, 0);
			}

			result = libusb_claim_interface(m_handle, 0);
			if (result != 0)
			{
				StripedHorse::Logging::Critical() << "Unable to claim USB device interface" << std::endl;
				StripedHorse::Logging::Critical() << libusb_strerror(static_cast<libusb_error>(result)) << std::endl;
				return std::make_tuple(0, 0);
			}

			uint8_t writeEndpoint = getWriteEndpoint();
			StripedHorse::Logging::Debug() << "Write endpoint is: " << static_cast<unsigned>(writeEndpoint) << std::endl;

			size_t totalBytesRead = 0;
			size_t totalBytesWritten = 0;
			while (in)
			{
				unsigned char buffer[4096];
				size_t bytesRead = 0;
				in.read(reinterpret_cast<char *>(&buffer[0]), sizeof buffer);

				bytesRead = in.gcount();
				totalBytesRead += bytesRead;

				int bytesToWrite = static_cast<int>(bytesRead);
				StripedHorse::Logging::Debug() << "Read " << bytesRead << " bytes..." << std::endl;
				
				unsigned char *pBuffer = buffer;
				while (bytesToWrite > 0)
				{
					int bytesWritten = 0;
					StripedHorse::Logging::Debug() << "Attempting to write " << bytesToWrite << " bytes..." << std::endl;
					result = libusb_bulk_transfer(m_handle, writeEndpoint, pBuffer, bytesToWrite, &bytesWritten, 10000);
					if (bytesWritten == 0)
					{
						StripedHorse::Logging::Critical() << "Unable to write data to USB device: " << std::endl;;
						StripedHorse::Logging::Critical() << libusb_strerror(static_cast<libusb_error>(result)) << std::endl;
						break;
					}

					StripedHorse::Logging::Debug() << "Wrote " << bytesWritten << " bytes..." << std::endl;
					bytesToWrite -= bytesWritten;
					totalBytesWritten += bytesWritten;
					pBuffer += bytesWritten;
				}
			}

			libusb_release_interface(m_handle, 0);
			return std::make_tuple(totalBytesRead, totalBytesWritten);
		}

		~LibUSBDevice()
		{
			close();
			if (m_device)
				libusb_unref_device(m_device);
		}

	private:
		libusb_device *m_device;
		libusb_device_handle *m_handle;
		libusb_device_descriptor m_descriptor;
		bool m_reattach;
	};

	class LibUSBManager
	{
	public:
		LibUSBManager() : m_context(nullptr)
		{
			if (libusb_init(&m_context) != 0)
				throw StripedHorse::InternalInconsistencyException("Unable to utilise libusb");
			libusb_set_debug(m_context, LIBUSB_LOG_LEVEL_NONE);
		}

		bool tryGetUSBDevices(const USBDeviceIDSet &supportedDevices, std::vector<LibUSBDevice> &result, const USBPortPath &portPath = { 0 })
		{
			libusb_device **devices = nullptr;
			ssize_t count = libusb_get_device_list(m_context, &devices);
			if (count <= 0)
				return false;

			for (ssize_t i = 0; i < count; i++)
			{
				libusb_device_descriptor descriptor;
				libusb_get_device_descriptor(devices[i], &descriptor);
				if (supportedDevices.contains(descriptor.idVendor, descriptor.idProduct))
				{
					if (portPath[0] == 0 || getPortPath(devices[i]) == portPath)
						result.emplace_back(devices[i]);
				}
			}

			libusb_free_device_list(devices, 1);
			return count > 0;
		}

		~LibUSBManager()
		{
			libusb_exit(m_context);
		}
	
	private:
		libusb_context *m_context;
	};

	using USBManager = LibUSBManager;
	using USBDevice = LibUSBDevice;
}