#include <usb.h>
#include <cstdlib>
#include <cstring>
#include <dirent.h>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <syslog.h>

namespace
{
	struct OldLibUSBQueryInformationSet
	{
	public:
		void add(uint8_t busNumber, uint8_t deviceNumber, const std::string &devicePath)
		{
			m_queryInformation[busNumber][deviceNumber] = devicePath;
		}

		bool contains(uint8_t busNumber) const
		{
			return m_queryInformation.find(busNumber) != m_queryInformation.end();
		}

		bool contains(uint8_t busNumber, uint8_t deviceNumber)
		{
			auto busIter = m_queryInformation.find(busNumber);
			if (busIter == m_queryInformation.end())
				return false;
			
			return busIter->second.find(deviceNumber) != busIter->second.end();
		}

		const std::string &getDevicePath(uint8_t busNumber, uint8_t deviceNumber) const
		{
			auto busIter = m_queryInformation.find(busNumber);
			if (busIter == m_queryInformation.end())
				throw std::out_of_range("busNumber");
			
			auto devIter = busIter->second.find(deviceNumber);
			if (devIter == busIter->second.end())
				throw std::out_of_range("deviceNumber");
			
			return devIter->second;
		}
		
		size_t size() const
		{
			size_t total = 0;
			for (const auto &info : m_queryInformation)
				total += info.second.size();
			return total;
		}

	private:
		std::unordered_map<uint8_t, std::unordered_map<uint8_t, std::string>> m_queryInformation;
	};

	bool readSysFile(const std::string &filename, std::string &result)
	{
		std::ifstream input(filename);
		return static_cast<bool>(std::getline(input, result));
	}

	enum class Base : int
	{
		kDecimal = 10,
		kHexadecimal = 16
	};

	template<typename T>
	bool readSysFileValueAsInteger(const std::string &filename, T &result, Base b)
	{
		std::string value;
		if (!readSysFile(filename, value))
			return false;
		
		size_t index;
		result = static_cast<T>(stoull(value, &index, static_cast<int>(b)));
		return index > 0;
	}

	USBPortPath parsePortPath(uint8_t busNumber, const std::string &devicePath)
	{
		syslog(LOG_DEBUG, "Parsing port path from: %s", devicePath.c_str());
	
		USBPortPath result = { 0, busNumber, 0, 0, 0, 0, 0, 0 };;
		uint8_t pathLength = 1;

		std::istringstream pathStream(devicePath);
		std::string pathComponent;

		while (std::getline(pathStream, pathComponent, '.'))
		{
			if (pathLength == 7)
				throw std::out_of_range("Device path is too long!");
			
			result[++pathLength] = static_cast<uint8_t>(std::stoul(pathComponent));
		}

		result[0] = pathLength;
		return result;
	}
	
	// The only way to get the physical topology of the USB connections is to query
	// the information directly from the filesystem, because it is not exposed through
	// oldlibusb.
	OldLibUSBQueryInformationSet enumerateUSBQueryInformation(const USBDeviceIDSet &supportedDevices)
	{
		const std::string &sysPath = "/sys/bus/usb/devices";

		OldLibUSBQueryInformationSet queryInformation;
		DIR *d = opendir(sysPath.c_str());

		struct dirent entry;
		struct dirent *result;
		while (readdir_r(d, &entry, &result) == 0)
		{
			if (result == nullptr)
				break;
			
			if (strcmp(result->d_name, ".") == 0 || strcmp(result->d_name, "..") == 0)
				continue;
			
			std::string thisPath = sysPath + "/" + result->d_name;
			uint16_t vendorID, productID;
			if (!(readSysFileValueAsInteger(thisPath + "/idVendor", vendorID, Base::kHexadecimal) && readSysFileValueAsInteger(thisPath + "/idProduct", productID, Base::kHexadecimal)))
				continue;
			
			if (!supportedDevices.contains(vendorID, productID))
				continue;
			
			uint8_t busNumber, deviceNumber;
			std::string devicePath;
			
			if (!readSysFileValueAsInteger(thisPath + "/devnum", deviceNumber, Base::kDecimal))
			{
				syslog(LOG_WARNING, "Unable to get the device number (%s)", result->d_name);
				continue;
			}
			
			if (!(readSysFileValueAsInteger(thisPath + "/busnum", busNumber, Base::kDecimal) || readSysFileValueAsInteger(thisPath + "/bus/devnum", busNumber, Base::kDecimal)))
			{
				char *endptr = nullptr;
				busNumber = static_cast<uint8_t>(strtoul(result->d_name, &endptr, 10));
				if (endptr == nullptr || *endptr != '-')
				{
					syslog(LOG_WARNING, "Unable to determine the bus number of the device (%s)", result->d_name);
					continue;
				}
				
				devicePath = endptr + 1;
			}
			else if (!readSysFile(thisPath + "/devpath", devicePath))
			{
				syslog(LOG_WARNING, "Unable to get the device path (%s)", result->d_name);
				continue;
			}
			
			queryInformation.add(busNumber, deviceNumber, devicePath);
		}

		closedir(d);

		return queryInformation;
	}

	bool endpointIsSuitableForWriting(const struct usb_endpoint_descriptor &endpointDescriptor)
	{
		return ((endpointDescriptor.bmAttributes & USB_ENDPOINT_TYPE_MASK) == USB_ENDPOINT_TYPE_BULK)
		    && ((endpointDescriptor.bEndpointAddress & USB_ENDPOINT_DIR_MASK) == 0);
	}

	class OldLibUSBDevice
	{
	public:
		OldLibUSBDevice(struct usb_device *device, const USBPortPath &portPath)
		  : m_device(device), m_handle(nullptr), m_reattach(false), m_portPath(portPath)
		{
		}

		OldLibUSBDevice(const OldLibUSBDevice &other) = delete;

		OldLibUSBDevice(OldLibUSBDevice &&other)
		  : m_device(other.m_device), m_handle(other.m_handle), m_reattach(other.m_reattach), m_portPath(other.m_portPath)
		{
			other.m_device = nullptr;
			other.m_handle = nullptr;
		}

		~OldLibUSBDevice()
		{
			close();
		}

		void close()
		{
			if (m_handle)
				usb_close(m_handle);

			m_handle = nullptr;
		}

		std::string getDescription()
		{
			auto parsed = parseIEEE1284DeviceID(getIEEE1284DeviceID());

			std::ostringstream description;
			description << "TOSHIBA ";

			auto modelIter = parsed.find("MDL");
			if (modelIter == parsed.end())
				modelIter = parsed.find("MODEL");
			
			if (modelIter == parsed.end())
				description << "Barcode Printer";
			else
				description << modelIter->second;
			
			return description.str();
		}

		USBDeviceID getDeviceID() const
		{
			return { m_device->descriptor.idVendor, m_device->descriptor.idProduct };
		}

		std::string getIEEE1284DeviceID()
		{
			static const std::string &defaultDeviceID = "MFG:TOSHIBA;MDL:Barcode Printer;CMD:TPCL;";
			int result = open();
			if (result != 0)
				return defaultDeviceID;
			
			result = usb_claim_interface(m_handle, 0);
			if (result != 0)
			{
				StripedHorse::Logging::Critical() << "Unable to claim USB device interface" << std::endl;
				StripedHorse::Logging::Critical() << usb_strerror() << std::endl;
				return defaultDeviceID;
			}

			char response[1024] = {0};
			int requestType = USB_TYPE_CLASS | USB_RECIP_INTERFACE | USB_ENDPOINT_IN;
			result = usb_control_msg(m_handle, requestType, 0, 0, 0, response, sizeof response, 10000);
			usb_release_interface(m_handle, 0);

			if (result < 0)
				return defaultDeviceID;
			
			uint16_t length = static_cast<uint8_t>(response[0]) << 8U | static_cast<uint8_t>(response[1]);
			const char *buf = response + 2;

			// IEEE1284 says that the length includes the two length bytes itself
			if (length < 2)
				return defaultDeviceID;
			length -= 2;

			return preprocessIEEE1284DeviceID(std::string { buf, buf + length });
		}

		std::tuple<uint8_t, uint8_t, uint8_t, uint8_t> getWriteEndpoint()
		{
			for (int c = 0; c < m_device->descriptor.bNumConfigurations; c++)
			{
				struct usb_config_descriptor *configuration = &m_device->config[c];
				
				for (int i = 0; i < configuration->bNumInterfaces; i++)
				{
					const struct usb_interface *interface = &configuration->interface[i];
					for (int altSetting = 0; altSetting < interface->num_altsetting; altSetting++)
					{
						const struct usb_interface_descriptor *interfaceDescriptor = &interface->altsetting[altSetting];
						for (uint8_t ep = 0; ep < interfaceDescriptor->bNumEndpoints; ep++)
						{
							if (endpointIsSuitableForWriting(interfaceDescriptor->endpoint[ep]))
							{
								StripedHorse::Logging::Debug() << "Config " << static_cast<int>(configuration->bConfigurationValue)
								                               << ", interface " << static_cast<int>(interfaceDescriptor->bInterfaceNumber)
								                               << ", altSetting " << static_cast<int>(interfaceDescriptor->bAlternateSetting)
								                               << ", bAlternateSetting " << static_cast<int>(interfaceDescriptor->bAlternateSetting)
								                               << ", ep " << static_cast<int>(interfaceDescriptor->endpoint[ep].bEndpointAddress) << std::endl;

								return std::make_tuple(configuration->bConfigurationValue,
								                       interfaceDescriptor->bInterfaceNumber,
								                       interfaceDescriptor->bAlternateSetting,
								                       interfaceDescriptor->endpoint[ep].bEndpointAddress);
							}
						}
					}
				}
			}

			return std::make_tuple<uint8_t, uint8_t, uint8_t, uint8_t>(0, 0, 0, 0);
		}

		int open()
		{
			if (m_handle)
				return 0;
			
			m_handle = usb_open(m_device);
			if (!m_handle)
			{
				StripedHorse::Logging::Error() << "Unable to query device: " << usb_strerror() << std::endl;
				return -1;
			}
			
			// libusb-0.1 always returns 0 for this function, however, libusb-compat returns the value of
			// libusb_detach_kernel_driver() which CAN return non-zero if the kernel driver is already
			// detached.
			int status = usb_detach_kernel_driver_np(m_handle, 0);
			if (status != 0)
				StripedHorse::Logging::Warning() << "Unable to detach kernel driver (might be already detached): " << usb_strerror() << std::endl;

			return 0;
		}

		int setConfiguration(uint8_t configNumber = 1)
		{
			// set the correct configuration
			int status = usb_set_configuration(m_handle, configNumber);
			if (status != 0)
				StripedHorse::Logging::Error() << "Unable to set device configuration: " << strerror(errno) << " (USB: " << usb_strerror() << ")" << std::endl;

			return status;
		}

		bool tryConstructURI(StripedHorse::URI &uri)
		{
			uint64_t uid = 0;
			for (size_t i = 0; i < m_portPath.size(); i++)
				uid |= static_cast<uint64_t>(m_portPath[i]) << ((7 - i) * 8);

			std::ostringstream uriString;
			uriString << "tec://" << uid << "/?p=" << m_device->descriptor.idProduct;
			uri = StripedHorse::URI(uriString.str());
			return true;
		}

		bool tryGetString(uint8_t id, std::string &out)
		{
			if (!open())
				return false;
			
			// some devices apparently choke on strings larger than 255
			std::vector<char> buffer(255);
			int status = usb_get_string_simple(m_handle, id, buffer.data(), buffer.size());
			if (static_cast<size_t>(status) <= buffer.size())
			{
				out = std::string { buffer.begin(), buffer.begin() + status };
				return true;
			}

			return false;   
		}

		std::tuple<size_t, size_t> write(std::istream &in)
		{
			StripedHorse::Logging::Debug() << "write() method" << std::endl;
			
			int result = open();
			if (result != 0)
			{
				StripedHorse::Logging::Critical() << "Unable to open USB device:" << std::endl;
				StripedHorse::Logging::Critical() << usb_strerror() << std::endl;
				return std::make_tuple(0, 0);
			}
			
			StripedHorse::Logging::Debug() << "Device opened successfully" << std::endl;

			uint8_t configurationValue, interfaceValue, altSetting, writeEndpoint;
			std::tie(configurationValue, interfaceValue, altSetting, writeEndpoint) = getWriteEndpoint();

			setConfiguration(configurationValue);

			result = usb_claim_interface(m_handle, interfaceValue);
			if (result != 0)
			{
				StripedHorse::Logging::Critical() << "Unable to claim USB device interface" << std::endl;
				StripedHorse::Logging::Critical() << usb_strerror() << std::endl;
				return std::make_tuple(0, 0);
			}
			
			StripedHorse::Logging::Debug() << "Interface ready" << std::endl;

			char response[1024] = {0};
			int requestType = USB_TYPE_CLASS | USB_RECIP_INTERFACE | USB_ENDPOINT_IN;
			result = usb_control_msg(m_handle, requestType, 0, 0, 0, response, sizeof response, 10000);
			if (result < 0)
				StripedHorse::Logging::Warning() << "Unable to communicate control message" << std::endl;

			StripedHorse::Logging::Debug() << "Write endpoint is: " << static_cast<unsigned>(writeEndpoint) << std::endl;

			size_t totalBytesRead = 0;
			size_t totalBytesWritten = 0;
			while (in)
			{
				unsigned char buffer[4096];
				size_t bytesRead = 0;

				in.read(reinterpret_cast<char *>(&buffer[0]), sizeof buffer);
				bytesRead = in.gcount();
				totalBytesRead += bytesRead;

				StripedHorse::Logging::Debug() << "Read " << bytesRead << " bytes..." << std::endl;

				// cast to int type because that's what the API wants (for some reason)
				int bytesToWrite = static_cast<int>(bytesRead);
				
				unsigned char *pBuffer = buffer;
				while (bytesToWrite > 0)
				{
					StripedHorse::Logging::Debug() << "Attempting to write " << bytesToWrite << " bytes..." << std::endl;
					int bytesWritten = usb_bulk_write(m_handle, writeEndpoint, reinterpret_cast<char *>(pBuffer), bytesToWrite, 10000);
					if (bytesWritten <= 0)
					{
						StripedHorse::Logging::Critical() << "Unable to write data to USB device: " << std::endl;;
						StripedHorse::Logging::Critical() << usb_strerror() << std::endl;
						break;
					}

					StripedHorse::Logging::Debug() << "Wrote " << bytesWritten << " bytes..." << std::endl;
					bytesToWrite -= bytesWritten;
					totalBytesWritten += bytesWritten;
					pBuffer += bytesWritten;
				}
			}

			usb_release_interface(m_handle, 0);
			return std::make_tuple(totalBytesRead, totalBytesWritten);
		}

	private:
		struct usb_device *m_device;
		usb_dev_handle *m_handle;
		bool m_reattach;
		USBPortPath m_portPath;
	};

	class OldLibUSBManager
	{
	public:
		OldLibUSBManager()
		{
			usb_init();
			usb_find_busses();
			usb_find_devices();
		}

		~OldLibUSBManager()
		{
		}

		bool tryGetUSBDevices(const USBDeviceIDSet &supportedDevices, std::vector<OldLibUSBDevice> &result, const USBPortPath &targetPortPath = { 0 })
		{
			OldLibUSBQueryInformationSet queryInformation = enumerateUSBQueryInformation(supportedDevices);
			size_t devicesFound = queryInformation.size();
			
			syslog(LOG_INFO, "Found %lu devices", static_cast<unsigned long>(devicesFound));
			
			if (devicesFound == 0)
				return false;
			
			struct usb_bus *bus = usb_get_busses();

			while (bus)
			{
				// libusb-0.1 does not set this for Linux, but libusb-compat does
				uint8_t location = bus->location;
				if (location == 0)
				{
					char *endptr;
					location = strtoul(bus->dirname, &endptr, 10);
					if (endptr != nullptr && *endptr == '\0')
					{
						syslog(LOG_DEBUG, "Using bus location #%d from dirname", location);
					}
					else if (bus->root_dev)
					{
						location = bus->root_dev->devnum;
						syslog(LOG_DEBUG, "Bus location not available from libusb, using bus #%d from root device", location);
					}
					else
					{
						syslog(LOG_ERR, "Unable to determine bus ID from any available source, cannot continue!");
						break;	
					}
				}
				
				struct usb_device *device = bus->devices;
				while (device)
				{
					uint8_t deviceNumber = device->devnum;
					if (deviceNumber == 0)
					{
						char *endptr;
						deviceNumber = strtoul(device->filename, &endptr, 10);
						if (endptr == nullptr || *endptr != '\0')
						{
							syslog(LOG_ERR, "Unable to determine device ID from any available source, cannot continue!");
							break;
						}
						
						syslog(LOG_DEBUG, "Using device number #%d from filename", deviceNumber);
					}
					
					if (queryInformation.contains(location, deviceNumber))
					{
						USBPortPath currentPortPath = parsePortPath(location, queryInformation.getDevicePath(location, deviceNumber));
						if (targetPortPath[0] == 0 || targetPortPath == currentPortPath)
						{
							result.emplace_back(device, currentPortPath);
						}
						else if (targetPortPath != currentPortPath)
						{
							std::string targetPortPathString = stringifyPortPath(targetPortPath);
							std::string currentPortPathString = stringifyPortPath(currentPortPath);
							syslog(LOG_DEBUG, "Found %s but looking for %s, skipping device", currentPortPathString.c_str(), targetPortPathString.c_str());
						}	
					}
					
					device = device->next;
				}

				bus = bus->next;
			}
			
			return result.size() > 0;
		}
	};

	using USBManager = OldLibUSBManager;
	using USBDevice = OldLibUSBDevice;
}
