#include <array>
#include <cstdint>
#include <sstream>
#include <unordered_map>
#include <unordered_set>
#include <libcore/exceptions.h>
#include <libcore/logging.h>
#include "device.h"

namespace
{
	struct USBDeviceID
	{
		uint16_t vendorID;
		uint16_t productID;

		USBDeviceID(uint16_t vendorID, uint16_t productID)
		  : vendorID(vendorID), productID(productID)
		{
		}
	};

	bool operator==(const USBDeviceID &d1, const USBDeviceID &d2)
	{
		return d1.vendorID == d2.vendorID && d1.productID == d2.productID;
	}

	struct USBDeviceIDHash
	{
		size_t operator() (const USBDeviceID &devID) const
		{
			return static_cast<size_t>(devID.vendorID) << 16 | devID.productID;
		}
	};

	class USBDeviceIDSet
	{
	public:
		USBDeviceIDSet(std::initializer_list<USBDeviceID> args)
		  : m_devices(args)
		{
		}

		bool contains(uint16_t vendorID, uint16_t productID) const
		{
			return contains({vendorID, productID});
		}

		bool contains(const USBDeviceID &devID) const
		{
			return m_devices.find(devID) != m_devices.end();
		}

	private:
		std::unordered_set<USBDeviceID, USBDeviceIDHash> m_devices;
	};

	const USBDeviceIDSet g_supportedDeviceIDs {
		{ 0x08A6, 0x0050 }, // B-EX series
		{ 0x08A6, 0x0051 }, // B-EX series
		{ 0x08A6, 0x0052 }, // B-EX series
		{ 0x08A6, 0x0053 }, // B-EX series
		{ 0x08A6, 0x0054 }, // B-EX series
		{ 0x08A6, 0x0055 }, // B-EX series
		{ 0x08A6, 0xC00B }, // DB-EA4D
		{ 0x08A6, 0xB000 }, // B-FV series
		{ 0x08A6, 0xB001 }, // BA-400 series
		{ 0x08A6, 0xB002 }, // BA-400 series
		{ 0x08A6, 0xB003 }, // BA-400 series
	};

	// Port paths are only 7 long max, but in this case, the first byte
	// represents the actual length of the path
	using USBPortPath = std::array<uint8_t, 8>;

	USBPortPath getPortPathFromURI(const StripedHorse::URI &uri)
	{
		uint64_t portPathInteger = std::stoull(uri.getHost());
		return USBPortPath { {
			static_cast<uint8_t>(portPathInteger >> 56),
			static_cast<uint8_t>(portPathInteger >> 48),
			static_cast<uint8_t>(portPathInteger >> 40),
			static_cast<uint8_t>(portPathInteger >> 32),
			static_cast<uint8_t>(portPathInteger >> 24),
			static_cast<uint8_t>(portPathInteger >> 16),
			static_cast<uint8_t>(portPathInteger >>  8),
			static_cast<uint8_t>(portPathInteger >>  0),
		} };
	}

	std::string stringifyPortPath(const USBPortPath &portPath)
	{
		std::ostringstream result;
		if (portPath[0] == 0)
			return "(null)";
		
		for (uint8_t i = 0; i < portPath[0] - 1; i++)
			result << static_cast<unsigned>(portPath[i + 1]) << "/";
		result << static_cast<unsigned>(portPath[portPath[0]]);
		return result.str();
	}

	std::unordered_map<std::string, std::string> parseIEEE1284DeviceID(const std::string &str)
	{
		std::unordered_map<std::string, std::string> parsed;
		std::istringstream controlTransferStream(str);
		std::string line;

		while (std::getline(controlTransferStream, line, ';'))
		{
			std::string::size_type colon = line.find(':');
			if (colon != std::string::npos)
				parsed.emplace(line.substr(0, colon), line.substr(colon + 1));
			else
				parsed.emplace(std::make_pair(line, ""));
		}

		return parsed;
	}

	// Ensure that the IEEE1284 device ID is consistent by forcing the
	// manufacturer to be the same. E.g. "Toshiba", "Toshiba TEC" and "TOSHIBA"
	// have been found reported from various Toshiba USB devices.
	std::string preprocessIEEE1284DeviceID(const std::string &str)
	{
		std::unordered_map<std::string, std::string> parsed = parseIEEE1284DeviceID(str);
		if (parsed.size() == 0)
			return str;

		auto it = parsed.find("MANUFACTURER");
		if (it == parsed.end())
			it = parsed.find("MFG");
		
		if (it == parsed.end() || it->second == "TOSHIBA")
			return str;
		
		it->second = "TOSHIBA";

		std::ostringstream output;
		for (auto kvp : parsed)
			output << kvp.first << ":" << kvp.second << ";";

		std::string result = output.str();
		if (result.size() > 0)
			return result.substr(0, result.size() - 1);
		else
			return result;
	}
}

#if defined(USE_LIBUSB_1_0)
#  include "device-libusb.inc.cpp"
#elif defined(USE_LIBUSB_0_1)
#  include "device-oldlibusb.inc.cpp"
#else
#  include "device-unsupported.inc.cpp"
#endif

StripedHorse::Device::Device(const URI &uri)
  : m_uri(uri)
{
}

StripedHorse::Device::Device(Device &&other)
  : m_uri(std::move(other.m_uri)),
    m_description(std::move(other.m_description)),
    m_deviceInfo(std::move(other.m_deviceInfo))
{
}

StripedHorse::Device::~Device()
{
}

bool StripedHorse::Device::PerformDiscovery(std::vector<Device> &result)
{
	USBManager manager;

	std::vector<USBDevice> supportedDevices;
	if (!manager.tryGetUSBDevices(g_supportedDeviceIDs, supportedDevices))
		return false;
	
	for (USBDevice &udev : supportedDevices)
	{
		StripedHorse::URI deviceURI;
		if (udev.tryConstructURI(deviceURI))
		{
			result.emplace_back(deviceURI);
			Device &dev = result.back(); // emplace_back doesn't return ref until C++17
			dev.m_description = udev.getDescription();
			dev.m_deviceID = udev.getIEEE1284DeviceID();
			dev.m_deviceInfo = dev.m_description;
		}
	}

	return true;
}

const std::string &StripedHorse::Device::getDescription() const
{
	return m_description;
}

const std::string &StripedHorse::Device::getDeviceID() const
{
	return m_deviceID;
}

const std::string &StripedHorse::Device::getDeviceInfo() const
{
	return m_deviceInfo;
}

const StripedHorse::URI &StripedHorse::Device::getURI() const
{
	return m_uri;
}

std::tuple<size_t, size_t> StripedHorse::Device::sendData(std::istream &in)
{
	USBPortPath portPath = getPortPathFromURI(m_uri);
	
	Logging::Debug() << "Looking for device at path " << stringifyPortPath(portPath) << std::endl;
	
	USBManager manager;

	std::vector<USBDevice> matchedDevices;
	if (!manager.tryGetUSBDevices(g_supportedDeviceIDs, matchedDevices, portPath))
	{
		Logging::Error() << "Device detection failed, unable to find printer device" << std::endl;
		return std::make_tuple(0, 0);
	}
	
	if (matchedDevices.size() == 0)
	{
		Logging::Error() << "No USB devices found" << std::endl;
		return std::make_tuple(0, 0);
	}
	
	USBDevice &device = matchedDevices.front();
	return device.write(in);
}
