#include <algorithm>
#include <cctype>
#include <functional>
#include <memory>
#include <set>
#include <sstream>
#include <syslog.h>
#include <unistd.h>
#include <unordered_map>

#define _IPP_PRIVATE_STRUCTURES 1
#include <cups/cups.h>
#include <cups/ppd.h>

#include <libcore/exceptions.h>

#include "device.h"

namespace StripedHorse
{
	class CUPSCommunicationException : public Exception
	{
	public:
		CUPSCommunicationException(const std::string &msg) : Exception(msg) {}
	};

	class USBCommunicationException : public Exception
	{
	public:
		USBCommunicationException(const std::string &msg) : Exception(msg) {}
	};
}

namespace
{
	std::string getStringFromAttribute(ipp_attribute_t *attr, int index)
	{
		if (attr == nullptr)
			throw std::invalid_argument("attr");
		
		if (index >= attr->num_values)
			throw std::out_of_range("index");
		
		if (attr->value_tag != IPP_TAG_TEXTLANG && attr->value_tag != IPP_TAG_NAMELANG && (attr->value_tag < IPP_TAG_TEXT || attr->value_tag > IPP_TAG_MIMETYPE))
			throw std::invalid_argument("attr");
		
		return attr->values[index].string.text;
	}

	class HTTPCloser
	{
	public:
		void operator()(http_t *http)
		{
			httpClose(http);
		}
	};

	class IPPCloser
	{
	public:
		void operator()(ipp_t *ipp)
		{
			ippDelete(ipp);
		}
	};

	class PPDCloser
	{
	public:
		void operator()(ppd_file_t *ppd)
		{
			ppdClose(ppd);
		}
	};

	struct CUPSPrinter
	{
		CUPSPrinter(const std::string &name, const StripedHorse::URI &uri)
		  : name(name), uri(uri)
		{
		}

		std::string name;
		StripedHorse::URI uri;
	};

	std::string flattenModelName(const std::string &input)
	{
		std::string result;
		result.reserve(input.size());

		for (const char &ch : input)
		{
			if (::isspace(ch) || !::isprint(ch) || ch == '(' || ch == ')')
				continue;
			
			result.push_back(::tolower(ch));
		}
		
		return result;
	}

	std::unique_ptr<ipp_t, IPPCloser> doRequest(http_t *cups, ipp_t *request, const std::string &resource = "/")
	{
		std::unique_ptr<ipp_t, IPPCloser> response(cupsDoRequest(cups, request, resource.c_str()));
		if (!response)
			throw StripedHorse::CUPSCommunicationException("Unable to query CUPS server");
		return response;
	}

	std::string formatPrinterName(const std::string &name)
	{
		std::string result;
		result.reserve(name.size());

		for (const char ch : name)
		{
			if (!::isprint(ch) || ::isspace(ch) || ch == '#' || ch == '/' || ch == '_')
				result.push_back('_');
			else
				result.push_back(ch);
		}

		return result;
	}

	std::vector<CUPSPrinter> getInstalledPrinters(http_t *cups, const std::string &scheme)
	{
		const char *requestedAttributes[] = {
			"printer-uri-supported",
			"device-uri"
		};

		ipp_t *request = ippNewRequest(CUPS_GET_PRINTERS);
		ippAddStrings(request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD, "requested-attributes", sizeof(requestedAttributes) / sizeof(requestedAttributes[0]), nullptr, requestedAttributes);

		auto response = doRequest(cups, request);
		std::vector<CUPSPrinter> printers;
		std::string name;
		StripedHorse::URI uri;
		
		for (ipp_attribute_t *attr = response->attrs; attr != nullptr; attr = attr->next)
		{
			if (attr->group_tag != IPP_TAG_PRINTER)
				continue;
			
			if (attr->value_tag == IPP_TAG_URI)
			{
				if (strcmp(attr->name, "device-uri") == 0)
				{
					uri = StripedHorse::URI(getStringFromAttribute(attr, 0));
				}
				else if (strcmp(attr->name, "printer-uri-supported") == 0)
				{
					StripedHorse::URI deviceURI(getStringFromAttribute(attr, 0));
					auto pathComponents = deviceURI.getPathComponents();
					if (pathComponents.size() == 0)
						throw StripedHorse::CUPSCommunicationException("URI returned from CUPS is unusable");
					
					name = pathComponents.back();
				}
			}
			
			if (name.size() > 0 && !uri.isEmpty())
			{
				if (uri.getScheme() == scheme)
					printers.emplace_back(name, uri);
				
				name.clear();
				uri.clear();
			}
		}

		return printers;
	}

	std::vector<std::string> getPPDsForManufacturer(http_t *cups, const std::string &manufacturer)
	{
		std::vector<std::string> ppdFilenames;

		ipp_t *request = ippNewRequest(CUPS_GET_PPDS);
		ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_TEXT, "ppd-make", nullptr, manufacturer.c_str());
		ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD, "requested-attributes", nullptr, "ppd-name");

		auto response = doRequest(cups, request);

		for (ipp_attribute_t *attr = response->attrs; attr != nullptr; attr = attr->next)
		{
			if (attr->group_tag != IPP_TAG_PRINTER)
				continue;
			
			ppdFilenames.emplace_back(getStringFromAttribute(attr, 0));
		}

		return ppdFilenames;
	}

	StripedHorse::URI getURIForCUPSPrinterName(const std::string &deviceName)
	{
		StripedHorse::URI result;
		result.setScheme("ipp");
		result.setHost("localhost");
		result.appendPathComponent("printers");
		result.appendPathComponent(deviceName);
		return result;
	}

	void doSimpleOperation(http_t *cups, const CUPSPrinter &device, ipp_op_t operation)
	{
		StripedHorse::URI printerURI = getURIForCUPSPrinterName(device.name);
		std::string uriString = static_cast<std::string>(printerURI);
		ipp_t *request = ippNewRequest(operation);
		ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI, "printer-uri", nullptr, uriString.c_str());

		auto response = doRequest(cups, request);
		ipp_status_t status = response->request.status.status_code;
		if (status > IPP_OK_CONFLICT)
			throw StripedHorse::CUPSCommunicationException("Unable to enable/disable the printer");
	}

	void setPrinterEnabledState(http_t *cups, const CUPSPrinter &device, bool enabled)
	{
		doSimpleOperation(cups, device, enabled ? IPP_ENABLE_PRINTER : IPP_DISABLE_PRINTER);
	}

	void setPrinterPausedState(http_t *cups, const CUPSPrinter &device, bool paused)
	{
		doSimpleOperation(cups, device, paused ? IPP_PAUSE_PRINTER : IPP_RESUME_PRINTER);
	}

	CUPSPrinter addQueueForDevice(http_t *cups, const StripedHorse::Device &device, const std::set<std::string> &bannedNames)
	{
		static const std::string absolutePPDDirectory = "/usr/share/cups/model/";
		
		std::unordered_map<std::string, std::string> ppdMap;

		for (const std::string &ppdFilename : getPPDsForManufacturer(cups, "TOSHIBA"))
		{
			std::string fullPath = absolutePPDDirectory + ppdFilename;
			if (access(fullPath.c_str(), R_OK) != 0)
				continue;

			std::unique_ptr<ppd_file_t, PPDCloser> ppd(ppdOpenFile(fullPath.c_str()));
			if (!ppd)
			{
				syslog(LOG_WARNING, "Unable to parse PPD: %s", ppdFilename.c_str());
				continue;
			}

			ppdMap[flattenModelName(ppd->modelname)] = ppdFilename;
			for (ppd_attr_t *attr = ppdFindAttr(ppd.get(), "Product", nullptr); attr != nullptr; attr = ppdFindNextAttr(ppd.get(), "Product", nullptr))
			{
				if (!attr->value)
					continue;
				
				std::string product(attr->value);
				ppdMap[flattenModelName(product)] = ppdFilename;
			}
		}

		const auto it = ppdMap.find(flattenModelName(device.getDeviceInfo()));
		if (it == ppdMap.end())
			throw StripedHorse::InternalInconsistencyException("Detected device does not have a valid PPD: " + device.getDeviceInfo());

		syslog(LOG_DEBUG, "Model '%s' found in this PPD: %s", device.getDeviceInfo().c_str(), it->second.c_str());

		std::string chosenPPDFilename = it->second;
		std::string printerInfo = device.getDeviceInfo();
		std::string formattedPrinterName = formatPrinterName(printerInfo);
		std::string printerName = formattedPrinterName;
		size_t retries = 1;
		while (bannedNames.find(printerName) != bannedNames.end())
		{
			std::ostringstream printerNameStream;
			printerNameStream << formattedPrinterName << "_" << (retries + 1);
			printerName = printerNameStream.str();
			retries++;
			if (retries == 100)
			{
				syslog(LOG_CRIT, "There are a lot of printers on this system with similar names. Consider removing some.");
				throw StripedHorse::Exception("Unable to generate a printer name");
			}
		}

		StripedHorse::URI printerURI = getURIForCUPSPrinterName(printerName);

		std::string printerURIString = static_cast<std::string>(printerURI);
		std::string deviceURIString = static_cast<std::string>(device.getURI());

		syslog(LOG_DEBUG, "Adding printer with name '%s' (printer-uri: %s -- device-uri: %s)", printerName.c_str(), printerURIString.c_str(), deviceURIString.c_str());

		ipp_t *request = ippNewRequest(CUPS_ADD_MODIFY_PRINTER);
		ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI, "printer-uri", nullptr, printerURIString.c_str());
		ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME, "requesting-user-name", nullptr, cupsUser());
		ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_URI, "device-uri", nullptr, deviceURIString.c_str());
		ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_NAME, "ppd-name", nullptr, chosenPPDFilename.c_str());
		ippAddString(request, IPP_TAG_PRINTER, IPP_TAG_TEXT, "printer-info", nullptr, printerInfo.c_str());
		ippAddBoolean(request, IPP_TAG_PRINTER, "printer-is-accepting-jobs", 1);

		auto response = doRequest(cups, request, "/admin/");

		ipp_status_t status = response->request.status.status_code;
		if (status > IPP_OK_CONFLICT)
		{
			syslog(LOG_DEBUG, "CUPS returned status %d", static_cast<int>(status));
			throw StripedHorse::CUPSCommunicationException("Unable to add a print queue for '" + device.getDeviceInfo() + "' because a CUPS-related error occurred, check the CUPS log for more info");
		}

		return CUPSPrinter(printerName, device.getURI());
	}

	void removeQueue(http_t *cups, const CUPSPrinter &printer)
	{
		StripedHorse::URI printerURI = getURIForCUPSPrinterName(printer.name);
		std::string printerURIString = static_cast<std::string>(printerURI);
		ipp_t *request = ippNewRequest(CUPS_DELETE_PRINTER);
		ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_URI, "printer-uri", nullptr, printerURIString.c_str());
		ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME, "requesting-user-name", nullptr, cupsUser());

		auto response = doRequest(cups, request, "/admin/");
		
		ipp_status_t status = response->request.status.status_code;
		if (status > IPP_OK_CONFLICT)
		{
			syslog(LOG_DEBUG, "CUPS returned status %d", static_cast<int>(status));
			throw StripedHorse::CUPSCommunicationException("Unable to remove the print queue for: " + printer.name);
		}
	}
}

int main(int argc, char *argv[])
{
	using namespace StripedHorse;

	enum class Mode
	{
		kAutoDetection,
		kDeleteAll
	} currentMode = Mode::kAutoDetection;

	if (argc > 1 && strcmp(argv[1], "--delete-all") == 0)
		currentMode = Mode::kDeleteAll;

	openlog("toshiba-usb", LOG_CONS, LOG_LPR);

	try
	{
		std::unique_ptr<http_t, HTTPCloser> cups(httpConnectEncrypt(cupsServer(), ippPort(), cupsEncryption()));
		if (!cups)
			throw StripedHorse::CUPSCommunicationException("Unable to connect to CUPS server");

		std::vector<CUPSPrinter> cupsPrinters = getInstalledPrinters(cups.get(), "tec");
		
		if (currentMode == Mode::kDeleteAll)
		{
			bool failed = false;
			for (const CUPSPrinter &printer : cupsPrinters)
			{
				try
				{
					removeQueue(cups.get(), printer);
				}
				catch (CUPSCommunicationException &ex)
				{
					syslog(LOG_DEBUG, "%s", ex.what());
					failed = true;
				}
			}

			return static_cast<int>(failed);
		}

		std::set<std::string> bannedNames;
		for (const CUPSPrinter &cupsPrinter : cupsPrinters)
		{
			syslog(LOG_DEBUG, "Found CUPS printer %s (uri: %s)", cupsPrinter.name.c_str(), static_cast<std::string>(cupsPrinter.uri).c_str());
			bannedNames.emplace(cupsPrinter.name);
		}

		std::vector<Device> discoveredDevices;
		if (!Device::PerformDiscovery(discoveredDevices))
			syslog(LOG_DEBUG, "No USB devices could be detected");
		
		for (const Device &dev : discoveredDevices)
		{
			const URI &uri = dev.getURI();

			syslog(LOG_DEBUG, "Found device: %s", dev.getDescription().c_str());

			std::vector<CUPSPrinter>::const_iterator it = std::find_if(cupsPrinters.cbegin(), cupsPrinters.cend(), [&](const CUPSPrinter &printer){
				return printer.uri == uri;
			});

			if (it == cupsPrinters.end())
			{
				syslog(LOG_DEBUG, "The following URI is not installed: %s", static_cast<std::string>(uri).c_str());
				
				CUPSPrinter newPrinter = addQueueForDevice(cups.get(), dev, bannedNames);
				setPrinterEnabledState(cups.get(), newPrinter, true);
				setPrinterPausedState(cups.get(), newPrinter, false);
			}
			else
			{
				syslog(LOG_DEBUG, "Printer '%s' was detected and is already installed, it is being enabled", it->name.c_str());

				// enable the printer and remove it from the list (any leftover printers in this
				// list are disabled because they are installed in CUPS but are not physically
				// detected)
				setPrinterEnabledState(cups.get(), *it, true);
				setPrinterPausedState(cups.get(), *it, false);
				cupsPrinters.erase(it);
			}
		}
		
		// Any printers that are installed in CUPS but were not detected should be disabled
		for (const CUPSPrinter &printer : cupsPrinters)
		{
			syslog(LOG_DEBUG, "Printer '%s' was not detected, it is being disabled", printer.name.c_str());
			setPrinterEnabledState(cups.get(), printer, false);
		}
	}
	catch (std::exception &ex)
	{
		syslog(LOG_CRIT, "An error occurred while managing CUPS print queues: %s", ex.what());
		return 1;
	}
}
