#include <cstdlib>
#include <fstream>
#include <iostream>

#include <libcore/logging.h>
#include "device.h"

namespace
{
	void doDiscovery()
	{
		std::vector<StripedHorse::Device> devices;
		if (StripedHorse::Device::PerformDiscovery(devices))
		{
			for (const auto &device : devices)
			{
				std::cout << "direct ";
				std::cout << device.getURI() << " ";
				std::cout << "\"" << device.getDescription() << "\" ";
				std::cout << "\"" << device.getDeviceInfo() << "\" ";
				std::cout << "\"" << device.getDeviceID() << "\"";
				std::cout << std::endl;
			}
		}
	}

	bool tryGetDeviceURI(StripedHorse::URI &result)
	{
		const char *deviceURI = ::getenv("DEVICE_URI");
		if (deviceURI == nullptr)
			return false;
		
		try
		{
			result = StripedHorse::URI(deviceURI);
		}
		catch (StripedHorse::Exception &ex)
		{
			StripedHorse::Logging::Error() << "Unable to get device URI: " << ex.what() << std::endl;
			return false;
		}

		return true;
	}

	enum CUPSStatus
	{
		kSuccess = 0,
		kFailed,
		kAuthenticationRequired,
		kHold,
		kStop,
		kCancel,
		kRetry,
		kRetryImmediately
	};
}

int main(int argc, char *argv[])
{
	using namespace StripedHorse;
	std::ifstream fileInput;
	std::istream *input = &std::cin;

	Logging::SetMessagePrefixes("DEBUG", "INFO", "WARNING", "ERROR", "CRIT");

	if (argc < 5)
	{
		try
		{
			doDiscovery();
		}
		catch (std::exception &ex)
		{
			Logging::Critical() << "Unable to perform device discovery: " << ex.what() << std::endl;
			return 1;
		}

		return 0;
	}
	else if (argc > 6)
	{
		fileInput.open(argv[6]);
		input = &fileInput;
	}
	
	Logging::Debug() << "Backend started for transport" << std::endl;
	
	URI deviceURI;
	if (!tryGetDeviceURI(deviceURI))
	{
		Logging::Critical() << "Unable to determine the device URI" << std::endl;
		return CUPSStatus::kStop;
	}

	try
	{
		Logging::Debug() << "Attempting to find device and send data..." << std::endl;
		
		Device device(deviceURI);
		size_t totalBytesRead;
		size_t totalBytesWritten;
		std::tie(totalBytesRead, totalBytesWritten) = device.sendData(*input);
		
		if (totalBytesRead == 0)
		{
			Logging::Error() << "A problem occurred before any data could be read from the input" << std::endl;
			return CUPSStatus::kStop;
		}
		else if (totalBytesWritten == 0)
		{
			Logging::Error() << "No data could be sent to the device" << std::endl;
			return CUPSStatus::kRetryImmediately;
		}
		else if (totalBytesWritten == totalBytesRead)
		{
			if (input->eof())
				Logging::Debug() << "All " << totalBytesWritten << " bytes written successfully" << std::endl;
			else
				Logging::Warning() << totalBytesWritten << " bytes written, but an IO error occurred before all bytes could be read" << std::endl;
			return CUPSStatus::kSuccess;
		}
		else
		{
			Logging::Warning() << "Only " << totalBytesWritten << " out of " << totalBytesRead << " bytes from the input were sent to the device" << std::endl;
			return CUPSStatus::kSuccess;
		}
	}
	catch (std::exception &ex)
	{
		Logging::Critical() << "An error occurred while connecting to or sending data to the printer: " << ex.what() << std::endl;
		return CUPSStatus::kStop;
	}
}
