#include <cstring>
#include <algorithm>
#include <sstream>
#include "cupsformat.h"

StripedHorse::CUPSFormat::CUPSFormat(const CUPSSync &sync, std::istream &in)
  : advanceDistance(0), advanceMedia(kActionNever), collate(false),
    cutMedia(kActionNever), duplex(false), insertSheet(false),
    jog(kActionNever), leadingEdge(kLeadingEdgeTop), manualFeed(false),
    mediaPosition(0), mediaWeight(0), mirrorPrint(false), negativePrint(false),
    numCopies(0), orientation(kOrientationRotate0), outputFaceUp(false),
    separations(0), traySwitch(false), tumble(0), cupsWidth(0), cupsHeight(0),
    cupsMediaType(0), cupsBitsPerColor(0), cupsBitsPerPixel(0),
    cupsBytesPerLine(0), cupsColorOrder(kCUPSColorOrderChunky),
    cupsColorSpace(kCUPSColorSpaceGray), cupsCompression(0), cupsRowCount(0),
    cupsRowFeed(0), cupsRowStep(0), cupsNumColors(0),
    cupsBorderlessScalingFactor(0.0f), m_version(sync.getVersion())
{
	std::vector<char> buf;
	if (m_version == 2 || m_version == 3)
		buf.resize(1796);
	else
		buf.resize(420);
	
	m_endian = sync.getFormatEndianness();
	
	if (!in.read(&buf[0], buf.size()))
		throw IOException("Unable to read from input");
	
	std::istringstream ss(std::string(&buf[0], buf.size()));
	
	readField(ss, mediaClass);
	readField(ss, mediaColor);
	readField(ss, mediaType);
	readField(ss, outputType);
	readField(ss, advanceDistance);
	readEnumField(ss, advanceMedia);
	readField(ss, collate);
	readEnumField(ss, cutMedia);
	readField(ss, duplex);
	readField(ss, resolution.horizontal);
	readField(ss, resolution.vertical);
	readField(ss, imagingBoundingBox.lowerLeft.x);
	readField(ss, imagingBoundingBox.lowerLeft.y);
	readField(ss, imagingBoundingBox.upperRight.x);
	readField(ss, imagingBoundingBox.upperRight.y);
	readField(ss, insertSheet);
	readEnumField(ss, jog);
	readEnumField(ss, leadingEdge);
	readField(ss, margins.x);
	readField(ss, margins.y);
	readField(ss, manualFeed);
	readField(ss, mediaPosition);
	readField(ss, mediaWeight);
	readField(ss, mirrorPrint);
	readField(ss, negativePrint);
	readField(ss, numCopies);
	readEnumField(ss, orientation);
	readField(ss, outputFaceUp);
	readField(ss, pageSize.x);
	readField(ss, pageSize.y);
	readField(ss, separations);
	readField(ss, traySwitch);
	readField(ss, tumble);
	readField(ss, cupsWidth);
	readField(ss, cupsHeight);
	readField(ss, cupsMediaType);
	readField(ss, cupsBitsPerColor);
	readField(ss, cupsBitsPerPixel);
	readField(ss, cupsBytesPerLine);
	readField(ss, cupsColorOrder);
	readEnumField(ss, cupsColorSpace);
	readField(ss, cupsCompression);
	readField(ss, cupsRowCount);
	readField(ss, cupsRowFeed);
	readField(ss, cupsRowStep);
	
	if (m_version == 1)
	{
		if (ss)
			return;
		else
			throw CUPSFormatException("Unable to read fields");
	}
	
	readField(ss, cupsNumColors);
	readField(ss, cupsBorderlessScalingFactor);
	readField(ss, cupsPageSize.x);
	readField(ss, cupsPageSize.y);
	readField(ss, cupsImagingBBox.lowerLeft.x);
	readField(ss, cupsImagingBBox.lowerLeft.y);
	readField(ss, cupsImagingBBox.upperRight.x);
	readField(ss, cupsImagingBBox.upperRight.y);
	
	cupsInteger.resize(16);
	for (size_t i = 0; i < cupsInteger.size(); i++)
		readField(ss, cupsInteger[i]);
	
	cupsReal.resize(16);
	for (size_t i = 0; i < cupsReal.size(); i++)
		readField(ss, cupsReal[i]);
	
	cupsString.resize(16);
	for (size_t i = 0; i < cupsString.size(); i++)
		readField(ss, cupsString[i]);
	
	readField(ss, cupsMarkerType);
	readField(ss, cupsRenderingIntent);
	readField(ss, cupsPageSizeName);
}

bool StripedHorse::CUPSFormat::readField(std::istream &in, bool &f)
{
	uint32_t i;
	if (in.read(reinterpret_cast<char *>(&i), sizeof i))
		f = static_cast<bool>(i);
	return in.good();
}

bool StripedHorse::CUPSFormat::readField(std::istream &in, float &f)
{
	if (in.read(reinterpret_cast<char *>(&f), sizeof f))
		f = swapBytes(f);
	return in.good();
}

bool StripedHorse::CUPSFormat::readField(std::istream &in, uint32_t &f)
{
	if (in.read(reinterpret_cast<char *>(&f), sizeof f))
		f = swapBytes(f);
	return in.good();
}

template<typename T>
bool StripedHorse::CUPSFormat::readEnumField(std::istream &in, T &f)
{
	uint32_t i;
	if (readField(in, i))
		f = (T) swapBytes(i);
	return in.good();
}

bool StripedHorse::CUPSFormat::readField(std::istream &in, std::string &f, size_t max)
{
	std::vector<char> buf(max);
	if (in.read(&buf[0], buf.size()))
	{
		std::vector<char>::iterator nul = std::find(buf.begin(), buf.end(), '\0');
		f = std::string(buf.begin(), nul);
	}
	
	return in.good();
}

uint32_t StripedHorse::CUPSFormat::swapBytes(uint32_t n)
{
	return FromEndian(m_endian, n);
}

float StripedHorse::CUPSFormat::swapBytes(float n)
{
	return FromEndian(m_endian, n);
}

std::ostream &StripedHorse::operator<<(std::ostream &out, const StripedHorse::CUPSFormat::Orientation &orientation)
{
	switch (orientation)
	{
		case CUPSFormat::kOrientationRotate0:
			out << "0 deg";
			break;
		
		case CUPSFormat::kOrientationRotate90:
			out << "90 deg";
			break;
		
		case CUPSFormat::kOrientationRotate180:
			out << "180 deg";
			break;
		
		case CUPSFormat::kOrientationRotate270:
			out << "270 deg";
			break;
	}
	
	return out;
}

