#ifndef STRIPEDHORSE_CUPSFORMAT_H
#define STRIPEDHORSE_CUPSFORMAT_H

#include <cstdint>
#include <functional>
#include <iostream>
#include <string>
#include <vector>

#include <libcore/endian.h>
#include <libcore/exceptions.h>
#include <libcore/types.h>

#include "cupssync.h"

namespace StripedHorse
{
	struct Resolution
	{
		uint32_t vertical;
		uint32_t horizontal;
		
		Resolution(uint32_t v = 0, uint32_t h = 0) : vertical(v), horizontal(h) {}
	};

	using CUPSPoint = BasicPoint<uint32_t>;
	using CUPSRectangle = BasicRectangle<uint32_t>;
	
	class CUPSFormat
	{
	public:
		enum Action
		{
			kActionNever,
			kActionAfterFile,
			kActionAfterJob,
			kActionAfterSet,
			kActionAfterPage
		};
		
		enum LeadingEdge
		{
			kLeadingEdgeTop,
			kLeadingEdgeRight,
			kLeadingEdgeBottom,
			kleadingEdgeLeft
		};
		
		enum Orientation
		{
			kOrientationRotate0,
			kOrientationRotate90,
			kOrientationRotate180,
			kOrientationRotate270
		};
		
		enum CUPSColorOrder
		{
			kCUPSColorOrderChunky, /* CMYK CMYK CMYK ... */
			kCUPSColorOrderBanded, /* CCC MMM YYY KKK ... */
			kCUPSColorOrderPlanar, /* CCC... MMM... YYY... KKK... */
		};
		
		enum CUPSColorSpace
		{
			kCUPSColorSpaceGray,
			kCUPSColorSpaceRGB,
			kCUPSColorSpaceRGBA,
			kCUPSColorSpaceBlack,
			kCUPSColorSpaceCMY,
			kCUPSColorSpaceYMC,
			kCUPSColorSpaceCMYK,
			kCUPSColorSpaceYMCK,
			kCUPSColorSpaceKCMY,
			kCUPSColorSpaceKCMYcm,
			kCUPSColorSpaceGMCK,
			kCUPSColorSpaceGMCS,
			kCUPSColorSpaceWhite,
			kCUPSColorSpaceGold,
			kCUPSColorSpaceSilver,
			kCUPSColorSpaceCIEXYZ,
			kCUPSColorSpaceCIELab,
			kCUPSColorSpaceRBGW,
			kCUPSColorSpace_sGray,
			kCUPSColorSpace_sRGB,
			kCUPSColorSpaceAdobeRGB,
			kCUPSColorSpaceICC1,
			kCUPSColorSpaceICC2,
			kCUPSColorSpaceICC3,
			kCUPSColorSpaceICC4,
			kCUPSColorSpaceICC5,
			kCUPSColorSpaceICC6,
			kCUPSColorSpaceICC7,
			kCUPSColorSpaceICC8,
			kCUPSColorSpaceICC9,
			kCUPSColorSpaceICCA,
			kCUPSColorSpaceICCB,
			kCUPSColorSpaceICCC,
			kCUPSColorSpaceICCD,
			kCUPSColorSpaceICCE,
			kCUPSColorSpaceICCF,
			kCUPSColorSpaceDevice1,
			kCUPSColorSpaceDevice2,
			kCUPSColorSpaceDevice3,
			kCUPSColorSpaceDevice4,
			kCUPSColorSpaceDevice5,
			kCUPSColorSpaceDevice6,
			kCUPSColorSpaceDevice7,
			kCUPSColorSpaceDevice8,
			kCUPSColorSpaceDevice9,
			kCUPSColorSpaceDeviceA,
			kCUPSColorSpaceDeviceB,
			kCUPSColorSpaceDeviceC,
			kCUPSColorSpaceDeviceD,
			kCUPSColorSpaceDeviceE,
			kCUPSColorSpaceDeviceF
		};
		
		CUPSFormat(const CUPSSync &sync, std::istream &);
		
		uint32_t getVersion() const;
		
		// CUPS v1 Raster Format fields
		std::string mediaClass;
		std::string mediaColor;
		std::string mediaType;
		std::string outputType;
		uint32_t advanceDistance;
		Action advanceMedia;
		bool collate;
		Action cutMedia;
		bool duplex;
		Resolution resolution;
		CUPSRectangle imagingBoundingBox;
		bool insertSheet;
		Action jog;
		LeadingEdge leadingEdge;
		CUPSPoint margins;
		bool manualFeed;
		uint32_t mediaPosition;
		uint32_t mediaWeight;
		bool mirrorPrint;
		bool negativePrint;
		uint32_t numCopies;
		Orientation orientation;
		bool outputFaceUp;
		CUPSPoint pageSize;
		bool separations;
		bool traySwitch;
		bool tumble;
		uint32_t cupsWidth;
		uint32_t cupsHeight;
		uint32_t cupsMediaType;
		uint32_t cupsBitsPerColor;
		uint32_t cupsBitsPerPixel;
		uint32_t cupsBytesPerLine;
		uint32_t cupsColorOrder;
		CUPSColorSpace cupsColorSpace;
		uint32_t cupsCompression;
		uint32_t cupsRowCount;
		uint32_t cupsRowFeed;
		uint32_t cupsRowStep;
		
		// CUPS Raster Format version 2/3 fields
		uint32_t cupsNumColors;
		float cupsBorderlessScalingFactor;
		BasicPoint<float> cupsPageSize;
		BasicRectangle<float> cupsImagingBBox;
		std::vector<uint32_t> cupsInteger;
		std::vector<float> cupsReal;
		std::vector<std::string> cupsString;
		std::string cupsMarkerType;
		std::string cupsRenderingIntent;
		std::string cupsPageSizeName;
		
	private:
		Endian m_endian;
		uint32_t m_version;
		
		bool readField(std::istream &, bool &);
		bool readField(std::istream &, float &);
		bool readField(std::istream &, uint32_t &);
		bool readField(std::istream &, std::string &, size_t max = 64);
		
		template<typename T>
		bool readEnumField(std::istream &, T &);
		
		uint32_t swapBytes(uint32_t);
		float swapBytes(float);
	};
	
	class CUPSFormatException : public Exception
	{
	public:
		CUPSFormatException(const std::string &m) : Exception(m) {}
	};
	
	std::ostream &operator<<(std::ostream &, const CUPSFormat::Orientation &);
}

#endif
