#ifndef STRIPEDHORSE_CUPSSYNC_H
#define STRIPEDHORSE_CUPSSYNC_H

#include <libcore/exceptions.h>
#include <libcore/endian.h>

namespace StripedHorse
{
	/**
	 CUPSSync object
	 
	 Before CUPS sends any raster data to the filter, it sends a sync word. This
	 word can be used to determine if the byte order of the data to follow is
	 in native endian or not. It also determines the version of the info to
	 follow (which determines whether extra fields need to be read from the
	 input stream).
	 */
	class CUPSSync
	{
	public:
		/**
		  Constructs a CUPSSync object
		  @param[in] input the stream to read from
		 */
		CUPSSync(std::istream &input);
		
		/**
		 The endianness of the format data from CUPS.
		 @return an Endian enum value (either kBigEndian or kLittleEndian)
		 
		 The data may have originated from a CUPS server running on a different
		 architecture to the one running this filter. If this is the case, and
		 the endianness of the architectures do not match, the format data from
		 CUPS needs to have all the bytes swapped for each field.
		 */
		enum Endian getFormatEndianness() const;
		
		/**
		 Returns the version of the information from CUPS.
		 @return either 1, 2 or 3 (the only supported formats so far)
		 */
		int getVersion() const;
		
	private:
		enum Endian m_endianness;
		int m_version;
	};
	
	class CUPSSyncException : public Exception
	{
	public:
		CUPSSyncException(const std::string &m) : Exception(m) {}
	};
}

#endif

