#include <algorithm>
#include <libcore/logging.h>
#include <libtecfilter/dbea4dprinter.h>
#include "dbea4dfilter.h"

namespace
{
	enum ASCII : char
	{
		FF = '\x0C',
		DC0 = '\x10',
		DC1 = '\x11',
		DC2 = '\x12',
		DC3 = '\x13',
		DC4 = '\x14',
		ESC = '\x1B',
		FS = '\x1C',
		GS = '\x1D',
		US = '\x1F'
	};

	// printer wants data in little-endian
	void writeToStream(std::ostream &out, uint16_t value)
	{
		out.put((value >> 0) & 0xFF);
		out.put((value >> 8) & 0xFF);
	}
}

StripedHorse::DBEA4DFilter::DBEA4DFilter(std::istream &i, std::ostream &o, const Options &op)
  : Filter(i, o, op), m_twoSidedPrintingMode(TwoSidedPrintingMode::kSimplex), m_printOnBackOnly(false), m_cuttingMode(CuttingMode::kNoCutting),
    m_lastPageHeight(0)
{
	const std::string &twoSidedPrintingModeOption = m_options.getString("Duplex", "None");
	
	if (twoSidedPrintingModeOption == "None")
		m_twoSidedPrintingMode = TwoSidedPrintingMode::kSimplex;
	else if (twoSidedPrintingModeOption == "DuplexNoTumble")
		m_twoSidedPrintingMode = TwoSidedPrintingMode::kDuplex;
	else if (twoSidedPrintingModeOption == "DuplexTumble")
		m_twoSidedPrintingMode = TwoSidedPrintingMode::kDuplexTumble;
	else
		Logging::Warning() << "Unable to determine two-sided printing mode, defaulting to front-only" << std::endl;
	
	const std::string &cuttingModeOption = m_options.getString("CuttingMode", "None");

	if (cuttingModeOption == "None")
		m_cuttingMode = CuttingMode::kNoCutting;
	else if (cuttingModeOption == "EveryPage")
		m_cuttingMode = CuttingMode::kCutAfterEveryPage;
	else if (cuttingModeOption == "EndOfDocument")
		m_cuttingMode = CuttingMode::kCutAfterDocument;
	else
		Logging::Warning() << "Unable to detemrine cutting mode, defaulting to none" << std::endl;
	
	const std::string &printOnBackOnlyOption = m_options.getString("PrintOnBackOnly", "False");
	m_printOnBackOnly = (printOnBackOnlyOption == "True");
}

StripedHorse::DBEA4DFilter::~DBEA4DFilter()
{
}

void StripedHorse::DBEA4DFilter::beginDocument()
{
	initialisePrinter();

	if (m_twoSidedPrintingMode == TwoSidedPrintingMode::kSimplex)
		setThermalPrintingMode(m_printOnBackOnly ? ThermalPrintingMode::kSimplexBack : ThermalPrintingMode::kSimplexFront);
	else
		setThermalPrintingMode(ThermalPrintingMode::kDuplex);
}

StripedHorse::DBEA4DFilter::PrintingSide StripedHorse::DBEA4DFilter::determinePrintingSide() const
{
	switch (m_twoSidedPrintingMode)
	{
		case TwoSidedPrintingMode::kSimplex:
			return m_printOnBackOnly ? PrintingSide::kBack : PrintingSide::kFront;
		
		case TwoSidedPrintingMode::kDuplex:
		case TwoSidedPrintingMode::kDuplexTumble:
			return (getPageCount() % 2 == 0) ? PrintingSide::kBack : PrintingSide::kFront;

		default:
			throw InternalInconsistencyException("Unknown two-sided printing mode");
	}
}

void StripedHorse::DBEA4DFilter::endDocument()
{
	// a duplex document with an odd number of pages must still be printed
	if (m_twoSidedPrintingMode != TwoSidedPrintingMode::kSimplex && getPageCount() % 2 == 1)
	{
		setPrintingSide(PrintingSide::kBack);
		print();
		if (m_cuttingMode == CuttingMode::kCutAfterEveryPage)
			performKnifeCut();
	}

	if (m_cuttingMode == CuttingMode::kCutAfterDocument)
		performKnifeCut();
}

void StripedHorse::DBEA4DFilter::feedRows(size_t numberOfRows)
{
	// previously, maximum rows per command was set to 64, but this
	// caused problems with some firmware.
	const int kMaximumRowsPerCommand = 0xFFFF;
	if (numberOfRows == 0)
		return;
	
	while (true)
	{
		m_out.put(ASCII::US);
		m_out.put('a');
		if (numberOfRows > kMaximumRowsPerCommand)
		{
			writeToStream(m_out, static_cast<uint16_t>(kMaximumRowsPerCommand));
			numberOfRows -= kMaximumRowsPerCommand;
		}
		else
		{
			writeToStream(m_out, static_cast<uint16_t>(numberOfRows));
			break;
		}
	}
}

void StripedHorse::DBEA4DFilter::initialisePrinter()
{
	Logging::Debug() << "Initialising printer" << std::endl;
	m_out.put(ASCII::ESC);
	m_out.put('@');
}

void StripedHorse::DBEA4DFilter::performKnifeCut()
{
	Logging::Debug() << "Performing knife cut" << std::endl;
	m_out.put(ASCII::ESC);
	m_out.put('i');
}

void StripedHorse::DBEA4DFilter::print()
{
	if (m_twoSidedPrintingMode == TwoSidedPrintingMode::kSimplex)
	{
		Logging::Debug() << "Issuing simplex print command" << std::endl;
		m_out.put(ASCII::ESC);
		m_out.put(ASCII::FF);
	}
	else
	{
		Logging::Debug() << "Issuing duplex print command" << std::endl;
		m_out.put(ASCII::US);
		m_out.put('d');
	}
}

void StripedHorse::DBEA4DFilter::processPageBitmap(const LemonGFX::Bitmap &bitmap, const CUPSFormat &format)
{
	// In order to match the output of the Windows driver, we only set the page height and thermal mode once.
	if (getPageCount() == 1)
	{
		m_lastPageHeight = format.cupsHeight;
		setPageHeight(format.cupsHeight);
	}
	else if (m_lastPageHeight != format.cupsHeight)
	{
		Logging::Warning() << "The document contains pages of varying heights, the result may not be accurate." << std::endl;
	}

	PrintingSide currentSide = determinePrintingSide();
	Logging::Debug() << "Printing page " << getPageCount() << " on the " << (currentSide == PrintingSide::kFront ? "front" : "back") << std::endl;
	setPrintingSide(currentSide);

	const Buffer &data = bitmap.getData();
	
	// number of bytes per data row that the printer expects
	const size_t bytesRequired = 108;
	
	// number of bytes to take from each row (should not exceed bytesRequired)
	size_t bytesFromRow = std::min(bytesRequired, bitmap.getBytesPerRow());

	// padding bytes when bytesFromRow < bytesRequired
	size_t leftPadding = 0;
	size_t rightPadding = 0;

	if (bytesFromRow <= bytesRequired)
	{
		// if totalPadding is odd, put the odd byte on the right
		size_t totalPadding = bytesRequired - bytesFromRow;
		leftPadding = totalPadding / 2;
		rightPadding = (totalPadding + 1) / 2;
		Logging::Debug() << "Total padding: " << totalPadding << " bytes (meaning " << leftPadding << " on the left and " << rightPadding << " on the right)" << std::endl;
	}
	else
	{
		Logging::Info() << "Bitmap is wider than the destination page" << std::endl;
	}

	size_t y = 0;
	while (y < bitmap.getHeight())
	{
		if (bitmap.isRowBlank(y))
		{
			size_t blankCount = 0;
			do
			{
				blankCount++;
				y++;
			}
			while (y < bitmap.getHeight() && bitmap.isRowBlank(y));

			if (y >= bitmap.getHeight())
				break;
			
			feedRows(blankCount);
		}

		m_out.put(ASCII::ESC);
		m_out.put('.');
		m_out.put('\0');
		writeToStream(m_out, 1);
		m_out.put('\0');

		for (size_t x = 0; x < leftPadding; x++)
			m_out.put('\0');
		
		for (size_t x = 0; x < bytesFromRow; x++)
			m_out.put(data.at(y * bitmap.getBytesPerRow() + x));
		
		for (size_t x = 0; x < rightPadding; x++)
			m_out.put('\0');
			
		y++;
	}

	// issue the print command if the job is simplex or we've just finished
	// processing the back for duplex
	if (m_twoSidedPrintingMode == TwoSidedPrintingMode::kSimplex || (currentSide == PrintingSide::kBack))
	{
		print();
		if (m_cuttingMode == CuttingMode::kCutAfterEveryPage)
			performKnifeCut();
	}
}

void StripedHorse::DBEA4DFilter::setPageHeight(size_t pageHeight)
{
	Logging::Debug() << "Setting page height: " << pageHeight << std::endl;
	m_out.put(ASCII::ESC);
	m_out.put('P');
	writeToStream(m_out, static_cast<uint16_t>(pageHeight));
}

void StripedHorse::DBEA4DFilter::setPrintingSide(PrintingSide printingSide)
{
	m_out.put(ASCII::US);
	m_out.put('c');
	switch (printingSide)
	{
		case PrintingSide::kFront:
			Logging::Debug() << "Setting printing side to front" << std::endl;
			m_out.put('\x00');
			break;
		
		case PrintingSide::kBack:
			Logging::Debug() << "Setting printing side to back" << std::endl;
			m_out.put('\x01');
			break;
		
		default:
			throw InternalInconsistencyException("Unknown printing side");
	}
	m_out.put('\x00');
}

void StripedHorse::DBEA4DFilter::setThermalPrintingMode(ThermalPrintingMode thermalPrintingMode)
{
	// Select thermal printing mode (duplex)
	m_out.put(ASCII::US);
	m_out.put('b');
	switch (thermalPrintingMode)
	{
		case ThermalPrintingMode::kSimplexFront:
			Logging::Debug() << "Setting thermal printing mode to simplex (front only)" << std::endl;
			m_out.put('\x00');
			break;
		
		case ThermalPrintingMode::kSimplexBack:
			Logging::Debug() << "Setting thermal printing mode to simplex (back only)" << std::endl;
			m_out.put('\x01');
			break;
		
		case ThermalPrintingMode::kDuplex:
			Logging::Debug() << "Setting thermal printing mode to duplex" << std::endl;
			m_out.put('\x02');
			break;
		
		default:
			throw InternalInconsistencyException("Unknown thermal printing mode");
	}
	m_out.put('\x00');
}

bool StripedHorse::DBEA4DFilter::shouldTumbleBitmap() const
{
	// Tumbling not necessary for regular duplex or simplex
	if (m_twoSidedPrintingMode != TwoSidedPrintingMode::kDuplexTumble)
		return false;
	
	// Tumbling only occurs for "back" pages (i.e. even pages)
	return (getPageCount() % 2) == 0;
}

bool StripedHorse::DBEA4DFilter::tryGetStatusCode(int &code) const
{
	try
	{
		DBEA4DPrinter printer = DBEA4DPrinter::ReadPrinterConfigurationAndStatus();
		code = printer.getStatusCode();
		return true;
	}
	catch (DBEA4DPrinterException &ex)
	{
		Logging::Info() << "The printer status could not be retrieved: " << ex.what() << std::endl;
	}

	return false;
}
