#include <libcore/logging.h>
#include <libtecfilter/commands.h>
#include "cupsformat.h"
#include "cupssync.h"
#include "filter.h"

namespace
{
	[[noreturn]] void invalidCall()
	{
		throw StripedHorse::FilterException("This method should not be called");
	}
}

StripedHorse::Filter::Filter(std::istream &in, std::ostream &out, const Options &opts)
  : m_in(in), m_out(out), m_options(opts), m_aborted(false), m_started(false), m_pageCount(0)
{
}

StripedHorse::Filter::~Filter()
{	
}

void StripedHorse::Filter::abort()
{
	m_aborted = true;
}

void StripedHorse::Filter::begin()
{
	using namespace StripedHorse;
	using namespace StripedHorse::TPCLCommands;
	
	if (m_started || m_aborted)
		return;
	
	try
	{
		CUPSSync sync(m_in);
		
		m_started = true;
		m_pageCount = 0;
		
		if (m_onDocumentBegin)
			m_onDocumentBegin(*this);
		
		beginDocument();

		while (!m_aborted)
		{
			if (m_in.peek() && !m_in.good())
				break;
			
			CUPSFormat format(sync, m_in);
			
			Logging::Debug() << "cupsWidth: " << format.cupsWidth << std::endl;
			Logging::Debug() << "cupsHeight: " << format.cupsHeight << std::endl;
			Logging::Debug() << "pageSize.x: " << format.pageSize.x << std::endl;
			Logging::Debug() << "pageSize.y: " << format.pageSize.y << std::endl;
			Logging::Debug() << "cupsBitsPerColor: " << format.cupsBitsPerColor << std::endl;
			Logging::Debug() << "cupsBitsPerPixel: " << format.cupsBitsPerPixel << std::endl;
			Logging::Debug() << "cupsBytesPerLine: " << format.cupsBytesPerLine << std::endl;
			Logging::Debug() << "orientation: " << format.orientation << std::endl;
			Logging::Debug() << "resolution: " << format.resolution.horizontal << "x" << format.resolution.vertical << std::endl;

			LemonGFX::Bitmap bitmap;

			Buffer imageData(format.cupsHeight * format.cupsBytesPerLine);
			if (!m_in.read(reinterpret_cast<char *>(&imageData[0]), imageData.size()))
				throw IOException("Unable to read from input");

			if (format.cupsBitsPerColor == 1 && format.cupsBitsPerPixel == 1)
			{
				bitmap = LemonGFX::Bitmap(imageData, format.cupsWidth, format.cupsHeight);
			}
			else if (format.cupsBitsPerColor == 8 && format.cupsBitsPerPixel == 8)
			{
				LemonGFX::RectangleU destination(0, 0, format.cupsWidth, format.cupsHeight);
				LemonGFX::Image image(format.cupsBitsPerPixel, 1, format.cupsWidth, format.cupsHeight, std::move(imageData));
				
				// typically, LemonGFX::Image objects assume 1 is white since this is the most
				// common convention among non-bilevel image formats, however, for CUPS, 1 is
				// considered black, so we have to reverse the image so that 1 is considered
				// white by the halftoner.
				image.invert();
				
				LemonGFX::HalftoningMethod hm = LemonGFX::HalftoningMethod::kThresholding;
				if (m_options.getString("DocumentType") == "Graphics")
				{
					Logging::Info() << "Using SH halftoner" << std::endl;
					hm = LemonGFX::HalftoningMethod::kStripedHorse9Toner;
				}
				else
				{
					Logging::Info() << "Using thresholding" << std::endl;
				}
				
				LemonGFX::HalftoningCurve hc = LemonGFX::HalftoningCurveFromName(m_options.getString("ImageDarkness", "Normal"));

				bitmap = LemonGFX::Bitmap::Downsample(image, destination, 10.0, LemonGFX::BitmapRotation::kRotate0, hm, hc);
			}
			else
			{
				throw FilterException("Incorrect raster image format");
			}

			m_pageCount++;

			if (m_onPageBegin)
				m_onPageBegin(*this);

			if (shouldTumbleBitmap())
			{
				Logging::Debug() << "The page needs to be tumbled" << std::endl;
				bitmap.rotate180();
			}
			else
			{
				Logging::Debug() << "The page does not need to be tumbled" << std::endl;
			}
			
			processPageBitmap(bitmap, format);
	
			if (m_onPageEnd)
				m_onPageEnd(*this);
			
			m_out.flush();
		}

		endDocument();
		
		if (m_onDocumentEnd)
			m_onDocumentEnd(*this);
	}
	catch (...)
	{
		m_aborted = true;
		throw;
	}
}

void StripedHorse::Filter::beginDocument()
{
}

void StripedHorse::Filter::endDocument()
{
}

uint32_t StripedHorse::Filter::getPageCount() const
{
	return m_pageCount;
}

bool StripedHorse::Filter::didAbort() const
{
	return m_aborted;
}

void StripedHorse::Filter::insert(const std::string &insertion)
{
	if (!insertion.empty())
		m_out << insertion;
}

void StripedHorse::Filter::processPageBitmap(const LemonGFX::Bitmap &bitmap, const CUPSFormat &format)
{
	(void) bitmap;
	(void) format;
	invalidCall();
}

void StripedHorse::Filter::setOnDocumentBegin(std::function<void(Filter &)> func)
{
	m_onDocumentBegin = func;
}

void StripedHorse::Filter::setOnDocumentEnd(std::function<void(Filter &)> func)
{
	m_onDocumentEnd = func;
}

void StripedHorse::Filter::setOnPageBegin(std::function<void(Filter &)> func)
{
	m_onPageBegin = func;
}

void StripedHorse::Filter::setOnPageEnd(std::function<void(Filter &)> func)
{
	m_onPageEnd = func;
}

bool StripedHorse::Filter::shouldTumbleBitmap() const
{
	return false;
}

bool StripedHorse::Filter::tryGetStatusCode(int &code) const
{
	(void) code;
	invalidCall();
}
