#ifndef STRIPEDHORSE_FILTER_H
#define STRIPEDHORSE_FILTER_H

#include <cstdint>
#include <functional>
#include <iostream>
#include <map>
#include <string>

#include <lemongfx/bitmap.h>
#include <libcore/exceptions.h>
#include <libtecfilter/options.h>
#include <libtecfilter/ppdfile.h>

#include "cupsformat.h"

namespace StripedHorse
{
	class Filter
	{
	public:
		/**
		 Constructs a filter object
		 @param[in] i the source of the data to filter
		 @param[in] o the destination of the filtered data
		 @param[in] ops options to fine-tune the filtering process
		 @param[in] ppd the PPD used for printing
		 */
		Filter(std::istream &i, std::ostream &o, const Options &op);
		Filter(const Filter &) = delete;
		virtual ~Filter();
		
		/**
		 Aborts the filter's processing. Once called, no further pages/labels
		 will be processed.
		 */
		void abort();
		
		/**
		 Begins the filter's processing. This does not occur straight away
		 upon construction to give a chance to set up event handlers.
		 */
		void begin();
		
		/**
		 Determines whether the filter was aborted
		 @return true if the filter was aborted (either by a call to abort() or
		         due to an uncorrectable error in processing).
		 */
		bool didAbort() const;
		
		/**
		 Gets the number of pages processed so far
		 @return the number of pages processed so far
		 */
		uint32_t getPageCount() const;

		/**
		 Gets the status code (a number returned by the printer or
		 generated by the filter that can be translated into a message string
		 based on information in a PPD. This method is const-qualified
		 because it should use CUPS sidechanneling to communicate with the
		 printer
		*/
		virtual bool tryGetStatusCode(int &code) const;
		
		/**
		 Arbitrarily inserts a string into the output
		 @param[in] str the string to insert
		 */
		void insert(const std::string &str);
		
		/**
		 Sets a handler that is called when the document has begun
		 @param[in] f the function to call when the filter begins processing
		            the document
		 */
		void setOnDocumentBegin(std::function<void(Filter &)> f);
		
		/**
		 Sets a handler that is called when the document has ended
		 @param[in] f the function to call when the filter has finished
		            processing the document
		 */
		void setOnDocumentEnd(std::function<void(Filter &)> f);
		
		/**
		 Sets a handler that is called when a page has begun
		 @param[in] f the function to call when the filter is about to process
		            the next page in the document
		 */
		void setOnPageBegin(std::function<void(Filter &)> f);
		
		/**
		 Sets a handler that is called after the page has ended
		 @param[in] f the function to call when the filter has finished
		            processing the page.
		 */
		void setOnPageEnd(std::function<void(Filter &)> f);
		
	protected:
		virtual void beginDocument();
		virtual void endDocument();
		virtual void processPageBitmap(const LemonGFX::Bitmap &bitmap, const CUPSFormat &format);
		virtual bool shouldTumbleBitmap() const;

		std::istream &m_in;
		std::ostream &m_out;
		Options m_options;

		bool m_aborted;
		bool m_started;
		uint32_t m_pageCount;
		
		std::function<void(Filter &)> m_onDocumentBegin;
		std::function<void(Filter &)> m_onDocumentEnd;
		std::function<void(Filter &)> m_onPageBegin;
		std::function<void(Filter &)> m_onPageEnd;
	};
	
	class FilterException : public Exception
	{
	public:
		FilterException(const std::string &m) : Exception(m) {}
	};
}

#endif

