#include <libcore/logging.h>
#include <fstream>
#include <regex>
#include <sstream>
#include "insertion.h"

namespace
{
	std::string performSubstitutions(const std::string &tmpl, const StripedHorse::Dictionary &dict)
	{
		static const std::regex escape("%[A-Za-z0-9%]");
		std::ostringstream result;
		
		std::string::size_type lastReplacement = 0;
		
		std::sregex_iterator templateStart(tmpl.begin(), tmpl.end(), escape);
		std::sregex_iterator templateEnd;
		
		for (std::sregex_iterator iter = templateStart; iter != templateEnd; iter++)
		{
			const std::string &word = iter->str();
			auto definition = dict.find(word);
			if (definition == dict.end())
			{
				StripedHorse::Logging::Info() << "Unknown value: " << word << std::endl;
				continue; // bypass it
			}
			
			result << tmpl.substr(lastReplacement, iter->position() - lastReplacement);
			result << definition->second;
			
			lastReplacement = iter->position() + iter->length();
		}
		
		result << tmpl.substr(lastReplacement);
		
		return result.str();		
	}
	
	std::string performSubstitutionsOnStream(std::istream &in, const StripedHorse::Dictionary &dict)
	{
		std::string line;
		std::ostringstream result;
		
		while (std::getline(in, line))
			result << performSubstitutions(line, dict) << '\n';
		
		return result.str();
	}
}

StripedHorse::Insertion::Insertion()
  : m_insertionType(None)
{
}

StripedHorse::Insertion::Insertion(const Insertion &other)
{
	(void) this->operator=(other);
}

StripedHorse::Insertion::Insertion(Insertion &&other)
  : m_insertionType(other.m_insertionType), m_template(std::move(other.m_template))
{
}

StripedHorse::Insertion::Insertion(enum InsertionType insertionType, const std::string &tmpl)
  : m_insertionType(insertionType), m_template(tmpl)
{
}

StripedHorse::Insertion &StripedHorse::Insertion::operator=(const Insertion &other)
{
	m_insertionType = other.m_insertionType;
	m_template = other.m_template;
	return *this;
}

StripedHorse::Insertion::InsertionType StripedHorse::Insertion::getInsertionType() const
{
	return m_insertionType;
}

std::string StripedHorse::Insertion::getInsertion(const Dictionary &dict) const
{
	if (m_insertionType == None || dict.size() == 0)
		return "";
	
	switch (m_insertionType)
	{
		case None:
		{
			return "";
		}
		
		case String:
		{
			return performSubstitutions(m_template, dict);
		}
		
		case File:
		{
			std::string filename = performSubstitutions(m_template, dict);
			std::ifstream input(filename);
			if (!input)
			{
				Logging::Info() << "Unable to read file: " << filename << std::endl;
				return "";
			}
			else
			{
				return performSubstitutionsOnStream(input, dict);
			}
		}
	}
	
	// Should not reach here...
	return "";
}

std::string StripedHorse::Insertion::getTemplate() const
{
	return m_template;
}
