#define __USE_POSIX 1

#include <cctype>
#include <cstring>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <fstream>
#include <limits.h>
#include <sstream>
#include <set>

#ifdef _WIN32
#include <winsock2.h>
#include <windows.h>
namespace
{
	enum { kMaxHostName = 256 };
}
#else
#include <unistd.h>
#ifdef _POSIX_HOST_NAME_MAX
#define SH_HOST_NAME_MAX _POSIX_HOST_NAME_MAX
#else
#define SH_HOST_NAME_MAX 256
#endif
namespace
{
	enum { kMaxHostName = SH_HOST_NAME_MAX };
}
#endif

#include <libcore/exceptions.h>
#include <libcore/logging.h>
#include <libtecfilter/options.h>
#include <libtecfilter/options-utilities.h>
#include <libtecfilter/ppdfile.h>
#include <libtecfilter/string-escape.h>
#include <libtecfilter/version.h>
#include "dbea4dfilter.h"
#include "tpclfilter.h"
#include "zplfilter.h"
#include "insertion.h"

#include "libtecfilter/runtimevalidator.inc.cpp"

namespace
{
	enum class PrinterLanguage
	{
		kDefer, // defer language selection
		kTPCL,
		kZPL,
		kDBEA4D
	};

	PrinterLanguage parsePrinterLanguage(const std::string &language)
	{
		if (language == "DBEA4D")
			return PrinterLanguage::kDBEA4D;
		else if (language == "TPCL")
			return PrinterLanguage::kTPCL;
		else if (language == "ZPL")
			return PrinterLanguage::kZPL;
		else
			return PrinterLanguage::kDefer;
	}

	struct RuntimeData
	{
		std::unordered_map<std::string, std::unordered_map<std::string, StripedHorse::Insertion>> insertions;
		std::unordered_map<int, std::string> statusText;
		std::unordered_set<int> nonErrors;
		std::string dateFormat = "%Y-%m-%d";
		std::string timeFormat = "%H:%M:%S";
		PrinterLanguage printerLanguage = PrinterLanguage::kTPCL;
	};
	
	std::string getHostName()
	{
		char buffer[kMaxHostName] = { '\0' };
		if (gethostname(buffer, sizeof buffer) != 0)
			return "unknown";
		
		// POSIX does not guarantee null termination if the host name was
		// too large to fit.
		buffer[sizeof buffer - 1] = '\0';
		return buffer;
	}
	
	RuntimeData loadRuntimeDataFromPPD(const StripedHorse::PPDFile &ppd)
	{
		RuntimeData result;
		using namespace StripedHorse;
		
		for (const PPDStatement &status : ppd.getAllStatementsWithMainKeyword("TECStatusText"))
		{
			try
			{
				if (status.hasOptionKeyword())
				{
					int code = std::stoi(status.getOptionKeyword());
					result.statusText.emplace(code, static_cast<std::string>(status.getValue()));
				}
			}
			catch (std::invalid_argument &)
			{
				// ignore
			}
		}
		
		PPDStatement st;
		if (ppd.tryGetAnyStatementWithMainKeyword("TECStatusNonError", st) && st.hasValue())
		{
			std::istringstream ss(static_cast<std::string>(st.getValue()));
			std::copy(std::istream_iterator<int>(ss), std::istream_iterator<int>(), std::inserter(result.nonErrors, result.nonErrors.begin()));
		}
		
		std::vector<std::string> split1 { "Pre", "Post" };
		std::vector<std::string> split2 { "Job", "Label" };
		
		for (const std::string &s1 : split1)
		{
			for (const std::string &s2 : split2)
			{
				std::string fullname = "TEC" + s1 + s2 + "Insertion";
				if (ppd.tryGetAnyStatementWithMainKeyword(fullname, st) && st.hasOptionKeyword())
				{
					Insertion::InsertionType insType = st.getOptionKeyword() == "File" ? Insertion::File : Insertion::String;
					result.insertions[s1][s2] = Insertion(insType, UnescapeString(static_cast<std::string>(st.getValue())));
				}
			}
		}
		
		if (ppd.tryGetAnyStatementWithMainKeyword("TECInsertionDateFormat", st) && st.hasValue())
			result.dateFormat = static_cast<std::string>(st.getValue());
		
		if (ppd.tryGetAnyStatementWithMainKeyword("TECInsertionTimeFormat", st) && st.hasValue())
			result.timeFormat = static_cast<std::string>(st.getValue());
		
		if (ppd.tryGetAnyStatementWithMainKeyword("TECBarcodePrinterLanguage", st) && st.hasValue())
		{
			std::string value = static_cast<std::string>(st.getValue());
			result.printerLanguage = parsePrinterLanguage(value);
		}

		return result;
	}

	/**
	 * If the user has attempted to set some dumb values through the command
	 * line (such as choosing a print speed that doesn't exist) then this
	 * function will attempt to correct it. If any settings were updated,
	 * then this function shall return true, otherwise it shall return false.
	 */
	bool fixOptionsWithPPD(StripedHorse::Options &options, const StripedHorse::PPDFile &ppdFile)
	{
		bool didAmendOptions = false;

		// Check to see whether the user has attempted to set these
		// values which may not be compatible with the target model,
		// and if this is the case, they are removed from the options
		std::vector<std::string> nonGenericOptions {
			"BackFeedAdjustment",
			"FeedAdjustment",
			"CutStripAdjustment",
			"ToneAdjustmentThermalTransfer",
			"ToneAdjustmentDirectThermal"
		};

		for (const auto &optionName : nonGenericOptions)
		{
			if (options.hasOption(optionName) && !ppdFile.hasOption(optionName))
			{
				StripedHorse::Logging::Warning() << "Unsupported option " << optionName << " is ignored for this model" << std::endl;
				options.remove(optionName);
				didAmendOptions = true;
			}
		}

		// the following options can drastically affect the output,
		// and in some cases may cause a jam, so we need to make sure
		// they are set to valid values
		std::vector<std::string> highPriorityOptions {
			"PrintSpeed",
			"TransferMode"
		};

		for (const auto &optionName : highPriorityOptions)
		{
			std::string oldValue = options.getString(optionName);
			if (ppdFile.hasOption(optionName) && !ppdFile.hasOptionWithChoice(optionName, oldValue))
			{
				std::string newValue = ppdFile.getDefaultChoiceForOption(optionName);
				StripedHorse::Logging::Warning() << "Invalid choice for " << optionName << ", replacing \"" << oldValue << "\" with default value \"" << newValue << "\"" << std::endl;
				options.set(optionName, newValue);
				didAmendOptions = true;
			}
		}

		return didAmendOptions;
	}
	
	std::string formatTime(struct tm *tm, const std::string &format)
	{
		std::ostringstream result;
		result << std::put_time(tm, format.c_str());
		return result.str();
	}
	
	void updateInsertionDictionary(StripedHorse::Dictionary &dict, const RuntimeData &runtimeData)
	{
		time_t t = time(nullptr);
		struct tm *tm = localtime(&t);
		
		dict["%t"] = formatTime(tm, runtimeData.timeFormat);
		dict["%d"] = formatTime(tm, runtimeData.dateFormat);
	}
}

int main(int argc, char *argv[])
{
	using namespace StripedHorse;

	if (argc >= 2 && strcmp(argv[1], "--version") == 0)
	{
		std::cerr << StripedHorse::g_version << std::endl;
		return 0;
	}
	else if (argc < 6)
	{
		std::cerr << "Usage\n";
		std::cerr << argv[0] << " job user title num-copies options [ filename ]\n";
		return 1;
	}
	
	Logging::SetMessagePrefixes("DEBUG", "INFO", "WARNING", "ERROR", "CRIT");

	std::ifstream fsrc;
	std::istream *src = &std::cin;
	if (argc > 6)
	{
		fsrc.open(argv[6]);
		src = &fsrc;
	}
		
	try
	{
		Dictionary dictionary;	
		dictionary["%%"] = "%";
		dictionary["%h"] = getHostName();
		dictionary["%i"] = argv[1];
		dictionary["%j"] = argv[3];
		dictionary["%p"] = "0";
		dictionary["%u"] = argv[2];

		Options options;
		RuntimeData runtimeData;

		std::unique_ptr<PPDFile> ppdFile;
		
		try
		{
			const char *ppdFilename = getenv("PPD");
			if (ppdFilename != nullptr)
			{
				ppdFile = std::make_unique<PPDFile>(ppdFilename);
				runtimeData = loadRuntimeDataFromPPD(*ppdFile);
				if (!OptionsUtilities::ParseDefaultValuesFromPPD(options, *ppdFile))
					Logging::Warning() << "PPD file was specified but no values were loaded" << std::endl;
				
				if (!OptionsUtilities::ParseMinAndMaxValuesFromPPD(options, *ppdFile))
					Logging::Warning() << "PPD file was specified but no minimum or maximum values for custom options were loaded" << std::endl;
			}
			else
			{
				Logging::Warning() << "Filter invoked without PPD environment variable set" << std::endl;
			}
		}
		catch (PPDException &ex)
		{
			Logging::Error() << "Exception occurred while processing PPD file: " << ex.what() << std::endl;
		}

		// update options with the ones provided by CUPS via command line argument
		options.set("JobID", argv[1]);
		options.set("User", argv[2]);
		options.set("Title", argv[3]);
		options.set("Copies", argv[4]);
		options.parse(argv[5]);
		
		if (ppdFile)
			if (fixOptionsWithPPD(options, *ppdFile))
				StripedHorse::Logging::Warning() << "Print settings were amended" << std::endl;
		
		std::unique_ptr<Filter> f;
		
		// The printer language may be chosen at runtime rather than at install time
		// so if the installed language says "defer" then we obtain the printer
		// language from the CUPS command line arguments instead
		if (runtimeData.printerLanguage == PrinterLanguage::kDefer)
			runtimeData.printerLanguage = parsePrinterLanguage(options.getString("PrinterLanguage", "TPCL"));
		
		switch (runtimeData.printerLanguage)
		{
			case PrinterLanguage::kTPCL:
				Logging::Debug() << "Printer language is TPCL" << std::endl;
				f = std::make_unique<TPCLFilter>(*src, std::cout, options);
				break;

			case PrinterLanguage::kDBEA4D:
				Logging::Debug() << "Printer language is DB-EA4D" << std::endl;
				f = std::make_unique<DBEA4DFilter>(*src, std::cout, options);
				break;
			
			case PrinterLanguage::kZPL:
				Logging::Debug() << "Printer language is ZPL" << std::endl;
				f = std::make_unique<ZPLFilter>(*src, std::cout, options);
				break;
			
			default:
				throw InternalInconsistencyException("Printer language is unknown");
		}
		
		f->setOnPageBegin([&dictionary,&runtimeData](Filter &filter)
		{
			updateInsertionDictionary(dictionary, runtimeData);
			dictionary["%p"] = std::to_string(filter.getPageCount());
			filter.insert(runtimeData.insertions["Pre"]["Label"].getInsertion(dictionary));

			Logging::Info() << "Processing page: " << filter.getPageCount() << std::endl;
		});
		
		f->setOnPageEnd([&dictionary,&runtimeData](Filter &filter)
		{
			updateInsertionDictionary(dictionary, runtimeData);
			dictionary["%p"] = std::to_string(filter.getPageCount());
			filter.insert(runtimeData.insertions["Post"]["Label"].getInsertion(dictionary));

			// tell CUPS the total number of pages issued
			std::cerr << "PAGE: " << filter.getPageCount() << " 1" << std::endl;
		});
		
		f->setOnDocumentBegin([&dictionary,&runtimeData](Filter &filter){
			updateInsertionDictionary(dictionary, runtimeData);
			filter.insert(runtimeData.insertions["Pre"]["Job"].getInsertion(dictionary));
		});
		
		f->setOnDocumentEnd([&dictionary,&runtimeData](Filter &filter){
			updateInsertionDictionary(dictionary, runtimeData);
			filter.insert(runtimeData.insertions["Post"]["Job"].getInsertion(dictionary));
		});
		
		f->begin();
		
		if (f->didAbort())
			return 1;
		else
			return 0;
	}
	catch (Exception &ex)
	{
		// All fixable exceptions should be caught before this handler, so if
		// we catch an exception here it means it's a serious problem.
		Logging::Critical() << "An exception occurred during filter execution: " << ex.what() << std::endl;
		return 1;
	}
}

