#include <algorithm>
#include <cmath>
#include <sstream>
#include <lemongfx/bitmap.h>
#include <libcore/logging.h>
#include <libtecfilter/commands.h>
#include <libtecfilter/options-utilities.h>
#include <libtecfilter/stream-utilities.h>
#include <libtecfilter/tpclprinterstatus.h>
#include "cupsformat.h"
#include "tpclfilter.h"

namespace
{
	uint32_t getThermalMode(const StripedHorse::Options &opts)
	{
		std::string method = opts.getString("PrintMethod");
		
		if (method == "DirectThermal")
			return 0;
		else if (method == "ThermalTransferRibbonSaving")
			return 1;
		else if (method == "ThermalTransfer")
			return 2;
		else if (method == "DirectThermalHeadUp")
			return 3;
		else
			throw StripedHorse::FilterException("The filter is unable to determine whether to use direct or transfer thermal mode.");
	}
}

StripedHorse::TPCLFilter::TPCLFilter(std::istream &i, std::ostream &o, const Options &op)
  : Filter(i, o, op), m_writer(o)
{
    m_thermalMode = getThermalMode(m_options);

	if (m_options.getString("CommandCharacters", "Readable") == "Unreadable")
		m_writer.setOutputStyle({ '\x1B' }, { '\n', '\x00' }, false);
}

bool StripedHorse::TPCLFilter::tryGetStatusCode(int &code) const
{
	try
	{
		TPCLPrinterStatus printerStatus = TPCLPrinterStatus::ReadPrinterStatus();
		code = printerStatus.getStatusCode();
		return true;
	}
	catch (EmptyTPCLPrinterStatusException &)
	{
		// This isn't really a problem
		Logging::Info() << "Printer did not report any abnormal status" << std::endl;
	}
	catch (InvalidTPCLPrinterStatusException &)
	{
		// This is unusual but also not really a problem
		Logging::Info() << "The printer's status could not be understood" << std::endl;
	}
	catch (TPCLPrinterStatusException &ex)
	{
		// Something completely unexpected happened - could be a problem
		Logging::Error() << ex.what() << std::endl;
	}

	return false;
}

void StripedHorse::TPCLFilter::processPageBitmap(const LemonGFX::Bitmap &bitmap, const CUPSFormat &format)
{
	// Obtain label gap (if custom, it is in millimetres, if not custom it is in 10th of mm)
	int32_t labelGap = 20;
	if (!OptionsUtilities::TryGetClampedValueIn10dpmm(m_options, "LabelGap", labelGap))
		Logging::Debug() << "Unable to determine label gap, using " << labelGap << std::endl;
	
	// need to convert page sizes from 72dpi to 10dpmm
	uint32_t pageWidth = 254 * format.pageSize.x / 72;
	uint32_t pageHeight = 254 * format.pageSize.y / 72;

	m_writer.writeCommand("Daaaa,bbbb,cccc", pageHeight + labelGap, pageWidth, pageHeight);

	// Issue feed amount, cut position, and backfeed
	m_writer.writeAdjustmentsFromOptions(m_options);
	
	uint32_t numCopies = format.numCopies;
	int32_t cutInterval = m_options.getInt32("CutInterval");
	uint32_t printSpeed = m_options.getUInt32("PrintSpeed", 3);
	uint32_t sensor = 0;
	if (m_options.getString("Sensor") == "Reflective")
		sensor = 1;
	else if (m_options.getString("Sensor") == "Transmissive")
		sensor = 2;
	
	if (cutInterval > 999)
		cutInterval = 999;
	else if (cutInterval < 0)
		cutInterval = 0;

	// Clear image buffer
	m_writer.writeCommand("C");

	RectangleU croppedDimensions = bitmap.getAutoCroppedDimensions();
	if (croppedDimensions.upperRight.x == 0 || croppedDimensions.upperRight.y == 0)
	{
		// Log for debugging purposes
		Logging::Debug() << "Page " << m_pageCount << " is blank" << std::endl;
	}
	else
	{
		LemonGFX::Bitmap cropped = bitmap.cut(croppedDimensions);
		
		uint32_t imageX = 254 * croppedDimensions.lowerLeft.x / format.resolution.horizontal;
		uint32_t imageY = 254 * croppedDimensions.lowerLeft.y / format.resolution.vertical;
		uint32_t imageWidth = cropped.getWidth();
		uint32_t imageHeight = cropped.getHeight();
		
		// Output page graphics
		char graphicsMode = '1';
		
		std::string compressionMode = m_options.getString("Compression", "Auto");
		std::stringstream topixCompressed;

		if (compressionMode == "Auto" || compressionMode == "TOPIX")
		{
			try
			{
				cropped.compress(LemonGFX::Bitmap::CompressionFormat::kTOPIX, topixCompressed);
			}
			catch (LemonGFX::BitmapException &ex)
			{
				// for whatever reason, the data couldn't be compressed
				compressionMode = "None";
			}
		}
		
		if (compressionMode == "Auto")
		{
			// If the TOPIX-compressed data has a greater size than the
			// uncompressed data then there's no point sending the
			// compressed data
			size_t topixSize = topixCompressed.tellp();
			size_t uncompressedSize = cropped.getDataSize();

			if (topixSize <= uncompressedSize)
				compressionMode = "TOPIX";
			else
				compressionMode = "None";

			StripedHorse::Logging::Debug() << "Compression mode is auto, TOPIX is " << topixSize
										<< " bytes and uncompressed it is " << uncompressedSize
										<< ", choosing " << compressionMode << std::endl;
		}
		
		if (compressionMode == "TOPIX")
		{
			// as a special case for TOPIX compressed images, the height
			// parameter is used as the scaling parameter.
			//
			// For 203dpi models and 300dpi models, a value of 300 here
			// means to print at normal resolution and 150 means to print
			// at double resolution.
			//
			// For 600dpi models, a value of 600 means normal resolution,
			// 300 means double, and 150 means quardruple.
			if (format.resolution.horizontal >= 600)
				imageHeight = 600;
			else
				imageHeight = 300;
			
			graphicsMode = '3';
			m_writer.writeCommand("SG;aaaa,bbbb,cccc,dddd,e,g/0", imageX, imageY, imageWidth, imageHeight, graphicsMode, StreamUtilities::ReadStreamToBuffer(topixCompressed));
		}
		else
		{
			graphicsMode = '1';
			m_writer.writeCommand("SG;aaaa,bbbb,cccc,dddd,e,g/0", imageX, imageY, imageWidth, imageHeight, graphicsMode, cropped.getData());
		}
	}
	
	// Any custom RFID user memory data?
	if (m_options.isCustom("RFIDUserMemData"))
	{
		std::string rfidUser = m_options.getString("RFIDUserMemData");
		if (rfidUser.size() > 0)
			m_writer.writeCommand("XBaa;bbbb,cccc,d,Tff,Gk,Vs,Btt=w/512", 0, 0, 0, "r", 24, 0, 1, 3, rfidUser);
	}
	
	// Any custom RFID EPC memory data?
	if (m_options.isCustom("RFIDEPCMemData"))
	{
		std::string rfidEPC = m_options.getString("RFIDEPCMemData");
		if (rfidEPC.size() > 0)
			m_writer.writeCommand("XBaa;bbbb,cccc,d,Tff,Gk,Vs,Btt=w/512", 0, 0, 0, "r", 24, 0, 1, 1, rfidEPC);
	}

	// Issue the label
	std::string issueMode = "C";
	if (cutInterval > 0)
	{
		std::string cutModule = m_options.getString("CutModule", "None");
		if (cutModule == "Partial")
			issueMode = "F";
		else if (cutModule == "Linerless")
			issueMode = "G";
	}
	else if (m_options.getBoolean("StripLabel"))
	{
		issueMode = "D";
	}
	
	m_writer.writeCommand("XS;I,aaaa,bbbcde$fgh", numCopies, cutInterval, sensor, issueMode, printSpeed, m_thermalMode, 0, 0);
}

