#include <sstream>
#include "zplfilter.h"

namespace
{
	inline void writeHexByte(std::ostream &out, unsigned char byte)
	{
		static const char nybbles[] = {
			'0', '1', '2', '3', '4', '5', '6', '7',
			'8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
		};

		out.put(nybbles[(byte >> 4) & 0xF]);
		out.put(nybbles[byte & 0xF]);
	}

	std::string formatDocumentNameForZPL(const std::string &name)
	{
		std::string result;
		result.reserve(20);

		for (const auto &ch : name)
		{
			if (ch == '^' || ch == '~')
				result.push_back('_');
			else
				result.push_back(ch);
			
			if (result.size() == 20)
				break;
		}

		return result;
	}
}

StripedHorse::ZPLFilter::ZPLFilter(std::istream &i, std::ostream &o, const Options &op)
  : Filter(i, o, op)
{

}

bool StripedHorse::ZPLFilter::tryGetStatusCode(int &status) const
{
	// in this case we don't know what we're doing
	status = 0;
	return false;
}

void StripedHorse::ZPLFilter::processPageBitmap(const LemonGFX::Bitmap &bmp, const CUPSFormat &format)
{
	(void) format;

	size_t bytesPerRow = bmp.getBytesPerRow();
	LemonGFX::RectangleU croppedDimensions = bmp.getAutoCroppedDimensions();

	m_out << "~DGR:PAGE.GRF," << (croppedDimensions.getHeight() * bytesPerRow) << "," << bytesPerRow << ",\n";

	// ZPL graphics data is written as hexadecimal
	const LemonGFX::Buffer &data = bmp.getData();
	for (RectangleU::type y = croppedDimensions.lowerLeft.y; y <= croppedDimensions.upperRight.y; y++)
	{
		for (RectangleU::type x = 0; x < bytesPerRow; x++)
		{
			uint32_t i = y * bmp.getBytesPerRow() + x;
			writeHexByte(m_out, data[i]);

            // output newline every 40 bytes (80 chars) unless it is the last byte
			// (newline will be appended later)
			if (((((y - croppedDimensions.lowerLeft.y) * bytesPerRow) + x) % 40) == 0)
				if (x > 0 && x < bytesPerRow - 1 && y <= croppedDimensions.upperRight.y)
					m_out << "\n";
		}
	}

	// start label format
	m_out << "\n^XA\n";

	// these ones only need to be done for the first label
	if (m_pageCount == 1)
	{
		m_out << "^FX JOB NAME: " << formatDocumentNameForZPL(m_options.getString("Title", "Untitled")) << "^FS\n";

		// cut interval and strip
		if (m_options.getUInt32("CutInterval") > 0)
			m_out << "^MMC,\n";
		else if (m_options.getBoolean("StripLabel"))
			m_out << "^MMP,\n";

		// print speed
		uint32_t printSpeed = m_options.getUInt32("PrintSpeed", 2);
		m_out << "^PR" << printSpeed << "," << printSpeed << ",\n";
	}

	// place the image on the page
	m_out << "^FO0," << croppedDimensions.lowerLeft.y << "\n";
	m_out << "^IMR:PAGE.GRF^FS\n";

	// end label format
	m_out << "^XZ\n";

	// delete page image
	m_out << "^IDR:PAGE.GRF^FS\n";
}
