#include <fstream>
#include <iostream>
#include <regex>
#include <lemongfx/bitmap.h>
#include <libcore/endian.h>
#include <libcore/types.h>
#include "cmd-render.h"

namespace
{
	StripedHorse::PointU parseForBitmapSize(std::istream &input, size_t dpmm)
	{
		StripedHorse::PointU result;
		
		std::regex labelSize("^\\{D(\\d{4,5}),(\\d{4}),(\\d{4,5})(,(\\d{4}))?\\|\\}\r?");
		
		std::string line;
		while (std::getline(input, line))
		{
			std::smatch match;
			if (std::regex_match(line, match, labelSize))
			{
				size_t width = std::stoi(match[2]);
				size_t height = std::stoi(match[1]);
				
				return StripedHorse::PointU(width * dpmm / 10, height * dpmm / 10);
			}
		}
		
		throw StripedHorse::InvalidFormatException("End of file reached before label size command found");
	}
	
	std::string parseNextArgument(std::istream &input)
	{
		char buf[7] = {0};
		
		if (!input.getline(buf, sizeof buf, ','))
			throw StripedHorse::IOException("Unexpected end of file");
		
		if (buf[0] == '\0')
			throw StripedHorse::InvalidFormatException("Unexpected value");
		
		return std::string { buf };
	}
	
	LemonGFX::Bitmap parseSG(std::istream &input, size_t dpmm)
	{
		using namespace StripedHorse;
		using namespace LemonGFX;
		
		int pch = input.peek();
		if (pch == std::istream::traits_type::eof())
			throw IOException("Unexpected end of input");
		
		bool expectSize = false;
		if (pch == '0')
		{
			input.get();
			expectSize = true;
		}
		
		char ch;
		if (!input.get(ch))
			throw IOException("Unexpected end of input");
		else if (ch != ';')
			throw InvalidFormatException("Syntax error");
		
		std::string xPosStr = parseNextArgument(input);
		std::string yPosStr = parseNextArgument(input);
		
		size_t width = std::stoul(parseNextArgument(input));
		size_t height = std::stoul(parseNextArgument(input));
		size_t xPos = std::stoul(xPosStr);
		size_t yPos = std::stoul(yPosStr);
		
		if (*(xPosStr.end() - 1) == 'D')
			xPos = xPos * dpmm / 10;
		
		if (*(yPosStr.end() - 1) == 'D')
			yPos = yPos * dpmm / 10;
		
		std::string formatStr = parseNextArgument(input);
		if (formatStr.size() > 1 && formatStr[0] == 'M')
		{
			// No idea what the M arg does, so we'll just skip it
			formatStr = parseNextArgument(input);
		}
		
		size_t format = std::stoul(formatStr, nullptr, 16);
		size_t expectedSize = 0;
		
		if (expectSize)
		{
			union
			{
				char c[4];
				uint32_t u;
			} sizeArg;
			 
			if (!input.read(&sizeArg.c[0], 4))
				throw IOException("Unexpected end of file");
			
			expectedSize = FromBigEndian(sizeArg.u);
			
			if (!input.get(ch)) // discard the next comma
				throw IOException("Unexpected end of file");
		}
		
		if (expectedSize == 0)
		{
			switch (format)
			{
				case 0x01:
				{
					size_t bytesPerRow = ((width + 7) / 8);
					expectedSize = bytesPerRow * height;
					break;
				}
				
				case 0x03:
				{
					uint16_t topixSize;
					if (input.read(reinterpret_cast<char *>(&topixSize), sizeof topixSize))
						expectedSize = FromBigEndian(topixSize);
					else
						throw IOException("Unable to read TOPIX data size");
					break;
				}
				
				case 0x0A:
				{
					// expected size already parsed
					break;
				}
				
				default:
					throw UnimplementedException("Graphics format unsupported (" + formatStr + ")");
			}
		}
		
		Buffer bitmapData(expectedSize);
		if (!input.read(reinterpret_cast<char *>(&bitmapData[0]), expectedSize))
			throw IOException("Unable to read expected amount of data");
		
		if (!input.get(ch) || ch != '|')
			throw InvalidFormatException("Graphics data does not match format");
		
		input.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
		
		Bitmap result;
		switch (format)
		{
			case 0x01: // Normal
			{
				result = Bitmap(bitmapData, width, height);
				break;
			}
			
			case 0x03: // TOPIX compressed
			{
				result = Bitmap::Load(bitmapData, width, height, Bitmap::CompressionFormat::kTOPIX);
				break;
			}
			
			default:
			{
				throw UnimplementedException("Graphics format unsupported (" + formatStr + ")");
			}
		}
		
		result.setPosition(PointU(xPos * dpmm / 10, yPos * dpmm / 10));
		return result;
	}
	
	LemonGFX::Bitmap parseTPCL(std::istream &input, const StripedHorse::PointU &size, size_t dpmm)
	{
		LemonGFX::Bitmap result(size);
		
		while (input)
		{
			char ch;
			if (!input.get(ch))
				break;
			
			if (ch != '{')
			{
				input.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
				continue;
			}
			
			char buf[3] = {0};
			if (!input.read(buf, 2) || input.gcount() != 2)
				throw StripedHorse::IOException("Unexpected end of file");
			
			if (strcmp(buf, "XS") == 0)
			{
				input.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
				break;
			}
			else if (strcmp(buf, "SG") == 0)
			{
				LemonGFX::Bitmap b = parseSG(input, dpmm);
				result.combine(b);
			}
			else
			{
				input.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
			}
		}
		
		return result;
	}
}

void StripedHorse::TECTool::Commands::Render(const ArgumentList &args)
{
	std::string inFilename;
	std::string outFilename;
	size_t dpmm = 0;

	for (size_t i = 0; i < args.size(); i++)
	{
		if (args[i] == "-o")
		{
			if (i == args.size() - 1)
				throw InvalidArgumentException("Expected filename after '-o'");
			
			if (!outFilename.empty())
				throw InvalidArgumentException("There can be only one output filename");
			
			i++;
			outFilename = args[i];
			continue;
		}
		else if (args[i] == "-r")
		{
			if (i == args.size() - 1)
				throw InvalidArgumentException("Expected resolution after '-r'");
			
			if (dpmm != 0)
				throw InvalidArgumentException("Multiple resolution arguments specified");
			
			i++;
			try
			{
				size_t end = 0;
				size_t val = std::stoul(args[i], &end, 10);
				
				if (end == args[i].size() || args[i].substr(end) == "dpi")
				{
					switch (val)
					{
						case 200:
						case 203:
							dpmm = 8;
							break;
						
						case 300:
						case 305:
							dpmm = 12;
							break;
						
						case 600:
						case 609:
							dpmm = 24;
							break;
						
						default:
							throw std::invalid_argument("Unsupported resolution");
					}
				}
				else if (args[i].substr(end) == "dpmm")
				{
					if (dpmm == 8 || dpmm == 12 || dpmm == 24)
						dpmm = val;
					else
						throw std::invalid_argument("Unsupported resolution");
				}
				else
				{
					throw std::invalid_argument("Unknown unit of measurement");
				}
			}
			catch (std::invalid_argument &ex)
			{
				throw InvalidArgumentException(args[i], "Invalid resolution");
			}
		}
		else
		{
			if (!inFilename.empty())
				throw InvalidArgumentException("Multiple input files specified");
			
			inFilename = args[i];
		}
	}
	
	if (dpmm == 0)
		dpmm = 8;
		
	LemonGFX::Bitmap b;
	std::ifstream infstream;
	std::ofstream outfstream;
	std::istream *instream = nullptr;
	std::ostream *outstream = nullptr;
	if (inFilename.empty())
	{
		instream = &std::cin;
	}
	else
	{
		infstream.open(inFilename, std::ios_base::binary | std::ios_base::in);
		if (!infstream)
			throw IOException("Unable to read from input file");
		instream = &infstream;
	}
	
	if (outFilename.empty())
	{
		outstream = &std::cout;
	}
	else
	{
		outfstream.open(outFilename, std::ios_base::binary | std::ios_base::out);
		if (!outfstream)
			throw IOException("Unable to read from output file");
		outstream = &outfstream;
	}
	
	PointU size = parseForBitmapSize(*instream, dpmm);
	b = parseTPCL(*instream, size, dpmm);

	b.write(LemonGFX::Image::Format::kTIFF, *outstream);
}
