#include <fstream>
#include <regex>
#include <libtecfilter/string-escape.h>
#include "cmd-setinsertion.h"

namespace
{
	void filter(std::istream &input, std::ostream &output, const std::regex &lookFor, const std::string &replacement)
	{
		bool replaced = false;
		std::string line;
		while (std::getline(input, line))
		{
			if (std::regex_match(line, lookFor))
			{
				output << replacement << '\n';
				replaced = true;
			}
			else
			{
				output << line << '\n';
			}
		}
		
		if (!replaced)
			throw StripedHorse::TECTool::Commands::InsertionNotSetException("could not find existing insertion");
	}
	
	std::string generateInsertionTooLongMessage(size_t len, size_t maxLen)
	{
		std::ostringstream msg;
		msg << "Insertion is too long (length is " << len << ", maximum length is " << maxLen << ").";
		return msg.str();
	}
	
	void setInsertion(StripedHorse::TECTool::Printer &printer, const std::string &name, const std::string &type, const std::string &tmpl)
	{
		if (!printer.isCUPS())
			throw StripedHorse::TECTool::InvalidArgumentException("Printer is not a CUPS printer");
		
		std::string ppdType;
		if (type == "string")
			ppdType = "String";
		else if (type == "file")
			ppdType = "File";
		
		if (ppdType.empty())
			throw StripedHorse::TECTool::InvalidArgumentException("Type must be \"file\" or \"string\"");	
		
		// perform round-trip to normalise escapes
		std::string escapedTmpl = StripedHorse::EscapeString(StripedHorse::UnescapeString(tmpl));
		std::regex insertionRegex("^\\*TEC" + name + "Insertion\\s+(String|File):.+$");
		std::string replacement = "*TEC" + name + "Insertion " + ppdType + ": \"" + escapedTmpl + "\"";
		
		if (replacement.size() >= 255)
			throw StripedHorse::TECTool::Commands::InsertionTooLongException(replacement.size(), 255);
		
		std::string ppdFilename = "/etc/cups/ppd/" + printer.getName() + ".ppd";
		std::ostringstream result;
		
		{
			std::ifstream ppdFile(ppdFilename);
			
			if (!ppdFile)
				throw StripedHorse::IOException("Unable to read from " + ppdFilename);
		
			filter(ppdFile, result, insertionRegex, replacement);
		}
		
		{
			std::ofstream ppdFile(ppdFilename);
			
			if (!ppdFile)
				throw StripedHorse::IOException("Unable to open " + ppdFilename + " for writing");
			
			ppdFile << result.str();
			
			if (!ppdFile)
				throw StripedHorse::IOException("Unable to write to " + ppdFilename + ". WARNING: Re-install printer to avoid using a possibly corrupt PPD");
		}
		
		std::cout << escapedTmpl << std::endl;
	}
}

void StripedHorse::TECTool::Commands::SetPreJobInsertion(Printer &p, const ArgumentList &args)
{
	if (args.size() != 2)
		throw InvalidNumberOfArgumentsException(args.size(), 2, 0);
	
	setInsertion(p, "PreJob", args[0], args[1]);
}

void StripedHorse::TECTool::Commands::SetPostJobInsertion(Printer &p, const ArgumentList &args)
{
	if (args.size() != 2)
		throw InvalidNumberOfArgumentsException(args.size(), 2, 0);
	
	setInsertion(p, "PostJob", args[0], args[1]);
}

void StripedHorse::TECTool::Commands::SetPreLabelInsertion(Printer &p, const ArgumentList &args)
{
	if (args.size() != 2)
		throw InvalidNumberOfArgumentsException(args.size(), 2, 0);
	
	setInsertion(p, "PreLabel", args[0], args[1]);
}

void StripedHorse::TECTool::Commands::SetPostLabelInsertion(Printer &p, const ArgumentList &args)
{
	if (args.size() != 2)
		throw InvalidNumberOfArgumentsException(args.size(), 2, 0);
	
	setInsertion(p, "PostLabel", args[0], args[1]);
}

StripedHorse::TECTool::Commands::InsertionTooLongException::InsertionTooLongException(size_t len, size_t maxLen)
  : Exception(generateInsertionTooLongMessage(len, maxLen))
{
}

StripedHorse::TECTool::Commands::InsertionNotSetException::InsertionNotSetException(const std::string &msg)
  : Exception("Insertion could not be set: " + msg)
{
}
