#include <cstdint>
#include <sstream>
#include "cmd-setipv4.h"

namespace
{
	class IPV4AddressParseException : public StripedHorse::Exception
	{
	public:
		IPV4AddressParseException() : Exception("Invalid IP address format") {}
	};
	
	class IPV4Address
	{
	public:
		IPV4Address(const std::string &ip)
		  : m_components{}
		{
			char dot[3];
			unsigned int cmp[4];
			std::istringstream ss(ip);
			
			if (!(ss >> cmp[0] >> dot[0] >> cmp[1] >> dot[1] >> cmp[2] >> dot[2] >> cmp[3]))
				throw IPV4AddressParseException();
			
			for (int i = 0; i < 4; i++)
			{
				if (cmp[i] > 255)
					throw IPV4AddressParseException();
				else
					m_components[i] = cmp[i];
			}
			
			if (dot[0] != '.' || dot[1] != '.' || dot[2] != '.')
				throw IPV4AddressParseException();
		}
		
		IPV4Address(uint32_t mask)
		  : m_components{}
		{
			if (mask == 0)
				return;
			
			if (mask > 32)
				throw IPV4AddressParseException();
			
			for (uint32_t i = 0; i < mask; i++)
				m_components[i / 8] |= 0x80U >> (i % 8);
		}
		
		unsigned int &operator[](int index)
		{
			return m_components[index];
		}
		
		const unsigned int &operator[](int index) const
		{
			return m_components[index];
		}
				
	private:
		unsigned int m_components[4];
	};
}

void StripedHorse::TECTool::Commands::SetIPV4(Printer &printer, const std::vector<std::string> &args)
{
	if (args.size() < 2 || args.size() > 3)
		throw InvalidNumberOfArgumentsException(args.size(), 2, 1);
	
	std::string::size_type slash = args[0].find('/');
	if (slash == std::string::npos)
		throw InvalidArgumentException("ip", args[0]);
	
	std::string ipPart = args[0].substr(0, slash);
	uint32_t nmPart;
	
	try
	{
		nmPart = std::stoi(args[0].substr(slash + 1));
	}
	catch (std::exception &)
	{
		throw InvalidArgumentException("ip", args[0]);
	}
	
	try
	{
		IPV4Address ip(ipPart);
		printer.writeCommand("IP;2,aaa,bbb,ccc,ddd", ip[0], ip[1], ip[2], ip[3]);
	}
	catch (IPV4AddressParseException &)
	{
		throw InvalidArgumentException("ip", args[0]);
	}
	
	try
	{
		IPV4Address nm(nmPart);
		printer.writeCommand("IP;3,aaa,bbb,ccc,ddd", nm[0], nm[1], nm[2], nm[3]);
	}
	catch (IPV4AddressParseException &)
	{
		throw InvalidArgumentException("ip", args[0]);
	}
	
	try
	{
		IPV4Address gw(args[1]);
		printer.writeCommand("IP;4,aaa,bbb,ccc,ddd", gw[0], gw[1], gw[2], gw[3]);
	}
	catch (IPV4AddressParseException &)
	{
		throw InvalidArgumentException("gw", args[1]);
	}
	
	if (args.size() == 3)
	{
		uint32_t port = 0;
		
		try
		{
			port = std::stoi(args[2]);
		}
		catch (std::exception &)
		{
			// ignore
		}
		
		if (port == 0 || port > 65535)
			printer.writeCommand("IS;a,bbbbb", 0, 0);
		else
			printer.writeCommand("IS;a,bbbbb", 1, port);
	}
}
