#include <iostream>
#include <string>
#include <unordered_map>
#include <vector>

#include <libtecfilter/version.h>

#include "tectool-command.h"

#include "cmd-feedlabel.h"
#include "cmd-getstatus.h"
#include "cmd-render.h"
#include "cmd-setfineadjustments.h"
#include "cmd-setinsertion.h"
#include "cmd-setipv4.h"
#include "cmd-setport.h"
#include "cmd-testlabel.h"
#include "cmd-tpcl.h"

namespace
{
	std::string g_progName;

#ifndef _WIN32
	void printUsage(std::ostream &out)
	{
		out << "Usage:\n\n";
		out << "  " << g_progName << " [ -p printer ] command [ arguments ]\n\n";
		out << "Refer to the man page for more information\n\n";
		out << "  man tectool\n";
	}
#else
	void printUsage(std::ostream &out)
	{
		out << "Usage:\n\n";
		out << "  " << g_progName << " render [ arguments ]\n\n";
	}
#endif
	
	std::unordered_map<std::string, StripedHorse::TECTool::Command> g_commands =
	{
#ifndef _WIN32
		{ "feedlabel", StripedHorse::TECTool::Commands::FeedLabel },
		{ "getstatus", StripedHorse::TECTool::Commands::GetStatus },
		{ "rawtpcl", StripedHorse::TECTool::Commands::RawTPCL },
		{ "setfineadjustments", StripedHorse::TECTool::Commands::SetFineAdjustments },
		{ "setipv4", StripedHorse::TECTool::Commands::SetIPV4 },
		{ "setport", StripedHorse::TECTool::Commands::SetPort },
		{ "setpostjob", StripedHorse::TECTool::Commands::SetPostJobInsertion },
		{ "setprejob", StripedHorse::TECTool::Commands::SetPreJobInsertion },
		{ "setpostlabel", StripedHorse::TECTool::Commands::SetPostLabelInsertion },
		{ "setprelabel", StripedHorse::TECTool::Commands::SetPreLabelInsertion },
		{ "testlabel", StripedHorse::TECTool::Commands::TestLabel },
		{ "tpcl", StripedHorse::TECTool::Commands::EscapedTPCL }
#endif
	};
	
	namespace ExitStatus
	{
		enum
		{
			Cromulent = 0,
			ProgramError = 1,
			PrinterError = 2
		};
	}
}

int main(int argc, char *argv[])
{
	using namespace StripedHorse;
	using namespace StripedHorse::TECTool;
	
	ArgumentList sargv;
	for (int i = 0; i < argc; i++)
		sargv.push_back(argv[i]);
	
	if (sargv.size() == 0)
	{
		std::cerr << "Invalid invocation\n";
		return ExitStatus::ProgramError;
	}
	else if (sargv.size() == 1)
	{
		// Since it's an error to invoke this program without arguments, usage
		// is printed to stderr.
		g_progName = sargv[0];
		printUsage(std::cerr);
		return ExitStatus::ProgramError;
	}
	else if (sargv.size() == 2 && sargv[1] == "--version")
	{
		std::cerr << StripedHorse::g_version << std::endl;
		return ExitStatus::Cromulent;
	}
	else
	{
		g_progName = sargv[0];
		sargv.erase(sargv.begin());
		
		if (sargv[0] == "--help" || sargv[0] == "-h" || sargv[0] == "-?")
		{
			// Since the user explicitly asked for help, it should go to stdout
			printUsage(std::cout);
			return ExitStatus::Cromulent;
		}
		else if (sargv[0] == "render")
		{
			// render is a special command that doesn't need a printer
			sargv.erase(sargv.begin());
			
			try
			{
				Commands::Render(sargv);
			}
			catch (Exception &ex)
			{
				std::cerr << "An unexpected error has occurred: " << std::endl;
				std::cerr << ex.what() << std::endl;
				return ExitStatus::ProgramError;
			}
			
			return ExitStatus::Cromulent;
		}
	}

#ifndef _WIN32
	bool networkPrinter = false;
	std::string printerName = "";
	
	if (sargv.size() && (sargv[0] == "-p" || sargv[0] == "-n"))
	{
		networkPrinter = (sargv[0] == "-n");
		if (sargv.size() == 1)
		{
			std::cerr << "Missing argument after '" << sargv[0] << "'" << std::endl;
			return ExitStatus::ProgramError;
		}
		
		printerName = sargv[1];
		sargv.erase(sargv.begin(), sargv.begin() + 2);
	}
	
	if (sargv.size() == 0)
	{
		std::cerr << "Missing command name" << std::endl;
		return ExitStatus::ProgramError;
	}
	
	try
	{
		Printer p = networkPrinter ? Printer::FromNetworkName(printerName) : Printer::FromName(printerName);
		
		std::cerr << "Printer: " << p.getName();
		if (p.isCUPS())
			std::cerr << " (from CUPS)";
		std::cerr << std::endl;
		
		ArgumentList args(sargv.begin() + 1, sargv.end());
		
		std::string commandName = sargv[0];
		
		auto commandIter = g_commands.find(commandName);
		if (commandIter == g_commands.end())
		{
			std::cerr << "Unknown command: " << commandName << std::endl;
			return ExitStatus::ProgramError;
		}
		
		commandIter->second(p, args);
	}
	catch (StripedHorse::TECTool::ResponseException &rex)
	{
		std::cerr << "A response from the printer was unexpected or invalid: " << std::endl;
		std::cerr << rex.what() << std::endl;
		return ExitStatus::PrinterError;
	}
	catch (StripedHorse::Exception &ex)
	{
		std::cerr << "An unexpected error has occurred: " << std::endl;
		std::cerr << ex.what() << std::endl;
		return ExitStatus::ProgramError;
	}
#else
	// In Windows if we get to this point, it's an error (the only supported
	// command is "render")
	return ExitStatus::ProgramError;
#endif
}
