#ifndef _WIN32

#include <cups/cups.h>
#include <cups/ipp.h>
#include <functional>
#include <iostream>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <vector>

#include <libtecfilter/network.h>
#include "printer.h"

namespace
{
	void checkedWrite(int fd, const void *buf, size_t len)
	{
		if (len == 0)
			return;
		
		if (fd < 0)
			throw StripedHorse::IOException("Invalid file descriptor");
		
		ssize_t bytesWritten = write(fd, buf, len);
		if (bytesWritten < 0)
			throw StripedHorse::IOException("Unable to write to device");
		
		if (static_cast<size_t>(bytesWritten) != len)
			throw StripedHorse::IOException("Unable to write all data to device");
	}
	
	class ippDeleter
	{
	public:
		void operator()(ipp_t *i)
		{
			if (i != nullptr)
				ippDelete(i);
		}
	};
	
	typedef std::unique_ptr<ipp_t, ippDeleter> IPPPtr;
	
	struct CUPSPrinter
	{
		CUPSPrinter(const std::string &name, const StripedHorse::URI &uri) : name(name), uri(uri) {}
		
		std::string name;
		StripedHorse::URI uri;
	};
	
	std::vector<CUPSPrinter> getCUPSPrinters()
	{
		static const char *pattrs[] = { "printer-name", "device-uri" };
		ipp_t *request = ippNewRequest(CUPS_GET_PRINTERS);
		
		ippAddStrings(request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD, "requested-attributes", sizeof(pattrs) / sizeof(pattrs[0]), nullptr, pattrs);
		ippAddString(request, IPP_TAG_OPERATION, IPP_TAG_NAME, "requesting-user-name", nullptr, cupsUser());
		
		IPPPtr response(cupsDoRequest(nullptr, request, "/"));
		
		std::vector<CUPSPrinter> printers;
		
		if (!response)
			return printers;
		
#if CUPS_VERSION_MAJOR > 1 || (CUPS_VERSION_MAJOR == 1 && CUPS_VERSION_MINOR >= 6)
		for (ipp_attribute_t *attr = ippFirstAttribute(response.get()); attr != nullptr; attr = ippNextAttribute(response.get()))
		{
			if (ippGetGroupTag(attr) != IPP_TAG_PRINTER)
				continue;
			
			std::string name;
			std::string uri;
			
			while (attr && ippGetGroupTag(attr) == IPP_TAG_PRINTER)
			{
				if (ippGetName(attr) == std::string("printer-name") && ippGetValueTag(attr) == IPP_TAG_NAME)
					name = ippGetString(attr, 0, nullptr);
				else if (ippGetName(attr) == std::string("device-uri") && ippGetValueTag(attr) == IPP_TAG_URI)
					uri = ippGetString(attr, 0, nullptr);
				
				if (!name.empty() && !uri.empty())
				{
					printers.emplace_back(name, StripedHorse::URI(uri, 9100));
					break;
				}
				
				attr = ippNextAttribute(response.get());
			}
#else
		for (ipp_attribute_t *attr = response->attrs; attr != nullptr; attr = attr->next)
		{
			if (attr->group_tag != IPP_TAG_PRINTER)
				continue;
			
			std::string name;
			std::string uri;
			
			while (attr && attr->group_tag == IPP_TAG_PRINTER)
			{
				if (attr->name == std::string("printer-name") && attr->value_tag == IPP_TAG_NAME)
					name = attr->values[0].string.text;
				else if (attr->name == std::string("device-uri") && attr->value_tag == IPP_TAG_URI)
					uri = attr->values[0].string.text;
				
				if (!name.empty() && !uri.empty())
				{
					printers.emplace_back(name, StripedHorse::URI(uri, 9100));
					break;
				}
				
				attr = attr->next;
			}
#endif
			if (attr == nullptr)
				break;
		}
		
		return printers;
	}
}

StripedHorse::TECTool::Printer::Printer()
  : m_fd(-1), m_isCUPS(false), m_connectionAttempted(false), m_name("unknown"),
    m_connector(nullptr), m_disconnector(nullptr)
{
}

StripedHorse::TECTool::Printer::Printer(Printer &&p)
  : m_fd(p.m_fd), m_isCUPS(p.m_isCUPS),
  m_connectionAttempted(p.m_connectionAttempted), m_name(std::move(p.m_name)),
  m_connector(p.m_connector), m_disconnector(p.m_disconnector)
{
}

StripedHorse::TECTool::Printer::~Printer()
{
	if (m_disconnector)
		m_disconnector(m_fd);
}

StripedHorse::TECTool::Printer StripedHorse::TECTool::Printer::FromName(const std::string &name)
{
	if (!name.empty())
	{	
		if (name == "-")
		{
			Printer p;
			p.m_connector = [](){ return 1; };
			p.m_disconnector = [](int fd){ (void) fd; return; };
			p.m_name = "stdout";
			return p;
		}
		
		if (name[0] == '/')
		{
			Printer p;
			p.m_name = name;
			p.m_connector = [name]()
			{
				struct stat st;
				
				int fd = open(name.c_str(), O_RDWR);
				if (fd >= 0)
				{
					if (fstat(fd, &st) == 0 && S_ISCHR(st.st_mode))
					{
						return fd;
					}
				}
				close(fd);
				return -1;
			};
			
			p.m_disconnector = [](int fd)
			{
				if (fd >= 0)
					close(fd);
			};
			
			return p;
		}
	}
	
	std::vector<CUPSPrinter> allPrinters = getCUPSPrinters();
	for (auto cp : allPrinters)
	{
		if (cp.name == name || name.empty())
		{
			Printer p = Printer::FromURI(cp.uri);
			p.m_isCUPS = true;
			p.m_name = cp.name;
			return p;
		}
	}
	
	if (name.empty())
		throw PrinterException("No suitable printer found");
	else
		throw PrinterException("Unknown printer: " + name);
}

StripedHorse::TECTool::Printer StripedHorse::TECTool::Printer::FromNetworkName(const std::string &name)
{
	using namespace StripedHorse::Network;

	Address addr(name);
	if (!addr.isValid())
		throw PrinterException("Unable to resolve \"" + name + "\"");
	
	Printer p;
	
	p.m_name = name;
	p.m_connector = [addr]()
	{
		return addr.connect();
	};
	
	p.m_disconnector = [](int fd)
	{
		close(fd);
	};
	
	return p;
}

StripedHorse::TECTool::Printer StripedHorse::TECTool::Printer::FromURI(const URI &uri)
{
	if (uri.getScheme() != "socket")
		throw PrinterException("Unsupported printer connection type, only \"socket\" printers accepted");
	std::string node = uri.getHost() + ":" + std::to_string(uri.getPort());
	return StripedHorse::TECTool::Printer::FromNetworkName(node);
}

void StripedHorse::TECTool::Printer::checkConnection()
{
	if (m_fd == -1)
	{
		if (m_connectionAttempted)
		{
			throw IOException("Unable to connect to device");
		}
		else if (m_connector)
		{
			m_fd = m_connector();
			m_connectionAttempted = true;
		}
		else
		{
			throw IOException("No method to connect to device");
		}
	}
	
	// If it's STILL -1, we can't connect
	if (m_fd == -1)
	{
		throw IOException("Unable to connect to device");
	}
}

const std::string &StripedHorse::TECTool::Printer::getName() const
{
	return m_name;
}

bool StripedHorse::TECTool::Printer::isCUPS() const
{
	return m_isCUPS;
}

void StripedHorse::TECTool::Printer::read(Buffer &buf, std::chrono::milliseconds timeout)
{
	checkConnection();
	
	size_t bytesRead = 0;
	fd_set rfds;
	FD_ZERO(&rfds);
	FD_SET(m_fd, &rfds);
	
	struct timeval tv;
	
	tv.tv_sec = std::chrono::duration_cast<std::chrono::seconds>(timeout).count();
	tv.tv_usec = std::chrono::duration_cast<std::chrono::microseconds>(timeout).count() % 1000000;
	
	while (bytesRead < buf.size())
	{
		int result = select(m_fd + 1, &rfds, NULL, NULL, &tv);
		
		if (result < 0)
			throw IOException("Unable to wait for data from printer");
		
		if (result == 0)
			throw IOException("Timed out waiting for data from printer");
		
		ssize_t bytesReadThisTime = ::read(m_fd, &buf[bytesRead], buf.size() - bytesRead);
		
		if (bytesReadThisTime <= 0)
			throw IOException("Unable to read any data from printer");
		
		bytesRead += static_cast<size_t>(bytesReadThisTime);
	}
}

void StripedHorse::TECTool::Printer::write(const std::string &str)
{
	checkConnection();
	checkedWrite(m_fd, &str[0], str.size());
}

void StripedHorse::TECTool::Printer::write(const Buffer &b)
{
	checkConnection();
	checkedWrite(m_fd, &b[0], b.size());
}

void StripedHorse::TECTool::Printer::write(std::istream &in)
{
	checkConnection();
	std::vector<char> buffer(4096);
	
	while (in)
	{
		in.read(&buffer[0], buffer.size());
		std::streamsize bytesRead = in.gcount();
		
		if (bytesRead > 0)
			checkedWrite(m_fd, &buffer[0], bytesRead);
		else
			break;
	}
	
	if (!in.eof())
		throw IOException("Unexpected read error before end of input");
}

#else

#include <libtecfilter/network.h>
#include "printer.h"

namespace
{
	[[noreturn]] void notSupported()
	{
		throw StripedHorse::IOException("This operating system does not support this operation");
	}
}

StripedHorse::TECTool::Printer StripedHorse::TECTool::Printer::FromNetworkName(const std::string &name)
{
	(void) name;
	notSupported();
}

StripedHorse::TECTool::Printer StripedHorse::TECTool::Printer::FromName(const std::string &name)
{
	(void) name;
	notSupported();
}

StripedHorse::TECTool::Printer::Printer()
{
	notSupported();
}

StripedHorse::TECTool::Printer::Printer(Printer &&p)
{
	(void) p;
	notSupported();
}

StripedHorse::TECTool::Printer::~Printer()
{
}

const std::string &StripedHorse::TECTool::Printer::getName() const
{
	notSupported();
}

bool StripedHorse::TECTool::Printer::isCUPS() const
{
	notSupported();
}

void StripedHorse::TECTool::Printer::read(Buffer &buf, std::chrono::milliseconds timeout)
{
	(void) buf;
	(void) timeout;
	notSupported();
}

void StripedHorse::TECTool::Printer::write(const std::string &str)
{
	(void) str;
	notSupported();
}

void StripedHorse::TECTool::Printer::write(const Buffer &buffer)
{
	(void) buffer;
	notSupported();
}

void StripedHorse::TECTool::Printer::write(std::istream &is)
{
	(void) is;
	notSupported();
}

#endif