#ifndef STRIPEDHORSE_TECTOOL_PRINTER_H
#define STRIPEDHORSE_TECTOOL_PRINTER_H

#include <chrono>
#include <functional>
#include <sstream>

#include <libcore/exceptions.h>
#include <libcore/uri.h>
#include <libtecfilter/commands.h>

namespace StripedHorse
{
	namespace TECTool
	{
		/**
		 A class representing a destination printer
		 */
		class Printer
		{
		public:
			/**
			 Gets a Printer object based on a network name (e.g. hostname or
			 IP address)
			 @param[in] name the hostname or IP address of the printer
			 @result if the supplied name is able to be resolved to an address
			         the result is a Printer object that will connect to that
			         address, otherwise a PrinterException is thrown
			 */
			static Printer FromNetworkName(const std::string &name);
			
			/**
			 Gets a Printer object based on a filename or CUPS printer name
			 @param[in] the filename or name of the CUPS printer. Filenames
			            must be absolute to avoid confusion with CUPS printer
			            names.
			 @result if the supplied name is a valid printer device or the name
			         of a CUPS printer, then the result is a Printer object
			         that will connect to that printer, otherwise a
			         PrinterException is thrown.
			 */
			static Printer FromName(const std::string &name);
			
			~Printer();
			
			/**
			 Printer objects are not copyable
			 */
			Printer(const Printer &) = delete;
			
			/**
			 Moves a printer object
			 */
			Printer(Printer &&);
			
			/**
			 Gets the display name of the printer.
			 @return the result is the display name that can be used to help
			         the user identify this printer object. This may or may not
			         be the same name used to create this printer object.
			 */
			const std::string &getName() const;
			
			/**
			 Determines whether the printer is a CUPS printer (a printer listed
			 in CUPS.
			 @return true if the Printer object was created to reference a CUPS
			         printer.
			 */
			bool isCUPS() const;
			
			/**
			 Reads data from the printer
			 @param[in,out] buf a buffer set to the required size of the number
			                of bytes to read
			 @param[in] timeout the number of milliseconds to wait for data
			 
			 If the buffer was not completely filled, or the timeout elapses,
			 then an IOException is thrown
			 */
			void read(Buffer &buf, std::chrono::milliseconds timeout = std::chrono::milliseconds(5000));
			
			/**
			 Sends a string to the printer
			 @param[in] s the string to send
			 
			 If there was an error communicating the entire string to the
			 printer, an IOException is thrown.
			 */
			void write(const std::string &s);
			
			/**
			 Sends a buffer o the printer
			 @param[in] b the buffer to send
			 
			 If there was an error communicating the entire buffer to the
			 printer, an IO exception is thrown.
			 */
			void write(const Buffer &b);
			
			/**
			 Sends data to the printer by reading from an input stream
			 @param[in] is the input stream to read from
			 
			 This function reads from the input stream, sending data to the
			 printer as it is read, until the input stream reaches end-of-file
			 or some other error occurs.
			 */
			void write(std::istream &is);
			
			/**
			 Encapsulates libtec's TPCLCommands system so that it can be used
			 to send TPCL commands to this printer
			 @param[in] cmd the command (and argument format) to send to the
			            printer
			 @param[in] args the arguments to send
			 */
			template<typename ... Args>
			void writeCommand(const std::string &cmd, const Args & ... args)
			{
				std::ostringstream ss;
				ss.put('{');
				TPCLCommands::WriteCommand(ss, cmd, args...);
				ss.put('|');
				ss.put('}');
				ss.put('\n');
				write(ss.str());
			}
			
		private:
			static Printer FromURI(const URI &uri);
			Printer();
			
			void checkConnection();
			
			int m_fd;
			bool m_isCUPS;
			bool m_connectionAttempted;
			std::string m_name = "unknown";
			std::function<int(void)> m_connector;
			std::function<void(int)> m_disconnector;
		};
		
		class PrinterException : public Exception
		{
		public:
			PrinterException(const std::string &msg) : Exception(msg) {}
		};
	}
}

#endif

