#!/usr/bin/env python

import ConfigParser
import errno
import glob
import os
import platform
import re
import shutil
import subprocess
import sys
import tempfile

INPUT_FILE_ARG="{input}"
OPTIONS_ARG="{options}"
FAILED_JOBOPTIONS_DIR="failed-joboptions"

normalise_re = re.compile("\\W+")
def normalise(s):
    return normalise_re.sub("_", s.lower())


class TestConfigurationError(Exception):
    pass

class TestConfigurationRunnerError(Exception):
    pass

class ExpectationError(Exception):
    pass


class Expectation(object):
    """This class provides the basis for creating different types of expectations"""
    def __init__(self):
        self.success_value = True
        pass
    
    def is_met(self, input):
        """Determines if the expectation is met based on the input (which is
        assumed to be a seekable file-like object"""
        raise NotImplementedError("is_met() not implemented in subclass")
    
    def negate(self):
        """Reverses the result of the expectation"""
        self.success_value = not self.success_value


class FixedStringExpectation(Expectation):
    """This class represents the expectation that a fixed string will
    appear in the input"""
    def __init__(self, s):
        super(FixedStringExpectation, self).__init__()
        self.string = s.strip().strip(" \"")
    

    def is_met(self, input):
        for line in input:
            if line.strip() == self.string:
                return self.success_value
        return not self.success_value


class RegexExpectation(Expectation):
    """This class represents the expectation that a regex will
    match somewhere in the input"""
    def __init__(self, r):
        super(RegexExpectation, self).__init__()
        try:
            self.regex = re.compile(r.strip(" \""))
        except re.error:
            raise TestConfigurationError("Invalid regular expression: " + r)
    
    def is_met(self, input):
        for line in input:
            if self.regex.search(line.strip()) is not None:
                return self.success_value
        return not self.success_value


class Test(object):
    """Pairs the options to use and the expectations to meet"""
    def __init__(self, index, options, expectations, expectation_failure_message=None):
        self.index = index
        self.options = options
        self.options_string = " ".join([ "%s=%s" % (k, v) for k, v in sorted(self.options.iteritems()) ])
        self.expectations = expectations
        if expectation_failure_message is None:
            self.expectation_failure_message = "no further description"
        else:
            self.expectation_failure_message = expectation_failure_message


class PPD(object):
    """Does a simple parse on the PPD. We don't use the CUPS
    parser because CUPS isn't availeble on Windows"""
    def __init__(self, filename):
        self.filename = filename
        self.supported = True
        self.detected_type = None
        self.model = "Unknown"

        ppd_file = open(filename, "r+")
        for line in ppd_file:
            line = line.strip()
            if line.startswith("*TECBarcodePrinterLanguage: "):
                self.detected_type = line[28:].strip(" \"")
            elif line.startswith("*TECTarget: "):
                if line == "*TECTarget: \"unsupported\"":
                    self.supported = False
            elif line.startswith("*ModelName: "):
                self.model = line[12:].strip(" \"")
        ppd_file.close()


class TestConfiguration(object):
    """Encapsulates a test configuration, as in, a number of tests to run
    on a particular file with particular models"""
    def __init__(self, filename, ppds):
        c = ConfigParser.ConfigParser()
        c.optionxform=str
        c.read(filename)
        
        self.name = c.get("Test", "Name")
        self.source = c.get("Test", "Source")
        self.tests = []
        self.model_ppds = []

        for model in c.get("Test", "Models").split(","):
            model = model.strip()
            if model.startswith("$"):
                self.model_ppds.extend([ ppd for ppd in ppds if ppd.supported and ppd.detected_type == model[1:] ])
            else:
                self.model_ppds.extend([ ppd for ppd in ppds if model in ppd.model ])

        if "ExcludeModels" in c.options("Test"):
            exclusions = [ e.strip() for e in c.get("Test", "ExcludeModels").split(",") ]
            new_model_ppds = []
            for ppd in self.model_ppds:
                # workaround for Python 2.5 not supporting any()
                if True not in [ excl in ppd.model for excl in exclusions ]:
                    new_model_ppds.append(ppd)
            self.model_ppds = new_model_ppds
        
        if len(self.model_ppds) == 0:
            raise RuntimeError("Test configuration %s does not include any models" % (filename,))

        i = 1
        while True:
            options = {}
            expectations = []

            options_section = "Options.%d" % i
            expectation_section = "Expectation.%d" % i
            if not c.has_section(options_section):
                break
            
            if not c.has_section(expectation_section):
                raise TestConfigurationError("Missing %s" % expectation_section)
        
            for option in c.options(options_section):
                value = c.get(options_section, option)
                options[option] = value.strip()
            
            failure_message=None
            for expectation_type in c.options(expectation_section):
                expectation_value = c.get(expectation_section, expectation_type)

                if expectation_type == "FailureMessage":
                    failure_message = expectation_value.strip(" \"")
                    continue

                if expectation_type.endswith(".Regex"):
                    ex = RegexExpectation(expectation_value)
                else:
                    ex = FixedStringExpectation(expectation_value)
                
                ex.failure_message = failure_message
                if expectation_type.startswith("Not"):
                    ex.negate()
                
                expectations.append(ex)
            
            self.tests.append(Test(i, options, expectations, expectation_failure_message=failure_message))
            i = i + 1

class TestConfigurationRunner(object):
    def __init__(self, executable, *args):
        self.executable = os.path.expandvars(executable)
        self.args = [ self.executable ]
        self.args.extend(args)
        self.keep_failed_files = False

        try:
            self.input_file_arg = self.args.index(INPUT_FILE_ARG)
        except ValueError:
            self.input_file_arg = None # read from stdin
        
        try:
            self.options_arg = self.args.index(OPTIONS_ARG)
        except ValueError:
            raise TestConfigurationRunnerError("Missing options argument for %s" % self.executable)
    

    def _check_result(self, spool_file):
        for i, expectation in enumerate(self.current_test.expectations):
            spool_file.seek(0)
            if not expectation.is_met(spool_file):
                raise ExpectationError(i)
        return True

    
    def _print_exception(self):
        sys.stderr.write("FAIL (%s: %s)\n" % (sys.exc_info()[0].__name__, sys.exc_info()[1].args[0]))


    def _print_failed_expectation(self):
        # if there's only one expectation, no point specifying the index
        if len(self.current_test.expectations) == 1:
            sys.stderr.write("FAIL (expectation not met: %s)\n" % (self.current_test.expectation_failure_message))
        else:
            failed_index = sys.exc_info()[1].args[0]
            sys.stderr.write("FAIL (expectation #%d not met: %s)\n" % (failed_index + 1, self.current_test.expectation_failure_message))

    
    def _print_pass(self):
        sys.stderr.write("PASS\n")


    def _print_test_prolog(self):
        sys.stderr.write("[%02d]: " % (self.current_test.index))
    

    def _run_test(self, input_filename, env):
        output_filename = None
        stderr_filename = None
        input_file = None
        output_file = None
        stderr_file = None
        success = False
        
        try:
            (ofd, output_filename) = tempfile.mkstemp()
            (efd, stderr_filename) = tempfile.mkstemp()
            os.close(ofd)
            os.close(efd)

            output_file = open(output_filename, "w+")
            stderr_file = open(stderr_filename, "w")

            these_args = list(self.args)
            these_args[self.options_arg] = self.current_test.options_string

            if self.input_file_arg is not None:
                these_args[self.input_file_arg] = input_filename
            else:
                input_file = open(input_filename, "r")

            result = subprocess.call(these_args, stdout=output_file, stdin=input_file, stderr=stderr_file, env=env)
            if result != 0:
                raise TestConfigurationRunnerError("%s returned non-zero status" % os.path.basename(self.executable))
            success = self._check_result(output_file)
        finally:
            if output_file is not None:
                output_file.close()

            if stderr_file is not None:
                stderr_file.close()
            
            if output_filename is not None:
                if self.keep_failed_files and not success:
                    filename = normalise("%s_%s_%03d.prn" % (self.current_config.name, self.current_ppd.model, self.current_test.index))
                    shutil.copy(output_filename, os.path.join(FAILED_JOBOPTIONS_DIR, filename))
                    if stderr_filename is not None:
                    	shutil.copy(stderr_filename, os.path.join(FAILED_JOBOPTIONS_DIR, filename + ".stderr"))
                os.unlink(output_filename)
            if stderr_filename is not None:
                os.unlink(stderr_filename)
        return success


    def run(self, config):
        """config is a TestConfiguration and input is the file to process"""
        sys.stderr.write(config.name.upper() + "\n")
        sys.stderr.write("=" * len(config.name) + "\n")

        self.current_config = config
        
        all_passed = True

        environment = dict(os.environ)
        environment["SOFTWARE"] = "CUPS/1.4"

        input_filename = os.path.join("joboptions", config.source)

        for ppd in config.model_ppds:
            self.current_ppd = ppd
            environment["PPD"] = ppd.filename
            sys.stderr.write(ppd.model + ":\n")
            for test in config.tests:
                success = False
                self.current_test = test
                self._print_test_prolog()
                try:
                    success = self._run_test(input_filename, environment)
                    self._print_pass()
                except ExpectationError:
                    self._print_failed_expectation()
                except TestConfigurationRunnerError:
                    self._print_exception()
                if not success:
                    all_passed = False
            sys.stderr.write("\n")
        sys.stderr.write("\n")
        return all_passed


def main():
    # always delete this directory on startup
    try:
        shutil.rmtree(FAILED_JOBOPTIONS_DIR)
    except OSError:
        pass
    
    # if "clean" was provided as an argument, stop here (already done with cleaning)
    if len(sys.argv) > 1 and sys.argv[1] == "clean":
        return 0

    if "ARCH" not in os.environ:
        os.environ.setdefault("ARCH", "x86-64")
    
    uname = platform.system()
    if uname == "Windows":
        sys.stderr.write("This operating system is not supported by this utility (ignoring)\n")
        return 0
    elif uname == "Darwin":
        os.environ.setdefault("OS_PLATFORM", "macos")
    elif uname == "Linux":
        os.environ.setdefault("OS_PLATFORM", "linux")
    else:
        raise RuntimeError("Unable to test on this platform")

    config_files = sorted(glob.glob("joboptions/*.ini"))
    ppds = [ PPD(p) for p in sorted(glob.glob("../drv/ppd/*.ppd")) ]

    if len(ppds) == 0:
        raise RuntimeError("No PPDs to test!")
    
    try:
        test_configurations = [ TestConfiguration(cfg, ppds) for cfg in config_files ]
    except TestConfigurationError, e:
        sys.stderr.write("%s could not be loaded: %s\n" % (cfg, e.args[0]))
        return 1

    # currently only tecfilter is tested, but the TestConfigurationRunner class will eventually be
    # configured by another ini file.
    tecfilter = TestConfigurationRunner("../tecfilter/build.${OS_PLATFORM}.${ARCH}/tecfilter-static", "1", "testuser", "testdoc", "1", OPTIONS_ARG, INPUT_FILE_ARG)

    try:
        os.makedirs(FAILED_JOBOPTIONS_DIR)
    except OSError:
        if sys.exc_info()[1].errno != errno.EEXIST:
            return 
    tecfilter.keep_failed_files = True

    test_results = [ tecfilter.run(cfg) for cfg in test_configurations ]
    if False in test_results:
        return 1
    else:
        return 0

if __name__ == "__main__":
    sys.exit(main())
