#!/usr/bin/env python

import errno
import filecmp
import glob
import os
import platform
import shutil
import subprocess
import sys
import zipfile

if os.name == 'nt':
    EXEC_CHECK=os.R_OK
    EXEC_SUFFIX='.exe'
else:
    EXEC_CHECK=os.X_OK
    EXEC_SUFFIX=''

INPUT_FILE="{INPUT}"
OUTPUT_FILE="{OUTPUT}"

class MissingFilterError(Exception):
    pass

class BasicTransition(object):
    def __init__(self, name, input_form, output_form, filter_exec, *args):
        self.name = name
        self.input_dir = os.path.dirname(input_form)
        self.input_ext = os.path.splitext(input_form)[1]
        self.output_dir = os.path.dirname(output_form)
        self.output_ext = os.path.splitext(output_form)[1]
        self.filter = os.path.expandvars(filter_exec) + EXEC_SUFFIX
        self.args = [ self.filter ]
        self.args.extend(args)

        try:
            self.input_arg_index = self.args.index(INPUT_FILE)
            self.output_arg_index = self.args.index(OUTPUT_FILE)
        except:
            raise ValueError("Missing input or ouput file name for %s => %s transition" % (input_form, output_form))
    
    def _fail(self, testname, message):
        sys.stdout.write("%-30s: FAIL (%s)\n" % (testname, message))

    def _pass(self, testname, messages):
        errors = 0
        warnings = 0
        for msg in messages:
            if msg.startswith("Warning:"):
                warnings = warnings + 1
            elif msg.startswith("Error:"):
                errors = errors + 1

        sys.stdout.write("%-30s: PASS" % (testname))
        if warnings > 0 and errors > 0:
            sys.stdout.write(" (%d warnings, %d errors)" % (warnings, errors))
        elif warnings > 0:
            sys.stdout.write(" (%d warnings)" % (warnings))
        elif errors > 0:
            sys.stdout.write(" (%d errors)" % (errors))
        sys.stdout.write("\n")

    def clean(self):
        """Removes all output files from the outputdir"""
        try:
            shutil.rmtree(self.output_dir)
        except OSError, e:
            if e.errno != errno.ENOENT:
                raise

    def check(self):
        if not os.access(self.filter, EXEC_CHECK):
            raise MissingFilterError(os.path.basename(self.filter))

    def execute_all(self, testnames):
        passed = 0
        sys.stdout.write(self.name + "\n")
        sys.stdout.write("=" * len(self.name) + "\n")
        for testname in testnames:
            result, messages = self.execute(testname)
            if result:
                passed = passed + 1
                self._pass(testname, messages)
            else:
                self._fail(testname, messages)
        sys.stdout.write("\n\n")
        return passed

    def execute(self, testname):
        """Runs the transition using the corresponding test names"""
        try:
            os.mkdir(self.output_dir)
        except OSError:
            pass

        try:
            these_args = list(self.args)
            if not os.path.exists(self.get_input_filename(testname)):
                return (False, "input file missing")
            
            these_args[self.input_arg_index] = self.get_input_filename(testname)
            these_args[self.output_arg_index] = self.get_output_filename(testname)

            proc = subprocess.Popen(these_args, stdin=None, stdout=None, stderr=subprocess.PIPE, env=os.environ)
            messages = proc.stderr.readlines()
            result = proc.wait()
            if result == 0:
                return (True, messages)
            else:
                return (False, os.path.basename(self.filter) + (" returned %d" % result))
        except Exception, e:
            return (False, str(e))
    
    def get_input_filename(self, testname):
        return os.path.join(self.input_dir, testname + self.input_ext)
    
    def get_output_filename(self, testname):
        return os.path.join(self.output_dir, testname + self.output_ext)

class PDFDecodeTransition(BasicTransition):
    def execute(self, testname):
        result, messages = super(PDFDecodeTransition, self).execute(testname)
        if result and not filecmp.cmp(self.get_input_filename(testname), self.get_output_filename(testname)):
            return (False, "input and output are not equal")
        else:
            return result, messages

class XPSTransition(BasicTransition):
    def execute(self, testname):
        result, messages = super(XPSTransition, self).execute(testname)
        if result:
            try:
                xps_archive = zipfile.ZipFile(self.get_output_filename(testname))
                bad_file = xps_archive.testzip()
                if bad_file is not None:
                    return (False, "%s is not valid in zip" % bad_file)
            except Exception, e:
                return (False, "unable to process file as zip: " + str(e))
        return result, messages


def main():
    os.environ.setdefault("ARCH", "x86-64")
    uname = platform.system()
    if uname == "Windows":
        os.environ.setdefault("OS_PLATFORM", "win32")
        paths = (os.environ.get("PATH") or "").split(";")
        paths.insert(0, os.path.normpath(os.path.join(os.getcwd(), "..\\3rdparty\\bin")))
        os.environ["PATH"] = ";".join(paths)
        os.environ["ARCH"] = "i386" # x86-64 is not supported (due to no official 64-bit build of libxml)
    elif uname == "Darwin":
        os.environ.setdefault("OS_PLATFORM", "macos")
    elif uname == "Linux":
        os.environ.setdefault("OS_PLATFORM", "linux")
    else:
        raise RuntimeError("Unable to test on this system")
    
    testnames = [ os.path.splitext(os.path.basename(x))[0] for x in sorted(glob.glob("pdf/*.pdf")) ]

    if len(testnames) == 0:
        sys.stderr.write("No test files\n")
        sys.exit(1)

    try:
        transitions = [
            BasicTransition("pdfdecode1", "pdf/*.pdf", "decoded1/*.pdf", "../contrib/pdfdecode/build.${OS_PLATFORM}.${ARCH}/pdfdecode-static", INPUT_FILE, OUTPUT_FILE),
            PDFDecodeTransition("pdfdecode2", "decoded1/*.pdf", "decoded2/*.pdf", "../contrib/pdfdecode/build.${OS_PLATFORM}.${ARCH}/pdfdecode-static", INPUT_FILE, OUTPUT_FILE),
            XPSTransition("pdftoxps", "pdf/*.pdf", "xps/*.xps", "../pdftoxps/build.${OS_PLATFORM}.${ARCH}/pdftoxps-static", "-c", "-o", OUTPUT_FILE, INPUT_FILE),
            BasicTransition("xpstotpcl", "xps/*.xps", "tpcl/*.tpcl", "../xpstotpcl/build.${OS_PLATFORM}.${ARCH}/xpstotpcl-static", "-o", OUTPUT_FILE, "-d", "8", "-W", "1016", "-H", "1524", "-O", "DocumentType=Graphics", INPUT_FILE),
            BasicTransition("tectool", "tpcl/*.tpcl", "bmp/*.tiff", "../tectool/build.${OS_PLATFORM}.${ARCH}/tectool-static", "render", "-o", OUTPUT_FILE, "-r", "203", INPUT_FILE)
        ]
    except ValueError, e:
        sys.stderr.write(str(e))
        sys.exit(1)

    if len(sys.argv) > 1 and sys.argv[1] == "clean":
        for transition in transitions:
            transition.clean()
        return 0

    some_failed = False
    for transition in transitions:
        try:
            transition.check()
        except MissingFilterError, e:
            some_failed = True
            sys.stderr.write("build %s for %s first or set ARCH correctly\n" % (e.args[0], os.environ["ARCH"]))
    
    if some_failed:
        return 1
    else:
        some_failed = False

    results = {}
    for transition in transitions:
        passed_count = transition.execute_all(testnames)
        results[transition.name] = passed_count
        if passed_count != len(testnames):
            some_failed = True
    
    sys.stdout.write("\n")
    for transition in transitions:
        sys.stdout.write("%s: %d/%d PASS\n" % (transition.name, results[transition.name], len(testnames)))
    
    if some_failed:
        return 1
    else:
        return 0

if __name__ == "__main__":
    sys.exit(main())

