# This Makefile is designed to be included into another Makefile for the purpose of generating
# static and dynamic libraries. The Makefile that includes this file should declare at least
# two variables, SOURCES and EXEC_NAME, containing the source files and name of the library
# respectively. It may also choose to override CFLAGS and LDFLAGS etc.

OBJECTS=$(addprefix $(BUILDDIR)/, $(SOURCES:.cpp=.o)) 
EXEC=$(BUILDDIR)/$(EXEC_NAME)
STATIC_EXEC=$(BUILDDIR)/$(EXEC_NAME)-static

ARCH?=x86-64
BUILDDIR=build.$(ARCH)
LDARCH=$(ARCH)

ifeq ($(OSNAME),Darwin)
	override CFLAGS+=$(DARWIN_CFLAGS)
	LDFLAGSEXTRA=$(DARWIN_LDFLAGS)
	STATIC_LDFLAGS=
	BUILDDIR=build.macos.$(ARCH)
else ifeq ($(OSNAME),Linux)
	override CFLAGS+=$(LINUX_CFLAGS)
	LDFLAGSEXTRA=$(LINUX_LDFLAGS) -latomic
	STATIC_LDFLAGS=-static-libstdc++ -static-libgcc -Wl,-Bstatic -latomic -Wl,-Bdynamic
	BUILDDIR=build.linux.$(ARCH)

	ifeq ($(ARCH),arm)
		LDARCH=armv6
	endif
else
	$(error Unable to build on this platform)
endif
override LDFLAGS+=-march=$(LDARCH) $(LDFLAGSEXTRA)

# Handle all dependencies
ifneq ($(DEPENDS),)
LIB_DEPENDS=$(patsubst %,-l%,$(patsubst lib%,%,$(notdir $(DEPENDS))))
STATIC_LIB_DEPENDS=$(shell for X in $(DEPENDS); do Y=$$(basename $${X}); [ -n "$${X}" ] && echo $${X}/$(BUILDDIR)/lib$${Y\#lib}.a; done)
DYNAMIC_LDFLAGS=$(addprefix -L,$(addsuffix /$(BUILDDIR),$(DEPENDS)))
endif

# Handle any manpages
ifneq ($(MANPAGE_SOURCES),)
MANPAGE_OBJECTS=$(addprefix $(BUILDDIR)/, $(MANPAGE_SOURCES:.1=.1.gz)) 
endif

.PHONY: all
all: $(EXEC) $(STATIC_EXEC) $(MANPAGE_OBJECTS)

$(BUILDDIR):
	mkdir -p "$@"

.PHONY: install
install:
	@if [ -z "$(INSTALL_DIR)" ]; then echo "Must set INSTALL_DIR"; exit 1; fi
	install -s "$(STATIC_EXEC)" "$(INSTALL_DIR)"
	install -s "$(EXEC)" "$(INSTALL_DIR)"

.PHONY: clean
clean:
	-rm -f $(OBJECTS) $(EXEC) $(STATIC_EXEC)
	-rm -rf $(BUILDDIR)

$(EXEC): $(OBJECTS)
	$(CXX) $^ $(DYNAMIC_LDFLAGS) $(LIB_DEPENDS) $(LDFLAGS) $(LIBS) -o $@

$(STATIC_EXEC): $(OBJECTS) $(STATIC_LIB_DEPENDS)
	$(CXX) $^ $(STATIC_LDFLAGS) $(LDFLAGS) $(LIBS) -o $@

$(OBJECTS) $(MANPAGE_OBJECTS): | $(BUILDDIR)

$(BUILDDIR)/%.o: %.cpp
	$(CXX) $(CPPFLAGS) -c $(CFLAGS) $(CXXFLAGS) $< -o $@

$(BUILDDIR)/%.1.gz: %.1
	gzip -9 < "$<" > "$@"
	chmod 644 "$@"
