# This Makefile is designed to be included into another Makefile for the purpose of generating
# static and dynamic libraries. The Makefile that includes this file should declare at least
# two variables, LIB_SOURCES and LIB_NAME, containing the source files and name of the library
# respectively. It may also choose to override CFLAGS and LDFLAGS etc.

STATIC_LIB_EXEC=$(BUILDDIR)/lib$(LIB_NAME).a

ARCH?=x86-64
LDARCH=$(ARCH)

ifeq ($(OSNAME),Darwin)
	ifeq ($(ARCH),i686)
		LDARCH=i386
	else ifeq ($(ARCH),x86-64)
		LDARCH=x86_64
	endif

	SHARED_LIB_EXEC=$(BUILDDIR)/lib$(LIB_NAME).1.0.dylib
	SHARED_LIB_NAME=$(BUILDDIR)/lib$(LIB_NAME).1.dylib
	LINKER_LIB_NAME=$(BUILDDIR)/lib$(LIB_NAME).dylib
	LDFLAGSEXTRA=-arch $(LDARCH) -framework CoreFoundation -framework CoreGraphics $(DARWIN_LDFLAGS)
	BUILDDIR=build.macos.$(ARCH)
	override CFLAGS+=$(DARWIN_CFLAGS)
else ifeq ($(OSNAME),Linux)
	ifeq ($(ARCH),arm)
		LDARCH=armv6
	endif

	SONAME=lib$(LIB_NAME).so
	SHARED_LIB_EXEC=$(BUILDDIR)/$(SONAME).1.0
	SHARED_LIB_NAME=$(BUILDDIR)/$(SONAME).1
	LINKER_LIB_NAME=$(BUILDDIR)/$(SONAME)
	LDFLAGSEXTRA=-Wl,-soname,$(SONAME).1 -ljpeg
	BUILDDIR=build.linux.$(ARCH)
	override CFLAGS+=$(LINUX_CFLAGS)
endif

# default flags
override CFLAGS+=-fPIC
override LDFLAGS+=-march=$(LDARCH) -lstdc++ $(LDFLAGSEXTRA)

# Handle all dependencies
ifneq ($(DEPENDS),)
LIB_DEPENDS=$(patsubst %,-l%,$(patsubst lib%,%,$(notdir $(DEPENDS))))
override LDFLAGS+=$(addprefix -L,$(addsuffix /$(BUILDDIR),$(DEPENDS)))
endif

LIB_OBJECTS=$(addprefix $(BUILDDIR)/, $(LIB_SOURCES:.cpp=.o))

.PHONY: all
all: $(BUILDDIR) $(SHARED_LIB_EXEC) $(STATIC_LIB_EXEC) $(LINKER_LIB_NAME)

$(BUILDDIR):
	mkdir -p "$@"

$(STATIC_LIB_EXEC): $(LIB_OBJECTS)
	$(AR) $(ARFLAGS) $@ $^

$(SHARED_LIB_EXEC): LDFLAGS+=-shared
$(SHARED_LIB_EXEC): $(LIB_OBJECTS)
	$(CXX) $(LDFLAGS) $(LDFLAGSEXTRA) $(LIB_DEPENDS) -o $(SHARED_LIB_EXEC) $^
	$(STRIP) -x $(SHARED_LIB_EXEC)

$(LIBOBJECTS): | $(BUILDDIR)

$(LINKER_LIB_NAME): $(SHARED_LIB_EXEC)
	cd "$(BUILDDIR)" && ln -sf "$(shell basename $(SHARED_LIB_EXEC))" "$(shell basename $(SHARED_LIB_NAME))"
	cd "$(BUILDDIR)" && ln -sf "$(shell basename $(SHARED_LIB_NAME))" "$(shell basename $(LINKER_LIB_NAME))"

$(BUILDDIR)/%.o: %.cpp
	$(CXX) $(CPPFLAGS) -c $(CFLAGS) $(CXXFLAGS) $< -o $@

install: $(SHARED_LIB_EXEC) $(STATIC_LIB_EXEC)
	@if [ -z "$(INSTALL_DIR)" ]; then echo "Must set INSTALL_DIR"; exit 1; fi
	install $(SHARED_LIB_EXEC) $(INSTALL_DIR)
	cp -PR $(LINKER_LIB_NAME) $(INSTALL_DIR)
	cp -PR $(SHARED_LIB_NAME) $(INSTALL_DIR)
	cp -PR $(STATIC_LIB_EXEC) $(INSTALL_DIR)

.PHONY: clean
clean:
	-rm -f $(LIB_OBJECTS) $(SHARED_LIB_EXEC) $(STATIC_LIB_EXEC)
	-rm -rf $(BUILDDIR)
