#include <cmath>
#include <iterator>
#include <sstream>
#include <unordered_map>
#include <vector>
#include <libcore/exceptions.h>
#include "bitmapfont.h"
#include "font-utilities.h"

namespace
{
	std::vector<StripedHorse::TPCLBitmapFont> g_bitmapFontTable {
		{ "Times Roman",    "normal", 12.0, "A"     },
		{ "Times Roman",    "normal", 15.0, "B"     },
		{ "Times Roman",    "bold",   15.0, "C"     },
		{ "Times Roman",    "bold",   18.0, "D"     },
		{ "Times Roman",    "bold",   21.0, "E"     },
		{ "Times Roman",    "italic", 18.0, "F"     },
		{ "Helvetica",      "normal",  9.0, "G", -2 }, // font is ridiculously wide
		{ "Helvetica",      "normal", 15.0, "H"     },
		{ "Helvetica",      "normal", 18.0, "I"     },
		{ "Helvetica",      "bold",   18.0, "J"     },
		{ "Helvetica",      "bold",   21.0, "K"     },
		{ "Helvetica",      "italic", 18.0, "L"     },
		{ "Presentation",   "bold",   27.0, "M"     },
		{ "Letter Gothic",  "normal", 14.3, "N"     },
		{ "Prestige Elite", "normal", 10.5, "O", -2 },
		{ "Prestige Elite", "bold",   15.0, "P"     },
		{ "Courier",        "normal", 15.0, "Q"     },
		{ "Courier",        "bold",   18.0, "R"     },
		{ "Courier",        "italic", 15.0, "Q"     }, // not italic but oh well
		{ "OCR-A",          "normal", 12.0, "S"     },
		{ "OCR-B",          "normal", 12.0, "T"     }
	};

	std::unordered_map<std::string, std::string> g_genericFontMapping {
		{ "fixed-width", "Prestige Elite" },
		{ "serif",       "Times Roman" },
		{ "sans-serif",  "Helvetica" }
	};

	std::unordered_map<std::string, std::string> g_specificFontMapping {
		{ "Arial",             "Helvetiva"      },
		{ "ArialPSMT",         "Helvetiva"      },
		{ "CourierNew",        "Courier"        },
		{ "CourierNewPSMT",    "Courier"        },
		{ "LiberationSerif",   "Times Roman"    },
		{ "NimbusMonL-Regu",   "Prestige Elite" },
		{ "NimbusSanL-Regu",   "Helvetica"      },
		{ "TimesNewRoman",     "Times Roman"    },
		{ "TimesNewRomanPSMT", "Times Roman"    },
	};

	struct TPCLScaledBitmapFont
	{
		StripedHorse::TPCLBitmapFont bitmapFont;
		uint32_t scale; // fixed point with 1 dp
	};

	TPCLScaledBitmapFont findBestFontAndScale(const std::vector<StripedHorse::TPCLBitmapFont> &availableFonts, double fontSize, double dpmm)
	{
		const double targetFontSize = fontSize * dpmm / 8.0;
		if (availableFonts.size() == 0)
			throw StripedHorse::InvalidArgumentException("No fonts to choose from");

		std::vector<StripedHorse::TPCLBitmapFont> sorted = availableFonts;
		std::sort(sorted.begin(), sorted.end(), [](const StripedHorse::TPCLBitmapFont &left, const StripedHorse::TPCLBitmapFont &right){
			return left.pointSizeAt8dpmm > right.pointSizeAt8dpmm;
		});

		std::vector<std::pair<double, uint32_t>> distances(sorted.size());
		size_t smallest = 0;
		for (size_t i = 0; i < sorted.size(); i++)
		{
			if (sorted[i].scales.size() == 0)
				throw StripedHorse::InternalInconsistencyException("Font scale table has not been established");

			std::map<double, uint32_t>::const_iterator it = sorted[i].scales.lower_bound(targetFontSize);
			if (it != sorted[i].scales.begin())
				it--;
			
			double distance = std::abs(targetFontSize - it->first);
			distances[i] = *it;
			if (distance < distances[smallest].first)
				smallest = i;
		}

		return { sorted[smallest], distances[smallest].second };
	}
}

StripedHorse::TPCLBitmapFont::TPCLBitmapFont(const std::string &name, const std::string &style, double pointSizeAt8dpmm, const std::string &code, int32_t extraDeviceSpacing)
  : name(name), style(style), pointSizeAt8dpmm(pointSizeAt8dpmm), code(code), valid(true),
    extraDeviceSpacing(extraDeviceSpacing)
{
	// valid scales are 5, 6, 7, 8, 9, 10, 15, 20, 25, ..., 95
	uint32_t scale = 5;
	while (scale < 100)
	{
		scales.insert(std::make_pair(pointSizeAt8dpmm * scale / 10.0, scale));
		if (scale < 10)
			scale++;
		else
			scale += 5;
	}
}

StripedHorse::TPCLBitmapFont::TPCLBitmapFont()
  : pointSizeAt8dpmm(0.0), valid(false)
{
}

StripedHorse::TPCLBitmapFontParameters::TPCLBitmapFontParameters()
  : horizontalMagnification(10), verticalMagnification(10)
{
}

StripedHorse::TPCLBitmapFontParameters::TPCLBitmapFontParameters(double dpmm, double fontSize, const std::string &fontDescription, bool allowSubstitution)
  : horizontalMagnification(10), verticalMagnification(10)
{
	std::string genericName, specificName, flags;
	bool isBold, isItalic;

	std::string finalFontName;

	try
	{
		std::tie(genericName, specificName, flags) = FontUtilities::ParseFontDescription(fontDescription);
		std::tie(isBold, isItalic) = FontUtilities::ParseFontDescriptionFlags(flags);
		
		// no bitmap font supports both bold and italic, but italic takes priority
		// because it is probably less noticeable that the bold is missing
		// compared to the other way around
		if (isBold && isItalic)
			isBold = false;
		
		// Look for specific substituted name first
		finalFontName = FontUtilities::MapFontName(specificName, g_specificFontMapping, genericName, g_genericFontMapping, g_bitmapFontTable, allowSubstitution);
	}
	catch (StripedHorse::UnimplementedException &)
	{
		return;
	}
	
	std::vector<TPCLBitmapFont> availableFonts;
	std::copy_if(g_bitmapFontTable.begin(), g_bitmapFontTable.end(), std::back_inserter(availableFonts), [=](const TPCLBitmapFont &bf) {
		return (bf.name == finalFontName) && ((isItalic && bf.style == "italic") || (isBold && bf.style == "bold") || (!isItalic && !isBold && (bf.style == "normal")));
	});

	TPCLScaledBitmapFont scaledFont = findBestFontAndScale(availableFonts, fontSize, dpmm);

	font = scaledFont.bitmapFont;
	horizontalMagnification = scaledFont.scale;
	verticalMagnification = scaledFont.scale;
}
