#include <cmath>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <memory>
#include <regex>
#include <string>
#include <typeinfo>

#include <lemonxps/file.h>
#include <libcore/logging.h>
#include <libtecfilter/argumentparser.h>
#include <libtecfilter/files.h>
#include <libtecfilter/managedfile.h>
#include <libtecfilter/options.h>
#include <libtecfilter/options-utilities.h>
#include <libtecfilter/version.h>
#include "tpclcontext.h"

#ifndef _WIN32
#include <unistd.h>
#else
#include <windows.h>
#define unlink _unlink
#endif

#include "libtecfilter/runtimevalidator.inc.cpp"

namespace
{
	// parses a dimensions in the form wwwwxhhhh (with optional units) into 10ths of millimetres
	std::tuple<double, double> parseDimensions(const std::string &dimension)
	{
		std::regex dimenRegex { "([0-9]+(\\.[0-9]+)?)x([0-9]+(\\.[0-9]+)?)(pt|cm|mm|in)?" };
		std::smatch result;
		if (std::regex_search(dimension, result, dimenRegex))
		{
			double width = std::stod(result.str(1));
			double height = std::stod(result.str(3));
			std::string unit = result.length(5) > 0 ? result.str(5) : "pt"; // default unit is points

			if (unit == "pt")
				return std::make_tuple(width * 254.0 / 72.0, height * 254.0 / 72.0);
			else if (unit == "cm")
				return std::make_tuple(width * 100.0, height * 100.0);
			else if (unit == "mm")
				return std::make_tuple(width * 10.0, height * 10.0);
			else if (unit == "in")
				return std::make_tuple(width * 254.0, height * 254.0);
		}

		throw StripedHorse::InvalidArgumentException("Not a valid size");
	}

	// ensures certain options are set in a format the the TPCLContext is happy with
	// this ensures that Resolution and Resolution.dpmm have valid values (which is
	// the most important) but also sets LabelWidth and LabelHeight if a valid
	// PageSize is set in the PPD
	void fixUpOptions(StripedHorse::Options &options, const StripedHorse::PPDFile &ppd)
	{
		std::string resolution = options.getString("Resolution", "305");
		size_t end = 0;
		double dpi = std::stod(resolution, &end);
		if (end != resolution.size())
			resolution.erase(end);
		if (resolution.size() == 0)
			resolution = "305";

		options.set("Resolution", resolution);
		options.set("Resolution.dpmm", std::to_string(dpi / 25.4));
		
		if (options.hasOption("PaperDimension"))
		{
			std::string targetPaperDimension = options.getString("PaperDimension");

			if (options.isCustom("PaperDimension"))
			{
				double width;
				double height;

				try
				{
					std::tie(width, height) = parseDimensions(targetPaperDimension);
					options.set("LabelWidth", std::to_string(width));
					options.set("LabelHeight", std::to_string(height));
				}
				catch (StripedHorse::InvalidArgumentException &)
				{
					// with these removed, the TPCLContext will derive the size from the input document
					options.remove("PaperDimension");
					options.remove("LabelWidth");
					options.remove("LabelHeight");
					StripedHorse::Logging::Warning() << "Unable to parse page size" << std::endl;
				}
			}
			else
			{
				std::vector<StripedHorse::PPDStatement> paperDimensions = ppd.getAllStatementsWithMainKeyword("PaperDimension");
				for (const auto &pd : paperDimensions)
				{
					if (pd.getOptionKeyword() != targetPaperDimension)
						continue;
					
					std::istringstream dimensions(static_cast<std::string>(pd.getValue()));
					double width, height;
					if (!(dimensions >> width >> height))
						std::cerr << "Warning: Unable to obtain page dimensions" << std::endl;
					
					options.set("LabelWidth", std::to_string(254.0 * width / 72.0));
					options.set("LabelHeight", std::to_string(254.0 * height / 72.0));
					break;
				}
			}
		}
	}

	std::unique_ptr<std::ofstream> g_logFile;
}

int main(int argc, char *argv[])
{
	int filenameArg = argc - 1;
	StripedHorse::ManagedFile infile { "-" };
	std::string outfilename = "-";
	size_t startPage = 0;
	size_t endPage = 0;
	
	bool debugMode = false;
	bool cupsInvocation = false;
	char *softwareEnv = ::getenv("SOFTWARE");
	StripedHorse::PPDFile ppd;
	if (softwareEnv != nullptr)
		cupsInvocation = strstr(softwareEnv, "CUPS/") == softwareEnv;
	
	if (::getenv("STRIPEDHORSE_DEBUG") != nullptr)
		debugMode = true;

	if (cupsInvocation)
		StripedHorse::Logging::SetMessagePrefixes("DEBUG", "INFO", "WARN", "ERROR", "CRIT");

	StripedHorse::Options options;

	char *ppdFilename = ::getenv("PPD");
	if (ppdFilename != nullptr)
	{
		try
		{
			std::ifstream ppdFile(ppdFilename);
			ppd.load(ppdFile);
			if (!StripedHorse::OptionsUtilities::ParseDefaultValuesFromPPD(options, ppd))
				StripedHorse::Logging::Error() << "PPD file was specified but no values were parsed" << std::endl;
			
			if (!StripedHorse::OptionsUtilities::ParseMinAndMaxValuesFromPPD(options, ppd))
				StripedHorse::Logging::Error() << "PPD file was specified but no minimum or maximum values were parsed" << std::endl;
		}
		catch (StripedHorse::Exception &ex)
		{
			StripedHorse::Logging::Warning() << "Unable to obtain default values from PPD (" << ex.what() << ")" << std::endl;
		}
	}

	if (cupsInvocation)
	{
		// argv[1] = job ID
		// argv[2] = username
		// argv[3] = job title
		// argv[4] = number of copies
		// argv[5] = options
		// argv[6] = filename (nullptr = read from stdin)

		if (argc < 5)
		{
			std::cerr << "Invalid invocation:" << std::endl;
			std::cerr << argv[0] << " job user title num-copies options [ filename ] " << std::endl;
			::exit(EXIT_FAILURE);
		}

		options.set("JobID", argv[1]);
		options.set("User", argv[2]);
		options.set("Title", argv[3]);
		options.set("Copies", argv[4]);
		options.parse(argv[5]);

		filenameArg = 6;
		
		StripedHorse::Logging::Debug() << "Parsed options: " << options << std::endl;

		debugMode = debugMode || options.getInt32("StripedHorseDebug") > 0;
	}
	else
	{
		if (argc < 2)
		{
			std::cerr << "This utility is not intended to be run manually" << std::endl;
			std::cerr << argv[0] << " job-id username title num-copies options [ filename ]" << std::endl;
			std::cerr << argv[0] << " --version" << std::endl;
			return 0;
		}
		else if (argc >= 2 && strcmp(argv[1], "--version") == 0)
		{
			std::cerr << StripedHorse::g_version << std::endl;
			return 0;
		}
		
		StripedHorse::ArgumentParser ap(argc, argv);
		
		ap.parse([&](char c){
			switch (c)
			{
				case 'c':
				{
					options.set("DestinationEncoding", ap.getArgumentValue());
					break;
				}

				case 'd':
				{
					const double epsilon = 0.0001;
					double dpmm = std::stod(ap.getArgumentValue());
					
					// There are only 3 known DPIs / DPMMs for TEC BCPs, so if they are "close enough"
					// to these then the values are hard-coded to the exact values, otherwise they are
					// calculated instead if the resolutions are not known.
					if (std::abs(dpmm - 8.0) < epsilon)
						options.set("Resolution", "203");
					else if (std::abs(dpmm - 11.8) < epsilon)
						options.set("Resolution", "300");
					else if (std::abs(dpmm - 12.0) < epsilon)
						options.set("Resolution", "305");
					else
						options.set("Resolution", std::to_string(dpmm * 25.4));
					
					break;
				}

				case 'e':
				{
					endPage = std::stoul(ap.getArgumentValue());
					break;
				}

				case 'H':
				{
					options.set("LabelHeight", ap.getArgumentValue());
					break;
				}

				case 'l':
				{
					g_logFile = std::make_unique<std::ofstream>(ap.getArgumentValue(), std::ios_base::out | std::ios_base::app);
					break;
				}

				case 'L':
				{
					g_logFile = std::make_unique<std::ofstream>(ap.getArgumentValue(), std::ios_base::out | std::ios_base::trunc);
					break;
				}

				case 'o':
				{
					outfilename = ap.getArgumentValue();
					break;
				}

				case 'O':
				{
					options.parse(ap.getArgumentValue());
					break;
				}

				case 's':
				{
					startPage = std::stoul(ap.getArgumentValue());
					break;
				}

				case 'W':
				{
					options.set("LabelWidth", ap.getArgumentValue());
					break;
				}

				default:
					return false;
			}

			return true;
		});

		filenameArg = ap.getCurrentPosition();

		if (g_logFile)
			StripedHorse::Logging::SetOutputStream(g_logFile.get());
	}

	if (argv[filenameArg] != nullptr)
	{
		infile.setFilename(argv[filenameArg]);
	}
	else
	{
		StripedHorse::ManagedFile::SetManagedFileToStandardInput(infile, "xpstotpcl");
		if (debugMode)
			infile.setDeleteWhenDone(false);
	}

	fixUpOptions(options, ppd);

	try
	{
		std::ifstream xpsFilename { infile.getFilename(), std::ios_base::binary | std::ios_base::out };
		LemonXPS::FilePtr xpsFile = std::make_shared<LemonXPS::File>(xpsFilename);
		
		std::ostream *output = &std::cout;
		std::ofstream outputFilestream;
		if (outfilename == "-")
		{
			StripedHorse::Files::SetStandardOutputToBinaryMode();
		}
		else
		{
			outputFilestream.open(outfilename, std::ios_base::binary | std::ios_base::out);
			output = &outputFilestream;
		}

		size_t pageIndex = 0;
		for (const LemonXPS::DocumentSequence &seq : xpsFile->getDocumentSequences())
		{
			for (LemonXPS::DocumentPtr doc : seq)
			{
				StripedHorse::TPCLContext ctx(*output, options);

				for (LemonXPS::PageCPtr page : doc->getPages())
				{
					pageIndex++;
					if ((startPage == 0 || pageIndex >= startPage) && (endPage == 0 || pageIndex <= endPage))
						page->render(ctx);
				}
			}
		}
	}
	catch (StripedHorse::Exception &ex)
	{
		StripedHorse::Logging::Critical() << "TPCL conversion failed due to an exception: " << ex.what() << std::endl;
		::exit(EXIT_FAILURE);
	}
	catch (LemonSqueezy::Exception &ex)
	{
		StripedHorse::Logging::Critical() << "TPCL conversion failed due to a compression exception: " << ex.what() << std::endl;
		::exit(EXIT_FAILURE);
	}
	catch (std::exception &ex)
	{
		StripedHorse::Logging::Critical() << "TPCL conversion failed due to a runtime exception: " << ex.what() << std::endl;
		::exit(EXIT_FAILURE);
	}
}
