#include <iostream>
#include <vector>
#include <libcore/exceptions.h>
#include <libcore/logging.h>
#include "stringapproximator.h"

StripedHorse::StringApproximator::StringApproximator()
  : m_passthrough(true), m_converter(iconv_t(-1))
{
}

StripedHorse::StringApproximator::StringApproximator(const std::string &destination)
  : m_passthrough(true), m_converter(iconv_t(-1))
{
	setDestinationEncoding(destination);
}

StripedHorse::StringApproximator::~StringApproximator()
{
	close();
}

std::string StripedHorse::StringApproximator::approximate(const std::string &str)
{
	if (m_passthrough || str.size() == 0)
		return str;

	std::vector<std::string::value_type> inbuffer(str.begin(), str.end());
	std::vector<std::string::value_type> outbuffer(str.size() * 4 / 3);

	size_t result = 0;
	size_t inBytesLeft = inbuffer.size();
	size_t outBytesLeft = outbuffer.size();

	iconv(m_converter, nullptr, nullptr, nullptr, nullptr);
#ifndef _WIN32
	char *inp = &inbuffer.at(0);
#else
	const char *inp = &inbuffer.at(0);
#endif
	char *outp = &outbuffer.at(0);

	while (true)
	{
		result = iconv(m_converter, &inp, &inBytesLeft, &outp, &outBytesLeft);
		if (inBytesLeft == 0)
			break;
		
		if (result == static_cast<size_t>(-1))
		{
			if (errno == E2BIG)
			{
				size_t oldSize = outbuffer.size();
				outbuffer.resize(outbuffer.size() * 4 / 3);
				outBytesLeft += outbuffer.size() - oldSize;
			}
			else
			{
				Logging::Critical() << "String conversion failed catastrophically: " << errno << std::endl;
				break;
			}
		}
		else if (result > 0)
		{
			Logging::Warning() << "String conversion resulted in character loss" << std::endl;
		}
	}

	// replace all tabs with spaces (since TPCL doesn't have tabstops)
	outbuffer.resize(outbuffer.size() - outBytesLeft);
	for (char &ch : outbuffer)
		if (ch == '\t')
			ch = ' ';

	return std::string { outbuffer.begin(), outbuffer.end() };
}

void StripedHorse::StringApproximator::close()
{
	if (m_converter != (iconv_t)-1)
		iconv_close(m_converter);
}

void StripedHorse::StringApproximator::setDestinationEncoding(const std::string &destination)
{
	close();

	if (destination == "UTF-8")
	{
		m_passthrough = true;
		return;
	}

	m_passthrough = false;
	m_converter = iconv_open((destination + "//TRANSLIT").c_str(), "UTF-8");
	if (m_converter == (iconv_t)-1)
		throw InternalInconsistencyException("Unable to create converter between UTF-8 and " + destination);
}
