#include <regex>
#include <sstream>
#include <libcore/endian.h>
#include <libcore/exceptions.h>
#include <libcore/logging.h>
#include <libtecfilter/commands.h>
#include <libtecfilter/options-utilities.h>
#include <lemongfx/bezier.h>
#include <lemongfx/draw.h>
#include <lemongfx/stroke.h>
#include <lemonxps/document.h>
#include <lemonxps/page.h>
#include <lemonxps/pathparser.h>
#include <iconv.h>
#include "bitmapfont.h"
#include "tpclcontext.h"
#include "tobj-code93.h"
#include "tobj-pdf417.h"
#include "tobj-qrcode.h"
#include "tobj-rfid.h"
#include "vectorfont.h"

namespace
{
	class TPCLPathParserContext : public LemonXPS::IPathParserContext
	{
	public:
		TPCLPathParserContext(StripedHorse::TPCLContext &ctx)
		  : m_context(ctx)
		{
		}

		void closePath() override
		{
			LemonGFX::Path::Subpath &topPath = m_context.getCurrentPath().back();
			StripedHorse::PointF p1 = topPath.front();
			if (p1 != topPath.back())
				topPath.emplace_back(p1);
		}

		void curveTo(double cx1, double cy1, double cx2, double cy2, double x3, double y3) override
		{
			const StripedHorse::TransformMatrix &ctm = m_context.topMatrix();
			LemonGFX::Path::Subpath topPath = m_context.getCurrentPath().back();

			StripedHorse::PointF p1 = topPath.back();
			StripedHorse::PointF cp1 = ctm * StripedHorse::PointF { cx1, cy1 };
			StripedHorse::PointF cp2 = ctm * StripedHorse::PointF { cx2, cy2 };
			StripedHorse::PointF p2 = ctm * StripedHorse::PointF { x3, y3 };
			
			StripedHorse::PointF lastPoint = p1;
			for (const StripedHorse::PointF &p : LemonGFX::Bezier::ApproximateCurve(p1, cp1, cp2, p2))
			{
				if (lastPoint == p)
					continue;
				
				topPath.emplace_back(p);
				lastPoint = p;
			}
		}

		void lineTo(double x, double y) override
		{
			m_context.getCurrentPath().back().emplace_back(m_context.topMatrix() * StripedHorse::PointF { x, y });
		}

		void moveTo(double x, double y) override
		{
			m_context.getCurrentPath().addSubpath(LemonGFX::Path::Subpath {m_context.topMatrix() * StripedHorse::PointF { x, y } });
		}

	private:
		StripedHorse::TPCLContext &m_context;
	};

	StripedHorse::TransformMatrix transformMatrixFromString(const std::string &string)
	{
		std::istringstream ms { string };
		std::string element;
		double elements[6];
		size_t i = 0;

		while (std::getline(ms, element, ',') && i < 6)
			elements[i++] = std::stod(element);

		if (i != 6)
			throw StripedHorse::TPCLContextException("Invalid matrix representation");

		return StripedHorse::TransformMatrix {
			elements[0],
			elements[1],
			elements[2],
			elements[3],
			elements[4],
			elements[5]
		};
	}

	std::vector<StripedHorse::PointF> pointsFromString(const std::string &string)
	{
		std::vector<StripedHorse::PointF> points;
		std::istringstream stream { string };

		while (stream)
		{
			double f1, f2;
			char ch;
			if ((stream >> f1 >> ch >> f2) && ch == ',')
				points.emplace_back(f1, f2);
			else if (stream.fail() && !stream.eof())
				throw StripedHorse::InvalidArgumentException("Not a valid point string");
		}

		return points;
	}

	std::vector<StripedHorse::RectangleF> rectanglesFromString(const std::string &string)
	{
		std::vector<StripedHorse::RectangleF> rectangles;
		std::istringstream stream { string };

		while (stream)
		{
			double x, y, w, h;
			char delim[3];
			if ((stream >> x >> delim[0] >> y >> delim[1] >> w >> delim[2] >> h) && delim[0] == ',' && delim[1] == ',' && delim[2] == ',')
				rectangles.emplace_back(x, y, x + w, y + h);
			else if (stream.fail() && !stream.eof())
				throw StripedHorse::InvalidArgumentException("Not a valid rectangle string");
		}

		return rectangles;
	}

	double convertXPSUnitsToTPCLUnits(double input)
	{
		// XPS unit = 96 dpi
		// TPCL unit = 10 dpmm
		return 254.0 * input / 96.0;
	}

	double detectScale(const StripedHorse::TransformMatrix &m)
	{
		StripedHorse::PointF p1 = m * StripedHorse::PointF { 0, 0 };
		StripedHorse::PointF p2 = m * StripedHorse::PointF { 1, 1 };

		return (std::abs(p2.x - p1.x) + std::abs(p2.y - p1.y)) / 2.0;
	}

	double getLuminance(double r, double g, double b)
	{
		return (0.299 * r) + (0.587 * g) + (0.114 * b);
	}

	std::tuple<uint32_t, uint32_t, uint32_t> parseColor(const std::string &color)
	{
		std::regex colorRegex { "^#([[:xdigit:]]{2})([[:xdigit:]]{2})([[:xdigit:]]{2})$" };
		std::smatch result;
		if (!std::regex_match(color, result, colorRegex))
			throw StripedHorse::InvalidArgumentException("Unknown color value: " + color);
		
		uint32_t red = std::stoul(result[1].str(), nullptr, 16);
		uint32_t green = std::stoul(result[2].str(), nullptr, 16);
		uint32_t blue = std::stoul(result[3].str(), nullptr, 16);
		return std::make_tuple(red, green, blue);
	}

	std::string getUserName()
	{
		const char *username = ::getenv("USER");
		if (username != nullptr)
			return username;
		else
			return "Unknown user";
	}

	LemonGFX::Image loadImageFromZipArchiveEntry(const LemonSqueezy::ZipArchiveEntry &zae)
	{
		LemonGFX::Image image;
		const std::vector<unsigned char> &buffer = zae.decompress();

		std::stringstream ss;
		ss.write(reinterpret_cast<const char *>(&buffer[0]), buffer.size());
		image.read(ss);
		return image;
	}

	void parseSpacingsAttribute(const std::string &unicodeString, const std::string &spacing, std::function<void(double, double, const std::string &)> callback)
	{
		std::string currentSpacing;
		std::istringstream spacingStream(spacing);
		size_t totalGlyphs = 0;

		while (std::getline(spacingStream, currentSpacing, ';'))
		{
			std::istringstream currentSpacingStream(currentSpacing);

			size_t glyphCount;
			double runningTally;
			double spacing;
			
			if (!(currentSpacingStream >> glyphCount >> runningTally >> spacing))
				break;
			
			callback(runningTally, spacing, unicodeString.substr(totalGlyphs, glyphCount));
			totalGlyphs += glyphCount;
		}
	}

	bool isWhitespace(const std::string &str)
	{
		static const std::regex whitespace { R"(^\s*$)" };
		return std::regex_match(str, whitespace);
	}

	std::string formatJobName(const std::string &jobName)
	{
		std::string result;
		result.reserve(jobName.size());
		for (const char &ch : jobName)
		{
			if (::isalnum(ch))
				result.push_back(ch);
			else
				result.push_back('_');
		}

		return result;
	}

	uint32_t getTPCLRotation(enum StripedHorse::Rotation r)
	{
		uint32_t result;
		switch (r)
		{
			case StripedHorse::Rotation::kRotate0:
				result = 0U;
				break;
			
			case StripedHorse::Rotation::kRotate90:
				result = 11U;
				break;
			
			case StripedHorse::Rotation::kRotate180:
				result = 22U;
				break;
			
			case StripedHorse::Rotation::kRotate270:
				result = 33U;
				break;

			case StripedHorse::Rotation::kRotateOther:
			default:
				throw StripedHorse::TPCLContextException("Invalid rotation for TPCL context");
		}

		return result;
	}
}

StripedHorse::TPCLContext::TPCLContext(std::ostream &out, const Options &options)
  : m_writer(out), m_options(options), m_bitmapCompressionMode(BitmapCompressionMode::kAuto), m_dpmm(12.0)
{
	m_matrixStack.emplace_back();

	m_beginHandlers.emplace("Canvas", &TPCLContext::beginCanvas);
	m_beginHandlers.emplace("DetectedShape", &TPCLContext::beginDetectedShape);
	m_beginHandlers.emplace("FixedPage", &TPCLContext::beginFixedPage);
	m_beginHandlers.emplace("Glyphs", &TPCLContext::beginGlyphs);
	m_beginHandlers.emplace("ImageBrush", &TPCLContext::beginImageBrush);
	m_beginHandlers.emplace("Path", &TPCLContext::beginPath);
	m_beginHandlers.emplace("Path.Data", &TPCLContext::beginPath_Data);
	m_beginHandlers.emplace("Path.Fill", &TPCLContext::beginPath_Fill);
	m_beginHandlers.emplace("PathFigure", &TPCLContext::beginPathFigure);
	m_beginHandlers.emplace("PathGeometry", &TPCLContext::beginPathGeometry);
	m_beginHandlers.emplace("PolyBezierSegment", &TPCLContext::beginPolyBezierSegment);
	m_beginHandlers.emplace("PolyLineSegment", &TPCLContext::beginPolyLineSegment);
	m_endHandlers.emplace("Canvas", &TPCLContext::endCanvas);
	m_endHandlers.emplace("FixedPage", &TPCLContext::endFixedPage);
	m_endHandlers.emplace("Path", &TPCLContext::endPath);
	m_endHandlers.emplace("PathFigure", &TPCLContext::endPathFigure);

	m_stringApproximator.setDestinationEncoding(options.getString("DestinationEncoding", "CP850"));

	if (!m_options.hasOption("Resolution") || !m_options.hasOption("Resolution.dpmm"))
		throw TPCLContextException("No resolution is specified, unable to print the job");

	std::string jobName = formatJobName(m_options.getString("Title", "Untitled"));
	std::string user = m_options.getString("User", getUserName());

	if (m_options.getString("CommandCharacters", "Readable") == "Unreadable")
		m_writer.setOutputStyle({ '\x1B' }, { '\n', '\x00' }, false);

	m_writer.writeCommandStart("AAA;");
	m_writer.writeString(jobName);
	m_writer.writeCommandEnd();

	std::string compressionMode = options.getString("Compression", "Auto");
	if (compressionMode == "None")
		m_bitmapCompressionMode = BitmapCompressionMode::kNone;
	else if (compressionMode == "Auto")
		m_bitmapCompressionMode = BitmapCompressionMode::kAuto;
	else if (compressionMode == "TOPIX")
		m_bitmapCompressionMode = BitmapCompressionMode::kTOPIX;

	m_dpmm = m_options.getDouble("Resolution.dpmm", 12.0);

	m_rasterAllLines = !m_options.getBoolean("UseNativeLineCommand", false);
}

StripedHorse::TPCLContext::~TPCLContext()
{
}

void StripedHorse::TPCLContext::beginPageElement(LemonXPS::PageElementCPtr element)
{
	std::string type = element->getType();
	HandlerMap::const_iterator i = m_beginHandlers.find(type);
	if (i != m_beginHandlers.end())
		(this->*(i->second))(element);
	else
		Logging::Warning() << "No begin-handler for " << type << std::endl;
}

void StripedHorse::TPCLContext::endPageElement(LemonXPS::PageElementCPtr element)
{
	std::string type = element->getType();
	HandlerMap::const_iterator i = m_endHandlers.find(type);
	if (i != m_endHandlers.end())
		(this->*(i->second))(element);
}

void StripedHorse::TPCLContext::flushPageBitmap()
{
	RectangleU croppedDimensions = m_pageBitmap.getAutoCroppedDimensions();
	if (croppedDimensions.getWidth() > 0 && croppedDimensions.getHeight() > 0)
	{
		LemonGFX::Bitmap cropped = m_pageBitmap.cut(croppedDimensions);

		// writeBitmap expects the position in the 10dpmm but the coordinates of the cropped image are in target dpmm
		PointF position = static_cast<PointF>(cropped.getPosition());
		position.x = position.x * 10 / m_dpmm;
		position.y = position.y * 10 / m_dpmm;
		cropped.setPosition(static_cast<PointU>(position));
		if (cropped.isValidSize())
			writeBitmap(cropped);
	}

	m_pageBitmap.clear();
}

LemonGFX::Path &StripedHorse::TPCLContext::getCurrentPath()
{
	return m_currentPath;
}

const LemonGFX::Path &StripedHorse::TPCLContext::getCurrentPath() const
{
	return m_currentPath;
}

StripedHorse::TransformMatrix &StripedHorse::TPCLContext::topMatrix()
{
	return m_matrixStack.back();
}

void StripedHorse::TPCLContext::writeBitmap(const LemonGFX::Bitmap &bmp)
{
	PointU32 pos = static_cast<PointU32>(bmp.getPosition());
	uint32_t width = static_cast<uint32_t>(bmp.getWidth());
	uint32_t height = static_cast<uint32_t>(bmp.getHeight());

	std::stringstream bmpData;

	m_writer.writeCommandStart("SG;");

	// If mode is "Auto" then that means to try TOPIX compression anyway
	// and if it is smaller (it typically will be) then it is used instead
	// of uncompressed mode.
	BitmapCompressionMode mode = m_bitmapCompressionMode;
	switch (mode)
	{
		case BitmapCompressionMode::kAuto:
		case BitmapCompressionMode::kTOPIX:
			try
			{
				bmp.compress(LemonGFX::Bitmap::CompressionFormat::kTOPIX, bmpData);
				if (bmpData.tellp() < 0)
					throw TPCLContextException("Unable to compress data");
				
				// Branch if the compressed data size is less than the bitmap's uncompressed size
				// or if the user has explicitly asked for TOPIX compression.
				if (static_cast<size_t>(bmpData.tellp()) < bmp.getDataSize() || mode == BitmapCompressionMode::kTOPIX)
				{
					// scaling for 203/300/305 dpi models is 1x if the scaling argument is 300
					uint32_t scaling = 300;
					if (m_options.getDouble("Resolution", 305.0) > 305.0)
						scaling = 600;
					
					m_writer.writeFormat("aaaa,bbbb,cccc,dddd,e,", pos.x, pos.y, width, scaling, '3');
					break;
				}
			}
			catch (LemonGFX::Exception &ex)
			{
				Logging::Warning() << "An error occurred while compressing graphics data: " << ex.what() << " (will attempt to output uncompressed data)" << std::endl;
			}

			bmpData.str("");
			// fallthrough
		case BitmapCompressionMode::kNone:
			bmp.compress(LemonGFX::Bitmap::CompressionFormat::kUncompressed, bmpData);
			m_writer.writeFormat("aaaa,bbbb,cccc,dddd,e,", pos.x, pos.y, width, height, '1');
			break;

		default:
			throw InternalInconsistencyException("Bitmap compression mode is invalid");
	}

	m_writer.getStream() << bmpData.rdbuf();
	m_writer.writeCommandEnd();
}


// Handlers
void StripedHorse::TPCLContext::beginCanvas(LemonXPS::PageElementCPtr canvas)
{
	std::string renderTransform;
	if (!canvas->tryGetAttribute("RenderTransform", renderTransform))
		renderTransform = "1,0,0,1,0,0";
	
	m_matrixStack.emplace_back(topMatrix() * transformMatrixFromString(renderTransform));
}

void StripedHorse::TPCLContext::beginDetectedShape(LemonXPS::PageElementCPtr detectedShape)
{
	std::string data = detectedShape->getAttribute("Data");
	std::string color = detectedShape->getAttribute("Color");
	double originX = std::stod(detectedShape->getAttribute("OriginX"));
	double originY = std::stod(detectedShape->getAttribute("OriginY"));
	double width = std::stod(detectedShape->getAttribute("Width"));
	double height = std::stod(detectedShape->getAttribute("Height"));
	double angle = std::stod(detectedShape->getAttribute("Angle"));

	RectangleF rect { originX, originY + height, originX + width, originY };
	RectangleF finalRect = topMatrix() * rect;
	if (finalRect.lowerLeft.x > finalRect.upperRight.x)
		std::swap(finalRect.lowerLeft.x, finalRect.upperRight.x);
	if (finalRect.lowerLeft.y > finalRect.upperRight.y)
		std::swap(finalRect.lowerLeft.y, finalRect.upperRight.y);
	
	TemplateObjectPtr templateObject;
	if (color == "Red")
		templateObject = std::make_shared<QRCodeTemplateObject>(finalRect, data);
	else if (color == "Green")
		templateObject = std::make_shared<Code93TemplateObject>(finalRect, data);
	else if (color == "Blue")
		templateObject = std::make_shared<PDF417TemplateObject>(finalRect, data);
	
	if (templateObject)
	{
		enum Rotation r = SnapAngleToFixedRotation(angle);
		if (r == Rotation::kRotateOther)
			Logging::Error() << "Unable to snap angle: " << angle << std::endl;
		
		templateObject->setRotation(r);
		templateObject->write(m_writer, m_dpmm);
	}
	else
	{
		Logging::Warning() << "Detected shape ignored!" << std::endl;
	}
}

void StripedHorse::TPCLContext::beginFixedPage(LemonXPS::PageElementCPtr fixedPage)
{
	double pageWidth = convertXPSUnitsToTPCLUnits(std::stod(fixedPage->getAttribute("Width")));
	double pageHeight = convertXPSUnitsToTPCLUnits(std::stod(fixedPage->getAttribute("Height")));

	double labelWidth = m_options.getDouble("LabelWidth", pageWidth);
	double labelHeight = m_options.getDouble("LabelHeight", pageHeight);
	
	// reset the default matrix
	m_matrixStack = std::vector<TransformMatrix> { { } };

	enum class Orientation
	{
		kLandscape,
		kPortrait,
	};
	
	Orientation sourceOrientation = pageWidth > pageHeight ? Orientation::kLandscape : Orientation::kPortrait;
	Orientation destinationOrientation = Orientation::kLandscape;
	
	if (m_options.getString("LabelOrientation", "Portrait") == "Portrait")
	{
		destinationOrientation = Orientation::kPortrait;
		if (labelWidth > labelHeight)
			std::swap(labelWidth, labelHeight);
	}
	
	if (sourceOrientation != destinationOrientation)
	{
		topMatrix().rotate(M_PI_2);
		topMatrix().translate(0.0, -labelWidth);
		std::swap(pageWidth, pageHeight);
	}

	double widthRatio = labelWidth / pageWidth;
	double heightRatio = labelHeight / pageHeight;
	double ratio = std::min(widthRatio, heightRatio);
	double xOffset = (labelWidth - ratio * pageWidth) / 2.0;
	double yOffset = (labelHeight - ratio * pageHeight) / 2.0;
	
	if (sourceOrientation != destinationOrientation)
		std::swap(xOffset, yOffset);
	
	topMatrix().translate(xOffset, yOffset);
	topMatrix().scale(ratio, ratio);

	double defaultScale = convertXPSUnitsToTPCLUnits(1.0);
	topMatrix().scale(defaultScale, defaultScale);

	int32_t labelGap = OptionsUtilities::GetValueIn10dpmm(m_options, "LabelGap", { 20, 200 });
	
	m_writer.writeCommand("Daaaa,bbbb,cccc", static_cast<uint32_t>(labelHeight) + labelGap, static_cast<uint32_t>(labelWidth), static_cast<uint32_t>(labelHeight));

	// only output these if at least one of them is non-zero (otherwise it is just a no-op)
	m_writer.writeAdjustmentsFromOptions(m_options);

	m_writer.writeCommand("C");

	// Any custom RFID user memory data?
	if (m_options.isCustom("RFIDUserMemData"))
	{
		// make the rectangle invalid so that it doesn't interfere with other template objects
		m_templateObjects.emplace_back(std::make_shared<RFIDUserMemoryDataTemplateObject>(RectangleF {-1, -1, -1, -1}, m_options.getString("RFIDUserMemData")));
	}
	
	// Any custom RFID EPC memory data?
	if (m_options.isCustom("RFIDEPCMemData"))
	{
		// make the rectangle invalid so that it doesn't interfere with other template objects
		m_templateObjects.emplace_back(std::make_shared<RFIDEPCMemoryDataTemplateObject>(RectangleF {-1, -1, -1, -1}, m_options.getString("RFIDEPCMemData")));
	}

	if (m_pageBitmap.isValidSize())
	{
		m_pageBitmap.clear();
	}
	else
	{
		m_pageBitmap = LemonGFX::Bitmap(labelWidth * m_dpmm / 10, labelHeight * m_dpmm / 10);

		LemonGFX::HalftoningCurve hc = LemonGFX::HalftoningCurveFromName(m_options.getString("ImageDarkness", "Normal"));
		m_pageBitmap.setHalftoningCurve(hc);
	}
}

void StripedHorse::TPCLContext::beginGlyphs(LemonXPS::PageElementCPtr glyphs)
{
	std::string unicodeString;
	if (!glyphs->tryGetAttribute("UnicodeString", unicodeString))
		Logging::Error() << "There is no text to print" << std::endl;
	
	// no point doing anything if the string is blank
	if (isWhitespace(unicodeString))
		return;

	std::string renderTransform;
	if (!glyphs->tryGetAttribute("RenderTransform", renderTransform))
		renderTransform = "1,0,0,1,0,0";
	
	TransformMatrix finalMatrix = topMatrix();
	finalMatrix.concat(transformMatrixFromString(renderTransform));
	PointF p = { std::stod(glyphs->getAttribute("OriginX")), std::stod(glyphs->getAttribute("OriginY")) };
	finalMatrix.translate(p.x, p.y);
	finalMatrix.m[1][1] *= -1.0;

	enum Rotation rotation = StripedHorse::DetectFixedRotation(finalMatrix);
	if (rotation == Rotation::kRotateOther)
	{
		StripedHorse::Logging::Error() << "Cannot arbitrarily rotate text (0/90/180/270 degrees only). Snapping to nearest right angle."<< std::endl;
		rotation = StripedHorse::DetectFixedRotation(finalMatrix, M_PI_2);
	}
	
	uint32_t finalRotation = getTPCLRotation(rotation);
	PointF point = finalMatrix * PointF { 0.0, 0.0 };
	PointU32 finalPoint = static_cast<PointU32>(point);

	flushPageBitmap();

	double emSize = std::stod(glyphs->getAttribute("FontRenderingEmSize"));
	double detectedScale = detectScale(finalMatrix);

	std::function<void(const PointU32 &, const std::string &)> textWriter = [](const PointU32 &p, const std::string &str){
		(void) p;
		(void) str;
		// do nothing as default action
	};

	std::string characterSpacingAttribute;
	if (!glyphs->tryGetAttribute("CharacterSpacing", characterSpacingAttribute))
		characterSpacingAttribute = "0.0";
	
	double characterSpacing = std::stod(characterSpacingAttribute);
	int32_t deviceSpacing = std::floor(m_dpmm * characterSpacing / (72.0 / 25.4));
	if (deviceSpacing > 99)
	{
		Logging::Warning() << "Device spacing for characters clamped to 99 dots" << std::endl;
		deviceSpacing = 99;
	}
	else if (deviceSpacing < -99)
	{
		Logging::Warning() << "Device spacing for characters clamped to -99 dots" << std::endl;
		deviceSpacing = -99;
	}

	try
	{
		double fontSize = detectedScale * emSize;
		std::string fontDescription = glyphs->getAttribute("FontUri");
		if (!m_options.hasOption("SubstituteFonts") || m_options.getString("SubstituteFonts") == "SubstituteForBitmapFonts")
		{
			TPCLBitmapFontParameters bfp(m_dpmm, fontSize / (72.0 / 25.4), fontDescription, true);
			deviceSpacing += bfp.font.extraDeviceSpacing;
			textWriter = [bfp,finalRotation,deviceSpacing,this](const PointU32 &p, const std::string &str) {
				if (deviceSpacing)
					m_writer.writeCommand("PCaaa;bbbb,cccc,dd,ee,f,+gg,ii,j=p/255", 0U, p.x, p.y, bfp.horizontalMagnification, bfp.verticalMagnification, bfp.font.code, deviceSpacing, finalRotation, 'B', str);
				else
					m_writer.writeCommand("PCaaa;bbbb,cccc,dd,ee,f,ii,j=p/255", 0U, p.x, p.y, bfp.horizontalMagnification, bfp.verticalMagnification, bfp.font.code, finalRotation, 'B', str);
			};
		}
		else
		{
			bool allowSubstitutions = m_options.getString("SubstituteFonts") == "SubstituteForOutlineFonts";
			TPCLVectorFontParameters vfp(m_dpmm, fontSize, fontDescription, allowSubstitutions);
			textWriter = [vfp,finalRotation,deviceSpacing,this](const PointU32 &p, const std::string &str) {
				if (vfp.font.type == TPCLVectorFont::Type::kTrueType)
					m_writer.writeCommand("PVaa;bbbb,cccc,dddd,eeee,ff,g,+iii,jj,k=p/255", 0U, p.x, p.y, vfp.characterWidth, vfp.characterHeight, vfp.font.code, 0U, deviceSpacing, finalRotation, 'B', str);
				else
					m_writer.writeCommand("PVaa;bbbb,cccc,dddd,eeee,f,+hhh,ii,j=p/255", 0U, p.x, p.y, vfp.characterWidth, vfp.characterHeight, vfp.font.code, deviceSpacing, finalRotation, 'B', str);
			};
		}
	}
	catch (StripedHorse::Exception &ex)
	{
		Logging::Error() << "The font in the document cannot be printed (" << ex.what() << ")" << std::endl;
	}

	double totalSpaceMoved = 0.0;
	std::string stringToOutput;
	std::string spacings;
	if (glyphs->tryGetAttribute("Spacings", spacings))
	{
		parseSpacingsAttribute(unicodeString, spacings, [&](double runningTally, double spacing, const std::string &str){
			std::string approximated = m_stringApproximator.approximate(str);
			totalSpaceMoved -= spacing;

			if (spacing <= -250)
			{
				textWriter(finalPoint, stringToOutput);
				finalPoint = static_cast<PointU32>(finalMatrix * PointF { totalSpaceMoved / 1000.0, 0.0 });
				stringToOutput = approximated;
			}
			else
			{
				// between -250 and -50 is typically a space. It's heuristic, but
				// it's pretty reliable (with the tests we have)
				if (spacing < -50)
					stringToOutput.push_back(' ');
				stringToOutput.append(approximated);
			}

			totalSpaceMoved += runningTally;
		});
	}
	else
	{
		stringToOutput = m_stringApproximator.approximate(unicodeString);
	}
	
	if (stringToOutput.size() > 0)
		textWriter(finalPoint, stringToOutput);
}

void StripedHorse::TPCLContext::beginImageBrush(LemonXPS::PageElementCPtr imageBrush)
{
	LemonXPS::PagePtr page = imageBrush->getPage().lock();
	if (!page)
	{
		Logging::Warning() << "Unable to get page object from page element" << std::endl;
		return;
	}
	
	LemonXPS::DocumentPtr document = page->getDocument().lock();
	if (!document)
	{
		Logging::Warning() << "Unable to get document object from page object" << std::endl;
		return;
	}

	std::string imageSource;
	if (!imageBrush->tryGetAttribute("ImageSource", imageSource))
	{
		Logging::Warning() << "Image source attribute is missing or invalid" << std::endl;
		return;
	}

	const LemonSqueezy::ZipArchive &zip = document->getArchive();
	LemonSqueezy::ZipArchive::const_iterator imageFile = zip.find(imageSource);

	if (imageFile == zip.end())
	{
		Logging::Warning() << "Image file (" << imageSource << ") could not be found in XPS archive" << std::endl;
		return;
	}

	std::string maskSource;
	LemonSqueezy::ZipArchive::const_iterator maskFile = zip.end();
	if (imageBrush->tryGetAttribute("MaskSource", maskSource))
	{
		maskFile = zip.find(maskSource);
		if (maskFile == zip.end())
		{
			Logging::Warning() << "Mask file (" << maskSource << ") could not be found in XPS archive" << std::endl;
			return;
		}
		else
		{
			Logging::Debug() << "Mask file is being used for this image" << std::endl;
		}
	}

	LemonGFX::Image image;
	LemonGFX::Image mask;

	try
	{
		image = loadImageFromZipArchiveEntry(*imageFile);
		if (maskFile != zip.end())
			mask = loadImageFromZipArchiveEntry(*maskFile);
	}
	catch (LemonGFX::ImageException &ex)
	{
		Logging::Error() << "The image file could not be used" << std::endl;
		return;
	}

	RectangleF sourceRect;
	RectangleF destinationRect;
	std::string viewBox;
	std::string viewPort;
	if (!(imageBrush->tryGetAttribute("Viewbox", viewBox) && imageBrush->tryGetAttribute("Viewport", viewPort)))
	{
		Logging::Warning() << "Image brush missing required attributes" << std::endl;
		return;
	}

	sourceRect = rectanglesFromString(viewBox).at(0);
	destinationRect = rectanglesFromString(viewPort).at(0);

	if (sourceRect.lowerLeft.x != 0 || sourceRect.lowerLeft.y != 0 || sourceRect.upperRight.x != image.getWidth() || sourceRect.upperRight.y != image.getHeight())
		throw StripedHorse::UnimplementedException("Cannot yet handle image subsetting");

	TransformMatrix rotater;
	double angle = DetectAngle(topMatrix());
	enum Rotation rotation = SnapAngleToFixedRotation(angle);

	switch (rotation)
	{
		case Rotation::kRotate0:
		case Rotation::kRotate180:
			// do nothing
			break;
		
		case Rotation::kRotate90:
			destinationRect = RectangleF(0, 0, 1, 1);
			break;

		case Rotation::kRotate270:
			destinationRect = RectangleF(1, 0, 0, 1);
			std::swap(destinationRect.lowerLeft.y, destinationRect.upperRight.y);
			break;

		case Rotation::kRotateOther:
			rotation = SnapAngleToFixedRotation(angle, M_PI_2);
			StripedHorse::Logging::Error() << "Cannot arbitrarily rotate images (0/90/180/270 degrees only). Snapping to nearest right angle" << std::endl;
			break;
	}

	RectangleF finalDestination = topMatrix() * destinationRect;
	double height = finalDestination.getHeight();
	if (height < 0)
	{
		std::swap(finalDestination.lowerLeft.y, finalDestination.upperRight.y);
		finalDestination.lowerLeft.y += height;
		finalDestination.upperRight.y += height;
	}

	double width = finalDestination.getWidth();
	if (width < 0)
	{
		std::swap(finalDestination.lowerLeft.x, finalDestination.upperRight.x);
		finalDestination.lowerLeft.x += width;
		finalDestination.upperRight.x += width;
	}

	LemonGFX::HalftoningMethod hm = LemonGFX::HalftoningMethod::kThresholding;
	if (m_options.getString("DocumentType") == "Graphics")
		hm = LemonGFX::HalftoningMethod::kStripedHorse9Toner;

	LemonGFX::HalftoningCurve hc = m_pageBitmap.getHalftoningCurve();
	m_imageBrush = std::make_unique<LemonGFX::Bitmap>(LemonGFX::Bitmap::Downsample(image, static_cast<RectangleU>(finalDestination), m_dpmm, static_cast<LemonGFX::BitmapRotation>(rotation), hm, hc));
	if (mask.isValidSize())
	{
		// halftoning method should not apply to this bitmap because it should already be a 1-bit bilevel bitmap
		LemonGFX::Bitmap maskBitmap = LemonGFX::Bitmap::Downsample(mask, static_cast<RectangleU>(finalDestination), m_dpmm);
		m_imageBrush->applyMask(maskBitmap);
	}
}

void StripedHorse::TPCLContext::beginPath(LemonXPS::PageElementCPtr path)
{
	if (m_currentPathElement != nullptr)
		throw UnimplementedException("Nested paths are not supported");

	m_currentPathElement = path;

	std::string pathData;
	if (path->tryGetAttribute("Data", pathData))
	{
		std::istringstream input { pathData };
		LemonXPS::PathParser pathParser(input);
		TPCLPathParserContext pathParserContext(*this);
		pathParser.render(pathParserContext);
	}
}

void StripedHorse::TPCLContext::beginPath_Data(LemonXPS::PageElementCPtr path_data)
{
	(void) path_data;
	// can ignore because it contains only PathGeometry elements
}

void StripedHorse::TPCLContext::beginPath_Fill(LemonXPS::PageElementCPtr path_fill)
{
	(void) path_fill;
	// can ignore because it only contains Brush elements
}

void StripedHorse::TPCLContext::beginPathFigure(LemonXPS::PageElementCPtr pathFigure)
{
	std::string startPoint;
	if (pathFigure->tryGetAttribute("StartPoint", startPoint))
	{
		std::vector<PointF> p = pointsFromString(startPoint);
		if (p.size() != 1)
			throw InvalidFormatException("StartPoint attribute is invalid");
		m_currentPath.emplaceSubpath(LemonGFX::Path::Subpath { topMatrix() * p.front() });
	}
	else
	{
		throw InvalidFormatException("Missing StartPoint attribute");
	}
}

void StripedHorse::TPCLContext::beginPathGeometry(LemonXPS::PageElementCPtr pathGeometry)
{
	if (!pathGeometry->tryGetAttribute("FillRule", m_currentFillRule))
		m_currentFillRule = "EvenOdd";
}

void StripedHorse::TPCLContext::beginPolyBezierSegment(LemonXPS::PageElementCPtr polyBezierSegment)
{
	std::vector<PointF> nextPoints = pointsFromString(polyBezierSegment->getAttribute("Points"));
	if (nextPoints.size() % 3 != 0)
		throw InvalidArgumentException("PolyBezierSegment contains an unusual number of points");
	
	if (nextPoints.size() == 0)
		Logging::Error() << "No points in line segment" << std::endl;

	LemonGFX::Path::Subpath &topPath = m_currentPath.back();
	for (LemonGFX::Path::Subpath::size_type i = 0; i < nextPoints.size(); i += 3)
	{
		PointF p1 = topPath.back();
		PointF cp1 = topMatrix() * nextPoints.at(i + 0);
		PointF cp2 = topMatrix() * nextPoints.at(i + 1);
		PointF p2 = topMatrix() * nextPoints.at(i + 2);
		std::vector<PointF> curvePoints = LemonGFX::Bezier::ApproximateCurve(p1, cp1, cp2, p2);
		
		PointF lastPoint = p1;
		for (const PointF &p : curvePoints)
		{
			if (lastPoint == p)
				continue;
			
			topPath.emplace_back(p);
			lastPoint = p;
		}
	}
}

void StripedHorse::TPCLContext::beginPolyLineSegment(LemonXPS::PageElementCPtr polyLineSegment)
{
	std::vector<PointF> nextPoints = pointsFromString(polyLineSegment->getAttribute("Points"));
	if (nextPoints.size() == 0)
		Logging::Error() << "No points in line segment" << std::endl;

	PointF lastPoint = m_currentPath.back().back();
	for (const PointF &p : nextPoints)
	{
		PointF finalPoint = topMatrix() * p;
		if (finalPoint == lastPoint)
			continue;
		
		m_currentPath.back().emplace_back(finalPoint);
		lastPoint = finalPoint;
	}
}

void StripedHorse::TPCLContext::endCanvas(LemonXPS::PageElementCPtr canvas)
{
	(void) canvas;
	m_matrixStack.pop_back();
}

void StripedHorse::TPCLContext::endFixedPage(LemonXPS::PageElementCPtr fixedPage)
{
	(void) fixedPage;

	flushPageBitmap();

	for (TemplateObjectPtr templateObject : m_templateObjects)
	{
		if (templateObject->getData().size() == 0)
			continue;
		
		templateObject->write(m_writer, m_dpmm);
	}

	m_templateObjects.clear();

	uint32_t numberOfCopies = m_options.getUInt32("Copies", 1);
	uint32_t cutInterval = m_options.getUInt32("CutInterval", 0);

	std::string sensorString = m_options.getString("Sensor");
	uint32_t sensor = 0;
	if (sensorString == "Reflective")
		sensor = 1;
	else if (sensorString == "Transmissive")
		sensor = 2;
	

	char issueModeChar = m_options.getString("StripLabel") == "True" ? 'D' : 'C';
	
	// although it is hexadecimal for now, it is only because all supported
	// print speeds are less than 16ips, but it's possible that future models
	// may support print speeds of 16ips or more, and it's more than likely
	// that to maintain backwards compatibility that 16 = G, 17 = H ...
	uint32_t printSpeed = m_options.getUInt32("PrintSpeed", 3);
	uint32_t printMethod;

	if (m_options.getString("PrintMethod") == "DirectThermalHeadUp")
		printMethod = 3;
	else if (m_options.getString("PrintMethod") == "ThermalTransfer")
		printMethod = 2;
	else if (m_options.getString("PrintMethod") == "ThermalTransferRibbonSaving")
		printMethod = 1;
	else 
		printMethod = 0;
	
	// printing directions
	//  0 - bottom first (only supported mode for now)
	//  1 - top first
	//  2 - bottom first mirror printing
	//  3 - top first mirror printing
	uint32_t printDirection = 0;

	// status response
	//  0 - don't send status response (only supported mode for now)
	//  1 - send status response
	uint32_t statusResponse = 0;

	m_writer.writeCommand("XS;I,aaaa,bbbcde$fgh", numberOfCopies, cutInterval, sensor, issueModeChar, printSpeed, printMethod, printDirection, statusResponse);
}

void StripedHorse::TPCLContext::endPath(LemonXPS::PageElementCPtr path)
{
	(void) path;

	if (m_imageBrush)
	{
		if (m_imageBrush->isValidSize())
		{
			// the page bitmap is in the target dpmm but the coordinates of the image brush are in 10dpmm
			PointF mover = static_cast<PointF>(m_imageBrush->getPosition());
			mover.x = mover.x * m_dpmm / 10;
			mover.y = mover.y * m_dpmm / 10;
			m_imageBrush->setPosition(static_cast<PointU>(mover));
			m_pageBitmap.combine(*m_imageBrush);
		}

		m_currentPath = LemonGFX::Path();
		m_currentPathElement = nullptr;
		m_imageBrush = nullptr;
		return;
	}

	double finalThickness = 0;
	std::string thicknessAttribute;

	if (m_currentPathElement->tryGetAttribute("StrokeThickness", thicknessAttribute))
	{
		// since we don't support horizontal/vertical scaling, average the scale
		finalThickness = std::stod(thicknessAttribute) * detectScale(topMatrix());
		m_currentPath.setStrokeWidth(finalThickness);
	}

	std::string fillColorString;
	if (m_currentPathElement->tryGetAttribute("Fill", fillColorString))
	{
		uint32_t red, green, blue;
		std::tie(red, green, blue) = parseColor(fillColorString);
		m_currentPath.setFillShade(1.0 - getLuminance(red / 255.0, green / 255.0, blue / 255.0));
		m_currentPath.setFillType(m_currentFillRule == "NonZero" ? LemonGFX::Fill::FillType::kNonzeroWindingFill : LemonGFX::Fill::FillType::kEvenOddFill);
	}

	if (m_currentPath.getStrokeWidth() > 0.0)
	{
		std::string strokeColorString;
		if (!m_currentPathElement->tryGetAttribute("Stroke", strokeColorString))
			strokeColorString = "#000000";
		
		uint32_t red, green, blue;
		std::tie(red, green, blue) = parseColor(strokeColorString);
		
		m_currentPath.setStrokeShade(1.0 - getLuminance(red / 255.0, green / 255.0, blue / 255.0));
	}

	// if the path is filled, or the stroke width is large, or the stroke shade is not black,
	// then the path should be rastered
	if (m_currentPath.getSize() > 0 && (m_currentPath.isFilled() || m_rasterAllLines || m_currentPath.getStrokeWidth() >= 5.0 || m_currentPath.getStrokeShade() != 1.0))
	{
		LemonGFX::Draw::DrawPathsOnBitmap(m_pageBitmap, std::vector<LemonGFX::Path> { m_currentPath }, m_dpmm);
	}
	else if (finalThickness >= 0.0 && finalThickness < 5.0)
	{
		uint32_t thicknessArgument = std::max(uint32_t(1), static_cast<uint32_t>(finalThickness));
		for (const LemonGFX::Path::Subpath &subpath : m_currentPath)
		{
			PointF lastPoint = subpath.at(0);
			for (const PointF &p : subpath)
			{
				if (lastPoint == p)
					continue;
				
				m_writer.writeCommand("LC;aaaa,bbbb,cccc,dddd,e,f", static_cast<uint32_t>(lastPoint.x), static_cast<uint32_t>(lastPoint.y), static_cast<uint32_t>(p.x), static_cast<uint32_t>(p.y), 0U, thicknessArgument);
				lastPoint = p;
			}
		}
	}

	m_currentPath = LemonGFX::Path();
	m_currentPathElement = nullptr;
}

void StripedHorse::TPCLContext::endPathFigure(LemonXPS::PageElementCPtr pathFigure)
{
	if (pathFigure->getAttribute("IsClosed") == "true")
	{
		LemonGFX::Path::Subpath &topSubpath = m_currentPath.back();
		if (topSubpath.front() != topSubpath.back())
			topSubpath.emplace_back(topSubpath.front());
	}

	m_currentPath.setIsFilled(pathFigure->getAttribute("IsFilled") == "true");
}
